Package: TreeTools
Title: Create, Modify and Analyse Phylogenetic Trees
Version: 2.4.0
Authors@R: c(
  person("Martin R.", 'Smith', role = c("aut", "cre", "cph"), 
         email = "martin.smith@durham.ac.uk",
         comment = c(ORCID = "0000-0001-5660-1727")),
  person(given = "Emmanuel", family = "Paradis", role = c("cph"),
         comment = c(ORCID = "0000-0003-3092-2199", "ape library")),
  person(given = "Robert", family = "Noble", role = c("cph"),
         comment = c(ORCID = "0000-0002-8057-4252", "RUtreebalance"))
  )
License: GPL (>= 3)
Copyright: Incorporates C/C++ code from 'ape' by Emmanuel Paradis
        <doi:10.1093/bioinformatics/bty633> and 'RUtreebalance' by
        Robert Noble <doi:10.5281/zenodo.5873857>.
Description: Efficient implementations of functions for the creation, 
  modification and analysis of phylogenetic trees.
  Applications include:
  generation of trees with specified shapes;
  tree rearrangement;
  analysis of tree shape;
  rooting of trees and extraction of subtrees;
  calculation and depiction of split support;
  plotting the position of rogue taxa (Klopfstein & Spasojevic 2019)
  <doi:10.1371/journal.pone.0212942>;
  calculation of ancestor-descendant relationships,
  of 'stemwardness' (Asher & Smith, 2022) <doi:10.1093/sysbio/syab072>,
  and of tree balance (Mir et al. 2013, Lemant et al. 2022)
  <doi:10.1016/j.mbs.2012.10.005>, <doi:10.1093/sysbio/syac027>;
  artificial extinction (Asher & Smith, 2022) <doi:10.1093/sysbio/syab072>;
  import and export of trees from Newick, Nexus (Maddison et al. 1997)
  <doi:10.1093/sysbio/46.4.590>,
  and TNT <https://www.lillo.org.ar/phylogeny/tnt/> formats;
  and analysis of splits and cladistic information.
URL: https://ms609.github.io/TreeTools/,
        https://github.com/ms609/TreeTools/
BugReports: https://github.com/ms609/TreeTools/issues/
SystemRequirements: C++17
Depends: R (>= 3.6.0), ape (>= 5.6),
Imports: bit64, fastmatch (>= 1.1.3), methods, PlotTools, Rdpack (>=
        2.6.6),
Suggests: RCurl, spelling, knitr, phangorn (>= 2.2.1), Rcpp (>= 1.0.8),
        rmarkdown, testthat (>= 3.0), TreeDist, TreeSearch, vdiffr (>=
        1.0.0),
Config/Needs/benchmark: bench, TreeDist
Config/Needs/check: rcmdcheck, testthat
Config/Needs/coverage: covr
Config/Needs/memcheck: pkgdown, testthat
Config/Needs/metadata: codemeta
Config/Needs/revdeps: revdepcheck
Config/Needs/website: pkgdown
Config/roxygen2/version: 8.0.0
Config/testthat/parallel: false
Config/testthat/edition: 3
LinkingTo: Rcpp
RdMacros: Rdpack
LazyData: true
ByteCompile: true
Encoding: UTF-8
Language: en-GB
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2026-06-02 15:02:39 UTC; pjjg18
Author: Martin R. Smith [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-5660-1727>),
  Emmanuel Paradis [cph] (ORCID: <https://orcid.org/0000-0003-3092-2199>,
    ape library),
  Robert Noble [cph] (ORCID: <https://orcid.org/0000-0002-8057-4252>,
    RUtreebalance)
Maintainer: Martin R. Smith <martin.smith@durham.ac.uk>
Repository: CRAN
Date/Publication: 2026-06-02 16:50:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-06-02 16:55:25 UTC; unix
Archs: TreeTools.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("!",Splits)
S3method("&",Splits)
S3method("+",Splits)
S3method("-",Splits)
S3method("<",MixedBase)
S3method("<",phylo)
S3method("==",MixedBase)
S3method("==",phylo)
S3method(">",MixedBase)
S3method(">",phylo)
S3method("[",phyDat)
S3method("[[",Splits)
S3method("names<-",Splits)
S3method("|",Splits)
S3method(.Edge,list)
S3method(.Edge,numeric)
S3method(.Edge,phylo)
S3method(.UpdateNodeLabel,numeric)
S3method(.UpdateNodeLabel,phylo)
S3method(AllTipLabels,"NULL")
S3method(AllTipLabels,Splits)
S3method(AllTipLabels,TreeNumber)
S3method(AllTipLabels,list)
S3method(AllTipLabels,matrix)
S3method(AllTipLabels,multiPhylo)
S3method(AllTipLabels,phylo)
S3method(ApePostorder,"NULL")
S3method(ApePostorder,list)
S3method(ApePostorder,multiPhylo)
S3method(ApePostorder,phylo)
S3method(ArtificialExtinction,matrix)
S3method(ArtificialExtinction,phyDat)
S3method(Cherries,numeric)
S3method(Cherries,phylo)
S3method(Cladewise,"NULL")
S3method(Cladewise,list)
S3method(Cladewise,matrix)
S3method(Cladewise,multiPhylo)
S3method(Cladewise,phylo)
S3method(CladisticInfo,Splits)
S3method(CladisticInfo,list)
S3method(CladisticInfo,multiPhylo)
S3method(CladisticInfo,phylo)
S3method(CollapseNode,phylo)
S3method(ConsensusWithout,"NULL")
S3method(ConsensusWithout,list)
S3method(ConsensusWithout,multiPhylo)
S3method(ConsensusWithout,phylo)
S3method(DropTip,"NULL")
S3method(DropTip,Splits)
S3method(DropTip,list)
S3method(DropTip,multiPhylo)
S3method(DropTip,phylo)
S3method(EdgeRatio,phylo)
S3method(KeepTip,"NULL")
S3method(KeepTip,Splits)
S3method(KeepTip,list)
S3method(KeepTip,multiPhylo)
S3method(KeepTip,phylo)
S3method(KeptPaths,data.frame)
S3method(KeptPaths,matrix)
S3method(KeptVerts,numeric)
S3method(KeptVerts,phylo)
S3method(LongBranch,"NULL")
S3method(LongBranch,list)
S3method(LongBranch,multiPhylo)
S3method(LongBranch,phylo)
S3method(MakeTreeBinary,"NULL")
S3method(MakeTreeBinary,list)
S3method(MakeTreeBinary,multiPhylo)
S3method(MakeTreeBinary,phylo)
S3method(MinimumSpanningTree,"NULL")
S3method(MinimumSpanningTree,dist)
S3method(MinimumSpanningTree,matrix)
S3method(NSplits,"NULL")
S3method(NSplits,ClusterTable)
S3method(NSplits,Splits)
S3method(NSplits,character)
S3method(NSplits,list)
S3method(NSplits,multiPhylo)
S3method(NSplits,numeric)
S3method(NSplits,phylo)
S3method(NTip,Splits)
S3method(NTip,default)
S3method(NTip,list)
S3method(NTip,matrix)
S3method(NTip,multiPhylo)
S3method(NTip,phyDat)
S3method(NTip,phylo)
S3method(NodeDepth,"NULL")
S3method(NodeDepth,list)
S3method(NodeDepth,matrix)
S3method(NodeDepth,multiPhylo)
S3method(NodeDepth,phylo)
S3method(NodeOrder,"NULL")
S3method(NodeOrder,list)
S3method(NodeOrder,matrix)
S3method(NodeOrder,multiPhylo)
S3method(NodeOrder,phylo)
S3method(Postorder,"NULL")
S3method(Postorder,list)
S3method(Postorder,multiPhylo)
S3method(Postorder,numeric)
S3method(Postorder,phylo)
S3method(PostorderOrder,"NULL")
S3method(PostorderOrder,numeric)
S3method(PostorderOrder,phylo)
S3method(Preorder,"NULL")
S3method(Preorder,list)
S3method(Preorder,multiPhylo)
S3method(Preorder,numeric)
S3method(Preorder,phylo)
S3method(Pruningwise,"NULL")
S3method(Pruningwise,list)
S3method(Pruningwise,multiPhylo)
S3method(Pruningwise,phylo)
S3method(RenumberTips,"NULL")
S3method(RenumberTips,Splits)
S3method(RenumberTips,list)
S3method(RenumberTips,multiPhylo)
S3method(RenumberTips,phylo)
S3method(Reweight,matrix)
S3method(Reweight,phyDat)
S3method(RootNode,"NULL")
S3method(RootNode,list)
S3method(RootNode,multiPhylo)
S3method(RootNode,numeric)
S3method(RootNode,phylo)
S3method(RootNodeDistance,"NULL")
S3method(RootNodeDistance,character)
S3method(RootNodeDistance,numeric)
S3method(RootOnNode,"NULL")
S3method(RootOnNode,list)
S3method(RootOnNode,multiPhylo)
S3method(RootOnNode,phylo)
S3method(RootTree,"NULL")
S3method(RootTree,list)
S3method(RootTree,matrix)
S3method(RootTree,multiPhylo)
S3method(RootTree,phylo)
S3method(RootedTreeWithShape,integer64)
S3method(RootedTreeWithShape,numeric)
S3method(SisterSize,"NULL")
S3method(SisterSize,character)
S3method(SisterSize,numeric)
S3method(SortTree,"NULL")
S3method(SortTree,list)
S3method(SortTree,multiPhylo)
S3method(SortTree,phylo)
S3method(SplitImbalance,"NULL")
S3method(SplitImbalance,Splits)
S3method(SplitImbalance,phylo)
S3method(SplitInformation,Splits)
S3method(SplitInformation,numeric)
S3method(SplitInformation,phylo)
S3method(SplitsInBinaryTree,"NULL")
S3method(SplitsInBinaryTree,Splits)
S3method(SplitsInBinaryTree,default)
S3method(SplitsInBinaryTree,list)
S3method(SplitsInBinaryTree,multiPhylo)
S3method(SplitsInBinaryTree,numeric)
S3method(SplitsInBinaryTree,phylo)
S3method(TCIContext,numeric)
S3method(TCIContext,phylo)
S3method(TipLabels,MixedBase)
S3method(TipLabels,Splits)
S3method(TipLabels,TreeNumber)
S3method(TipLabels,character)
S3method(TipLabels,default)
S3method(TipLabels,list)
S3method(TipLabels,logical)
S3method(TipLabels,matrix)
S3method(TipLabels,multiPhylo)
S3method(TipLabels,numeric)
S3method(TipLabels,phyDat)
S3method(TipLabels,phylo)
S3method(TipsInSplits,Splits)
S3method(TipsInSplits,phylo)
S3method(TntOrder,"NULL")
S3method(TntOrder,list)
S3method(TntOrder,multiPhylo)
S3method(TntOrder,numeric)
S3method(TntOrder,phylo)
S3method(TopologyOnly,default)
S3method(TopologyOnly,list)
S3method(TopologyOnly,multiPhylo)
S3method(TopologyOnly,phylo)
S3method(TotalCopheneticIndex,"NULL")
S3method(TotalCopheneticIndex,list)
S3method(TotalCopheneticIndex,multiPhylo)
S3method(TotalCopheneticIndex,phylo)
S3method(Treeness,"NULL")
S3method(Treeness,list)
S3method(Treeness,multiPhylo)
S3method(Treeness,phylo)
S3method(UnrootTree,"NULL")
S3method(UnrootTree,list)
S3method(UnrootTree,multiPhylo)
S3method(UnrootTree,phylo)
S3method(WriteTntCharacters,matrix)
S3method(WriteTntCharacters,phyDat)
S3method(as.ClusterTable,"NULL")
S3method(as.ClusterTable,list)
S3method(as.ClusterTable,multiPhylo)
S3method(as.ClusterTable,phylo)
S3method(as.MixedBase,"NULL")
S3method(as.MixedBase,MixedBase)
S3method(as.MixedBase,TreeNumber)
S3method(as.MixedBase,integer64)
S3method(as.MixedBase,multiPhylo)
S3method(as.MixedBase,numeric)
S3method(as.MixedBase,phylo)
S3method(as.Newick,list)
S3method(as.Newick,multiPhylo)
S3method(as.Newick,phylo)
S3method(as.Splits,Splits)
S3method(as.Splits,character)
S3method(as.Splits,integer)
S3method(as.Splits,list)
S3method(as.Splits,logical)
S3method(as.Splits,matrix)
S3method(as.Splits,multiPhylo)
S3method(as.Splits,numeric)
S3method(as.Splits,phylo)
S3method(as.TreeNumber,"NULL")
S3method(as.TreeNumber,MixedBase)
S3method(as.TreeNumber,TreeNumber)
S3method(as.TreeNumber,character)
S3method(as.TreeNumber,multiPhylo)
S3method(as.TreeNumber,phylo)
S3method(as.character,Splits)
S3method(as.integer64,TreeNumber)
S3method(as.logical,Splits)
S3method(as.logical,phylo)
S3method(as.matrix,ClusterTable)
S3method(as.matrix,phyDat)
S3method(as.matrix,phylo)
S3method(as.multiPhylo,Splits)
S3method(as.multiPhylo,list)
S3method(as.multiPhylo,phyDat)
S3method(as.multiPhylo,phylo)
S3method(as.phylo,MixedBase)
S3method(as.phylo,Splits)
S3method(as.phylo,TreeNumber)
S3method(as.phylo,integer64)
S3method(as.phylo,numeric)
S3method(c,Splits)
S3method(duplicated,Splits)
S3method(head,Splits)
S3method(length,Splits)
S3method(names,Splits)
S3method(print,ClusterTable)
S3method(print,Splits)
S3method(print,TreeNumber)
S3method(rev,Splits)
S3method(sort,Splits)
S3method(sort,multiPhylo)
S3method(summary,ClusterTable)
S3method(summary,Splits)
S3method(t,Splits)
S3method(tail,Splits)
S3method(unique,Splits)
S3method(xtfrm,Splits)
export(.CompatibleRaws)
export(.CompatibleSplit)
export(.RandomParent)
export(.UnrootedKeys)
export(AddTip)
export(AddTipEverywhere)
export(AddUnconstrained)
export(AllAncestors)
export(AllTipLabels)
export(AncestorEdge)
export(ApePostorder)
export(ApeTime)
export(ArtEx)
export(ArtificialExtinction)
export(BalancedTree)
export(CharacterInformation)
export(Cherries)
export(CladeSizes)
export(Cladewise)
export(CladisticInfo)
export(CladisticInformation)
export(CollapseEdge)
export(CollapseNode)
export(CompatibleSplits)
export(Consensus)
export(ConsensusWithout)
export(ConstrainedNJ)
export(Decompose)
export(DescendantEdges)
export(DescendantTips)
export(DoubleFactorial)
export(DoubleFactorial64)
export(DropTip)
export(DropTipPhylo)
export(EdgeAncestry)
export(EdgeDistances)
export(EdgeRatio)
export(EndSentence)
export(ExtractTaxa)
export(FirstMatchingSplit)
export(Hamming)
export(IC1Spr)
export(ImposeConstraint)
export(J1Index)
export(JQIndex)
export(KeepTip)
export(KeepTipPostorder)
export(KeepTipPreorder)
export(KeptPaths)
export(KeptVerts)
export(LabelSplits)
export(LeafLabelInterchange)
export(ListAncestors)
export(LnDoubleFactorial)
export(LnDoubleFactorial.int)
export(LnRooted)
export(LnRooted.int)
export(LnSplitMatchProbability)
export(LnTreesMatchingSplit)
export(LnTreesMatchingTree)
export(LnUnrooted)
export(LnUnrooted.int)
export(LnUnrootedMult)
export(LnUnrootedSplits)
export(LnUnrootedTreesMatchingSplit)
export(Log2DoubleFactorial)
export(Log2Rooted)
export(Log2Rooted.int)
export(Log2TreesMatchingSplit)
export(Log2TreesMatchingTree)
export(Log2Unrooted)
export(Log2Unrooted.int)
export(Log2UnrootedMult)
export(Log2UnrootedSplits)
export(Log2UnrootedTreesMatchingSplit)
export(LogDoubleFactorial)
export(LogDoubleFactorial.int)
export(LongBranch)
export(MRCA)
export(MSTEdges)
export(MSTLength)
export(MakeTreeBinary)
export(MarkMissing)
export(MatchEdges)
export(MatchNodes)
export(MatchStrings)
export(MatrixToPhyDat)
export(MorphoBankDecode)
export(MrBayesTrees)
export(MultiSplitInformation)
export(N1Spr)
export(NDescendants)
export(NJTree)
export(NPartitionPairs)
export(NPartitions)
export(NRooted)
export(NRooted64)
export(NRootedShapes)
export(NSplits)
export(NTip)
export(NUnrooted)
export(NUnrooted64)
export(NUnrootedMult)
export(NUnrootedShapes)
export(NUnrootedSplits)
export(NewickTree)
export(NeworderPhylo)
export(NeworderPruningwise)
export(NexusTokens)
export(NexusTokensToInteger)
export(NodeDepth)
export(NodeNumbers)
export(NodeOrder)
export(PaintTree)
export(PairwiseDistances)
export(PathLengths)
export(PectinateTree)
export(PhyDat)
export(PhyDatToMatrix)
export(PhyDatToString)
export(PhyToString)
export(PhydatToString)
export(PolarizeSplits)
export(Postorder)
export(PostorderOrder)
export(Preorder)
export(Pruningwise)
export(RandomTree)
export(ReadAsPhyDat)
export(ReadCharacters)
export(ReadMrBayes)
export(ReadMrBayesTrees)
export(ReadNotes)
export(ReadTNTAsPhyDat)
export(ReadTNTCharacters)
export(ReadTntAsPhyDat)
export(ReadTntCharacters)
export(ReadTntTree)
export(Renumber)
export(RenumberEdges)
export(RenumberTips)
export(RenumberTree)
export(Reweight)
export(RightmostCharacter)
export(RoguePlot)
export(RootNode)
export(RootNodeDist)
export(RootNodeDistance)
export(RootOnNode)
export(RootTree)
export(RootedTreeShape)
export(RootedTreeWithShape)
export(SampleOne)
export(SingleTaxonTree)
export(SisterSize)
export(SortTree)
export(SplitConflicts)
export(SplitConsistent)
export(SplitFrequency)
export(SplitImbalance)
export(SplitInformation)
export(SplitMatchProbability)
export(SplitsInBinaryTree)
export(StarTree)
export(Stemminess)
export(StringToPhyDat)
export(Subsplit)
export(Subtree)
export(SupportColor)
export(SupportColour)
export(TCIContext)
export(TNTOrder)
export(TNTText2Tree)
export(TipLabels)
export(TipTimedTree)
export(TipsInSplits)
export(TntOrder)
export(TntText2Tree)
export(TopologyOnly)
export(TotalCopheneticIndex)
export(TreeIsRooted)
export(Treeness)
export(TreesMatchingSplit)
export(TreesMatchingTree)
export(TrivialSplits)
export(Unquote)
export(UnrootTree)
export(UnrootedKeys)
export(UnrootedTreeKey)
export(UnrootedTreeShape)
export(UnrootedTreeWithKey)
export(UnrootedTreeWithShape)
export(UnrootedTreesMatchingSplit)
export(UnshiftTree)
export(WithoutTrivialSplits)
export(WriteTntCharacters)
export(YuleTree)
export(ZeroTaxonTree)
export(as.ClusterTable)
export(as.MixedBase)
export(as.Newick)
export(as.Splits)
export(as.TreeNumber)
export(as.multiPhylo)
export(doubleFactorials)
export(edge_to_splits)
export(is.TreeNumber)
export(logDoubleFactorials)
export(replicate64)
export(root_on_node)
export(sapply64)
export(vapply64)
exportMethods("%in%")
exportMethods(match)
exportMethods(xor)
importFrom(PlotTools,SpectrumLegend)
importFrom(Rdpack,reprompt)
importFrom(ape,all.equal.phylo)
importFrom(ape,as.phylo)
importFrom(ape,cophenetic.phylo)
importFrom(ape,edgelabels)
importFrom(ape,is.rooted)
importFrom(ape,multi2di)
importFrom(ape,nj)
importFrom(ape,read.nexus)
importFrom(ape,read.tree)
importFrom(ape,write.tree)
importFrom(bit64,as.integer64)
importFrom(bit64,integer64)
importFrom(fastmatch,"%fin%")
importFrom(fastmatch,fmatch)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRamp)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,hcl)
importFrom(grDevices,rgb)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,par)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setMethod)
importFrom(methods,setOldClass)
importFrom(stats,as.dist)
importFrom(stats,cophenetic)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(utils,combn)
importFrom(utils,globalVariables)
importFrom(utils,head)
importFrom(utils,tail)
useDynLib(TreeTools, .registration = TRUE)

## ----Contrast matrix----------------------------------------------------------
contrast.matrix <- matrix(data = c(
# 0 1 -  # Each column corresponds to a character-state
  1, 0, 0, # Each row corresponds to a token, here 0, denoting the 
           # character-state set {0} 
  0, 1, 0, # 1 | {1}
  0, 0, 1, # - | {-}
  1, 1, 0, # A | {01}
  1, 1, 0, # + | {01}
  1, 1, 1  # ? | {01-}
), ncol = 3, # ncol corresponds to the number of columns in the matrix
byrow = TRUE)
dimnames(contrast.matrix) <- list(
  c(0, 1, "-", "A", "+", "?"), # A list of the tokens corresponding to each row
                               # in the contrast matrix
  c(0, 1, "-") # A list of the character-states corresponding to the columns 
               # in the contrast matrix
)

contrast.matrix

## ----text-entry, fig.align='center', fig.width=6, out.width='75%'-------------
myTree <- ape::read.tree(text = "((A, B), ((C, D), (E, F)));")
plot(myTree)

## ----double-bracket, fig.align='center', fig.width=6, out.width='75%'---------
badTree <- ape::read.tree(text = "((A, B), (((C, D), ((E), F))));")
plot(badTree)
ape::nodelabels(bg = c(3, 3, 2, 2, 3, 3, 2))

## ----branch-lengths, fig.align='center', fig.width=6, out.width='75%'---------
myTree <- ape::read.tree(
  text = "((A:1, B:1):2, ((C:1, D:1):2, (E:1, F:1):2):4);"
)
plot(myTree)

#' Number of trees matching a bipartition split
#'
#' Calculates the number of unrooted bifurcated trees that are consistent with
#' a bipartition split that divides taxa into groups of size `A` and `B`.
#'
#' @param A,B Integer specifying the number of taxa in each partition.
#'
#' @return `TreesMatchingSplit()` returns a numeric specifying the number of trees
#' that are compatible with the given split.
#'
#' `LnTreesMatchingSplit()` and `Log2TreesMatchingSplit()` give the natural
#' and base-2 logarithms of this number.
#'
#' @examples
#' TreesMatchingSplit(5, 6)
#' LnTreesMatchingSplit(5, 6)
#' Log2TreesMatchingSplit(5, 6)
#'
#' @template MRS
#'
#' @family split information functions
#' @export
TreesMatchingSplit <- function(A, B = A[2]) {
  if (A[1] == 0) NUnrooted(B) else
  if (B == 0) NUnrooted(A[1]) else
  NRooted(A[1]) * NRooted(B)
}

#' @rdname TreesMatchingSplit
#' @export
LnTreesMatchingSplit <- function(A, B = A[2]) {
  if (A[1] == 0) LnUnrooted.int(B) else
  if (B == 0) LnUnrooted.int(A[1]) else
  LnRooted.int(A[1]) + LnRooted.int(B)
}

#' @rdname TreesMatchingSplit
#' @export
Log2TreesMatchingSplit <- function(A, B = A[2]) {
  if (A[1] == 0) Log2Unrooted.int(B) else
  if (B == 0) Log2Unrooted.int(A[1]) else
  Log2Rooted.int(A[1]) + Log2Rooted.int(B)
}

#' Character information content
#'
#' `CharacterInformation()` calculates the cladistic information content
#' \insertCite{Steel2006}{TreeTools} of a given character, in bits.
#' The total information in all characters gives a measure of the potential
#' utility of a dataset \insertCite{Cotton2008}{TreeTools}, which can be
#' compared with a profile parsimony score \insertCite{Faith2001}{TreeTools} to
#' evaluate the degree of homoplasy within a dataset.
#'
#' @param tokens Character vector specifying the tokens assigned to each taxon for
#' a character.  Example: `c(0, 0, 0, 1, 1, 1, "?", "-")`.
#'
#' Note that ambiguous tokens such as `(01)` are not supported, and should be
#' replaced with `?`.
#'
#' @return `CharacterInformation()` returns a numeric specifying the
#' phylogenetic information content of the character
#' (_sensu_ \insertCite{Steel2006;nobrackets}{TreeTools}), in bits.
#'
#' @references \insertAllCited{}
#' @family split information functions
#' @template MRS
#' @importFrom fastmatch %fin%
#' @export
CharacterInformation <- function(tokens) {
  tokenCounts <- table(tokens)
  # Our character splits our taxa into groups with the same token
  # ?s and -s are best ignored
  splits <- tokenCounts[!(names(tokenCounts) %fin% c("?", "-"))]

  # Return:
  MultiSplitInformation(splits)
}

#' Phylogenetic information content of splitting leaves into two partitions
#'
#' Calculate the phylogenetic information content (_sensu_ 
#' \insertCite{Steel2006;nobrackets}{TreeTools}) of a split, which
#' reflects the probability that a uniformly selected random tree will contain
#' the split: a split that is consistent with a smaller number of trees will
#' have a higher information content.
#'
#' `SplitInformation()` addresses bipartition splits, which correspond to
#' edges in an unrooted phylogeny; `MultiSplitInformation()` supports splits
#' that subdivide taxa into multiple partitions, which may correspond to
#' multi-state characters in a phylogenetic matrix.
#'
#' A simple way to characterise trees is to count the number of edges.
#' (Edges are almost, but not quite, equivalent to nodes.)
#' Counting edges (or nodes) provides a quick measure of a tree's resolution,
#' and underpins the Robinson-Foulds tree distance measure.
#' Not all edges, however, are created equal.
#'
#' An edge splits the leaves of a tree into two subdivisions.  The more equal
#' these subdivisions are in size, the more instructive this edge is.
#' Intuitively, the division of mammals from reptiles is a profound revelation
#' that underpins much of zoology; recognizing that two species of bat are more
#' closely related to each other than to any other mammal or reptile is still
#' instructive, but somewhat less fundamental.
#'
#' Formally, the phylogenetic (Shannon) information content of a split _S_,
#' _h(S)_, corresponds to the probability that a uniformly selected random tree
#' will contain the split, _P(S)_: _h(S)_ = -log _P(S)_.
#' Base 2 logarithms are typically employed to yield an information content in
#' bits.
#'
#' As an example, the split `AB|CDEF` occurs in 15 of the 105 six-leaf trees;
#' _h_(`AB|CDEF`) = -log _P_(`AB|CDEF`) = -log(15/105) ~ 2.81 bits.  The split
#' `ABC|DEF` subdivides the leaves more evenly, and is thus more instructive:
#' it occurs in just nine of the 105 six-leaf trees, and
#' _h_(`ABC|DEF`) = -log(9/105) ~ 3.54 bits.
#'
#' As the number of leaves increases, a single even split may contain more
#' information than multiple uneven splits -- see the examples section below.
#'
#' Summing the information content of all splits within a tree, perhaps using
#' the '[TreeDist](https://ms609.github.io/TreeDist/)' function
#' [`SplitwiseInfo()`](https://ms609.github.io/TreeDist/reference/TreeInfo.html),
#' arguably gives a more instructive picture of its resolution than simply
#' counting the number of splits that are present -- though with the caveat
#' that splits within a tree are not independent of one another, so some
#' information may be double counted.  (This same charge applies to simply
#' counting nodes, too.)
#'
#' Alternatives would be to count the number of quartets that are resolved,
#' perhaps using the '[Quartet](https://ms609.github.io/Quartet/)' function
#' [`QuartetStates()`](https://ms609.github.io/Quartet/reference/QuartetState.html),
#' or to use a different take on the information contained within a split, the
#' clustering information: see the 'TreeDist' function
#' [`ClusteringInfo()`](https://ms609.github.io/TreeDist/reference/TreeInfo.html)
#' for details.
#'
#'
#' @inheritParams TreesMatchingSplit
#'
#' @return `SplitInformation()` and `MultiSplitInformation()` return the
#' phylogenetic information content, in bits, of a split that subdivides leaves
#' into partitions of the specified sizes.
#'
#' @examples
#' # Eight leaves can be split evenly:
#' SplitInformation(4, 4)
#'
#' # or unevenly, which is less informative:
#' SplitInformation(2, 6)
#'
#' # A single split that evenly subdivides 50 leaves contains more information
#' # that seven maximally uneven splits on the same leaves:
#' SplitInformation(25, 25)
#' 7 * SplitInformation(2, 48)
#' @references \insertAllCited{}
#'
#' @family split information functions
#'
#' @seealso
#'
#' Sum the phylogenetic information content of splits within a tree:
#'  [`TreeDist::SplitwiseInfo()`](https://ms609.github.io/TreeDist/reference/TreeInfo.html)
#'
#'
#' Sum the clustering information content of splits within a tree:
#'  [`TreeDist::ClusteringInfo()`](https://ms609.github.io/TreeDist/reference/TreeInfo.html)
#'
#'
#'
#' @template MRS
#' @export
SplitInformation <- function(A, B = A[1]) {
  UseMethod("SplitInformation")
}

#' @rdname SplitInformation
#' @export
SplitInformation.numeric <- function(A, B = A[1]) {
  -(Log2TreesMatchingSplit(A, B) - Log2Unrooted.int(A + B))
}

#' @rdname SplitInformation
#' @export
SplitInformation.Splits <- function(A, B) {
  nTip <- NTip(A)
  tis <- TipsInSplits(A)
  Log2Unrooted.int(nTip) - Log2Rooted.int(tis) - Log2Rooted.int(nTip - tis)
}

#' @rdname SplitInformation
#' @export
SplitInformation.phylo <- function(A, B) {
  SplitInformation(as.Splits(A))
}

#' @rdname SplitInformation
#' @param partitionSizes Integer vector specifying the number of taxa in each
#' partition of a multi-partition split.
#' @examples
#' # Three ways to split eight leaves into multiple partitions:
#' MultiSplitInformation(c(2, 2, 4))
#' MultiSplitInformation(c(2, 3, 3))
#' MultiSplitInformation(rep(2, 4))
#'
#'
#' @export
MultiSplitInformation <- function(partitionSizes) {
  Log2Unrooted.int(sum(partitionSizes)) - Log2UnrootedMult(partitionSizes)
}

#' Number of trees consistent with split
#'
#' Calculates the number of unrooted bifurcating trees consistent with the
#' specified multi-partition split, using theorem two of
#' \insertCite{Carter1990;textual}{TreeTools}.
#'
#' @param \dots A series or vector of integers listing the number of tips in
#' each of a number of tree splits (e.g. bipartitions).
#' For example, `3, 5` states that a character divides a set of eight tips into
#' a group of three and a group of five.
#TODO Support Splits objects, using a Method.
#'
#' @return `UnrootedTreesMatchingSplit()` returns an integer specifying the
#'  number of unrooted bifurcating trees consistent with the specified split.
#'
#'
#' @references \insertAllCited{}
#' 
#' @examples
#' UnrootedTreesMatchingSplit(c(3, 5))
#' UnrootedTreesMatchingSplit(3, 2, 1, 2)
#' @template MRS
#' @family split information functions
#' @export
UnrootedTreesMatchingSplit <- function(...) {
  # use exp and log as it's just as fast, but less likely to overflow to Inf
  exp(LnUnrootedTreesMatchingSplit(...))
}

.LogUTMS <- function(LogXDoubleFactorial, splits) {

  splits <- splits[splits > 0L]
  totalTips <- sum(splits)
  tipsMinusLengthSplits <- totalTips - length(splits)

  sum(LogXDoubleFactorial(totalTips + totalTips - 5L),
      LogXDoubleFactorial(splits + splits - 3L)) -
      LogXDoubleFactorial(tipsMinusLengthSplits + tipsMinusLengthSplits - 1L)
}

#' @rdname UnrootedTreesMatchingSplit
#' @export
LnUnrootedTreesMatchingSplit <- function(...) {
  .LogUTMS(LnDoubleFactorial, c(...))
}

#' @rdname UnrootedTreesMatchingSplit
#' @export
Log2UnrootedTreesMatchingSplit <- function(...) {
  .LogUTMS(Log2DoubleFactorial, c(...))
}

test_that("Trees matching splits calculated correctly", {
  expect_equal(NUnrooted(9), TreesMatchingSplit(0, 9))
  expect_equal(NUnrooted(9), TreesMatchingSplit(9, 0))
  expect_equal(NUnrooted(9), TreesMatchingSplit(8, 1))
  expect_equal(LnUnrooted(9), LnTreesMatchingSplit(0, 9))
  expect_equal(LnUnrooted.int(9), LnTreesMatchingSplit(9, 0))
  expect_equal(LnUnrooted.int(9), LnTreesMatchingSplit(1, 8))
  expect_equal(NUnrooted(9), TreesMatchingSplit(8, 1))
  expect_equal(Log2Unrooted(9), Log2TreesMatchingSplit(0, 9))
  expect_equal(Log2Unrooted.int(9), Log2TreesMatchingSplit(9, 0))
  expect_equal(Log2Unrooted.int(9), Log2TreesMatchingSplit(1, 8))
  expect_equal(log(315/10395)/-log(2), SplitInformation(3, 5))
})

test_that("SplitInformation() handles Splits", {
  t6 <- BalancedTree(6)
  expect_equal(SplitInformation(t6), SplitInformation(as.Splits(t6)))
  expect_equal(SplitInformation(t6), c(`8` = SplitInformation(3, 3),
                                       `9` = SplitInformation(2, 4),
                                       `11` = SplitInformation(4, 2)))
})

test_that("UnrootedTreesMatchingSplit() correct", {
  expect_equal(NRooted(3) * NRooted(5), UnrootedTreesMatchingSplit(c(3, 5)))
  expect_equal(LnRooted(30) + LnRooted(50), LnUnrootedTreesMatchingSplit(30, 50))
  expect_equal(sum(log2DoubleFactorials[15],
                   Log2DoubleFactorial(1:4 + 1:4 - 3L))
               - log2DoubleFactorials[11],
               Log2UnrootedTreesMatchingSplit(1:4))
})

test_that("MultiSplitInformation() works", {
  expect_equal(12.8323, MultiSplitInformation(3:5), tolerance = 6)
  expect_equal(CharacterInformation(rep(c("-", "?", 0:2), 1:5)),
               MultiSplitInformation(3:5))
})

test_that("TreesMatchingSplit() accepts different formats", {
  expect_equal(TreesMatchingSplit(4, 5), TreesMatchingSplit(4:5))
  expect_equal(LnTreesMatchingSplit(4, 5), LnTreesMatchingSplit(4:5))
  expect_equal(Log2TreesMatchingSplit(4, 5), Log2TreesMatchingSplit(4:5))
})

#' Drop leaves from tree
#'
#' `DropTip()` removes specified leaves from a phylogenetic tree, collapsing
#' incident branches.
#'
#' This function differs from [`ape::drop.tip()`], which roots unrooted trees,
#' and which can crash when trees' internal numbering follows unexpected schema.
#'
#' @template treeParam
#' @param tip Character vector specifying labels of leaves in tree to be
#' dropped or kept, or integer vector specifying the indices of leaves to be
#' dropped or kept.
#' Specifying the index of an internal node will drop all descendants of that
#' node.
#' @param preorder Logical specifying whether to [Preorder] `tree` before
#' dropping tips.  Specifying `FALSE` saves a little time, but will result in
#' undefined behaviour if `tree` is not in preorder.
#' @param check Logical specifying whether to check validity of `tip`. If
#' `FALSE` and `tip` contains entries that do not correspond to leaves of the
#' tree, undefined behaviour may occur.
#'
#' @return `DropTip()` returns a tree of class `phylo`, with the requested
#' leaves removed. The edges of the tree will be numbered in preorder,
#' but their sequence may not conform to the conventions of [`Preorder()`].
#'
#' @examples
#' tree <- BalancedTree(9)
#' plot(tree)
#' plot(DropTip(tree, c("t5", "t6")))
#' 
#' unrooted <- UnrootTree(tree)
#' plot(unrooted)
#' plot(DropTip(unrooted, 4:5))
#'
#' @family tree manipulation
#' @template MRS
#' @export
DropTip <- function(tree, tip, preorder = TRUE, check = TRUE) {
  UseMethod("DropTip")
}

#' @rdname DropTip
#' @return `KeepTip()` returns `tree` with all leaves not in `tip` removed,
#' in preorder.
#' @export
KeepTip <- function(tree, tip, preorder = TRUE, check = TRUE) {
  UseMethod("KeepTip")
}

#' @rdname DropTip
#' @export
DropTip.phylo <- function(tree, tip, preorder = TRUE, check = TRUE) {
  if (preorder && (length(attr(tree, "order")) == 0 || 
                   attr(tree, "order")[[1]] != "preorder")) {
    tree <- Preorder(tree)
  }
  labels <- tree[["tip.label"]]
  nTip <- length(labels)
  
  # Establish which tips should be dropped
  if (is.null(tip) || !length(tip) || any(is.na(tip))) {
    drop <- logical(nTip)
  } else if (is.character(tip)) {
    if (check) {
      if (sum(unique(labels) %in% tip) != length(tip)) {
        warning(paste(tip[!tip %in% labels], collapse = ", "),
                " not present in tree")
      }
    }
    drop <- labels %in% tip
  } else if (is.numeric(tip)) {
    if (check) {
      nNodes <- nTip + tree[["Nnode"]]
      if (any(tip > nNodes)) {
        warning("Tree only has ", nNodes, " nodes")
        tip <- tip[tip <= nNodes]
      }
      if (any(tip < 1L)) {
        warning("`tip` must be > 0")
        tip <- tip[tip > 0L]
      }
      
      if (any(tip > nTip)) {
        edge <- tree[["edge"]]
        parent <- edge[, 1]
        child <- edge[, 2]
        drop <- tabulate(c(tip[tip <= nTip],
                           .DescendantTips(parent, child, nTip,
                                           tip[tip > nTip])),
                         nbins = nTip)
      } else {
        drop <- as.logical(tabulate(tip, nbins = nTip))
      }
    } else {
      drop <- as.logical(tabulate(tip, nbins = nTip))
    }
  } else if (is.logical(tip)) {
    if (check) {
      if (length(tip) != nTip) {
        stop("`tip` must list `TRUE` or `FALSE` for each leaf.")
      }
    }
    drop <- tip
  } else {
    stop("`tip` must be of type character or numeric")
  }
  
  # Having established which tips to drop,
  # work out what to do about it
  if (all(drop)) {
    return(structure(list(edge = matrix(0, 0, 2), tip.label = character(0),
                          Nnode = 0), class = "phylo"))
  }
  
  if (any(drop)) {
    weights <- tree[["edge.length"]]
    keep <- !drop
    nodeLabel <- tree[["node.label"]]
    if (!is.null(weights)) {
      original <- PathLengths(tree, fullMatrix = TRUE)
    }
    if (!is.null(weights) || !is.null(nodeLabel)) {
      verts <- KeptVerts(tree, keep)
      tree[["node.label"]] <- nodeLabel[verts[-seq_len(nTip)]]
    }
    tree[["edge"]] <- keep_tip(tree[["edge"]], keep)
    tree[["tip.label"]] <- labels[keep]
    tree[["Nnode"]] <- dim(tree[["edge"]])[1] + 1L - sum(keep)
    
    if (!is.null(weights)) {
      newVerts <- matrix(which(verts)[tree[["edge"]]], ncol = 2L)
      tree[["edge.length"]] <- original[newVerts]
    }
  }
  
  # Return:
  tree
}

#' @rdname DropTip
#' @export
KeepTip.phylo <- function(tree, tip, preorder = TRUE, check = TRUE) {
  drop <- switch(
    mode(tip),
    "logical" = {
      !tip
    }, "character" = {
      setdiff(TipLabels(tree), MatchStrings(tip, TipLabels(tree), warning))
    }, "numeric" = {
      nTip <- NTip(tree)
      nValid <- if (inherits(tree, "phylo")) {
        nTip + tree[["Nnode"]]
      } else {
        nTip
      }
      badTips <- tip > nValid | tip < 1
      if (any(badTips)) {
        warning("`tip` must be between 1 and ", nValid)
        tip <- tip[!badTips]
      }
      nodes <- tip > nTip
      keep <- if (any(nodes)) {
        edge <- tree[["edge"]]
        theNodes <- tip[nodes]
        nodeDesc <- DescendantTips(edge[, 1], edge[, 2], node = theNodes)
        union(tip[!nodes], which(
          if (is.null(dim(nodeDesc))) nodeDesc else colSums(nodeDesc) > 0))
      } else {
        tip
      }
      setdiff(seq_len(nTip), keep)
    },
    stop("`tip` must be of type character, logical or numeric")
  )
  DropTip(tree, drop, preorder, check)
}

#' @rdname DropTip
#' @export
KeepTip.Splits <- KeepTip.phylo

#' @rdname DropTip
#' @examples summary(DropTip(as.Splits(tree), 4:5))
#' @family split manipulation functions
#' @export
DropTip.Splits <- function(tree, tip, preorder, check = TRUE) {
  labels <- TipLabels(tree)
  if (is.null(tip) || !length(tip) || any(is.na(tip))) {
    return(tree)
  } else if (is.character(tip)) {
    if (check) {
      drop <- match(tip, labels)
      missing <- is.na(drop)
      if (any(missing)) {
        warning(paste(tip[missing], collapse = ", "), " not present in tree")
        drop <- drop[!missing]
      }
      drop <- as.logical(tabulate(drop, NTip(tree)))
    } else {
      drop <- labels %in% tip
    }
  } else if (is.numeric(tip)) {
    nTip <- NTip(tree)
    if (check) {
      if (any(tip > nTip)) {
        warning("Tree only has ", nTip, " leaves")
        tip <- tip[tip <= nTip]
      }
      if (any(tip < 1L)) {
        warning("`tip` must be > 0")
        tip <- tip[tip > 0L]
      }
    }
    drop <- as.logical(tabulate(tip, nTip))
  } else if (is.logical(tip)) {
    if (check && length(tip) != NTip(tree)) {
      stop("`tip` must contain an entry for each leaf of `tree`.")
    }
    drop <- tip
  } else {
    stop("`tip` must be of type character, logical or numeric")
  }
  
  keep <- !drop
  thinner <- structure(thin_splits(tree, drop),
                       nTip = sum(keep),
                       tip.label = labels[keep],
                       class = "Splits")
  
  # Return:
  unique(thinner, fromLast = TRUE)
}

# nodes must all be internal
.DescendantTips <- function(parent, child, nTip, nodes, isDesc = logical(nTip)) {
  newDescs <- child[parent %in% nodes]
  recurse <- newDescs > nTip
  
  # Return:
  if (any(recurse)) {
    isDesc[newDescs[!recurse]] <- TRUE
    .DescendantTips(parent, child, nTip, newDescs[recurse], isDesc)
  } else {
    isDesc[newDescs] <- TRUE
    which(isDesc)
  }
}

#' @describeIn DropTip Direct call to `DropTip.phylo()`, to avoid overhead of
#' querying object's class.
#' @export
DropTipPhylo <- DropTip.phylo

#' @rdname DropTip
#' @export
DropTip.multiPhylo <- function(tree, tip, preorder = TRUE, check = TRUE) {
  at <- attributes(tree)
  tree <- lapply(tree, DropTip, tip, preorder)
  attributes(tree) <- at
  if (!is.null(at[["TipLabel"]])) {
    attr(tree, "TipLabel") <- setdiff(at[["TipLabel"]], tip)
  }
  tree
}

#' @rdname DropTip
#' @export
KeepTip.multiPhylo <- function(tree, tip, preorder = TRUE, check = TRUE) {
  at <- attributes(tree)
  tree <- lapply(tree, KeepTip, tip = tip, preorder = preorder, check = check)
  attributes(tree) <- at
  if (!is.null(at[["TipLabel"]])) {
    attr(tree, "TipLabel") <- intersect(at[["TipLabel"]], tip)
  }
  tree
}

#' @rdname DropTip
#' @export
DropTip.list <- function(tree, tip, preorder = TRUE, check = TRUE) {
  if (all(vapply(tree, inherits, logical(1), "phylo"))) {
    DropTip.multiPhylo(tree, tip, preorder, check)
  } else {
    stop("Expected a list of `phylo` objects, but not all elements inherit ",
         "from \"phylo\".\n  Classes found: ",
         paste(unique(vapply(tree, function(x) paste(class(x), collapse = "/"),
                             character(1))),
               collapse = ", "))
  }
}

#' @rdname DropTip
#' @export
KeepTip.list <- function(tree, tip, preorder = TRUE, check = TRUE) {
  if (all(vapply(tree, inherits, logical(1), "phylo"))) {
    KeepTip.multiPhylo(tree, tip, preorder, check)
  } else {
    stop("Expected a list of `phylo` objects, but not all elements inherit ",
         "from \"phylo\".\n  Classes found: ",
         paste(unique(vapply(tree, function(x) paste(class(x), collapse = "/"),
                             character(1))),
               collapse = ", "))
  }
}

#' @rdname DropTip
#' @export
DropTip.NULL <-  function(tree, tip, preorder = TRUE, check = TRUE) {
  NULL
}

#' @rdname DropTip
#' @export
KeepTip.NULL <-  function(tree, tip, preorder = TRUE, check = TRUE) {
  NULL
}

#' @describeIn DropTip Faster version with no checks.
#' Does not retain labels or edge weights.
#' Edges must be listed in preorder.
#' May crash if improper input is specified.
#' @export
KeepTipPreorder <- function(tree, tip) {
  if (!any(tip)) {
    structure(list(edge = matrix(0, 0, 2), tip.label = character(0),
                          Nnode = 0), class = "phylo")
  } else if (all(tip)) {
    tree
  } else {
    tree[["edge"]] <- keep_tip(tree[["edge"]], tip)
    tree[["tip.label"]] <- as.character(which(tip))
    tree[["Nnode"]] <- dim(tree[["edge"]])[1] + 1L - sum(tip) 
    tree
  }
}
  
#' @describeIn DropTip Faster version with no checks.
#' Does not retain labels or edge weights.
#' Edges must be listed in postorder.
#' May crash if improper input is specified.
#' @export
KeepTipPostorder <- function(tree, tip) {
  if (!any(tip)) {
    structure(list(edge = matrix(0, 0, 2), tip.label = character(0),
                          Nnode = 0), class = "phylo")
  } else if (all(tip)) {
    tree
  } else {
    edge <- tree[["edge"]]
    tree[["edge"]] <- keep_tip(edge[dim(edge)[1]:1, ], tip)
    tree[["tip.label"]] <- tree[["tip.label"]][tip]
    tree[["Nnode"]] <- dim(tree[["edge"]])[1] + 1L - sum(tip) 
    tree
  }
}

nasty <- structure(list(edge = structure(
  c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
    5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
  .Dim = c(12, 2)),
  Nnode = 5L,
  tip.label = letters[1:8]),
  class = "phylo") # Danger: Do not plot!

test_that("keep_tip() fails with small edge", {
  # False positives with ASaN
  skip_if(Sys.getenv("USING_ASAN") != "")
  expect_error(keep_tip(BalancedTree(8)$edge[, 1, drop = FALSE],
                        !tabulate(6:4, 8)))
  expect_error(keep_tip(BalancedTree(8)$edge[, 1], !tabulate(6:4, 8)))
})

test_that("keep_tip() works", {
  expect_error(keep_tip(BalancedTree(8)$edge[, c(1, 2, 1)], !tabulate(6:4, 8)))
  
  expect_equal(keep_tip(BalancedTree(9)$edge, !tabulate(5:8, 9)),
               matrix(c(6, 7, 8, 9, 9, 8, 7, 6,
                        7, 8, 9, 1, 2, 3, 4, 5), 8, 2))
  
  expect_equal(keep_tip(BalancedTree(8)$edge, !tabulate(6:4, 8)),
               matrix(c(6, 7, 8, 8, 7, 6, 9, 9,
                        7, 8, 1, 2, 3, 9, 4, 5), 8, 2))
  expect_equal(keep_tip(BalancedTree(8)$edge, !tabulate(1:4, 8)),
               BalancedTree(4)$edge)
  
  expect_equal(keep_tip(BalancedTree(8)$edge, !tabulate(3:8, 8)),
               BalancedTree(2)$edge)
  
  expect_equal(keep_tip(UnrootTree(BalancedTree(4))$edge, !tabulate(1, 4)),
               matrix(c(4, 4, 4, 1, 2, 3), 3, 2))
  
  expect_equal(Preorder(keep_tip(as.phylo(3, 7)$edge, !tabulate(1:4, 7))),
               PectinateTree(3)$edge)
  
  # Rooted tree may lose root when reaching a polytomy:
  expect_equal(keep_tip(root(StarTree(6), 4, resolve.root = TRUE)$edge,
                        !tabulate(4, 6)),
               StarTree(5)$edge)
  
  # But need not:
  expect_equal(keep_tip(RootTree(BalancedTree(5), 3)$edge, !tabulate(3, 5)),
               BalancedTree(4)$edge)
  
  unrooted <- ape::read.tree(text = "(a, b, (c, d, ((e1, e2), (f, g))));")
  expect_equal(keep_tip(unrooted$edge, !tabulate(1:4, 8)),
               UnrootTree(BalancedTree(4))$edge)
})

test_that("DropTip() works", {
  bal8 <- BalancedTree(8)
  expect_equal(NTip(DropTip(bal8, 1:8, check = FALSE)), 0)
  expect_equal(NTip(KeepTip(bal8, integer(0), check = FALSE)), 0)
  expect_warning(expect_true(all.equal(bal8, DropTip(bal8, -1))),
                 "`tip` must be > 0")
  expect_warning(expect_true(all.equal(bal8, KeepTip(bal8, 0:8))),
                 "`tip` must be between 1 and 15")
  expect_warning(expect_true(all.equal(bal8, KeepTip(bal8, 9:16))),
                 "`tip` must be between 1 and 15")
  expect_warning(expect_true(all.equal(bal8, DropTip(bal8, 99))),
                 "Tree only has 15 nodes")
  expect_warning(expect_true(all.equal(bal8, DropTip(bal8, "MissingTip"))),
                 "not present in tree")
  expect_warning(expect_true(
    all.equal(bal8, KeepTip(bal8, c(TipLabels(bal8), "MissingTip")))),
                 "Could not find 'MissingTip'")
  expect_warning(expect_identical(
    DropTip(bal8, c("t8", "NotThere"), check = TRUE),
    DropTip(bal8, c("t8", "NotThere"), check = FALSE)), "not present in tree")
  expect_error(DropTip(bal8, TRUE),
               "`tip` must list `TRUE` or `FALSE` for each leaf")
  expect_error(KeepTip(bal8, TRUE),
               "`tip` must list `TRUE` or `FALSE` for each leaf")
  expect_error(DropTip(bal8, list("Invalid format")), "`tip` must be of type")
  expect_error(KeepTip(bal8, list("Invalid format")), "`tip` must be of type")
  
  expect_equal(DropTip(bal8, 7:8), DropTip(bal8, 15L))
  expect_true(all.equal(ape::drop.tip(bal8, 6:8), DropTip(bal8, 6:8)))
  expect_true(all.equal(ape::drop.tip(bal8, c(3, 5, 7)),
                        DropTip(bal8, c(3, 5, 7))))
  
  expect_equal(KeepTip(bal8, 9), bal8)
  expect_equal(KeepTip(bal8, c(9, 1)), bal8)
  expect_equal(KeepTip(bal8, 10), BalancedTree(4))
  expect_equal(KeepTip(bal8, c(6:8, 14)), KeepTip(bal8, 13))
  
  expect_equal(Preorder(DropTip(Preorder(nasty), c(1, 3))),
               Preorder(DropTip(nasty, c(1, 3))))
  
  expect_null(DropTip(NULL))
  expect_null(KeepTip(NULL, "tip"))
})

test_that("DropTip() supports node labels", {
  bal8 <- RootTree(BalancedTree(8), 1)
  startLabel <- paste("Node", 9:15)
  bal8$node.label <- startLabel
  expect_equal(DropTip(bal8, 1)[["node.label"]], startLabel[-1])
  expect_equal(DropTip(bal8, 5:6)[["node.label"]], startLabel[-(5:6)])
})

test_that("DropTip() root relocation", {
  nTip <- 12
  nKept <- nTip / 2L
  for (i in c(8, 30, 324)) { # A selection of interesting failure cases
    set.seed(i)
    
    bigTree <- RandomTree(nTip)
    bigDrop <- sample.int(nTip, nKept)
    bigKeep <- setdiff(seq_len(nTip), bigDrop)
    
    reduced <- DropTip(bigTree, bigDrop)
    expect_true(all.equal(reduced, unroot(drop.tip(bigTree, bigDrop))))
    
    map <- which(KeptVerts(bigTree, !tabulate(bigDrop, nTip)))
    expect_equal(map[seq_len(nKept)], sort(bigKeep))
    newDesc <- .ListDescendants(reduced)
    bigDesc <- .ListDescendants(bigTree)
    
    for (newNode in (nKept + 1):(nKept * 2 - 2)) {
      expect_equal(sort(map[newDesc[[newNode]]]),
                   sort(intersect(bigDesc[[map[newNode]]], bigKeep)))
    }
  }
  #microbenchmark::microbenchmark(ape::drop.tip(bigTree, bigTip), DropTip(bigTree, bigTip))
  
  nTip <- 1284
  set.seed(43)
  bigTree <- RandomTree(nTip)
  bigDrop <- sample.int(nTip, nKept)
  bigKeep <- setdiff(seq_len(nTip), bigDrop)
  
  reduced <- DropTip(bigTree, bigDrop)
  expect_true(all.equal(reduced, unroot(drop.tip(bigTree, bigDrop))))
})

test_that("DropTip.multiPhylo() with attributes", {
  multi <- c(bal8 = BalancedTree(8), pec8 = PectinateTree(8))
  attr(multi, "TipLabel") <- paste0("t", 1:8)
  not6 <- setdiff(TipLabels(multi[[1]]), "t6")
  not8 <- setdiff(TipLabels(multi[[1]]), "t8")
  
  expect_equal(DropTip(unclass(multi), "t6"), unclass(DropTip(multi, "t6")))
  expect_equal(KeepTip(unclass(multi), not6), unclass(KeepTip(multi, not6)))
  
  expect_equal(attr(DropTip(multi, "t8"), "TipLabel"),
               paste0("t", 1:7))
  expect_equal(attr(KeepTip(multi, not8), "TipLabel"),
               paste0("t", 1:7))
  expect_equal(names(DropTip(multi, "t8")), names(multi))
  expect_equal(names(KeepTip(multi, not8)), names(multi))
  expect_equal(DropTip(multi[1], "t1")[[1]], DropTip(multi[[1]], "t1"))
  expect_equal(KeepTip(multi[1], not8)[[1]], KeepTip(multi[[1]], not8))
})

test_that("DropTip.Splits()", {
  bal9 <- BalancedTree(9)
  s9 <- as.Splits(bal9)
  s19 <- as.Splits(PectinateTree(19))
  
  expect_error(DropTip(s9, c(TRUE, FALSE)), "each leaf\\b")
  expect_warning(
    expect_warning(
      expect_equal(DropTip(s9, c(0, 9:10)), DropTip(s9, !tabulate(1:8, 9))),
      "only has 9 leaves"),
    "`tip` must be > 0"
  )
  expect_error(DropTip(s9, raw(1)), "`tip` must be of type")
  
  expect_equal(s9, DropTip(s9, NULL))
  expect_warning(
    expect_equal(DropTip(s9, c("t7", "missing")),
                 DropTip(s9, "t7", check = FALSE)),
    "not present in tree")
  
  expect_equal(unname(DropTip(s9, 4:5)), unname(as.Splits(DropTip(bal9, 4:5))))
  expect_equal(unname(KeepTip(s9, c(1:4, 7:9))),
               unname(as.Splits(DropTip(bal9, 6:5))))
  expect_warning(expect_equal(KeepTip(s9, 1:10), s9),
                 "`tip` must be between 1 and 9")
  
  expect_equal(DropTip(s9[[1:5]], 8:9),
               DropTip(s9, 8:9))
  
  expect_equal(thin_splits(s9, !logical(9)),
                           structure(raw(0), .Dim = c(0L, 0L)))
  expect_equal(thin_splits(s19, 1:19 %in% 2:19),
               structure(raw(0), .Dim = c(0L, 1L)))
  expect_equal(thin_splits(s9, logical(9)), s9, ignore_attr = TRUE)
  
  expect_equal(DropTip(s9, TipLabels(s9)), as.Splits(ZeroTaxonTree()))
})

test_that("KeepTip() works", {
  # Some of this now duplicates newer tests in DropTip() above -
  # TODO identify and remove duplication.
  expect_warning(expect_true(all.equal(
    BalancedTree(paste0("t", 5:8)),
    KeepTip(BalancedTree(8), paste0("t", 5:9))
  )))
  
  expect_true(all.equal(
    KeepTip(Postorder(UnrootTree(BalancedTree(letters[1:12]))), 6:12),
    read.tree(text = "(f, (i, (g, h)), (l, (j, k)));")
  ))
  
  expect_identical(
    KeepTip(BalancedTree(12), !tabulate(1:5, 12)),
    DropTip(BalancedTree(12), !tabulate(6:12, 12))
    )
  
  s9 <- as.Splits(BalancedTree(9))
  expect_equal(KeepTip(s9, character(0)), DropTip(s9, TipLabels(s9)))
  expect_equal(KeepTip(s9, TipLabels(s9)), DropTip(s9, character(0)))
})

test_that("KeepTip() retains edge lengths", {
  bal9 <- BalancedTree(9)
  bal9$edge.length <- 10 * 1:16
  expect_equal(KeepTip(bal9, c(1, 3, 5, 7, 9))[["edge.length"]],
               c(10, 20, 30 + 40,
                 60,
                 70 + 90,
                 100, 110 + 130,
                 140 + 160))
})

test_that("KeepTipPreorder()/Postorder()", {
  pre <- Preorder(BalancedTree(1:9))
  post <- Postorder(pre)
  keep <- !tabulate(7, 9)
  expect_true(all.equal(KeepTipPreorder(pre, keep),
                        KeepTip(pre, keep)))
  expect_true(all.equal(KeepTipPostorder(post, keep),
                        KeepTip(post, keep)))
  
  emptyTree <- structure(list(edge = matrix(0, 0, 2),
                              tip.label = character(0),
                              Nnode = 0), class = "phylo")
  expect_equal(KeepTipPreorder(pre, logical(9)),
               emptyTree)
  expect_equal(KeepTipPostorder(post, logical(9)),
               emptyTree)

  expect_equal(KeepTipPreorder(pre, !logical(9)),
               pre)
  expect_equal(KeepTipPostorder(post, !logical(9)),
               post)
})

test_that("DropTip.list / KeepTip.list error on non-phylo elements", {
  bad_list <- list(list(a = 1), list(b = 2))
  expect_error(DropTip(bad_list, "x"), "not all elements inherit")
  expect_error(KeepTip(bad_list, "x"), "not all elements inherit")
})

test_that("KeptPaths() works", {
  tree <- BalancedTree(7)
  tree$edge.length <- 1:12
  paths <- PathLengths(tree)
  keptTips <- c(2, 4, 6)
  kept <- KeptVerts(tree, 1:7 %in% keptTips)
  lengths <- c(1 + 2 + 4,
               2 + 4,
               1 + 5 + 7,
               5 + 7,
               8 + 9 + 11,
               1)
  expect_equal(paths[KeptPaths(paths, kept), "length"], lengths)
  expect_equal(paths[KeptPaths(paths, kept, FALSE), "length"],
               lengths[c(2, 4:6)])
  expected <- PathLengths(KeepTip(tree, keptTips))
  expected[, 1:2] <- which(kept)[unlist(expected[, 1:2])]
  
  KeptPaths(PathLengths(tree, TRUE), kept)
})
  

#' Select element at random
#'
#' `SampleOne()` is a fast alternative to  [`sample()`] that avoids some checks.
#'
#' @param x A vector to sample.
#' @param len (Optional) Integer specifying length of `x`.
#'
#' @return `SampleOne()` returns a length one vector, randomly sampled from `x`.
#'
#' @examples
#' SampleOne(9:10)
#' SampleOne(letters[1:4])
#'
#' @template MRS
#' @family utility functions
#' @export
SampleOne <- function(x, len = length(x)) {
  x[[sample.int(len, 1L, FALSE, NULL, FALSE)]]
}

#' Add tree to start of list
#'
#' `UnshiftTree()` adds a phylogenetic tree to the start of a list of trees.
#' This is useful where the class of a list of trees is unknown, or where
#' names of trees should be retained.
#'
#' Caution: adding a tree to a `multiPhylo` object whose own attributes apply
#' to all trees, for example trees read from a Nexus file, causes data to be
#' lost.
#'
#' @param add Tree to add to the list, of class [`phylo`][ape::read.tree].
#' @param treeList A list of trees, of class `list`,
#' [`multiPhylo`][ape::multiphylo], or, if a single tree,
#' [`phylo`][ape::read.tree].
#'
#' @return `UnshiftTree()` returns a list of class `list` or `multiPhylo`
#' (following the original class of `treeList`), whose first element is the
#' tree specified as `add.
#'
#' @seealso
#' [`c()`] joins a tree or series of trees to a `multiPhylo` object, but loses
#' names and does not handle lists of trees.
#'
#' @examples
#' forest <- as.phylo(0:5, 6)
#' tree <- BalancedTree(6)
#'
#' UnshiftTree(tree, forest)
#' UnshiftTree(tree, tree)
#' @template MRS
#' @family utility functions
#'
#' @export
UnshiftTree <- function(add, treeList) {
  if (inherits(treeList, "multiPhylo")) {
    structure(c(list(add), lapply(treeList, I)), class = "multiPhylo")
  } else if (inherits(treeList, "phylo")) {
    structure(list(add, treeList), class = "multiPhylo")
  } else { # including: if (is.list(trees)) {
    c(list(add), treeList)
  }
}

#' Apply a function that returns 64-bit integers over a list or vector
#'
#' Wrappers for members of the [`lapply()`] family intended for use when a
#' function `FUN` returns a vector of `integer64` objects.
#' `vapply()`, `sapply()` or `replicate()` drop the `integer64` class,
#' resulting in a vector of numerics that require conversion back to
#' 64-bit integers.  These functions restore the missing `class` attribute.
#'
#' @inheritParams base::lapply
#' @param FUN.LEN Integer specifying the length of the output of `FUN`.
#' @details For details of the underlying functions, see [`base::lapply()`].
#' @examples
#' sapply64(as.phylo(1:6, 6), as.TreeNumber)
#' vapply64(as.phylo(1:6, 6), as.TreeNumber, 1)
#' set.seed(0)
#' replicate64(6, as.TreeNumber(RandomTree(6)))
#' @template MRS
#' @seealso \code{\link[bit64]{integer64}()}
#' @family utility functions
#' @export
sapply64 <- function(X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE) {
  structure(sapply(X, FUN, ..., simplify, USE.NAMES), class = "integer64")
}

#' @rdname sapply64
#' @export
vapply64 <- function(X, FUN, FUN.LEN = 1, ...) {
  structure(vapply(X, FUN, FUN.VALUE = numeric(FUN.LEN), ...),
            class = "integer64")
}

#' @rdname sapply64
#' @export
replicate64 <- function(n, expr, simplify = "array") {
  sapply64(integer(n), eval.parent(substitute(function(...) expr)),
           simplify = simplify)
}

test_that("Random trees are set by R seed", {

  set.seed(0)
  tr1 <- RandomTree(123)
  tr2 <- RandomTree(123)
  set.seed(0)
  tr3 <- RandomTree(123)
  tr4 <- RandomTree(123)
  expect_identical(tr1, tr3)
  expect_identical(tr2, tr4)
  expect_false(identical(tr1, tr2))

  set.seed(1)
  tr5 <- RandomTree(123)
  set.seed(1)
  tr6 <- RandomTree(123)
  expect_identical(tr5, tr6)
  expect_false(identical(tr1, tr5))

})

test_that("Random trees drawn from uniform distribution", {
  # NB We're also testing sapply64 and vapply64 in this chunk
  expect_error(.RandomParent(1))
  expect_error(.RandomParent(0))
  expect_error(.RandomParent(-1))

  nSamples <- 100
  # Ape's trees are not uniformly distributed:
  ape_counts <- table(vapply64(lapply(rep(5, nSamples), ape::rtree),
                               as.TreeNumber, 1))
  expect_lt(chisq.test(ape_counts)$p.value, 0.005)

  # Our trees are:
  counts <- table(sapply64(lapply(rep(5, nSamples), RandomTree), as.TreeNumber))
  expect_gt(chisq.test(counts)$p.value, 0.001)

  if (FALSE) { # Takes many seconds - but worth checking manually?
    counts <- table(sapply64(lapply(rep(8, 200000), RandomTree), as.TreeNumber))
    expect_gt(chisq.test(counts)$p.value, 0.001)
  }

  expect_false(is.rooted(RandomTree(6, root = FALSE)))
  expect_true(is.rooted(RandomTree(6, root = TRUE)))
})

test_that("UnshiftTree() works", {
  t1 <- as.phylo(1, 8)
  t2..9 <- setNames(as.phylo(2:3, 8), letters[2:3])
  t1..9 <- setNames(as.phylo(1:3, 8), c("", letters[2:3]))
  attr(t1..9, "tip.label") <- NULL
  expect_true(all.equal(UnshiftTree(t1, t2..9), t1..9))
  expect_equal(unclass(t1..9), UnshiftTree(t1, unclass(t2..9)))
  expectation <- as.phylo(1:2, 8)
  attr(expectation, "tip.label") <- NULL
  expect_equal(expectation, UnshiftTree(t1, as.phylo(2, 8)))
})

#' Count cherries in a tree
#' 
#' `Cherries()` counts the number of vertices in a binary tree whose children
#' are both leaves.
#' 
#' @param tree A binary tree, of class `phylo`; or a matrix corresponding to its
#' edge matrix.
#' @param nTip Number of leaves in tree.
#' @return `Cherries()` returns an integer specifying the number of nodes whose
#' children are both leaves.
#' 
#' Equations for the number of cherries expected under uniform and Yule tree
#' models are derived by \insertCite{McKenzie2000;textual}{TreeTools}.
#' 
#' @references \insertAllCited{}
#' @family tree properties
#' @template MRS
#' @export
Cherries <- function(tree, nTip) UseMethod("Cherries")

#' @rdname Cherries
#' @export
Cherries.phylo <- function(tree, nTip = NTip(tree)) {
  n_cherries_wrapper(tree[["edge"]][, 1], tree[["edge"]][, 2], nTip)
}

#' @rdname Cherries
#' @export
Cherries.numeric <- function(tree, nTip) {
  if (is.null(dim(tree)) || dim(tree)[[2]] != 2) {
    stop("`tree` must be the edge matrix of a tree of class phylo")
  }
  n_cherries_wrapper(tree[, 1], tree[, 2], nTip)
}

#' Total Cophenetic Index
#'
#' `TotalCopheneticIndex()` calculates the total cophenetic index
#' \insertCite{Mir2013}{TreeTools} for any tree, a measure of its balance;
#' `TCIContext()` lists its possible values.
#'
#' The Total Cophenetic Index is a measure of tree balance -- i.e. whether
#' a (phylogenetic) tree comprises symmetric pairs of nodes, or has a pectinate
#' "caterpillar" shape.
#' The index has a greater resolution power than Sackin's and Colless' indices,
#' and can be applied to trees that are not perfectly resolved.
#'
#' For a tree with _n_ leaves, the Total Cophenetic Index can take values of
#' 0 to `choose(n, 3)`.
#' The minimum value is higher for a perfectly resolved (i.e. dichotomous) tree
#' (see Lemma 14 of Mir _et al._ 2013).
#' Formulae to calculate the expected values under the Yule and Uniform models
#' of evolution are given in Theorems 17 and 23.
#'
#' Full details are provided by \insertCite{Mir2013;textual}{TreeTools}.
#' 
#' The \ifelse{html}{\out{J<sup>1</sup>}}{\eqn{J^1}} index
#' \insertCite{Lemant2022}{TreeTools} has advantages over the Total Cophenetic
#' Index, particularly when comparing trees with different numbers of leaves,
#' or where the population size of nodes is meaningful; see [`J1Index()`].
#'  
#'
#' @template xPhylo
#'
#' @return
#' `TotalCopheneticIndex()` returns an integer denoting the total cophenetic index.
#'
#' `TCIContext()` returns a data frame detailing the maximum and minimum value
#' obtainable for the Total Cophenetic Index for rooted binary trees with the
#' number of leaves of the given tree, and the expected value under the Yule
#' and Uniform models.
#' The variance of the expected value is given under the Yule model, but cannot
#' be obtained by calculation for the Uniform model.
#'
#' @seealso
#' - [`J1Index()`] provides a more robust, universal tree balance index.
#' - `cophen.index()` in the package
#' [\pkg{CollessLike}](https://github.com/LuciaRotger/CollessLike)
#' provides an alternative implementation of this index and its predecessors.
#'
#' @references \insertAllCited{}
#'
#' @examples
#' # Balanced trees have the minimum index for a binary tree;
#' # Pectinate trees the maximum:
#' TCIContext(8)
#' TotalCopheneticIndex(PectinateTree(8))
#' TotalCopheneticIndex(BalancedTree(8))
#' TotalCopheneticIndex(StarTree(8))
#'
#'
#' # Examples from Mir et al. (2013):
#' tree12 <- ape::read.tree(text="(1, (2, (3, (4, 5))));")  #Fig. 4, tree 12
#' TotalCopheneticIndex(tree12) # 10
#' tree8  <- ape::read.tree(text="((1, 2, 3, 4), 5);")      #Fig. 4, tree 8
#' TotalCopheneticIndex(tree8)  # 6
#' TCIContext(tree8)
#' TCIContext(5L) # Context for a tree with 5 leaves.
#'
#' @family tree characterization functions
#'
#' @encoding UTF-8
#' @template MRS
#' @export
TotalCopheneticIndex <- function(x) UseMethod("TotalCopheneticIndex")

.Depth <- function(parent, child) {
  depth  <- integer(max(parent))
  for (i in seq_along(parent)) {
    depth[child[i]] <- depth[parent[i]] + 1L
  }

  # Return:
  depth
}

#' @export
TotalCopheneticIndex.NULL <- function(x) NULL

#' @importFrom fastmatch %fin%
#' @export
TotalCopheneticIndex.phylo <- function(x) {
  nTip   <- NTip(x)
  edge   <- Preorder(x)[["edge"]]
  parent <- edge[, 1]
  child  <- edge[, 2]
  depth  <- .Depth(parent, child)

  ancestors <- lapply(seq_len(nTip),
                      function(node) ListAncestors(parent, child, node))

  lca.depth <- vapply(seq_len(nTip), function(i) {
    vapply(seq_len(nTip), function(j) {
      anc.i <- ancestors[[i]]
      anc.j <- ancestors[[j]]
      lca <- max(anc.i[anc.i %fin% anc.j])

      # Return:
      depth[lca]
    }, integer(1L))
  }, integer(nTip))

  # Return:
  sum(lca.depth[upper.tri(lca.depth)])
}

#' @export
TotalCopheneticIndex.list <- function(x) {
  vapply(x, TotalCopheneticIndex, integer(1))
}

#' @export
TotalCopheneticIndex.multiPhylo <- TotalCopheneticIndex.list

#' @rdname TotalCopheneticIndex
#' @export
TCIContext <- function(x) UseMethod("TCIContext")

#' @export
TCIContext.phylo <- function(x) {
  TCIContext.numeric(NTip(x))
}

.MCI <- function(n) { # Lemma 14 in Mir er al 2013
  if (n < 3L) return(0L)
  halfN <- n / 2L
  topHalf <- ceiling(halfN)
  btmHalf <- floor(halfN)
  .MCI(topHalf) + .MCI(btmHalf) +
    choose(topHalf, 2L) + choose(btmHalf, 2L)
}

#' @rdname TotalCopheneticIndex
#' @export
TCIContext.numeric <- function(x) {
  H  <- function(n) sum(1 / seq_len(n))
  H2 <- function(n) sum(1 / (seq_len(n) ^ 2))

  maximum <- choose(x, 3L)
  minimum <- .MCI(x)

  # Theorem 17
  uniform.expected <- exp(lchoose(x, 2) - log(2) + 
                            LnDoubleFactorial((x + x) - 2L) - 
                            LnDoubleFactorial((x + x) - 3L)
                          ) - choose(x, 2)
  yule.expected    <- (x * (x + 1)) - (2 * x * H(x))
  yule.variance    <- ((1 / 12) * (x^4 - (10 * x^3) + (131 * x^2) - (2 * x))) -
    (4 * x^2 * H2(x)) - (6 * x * H(x))

  # Return:
  data.frame(maximum, minimum, uniform.expected, yule.expected, yule.variance)
}

test_that("Trees from Mir et al. 2013 are scored correctly", {
  Tree <- function(text) ape::read.tree(text = text)
  expect_identical(0L,  TotalCopheneticIndex(Tree("(1,2,3,4,5);")))
  expect_identical(1L,  TotalCopheneticIndex(Tree("((1,2),3,4,5);")))
  expect_identical(2L,  TotalCopheneticIndex(Tree("((1,2),(3,4),5);")))
  expect_identical(3L,  TotalCopheneticIndex(Tree("((1,2,3),4,5);")))
  expect_identical(4L,  TotalCopheneticIndex(Tree("(((1,2),3),4,5);")))
  expect_identical(4L,  TotalCopheneticIndex(Tree("((1,2,3),(4,5));")))
  expect_identical(5L,  TotalCopheneticIndex(Tree("(((1,2),3),(4,5));")))
  expect_identical(6L,  TotalCopheneticIndex(Tree("((1,2,3,4),5);")))
  expect_identical(7L,  TotalCopheneticIndex(Tree("(((1,2),3,4),5);")))
  expect_identical(8L,  TotalCopheneticIndex(Tree("(((1,2),(3,4)),5);")))
  expect_identical(9L,  TotalCopheneticIndex(Tree("(((1,2,3),4),5);")))
  expect_identical(10L, TotalCopheneticIndex(PectinateTree(5)))
  expect_identical(c(10L, 8L, 10L), TotalCopheneticIndex(as.phylo(0:2, 5)))

  expect_equal(TotalCopheneticIndex(BalancedTree(13)),
               TCIContext(13)[1, "minimum"])
  expect_equal(TotalCopheneticIndex(PectinateTree(13)),
               TCIContext(13)[1, "maximum"])

  nasty <- structure(list(edge = structure(
    c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
      5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
    .Dim = c(12, 2)),
    Nnode = 5L,
    tip.label = letters[1:8]),
    class = "phylo")
  expect_equal(28L, TotalCopheneticIndex(nasty))

  expect_equal(TCIContext(BalancedTree(5)), TCIContext(5L))
})

test_that("Expectations are correct", {
  nTip <- 5
  uniform <- DropTip(
    unlist(recursive = FALSE,
           lapply(
             lapply(as.phylo(1:NUnrooted(nTip) - 1, nTip),
                    AddTipEverywhere, "ROOT"),
             RootTree, "ROOT")
           ),
    "ROOT")
  tci <- TotalCopheneticIndex(uniform)
  context <- TCIContext(nTip)
  expect_equal(range(tci), range(context[1:2]))
  expect_equal(mean(tci), context[["uniform.expected"]])
  
  set.seed(1)
  unif <- replicate(100, TotalCopheneticIndex(RandomTree(10, root = TRUE)))
  expect_equal(
    mean(unif),
    TCIContext(10)$uniform.expected, # 76
    tolerance = 0.075
  )
})

test_that("Large trees work", {
  expect_equal(TotalCopheneticIndex(BalancedTree(164)), .MCI(164))
  H  <- function(n) sum(1 / seq_len(n))
  H2 <- function(n) sum(1 / (seq_len(n) ^ 2))
  
  x <- 1024
  
  expect_equal(unlist(TCIContext(x)),
               c(maximum = choose(x, 3), minimum = .MCI(x),
                 uniform.expected = exp(lchoose(x, 2) - log(2) +
                                          LnDoubleFactorial((x + x) - 2L) - 
                                          LnDoubleFactorial((x + x) - 3L)
                                        ) - choose(x, 2),
                 yule.expected =  (x * (x + 1)) - (2 * x * H(x)),
                 yule.variance = (
                   (1 / 12) * (x^4 - (10 * x^3) + (131 * x^2) - (2 * x))
                   ) - (4 * x^2 * H2(x)) - (6 * x * H(x))
               ))
})

#' Visualize position of rogue taxa
#'
#' Plots a consensus of trees with a rogue taxon omitted, with edges coloured
#' according to the proportion of trees in which the taxon attaches to that
#' edge, after \insertCite{Klopfstein2019;textual}{TreeTools}.
#' 
#' Rogue taxa can be identified using the package \pkg{Rogue}
#' \insertCite{SmithCons}{TreeTools}.
#'
#' @param trees List or `multiPhylo` object containing phylogenetic trees
#' of class `phylo` to be summarized.
#' @param tip Numeric or character identifying rogue leaf, in format accepted
#' by [`DropTip()`].
#' @param p A numeric value between 0.5 and 1 giving the proportion for a clade
#' to be represented in the consensus tree (see [`Consensus()`]).
#' @param \dots Additional parameters to `plot.phylo()`.
#' @param plot Logical specifying whether to plot the tree.
#' @param Palette Function that takes a parameter `n` and generates a colour
#' palette with `n` entries.
#' @param nullCol Colour to paint regions of the tree on which the rogue is
#' never found.
#' @template edgeLengthParam
#' @inheritParams RootTree
#' @param thin,fat Numeric specifying width to plot edges if the rogue tip
#' never / sometimes does attach to them.
#' @param sort Logical specifying whether to sort consensus tree using
#' [`SortTree()`].
#' @param legend Character vector specifying position of legend (e.g.
#' `"bottomleft"`), or `"none"` to suppress legend.
#' For fine-grained control of legend, use [`PlotTools::SpectrumLegend()`].
#' @param legend.inset Numeric specifying fraction of plot width / height
#' by which the legend's position should be inset.
#'
#' @return `RoguePlot()` invisibly returns a list whose elements are:
#' - `cons`: The reduced consensus tree, sorted if `sort = TRUE`, otherwise 
#' in preorder;
#' - `onEdge`: a vector of integers specifying the number of
#' trees in `trees` in which the rogue leaf is attached to each edge in turn
#' of the consensus tree;
#' - `atNode`: a vector of integers specifying the number of trees in `trees`
#' in which the rogue leaf is attached to an edge collapsed into each node
#' of the consensus tree.
#' - `legendLabels`: A character vector suggesting labels for a plot legend;
#' suitable for `PlotTools::SpectrumLegend(legend = x$legendLabels)`.
#' @references \insertAllCited{}
#' @examples
#' trees <- list(read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),
#'               read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),
#'               read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),
#'               read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),
#'               read.tree(text = "(rogue, (a, (b, (c, (d, (e, f))))));"),
#'               read.tree(text = "((rogue, a), (b, (c, (d, (e, f)))));"),
#'               read.tree(text = "(a, (b, ((c, d), (rogue, (e, f)))));"),
#'               read.tree(text = "(a, (b, ((c, (rogue, d)), (e, f))));"),
#'               read.tree(text = "(a, (b, (c, (d, (rogue, (e, f))))));"))
#' plotted <- RoguePlot(trees, "rogue", legend = "topleft", legend.inset = 0.02)
#' PlotTools::SpectrumLegend(
#'   "bottomleft",
#'   palette = colorRampPalette(c(par("fg"), "#009E73"), space = "Lab")(100),
#'   legend = plotted$legendLabels,
#'   cex = 0.4
#' )
#' @template MRS
#' @importFrom fastmatch fmatch %fin%
#' @importFrom graphics par
#' @importFrom grDevices colorRamp colorRampPalette rgb
#' @importFrom PlotTools SpectrumLegend
#' @family consensus tree functions
#' @export
RoguePlot <- function(trees, tip, p = 1, plot = TRUE,
                      Palette = colorRampPalette(c(par("fg"), "#009E73"),
                                                 space = "Lab"),
                      nullCol = rgb(colorRamp(unlist(par(c("fg", "bg"))),
                                              space = "Lab")(0.8) / 255),
                      edgeLength = NULL,
                      thin = par("lwd"), fat = thin + 1L,
                      outgroupTips,
                      sort = FALSE,
                      legend = "none",
                      legend.inset = 0,
                      ...) {
  tipLabels <- TipLabels(trees[[1]])
  nTip <- length(tipLabels)

  at <- attributes(trees)
  if (!missing(outgroupTips)) {
    trees <- lapply(trees, RootTree, intersect(outgroupTips, tipLabels))
  }
  trees <- RenumberTips(trees, tipLabels)
  attributes(trees) <- at

  noRogue <- trees
  attr(noRogue, "TipLabel") <- NULL
  dummyRoot <- "xxTREETOOLSxxDUMMYxxROOTxx"
  noRogue[] <- lapply(lapply(noRogue, DropTip, tip),
                      AddTip, 0, dummyRoot)
  class(noRogue) <- "multiPhylo"
  cons <- RootTree(Consensus(noRogue, p = p, check.labels = FALSE),
                   dummyRoot) # RootTree gives Preorder
  
  if (isTRUE(sort)) {
    cons <- SortTree(cons)
  }
  
  if (is.character(tip)) {
    tip <- fmatch(tip, tipLabels)
  }
  
  keptTips <- c(tipLabels[-tip], dummyRoot)
  
  cons <- RenumberTips(cons, keptTips)
  if (!isTRUE(sort)) {
    # RenumberTips does not guarantee preorder output:
    # Preorder now to avoid renumbering of edges in `DropTip()` later.
    cons <- Preorder(cons)
  }
  
  consTip <- NTip(cons)

  pole <- nTip

  #allTips <- logical(ceiling((nTip) / 8) * 8L + 2L) # Multiple of 8 for packBits
  allTips <- logical(nTip + 1L) # including dummy
  # Identify whether each leaf is in the sister clade to the rogue taxon
  # dummyRoot is, by definition, always below rogue.
  besideRogue <- .vapply(trees, function(tr) {
    edge <- AddTip(tr, 0, dummyRoot)[["edge"]]
    parent <- edge[, 1]
    child <- edge[, 2]
    rogueEdge <- fmatch(tip, child)
    rogueParent <- parent[[rogueEdge]]
    beside <- fmatch(rogueParent, child) # edge above parent of rogue tip
    splitTips <- allTips
    stopifnot(!is.na(beside))
    edgeInSplit <- DescendantEdges(parent = parent, child = child,
                                   edge = beside)
    splitKids <- child[edgeInSplit]
    splitTips[splitKids[splitKids <= nTip + 1L]] <- TRUE
    splitTips <- splitTips[-tip]

    #ret <- packBits(splitTips[-1]) # TODO can do something clever like this?
    
    # `pole` is the dummy root, so cannot have been in the rogue's sister clade
    # We thus guarantee the polarity of splitTips
    stopifnot(!splitTips[[pole]]) # if NOT, return !splitTips[-pole]
    splitTips[-pole]
  #}, raw((length(allTips) - 2L)  / 8L))
  }, logical((length(allTips) - 2L)))
  
  # Vector, each entry corresponds to a tree
  tipsBesideRogue <- colSums(besideRogue) # includes dummy root
  
  # Initialize count of trees where rogue occurs alongside a leaf
  nAtTip <- c(double(nTip - 1L),
              sum(tipsBesideRogue == nTip - 1L)) # At pole
  # Populate nAtTip
  atTip <- tipsBesideRogue == 1L
  tipMatches <- apply(besideRogue[, atTip, drop = FALSE], 2, which)
  tab <- tabulate(as.integer(tipMatches))
  nAtTip[seq_along(tab)] <- tab

  unmatchedTrees <- !(tipsBesideRogue %fin% c(0L, 1L, nTip - 1L))
  consSplits <- PolarizeSplits(as.Splits(cons, tipLabels = keptTips), pole)
  splits <- !as.logical(consSplits)
  nSplits <- nrow(splits)
  # Had previously used fmatch here, but encountered unexpected error
  # Error in fmatch: cannot take a writable DATAPTR of an ALTLIST
  #  [class: wrap_list, pkg: base]
  # Replaced with match for safety (?)
  edgeMatches <- match(data.frame(besideRogue[, unmatchedTrees, drop = FALSE]),
                       data.frame(t(splits[, -pole, drop = FALSE])))

  nAtSplit <- double(nSplits)
  tab <- tabulate(edgeMatches)
  nAtSplit[which(as.logical(tab))] <- tab[as.logical(tab)]
  
  decipher <- c(nAtTip, rep.int(NA, cons[["Nnode"]]))
  consNodes <- as.integer(names(consSplits))
  decipher[consNodes] <- nAtSplit
  nOnEdge <- decipher[cons[["edge"]][, 2]]


  nAtNode <- c(nAtTip[length(nAtTip)], double(cons[["Nnode"]] - 2L))
  unmatchedTrees[unmatchedTrees] <- is.na(edgeMatches)
  if (any(unmatchedTrees)) {
    nodeDescs <- .NodeDescendants(cons, nTip, pole)
    nNode <- nrow(nodeDescs)

    unmatchedGroup <- besideRogue[, unmatchedTrees, drop = FALSE]
    nUnmatched <- sum(unmatchedTrees)

    # nodeOverlapsGroup <- apply(unmatchedGroup, 2, function(gp) {
    #   apply(nodeDescs[gp, , drop = FALSE], 2, any)
    # })
    floors <- .apply(unmatchedGroup, 2, function(gp) {
       nodeContainsAllGroup <- apply(nodeDescs[, gp, drop = FALSE], 1, all)
       nodeContainsNonGroup <- apply(nodeDescs[, !gp, drop = FALSE], 1, any)
       nodeContainsAllGroup & nodeContainsNonGroup
    })

    # active[i, ] != within[i, ], or we'd be on an edge
    atNode <- table(apply(cbind(floors), 2,
                          function(x) 1 + length(x) - which.max(rev(x))))
    nAtNode[as.integer(names(atNode))] <- atNode

  }
  
  if (isTRUE(sort)) {
    cons <- DropTip(cons, dummyRoot, preorder = FALSE, check = FALSE)
  } else {
    stopifnot(attr(cons, "order") == "preorder")
    # We promise to return `cons` in preorder
    cons <- DropTip(cons, dummyRoot, preorder = TRUE, check = FALSE)
  }
  if (!is.null(edgeLength)) {
    cons[["edge.length"]] <- rep_len(edgeLength, dim(cons[["edge"]])[[1]])
  }
  # We've deleted the edge to the dummy root
  nOnEdge <- nOnEdge[2:(length(nOnEdge) - 1L)]
  maxVal <- max(c(nOnEdge, nAtNode)) + 1L

  legendLabels <- if (maxVal < 8) {
    maxVal:1
  } else if ((maxVal - 1) %% 5 == 0) {
    seq(maxVal, 1, length.out = 6)
  } else if ((maxVal - 1) %% 4 == 0) {
    seq(maxVal, 1, length.out = 5)
  } else if ((maxVal - 1) %% 3 == 0) {
    seq(maxVal, 1, length.out = 4)
  } else if ((maxVal - 1) %% 6 == 0) {
    seq(maxVal, 1, length.out = 7)
  } else {
    ceiling(seq(maxVal, 1, length.out = 5))
  }
  legendLabels <- c(paste(legendLabels[-length(legendLabels)], "trees"),
                    "1 tree")
  
  if (plot) {
    #pal <- c(NA, Palette(length(trees)))
    pal <- Palette(maxVal)
    plot(cons,
         edge.color = ifelse(nOnEdge > 0, pal[nOnEdge + 1L], nullCol),
         node.color = c(double(consTip - 1L),
                        ifelse(nAtNode > 0, pal[nAtNode + 1L], nullCol)),
         edge.width = ifelse(nOnEdge > 0, fat, thin),
         node.width = ifelse(c(double(consTip - 1L), nAtNode) > 0,
                             fat, thin),
         ...)
    if (legend != "none") {
      SpectrumLegend(
        legend,
        bty = "n",
        palette = pal,
        legend = legendLabels,
        inset = legend.inset
      )
    }
  }

  # Return:
  invisible(list(cons = cons, onEdge = nOnEdge, atNode = nAtNode,
                 legendLabels = legendLabels))
}

# `tree` must be in preorder
.NodeDescendants <- function(tree, nTip, pole) {
  edge <- tree[["edge"]]
  parent <- edge[, 1]
  child <- edge[, 2]
  descs <- matrix(FALSE, max(parent), nTip)
  diag(descs[seq_len(nTip), ]) <- TRUE
  
  for (i in rev(seq_along(parent))) {
    descs[parent[i], ] <- descs[parent[i], ] | descs[child[i], ]
  }
  
  # Internal nodes only
  descs <- descs[-seq_len(nTip), , drop = FALSE]
  
  # Ignore the dummy root node
  descs <- descs[-1, , drop = FALSE]
  
  flip <- descs[, pole]
  descs[flip, ] <- !descs[flip, ]
  
  # Return:
  descs[, -pole, drop = FALSE]
}

# TODO in r4.1.0 use apply(simplify = false)?
.vapply <- function(...) {
  x <- vapply(...)
  if (is.null(dim(x))) {
    x <- matrix(x, 1L)
  }

  # Return:
  x
}

.apply <- function(...) {
  x <- apply(...)
  if (is.null(dim(x))) {
    x <- matrix(x, 1L)
  }

  # Return:
  x
}

AllTreesCounted <- function(trees, rogue) {
  x <- RoguePlot(trees, rogue, plot = FALSE)
  expect_equal(length(trees), sum(x$onEdge, x$atNode))
}

ExpectedCons <- function(text) {
  Preorder(read.tree(text = text))
}

LegendLabels <- function(n) {
  paste0(n:1, " tree", c(rep("s", n - 1), ""))
}

test_that(".apply() helper function", {
  Splat <- function (...) paste(..., collapse = " ")
  mat <- matrix(1:12, 3, 4)
  expect_equal(.apply(mat, 2, paste), apply(mat, 2, paste))
  expect_equal(.apply(mat, 2, Splat), matrix(apply(mat, 2, Splat), 1, 4))
  mat1 <- mat[1, , drop = FALSE]
  expect_equal(.apply(mat1, 2, paste), matrix(apply(mat1, 2, paste), 1, 4))
})

test_that("Simple rogue plot", {
  trees <- list(read.tree(text = "(a, (b, (c, (rogue, (d, e)))));"),
                read.tree(text = "(a, (b, (c, ((d, e), rogue))));"),
                read.tree(text = "(a, (b, (c, (rogue, (d, e)))));"),
                read.tree(text = "(a, (b, (c, (d, (rogue, e)))));"))
  expect_equal(
    RoguePlot(trees, tip = "rogue", plot = FALSE),
    list(cons = ExpectedCons("(a, (b, (c, (d, e))));"),
         onEdge = c(0, 0, 0, 0, 0, 3, 0, 1),
         atNode = double(4),
         legendLabels = LegendLabels(4)
         )
  )
  expect_equal(
    RoguePlot(trees, tip = "rogue", plot = FALSE, sort = TRUE)[["cons"]],
    RenumberTips(
      SortTree(RoguePlot(trees, tip = "rogue", plot = FALSE)[["cons"]]),
      TipLabels(trees[[1]]))
  )

  skip_if_not_installed("vdiffr", "1.0")
  RoguePlotTest <- function() {
    par(mar = rep(0, 4))
    RoguePlot(trees, "rogue",
              Palette = function(...) hcl.colors(..., palette = "inferno"),
              thin = 2, fat = 4, edge.lty = 2,
              legend = "topleft", legend.inset = 0.02
              )
  }
  vdiffr::expect_doppelganger("RoguePlot(simple)", RoguePlotTest)
})

test_that("RoguePlot(sort = TRUE)", {
  trees <- c(PectinateTree(7), PectinateTree(7))
  rp <- RoguePlot(trees, "t5", sort = TRUE, plot = FALSE)
  expect_equal(rp[["cons"]],
               RenumberTips(SortTree(ConsensusWithout(trees, "t5")),
                            PectinateTree(7)))
  expect_equal(rp[["atNode"]], rep(0, 5))
  expect_equal(rp[["onEdge"]], `[<-`(double(10), 4, 2))
  
  
  trees <- list(read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),
                read.tree(text = "(rogue, (a, (b, (c, (d, (e, f))))));"),
                read.tree(text = "((rogue, a), (b, (c, (d, (e, f)))));"),
                read.tree(text = "(a, (b, ((c, d), (rogue, (e, f)))));"),
                read.tree(text = "(a, (b, ((c, (rogue, d)), (e, f))));"),
                read.tree(text = "(a, (b, (c, (d, (rogue, (e, f))))));"))[
                  c(1, 1, 1, 1:6)]
  unsorted <- RoguePlot(trees, "rogue", plot = FALSE)
  sorted <- RoguePlot(trees, "rogue", plot = FALSE, sort = TRUE)
  expect_equal(sorted$onEdge, unsorted$onEdge[c(2, 4, 7, 9, 8, 6, 5, 3, 1)])
  expect_equal(sorted$atNode, unsorted$atNode)
})

test_that("polytomy id", {
  trees <- list(read.tree(text = "(a,(((b,(c,d)),e),rogue));"),
                read.tree(text = "(a,((((c,d),e),rogue),b));"),
                read.tree(text = "(a,(rogue,(e,((b,d),c))));"))
  AllTreesCounted(trees, "rogue")

  expect_equal(
    RoguePlot(trees, "rogue", plot = FALSE),
    list(cons = ExpectedCons("(a, (b, c, d, e));"),
         onEdge = c(0, 2, 0, 0, 0, 0),
         atNode = c(0, 1),
         legendLabels = LegendLabels(3))
  )

  skip_if_not_installed("vdiffr", "1.0")
  RoguePlotTest <- function() {
    par(mar = rep(0, 4))
    RoguePlot(trees, "rogue", legend = "bottomleft")
  }
  vdiffr::expect_doppelganger("RoguePlot(poly)", RoguePlotTest)
})

# TODO test tree with rogue to left and right of balanced root
test_that("Complex rogue plot", {

  trees1 <- list(read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),  # node 9
                 read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),  # node 9
                 read.tree(text = "(a, (b, (c, (rogue, (d, (f, e))))));"),  # node 9
                 read.tree(text = "(a, (b, (c, (rogue, ((e, f), d)))));"),  # node 9
                 
                 read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),  # node 9 x 5
                 read.tree(text = "(rogue, (a, (b, (c, (d, (e, f))))));"),  # node 7 x 1
                 read.tree(text = "(a, (rogue, (b, (c, (d, (e, f))))));"),  # edge 2 x 1
                 read.tree(text = "((rogue, a), (b, (c, (d, (e, f)))));"),  # edge 1
                 
                 read.tree(text = "((rogue, a), (b, (c, (d, (e, f)))));"),  # edge 1 x 2
                 read.tree(text = "(a, (b, ((c, d), (rogue, (f, e)))));"),  # edge 7
                 read.tree(text = "(a, (b, (((rogue, d), c), (e, f))));"),  # edge 6 x 1
                 read.tree(text = "(a, (b, (c, (d, (rogue, (e, f))))));"))  # edge 7 x 2
  expect_equal(
    RoguePlot(trees = trees1, tip = "rogue", plot = FALSE),
    list(cons = ExpectedCons("(a, (b, (c, d, (e, f))));"),
         onEdge = c(2, 1, 0, 0, 0, 1, 2, 0, 0),
         atNode = c(1, 0, 5, 0),
         legendLabels = LegendLabels(6))
    )

  expect_equal(
    RoguePlot(trees = trees1, tip = "rogue", plot = FALSE, sort = TRUE),
    list(cons = RenumberTips(SortTree(ExpectedCons("(a, (b, (c, d, (e, f))));")), trees1),
         onEdge = c(2, 1, 0, 0, 0, 1, 2, 0, 0)[c(2, 4, 7, 9, 8, 6, 5, 3, 1)],
         atNode = c(1, 0, 5, 0),
         legendLabels = LegendLabels(6))
    )

  expectedCons <- Preorder(
    RenumberTips(read.tree(text = "(f, (e, (d, c, (b, a))));"), letters[1:6])
  )
  
  expect_equal(
    RoguePlot(trees1, "rogue", outgroupTips = "f", plot = FALSE),
    list(cons = expectedCons,
         onEdge = c(0, 2, 0, 4, 0, 0, 1, 0, 0),
         atNode = c(0, 0, 5, 0),
         legendLabels = LegendLabels(6)
         )
  )

  trees2 <- list(read.tree(text = "(a, (b, (rogue, ((d, c), (e, f)))));"),
                 read.tree(text = "(a, (b, ((rogue, (d, c)), (e, f))));"),
                 read.tree(text = "(a, (b, ((rogue, (d, c)), (e, f))));"),
                 read.tree(text = "(a, ((b, (c, d)), (rogue, (e, f))));"),
                 read.tree(text = "(a, (b, ((c, d), (rogue, (e, f)))));"),
                 read.tree(text = "(a, (b, ((c, d), (rogue, (e, f)))));"))
  expected <- list(cons = Preorder(read.tree(text = "(a, (b, (c, d), (e, f)));")),
                   onEdge = c(0, 0, 0, 2, 0, 0, 3, 0, 0),
                   atNode = c(0, 1, 0, 0),
                   legendLabels = LegendLabels(5))
  actual <- RoguePlot(trees = trees2, tip = "rogue", plot = FALSE)
  expect_equal(names(actual), names(expected))
  expect_true(all.equal(actual$cons, expected$cons))
  expect_equal(actual$onEdge, expected$onEdge)
  expect_equal(actual$atNode, expected$atNode)

  trees3 <- lapply(c(9, 10, 13), AddTip, tree = BalancedTree(8), label = "rogue")
  AllTreesCounted(trees3, "rogue")

  skip_if_not_installed("vdiffr", "1.0")
  RoguePlotTest <- function() {
    par(mar = rep(0, 4))
    RoguePlot(trees1, "rogue",
              Palette = function(...) hcl.colors(..., palette = "inferno"),
              edgeLen = 1, thin = 2, fat = 4,
              legend = "bottomright"
              )
  }
  vdiffr::expect_doppelganger("RoguePlot(trees1)", RoguePlotTest)

  RoguePlotTest <- function() {
    par(mar = rep(0, 4))
    RoguePlot(trees2, "rogue", edgeLength = 1:7, thin = 2, fat = 4)
  }
  vdiffr::expect_doppelganger("RoguePlot(trees2)", RoguePlotTest)
})

#' Renumber a tree's nodes and tips
#'
#' `Renumber()` numbers the nodes and tips in a tree to conform with the
#' `phylo` standards.
#'
#' The \pkg{ape} class `phylo` is not formally defined, but expects trees' internal
#' representation to conform to certain principles: for example, nodes should
#' be numbered sequentially, with values increasing away from the root.
#'
#' `Renumber()` attempts to reformat any tree into a representation that will
#' not cause \pkg{ape} functions to produce unwanted results or to crash R.
#'
#' @template treeParam
#'
#' @examples
#' tree <- RandomTree(letters[1:10])
#' Renumber(tree)
#'
#' @return `Renumber()` returns a tree of class `phylo`, numbered in a
#' [Cladewise] fashion consistent with the expectations of \pkg{ape} functions.
#'
#' @seealso `Preorder()` provides a faster and simpler alternative, but also
#' rotates nodes.
#'
#' @template MRS
#' @family tree manipulation
#' @export
Renumber <- function(tree) {
  tree   <- ApePostorder(tree)
  edge   <- tree[["edge"]]
  nTip   <- NTip(tree)
  parent <- edge[, 1L]
  child  <- edge[, 2L]
  NODES  <- child > nTip
  TIPS   <- !NODES
  nNode  <- sum(NODES) + 1 # Root node has no edge leading to it, so add 1
  nodeLb <- tree[["node.label"]]

  tip <- child[TIPS]
  tree[["tip.label"]] <- tree[["tip.label"]][tip]
  child[TIPS] <- seq_len(nTip)

  old.node.number <- unique(parent)
  new.node.number <- rev(nTip + seq_along(old.node.number))
  renumbering.schema <- integer(nNode)
  renumbering.schema[old.node.number - nTip] <- new.node.number
  child[NODES] <- renumbering.schema[child[NODES] - nTip]
  nodeseq <- seq_len(nNode) * 2L
  parent <- renumbering.schema[parent - nTip]
  if (!is.null(nodeLb)) {
    tree[["node.label"]][new.node.number - nTip] <- 
      nodeLb[old.node.number - nTip]
  }

  tree[["edge"]][, 1] <- parent
  tree[["edge"]][, 2] <- child
  Cladewise(tree)
}

#' Generate trivial trees
#'
#' `SingleTaxonTree()` creates a phylogenetic "tree" that contains a single
#' taxon. 
#' `ZeroTaxonTree()` creates an empty `phylo` object with zero leaves or edges.
#'
#' @param label a character vector specifying the label of the tip.
#' @param lengths Numeric vector of edge lengths, or a function that returns
#' such a vector when passed the number of edges as its argument (e.g. `runif`).
#' @return `SingleTaxonTree()` returns a \code{phylo} object containing a single
#' tip with the specified label.
#'
#' @examples
#' SingleTaxonTree("Homo_sapiens")
#' plot(SingleTaxonTree("root") + BalancedTree(4))
#'
#' @keywords tree
#' @family tree manipulation
#' @family tree generation functions
#' @name TrivialTree
NULL

#' @rdname TrivialTree
#' @export
SingleTaxonTree <- function(label = "t1", lengths = NULL) {
  tr <- list(edge = matrix(c(2L, 1L), 1, 2), tip.label = label, Nnode = 1L)
  if (!is.null(lengths)) {
    tr[["edge.length"]] <- .EdgeLengths(lengths, 1L)[[1L]]
  }
  # Return:
  structure(tr, class = "phylo", order = "preorder")
}

#' @rdname TrivialTree
#' @return `ZeroTaxonTree()` returns an empty \code{phylo} object.
#' @examples
#' ZeroTaxonTree()
#' @export
ZeroTaxonTree <- function() {
  structure(list(
    # Order is consistent with ape::read.tree (but not ape::rtree...)
    edge = structure(numeric(0), dim = c(0L, 2L)),
    Nnode = 0, tip.label = character(0)
  ), class = "phylo", order = "preorder")
}

#' Extract a subtree
#'
#' `Subtree()` safely extracts a clade from a phylogenetic tree.
#'
#' Modified from the \pkg{ape} function \code{\link[ape]{extract.clade}}, which
#' sometimes behaves unpredictably.
#' Unlike extract.clade, this function supports the extraction of "clades"
#' that constitute a single tip.
#'
#' @template preorderTreeParam
#' @param node The number of the node at the base of the clade to be extracted.
#'
#' @return `Subtree()` returns a tree of class \code{phylo} that represents a
#' clade extracted from the original tree.
#'
#' @examples
#' tree <- Preorder(BalancedTree(8))
#' plot(tree)
#' ape::nodelabels()
#' ape::nodelabels(13, 13, bg="yellow")
#'
#' plot(Subtree(tree, 13))
#'
#' @template MRS
#' @family tree manipulation
#' @export
Subtree <- function(tree, node) {
  if (is.null(treeOrder <- attr(tree, "order")) || treeOrder != "preorder") {
    stop("Tree must be in preorder")
  }
  tipLabel <- tree[["tip.label"]]
  nTip <- length(tipLabel)
  if (node <= nTip) {
    return(SingleTaxonTree(tipLabel[node]))
  }
  if (node == nTip + 1L) {
    return(tree)
  }

  edge <- tree[["edge"]]
  parent <- edge[, 1]
  child <- edge[, 2]
  subtreeParentEdge <- match(node, child)
  keepEdge <- DescendantEdges(parent, child, edge = subtreeParentEdge)
  keepEdge[subtreeParentEdge] <- FALSE

  edge <- edge[keepEdge, ]
  edge1 <- edge[, 1]
  edge2 <- edge[, 2]

  isTip <- edge2 <= nTip
  tips  <- edge2[isTip]
  new.nTip <- length(tips)
  name <- character(new.nTip)
  # method="radix" typically a few % faster than "auto"
  tipOrder <- order(tips, method = "radix")
  name[tipOrder] <- tipLabel[tips]
  edge2[isTip] <- tipOrder

  ## renumber nodes:
  nodeAdjust <- new.nTip + 1 - node
  keptLabels <- c(node, edge2[!isTip]) - nTip
  edge2[!isTip] <- edge2[!isTip] + nodeAdjust
  edge[, 1] <- edge1 + nodeAdjust
  edge[, 2] <- edge2

  ret <- structure(list(
    tip.label = name,
    Nnode = dim(edge)[[1]] - new.nTip + 1L,
    edge = edge
  ), class = "phylo", order = "preorder")
  if (!is.null(tree[["node.label"]])) {
    ret[["node.label"]] <- tree[["node.label"]][keptLabels]
  }
  
  # Return:
  ret
}

#' List ancestors
#'
#' `ListAncestors()` reports all ancestors of a given node.
#'
#' Note that if `node = NULL`, the tree's edges must be listed such that each
#' internal node (except the root) is listed as a child before it is listed
#' as a parent, i.e. its index in `child` is less than its index in `parent`.
#' This will be true of trees listed in [Preorder].
#'
#'
#' @template treeParent
#' @template treeChild
#' @param node Integer giving the index of the node or tip whose ancestors are
#'  required, or `NULL` to return ancestors of all nodes.
#'
#' @return
#' If `node = NULL`, `ListAncestors()` returns a list. Each entry _i_ contains
#' a vector containing, in order, the nodes encountered when traversing the tree
#' from node _i_ to the root node.
#' The last entry of each member of the list is therefore the root node,
#' with the exception of the entry for the root node itself, which is a
#' zero-length integer.
#'
#' If `node` is an integer, `ListAncestors()` returns a vector of the numbers of
#' the nodes ancestral to the given \code{node}, including the root node.
#' @template MRS
#'
#' @seealso
#' Implemented less efficiently in \code{phangorn:::Ancestors}, on which this
#' code is based.
#'
#' @examples
#' tree <- PectinateTree(5)
#' edge <- tree[["edge"]]
#'
#' # Identify desired node with:
#' plot(tree)
#' nodelabels()
#' tiplabels()
#'
#' # Ancestors of specific nodes:
#' ListAncestors(edge[, 1], edge[, 2], 4L)
#' ListAncestors(edge[, 1], edge[, 2], 8L)
#'
#' # Ancestors of each node, if tree numbering system is uncertain:
#' lapply(seq_len(max(edge)), ListAncestors,
#'        parent = edge[, 1], child = edge[, 2])
#'
#' # Ancestors of each node, if tree is in preorder:
#' ListAncestors(edge[, 1], edge[, 2])
#'
#' @family tree navigation
#' @export
ListAncestors <- function(parent, child, node = NULL) {
  if (is.null(node)) {
    AllAncestors(parent, child)
  } else {
    pvector <- integer(max(parent))
    pvector[child] <- parent
    anc <- function(pvector, node) {
      res <- integer(0)
      repeat {
        anc <- pvector[node]
        if (anc == 0)
          break
        res <- c(res, anc)
        node <- anc
      }
      res
    }

    # Return:
    anc(pvector, node)
  }
}

#' @describeIn ListAncestors Alias for `ListAncestors(node = NULL)`.
#'
#' @examples
#' # Alias:
#' AllAncestors(edge[, 1], edge[, 2])
#'
#' @family tree navigation
#' @export
AllAncestors <- function(parent, child) {
  res <- lapply(integer(max(parent)), integer)
  for (i in seq_along(parent)) {
    pa <- parent[i]
    res[[child[i]]] <- c(pa, res[[pa]])
  }
  res
}

#' Clade sizes
#'
#' `CladeSizes()` reports the number of nodes in each clade in a tree.
#'
#' @template treeParam
#' @param internal Logical specifying whether internal nodes should be counted
#' towards the size of each clade.
#' @param nodes Integer specifying indices of nodes at the base of clades
#' whose sizes should be returned.
#' If unspecified, counts will be provided for all nodes (including leaves).
#'
#' @return `CladeSizes()` returns the number of nodes (including leaves) that
#' are descended from each node, not including the node itself.
#'
#' @examples
#' tree <- BalancedTree(6)
#' plot(tree)
#' ape::nodelabels()
#' CladeSizes(tree, nodes = c(1, 8, 9))
#'
#' @family tree navigation
#' @export
CladeSizes <- function(tree, internal = FALSE, nodes = NULL) {
  if (length(internal) > 1 || !is.logical(internal)) {
    warning("`internal` should be a single logical value.")
    internal <- isTRUE(internal)
  }

  nTip <- NTip(tree)
  edge <- tree[["edge"]]

  size <- c(rep.int(1L, nTip), rep.int(internal, max(edge[, 1]) - nTip))
  for (i in postorder_order(tree[["edge"]])) {
    edgeI <- edge[i, ]
    parent <- edgeI[1]
    child <- edgeI[2]
    size[parent] <- size[parent] + size[child]
  }


  # Return:
  (if (is.null(nodes)) size else size[nodes]) - internal
}

test_that("TipTimedTree() example works", {
  expect_equal(
    TipTimedTree(BalancedTree(6), tipAge = 1:6, minEdge = 2)[["edge.length"]],
    c(5, 2, 3, 2, 3, 2, 2, 3, 2, 3)
  )
  expect_equal(
    TipTimedTree(Postorder(BalancedTree(6)), tipAge = 1:6, minEdge = .5)$edge.l,
    c(0.5, 1.5, 1, 0.5, 0.5, 1.5, 1, 0.5, 0.5, 3.5)
  )
  expect_error(TipTimedTree(BalancedTree(5), 1:10), "one age per leaf")
  expect_error(TipTimedTree(BalancedTree(5), double(0)), "one age per leaf")
  expect_warning(expect_equal(TipTimedTree(SingleTaxonTree(1), 1, 100)$edge.l,
                              100), "does not contain multiple edges")
})

nasty <- structure(list(edge = structure(
  c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
    5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
  .Dim = c(12, 2)),
  Nnode = 5L,
  tip.label = letters[1:8]),
  class = "phylo") # Danger: Do not plot!

test_that("Subtree() works", {
  expect_error(Subtree(read.tree(text = "((a,b),(c,d));", 7)),
               " be in preorder")
  t4 <- Subtree(Preorder(BalancedTree(8)), 10)
  expect_true(all.equal(BalancedTree(4), t4))
  expect_true(all.equal(BalancedTree(4), Subtree(t4, 5)))
  expect_true(all.equal(SingleTaxonTree("t1"), Subtree(t4, 1)))
})

test_that("Subtree() handles node labels", {
  bal8 <- ape::makeNodeLabel(RootTree(BalancedTree(8), 1), prefix = "Node ")
  expect_equal(Subtree(bal8, 5 + 8)[["node.label"]],
               paste("Node", 5:7))
})


test_that("ListAncestors() works", {
  edge <- nasty$edge
  expect_equal(c(10L, 12L), ListAncestors(edge[, 1], edge[, 2], 11))
  tr <- PectinateTree(4)
  expect_equal(list(5, 6:5, 7:5, 7:5, integer(0), 5, 6:5),
               ListAncestors(tr$edge[, 1], tr$edge[, 2]))
  expect_equal(integer(0), ListAncestors(tr$edge[, 1], tr$edge[, 2], 5))
})

test_that("CladeSizes() works", {
  #plot(Preorder(nasty)); nodelabels(c(12, 10, 13, 11, 9)); tiplabels(1:8)
  #edgelabels(c(2, 3, 6, 4, 8, 10, 9, 5, 7, 1, 12, 11))
  
  expect_equal(c(3, 8 + 4, 3 + 1, 7 + 3, 2),
               CladeSizes(nasty, internal = TRUE, 13:9))
  expect_equal(c(3, 8, 3, 7, 2), CladeSizes(nasty, internal = FALSE, 13:9))
  
  # Misspecification:
  expect_warning(expect_equal(CladeSizes(BalancedTree(7), internal = FALSE),
                              CladeSizes(BalancedTree(7), internal = 8:9)))
})

test_that("Tree list entries are in consistent order", {
  expected <- names(ape::read.tree(text = "(a, b);"))
  expect_equal(names(ZeroTaxonTree()), expected)
  expect_equal(names(BalancedTree(2)), expected)
  expect_equal(names(PectinateTree(2)), expected)
  expect_equal(names(StarTree(2)), expected)
  expect_equal(names(RandomTree(3)), expected)
  expect_equal(names(RootTree(BalancedTree(4))), expected)
  expect_equal(names(RootTree(KeepTip(BalancedTree(14), 1:4))), expected)
})

#' Construct consensus trees
#'
#' `Consensus()` calculates the majority-rule or strict consensus of a set of
#' trees, using the cluster-table approach of \insertCite{Day1985}{TreeTools}.
#'
#' The strict consensus (`p = 1`) compares the clusters of the first tree
#' against every other tree in linear time.  The majority-rule and threshold
#' consensus (`0.5 <= p < 1`) instead count the frequency of every split across
#' all trees in a single pass and retain those occurring in a proportion `p` or
#' more of trees; this runs in time linear in the number of trees, after
#' \insertCite{Jansson2016}{TreeTools}.  By default the count uses
#' a 128-bit hash, whose results are exact with overwhelming probability; set
#' `hash = FALSE` for a slower but guaranteed-exact count.
#'
#' @param trees List of trees, optionally of class `multiPhylo`.
#' @param p Proportion of trees that must contain a split for it to be reported
#' in the consensus.  `p = 0.5` gives the majority-rule consensus; `p = 1` (the
#' default) gives the strict consensus.
#' @param check.labels Logical specifying whether to check that all trees have
#' identical labels.  Defaults to `TRUE`, which is slower.
#' @param hash Logical; if `TRUE` (default), majority/threshold consensus
#' counts splits using 128-bit hashing, which is exact with overwhelming
#' probability (a collision conflating two distinct splits is vanishingly
#' unlikely).  Set `hash = FALSE` for a slower but guaranteed-exact count.
#' Ignored when `p = 1`, which is always exact.
#'
#' @return `Consensus()` returns an object of class `phylo`, rooted as in the
#' first entry of `trees`.
#' @examples
#' Consensus(as.phylo(0:2, 8))
#' @seealso
#' * [\pkg{ConsTree}](https://constree.github.io/) implements
#' other consensus tree algorithms.
#' 
#' * [\pkg{Rogue}](https://ms609.github.io/Rogue/) increases the resolution of
#' consensus trees by dropping wildcard taxa.
#' 
#' * `TreeDist::ConsensusInfo()` calculates the information content of a
#' consensus tree.
#' @template MRS
#' @family consensus tree functions
#' @family tree characterization functions
#' @references
#' \insertAllCited{}
#' @export
Consensus <- function(trees, p = 1, check.labels = TRUE, hash = TRUE) {
  if (length(trees) == 1L) {
    return(trees[[1]])
  }
  if (inherits(trees, "phylo")) {
    return(trees)
  }
  if (!is.list(trees) || is.data.frame(trees)) {
    stop("Expecting `trees` to be a list.")
  }
  
  # Remove irrelevant metadata so we don't waste time processing it
  trees <- lapply(c(trees), function(tr) {
    tr[["edge.length"]] <- NULL
    tr[["node.label"]] <- NULL
    tr
  })
  
  repeat {
    nTip <- NTip(trees)
    if (length(unique(nTip)) > 1) {
      warning("Tree sizes differ; removing leaves not in smallest.")
      trees <- lapply(trees, KeepTip, trees[[which.min(nTip)]][["tip.label"]])
    } else {
      nTip <- nTip[[1]]
      break
    }
  }
  if (nTip < 4L) {
    return(trees[[1]])
  }
  if (check.labels) {
    trees <- RenumberTips(trees, trees[[1]])
  }
  if (p < 0.5 || p > 1) {
    stop("`p` must be between 0.5 and 1.")
  }
  trees <- Preorder(trees) # Per #168; could be dispensed with with further
                           # investigation of consensus_tree
  tree1 <- trees[[1]] # Must be in Preorder for DescendantEdges()
  edg <- tree1[["edge"]]
  root <- edg[DescendantEdges(edg[, 1], edg[, 2], edge = 1), 2]
  root <- root[root <= NTip(tree1)]

  # Return:
  RootTree(.PreorderTree(
    edge = splits_to_edge(consensus_tree(trees, p, exact = !isTRUE(hash)), nTip),
    tip.label = TipLabels(trees[[1]])
  ), root)
}

#' Reduced consensus, omitting specified taxa
#'
#' `ConsensusWithout()` displays a consensus plot with specified taxa excluded,
#' which can be a useful way to increase the resolution of a consensus tree
#' when a few wildcard taxa obscure a consistent set of relationships.
#' `MarkMissing()` adds missing taxa as loose leaves on the plot.
#'
#' @param trees A list of phylogenetic trees, of class `multiPhylo` or `list`.
#' @param tip A character vector specifying the names (or numbers) of tips to
#' drop (using `ape::drop.tip()`).
#'
#' @return `ConsensusWithout()` returns a consensus tree (of class `phylo`)
#' without the excluded taxa.
#'
#' @examples
#' oldPar <- par(mfrow = c(1, 2), mar = rep(0.5, 4))
#'
#' # Two trees differing only in placement of tip 2:
#' trees <- as.phylo(c(0, 53), 6)
#' plot(trees[[1]])
#' plot(trees[[2]])
#'
#' # Strict consensus (left panel) lacks resolution:
#' plot(ape::consensus(trees))
#'
#' # But omitting tip two (right panel) reveals shared structure in common:
#' plot(ConsensusWithout(trees, "t2"))
#' MarkMissing("t2")
#'
#' par(oldPar)
#' @family tree manipulation
#' @family tree properties
#' @family consensus tree functions
#'
#' @template MRS
#' @export
ConsensusWithout <- function(trees, tip = character(0), ...) {
  UseMethod("ConsensusWithout")
}

#' @rdname ConsensusWithout
#' @export
ConsensusWithout.phylo <- function(trees, tip = character(0), ...) {
  DropTip(trees, tip = tip)
}

#' @export
ConsensusWithout.NULL <- function(trees, tip, ...) NULL

#' @rdname ConsensusWithout
#' @export
ConsensusWithout.multiPhylo <- function(trees, tip = character(0), ...) {
  Consensus(lapply(trees, DropTip, tip = tip), ...)
}

#' @rdname ConsensusWithout
#' @export
ConsensusWithout.list <- ConsensusWithout.multiPhylo

#' @rdname ConsensusWithout
#' @param position Where to plot the missing taxa.
#' See [`legend()`] for options.
#' @param \dots Additional parameters to pass on to [`ape::consensus()`] or
#' [`legend()`].
#' @return `MarkMissing()` provides a null return, after plotting the specified
#' `tip`s as a legend.
#' @importFrom graphics legend
#' @export
MarkMissing <- function(tip, position = "bottomleft", ...) {                   # nocov start
  if (length(tip) > 0) {
    legend(position, legend = gsub("_", " ", tip, fixed = TRUE),
           lwd = 1, lty = 2, bty = "n", ...)
  }
}                                                                               # nocov end

#' Reweight phylogenetic characters
#' 
#' `Reweight()` allows the weights of specific characters in phylogenetic
#' datasets to be arbitrarily adjusted.
#' 
#' This functionality should be employed with care.
#' The underlying principle of parsimony is that all evolutionary steps are
#' equivalent.
#' Setting different weights to different characters is at odds with that
#' principle, so analysis of a re-weighted matrix using a parsimony-based
#' framework is arguably no longer parsimony analysis; on the most permissive
#' view, the criteria used to determine a weighting scheme will always
#' be arbitrary.
#' 
#' It can be useful to relax the criterion that all evolutionary steps are
#' equivalent -- for example, implied weighting 
#' \insertCite{Goloboff1997}{TreeTools} typically recovers better trees than
#' equal-weights parsimony \insertCite{Smith2019}{TreeTools}.
#' This said, assigning different weights to different characters tacitly
#' imposes a model of evolution that differs from that implicit in equal-weights
#' parsimony.  Whereas probabilistic models can be evaluated by various methods
#' (e.g. fit, marginal likelihood, posterior predictive power), there are no
#' principled methods of comparing different models under a parsimony framework.
#' 
#' As such, `Reweight()` is likely to be useful for a narrow set of uses.
#' Examples may include: 
#'  - informal robustness testing, to explore whether certain characters are 
#'  more or less influential on the resulting tree;
#'  - Imposing constraints on a dataset, by adding each constraint as a column
#'  in a dataset whose weight exceeds the total amount of data.
#' 
#' 
#' @param dataset A phylogenetic data matrix of \pkg{phangorn} class `phyDat`,
#' or as a matrix in the format produced by [`PhyDatToMatrix()`].
#' @param weights Unnamed integer vector specifying desired weight of each
#' character in turn; or named integer vector specifying weights of each
#' character; unnamed entries will be assigned weight 1.
#' 
#' @return `Reweight()` returns `dataset` after adjusting the weights of
#' the specified characters.
#' For a matrix, this is attained by repeating each column the `weights` times.
#' For a `phyDat` object, the "weight" attribute will be modified.
#' 
#' @references \insertAllCited{}
#' @examples
#' mat <- rbind(a = c(0, 2, 0), b = c(0, 2, 0), c = c(1, 3, 0), d = c(1, 3, 0))
#' dat <- MatrixToPhyDat(mat)
#' 
#'  # Set character 1 to weight 1, character 2 to weight 2; omit character 3
#' Reweight(mat, c(1, 2, 0))
#' # Equivalently:
#' Reweight(dat, c("3" = 0, "2" = 2))
#' @template MRS
#' @family phylogenetic matrix conversion functions
#' @export
Reweight <- function(dataset, weights) {
  UseMethod("Reweight")
}

.MakeWeights <- function(weights, nChar) {
  if (!is.numeric(weights)) {
    stop("`weights` must be a numeric vector")
  }
  
  keys <- as.integer(names(weights))
  if (!is.null(keys) && length(keys) > 0) {
    if (any(keys < 1 | keys > nChar)) {
      stop("Indices in `names(weights)` must be between 1 and ", nChar, 
           ". Found ", paste(keys[keys < 1 | keys > nChar], collapse = ", "))
    }
    w <- rep(1, nChar)
    w[as.integer(keys)] <- weights
    weights <- w
  }
  
  if (length(weights) != nChar) {
    stop("Length of `weights` (", length(weights),
         ") must match number of characters (", nChar, ")")
  }
  
  # Return:
  weights
}

#' @export
Reweight.matrix <- function(dataset, weights) {
  nChar <- dim(dataset)[[2]]
  weights <- .MakeWeights(weights, nChar)
  
  dataset <- dataset[, rep(seq_len(nChar), times = weights), drop = FALSE]
  `colnames<-`(dataset,
               paste0(colnames(dataset), "_", 
                      unlist(sapply(weights, seq_len, USE.NAMES = FALSE),
                             recursive = FALSE, use.names = FALSE)))
}

#' @export
Reweight.phyDat <- function(dataset, weights) {
  nChar <- length(attr(dataset, "index"))
  weights <- .MakeWeights(weights, nChar)
  wAtt <- double(attr(dataset, "nr"))
  index <- attr(dataset, "index")
  for (i in seq_along(weights)) {
    wAtt[index[[i]]] <- wAtt[index[[i]]] + weights[[i]]
  }
  attr(dataset, "weight") <- wAtt
  
  # Return:
  dataset
}

.onUnload <- function(libpath) {
  library.dynam.unload("TreeTools", libpath)
}

## Reminders when releasing for CRAN
release_questions <- function() {
  c(
    "Is the code free of #TODOs?",
    "Have you checked the Vignettes for sanity?"
  )
}


# Additional steps:
#
# Propagate changes in README.md to R/TreeTools-package.R


# Additional tests:
#
# spell_check()
# pkgdown::build_reference_index()
#
# run_examples()
# build_vignettes()
#
# devtools::check_win_devel(quiet = TRUE); rhub::check_for_cran()
# Check valgrind results on Github Actions
# revdepcheck::revdep_check()
#
# codemeta::write_codemeta()
#
# tools::resaveRdaFiles("R", compress="auto") - is default bzip2 the optimal?
# tools::checkRdaFiles("R") - set optimal compression in `data-raw`

#' Ancestral edge
#'
#' @param edge Number of an edge
#' @template treeParent
#' @template treeChild
#' @return `AncestorEdge` returns a logical vector identifying whether each edge
#' is the immediate ancestor of the given edge.
#' @examples
#' tree <- BalancedTree(6)
#' parent <- tree$edge[, 1]
#' child <- tree$edge[, 2]
#' plot(tree)
#' ape::edgelabels()
#' AncestorEdge(5, parent, child)
#' which(AncestorEdge(5, parent, child))
#'
#' @keywords internal
#' @family tree navigation
#' @export
AncestorEdge <- function(edge, parent, child) child == parent[edge]

#' Ancestors of an edge
#'
#' Quickly identify edges that are "ancestral" to a particular edge in a tree.
#'
#' @param edge Integer specifying the number of the edge whose child edges
#' should be returned.
#' @template treeParent
#' @template treeChild
#' @param stopAt Integer or logical vector specifying the edge(s) at which to
#' terminate the search; defaults to the edges with the smallest parent,
#' which will be the root edges if nodes are numbered [Cladewise] or in
#' [Preorder].
#'
#' @return `EdgeAncestry()` returns a logical vector stating whether each edge
#' in turn is a descendant of the specified edge.
#'
#' @examples
#' tree <- PectinateTree(6)
#' plot(tree)
#' ape::edgelabels()
#' parent <- tree$edge[, 1]
#' child <- tree$edge[, 2]
#' EdgeAncestry(7, parent, child)
#' which(EdgeAncestry(7, parent, child, stopAt = 4))
#'
#' @template MRS
#' @family tree navigation
#' @export
EdgeAncestry <- function(edge, parent, child,
                         stopAt = (parent == min(parent))) {
  ret <- edge <- AncestorEdge(edge, parent, child)
  if (any(ret)) repeat {
    if (any(ret[stopAt])) return(ret)
    ret[edge <- AncestorEdge(edge, parent, child)] <- TRUE
  }
  # Return:
  ret
}

#' Most recent common ancestor
#'
#' `MRCA()` calculates the last common ancestor of specified nodes.
#'
#' `MRCA()` requires that node values within a tree increase away from the root,
#' which will be true of trees listed in `Preorder`.
#' No warnings will be given if trees do not fulfil this requirement.
#'
#' @param x1,x2 Integer specifying index of leaves or nodes whose most
#' recent common ancestor should be found.
#' @param ancestors List of ancestors for each node in a tree. Perhaps
#' produced by [`ListAncestors()`].
#'
#' @return `MRCA()` returns an integer specifying the node number of the last
#' common ancestor of `x1` and `x2`.
#'
#' @family tree navigation
#' @template MRS
#'
#' @examples
#' tree <- BalancedTree(7)
#'
#' # Verify that node numbering increases away from root
#' plot(tree)
#' nodelabels()
#'
#' # ListAncestors expects a tree in Preorder
#' tree <- Preorder(tree)
#' edge <- tree$edge
#' ancestors <- ListAncestors(edge[, 1], edge[, 2])
#' MRCA(1, 4, ancestors)
#'
#' # If a tree must be in postorder, use:
#' tree <- Postorder(tree)
#' edge <- tree$edge
#' ancestors <- lapply(seq_len(max(edge)), ListAncestors,
#'                     parent = edge[, 1], child = edge[, 2])

#'
#' @export
MRCA <- function(x1, x2, ancestors) {
  if (x1 == x2) {
    x1
  } else {
    anc1 <- ancestors[[x1]]
    anc2 <- ancestors[[x2]]
    max(intersect(anc1, anc2))
  }
}

test_that("EdgeAncestry() works", {
  tree <- BalancedTree(10)
  edge <- tree$edge
  parent <- edge[, 1]
  child <- edge[, 2]
  expect_equal(which(EdgeAncestry(4L, parent, child)), 1:3)
  expect_equal(which(EdgeAncestry(2, parent, child)), 1)
  expect_equal(which(EdgeAncestry(1, parent, child)), integer(0))
  expect_equal(EdgeAncestry(10, parent, child), logical(18))
})

test_that("MRCA() works", {
  bal7 <- BalancedTree(7)
  allAnc <- AllAncestors(bal7$edge[, 1], bal7$edge[, 2])
  expect_equal(9, MRCA(1, 4, allAnc))
  expect_equal(8, MRCA(1, 6, allAnc))
  expect_equal(8, MRCA(1, 7, allAnc))
  expect_equal(1, MRCA(1, 1, allAnc))
  expect_equal(9, MRCA(1, 11, allAnc))
})

#' Calculate length of paths between each pair of vertices within tree
#' 
#' Given a weighted rooted tree `tree`, `PathLengths()` returns the distance
#' from each vertex to each of its descendant vertices.
#' 
#' @param tree Original tree of class `phylo`, in [`Preorder`].
#' @param fullMatrix Logical specifying return format; see "value" section`.
#' @param use.na Logical specifying whether to set non-existent paths to `NA`,
#' or to leave uninitialized.  Set to `FALSE` to maximize performance.
#' @return If `fullMatrix = TRUE`, `PathLengths()` returns a square matrix in
#' which entry `[i, j]` denotes the distance from internal node `i` to the
#' descendant vertex `j`. 
#' Vertex pairs without a continuous directed path are denoted `NA` if `use.na`
#' is `TRUE`.
#' If `fullMatrix = FALSE`, `PathLengths()` returns a `data.frame` with three
#' columns: `start` lists the deepest node in each path (i.e. that closest
#' to the root); `end` lists the shallowest node (i.e. that closest to a leaf);
#' `length` lists the total length of that path.
#' @examples
#' tree <- rtree(6)
#' plot(tree)
#' add.scale.bar()
#' nodelabels()
#' tiplabels()
#' PathLengths(tree)
#' @template MRS
#' @export
#' @family tree properties
PathLengths <- function(tree, fullMatrix = FALSE, use.na = TRUE) {
  if (!inherits(tree, "phylo")) {
    stop("`tree` must be an object of class `phylo`.")
  }
  weights <- tree[["edge.length"]]
  if (is.null(weights)) {
    weights <- rep_len(1L, dim(tree[["edge"]])[1])
  }
  mat <- path_lengths(tree[["edge"]], weights,
                      init_nas = isFALSE(fullMatrix) || isTRUE(use.na))
  
  # Return:
  if (fullMatrix) {
    mat
  } else {
    paths <- !is.na(mat)
    data.frame(start = row(paths)[paths],
               end = col(paths)[paths],
               length = mat[paths])
  }
}

#' Ratio of external:internal edge length
#' 
#' Reports the ratio of tree length associated with external edges (i.e.
#' edges whose child is a leaf) and internal edges.
#' Where tree length is dominated by internal edges, variation between tips
#' is dominantly controlled by phylogenetic history.
#' 
#' @param x A tree of class \code{\link[ape:read.tree]{phylo}}.
#' @returns `EdgeRatio()` returns a numeric specifying the ratio of external
#' to internal edge length (> 1 means the length of a tree is predominantly
#' in external edges), with attributes `external`, `internal`, and `total`
#' specifying the total length associated with edges of that nature.
#' @template MRS
#' @family tree properties
#' @export
EdgeRatio <- function(x) UseMethod("EdgeRatio")
  
#' @rdname EdgeRatio
#' @export
EdgeRatio.phylo <- function(x) {
  el <- x[["edge.length"]]
  if (is.null(el)) {
    warning("Edge lengths not specified")
    return(NA_real_)
  }
  ed <- x[["edge"]]
  nTip <- NTip(x)
  external <- ed[, 2] <= nTip
  exLen <- sum(el[external])
  inLen <- sum(el[!external])
  structure(exLen / inLen,
            external = exLen,
            internal = inLen,
            total = sum(exLen, inLen))
}

test_that("EdgeRatio() works", {
  expect_warning(expect_equal(EdgeRatio(BalancedTree(5)), NA_real_),
                 "[Ee]dge lengths not spec")
  
  ex2 <- BalancedTree(4)
  ex2[["edge.length"]] <- c(1, 2, 2, 1, 2, 2)
  expect_equal(EdgeRatio(ex2),
               structure(8/2, external = 8, internal = 2, total = 10))
})

#' Root or unroot a phylogenetic tree
#'
#' `RootTree()` roots a tree on the smallest clade containing the specified
#' tips;
#' `RootOnNode()` roots a tree on a specified internal node;
#' `UnrootTree()` collapses a root node, without the undefined behaviour
#' encountered when using \code{\link[ape:root]{ape::unroot}()} on trees in
#' preorder.
#'
#' @template tree(s)Param
#' @param outgroupTips Vector of type character, integer or logical, specifying
#' the names or indices of the tips to include in the outgroup.
#' If `outgroupTips` is a of type character, and a tree contains multiple tips
#' with a matching label, the first will be used.
#' @param fallback Vector corresponding to `outgroupTips` determining behaviour
#' when `outgroupTips` do not root a tree.
#'
#' Where the smallest clade that contains `outgroupTips` includes all taxa,
#' `RootTree()` will not change the topology of a tree.
#' If `fallback = NULL`, `RootTree()` will return `tree`.
#' Otherwise, taxa will be excluded from `outgroupTips` in the sequence
#' specified (`fallback[1]` first) by until the clade containing all outgroup
#' tips is smaller than `tree`.
#'
#' @return `RootTree()` returns a tree of class `phylo`, rooted on the smallest
#' clade that contains the specified tips, with edges and nodes numbered in
#' preorder. Node labels are not retained.
#'
#' @examples
#' tree <- PectinateTree(8)
#' plot(tree)
#' ape::nodelabels()
#'
#' plot(RootTree(tree, c("t6", "t7")))
#'
#' plot(RootOnNode(tree, 12))
#' plot(RootOnNode(tree, 2))
#'
#' @seealso
#' - [`ape::root()`]
#'
#' @family tree manipulation
#'
#' @template MRS
#' @export
RootTree <- function(tree, outgroupTips, fallback = NULL) {
  UseMethod("RootTree")
}

.MatchTips <- function(tips, tipLabels) {
  if (is.character(tips)) {
    chosenTips <- match(tips, tipLabels)
    if (any(is.na(chosenTips))) {
      stop("Outgroup tips [",
           paste(tips[is.na(chosenTips)], collapse = ", "),
           "] not found in tree's tip labels.")
    }
    chosenTips
  } else if (is.logical(tips)) {
    which(tips)
  } else {
    tips
  }
}

#' @export
RootTree.phylo <- function(tree, outgroupTips, fallback = NULL) {
  if (missing(outgroupTips) ||
      is.null(outgroupTips) ||
      length(outgroupTips) == 0) {
    return(tree)
  }
  tipLabels <- tree[["tip.label"]]
  outgroupTips <- .MatchTips(outgroupTips, tipLabels)
  tree <- Preorder(tree)
  nTip <- NTip(tree)
  if (length(outgroupTips) == 0) {
    stop("No outgroup tips selected")
  } else if (length(outgroupTips) == 1L) {
    outgroup <- outgroupTips
  } else if (length(outgroupTips) == nTip - 1L) {
    outgroup <- setdiff(seq_len(nTip), outgroupTips)
  } else {
    ancestryTable <- .AncestorTable(tree, outgroupTips)
    lineage <- ancestryTable[1, ]
    lca <- max(lineage[ancestryTable[2, ] == length(outgroupTips)])
    nTip <- length(tipLabels)
    rootNode <- nTip + 1L
    if (lca == rootNode) {
      if (tree[["Nnode"]] > 2L) {
        lca <- lineage[lineage - c(lineage[-1], 0) != -1][[1]] + 1L
        if (lca > rootNode + nTip - 2L) {
          if (is.null(fallback) || length(fallback) == 0) {
            return(tree)
          } else {
            fallback <- .MatchTips(fallback, tipLabels)
            # TODO It might be more efficient to avoid a new call to RootTree
            return(RootTree(tree, setdiff(outgroupTips, fallback[[1]]),
                            fallback = fallback[-1]))
          }
        }
      } else {
        return(tree)
      }
    }
    outgroup <- lca
  }
  if (outgroup > nTip + tree[["Nnode"]]) {
    return(tree)
  }

  root_on_node(tree, outgroup)
}

.AllAncestors <- function(edge) {
  edge <- Preorder(edge)
  parents <- edge[, 1]
  child <- edge[, 2]
  nEdge <- length(child)
  res <- vector("list", max(parents))
  for (i in seq_len(nEdge)) {
    pa <- parents[i]
    res[[child[i]]] <- c(pa, res[[pa]])
  }

  # Return:
  res
}

.AncestorTable <- function(tree, outgroupTips) {
  edge <- tree[["edge"]]
  parent <- edge[, 1]
  child <- edge[, 2]
  nVert <- max(parent)
  parentOf <- `length<-`(integer(), nVert)
  parentOf[child] <- parent
  
  counts <- integer(nVert)
  i <- outgroupTips
  while (length(i)) {
    i <- parentOf[i]
    i <- i[!is.na(i)]
    for (j in i) {
      counts[j] <- counts[j] + 1L
    }
  }
  
  counted <- counts > 0
  
  # Return:
  rbind(node = which(counted),
        count = counts[counted])
}

#' @export
RootTree.matrix <- function(tree, outgroupTips, fallback = NULL) {
  tree <- Preorder(tree)
  if (missing(outgroupTips) ||
      is.null(outgroupTips) ||
      length(outgroupTips) == 0L) {
    return(tree)
  }
  rootNode <- tree[1]
  nNode <- max(tree[, 1]) - rootNode + 1L
  if (length(outgroupTips) == 1L) {
    outgroup <- outgroupTips
  } else {
    ancestry <- unlist(.AllAncestors(tree)[outgroupTips])
    ancestryTable <- tabulate(ancestry)
    lca <- max(which(ancestryTable == length(outgroupTips)))
    
    if (lca == rootNode) {
      if (nNode > 2L) {
        lineage <- which(as.logical(ancestryTable))
        lca <- lineage[lineage - c(lineage[-1], 0) != -1][1] + 1L
      } else {
        return(tree)
      }
    }
    outgroup <- lca
  }

  root_on_node(list(edge = tree, Nnode = nNode), outgroup)[["edge"]]
}

#' @export
RootTree.list <- function(tree, outgroupTips, fallback = NULL) {
  if (missing(outgroupTips)
      || is.null(outgroupTips)
      || length(outgroupTips) == 0) {
    return(tree)
  }
  lapply(tree, RootTree, outgroupTips, fallback)
}

#' @export
RootTree.multiPhylo <- function(tree, outgroupTips, fallback = NULL) {
  tree[] <- RootTree.list(tree, outgroupTips, fallback)
  tree
}

#' @export
RootTree.NULL <- function(tree, outgroupTips, fallback = NULL) NULL

#' @rdname RootTree
#' @param node Integer specifying node (internal or tip) to set as the root.
#' @param resolveRoot Logical specifying whether to resolve the root node.
#'
#' @return `RootOnNode()` returns a tree of class `phylo`, rooted on the
#' requested `node` and ordered in [`Preorder`].
#'
#' @export
RootOnNode <- function(tree, node, resolveRoot = FALSE) {
  UseMethod("RootOnNode", tree)
}

#' @importFrom fastmatch %fin%
#' @export
RootOnNode.phylo <- function(tree, node, resolveRoot = FALSE) {
  edge <- tree[["edge"]]
  parent <- edge[, 1]
  child <- edge[, 2]
  nodeParentEdge <- child == node
  rootEdges <- !parent %fin% child
  rootNode <- parent[rootEdges][1]
  rooted <- sum(rootEdges) == 2L

  if (any(nodeParentEdge)) {
    if (rooted) {
      # Do before editing parent:
      ancestorEdges <- EdgeAncestry(which(nodeParentEdge), parent, child,
                                    stopAt = rootEdges)

      rootChildren <- child[rootEdges]
      if (node %fin% rootChildren) {
        if (!resolveRoot) {
          if (node < NTip(tree)) {
            node <- rootChildren[rootChildren != node]
          }
          deletedEdge  <- child == node
          parent[parent == node] <- parent[deletedEdge]
          parent <- parent[!deletedEdge]
          child <- child[!deletedEdge]
          parent[parent > node] <- parent[parent > node] - 1L
          child[child > node] <- child[child > node] - 1L
          tree[["Nnode"]] <- tree[["Nnode"]] - 1L
        }
        inverters <- logical(length(parent))
      } else {
        spareRoot <- rootChildren == min(rootChildren) # Hit tip if present
        parent[rootEdges][!spareRoot] <- rootChildren[spareRoot]

        inverters <- ancestorEdges | rootEdges
        if (!ancestorEdges[rootEdges][!spareRoot]) {
          inverters[which(rootEdges)[!spareRoot]] <- FALSE
        }
        if (resolveRoot) {
          inverters <- inverters | nodeParentEdge
          parent[rootEdges][spareRoot] <- node
          child[rootEdges][spareRoot] <- rootNode
          child[nodeParentEdge] <- rootNode
        } else {
          if (node > rootNode) inverters <- inverters | nodeParentEdge
          deletedEdge <- which(rootEdges)[spareRoot]

          parent <- parent[-deletedEdge]
          child <- child[-deletedEdge]
          inverters <- inverters[-deletedEdge]

          parent[parent > rootNode] <- parent[parent > rootNode] - 1L
          child[child > rootNode] <- child[child > rootNode] - 1L

          tree[["Nnode"]] <- tree[["Nnode"]] - 1L
        }
      }
    } else {
      if (resolveRoot) {
        inverters <- c(EdgeAncestry(which(nodeParentEdge), parent, child), FALSE)
        newNode <- max(parent) + 1L
        parent <- c(parent, newNode)
        child <- c(child, parent[nodeParentEdge])
        parent[nodeParentEdge] <- newNode
        tree[["Nnode"]] <- 1L + tree[["Nnode"]]
      } else {
        inverters <- EdgeAncestry(which(nodeParentEdge), parent, child,
                                  stopAt = rootEdges)
      }
    }
    tree[["edge"]] <- RenumberTree(ifelse(inverters, child, parent),
                                   ifelse(inverters, parent, child))
    attr(tree, "order") <- "preorder"
    tree
  } else {
    # Root position is already correct
    if (rooted && !resolveRoot) {
      UnrootTree(tree)
    } else if (!rooted && resolveRoot) {
      RootOnNode(tree, max(child[rootEdges]), resolveRoot = TRUE)
    } else {
      Preorder(tree)
    }
  }
}

#' @export
RootOnNode.list <- function(tree, node, resolveRoot = FALSE) {
  lapply(tree, RootOnNode, node, resolveRoot)
}

#' @export
RootOnNode.multiPhylo <- function(tree, node, resolveRoot = FALSE) {
  tree[] <- RootOnNode.list(tree, node, resolveRoot)
  tree
}

#' @export
RootOnNode.NULL <- function(tree, node, resolveRoot = FALSE) NULL

#' @rdname RootTree
#' @return `UnrootTree()` returns `tree`, in preorder,
#' having collapsed the first child of the root node in each tree.
#' @export
UnrootTree <- function(tree) UseMethod("UnrootTree")

#' @export
UnrootTree.phylo <- function(tree) {
  tree <- Preorder(tree)
  edge <- tree[["edge"]]
  if (dim(edge)[1] < 3) {
    return(tree)
  }

  parent <- edge[, 1]
  rootNode <- parent[1]
  rootEdge2 <- parent[-1] == rootNode
  if (sum(rootEdge2) > 1L) {
    return(tree)
  }

  if (edge[1, 2] < rootNode) {
    deleted <- 1L + which(rootEdge2)
    weightTo <- 1L
  } else {
    deleted <- 1L
    weightTo <- 1L + which(rootEdge2)
  }
  renumber <- edge > rootNode
  edge[renumber] <- edge[renumber] - 1L
  tree[["edge"]] <- edge[-deleted, ]
  tree[["Nnode"]] <- tree[["Nnode"]] - 1L
  weight <- tree[["edge.length"]]
  if (!is.null(weight)) {
    weight[weightTo] <- weight[weightTo] + weight[deleted]
    tree[["edge.length"]] <- weight[-deleted]
  }

  # Return:
  tree
}

#' @export
UnrootTree.list <- function(tree) lapply(tree, UnrootTree)

#' @export
UnrootTree.multiPhylo <- function(tree) {
  tree[] <- UnrootTree.list(tree)
  tree
}

#' @export
UnrootTree.NULL <- function(tree) NULL

#' Collapse nodes on a phylogenetic tree
#'
#' Collapses specified nodes or edges on a phylogenetic tree, resulting in
#' polytomies.
#'
#' @template treeParam
#' @param nodes,edges Integer vector specifying the nodes or edges in the tree
#'  to be dropped.
#' (Use \code{\link[ape]{nodelabels}()} or
#' \code{\link[ape:nodelabels]{edgelabels}()}
#' to view numbers on a plotted tree.)
#'
#' @return `CollapseNode()` and `CollapseEdge()` return a tree of class `phylo`,
#' corresponding to `tree` with the specified nodes or edges collapsed.
#' The length of each dropped edge will (naively) be added to each descendant
#' edge.
#'
#' @examples
#' oldPar <- par(mfrow = c(3, 1), mar = rep(0.5, 4))
#'
#' tree <- as.phylo(898, 7)
#' tree$edge.length <- 11:22
#' plot(tree)
#' nodelabels()
#' edgelabels()
#' edgelabels(round(tree$edge.length, 2),
#'            cex = 0.6, frame = "n", adj = c(1, -1))
#'
#' # Collapse by node number
#' newTree <- CollapseNode(tree, c(12, 13))
#' plot(newTree)
#' nodelabels()
#' edgelabels(round(newTree$edge.length, 2),
#'            cex = 0.6, frame = "n", adj = c(1, -1))
#'
#' # Collapse by edge number
#' newTree <- CollapseEdge(tree, c(2, 4))
#' plot(newTree)
#'
#' par(oldPar)
#'
#' @family tree manipulation
#' @author  Martin R. Smith
#' @export
CollapseNode <- function(tree, nodes) UseMethod("CollapseNode")

#' @rdname CollapseNode
#' @importFrom fastmatch %fin%
#' @export
CollapseNode.phylo <- function(tree, nodes) {
  if (length(nodes) == 0) return(tree)

  edge <- tree[["edge"]]
  lengths <- tree[["edge.length"]]
  hasLengths <- !is.null(lengths)
  nodeLabels <- tree[["node.label"]]
  hasLabels <- !is.null(nodeLabels)
  parent <- edge[, 1]
  child <- edge[, 2]
  root <- RootNode(edge)
  nTip <- NTip(tree)
  maxNode <- max(parent)
  edgeBelow <- order(child, method = "radix") # a little faster than "auto"
  tips <- seq_len(nTip)
  preRoot <- seq_len(root - 1L)
  edgeBelow <- c(edgeBelow[preRoot], NA, edgeBelow[-preRoot])
  nodes <- unique(nodes)

  if (!all(nodes %fin% (nTip + 1L):maxNode)) {
    stop("nodes must be integers between ", nTip + 1L, " and ", maxNode)
  }
  if (any(nodes == root)) {
    warning("Cannot collapse root node")
    nodes <- nodes[nodes != root]
    if (length(nodes) == 0L) return(tree)
  }

  keptEdges <- -edgeBelow[nodes]
  depths <- NodeDepth(edge)[nodes]

  for (node in nodes[order(depths)]) {
    newParent <- parent[edgeBelow[node]]
    if (hasLengths) {
      lengths[parent == node] <- lengths[parent == node] + lengths[child == node]
    }
    parent[parent == node] <- newParent
  }

  newNumber <- c(seq_len(nTip), nTip + cumsum((nTip + 1L):maxNode %fin% parent))

  tree[["edge"]] <- cbind(newNumber[parent[keptEdges]],
                          newNumber[child[keptEdges]])
  tree[["edge.length"]] <- lengths[keptEdges]
  tree[["Nnode"]] <- tree[["Nnode"]] - length(nodes)
  if (hasLabels) {
    tree[["node.label"]] <- nodeLabels[-(nodes - nTip)]
  }

  # TODO Renumber nodes sequentially
  # TODO Re-write this in C++.
  tree
}

#' @rdname CollapseNode
#' @export
CollapseEdge <- function(tree, edges) {
  nodesToCollapse <- tree[["edge"]][edges, 2]
  if (any(nodesToCollapse < NTip(tree))) {
    stop("Cannot collapse external edges: ",
         paste(edges[nodesToCollapse <= NTip(tree)], collapse = ", "))
  }
  CollapseNode(tree, nodesToCollapse)
}

#' Generate binary tree by collapsing polytomies
#'
#' `MakeTreeBinary()` resolves, at random, all polytomies in a tree or set of
#' trees, such that all trees compatible with the input topology are drawn
#' with equal probability. Edge lengths are not yet supported, so are removed.
#'
#' @seealso Since ape v5.5, this functionality is available through
#' [`ape::multi2di()`]; previous versions of "ape" did not return topologies
#' in equal frequencies.  `MakeTreeBinary()` is often somewhat faster;
#' `multi2di()` retains edge lengths.
#'
#' @return `MakeTreeBinary()` returns a rooted binary tree of class `phylo`,
#' corresponding to tree uniformly selected from all those compatible with
#' the input tree topologies.
#'
#' @examples
#' MakeTreeBinary(CollapseNode(PectinateTree(7), c(9, 11, 13)))
#' UnrootTree(MakeTreeBinary(StarTree(5)))
#' @template MRS
#' @template treeParam
#' @family tree manipulation
#' @export
MakeTreeBinary <- function(tree) {
  UseMethod("MakeTreeBinary")
}

#' @export
MakeTreeBinary.NULL <- function(tree) NULL

#' @export
MakeTreeBinary.phylo <- function(tree) {
  tree <- Preorder(`[[<-`(tree, "edge.length", NULL))
  degree <- NodeOrder(tree, internalOnly = TRUE)
  degree[[1]] <- degree[[1]] + 1L # Root node is degree 2
  polytomies <- degree > 3L
  if (!any(polytomies)) {
    return(tree)
  }
  edge <- tree[["edge"]]

  nTip <- edge[[1]] - 1L
  polytomyN <- which(polytomies) + nTip
  degree <- degree[polytomies]
  for (i in seq_len(sum(polytomies))) {
    n <- polytomyN[[i]]
    nKids <- degree[[i]] - 1L
    newParent <- .RandomParent(nKids + 1L) # Tip 1 is the "root"
    newEdges <- RenumberEdges(newParent, seq_len(nKids + nKids))

    nNewNodes <- nKids - 2L

    childEdges <- edge[, 1] == n

    keep <- edge[!childEdges, ]
    increase <- keep > n
    keep[increase] <- keep[increase] + nNewNodes

    children <- edge[childEdges, 2L]
    increase <- children > n
    children[increase] <- children[increase] + nNewNodes

    polytomyN <- polytomyN + nNewNodes

    newEdges2 <- newEdges[[2]][c(-1, -2)] - 1L
    decrease <- newEdges2 > nKids
    newEdges2[decrease] <- newEdges2[decrease] - 2L

    add <- cbind(newEdges[[1]][-(1:2)] + n - nKids - 3L,
                 c(children, n + seq_len(nNewNodes))[newEdges2])

    edge <- rbind(keep, add)
  }
  nodeLabel <- tree[["node.label"]]
  if (!is.null(nodeLabel)) {
    # Inefficient but pragmatic
    tree[["node.label"]] <- .UpdateNodeLabel.numeric(edge, tree, nodeLabel)
  }
  tree[["edge"]] <- edge
  tree[["Nnode"]] <- nTip - 1L
  attr(tree, "order") <- NULL
  tree
}

#' @export
MakeTreeBinary.list <- function(tree) lapply(tree, MakeTreeBinary)

#' @export
MakeTreeBinary.multiPhylo <- function(tree) {
  structure(MakeTreeBinary.list(tree), class = "multiPhylo")
}


#' Leaf label interchange
#'
#' `LeafLabelInterchange()` exchanges the position of leaves within a tree.
#'
#' Modifies a tree by switching the positions of _n_ leaves.  To avoid
#' later swaps undoing earlier exchanges, all _n_ leaves are guaranteed
#' to change position.  Note, however, that no attempt is made to avoid
#' swapping equivalent leaves, for example, a pair that are each others'
#' closest relatives.  As such, the relationships within a tree are not
#' guaranteed to be changed.
#'
#' @template treeParam
#' @param n Integer specifying number of leaves whose positions should be
#' exchanged.
#'
#' @return `LeafLabelInterchange()` returns a tree of class `phylo` on which
#' the position of `n` leaves have been exchanged.
#' The tree's internal topology will not change.
#'
#' @examples
#' tree <- PectinateTree(8)
#' plot(LeafLabelInterchange(tree, 3L))
#'
#' @template MRS
#' @family tree manipulation
#' @export
LeafLabelInterchange <- function(tree, n = 2L) {

  if (n < 2L) return(tree)
  tipLabel <- tree[["tip.label"]]
  nTip <- length(tipLabel)
  if (n > nTip) {
    stop("`n` cannot exceed number of tips in tree (", nTip, ")")
  }

  from <- sample.int(nTip, n)
  # We need to ensure that a tip's new position is different from its old one.
  # A simple way would be to shift each moved tip to the next one in turn:
  # 1, 2, 3, 4 -> 2, 3, 4, 1
  # But this cyclicity is non-random: we may wish to introduce some smaller
  # cycles, e.g.
  # 1, 2, 3, 4 -> 2, 1, 4, 3.

  cycles <- integer(n / 2L)
  remaining <- n
  i <- 1L
  while (remaining > 0L) {
    if (remaining == 2L) {
      cycles[i] <- 2L
      break
    }
    thisCycle <- SampleOne(1L + seq_len(remaining - 2L))
    if (thisCycle == remaining - 1L) thisCycle <- remaining
    remaining <- remaining - thisCycle
    cycles[i] <- thisCycle
    i <- i + 1L
  }
  cycles <- cycles[cycles > 0L]
  nCycles <- length(cycles)
  start <- cumsum(c(0L, cycles[-nCycles]))
  to <- unlist(lapply(seq_along(cycles), function(i) {
     c(seq_len(cycles[i] - 1L) + 1L, 1L) + start[i]
  }))

  tree[["tip.label"]][from] <- tipLabel[from[to]]
  
  # Return:
  tree
}

test_that("KeptVerts() works", {
  Keepers <- function(n, nTip) as.logical(tabulate(n, nTip))
  
  bal12 <- BalancedTree(12)
  expect_error(KeptVerts(bal12, raw(0)), "`keptTips`")
  Y <- TRUE
  N <- FALSE
  expect_error(KeptVerts(matrix(FALSE, 4, 2), c(Y, Y, N, Y)),
               "no applicable method")
  expect_error(KeptVerts(matrix(0, 4, 3), c(Y, Y, N, Y)),
               "edge matrix of a `phylo` obj")
  expect_error(KeptVerts(1:8, c(Y, Y, N, Y)),
               "edge matrix of a `phylo` obj")
  expect_error(KeptVerts(Postorder(bal12), Keepers(1:10, 12)), "in preorder")

  expect_equal(KeptVerts(bal12, 1:12 %in% 1:9),
               c(1:12 %in% 1:9, rep(Y, 6), N, Y, Y, N, N))
  expect_equal(KeptVerts(bal12, 1:12 %in% 7:12),
               c(1:12 %in% 7:12, rep(N, 6), rep(Y, 5)))
  expect_equal(KeptVerts(bal12, 1:12 %in% 4:6),
               c(1:12 %in% 4:6, rep(N, 4), rep(Y, 2), rep(N, 5)))
  expect_equal(KeptVerts(bal12, 1:12 %in% c(1, 2, 6, 12)),
               c(1:12 %in% c(1, 2, 6, 12), Y, Y, N, Y, N, N, N, N, N, N, N))
  expect_equal(KeptVerts(bal12, rep(Y, 12)), rep(TRUE, 23))
  
  # Input format
  keeps <- c(1, 3, 5, 7, 9, 10)
  expect_equal(KeptVerts(bal12, keeps),
               KeptVerts(bal12, 1:12 %in% keeps))
  expect_equal(KeptVerts(bal12, bal12$tip.label[keeps]),
               KeptVerts(bal12, as.integer(keeps)))
  
  pec9 <- UnrootTree(PectinateTree(9))
  kept <- c(1, 3, 8, 9)
  expect_equal(which(KeptVerts(pec9, 1:9 %in% kept)),
               c(kept, 10, 16))
  
  real <- Preorder(read.tree(text = 
  "(t1:3,t2:1,((t3:1,t4:1):3,(t5:2,(t6:1,(t7:6,((t8:8,(t9:2,t10:4):3):2,(t11:1,t12:2):5):1):2):1):1):2);"))
  kept <- c(1L, 3L, 4L, 8L, 9L)
  # Duplicate root is not retained.
  expect_equal(
    which(KeptVerts(real, Keepers(kept, 12))),
    c(kept, 13, 15, 20))
  
  # Basic tests with unrooted trees
  urb12 <- UnrootTree(bal12)
  expect_equal(which(KeptVerts(urb12, Keepers(7:12, 12))),
               c(7:12, 18, 20:22))
  expect_equal(which(KeptVerts(urb12, Keepers(10:12, 12))),
               c(10:12, 21))
  expect_equal(which(KeptVerts(urb12, Keepers(1:6, 12))),
               c(1:6, 13, 15:17))
  col19 <- Preorder(CollapseNode(bal12, 19))
  expect_equal(which(KeptVerts(col19, Keepers(6:12, 12))),
               c(6:12, 13, 19:22))
  
  
  # Duplicate keep_tip tests
  expect_equal(which(KeptVerts(BalancedTree(9)$edge, !tabulate(5:8, 9))),
               c(1:4, 9, 10:13))
  
  expect_equal(which(KeptVerts(BalancedTree(8)$edge, !tabulate(6:4, 8))),
               c(1:3, 7:8, 9:11, 15))
  expect_equal(which(KeptVerts(BalancedTree(8)$edge, !tabulate(1:4, 8))),
               c(5:8, 13:15))
  
  expect_equal(which(KeptVerts(BalancedTree(8)$edge, !tabulate(3:8, 8))),
               c(1:2, 11))
  
  expect_equal(which(KeptVerts(ape::unroot(BalancedTree(4)), !tabulate(1, 4))),
               c(2:4, 5))
  
  expect_equal(which(KeptVerts(as.phylo(3, 7)$edge, !tabulate(1:4, 7))),
               c(5:7, 9, 11))
  
  # Rooted tree may lose root when reaching a polytomy:
  expect_equal(which(KeptVerts(RootTree(StarTree(6), 4), !tabulate(4, 6))),
               c(1:3, 5:6, 8))
  
  # But need not:
  expect_equal(which(KeptVerts(RootTree(BalancedTree(5), 3)$edge,
                               !tabulate(3, 5))),
               c(1:2, 4:5, 7:9))
  
})

nasty <- structure(list(edge = structure(
  c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
    5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
  .Dim = c(12, 2)),
  Nnode = 5L,
  tip.label = letters[1:8]),
  class = "phylo") # Danger: Do not plot!

test_that(".DescendantTips() recurses", {
  nTip <- 19L
  tree <- BalancedTree(nTip)
  edge <- tree$edge
  expect_equal(.DescendantTips(edge[, 1], edge[, 2], nTip, c(23, 32, 35)),
               c(1:3, 11:13, 16:19))
})

test_that("RootOnNode() works", {
  expect_null(RootOnNode(NULL, 1))

  tree <- structure(list(edge = structure(c(6L, 9L, 9L, 7L, 7L, 8L, 8L,
                                            6L, 9L, 2L, 7L, 3L, 8L, 4L, 5L, 1L),
                                          .Dim = c(8L, 2L)),
                         tip.label = c("t3", "t4", "t1", "t2", "t5"),
                         Nnode = 4L), class = "phylo", order = "cladewise")
  
  exp8 <- structure(list(edge = structure(c(6L, 7L, 8L, 8L, 7L, 6L, 9L, 9L, 7L,
                                            8L, 1L, 2L, 3L, 9L, 4L, 5L),
                                          .Dim = c(8L, 2L)),
                         tip.label = c("t3", "t4", "t1", "t2", "t5"),
                         Nnode = 4L), class = "phylo", order = "preorder")
  exp7 <- structure(list(edge = structure(c(6L, 7L, 7L, 6L, 8L, 8L, 9L, 9L, 7L,
                                            1L, 2L, 8L, 3L, 9L, 4L, 5L),
                                          .Dim = c(8L, 2L)),
                         tip.label = c("t3", "t4", "t1", "t2", "t5"),
                         Nnode = 4L), class = "phylo", order = "preorder")
  exp5 <- structure(list(edge = structure(c(6L, 7L, 8L, 9L, 9L, 8L, 7L, 6L, 7L,
                                            8L, 9L, 1L, 2L, 3L, 4L, 5L),
                                          .Dim = c(8L, 2L)),
                         tip.label = c("t3", "t4", "t1", "t2", "t5"),
                         Nnode = 4L), class = "phylo", order = "preorder")
  exp4 <- structure(list(edge = structure(c(6L, 7L, 8L, 9L, 9L, 8L, 7L, 6L, 7L,
                                            8L, 9L, 1L, 2L, 3L, 5L, 4L),
                                          .Dim = c(8L, 2L)),
                         tip.label = c("t3", "t4", "t1", "t2", "t5"),
                         Nnode = 4L), class = "phylo", order = "preorder")
  exp3 <- structure(list(edge = structure(c(6L, 7L, 8L, 8L, 7L, 9L, 9L, 6L, 7L,
                                            8L, 1L, 2L, 9L, 4L, 5L, 3L),
                                          .Dim = c(8L, 2L)),
                         tip.label = c("t3", "t4", "t1", "t2", "t5"),
                         Nnode = 4L), class = "phylo", order = "preorder")
  exp2 <- structure(list(edge = structure(c(6L, 7L, 7L, 8L, 8L, 9L, 9L, 6L, 7L,
                                            1L, 8L, 3L, 9L, 4L, 5L, 2L),
                                          .Dim = c(8L, 2L)),
                         tip.label = c("t3", "t4", "t1", "t2", "t5"),
                         Nnode = 4L), class = "phylo", order = "preorder")
  #t2 <- Preorder(t2)
  expect_true(all.equal(tree, RootOnNode(tree, node = 9L, TRUE)))
  expect_true(all.equal(exp8, RootOnNode(tree, node = 8L, TRUE)))
  expect_true(all.equal(exp7, RootOnNode(tree, node = 7L, TRUE)))
  expect_true(all.equal(tree, RootOnNode(tree, node = 6L, TRUE)))

  expect_true(all.equal(exp5, RootOnNode(tree, node = 5L, TRUE)))
  expect_true(all.equal(exp4, RootOnNode(tree, node = 4L, TRUE)))
  expect_true(all.equal(exp3, RootOnNode(tree, node = 3L, TRUE)))
  expect_true(all.equal(exp2, RootOnNode(tree, node = 2L, TRUE)))
  expect_true(all.equal(tree, RootOnNode(tree, node = 1L, TRUE)))
  expect_true(all.equal(tree, RootOnNode(UnrootTree(tree), node = 1L, TRUE)))

  expb6_8 <- structure(list(edge = structure(c(8, 8, 10, 10, 9, 9, rep(7, 3),
                                               1:2, 4:6, 10, 3, 8:9),
                                             .Dim = c(9L, 2L)),
                            tip.label = paste0("t", 1:6), Nnode = 4L),
                       class = "phylo", order = "preorder")
  expp6_8 <- structure(list(edge = structure(c(7, rep(7:10, each = 2), 1:2, 8,
                                               3, 9, 4, 10, 5:6),
                                             .Dim = c(9L, 2L)),
                            tip.label = paste0("t", 1:6), Nnode = 4L),
                       class = "phylo", order = "preorder")

  expect_true(all.equal(expb6_8, Preorder(RootOnNode(BalancedTree(6L), 8L))))
  expect_true(all.equal(expp6_8, Preorder(RootOnNode(PectinateTree(6L), 8L))))

  expect_true(all.equal(UnrootTree(PectinateTree(6L)),
                        RootOnNode(PectinateTree(6L), 1)))

  TestTip <- function(tr, node, rr) {
    expect_true(all.equal(
      Preorder(ape::root(tr, outgroup = node, resolve.root = rr)),
      RootOnNode(tr, node, rr))
    )
  }
  TestInternal <- function(tr, node, rr) {
    expect_true(all.equal(
      Preorder(ape::root(tr, node = node, resolve.root = rr)),
      RootOnNode(tr, node, rr)))
  }

  TestInternal(PectinateTree(8), 14, TRUE)
  TestInternal(PectinateTree(8), 14, FALSE)
  TestTip(PectinateTree(8), 4L, TRUE)
  TestTip(PectinateTree(8), 4L, FALSE)

  TestInternal(BalancedTree(8L), 11L, TRUE)
  TestInternal(BalancedTree(8L), 11L, FALSE)

  urt <- UnrootedTreeWithShape(3, 8, letters[1:8])
  TestInternal(urt, 12, TRUE)
  TestInternal(urt, 12, FALSE)
  TestTip(urt, 4L, TRUE)
  TestTip(urt, 4L, FALSE)

  # Children of an unresolved root
  TestInternal(urt, 10, TRUE)
  TestTip(urt, 1, TRUE)

  expect_true(all.equal(PectinateTree(8), RootOnNode(PectinateTree(8), 9L, TRUE)))
  expect_true(all.equal(UnrootTree(PectinateTree(8)),
                        RootOnNode(PectinateTree(8), 9L, FALSE)))
  expect_true(all.equal(urt, RootOnNode(urt, 9L, FALSE)))
  expect_true(all.equal(Preorder(RootTree(urt, letters[1:2])),
                        RootOnNode(urt, 9L, TRUE)))

})

test_that("root_on_node() works", {
  tree <- Preorder(BalancedTree(15))
  edge <- tree$edge
  TipTest <- function(i) {
    tr.rooted <- root_on_node(tree, i)
    expect_equal(SortTree(root(tree, i, resolve.root = TRUE)),
                 SortTree(tr.rooted))
  }
  StaticTest <- function(i) expect_equal(tree, root_on_node(tree, i))
  NodeTest <- function(i) {
    tr.rooted <- root_on_node(tree, i)
    expect_equal(SortTree(root(tree, node = i, resolve.root = TRUE)),
                 SortTree(tr.rooted))
  }
  expect_error(root_on_node(edge, 0))
  for (i in 1:15) TipTest(i)
  StaticTest(16)
  StaticTest(17)
  for (i in 18:23) NodeTest(i)
  StaticTest(24)
  for (i in 24:29) NodeTest(i)
  expect_error(root_on_node(edge, 30))

  tree <- Preorder(root(BalancedTree(15), "t1", resolve.root = TRUE))
  edge <- tree$edge
  expect_error(root_on_node(edge, 0))
  for (i in 1:15) TipTest(i)
  StaticTest(16)
  StaticTest(17)
  for (i in 18:29) NodeTest(i)
  expect_error(root_on_node(edge, 30))
})

test_that("RootOnNode() supports lists of trees", {
  rootOn <- 8L
  expect_equal(structure(list(RootOnNode(as.phylo(1, 5), rootOn),
                              RootOnNode(as.phylo(2, 5), rootOn)),
                         class = "multiPhylo",
                         tip.label = paste0("t", 1:5)),
               RootOnNode(as.phylo(1:2, 5), rootOn))
})

test_that("RootTree() supports star trees", {
  star <- StarTree(8)
  expect_true(all.equal(star, RootTree(star, 5:7)))
  expect_true(all.equal(star$edge, RootTree(star$edge, 5:7)))
})

test_that("RootOnNode() supports nasty node ordering", {
  expect_equal(Preorder(nasty),
               RootOnNode(nasty, 12L, resolveRoot = TRUE))
  expect_equal(RootOnNode(Preorder(nasty), 11L),
               RootOnNode(nasty, 13L))
})

test_that("RootTree() handles null outgroups", {
  bal8 <- BalancedTree(8)
  expect_equal(RootTree(bal8), bal8)
  expect_equal(RootTree(bal8$edge), bal8$edge)
  expect_equal(RootTree(bal8, NULL), bal8)
  expect_equal(RootTree(c(bal8, bal8)), c(bal8, bal8))
  expect_equal(RootTree(list(bal8, bal8)), list(bal8, bal8))
  expect_equal(RootTree(bal8$edge, NULL), bal8$edge)
  expect_equal(RootTree(bal8, character(0)), bal8)
  expect_null(RootTree(NULL))
  expect_null(RootTree(NULL, "tip"))
})

test_that("RootTree() works", {
  bal8 <- BalancedTree(8)
  bal15 <- BalancedTree(15)
  expect_error(RootTree(bal8, 1:8 %in% 0))
  expect_error(RootTree(bal8, "tip_not_there"))
  expect_equal(RootTree(bal8, 5:6), RootTree(bal8, 1:8 %in% 5:6))
  expect_equal(RootTree(bal8$edge, 5:6), RootTree(bal8, 5:6)$edge)
  expect_equal(RootTree(bal15$edge, 9:11), RootTree(bal15, 9:11)$edge)
  expect_equal(RootTree(bal8, 5:6), RootTree(bal8, c("t5", "t6")))
  expect_true(all.equal(
               as.phylo(5518, 8, paste0("t", rev(c(7,8,3,4,1,2,6,5)))),
               RootTree(bal8, "t5")))
  expect_equal(RootTree(bal8, 5), RootTree(bal8, "t5"))
  expect_equal(RootTree(bal8, 5)$edge, RootTree(bal8$edge, 5))
  expect_equal(RootTree(bal8, c("t1", "t2")), RootTree(bal8, c("t4", "t5")))

  expect_equal(structure(list(RootTree(as.phylo(1, 5), "t5"),
                              RootTree(as.phylo(2, 5), "t5")),
                         class = "multiPhylo",
                         tip.label = paste0("t", 1:5)),
               RootTree(as.phylo(1:2, 5), "t5"))

  expect_true(all.equal(read.tree(text = "((b, c), (a, (d, e)));"),
               RootTree(read.tree(text = "(a, ((b, c), (d, e)));"), c(1, 4))))

  pec4 <- PectinateTree(4)
  expect_true(all.equal(pec4, RootTree(pec4, c(1, 3))))
  pec5 <- PectinateTree(5)
  expect_true(all.equal(pec5, RootTree(pec5, c(1, 4))))

  tree <- structure(list(edge = structure(c(7L, 8L, 8L, 7L, 7L, 9L, 9L, 10L,
                                            10L, 8L, 1L, 6L, 2L, 9L, 3L, 10L,
                                            4L, 5L), .Dim = c(9L, 2L)),
                         Nnode = 4L, tip.label = letters[1:6]),
                    class = "phylo", order = "preorder")
  expect_equal(RootTree(tree, 1:5), RootTree(tree, 6))
  # non-contiguous outgroup
  expect_equal(RootTree(tree, 1:4), RootTree(tree, 5:6))
})

test_that("RootTree(fallback)", {
  tangle <- ape::read.tree(text = "((a1, (a2, a3)), ((b4, (b5, b6)), (c7, (c8, c9))));")
  outgroup <- c("a3", "b6", "c7", "c8", "c9")
  fallback <- c("c7", "c8", "b4", "b5", "c9", "a1", "a2")
  toN <- function(x) as.integer(substr(x, 2, 2))

  expect_equal(
    RootTree(tangle, outgroup, fallback),
    RootTree(tangle, c("a3", "b6"))
  )
  expect_equal(
    RootTree(tangle, outgroup, toN(fallback)),
    RootTree(tangle, c("a3", "b6"))
  )
  expect_equal(
    RootTree(tangle, toN(outgroup), fallback),
    RootTree(tangle, c("a3", "b6"))
  )
  expect_equal(
    RootTree(c(tangle, tangle), outgroup, fallback),
    RootTree(c(tangle, tangle), c("a3", "b6"))
  )
})

test_that("RootTree() & UnrootTree() retain edge lengths", {
  bal7 <- BalancedTree(7)
  bal7$edge.length <- 1:12 * 10
  attr(bal7, "order") <- NULL
  expect_equal(RootTree(bal7, 1:4),
               structure(bal7, order = "preorder"))
  expect_equal(RootTree(RootTree(bal7, 1), 1:4),
               structure(bal7, order = "preorder"))
  exp <- structure(bal7, order = "preorder")
  exp$edge.length = 10 * c(1 + 8, 2:7, 0, 9:12)
  expect_equal(RootTree(UnrootTree(bal7), 1:4), exp)
})

test_that("UnrootTree() works", {
  expect_null(UnrootTree(NULL))
  expect_equal(matrix(c(7, 8, 8, 7, 7, 9, 10, 10, 9,
                        8, 1, 2, 3, 9, 10, 4,  5, 6), ncol = 2L),
               UnrootTree(BalancedTree(6))$edge)
  expect_equal(matrix(c(7, 7, 7, 8, 8, 9, 9, 10, 10,
                        1, 2, 8, 3, 9, 4, 10, 5,  6), ncol = 2L),
               UnrootTree(PectinateTree(6))$edge)
  expect_true(all.equal(BalancedTree(2), UnrootTree(BalancedTree(2))))
  expect_true(all.equal(BalancedTree(1), UnrootTree(BalancedTree(1))))
  expect_equal(matrix(c(9, 9, 10, 10, 10, 9, 11, 11, 12, 12, 9,
                        1, 10, 2,  4,  7, 11, 3, 12,  5,  6, 8), ncol = 2L),
               UnrootTree(nasty)$edge)
  # Unrooting when already unrooted
  expect_equal(UnrootTree(PectinateTree(6)),
               UnrootTree(UnrootTree(PectinateTree(6))))

  expList <- list(UnrootTree(as.phylo(1, 5)), UnrootTree(as.phylo(2, 5)))
  expect_equal(expList, UnrootTree(list(as.phylo(1, 5), as.phylo(2, 5))))
  exp <- structure(expList, class = "multiPhylo")
  attr(exp, "tip.label") <- paste0("t", 1:5)
  expect_equal(exp, UnrootTree(as.phylo(1:2, 5)))
})

test_that("CollapseNode() works", {
  tree8  <- read.tree(text="(((a, (b, (c, d))), (e, f)), (g, h));")
  expect_error(CollapseNode(1:5, tree8))
  expect_error(CollapseNode(tree8, 1))
  expect_warning(CollapseNode(tree8, 9L))

  tree <- as.phylo(123, 7)
  tree$edge.length <- 12:1
  expect_true(all.equal(tree, CollapseNode(tree, integer(0))))

  exp1213 <- matrix(c(8, 8, 9, 10, 10, 9, 11, 11, 11, 11,
                      1, 9, 10, 2, 4, 11, 3, 7, 6, 5), ncol=2)
  no1213 <- CollapseNode(tree, c(12, 13))
  expect_equal(exp1213, no1213$edge)

  el <- tree$edge.length
  expect_equal(no1213$edge.length, c(el[1:6],
                                     el[7] + c(c(el[8] + el[9:10]), el[11]),
                                     el[12]))

  no11 <- CollapseEdge(tree, c(7, 8))
  expect_equal(exp1213, no11$edge)

  expect_true(all.equal(CollapseNode(Preorder(nasty), c(11, 12)),
                        Preorder(CollapseNode(nasty, c(11, 13)))))

  expect_error(CollapseEdge(tree, 9))
})

test_that("CollapseNode() handles node labels", {
  bal6 <- BalancedTree(6)
  startLabels <- paste("Node", 7:11)
  bal6[["node.label"]] <- startLabels
  if (interactive()) {
    plot(bal6, show.node.label = TRUE)
  }
  
  # Collapse a cherry
  expect_equal(
    CollapseNode(bal6, 9)[["node.label"]],
    startLabels[-3]
  )
  
  # Collapse an internal node
  expect_equal(
    CollapseNode(bal6, 8)[["node.label"]],
    startLabels[-2]
  )
  
  # Collapse an internal node
  expect_equal(
    CollapseEdge(bal6, 1)[["node.label"]],
    startLabels[-2]
  )
  
  # case = 3 -> y is bound on an internal edge
  expect_equal(
    CollapseNode(bal6, c(8, 11))[["node.label"]],
    startLabels[-c(2, 5)]
  )
  
  expect_equal(AddTipEverywhere(bal6)[[1]][["node.label"]],
               AddTip(bal6, where = 1)[["node.label"]])
})

test_that("MakeTreeBinary() edge lengths", {
  tree <- ape::read.tree(text = "(a:1, b:2, c:3, d:4);")
  # Not yet supported; ensure they are removed
  expect_null(MakeTreeBinary(tree)$edge.length)
})

test_that("Binarification is uniform", {
  set.seed(0)
  Test <- function(tree, nTree, nSamples = 200L, ape = FALSE) {
    counts <- table(replicate64(nSamples, as.TreeNumber(MakeTreeBinary(tree))))
    expect_equal(nTree, length(counts))
    expect_gt(chisq.test(counts)$p.value, 0.001)
  }

  Test(CollapseNode(PectinateTree(6), 8:9), NUnrooted(4))
  Test(CollapseNode(PectinateTree(6), 9:10), NRooted(4))
  Test(CollapseNode(PectinateTree(6), c(8, 10)), NUnrooted(3) * NRooted(3))
  Test(CollapseNode(BalancedTree(8), c(10:12)), NUnrooted(5))
  Test(CollapseNode(BalancedTree(7), c(10, 13)), NRooted(3) * NRooted(3))

  bal7 <- BalancedTree(7)
  bal7[["node.label"]] <- paste("Node", 8:13)
  expect_true(all.equal(bal7, MakeTreeBinary(bal7)))
  expect_true(all.equal(list(bal7, bal7), MakeTreeBinary(list(bal7, bal7))))
  expect_true(all.equal(
    structure(list(bal7, bal7), class = "multiPhylo"),
    MakeTreeBinary(structure(list(bal7, bal7), class = "multiPhylo"))))

  set.seed(1)
  binNodes <- MakeTreeBinary(CollapseNode(bal7, 9:10))[["node.label"]]
  expect_equal(binNodes[!is.na(binNodes)], bal7[["node.label"]][-(2:3)])
  
})

test_that("LeafLabelInterchange() fails", {
  skip_if(Sys.getenv("USING_ASAN") != "")
  expect_error(LeafLabelInterchange(BalancedTree(4), 5)) # n too many
})

test_that("LeafLabelInterchange() works", {
  expect_equal(PectinateTree(40), LeafLabelInterchange(PectinateTree(40), 1))
  expect_true(all.equal(BalancedTree(2),
                        LeafLabelInterchange(BalancedTree(2), 2)))
  expect_equal(rev(BalancedTree(2)$tip),
                   LeafLabelInterchange(BalancedTree(2), 2)$tip)

  abcd <- letters[1:4]
  sapply(1 + seq_len(100), function(i) {
    # Check all perturbations
    set.seed(i)
    expect_false(any(abcd == LeafLabelInterchange(BalancedTree(abcd), 4)$tip))

    # Check lots of sizes
    expect_equal(i, sum(paste0("t", 1:128) !=
                        LeafLabelInterchange(BalancedTree(128), i)$tip))
  })
})

test_that("descendant_edges() handles errors", {
  expect_error(descendant_edges(1:2, 1:3, 1:2),
               "`parent` and `child` must be the same length")
  expect_error(descendant_tips(1:2, 1:3, 1:2),
               "`parent` and `child` must be the same length")
  expect_error(descendant_edges(1:2, 1:2, 1:3),
               "`postorder` must list each edge once")
  expect_error(descendant_tips(1:2, 1:2, 1:3),
               "`postorder` must list each edge once")
})

test_that("DescendantEdges() works", {
  pec5 <- UnrootTree(PectinateTree(5))
  V <- TRUE
  x <- FALSE
  answer <- matrix(c(V, x, x, x, x, x, x,
                     x, V, x, x, x, x, x,
                     x, x, V, V, V, V, V,
                     x, x, x, V, x, x, x,
                     x, x, x, x, V, V, V,
                     x, x, x, x, x, V, x,
                     x, x, x, x, x, x, V), 7, 7, byrow = T)
  expect_equal(
    DescendantEdges(edge = NULL, pec5$edge[, 1], pec5$edge[, 2]),
    answer)
  expect_equal(
    apply(DescendantEdges(node = 0, pec5$edge[, 1], pec5$edge[, 2]), 1, which),
    list(1:7, 4:7, 6:7)
  )
  expect_equal(
    apply(DescendantEdges(node = 8:7, pec5$edge[, 1], pec5$edge[, 2]), 1, which),
    list(6:7, 4:7)
  )
})

test_that("DescendantTips() works", {
  tree <- as.phylo(0, 6)
  parent <- tree$edge[, 1]
  child <- tree$edge[, 2]
  expect_equal(DescendantTips(parent, child, 5), 1:6 %in% c(2, 6))
  expect_equal(
    DescendantTips(parent, child),
    t(vapply(1:10, function(x) DescendantTips(parent, child, edge = x),
             logical(6)))
  )
  
  polytomies <- CollapseNode(BalancedTree(9), c(12, 13, 16))
  if (interactive()) {
    plot(polytomies)
    edgelabels()
    nodelabels()
  }
  pol <- polytomies[["edge"]]
  expect_equal(which(DescendantTips(pol[, 1], pol[, 2], edge = 5)), 4:5)
  expect_equal(which(DescendantTips(pol[, 1], pol[, 2], edge = 1)), 1:5)
  expect_equal(which(DescendantTips(pol[, 1], pol[, 2], edge = 8)), 6:9)
  expect_equal(which(DescendantTips(pol[, 1], pol[, 2], edge = 10)), 7)
  
  expect_equal(which(DescendantTips(pol[, 1], pol[, 2], node = 5)), 5)
  expect_equal(which(DescendantTips(pol[, 1], pol[, 2], node = 11)), 1:5)
  expect_equal(which(DescendantTips(pol[, 1], pol[, 2], node = 0)[11, ]), 1:5)
})

test_that("DescendantTips() handles postorder", {
  post6 <- Postorder(BalancedTree(6))
  if (interactive()) {
    oPar <- par(mar = rep(0.5, 4), cex = 0.9)
    on.exit(par(oPar))
    plot(post6)
    nodelabels()
    edgelabels()
    tiplabels()
  }
  parent <- post6[["edge"]][, 1]
  child <- post6[["edge"]][, 2]
  expect_equal(DescendantTips(parent, child, edge = 7), 1:6 %in% 1:2)
  expect_equal(DescendantTips(parent, child, edge = 9), 1:6 %in% 4:6)
  # edge = NULL is handled by .AllDescendantEdges
  expect_equal(DescendantTips(parent, child),
               t(vapply(list(5, 4, 4:5, 6, 2, 1, 1:2, 3, 4:6, 1:3),
                        function (table) 1:6 %in% table,
                        logical(6)))
  )
})

test_that("SortTree() works", {
  expect_equal(SortTree(as.phylo(10, 6))$edge,
               matrix(c(7:10, 10, 9, 11, 11, 8, 7:10, 3:4, 11, 1:2, 5:6), 10))
  expect_warning(
    expect_equal(
      SortTree(as.phylo(10, 6), how = "unrecognized"),
      SortTree(as.phylo(10, 6))
    )
  )
  
  start <- makeNodeLabel(as.phylo(10, 6))
  if(interactive()) {
    oPar <- par(mfrow = c(2, 1), mar = rep(0.5, 4), cex = 0.9)
    on.exit(par(oPar))
    plot(start, show.node.label = TRUE, xpd = NA)
    nodelabels()
    plot(SortTree(start), show.node.label = TRUE, xpd = NA)
    nodelabels()
  }
  expect_equal(SortTree(start)[["node.label"]], paste0("Node", c(1:3, 5, 4)))
  
  tipSorted <- SortTree(as.phylo(10, 6), how = "tip")
  children <- tipSorted[["edge"]][, 2]
  expect_equal(children[children < 7],
               order(TipLabels(tipSorted))[c(1, 2, 6, 3, 4, 5)])
  
  tipSorted <- SortTree(as.phylo(10, 6), how = "tip",
                        order = paste0("t", c(6, 4, 7, 2)))
  children <- tipSorted[["edge"]][, 2]
  expect_equal(children[children < 7],
               order(TipLabels(tipSorted))[c(6, 2, 4, 3, 5, 1)])
  
  
  unrooted5 <- UnrootTree(PectinateTree(5))
  unrooted5$edge.length <- 1:7
  sorted5 <- SortTree(unrooted5)
  expect_equal(sorted5$edge,
               matrix(c(6, 7, 8, 8, 7, 6, 6,
                        7, 8, order(sorted5[["tip.label"]], decreasing = TRUE)),
                      ncol = 2))
  expect_equal(sorted5[["edge.length"]], c(3, 5, 7, 6, 4, 2, 1))
  expect_equal("cladewise", attr(SortTree(PectinateTree(5)), "order"))
  expect_equal(SortTree(sorted5, "tip",
                        paste0("t", c(1, 2, 4, 5, 3)))[["edge.length"]],
               c(1:3, 5:7, 4))

  
  skip_if_not_installed("vdiffr", "1.0")
  vdiffr::expect_doppelganger("sorted-tree", function() {
    par(mar = rep(0, 4))
    plot(SortTree(UnrootTree(CollapseNode(BalancedTree(17), c(28, 31, 32)))))
    edgelabels(adj = c(1, 1/2))
  })
})


test_that("SortTree.multiPhylo()", {
  t1 <- as.phylo(123, 12)
  t2 <- as.phylo(921, 12)

  expect_identical(structure(list(SortTree(t1), SortTree(t2)),
                             class = "multiPhylo"),
                   SortTree(c(t1, t2)))
})

#' Convert object to `multiPhylo` class
#'
#' Converts representations of phylogenetic trees to an object of the "ape"
#' class `multiPhylo`.
#' @param x Object to be converted
#' @return `as.multiPhylo` returns an object of class `multiPhylo`
#' @examples
#' as.multiPhylo(BalancedTree(8))
#' as.multiPhylo(list(BalancedTree(8), PectinateTree(8)))
#' data("Lobo")
#' as.multiPhylo(Lobo.phy)
#' @family utility functions
#' @export
as.multiPhylo <- function(x) UseMethod("as.multiPhylo")

#' @rdname as.multiPhylo
#' @export
as.multiPhylo.phylo <- function(x) c(x)


#' @rdname as.multiPhylo
#' @export
as.multiPhylo.list <- function(x) structure(x, class = "multiPhylo")

#' @rdname as.multiPhylo
#' @return `as.multiPhylo.phyDat()` returns a list of trees, each corresponding
#' to the partitions implied by each non-ambiguous character in `x`.
#' @importFrom ape read.tree
#' @export
as.multiPhylo.phyDat <- function(x) {
  at <- attributes(x)
  cont <- at[["contrast"]]
  if ("-" %fin% colnames(cont)) {
    cont[cont[, "-"] > 0, ] <- 1
  }
  ambiguous <- rowSums(cont) != 1
  labels <- names(x)

  mat <- matrix(unlist(x), length(x), byrow = TRUE)
  mat[ambiguous[mat]] <- NA
  mat <- apply(mat, 2, function(x) {
    uniques <- table(x) == 1
    x[x %fin% names(uniques[uniques])] <- NA
    x
  })
  structure(apply(mat, 2, function(split) {
    a <- !is.na(split)
    aSplit <- split[a]
    tokens <- unique(aSplit)
    aLabels <- labels[a]
    if (length(tokens) == 1L) {
      read.tree(text = paste0("(", paste(aLabels, collapse = ", "), ");"))
    } else {
      read.tree(text = paste0("((",
                              paste(vapply(unique(aSplit), function(token) {
                                paste(aLabels[aSplit == token], collapse = ", ")
                              }, character(1)), collapse = "), ("),
                              "));"))
    }
  })[at[["index"]]], class = "multiPhylo")
}

#' @rdname as.multiPhylo
#' @importFrom ape read.tree
#' @export
as.multiPhylo.Splits <- function(x) {
  labels <- TipLabels(x)
  structure(apply(as.logical(x), 1, function(a) {
    read.tree(text = paste0(
      "((",
      paste0(labels[a], collapse = ","),
      "),(",
      paste0(labels[!a], collapse = ","),
      "));"))
  }), class = "multiPhylo")
}

#' @importFrom methods new setClass
setClass("Splits", representation("raw"))

#' Convert object to `Splits`
#'
#' `as.Splits()` converts a phylogenetic tree to a `Splits` object representing
#' its constituent bipartition splits.
#'
#' @param x Object to convert into splits: perhaps a tree of class
#' \code{\link[ape:read.tree]{phylo}}.
#' If a logical matrix is provided, each row will be considered as a separate
#' split.
#' If an integer matrix is provided, entries assigned the same integer will
#' be assigned to the same split.
#'  
#' @param tipLabels Character vector specifying sequence in which to order
#' tip labels.  Label order must (currently) match to combine or compare separate
#' `Splits` objects.
#' @param \dots Presently unused.
#'
#' @return `as.Splits()` returns an object of class `Splits`, or
#' (if `asSplits = FALSE`) a two-dimensional array of `raw` objects, with
#' each bit specifying whether or not the leaf corresponding to the respective
#' bit position is a member of the split.
#' Splits are named according to the node at the non-root end of the edge that
#' defines them. In rooted trees, the child of the rightmost root edge names
#' the split.
#'
#' @template MRS
#'
#' @examples
#' splits <- as.Splits(BalancedTree(letters[1:6]))
#' summary(splits)
#' TipsInSplits(splits)
#' summary(!splits)
#' TipsInSplits(!splits)
#'
#' length(splits + !splits)
#' length(unique(splits + !splits))
#'
#' summary(c(splits[[2:3]], !splits[[1:2]]))
#'
#' moreSplits <- as.Splits(PectinateTree(letters[6:1]), tipLabel = splits)
#' print(moreSplits, details = TRUE)
#' match(splits, moreSplits)
#' moreSplits %in% splits
#' 
#' as.Splits("....**", letters[1:6])
#'
#' @family Splits operations
#' @name Splits
#' @export
as.Splits <- function(x, tipLabels = NULL, ...) UseMethod("as.Splits")

#' @rdname Splits
#' @param asSplits Logical specifying whether to return a `Splits` object,
#'   or an unannotated two-dimensional array (useful where performance is
#'   paramount).
#' @export
as.Splits.phylo <- function(x, tipLabels = NULL, asSplits = TRUE, ...) {
  
  if (length(tipLabels) && !identical(x[["tip.label"]], tipLabels)) {
    x <- RenumberTips(x, tipLabels)
  }
  
  edge <- x[["edge"]]
  nEdge <- dim(edge)[[1]]
  order <- attr(x, "order")[[1]]
  edgeOrder <- if (length(order) == 0) {
    postorder_order(edge)
  } else {
    switch(order,
           "preorder" = rev(seq_len(nEdge)),
           "postorder" = seq_len(nEdge),
           postorder_order(edge))
  }

  # Return:
  edge_to_splits(edge, edgeOrder, tipLabels = x[["tip.label"]],
                 asSplits = asSplits, nTip = NTip(x), ...)
}

#' Efficiently convert edge matrix to splits
#' 
#' Wrapper for internal C++ function for maximum efficiency.
#' Improper input may crash R.  Behaviour not guaranteed.
#' It is advisable to contact the package maintainers before
#' relying on this function.
#' 
#' @template edgeParam
#' @param edgeOrder Integer vector such that `edge[edgeOrder, ]` returns a
#' postorder ordering of edges.
#' @param nTip Integer specifying number of leaves in tree.
#' @inheritParams Splits
#' @return `edge_to_splits()` uses the same return format as `as.Splits()`.
#' 
#' @seealso [`as.Splits()`][Splits] offers a safe access point to this
#' function that should be suitable for most users.
#' 
#' @family C++ wrappers
#' @export
edge_to_splits <- function(edge, edgeOrder, tipLabels = NULL, asSplits = TRUE,
                           nTip = NTip(edge), ...) {
  splits <- cpp_edge_to_splits(edge, edgeOrder - 1L, nTip)
  nSplits <- dim(splits)[[1]]

  # Return:
  if (asSplits) {
    structure(splits,
              nTip = nTip,
              tip.label = tipLabels,
              class = "Splits")
  } else {
    splits
  }
}

#' @rdname Splits
#' @export
as.Splits.multiPhylo <- function(x, tipLabels = unique(unlist(TipLabels(x))),
                                  asSplits = TRUE, ...) {
  lapply(x, as.Splits.phylo, tipLabels = tipLabels, asSplits = asSplits)
}

#' @rdname Splits
#' @export
as.Splits.Splits <- function(x, tipLabels = NULL, ...) {
  if (is.null(tipLabels)) {
    # Nothing needs doing.
    # Return:
    x
  } else {
    tipLabels <- TipLabels(tipLabels, single = TRUE)
    oldLabels <- attr(x, "tip.label")
    if (is.null(oldLabels)) {
      nTip <- attr(x, "nTip")
      if (length(tipLabels) == nTip) {
        attr(x, "tip.label") <- tipLabels
        x
      } else {
        stop(length(tipLabels), " labels provided; expecting ", nTip)
      }
    } else if (!identical(oldLabels, tipLabels)) {
      nTip <- attr(x, "nTip")
      if (length(x) == 0) {
        attr(x, "tip.label") <- tipLabels
        x
      } else {
        if (all(oldLabels %in% tipLabels) && all(tipLabels %in% oldLabels)) {
          as.Splits.logical(as.logical(x, tipLabels = tipLabels)
                            [, match(tipLabels, oldLabels), drop = FALSE],
                            tipLabels = tipLabels)
        } else {
          stop("Old and new labels must match")
        }
      }
    } else {
      x
    }
  }
}

#' @rdname Splits
#' @export
as.Splits.list <- function(x, tipLabels = NULL, asSplits = TRUE, ...) {
  if (inherits(x[[1]], "phylo")) {
    if (is.null(tipLabels)) {
      tipLabels <- unique(unlist(TipLabels(x)))
    }
    lapply(x, function(sp) as.Splits(
      sp,
      tipLabels = intersect(tipLabels, TipLabels(sp)),
      asSplits = asSplits
      ))
  } else if (inherits(x[[1]], "Splits")) {
    if (is.null(tipLabels)) {
      tipLabels <- attr(x, "tip.label")
      if (is.null(tipLabels)) {
        tipLabels <- attr(x[[1]], "tip.label")
      }
    }
    lapply(x, as.Splits, tipLabels = tipLabels, asSplits = asSplits)
  } else {
    stop("Unsupported list type; first item has class '",
         paste0(class(x[[1]]), collapse = ", "), "'")
  }
}

#' @rdname Splits
#' @export
as.Splits.matrix <- function(x, tipLabels = NULL, ...) {
  if (all(c("edge", "Nnode") %in% rownames(x))) {
    col1 <- x[, 1]
    if (is.list(col1)) {
      if (is.null(tipLabels)) {
        tipLabels <- col1[["tip.label"]]
        if (is.null(tipLabels)) {
          nTip <- dim(col1[["edge"]])[1] - col1[["Nnode"]] + 1L
          tipLabels <- seq_len(nTip)
          x <- rbind(x, replicate(ncol(x), list(tip.label = tipLabels)))
          rownames(x)[nrow(x)] <- "tip.label"
        }
      }
      lapply(seq_len(ncol(x)), function(i) {
        as.Splits(structure(x[, i], class = "phylo"), tipLabels = tipLabels)
      })
    } else {
      stop("Unsupported matrix. Columns should correspond to trees.")
    }
  } else if (is.numeric(x) && dim(x)[2] == 2) {
    edge_to_splits(x, postorder_order(x),
                   tipLabels = tipLabels, asSplits = TRUE, ...)
  } else {
    NextMethod()
  }
}

#' @rdname Splits
#' @export
as.Splits.integer <- function(x, tipLabels = NULL, ...) {
  membership <- t(vapply(sort(unique(x)), `==`, logical(length(x)), x))
  if (dim(membership)[[1]] == 2) {
    as.Splits.logical(membership[1, ], tipLabels = tipLabels, ...)
  } else {
    as.Splits.logical(membership, tipLabels = tipLabels, ...)
  }
}

#' @rdname Splits
#' @export
as.Splits.numeric <- function(x, tipLabels = NULL, ...) {
  if (all(x == as.integer(x))) {
    as.Splits.integer(as.integer(x))
  } else {
    stop("no applicable method for 'as.Splits' applied to an object of class \"",
         class(x), "\"")
  }
}

#' @rdname Splits
#' @export
as.Splits.logical <- function(x, tipLabels = NULL, ...) {
  dimX <- dim(x)
  
  if (is.null(dimX)) {
    nTip <- length(x)
    ret <- if (nTip == 0L) {
      matrix(raw(0), 0, 0)
    } else {
      pack_splits_logical_vec(x)
    }
  } else {
    nTip <- dimX[2L]
    ret <- `rownames<-`(pack_splits_logical(x), dimnames(x)[[1]])
  }
  
  if (is.null(tipLabels)) {
    # TODO when require R 4.1
    # tipLabels <- TipLabels(x) %||% TipLabels(nTip)
    tipLabels <- TipLabels(x)
    if (is.null(tipLabels)) {
      tipLabels <- TipLabels(nTip)
    }
  } else {
    tipLabels <- TipLabels(tipLabels)
  }
  
  structure(
    ret,
    nTip = nTip,
    tip.label = tipLabels,
    class = "Splits"
  )
}

#' @rdname Splits
#' @export
as.Splits.character <- function(x, tipLabels = NULL, ...) {
  nTip <- nchar(x[[1]])
  
  if (is.null(tipLabels)) {
    tipLabels <- TipLabels(x)
    if (length(tipLabels) == 1 && tipLabels == x) {
      tipLabels <- paste0("t", seq_len(nTip))
    }
  } else {
    tipLabels <- TipLabels(tipLabels)
  }
  
  sp <- .vapply(gregexpr("*", x, fixed = TRUE), tabulate, integer(nTip),
                nTip) > 0
  structure(t(.apply(sp, 2, as.Splits.logical)),
    nTip = nTip,
    tip.label = tipLabels,
    class = "Splits"
  )
}

#' @rdname Splits
#' @export
as.logical.Splits <- function(x, tipLabels = attr(x, "tip.label"), ...) {
  nTip <- attr(x, "nTip")
  if (dim(x)[1] == 0) {
    ret <- matrix(logical(0), 0, nTip)
  } else {
    ret <- matrix(as.logical(rawToBits(t(x))),
                  nrow = nrow(x), byrow = TRUE)[, seq_len(nTip), drop = FALSE]
  }
  dimnames(ret) <- list(rownames(x), tipLabels)
  ret
}

#' @family Splits operations
#' @export
print.Splits <- function(x, details = FALSE, ...) {
  nTip <- attr(x, "nTip")
  tipLabels <- attr(x, "tip.label")
  count <- attr(x, "count")
  trivial <- TrivialSplits(x)
  cat(dim(x)[1], "bipartition", ifelse(dim(x)[1] == 1, "split", "splits"),
      if(any(trivial)) paste0("(", sum(trivial), " trivial)"),
      "dividing", nTip,
      if(is.null(tipLabels)) {
        "unlabelled tips."
      } else {
        if (nTip) {
          if (nTip == 1) {
            paste("tip,", tipLabels[[1]])
          } else {
            paste("tips,", tipLabels[[1]], "..", tipLabels[[nTip]])
          }
        } else {
          "tips"
        }
      }
      )
  if (details) {
    splitNames <- rownames(x)
    if (!is.null(splitNames)) {

      nameLengths <- vapply(splitNames, nchar, 0)
      namePads <- vapply(nameLengths, function(thisLength)
        paste0(rep.int(" ", max(nameLengths) - thisLength), collapse=""), character(1))
      splitNames <- paste0(splitNames, namePads)
    } else {
      splitNames <- character(length(x))
      nameLengths = 0L
    }
    if (length(splitNames) > 0) {
      cat("\n ", paste0(rep.int(" ", max(nameLengths)), collapse = ""),
        paste0(rep_len(c(1:9, " "), nTip), collapse = ""))
      
      
      nSplits <- dim(x)[[1]]
      splitCounts <- if (!is.null(count)) {
        if (length(count) != nSplits) {
          warning("\"count\" attribute does not match number of splits")
        }
        paste0("\UD7 ", count)
      } else {
        rep("", nSplits)
      }
      
      for (i in seq_len(nSplits)) {
        split <- x[i, , drop = FALSE]
        cat("\n", splitNames[i], "",
            paste(ifelse(as.logical(rawToBits(split)[seq_len(nTip)]), "*", "."),
                  collapse = ""),
            splitCounts[i])
      }
    }
  }
}

#' @family Splits operations
#' @export
sort.Splits <- function(x, decreasing = TRUE, ...) {
  newOrder <- order(x, decreasing = decreasing, ...)
  count <- attr(x, "count")
  if (is.null(count)) {
    x[[newOrder]]
  } else {
    structure(x[[newOrder]], count = count[newOrder])
  }
}

# Underpins `order`
#' @family Splits operations
#' @export
xtfrm.Splits <- function(x) {
  count <- attr(x, "count")
  splitRanking <- as.integer(x)
  if (is.null(count)) {
    splitRanking
  } else {
    count + (splitRanking / max(splitRanking))
  }
}

#' @family Splits operations
#' @importFrom utils head
#' @export
head.Splits <- function(x, n = 6L, ...) {
  if (n < 0) {
    n <- max(0L, length(x) + n)
  }
  if (length(x) > n) {
    x[[seq_len(n)]]
  } else {
    x
  }
}

#' @family Splits operations
#' @importFrom utils tail
#' @export
tail.Splits <- function(x, n = 6L, ...) {
  if (n < 0) {
    n <- max(0L, length(x) + n)
  }
  if (length(x) > n) {
    x[[length(x) + 1L - rev(seq_len(n))]]
  } else {
    x
  }
}

#' @family Splits operations
#' @export
summary.Splits <- function(object, ...) {
  print(object, details = TRUE, ...)
  nTip <- attr(object, "nTip")
  tipLabels <- attr(object, "tip.label")
  if (is.null(tipLabels)) {
    cat("\n\nTips not labelled.")
  } else {
    if (length(tipLabels) > 0) {
      cat("\n\n", paste0("Tip ", seq_len(nTip), ": ", tipLabels,
                        "\t", c(character(4L), "\n")[seq_len(min(nTip, 5L))]))
    }
  }
}

#' @family Splits operations
#' @export
names.Splits <- function(x) rownames(x)


#' @family Splits operations
#' @export
`names<-.Splits` <- `rownames<-`


#' @family Splits operations
#' @export
as.character.Splits <- function(x, ...) {
  tipLabels <- attr(x, "tip.label")
  if (is.null(tipLabels)) {
    tipLabels <- paste0("t", seq_len(attr(x, "nTip")))
  }
  logx <- as.logical(x)
  ret <- splits_to_char(logx, tipLabels)
  names(ret) <- rownames(logx)
  ret
}

#' @family Splits operations
#' @export
as.phylo.Splits <- function(x, ...) {
  .PreorderTree(
    edge = splits_to_edge(x, NTip(x)),
    tip.label = TipLabels(x)
  )
}

#' @family Splits operations
#' @export
c.Splits <- function(...) {
  splits <- list(...)
  nTip <- unique(vapply(splits, attr, 1, "nTip"))
  if (length(nTip) > 1L) {
    stop("Splits must relate to identical tips.")
  }
  tips <- lapply(splits, attr, "tip.label")
  if (length(unique(lapply(tips, sort))) > 1L) {
    stop("All splits must bear identical tips")
  }
  tipLabels <- tips[[1]]
  splits <- c(splits[1], lapply(splits[seq_along(splits)[-1]], as.Splits,
                                  tipLabels = tipLabels))

  structure(do.call(rbind, splits),
            nTip = nTip,
            tip.label = tipLabels,
            class = "Splits")
}

#' @family Splits operations
#' @export
`+.Splits` <- function(...) c(...)


#' @family Splits operations
#' @export
`-.Splits` <- function(x, y) {
  x[[!x %in% y]]
}

#' @family Splits operations
#' @export
`[[.Splits` <- function(x, ..., drop = TRUE) {
  elements <- list(...)
  if (length(elements) == 0L) {
    x[]
  } else if (length(elements) == 1L) {
    ret <- x[elements[[1]], , drop = FALSE]
    at <- attributes(x)
    at[["dim"]] <- dim(ret)
    at[["dimnames"]][[1]] <- rownames(x)[elements[[1]]]
    attributes(ret) <- at
    ret
  } else {
    stop("Too many dimensions specified")
  }
}

.MaskSplits <- function(x) {
  mask_splits(x)
}

#' @family Splits operations
#' @method ! Splits
#' @export
`!.Splits` <- function(x) {
  not_splits(x)
}

#' @family Splits operations
#' @method & Splits
#' @export
`&.Splits` <- function(e1, e2) {
  and_splits(e1, e2)
}

#' @family Splits operations
#' @method | Splits
#' @export
`|.Splits` <- function(e1, e2) {
  or_splits(e1, e2)
}


#' Exclusive OR operation
#' @rdname xor
#' @param x,y Objects to be compared.
setGeneric("xor")

#' @family Splits operations
#' @rdname xor
#' @importFrom methods setMethod
#' @export
setMethod("xor", signature = representation(x = "Splits", y = "Splits"),
          function(x, y) xor_splits(x, y))

#' @export
t.Splits <- function(x) t(x[])

#' @family Splits operations
#' @export
length.Splits <- function(x) nrow(x)

#' @family Splits operations
#' @importFrom stats setNames
#' @export
duplicated.Splits <- function(x, incomparables, fromLast = FALSE,
                              withNames = TRUE, ...) {
  if (missing(incomparables)) {
    ret <- duplicated_splits(x, isTRUE(fromLast))
    if (withNames) names(ret) <- names(x)
    ret
  } else {
    dupX <- !x
    useOrig <- x[, 1] < dupX[, 1]
    dupX[useOrig, ] <- x[useOrig, ]
    
    if (dim(dupX)[2] == 1) {
      ret <- duplicated.default(dupX, incomparables = incomparables, 
                         fromLast = fromLast, ...)
      if (withNames) {
        setNames(ret, names(dupX))
      } else {
        ret
      }
    } else {
      ret <- duplicated.array(dupX, MARGIN = 1, incomparables = incomparables,
                              fromLast = fromLast, ...)
      if (withNames) {
        ret
      } else {
        as.logical(ret)
      }
    }
  }
}

#' @family Splits operations
#' @export
unique.Splits <- function(x, incomparables, ...) {
  at <- attributes(x)
  dups <- duplicated(x, incomparables, ...)
  x <- x[!dups, , drop = FALSE]
  at[["dim"]] <- dim(x)
  at[["dimnames"]] <- dimnames(x)
  attributes(x) <- at
  x
}

#' @family Splits operations
#' @export
rev.Splits <- function(x) {
  newOrder <- seq.int(dim(x)[1], 1)
  x[] <- x[newOrder, ]
  rownames(x) <- rownames(x)[newOrder]
  x
}

#' Polarize splits on a single taxon
#'
#' @param x Object that can be coerced into class [`Splits`].
#' @param pole Numeric, character or logical vector identifying tip that will
#' polarize each split.
#'
#' @return `PolarizeSplits()` returns a `Splits` object in which `pole` is
#' represented by a zero bit
#' @family Splits operations
#' @export
PolarizeSplits <- function(x, pole = 1L) {
  x <- as.Splits(x)
  nTip <- attr(x, "nTip")
  if (is.logical(pole)) {
    pole <- which(pole)[[1]]
  } else if (!is.numeric(pole)) {
    pole <- match(pole, attr(x, "tip.label"))[[1]]
  }
  if (is.na(pole)) {
    stop("Could not find `pole` in labels of `x`")
  }
  if (pole < 1) {
    stop("`pole` must be positive")
  }
  if (pole > nTip) {
    stop("`pole` must correspond to a leaf")
  }

  poleMask <- logical(8 * ceiling(nTip / 8))
  poleMask[pole] <- TRUE
  poleMask <- packBits(poleMask)
  flip <- !apply(t(x) & poleMask, 2, function(x) any(as.logical(x)))

  x[flip, ] <- !x[flip, ]
  remainder <- (8L - nTip) %% 8L
  if (remainder) {
    lastSplit <- dim(x)[2]
    endMask <- packBits(c(!logical(8L - remainder), logical(remainder)))
    x[, lastSplit] <- x[, lastSplit] & endMask
  }
  x
}

expect_splits_equal <- function(s1, s2, ...) {
  expect_equal(as.character(PolarizeSplits(s1)),
               as.character(PolarizeSplits(s2)), ...)
}

test_that("as.Splits()", {
  A <- FALSE
  B <- TRUE
  expect_equal(as.Splits(BalancedTree(0)), as.Splits(logical(0)))
  expect_equal(TipsInSplits(as.Splits(logical(0))), integer(0))
  expect_equal(strsplit(capture_output(summary(as.Splits(c(A, A, B, B)))),
                        "\n")[[1]],
               c("1 bipartition split dividing 4 tips, t1 .. t4",
                 "   1234",
                 "   ..** ", "",
                 " Tip 1: t1\t Tip 2: t2\t Tip 3: t3\t Tip 4: t4\t"))
  logical80 <- c(rep(TRUE, 40), rep(FALSE, 16), rep(TRUE, 24))
  expect_equal(strsplit(capture_output(print(
    as.Splits(logical80), detail = TRUE)), "\n")[[1]][3],
    paste0(c("   ", ifelse(logical80, "*", "."), " "), collapse = "")
  )
  expect_equal(as.logical(as.logical(as.Splits(logical80))), logical80)
  expect_equal(as.logical(as.Splits(c(A, A, B, B))),
               t(matrix(c(A, A, B, B), dimnames = list(paste0("t", 1:4)))))
  tree1 <- BalancedTree(letters[1:5])
  splits1 <- as.Splits(tree1)
  expect_equal(as.character(splits1),
               c("7" = "a b c | d e", "8" = "a b | c d e"))
  
  logicalSplits <- as.Splits(matrix(c(B, B, B, A, A,  B, B, A, A, A),
                                    nrow = 2, byrow = TRUE),
                             tipLabels = letters[1:5])
  rownames(logicalSplits) <- rownames(splits1)
  expect_equal(splits1, logicalSplits)
  expect_equal(splits1, as.Splits(splits1))
  
  pec6 <- as.Splits(PectinateTree(letters[1:6]))
  expect_equal(as.character(as.Splits(pec6)),
               c("9" = "c d e f | a b",
                 "10" = "d e f | a b c",
                 "11" = "e f | a b c d"))
  
  splitsC <- as.Splits(ape::read.tree(text = "(((a, d), e), (b, (f, c)));"))
  splitsD <- as.Splits(ape::read.tree(text = "((a, b, c), (d, (e, f)));"))
  splitsU <- as.Splits(ape::read.tree(text = "(a, b, c, d, e, f);"))
  oneSplit <- as.Splits(ape::read.tree(text = "((a, b, c), (d, e, f));"))
  expect_equal(attr(splitsC, "tip.label"),
               attr(as.Splits(splitsU, splitsC), "tip.label"))
  expect_equal(attr(splitsC, "tip.label"),
               attr(as.Splits(oneSplit, splitsC), "tip.label"))
  expect_equal(as.Splits(splitsD, splitsC),
               as.Splits(list(splitsC, splitsD, splitsU, oneSplit))[[2]])

  expect_equal(letters[1:5], colnames(as.logical(logicalSplits)))

  polytomy <- ape::read.tree(text = "(a, b, c, d, e);")
  expect_equal("0 bipartition splits dividing 5 tips, a .. e",
               capture_output(print(as.Splits(polytomy))))

  notPreorder <- structure(list(
    edge = structure(c(6L, 9L, 8L, 7L, 7L, 8L, 9L, 6L,
                       9L, 8L, 7L, 2L, 3L, 5L, 4L, 1L), .Dim = c(8L, 2L)),
    Nnode = 4L, tip.label = 1:5), class = "phylo", order = "cladewise")
  expect_equal(c("7" = packBits(c(A, B, B, A, A, rep(FALSE, 3))),
                 "8" = packBits(c(A, B, B, A, B, rep(FALSE, 3)))),
               as.Splits(notPreorder)[, 1])
  
  expect_equal(as.Splits(list(BalancedTree(3:9), BalancedTree(7:1))),
               list(as.Splits(BalancedTree(3:9)),
                    as.Splits(BalancedTree(7:1),
                              tipLabels = as.character(c(3:7, 2:1)))))
  
  expect_equal(as.Splits(c(BalancedTree(3:9), BalancedTree(7:1))),
               list(as.Splits(BalancedTree(3:9)),
                    as.Splits(BalancedTree(7:1),
                              tipLabels = as.character(c(3:7, 2:1)))))
  
  noNode <- structure(
    list(edge = structure(integer(0), dim = c(0L, 2L)),
         Nnode = 0L, tip.label = "a"), order = "preorder", class = "phylo")
  expect_equal(as.Splits(noNode), as.Splits(SingleTaxonTree("a")))
})

test_that("as.Splits.phylo()", {
  rootedStar <- structure(list(edge = structure(c(7L, 8L, 8L, 8L, 8L, 8L, 7L,
                                                  8L, 1L, 2L, 3L, 4L, 5L, 6L),
                                                .Dim = c(7L, 2L)), Nnode = 2L,
                               tip.label = letters[1:6]), class = "phylo")

  rootedStar2 <- structure(list(edge = structure(c(8L, 8L, 8L, 8L, 8L, 7L, 7L,
                                                   1L, 2L, 3L, 4L, 5L, 6L, 8L),
                                                .Dim = c(7L, 2L)), Nnode = 2L,
                               tip.label = letters[1:6]), class = "phylo")
  nasty <- structure(list(edge = structure( # Danger: Do not plot!
    c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
      5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
    .Dim = c(12, 2)), Nnode = 5L, tip.label = letters[1:8]), class = "phylo")


  expect_equal(c(0L, 1L), dim(as.Splits(rootedStar)))
  expect_equal(c(0L, 1L), dim(as.Splits(rootedStar2)))
  expected <- as.raw(as.Splits(Preorder(nasty)))
  actual <- as.raw(as.Splits(nasty))
  expect_equal(length(expected), length(actual))
  expect_equal(length(actual), length(unique(actual)))
  expect_true(all(actual %in% expected))
  expect_true(all(expected %in% actual))

  expect_equal(c(2L, 1L), dim(as.Splits(PectinateTree(5L))))
  expect_equal(c(2L, 1L), dim(as.Splits(UnrootTree(PectinateTree(5L)))))

  twoCherriesInGrass <- PectinateTree(3) + PectinateTree(3)
  expect_equal(c(2L, 1L), dim(as.Splits(twoCherriesInGrass)))

  unresolvedRoot <- CollapseNode(twoCherriesInGrass, 8)
  expect_equal(c(1L, 1L), dim(as.Splits(unresolvedRoot)))
  expect_equal(c(3L, 1L), dim(as.Splits(BalancedTree(6))))


  expect_equal(c(61L, 8L), dim(as.Splits(PectinateTree(64L))))
  expect_equal(c(62L, 9L), dim(as.Splits(PectinateTree(65L))))
  expect_equal(c(125L, 16L), dim(as.Splits(PectinateTree(128L))))
  expect_equal(as.Splits(Postorder(PectinateTree(64L))),
               as.Splits(PectinateTree(64L)))
  expect_true(all(
    as.Splits(ape::reorder.phylo(BalancedTree(66L), "postorder")) %in% 
               as.Splits(BalancedTree(66L))))
  expect_false(all(
    as.Splits(ape::reorder.phylo(PectinateTree(66L), "postorder")) %in% 
               as.Splits(BalancedTree(66L))))

  # Should be viable on 32-bit systems
  expect_equal(c(7997L, 1000L), dim(as.Splits(PectinateTree(8000L))))

  bigTree <- PectinateTree(16384L) # Max size with int16_t on 64-bit systems
  bigMemory <- try(Postorder(bigTree$edge), silent = TRUE)
  if (inherits(bigMemory, "try-error")) {
    expect_error(dim(as.Splits(bigTree))) # Likely on 32-bit systems
  } else {
    rm(bigMemory)
    gc()
    expect_equal(c(16381L, 2048L), dim(as.Splits(bigTree)))
  }

})

test_that("as.Splits.multiPhylo()", {
  randomTreeIds <- c(30899669, 9149275, 12823175, 19740197, 31296318,
                     6949843, 30957991, 32552966, 22770711, 21678908)
  randomTrees <- as.phylo(randomTreeIds, 11L, seq_len(11L))

  expect_equal(as.Splits(randomTrees[[1]]),
               as.Splits(randomTrees)[[1]])
})

test_that("as.Splits.Splits()", {
  # n tip > 8L
  splitsA <- as.Splits(ape::read.tree(text="((((a, b, c, c2), g), h), (d, (e, f)));"))
  splitsB <- as.Splits(ape::read.tree(text="(((((a, b), (c, c2)), h), g), (d, e, f));"))
  expectedSplits <- structure(matrix(as.raw(c(0x3f, 0x2f, 0x0f, 0x03, 0x0c, 0,
                                              0, 0, 0, 0)), ncol=2), nTip = 9,
                              tip.labels = TipLabels(splitsA), class="Splits")
  actualSplits <- as.Splits(splitsB, splitsA)
  expect_true(all(expectedSplits %in% actualSplits))
  expect_true(all(actualSplits %in% expectedSplits))
  expect_equal(NSplits(expectedSplits), NSplits(actualSplits))

  splitsABare <- splitsA
  attr(splitsABare, "tip.label") <- NULL
  expect_error(as.Splits(splitsABare, 1:2))
  expect_equal(splitsA, as.Splits(splitsABare, splitsA))
})

test_that("as.Splits.matrix()", {
  expect_error(as.Splits(matrix(1, 2, 2,
                                dimnames = list(c("edge", "Nnode"), NULL))))
  trees <- list(BalancedTree(8), PectinateTree(8),
                CollapseNode(BalancedTree(8), 10:13))
  arr <- sapply(1:3, function(i) trees[[i]])

  expect_error(as.Splits(arr[-2, ]))
  expect_equal(as.Splits(trees), as.Splits(arr))

  for (i in seq_along(trees)) {
    trees[[i]]$tip.label <- 1:8
  }
  expect_equal(as.Splits(trees), as.Splits(arr[-3, ]))
  
  exp <- as.Splits(BalancedTree(8))
  expect_equal(as.Splits(BalancedTree(8)$edge, TipLabels(BalancedTree(8))), exp)

  exp <- as.Splits(BalancedTree(8))
  attr(exp, "tip.label") <- NULL
  expect_equal(capture.output(summary(exp))[[9]], "Tips not labelled.")
  expect_equal(as.Splits(BalancedTree(8)$edge), exp)
})

test_that("as.Splits.edge()", {
  
  expect_one_of <- function(object, options) {
    val <- object
    ok  <- vapply(options, function(opt) identical(val, opt), logical(1L))
    
    expect(
      any(ok),
      sprintf(
        "Result is not one of the accepted options: %s",
        paste(vapply(options, toString, character(1L)), collapse = ", ")
      )
    )
    
    invisible(val)
  }
  
  # Test expect_one_of
  expect_success(expect_one_of(1, list(1, 2)))
  expect_failure(expect_one_of(0, list(1, 2)), "not one of")
  
  test <- unname(as.Splits(BalancedTree(4), asSplits = FALSE))
  expect_one_of(test, list(matrix(as.raw(0x0c)), matrix(as.raw(0x03))))
})

test_that("Logical splits don't get caught by as.matrix", {
  expect_equal(as.raw(as.Splits(matrix(FALSE, 2, 2))), raw(2))
  expect_equal(as.raw(as.Splits(matrix(TRUE, 3, 3))), as.raw(rep(7, 3)))
})

test_that("as.Splits.logical()", {
  FFTT <- c(FALSE, FALSE, TRUE, TRUE)
  a..d <- letters[1:4]
  expect_splits_equal(unname(as.Splits(BalancedTree(a..d))),
                      as.Splits(!FFTT, a..d))
  expect_splits_equal(as.Splits(FFTT, a..d), as.Splits(t(matrix(FFTT)), a..d))
  expect_splits_equal(as.Splits(FFTT), as.Splits(t(matrix(FFTT))))
})

test_that("as.Splits.integer()", {
  . <- FALSE
  X <- TRUE
  expect_splits_equal(as.Splits(c(rep(1L, 4)), tipLabels = letters[2:5]),
                      as.Splits(c(., ., ., .), tipLabels = letters[2:5]))
  expect_splits_equal(as.Splits(c(rep(1L, 4), rep(2L, 4))),
                      as.Splits(c(., ., ., ., X, X, X, X)))
  expect_splits_equal(as.Splits(c(rep(1L, 4), rep(2L, 4), 3L)),
                      as.Splits(rbind(c(X, X, X, X, ., ., ., ., .),
                                      c(., ., ., ., X, X, X, X, .),
                                      c(., ., ., ., ., ., ., ., X))))
  expect_splits_equal(as.Splits(c(rep(33, 4), 0, rep(-1, 3), 0)),
                      as.Splits(rbind(c(., ., ., ., ., X, X, X, .),
                                      c(., ., ., ., X, ., ., ., X),
                                      c(X, X, X, X, ., ., ., ., .))))
  expect_error(as.Splits(c(rep(1.45, 4), rep(sqrt(2), 3))), 
               "no applicable method")
})


test_that("as.Splits.character()", {
  a..f <- letters[1:6]
  a..i <- letters[1:9]
  . <- FALSE
  X <- TRUE
  
  expect_splits_equal(
    as.Splits("...***", a..f),
    as.Splits(c(., ., ., X, X, X), a..f)
  )
  expect_splits_equal(
    as.Splits(c("...***", "*.*.*."), a..f),
    as.Splits(rbind(c(., ., ., X, X, X), c(X, ., X, ., X, .)), a..f)
  )
  expect_splits_equal(
    as.Splits("****.....", a..i),
    as.Splits(c(X, X, X, X, ., ., ., ., .), a..i)
  )
  expect_splits_equal(
    as.Splits(c("****.....", ".*.*.*.*."), a..i),
    as.Splits(rbind(c(X, X, X, X, ., ., ., ., .),
                    c(., X, ., X, ., X, ., X, .)),
              a..i)
  )
  
  
  expect_equal(TipLabels(as.Splits("...***")), paste0("t", 1:6))
  
  expect_equal(
    TipLabels(as.Splits(structure("...***", tip.label = letters[1:6]))),
    letters[1:6]
  )
})

test_that("!.Splits() errors", {
  skip_if(Sys.getenv("USING_ASAN") != "")
  x <- as.Splits(as.phylo(8, 7))
  expect_error(!(structure(x, nTip = NULL)),
               "lacks nTip attrib")
})

test_that("xor, |, &.Splits()", {
  splits <- structure(as.raw(c(0x07, 0x03, 0x18, 0xe0, 0x60, 0x80)),
                      .Dim = c(3L, 2L), nTip = 9L, class = "Splits")
  expect_equal(splits & splits, splits)
  expect_equal(splits | splits, splits)
  expect_equal(xor(splits, splits),
               structure(matrix(raw(6), 3, 2), nTip = 9, class = "Splits"))
  
  mask <- structure(as.raw(c(0x0f, 0x00)), .Dim = c(1L, 2L), nTip = 9L,
                    class = "Splits")
  mask <- c(mask, mask, mask)
  expect_equal(t(splits[] & mask[]),
               structure(as.raw(c(0x07, 0x00, 0x03, 0x00, 0x08, 0x00)),
                         .Dim = c(2L, 3L)))
  expect_equal(splits & mask,
               structure(as.raw(c(0x07, 0x03, 0x08, 0x00, 0x00, 0x00)),
                         .Dim = c(3L, 2L), nTip = 9L, class = "Splits"))
  expect_equal(splits | mask, 
               structure(unclass(splits) | unclass(mask),
                         nTip = 9L, class = "Splits"))
  expect_equal(xor(splits, mask),
               (splits | mask) & !(splits & mask))
  
  Test <- function (s1, s2) {
    s1 <- as.Splits(s1)
    s2 <- as.Splits(s2)
    expect_equal((s1 & s2)[], s2[] & s1[], ignore_attr = TRUE)
    expect_equal((s1 | s2)[], s2[] | s1[], ignore_attr = TRUE)
    expect_equal(xor(s1,  s2)[], xor(s2[], s1[]), ignore_attr = TRUE)
    expect_equal((!s1)[], mask_splits(!(s1[])))
    
    expect_equal(attributes(!s1), attributes(s1))
    expect_equal(attributes(s1 | s2), attributes(s1))
    expect_equal(attributes(s1 & s2), attributes(s1))
    expect_equal(attributes(xor(s1, s2)), attributes(s1))
  }
  Test(PectinateTree(145), BalancedTree(145))
  Test(PectinateTree(64 * 3), BalancedTree(64 * 3))
  Test(PectinateTree(64 * 3 + 1), BalancedTree(64 * 3 + 1))
  Test(PectinateTree(64 * 3 - 1), BalancedTree(64 * 3 - 1))
})

test_that("empty as.X.Splits()", {
  someSplit <- as.Splits(BalancedTree(6))
  noSplit <- someSplit[[logical(0)]]
  expect_equal(matrix(logical(0), 0, 6,
                      dimnames = list(NULL, paste0("t", 1:6))),
               as.logical(noSplit))
  expect_equal(character(0), as.character(noSplit))
})


test_that("Renaming splits", {
  tree1 <- PectinateTree(1:8)
  tree2 <- BalancedTree(8:1)
  splits1 <- as.Splits(tree1)
  splits2 <- as.Splits(tree2)

  expect_error(as.Splits(splits1, tipLabel = 1:4))
  expect_error(as.Splits(splits1, tipLabel = LETTERS[1:8]))

  expect_equal(as.Splits(tree1, tipLabel = as.character(8:1)),
               as.Splits(tree1, tipLabel = tree2))

  expect_equal(as.Splits(tree1, tipLabel = tree2),
               as.Splits(splits1, tipLabel = tree2))

  as.Splits(tree1, tipLabel = tree2)[]
  as.Splits(splits1, tipLabel = tree2)[]

})

test_that("match.Splits()", {
  tree1 <- PectinateTree(1:8)
  tree2 <- PectinateTree(8:1)
  col2 <- as.Splits(CollapseNode(tree2, 13))

  expect_equal(5:1, match(as.Splits(tree1), as.Splits(tree2, tree1)))
  expect_equal(c(4, 3, NA, 2, 1), match(as.Splits(tree1, tree2), col2))
  expect_equal(c(4, 3, 999, 2, 1),
               match(as.Splits(tree1, tree2), col2, nomatch = 999))
  expect_equal(c(5, 4, 2, 1), match(col2, as.Splits(tree1, tree2)))
  
  expect_equal(match(sort(as.Splits(tree1)),
                     sort(as.Splits(tree2))), 5:1)
  expect_equal(match(sort(as.Splits(tree1)),
                     sort(as.Splits(tree2, tree1))), 5:1)
  expect_equal(match(sort(PolarizeSplits(as.Splits(tree1), 1)),
                     sort(PolarizeSplits(as.Splits(tree2, tree1), 1))), 1:5)
})

test_that("duplicated.Splits(internal)", {
  for (nTip in c(7:8, 23:26)) { # n %% BIN_SIZE ==0, ==1, >1
    x <- c(as.Splits(PectinateTree(nTip)), !as.Splits(BalancedTree(nTip)))
    
    expect_equal(duplicated(x, withNames = FALSE),
                 duplicated(x, FALSE, withNames = FALSE))
    
    expect_equal(as.logical(duplicated(x, fromLast = TRUE)),
                 as.logical(duplicated(x, FALSE, fromLast = TRUE)))
    
    expect_equal(names(duplicated(x, fromLast = TRUE)),
                 names(duplicated(x, FALSE, fromLast = TRUE)))
    
    expect_equal(duplicated(PolarizeSplits(x, nTip / 2)), duplicated(x))
    
    expect_equal(duplicated(x, fromLast = TRUE),
                 rev(duplicated(rev(x), fromLast = FALSE)))
  }
})

test_that("%in%.Splits()", {
  splits5 <- as.Splits(PectinateTree(5))
  x <- splits5
  table <- splits5[[-1]]
  expect_null(dim(duplicated(c(splits5, splits5[[-1]]))))
  expect_equal(names(splits5 %in% splits5[[-1]]), names(splits5))
  
  splits9 <- as.Splits(PectinateTree(9))
  expect_equal(names(splits9 %in% splits9[[-4]]), names(splits9))
})

test_that("print.Splits()", {
  sp4 <- as.Splits(BalancedTree(4))
  num <- names(sp4)
  expect_equal(
    capture.output(print(PolarizeSplits(sp4, 1), details = TRUE)),
    c( "1 bipartition split dividing 4 tips, t1 .. t4", "    1234",
       paste0(" ", num, "  **.. ")))
  expect_equal(capture.output(print(PolarizeSplits(sp4, 4), details = TRUE)),
               c( "1 bipartition split dividing 4 tips, t1 .. t4", "    1234",
                  paste0(" ", num, "  ..** ")))
  expect_equal(
    capture.output(print(structure(PolarizeSplits(sp4, 4), count = 2),
                         details = TRUE)),
    c( "1 bipartition split dividing 4 tips, t1 .. t4", "    1234",
                  paste0(" ", num, "  ..** \UD7 2")))
  
  expect_warning(expect_equal(
    capture.output(print(
      structure(PolarizeSplits(BalancedTree(5), 5), count = 2),
      details = TRUE)),
    c( "2 bipartition splits dividing 5 tips, t1 .. t5", "    12345",
       " 7  ...** \UD7 2",
       " 8  ..*** NA")))
  
  expect_equal(
    capture.output(summary(as.Splits(SingleTaxonTree()))),
    c("0 bipartition splits dividing 1 tip, t1", "", " Tip 1: t1\t")
  )
  
  expect_equal(
    capture.output(summary(as.Splits(ZeroTaxonTree()))),
    c("0 bipartition splits dividing 0 tips")
  )
})

test_that("head,tail.Splits()", {
  sp12 <- as.Splits(BalancedTree(12)) # 9 splits
  expect_equal(head(sp12), sp12[[1:6]])
  expect_equal(head(sp12, 2), sp12[[1:2]])
  expect_equal(head(sp12, 24), sp12)
  expect_equal(head(sp12, -7), sp12[[1:2]])
  
  expect_equal(tail(sp12), sp12[[4:9]])
  expect_equal(tail(sp12, 2), sp12[[8:9]])
  expect_equal(tail(sp12, 24), sp12)
  expect_equal(tail(sp12, -7), sp12[[8:9]])
  
  sp0 <- sp12 - sp12
  expect_equal(head(sp0), sp0)
  expect_equal(head(sp0, 1), sp0)
  expect_equal(tail(sp0), sp0)
  expect_equal(tail(sp0, 1), sp0)
  
  expect_equal(head(sp12, -24), sp0)
  expect_equal(tail(sp12, -24), sp0)
})

test_that("Split operations", {
  split1 <- as.Splits(c(rep(TRUE, 30), rep(FALSE, 70), rep(TRUE, 20)))
  notSplit1 <- as.Splits(c(rep(FALSE, 30), rep(TRUE, 70), rep(FALSE, 20)))
  expect_equal(!split1, notSplit1)
  expect_equal(split1, !notSplit1)

  expect_equal(notSplit1[1, ], (split1 + notSplit1)[2, ])
  expect_equal(split1 + notSplit1, !(notSplit1 + split1))

  expect_equal(notSplit1, (split1 + notSplit1)[[2]])
  expect_equal(notSplit1 + split1 + split1, (split1 + notSplit1)[[c(2, 1, 1)]])

  split2 <- as.Splits(c(rep(TRUE, 4), FALSE, FALSE))
  notSplit2 <- as.Splits(c(rep(FALSE, 4), TRUE, TRUE))

  expect_equal(split2, !notSplit2)
  expect_equal(!split2, notSplit2)
  expect_equal(notSplit2[1, ], (split2 + notSplit2)[2, ])
  expect_equal(split2 + notSplit2, !(notSplit2 + split2))

  expect_error(split1 + split2)

  namedSplits <- as.Splits(BalancedTree(8))
  if (isTRUE(all.equal(rownames(namedSplits), 
                       as.character(c(10:12, 14:15))))) {
    expect_true(TRUE)
  } else {
    expect_equal(rownames(namedSplits), as.character(c(11:15)))
  }
  
  expect_equal(split1[], `[[.Splits`(split1))
  expect_error(c(split1, notSplit1)[[2:1, 1]])
})

test_that("Split subtraction", {
  splits <- as.Splits(BalancedTree(8))
  expect_equal(splits[[c(1, 2, 4, 5)]], splits - splits[[3]])
})

test_that("unique.Splits()", {
  noSplits <- WithoutTrivialSplits(as.Splits(matrix(FALSE, 2, 2)))
  
  expect_equal(duplicated(noSplits), logical(0))
  expect_equal(unique(noSplits), noSplits)
})

test_that("Split combination", {
  tree1 <- BalancedTree(letters[1:5])
  splits1 <- as.Splits(tree1)
  tree2 <- PectinateTree(letters[1:5])
  splits12 <- c(splits1, as.Splits(tree2))
  tree3 <- PectinateTree(letters[c(1:3, 5, 4)])
  tree4 <- PectinateTree(letters[c(1:4, 6)])
  tree5 <- PectinateTree(letters[1:6])

  expect_equal(4L, length(splits12))
  expect_equal(c(FALSE, FALSE, TRUE, TRUE), as.logical(duplicated(splits12)))
  expect_equal(2L, length(unique(c(splits1, as.Splits(tree3)))))
  expect_error(c(splits1, as.Splits(tree4)))
  expect_error(c(splits1, as.Splits(tree5)))
  expect_equal(c(as.raw(splits1), as.raw(as.Splits(PectinateTree(5)))),
               unname(c(splits1,
                        as.Splits(RenumberTips(tree3, letters[1:5])))[, 1]))
  expect_equal(2L, length(unique(c(splits1, as.Splits(tree2)))))

  #TODO: Fully test splits with large (> 8 tip) trees
})

test_that("as.phylo.Splits() supports large trees", {
  tree3000 <- BalancedTree(3000)
  expect_no_error(splits3000 <- as.Splits(tree3000))
  result <- as.phylo(splits3000)
  
  # Verify it's a valid phylo object
  expect_s3_class(result, "phylo")
  expect_equal(length(result$tip.label), 3000)
  expect_equal(NTip(result), 3000)
})

test_that("as.phylo.Splits()", {
  Test <- function(tr) {
    expect_true(all.equal(as.phylo(as.Splits(tr)), UnrootTree(tr)))
  }
  Test(read.tree(text = "((a, b, c), (d, e, f, g));"))
  Test(read.tree(text = "((a, b, c), (d, e, (f, g)));"))
  Test(BalancedTree(64))
  Test(BalancedTree(65)) # multiple bins
  Test(PectinateTree(65*6)) # big!

  A <- FALSE
  B <- TRUE
  splits <- as.Splits(matrix(c(A, B, A, A, A, A, A, B,
                               A, B, A, A, A, A, B, B,
                               A, B, A, A, A, B, B, B,
                               A, B, A, A, B, B, B, B), 4, byrow = TRUE),
                      tipLabels = letters[1:8])
  expect_true(all.equal(as.phylo(splits),
                        read.tree(text = "(a, c, d, (e, (f, (g, (h, b)))));")))
})

test_that("PolarizeSplits()", {
  bal6 <- as.Splits(BalancedTree(6))
  expect_error(PolarizeSplits(bal6, "ERROR"))
  expect_error(PolarizeSplits(bal6, 0))
  expect_error(PolarizeSplits(bal6, 7))
  expect_equal(PolarizeSplits(bal6, 4), PolarizeSplits(bal6, 1:6 == 4))
})

test_that(".MaskSplits()", {
  x <- as.Splits(as.phylo(5, 5))
  expect_equal(.MaskSplits(x), mask_splits(x))
  
  skip_if(Sys.getenv("USING_ASAN") != "")
  expect_error(mask_splits(structure(x, nTip = NULL)),
               "lacks nTip attrib")
})

test_that("as.matrix(phylo)", {
  bal8 <- BalancedTree(8)
  sp8 <- as.Splits(bal8)
  expect_equal(as.matrix(bal8), 1 * as.logical(sp8))
  expect_equal(TipLabels(as.matrix(bal8)), TipLabels(bal8))
  expect_equal(as.logical(bal8), as.logical(sp8))
})

#' Colour a tree by topology
#'
#' `PaintTree()` assigns a colour to every edge, leaf, and internal node of a
#' tree such that sister clades occupy adjacent hue bands proportional to their
#' tip counts, and saturation grows from zero at the root to one at every tip.
#' The result is a list of vectors that correspond to 
#' [`plot.phylo()`][ape::plot.phylo]'s
#' `edge.color`, `tip.color`, and `node.color` arguments.
#'
#' Hue is allocated recursively from a 0–360° budget at the root.  At each
#' internal node, the parent's budget is partitioned across its descendant
#' edges in proportion to the number of leaves descended from each child; the
#' colour reported for an edge or node is the midpoint of its assigned range.
#' Saturation `s = (nTip - nDesc) / (nTip - 1)`, where `nDesc` is the number of
#' leaves descended from the node (including itself for tips), so the root is
#' achromatic (`s = 0`) and every leaf is fully saturated (`s = 1`).
#'
#' @template treeParam
#' @param palette Either a character string naming one of the built-in
#' palettes (`"default"`, `"protanopia"`, `"tritanopia"`,
#' or a function `function(h, s)` taking vectors of hues in
#' `[0, 360]` and saturations in `[0, 1]` and returning a character vector of
#' colours.  The `"protanopia"` and `"tritanopia"` palettes simulate
#' dichromatic perception of the default HCL spectrum using the
#' Viénot–Brettel–Mollon projection.
#'
#' @return A list with three character vectors of hex colours:
#' \describe{
#'   \item{`edgeCol`}{One colour per edge in `tree$edge`, taken from the
#'   child node's `(hue, saturation)`.}
#'   \item{`tipCol`}{One colour per leaf, indexed `1:NTip(tree)`.}
#'   \item{`nodeCol`}{One colour per internal node, indexed so that entry `i`
#'   corresponds to node `NTip(tree) + i`; the root (entry 1) is grey
#'   because its saturation is zero.}
#' }
#'
#' @examples
#' tree <- BalancedTree(1:8) + PectinateTree(9:14)
#' tree <- ape::bind.tree(BalancedTree(1:8), PectinateTree(9:12), 8, 4)
#' paint <- PaintTree(tree)
#' plot(tree,
#'      edge.color = paint$edgeCol,
#'      tip.color = paint$tipCol,
#'      edge.width = 3)
#'
#' # Colour-blind-safe variants
#' paintP <- PaintTree(tree, "protanopia")
#' plot(tree,
#'      edge.color = paintP$edgeCol,
#'      tip.color = paintP$tipCol,
#'      edge.width = 3)
#'
#' paintT <- PaintTree(tree, "tritanopia")
#' plot(tree,
#'      edge.color = paintT$edgeCol,
#'      tip.color = paintT$tipCol,
#'      edge.width = 3)
#'
#' # User-supplied palette function (greyscale)
#' grey_pal <- function(h, s) grey(1 - s * 0.8)
#' paintG <- PaintTree(tree, grey_pal)
#' plot(tree, edge.color = paintG$edgeCol, edge.width = 3)
#'
#' @seealso [`CladeSizes()`], [`DescendantEdges()`]
#' @template MRS
#' @family tree navigation
#' @importFrom grDevices hcl col2rgb rgb
#' @export
PaintTree <- function(tree, palette = "default") {
  edge <- tree[["edge"]]
  parent <- edge[, 1]
  child <- edge[, 2]
  nTip <- NTip(tree)
  nNode <- tree[["Nnode"]]
  maxNode <- nTip + nNode

  Pal <- .ResolvePalette(palette)

  if (nTip < 2L) {
    return(list(
      edgeCol = if (length(parent)) Pal(rep_len(180, length(parent)),
                                        rep_len(0, length(parent)))
                else character(0),
      tipCol = Pal(rep_len(180, nTip), rep_len(1, nTip)),
      nodeCol = Pal(rep_len(180, nNode), rep_len(0, nNode))
    ))
  }

  nDesc <- CladeSizes(tree, internal = FALSE)

  loH <- numeric(maxNode)
  hiH <- numeric(maxNode)
  root <- setdiff(parent, child)
  loH[root] <- 0
  hiH[root] <- 360

  # Order in which to process parents: preorder (root first, descendants later).
  # Within each parent, children are taken in their original `tree$edge` order
  # so the result does not depend on whether the tree was passed in preorder,
  # postorder, or any other valid edge ordering.
  parentOrder <- unique(parent[rev(PostorderOrder(tree))])
  groups <- split(seq_along(parent),
                  factor(parent, levels = parentOrder))

  for (g in groups) {
    p <- parent[g[1]]
    lo <- loH[p]
    hi <- hiH[p]
    kids <- child[g]
    w <- nDesc[kids]
    cuts <- lo + (hi - lo) * cumsum(c(0, w)) / sum(w)
    loH[kids] <- cuts[-length(cuts)]
    hiH[kids] <- cuts[-1]
  }

  hue <- (loH + hiH) / 2
  sat <- (nTip - nDesc) / (nTip - 1)

  list(
    edgeCol = Pal(hue[child], sat[child]),
    tipCol  = Pal(hue[seq_len(nTip)], rep_len(1, nTip)),
    nodeCol = Pal(hue[nTip + seq_len(nNode)], sat[nTip + seq_len(nNode)])
  )
}

.ResolvePalette <- function(palette) {
  if (is.function(palette)) {
    return(palette)
  }
  if (!is.character(palette) || length(palette) != 1L) {
    stop("`palette` must be a single string or a `function(h, s)`.")
  }
  choices <- c("default", "protanopia", "tritanopia")
  idx <- pmatch(tolower(palette), choices)
  if (is.na(idx)) {
    stop("`palette` must match one of: ",
         paste(choices, collapse = ", "),
         " (or be a function).")
  }
  switch(choices[idx],
         default    = .PaletteDefault,
         protanopia = .PaletteProtanopia,
         tritanopia = .PaletteTritanopia)
}

.PaletteDefault <- function(h, s) {
  hcl(h = h %% 360, c = s * 70, l = 70, fixup = TRUE)
}

.srgb2linear <- function(x) {
  ifelse(x <= 0.04045, x / 12.92, ((x + 0.055) / 1.055) ^ 2.4)
}

.linear2srgb <- function(x) {
  x <- pmin(pmax(x, 0), 1)
  ifelse(x <= 0.0031308, 12.92 * x, 1.055 * x ^ (1 / 2.4) - 0.055)
}

# Viénot, Brettel & Mollon (1999) sRGB projection for protanopia.
.protanopiaMatrix <- matrix(c(
  0.567, 0.433, 0.000,
  0.558, 0.442, 0.000,
  0.000, 0.242, 0.758
), nrow = 3, byrow = TRUE)

.tritanopiaMatrix <- matrix(c(
  0.950, 0.050, 0.000,
  0.000, 0.433, 0.567,
  0.000, 0.475, 0.525
), nrow = 3, byrow = TRUE)

.ApplyDichromat <- function(h, s, M) {
  base <- .PaletteDefault(h, s)
  rgbMat <- col2rgb(base) / 255
  lin <- matrix(.srgb2linear(as.numeric(rgbMat)), nrow = 3)
  proj <- M %*% lin
  out <- matrix(.linear2srgb(as.numeric(proj)), nrow = 3)
  rgb(out[1, ], out[2, ], out[3, ])
}

.PaletteProtanopia <- function(h, s) {
  .ApplyDichromat(h, s, .protanopiaMatrix)
}

.PaletteTritanopia <- function(h, s) {
  .ApplyDichromat(h, s, .tritanopiaMatrix)
}

test_that("PaintTree returns correct structure", {
  tree <- BalancedTree(15)
  paint <- PaintTree(tree)

  expect_named(paint, c("edgeCol", "tipCol", "nodeCol"))
  expect_type(paint$edgeCol, "character")
  expect_type(paint$tipCol, "character")
  expect_type(paint$nodeCol, "character")

  expect_length(paint$edgeCol, nrow(tree$edge))
  expect_length(paint$tipCol, NTip(tree))
  expect_length(paint$nodeCol, tree$Nnode)
})

test_that("PaintTree is deterministic", {
  tree <- BalancedTree(8)
  expect_identical(PaintTree(tree), PaintTree(tree))
})

test_that("PaintTree hue partitioning matches specification", {
  tree <- ape::read.tree(text = "(a, (b, (c, d)));")
  # NB this is a binary rooted tree (a, (b, (c, d))) with 4 tips
  # nDesc: a=1, b=1, c=1, d=1; (c,d)=2; (b,(c,d))=3; root=4
  # Root range 0-360; children weights a:1, (b,(c,d)):3
  #   a:        0-90 (mid 45)
  #   (b,(c,d)):90-360 (mid 225)
  # (b,(c,d)) range 90-360; children weights b:1, (c,d):2
  #   b:    90-180 (mid 135)
  #   (c,d):180-360 (mid 270)
  # (c,d) range 180-360; children c:1, d:1
  #   c: 180-270 (mid 225)
  #   d: 270-360 (mid 315)
  paint <- PaintTree(tree, palette = function(h, s) h)
  nTip <- NTip(tree)
  # tipCol is hue at each tip in tip order
  expect_equal(paint$tipCol[match("a", tree$tip.label)], 45)
  expect_equal(paint$tipCol[match("b", tree$tip.label)], 135)
  expect_equal(paint$tipCol[match("c", tree$tip.label)], 225)
  expect_equal(paint$tipCol[match("d", tree$tip.label)], 315)
  # Root is the first internal node (5); (b,(c,d)) is 6; (c,d) is 7
  expect_equal(paint$nodeCol[1], 180)   # root midpoint
  expect_equal(paint$nodeCol[2], 225)   # (b,(c,d)) midpoint
  expect_equal(paint$nodeCol[3], 270)   # (c,d) midpoint
})

test_that("PaintTree saturation invariants", {
  tree <- BalancedTree(8)
  sat <- PaintTree(tree, palette = function(h, s) s)

  # Tips fully saturated
  expect_equal(sat$tipCol, rep_len(1, NTip(tree)))
  # Root achromatic
  expect_equal(sat$nodeCol[1], 0)
  # Saturation strictly increases from root to tip along every edge:
  # for each edge, child's sat >= parent's sat
  parent <- tree$edge[, 1]
  child <- tree$edge[, 2]
  allSat <- c(rep_len(1, NTip(tree)), sat$nodeCol)
  expect_true(all(allSat[child] >= allSat[parent] - 1e-9))
})

test_that("PaintTree root colour is grey under default palette", {
  paint <- PaintTree(BalancedTree(8))
  root <- as.integer(col2rgb(paint$nodeCol[1]))
  expect_equal(root[1], root[2])
  expect_equal(root[2], root[3])
})

test_that("edgeCol always reports the colour of its child", {
  tree <- BalancedTree(10)
  paint <- PaintTree(tree)
  child <- tree$edge[, 2]
  nTip <- NTip(tree)
  isTip <- child <= nTip
  childCol <- character(length(child))
  childCol[isTip] <- paint$tipCol[child[isTip]]
  childCol[!isTip] <- paint$nodeCol[child[!isTip] - nTip]
  expect_equal(paint$edgeCol, childCol)
})

test_that("PaintTree preserves edge identity when $edge is not in preorder", {
  # Build a non-preorder tree by reversing the edge matrix without
  # renumbering any nodes.  Each row of edgeCol must still describe its
  # corresponding row of $edge.
  tree <- BalancedTree(10)
  rev_tree <- tree
  rev_tree$edge <- tree$edge[nrow(tree$edge):1, , drop = FALSE]
  attr(rev_tree, "order") <- NULL

  paint <- PaintTree(rev_tree)
  expect_length(paint$edgeCol, nrow(rev_tree$edge))

  child <- rev_tree$edge[, 2]
  nTip <- NTip(rev_tree)
  isTip <- child <= nTip
  childCol <- character(length(child))
  childCol[isTip] <- paint$tipCol[child[isTip]]
  childCol[!isTip] <- paint$nodeCol[child[!isTip] - nTip]
  expect_equal(paint$edgeCol, childCol)
})

test_that("PaintTree palette dispatch", {
  tree <- BalancedTree(8)
  defaultPaint <- PaintTree(tree, "default")
  protPaint <- PaintTree(tree, "Prot")  # pmatch / tolower
  triPaint <- PaintTree(tree, "TRITANOPIA")

  expect_false(identical(defaultPaint$tipCol, protPaint$tipCol))
  expect_false(identical(defaultPaint$tipCol, triPaint$tipCol))

  redPaint <- PaintTree(tree, function(h, s) rep("#ff0000", length(h)))
  expect_true(all(redPaint$edgeCol == "#ff0000"))
  expect_true(all(redPaint$tipCol == "#ff0000"))
})

test_that("PaintTree rejects bad palette arguments", {
  tree <- BalancedTree(4)
  expect_error(PaintTree(tree, "bogus"), "must match")
  expect_error(PaintTree(tree, c("default", "default")), "single string")
  expect_error(PaintTree(tree, 42), "single string")
})

test_that("PaintTree handles single-tip tree", {
  tree <- structure(list(
    edge = matrix(c(2L, 1L), ncol = 2),
    tip.label = "a",
    Nnode = 1L
  ), class = "phylo")
  paint <- PaintTree(tree)
  expect_length(paint$tipCol, 1)
  expect_length(paint$nodeCol, 1)
})

#' Tips contained within splits
#'
#' `TipsInSplits()` specifies the number of tips that occur within each
#' bipartition split in a `Splits` object.
#'
#' @param splits Object of class `Splits` or `phylo`.
#' @param keep.names Logical specifying whether to include the names of `splits`
#' in the output.
#' @param smallest Logical; if `TRUE`, return the number of leaves in the
#' smaller bipartition.
#' @param \dots Additional parameters to pass to `as.Splits()`.
#'
#' @return `TipsInSplits()` returns a named vector of integers, specifying the
#' number of tips contained within each split in `splits`.
#'
#' @examples
#' tree <- PectinateTree(8)
#' splits <- as.Splits(tree)
#' TipsInSplits(splits)
#'
#' plot(tree)
#' LabelSplits(tree, as.character(splits), frame = "none", pos = 3L, cex = 0.7)
#' LabelSplits(tree, TipsInSplits(splits), unit = " tips", frame = "none",
#'             pos = 1L)
#'
#' @family Splits operations
#' @export
TipsInSplits <- function(splits, keep.names = TRUE, smallest = FALSE, ...) {
  UseMethod("TipsInSplits")
}

#' @rdname TipsInSplits
#' @export
TipsInSplits.Splits <- function(splits, keep.names = TRUE, smallest = FALSE,
                                ...) {
  ret <- tips_in_splits(splits)
  if (smallest) {
    nTip <- NTip(splits)
    big <- ret > nTip / 2
    ret[big] <- nTip - ret[big]
  }
  if (keep.names) {
    names(ret) <- names(splits)
  }
  ret
}

#' @rdname TipsInSplits
#' @export
TipsInSplits.phylo <- function(splits, keep.names = TRUE, smallest = FALSE,
                               ...) {
  TipsInSplits(as.Splits(splits, ...),
               keep.names = keep.names,
               smallest = smallest)
}

#' @rdname TipsInSplits
#' @return `SplitImbalance()` returns a named vector of integers, specifying the
#' number of leaves within a split that are not "balanced" by a leaf outside it;
#' i.e. a split that divides leaves evenly has an imbalance of zero; one that
#' splits two tips from ten has an imbalance of 10 - 2 = 8.
#' @export
SplitImbalance <- function(splits, keep.names = TRUE, ...) UseMethod("SplitImbalance")

#' @export
SplitImbalance.NULL <- function(splits, keep.names, ...) NULL
  
#' @rdname TipsInSplits
#' @export
SplitImbalance.Splits <- function(splits, keep.names = TRUE, ...) {
  nTip <- NTip(splits)
  inSplit <- TipsInSplits(splits, keep.names = TRUE, ...)
  outSplit <- nTip - inSplit

  # Return:
  abs(inSplit - outSplit)
}

#' @rdname TipsInSplits
#' @export
SplitImbalance.phylo <- SplitImbalance.Splits

test_that("TipsInSplits() family", {
  test <- TipsInSplits(BalancedTree(letters[1:5]))
  expect_identical(test, c("7" = 3L, "8" = 2L, "9" = 2L)[names(test)])
  expect_equal(TipsInSplits(PectinateTree(7), smallest = TRUE),
               c("10" = 2, "11" = 3, "12" = 3, "13" = 2))
  expect_identical(TipsInSplits(PectinateTree(17), keep.names = FALSE), 15:2)

  test <- SplitImbalance(BalancedTree(7))
  expectation <- c("9" = 1L, "10" = 3L, "11" = 3L, "12" = 1L, "13" = 3L)
  expect_identical(test, expectation[names(test)])
})

#' Split matching
#'
#' `match()` returns a vector of the positions of (first) matches of splits in
#' its first argument in its second.
#' `%in%` is a more intuitive interface as a binary operator, which returns
#' a logical vector indicating whether there is a match or not for each
#' split in its left operand.
#'
#' @param x,table Object of class `Splits`.
#' @param nomatch Integer value that will be used in place of `NA` in the case
#' where no match is found.
#' @param incomparables Ignored. (Included for consistency with generic.)
# @param incomparables A vector of values that cannot be matched. Any value in
# `x` matching a value in this vector is assigned the `nomatch` value.
# For historical reasons, `FALSE` is equivalent to `NULL`.
#'
#' @return `match()` returns an integer vector specifying the position in
#' `table` that matches each element in `x`, or `nomatch` if no match is found.
#'
#' @examples
#' splits1 <- as.Splits(BalancedTree(7))
#' splits2 <- as.Splits(PectinateTree(7))
#'
#' match(splits1, splits2)
# Turns base functions into S3 generics that can handle `Splits` objects
# (as well as `integer64`s).  This follows equivalent functions in the
# "\pkg{bit64}" package.
#
#' @seealso Corresponding base functions are documented in
#' [`match()`][base::match].
#'
#' @aliases match,Splits,Splits-method
#' @family Splits operations
#' @keywords methods
#' 
#' @rdname match.Splits
#' @export
setMethod("match",
          signature(x = "Splits", table = "Splits"),
          function(x, table, nomatch, incomparables) {
            if(missing("nomatch")) {
              nomatch <- NA_integer_
            }
            nomatch <- as.integer(nomatch)
            if (length(nomatch) != 1L) {
              nomatch <- NA_integer_
            }
            vapply(seq_along(x), function(i) {
              ret <- which(table %in% x[[i]])
              if (length(ret) == 0) ret <- nomatch
              ret
            }, integer(1))
          })

.in.Splits <- function(x, table) {
  duplicated(c(x, table), fromLast = TRUE)[seq_along(x)]
}

#' @rdname match.Splits
setGeneric("match")


#' @rdname match.Splits
#' @export
#' @keywords methods
setMethod("%in%",
          signature(x = "Splits", table = "Splits"),
          .in.Splits)


#' @importFrom methods setOldClass
#' @importFrom ape as.phylo
setOldClass(c("phylo", "multiPhylo"))

#' Tree matching
#'
#' `match()` returns a vector of the positions of (first) matches of trees in
#' its first argument in its second.
#' `%in%` is a more intuitive interface as a binary operator, which returns
#' a logical vector indicating whether there is a match or not for each
#' tree in its left operand.
#'
#' @param x,table Object of class `phylo` or `multiPhylo`.
#' @param nomatch Integer value that will be used in place of `NA` in the case
#' where no match is found.
#' @param incomparables Ignored. (Included for consistency with generic.)
# @param incomparables A vector of values that cannot be matched. Any value in
# `x` matching a value in this vector is assigned the `nomatch` value.
# For historical reasons, `FALSE` is equivalent to `NULL`.
#'
#' @return `match()` returns an integer vector specifying the position in
#' `table` that matches each element in `x`, or `nomatch` if no match is found.
#'
#' @examples
#' tree1 <- BalancedTree(7)
#' trees <- c(PectinateTree(7), BalancedTree(7))
#'
#' match(tree1, trees)
#
#' @seealso Corresponding base functions are documented in
#' [`match()`][base::match].
#'
#' @family utility functions
#' @keywords methods
#' 
#' @rdname match.multiPhylo
#' @aliases match,phylo,phylo-method
#' @export
setMethod("match",
          signature(x = "phylo", table = "phylo"),
          function(x, table, nomatch = NA_integer_, incomparables = NULL) {
            if (all.equal(x, table)) {
              1L
            } else {
              if (missing("nomatch")) {
                NA_integer_
              } else {
                nomatch <- as.integer(nomatch)
                if (length(nomatch) != 1L) {
                  NA_integer_
                } else {
                  nomatch
                }
              }
            }
          })


#' @rdname match.multiPhylo
#' @export
setMethod("match",
          signature(x = "multiPhylo", table = "phylo"),
          function(x, table, nomatch = NA_integer_, incomparables = NULL) {
            if(missing("nomatch")) {
              nomatch <- NA_integer_
            }
            nomatch <- as.integer(nomatch[[1]])
            if (length(nomatch) != 1L) {
              nomatch <- NA_integer_
            }
            ifelse(x %in% table, 1L, nomatch)
          })

#' @rdname match.multiPhylo
#' @export
setMethod("match",
          signature(x = "phylo", table = "multiPhylo"),
          function(x, table, nomatch = NA_integer_, incomparables = NULL) {
            index <- table %in% x
            if (any(index)) {
              which.max(index)
            } else if (missing("nomatch")) {
              NA_integer_
            } else {
              nomatch <- as.integer(nomatch)
              if (length(nomatch) != 1L) {
                NA_integer_
              } else {
                nomatch
              }
            }
          })

#' @rdname match.multiPhylo
#' @aliases match,multiPhylo,multiPhylo-method
#' @export
setMethod("match",
          signature(x = "multiPhylo", table = "multiPhylo"),
          function(x, table, nomatch = NA_integer_, incomparables = NULL) {
            if (missing("nomatch")) {
              nomatch <- NA_integer_
            }
            nomatch <- as.integer(nomatch)
            if (length(nomatch) != 1L) {
              nomatch <- NA_integer_
            }
            
            vapply(x, function(i) {
              match(i, table, nomatch = nomatch)
            }, integer(1))
          })



#' @rdname match.multiPhylo
#' @export
#' @keywords methods
setMethod("%in%",
          signature(x = "multiPhylo", table = "multiPhylo"),
          function(x, table) {
            vapply(x, function (i) any(i %in% table), logical(1))
          })

#' @rdname match.multiPhylo
#' @export
#' @keywords methods
setMethod("%in%",
          signature(x = "multiPhylo", table = "phylo"),
          function(x, table) {
            vapply(x, function (i) all.equal(table, i), logical(1))
          }
          )

#' @rdname match.multiPhylo
#' @export
#' @keywords methods
setMethod("%in%",
          signature(x = "phylo", table = "multiPhylo"),
          function(x, table) {
            any(table %in% x)
          })

#' @rdname match.multiPhylo
#' @export
#' @importFrom ape all.equal.phylo
#' @keywords methods
setMethod("%in%",
          signature(x = "phylo", table = "phylo"),
          function(x, table) all.equal(x, table))

#' @rdname match.Splits
#' @param x,table Splits objects
#' @param return Which index to return: in `x`, in `table`, or both
#' @return `FirstMatchingSplit()` returns an integer
#'  (or length-2 integer if `return = "both"`) specifying the first split in `x`
#'  to have a match in `table` (`return = "x"`),
#'  or the index of that match (`return = "table"`).
#'  `nomatch` (default `0`) is returned in the absence of a match.
#' @export
FirstMatchingSplit <- function(x, table, nomatch,
                               return = c("x", "table", "both")) {
  if (!inherits(x, "Splits")) {
    x <- as.Splits(x)
  }
  table <- as.Splits(table, x)
  ij <- first_matching_split_pair(x, table)
  
  if (!missing(nomatch)) {
    ij[ij == 0] <- nomatch
  }
  
  # Return:
  return <- match.arg(return)
  switch(return,
         x     = ij[[1L]],
         table = ij[[2L]],
         both  = ij)
}

#' Add a tip to a phylogenetic tree
#'
#' `AddTip()` adds a tip to a phylogenetic tree at a specified location.
#'
#' `AddTip()` extends \code{\link[ape]{bind.tree}}, which cannot handle
#'   single-taxon trees.
#'
#' @template treeParam
#' @param where The node or tip that should form the sister taxon to the new
#' node.  To add a new tip at the root, use `where = 0`.  By default, the
#' new tip is added to a random edge.
#' @param label Character string providing the label to apply to the new tip.
#' @param nodeLabel Character string providing a label to apply to the newly
#' created node, if `tree$node.label` is specified.
#' @param edgeLength Numeric specifying length of new edge. If `NULL`,
#' defaults to `lengthBelow`.
# Notice added in v1.10.0.9001, 2024-02-20:
#' This will become the default behaviour in a
#' future release; please manually specify the desired behaviour in your code.
#' @param lengthBelow Numeric specifying length below neighbour at which to
#' graft new edge. Values greater than the length of the edge will result
#' in negative edge lengths. If `NULL`, the default, the new tip will be added
#' at the midpoint of the broken edge. If inserting at the root (`where = 0`),
#' a new edge of length `lengthBelow` will be inserted.
#' If `NA`, the new leaf will be attached adjacent to `where`; at internal
#' nodes, this will result in polytomy.
#' 
#' @param nTip,nNode,rootNode Optional integer vectors specifying number of tips
#' and nodes in `tree`, and index of root node.
#' Not checked for correctness: specifying values here yields a marginal speed
#' increase at the cost of code safety.
#'
#' @return `AddTip()` returns a tree of class `phylo` with an additional tip
#' at the desired location.
#'
#' @template MRS
#'
#' @seealso Add one tree to another: \code{\link[ape]{bind.tree}()}
#'
#' @examples
#' tree <- BalancedTree(10)
#' 
#' # Add a leaf below an internal node
#' plot(tree)
#' ape::nodelabels() # Identify node numbers 
#' node <- 15        # Select location to add leaf
#' ape::nodelabels(bg = ifelse(NodeNumbers(tree) == node, "green", "grey"))
#'
#' plot(AddTip(tree, 15, "NEW_TIP"))
#' 
#' # Add edge lengths for an ultrametric tree
#' tree$edge.length <- rep(c(rep(1, 5), 2, 1, 2, 2), 2)
#' 
#' # Add a leaf to an external edge
#' leaf <- 5
#' plot(tree)
#' ape::tiplabels(bg = ifelse(seq_len(NTip(tree)) == leaf, "green", "grey"))
#' 
#' plot(AddTip(tree, 5, "NEW_TIP", edgeLength = NULL))
#' 
#' # Create a polytomy, rather than a new node
#' plot(AddTip(tree, 5, "NEW_TIP", edgeLength = NA))
#' 
#' @keywords tree
#' @family tree manipulation
#' @export
AddTip <- function(tree,
                   where = sample.int(tree[["Nnode"]] * 2 + 2L, size = 1) - 1L,
                   label = "New tip",
                   nodeLabel = "",
                   edgeLength = 0,
                   lengthBelow = NULL,
                   nTip = NTip(tree),
                   nNode = tree[["Nnode"]],
                   rootNode = RootNode(tree)
) {
  newTipNumber <- nTip + 1L
  treeEdge <- tree[["edge"]]
  edgeLengths <- tree[["edge.length"]]
  hasLengths <- !is.null(edgeLengths)
  
  if (is.character(where)) {
    tmp <- match(where, TipLabels(tree))
    if (is.na(tmp)) {
      stop("No tip labelled '", where, "'")
    }
    where <- tmp
  }
  
  # Find the row of "where" before renumbering
  if (where < 1L || where == rootNode) {
    case <- 1L
  } else {
    insertionEdge <- which(treeEdge[, 2] == where)
    case <- if (where <= nTip) 2L else 3L
  }
  # case = 1 -> y is bound on the root of x
  # case = 2 -> y is bound on a tip of x
  # case = 3 -> y is bound on a node of x
  
  # Because in all situations internal nodes need to be
  # renumbered, they are changed to negatives first, and
  # nodes eventually added will be numbered sequentially
  nodes <- treeEdge > nTip
  treeEdge[nodes] <- nTip - treeEdge[nodes]  # -1, ..., -nTip
  nextNode <- -nNode - 1L
  rootNode <- nTip - rootNode
  # Determine now, before we overwrite lengthBelow
  addingNode <- is.null(lengthBelow) || 
    (!is.null(lengthBelow) && !is.na(lengthBelow)) ||
    case == 2 
  
  switch(case, { # case = 1 -> y is bound on the root of x
    if (addingNode) {
      treeEdge <- rbind(c(nextNode, treeEdge[[1]]),
                        treeEdge,
                        c(nextNode, newTipNumber))
    } else {
      treeEdge <- rbind(treeEdge, c(rootNode, newTipNumber))
    }
    if (hasLengths) {
      if (is.null(lengthBelow) || is.na(lengthBelow)) {
        lengthBelow <- 0
      }
      edgeLengths <- c(if (addingNode) lengthBelow, edgeLengths,
                       if(is.null(edgeLength)) lengthBelow else edgeLength)
    }
    rootNode <- nextNode
  }, { # case = 2 -> y is bound on a tip of x
    beforeInsertion <- seq_len(insertionEdge)
    treeEdge[insertionEdge, 2] <- nextNode
    treeEdge <- rbind(treeEdge[beforeInsertion, ],
                      c(nextNode, where),
                      c(nextNode, newTipNumber),
                      treeEdge[-beforeInsertion, ])
    if (hasLengths) {
      if (is.null(lengthBelow)) {
        lengthBelow <- edgeLengths[insertionEdge] / 2L
      } else if (is.na(lengthBelow)) {
        lengthBelow <- 0
      }
      edgeLengths <- c(edgeLengths[beforeInsertion[-insertionEdge]],
                       edgeLengths[insertionEdge] - lengthBelow,
                       lengthBelow,
                       if(is.null(edgeLength)) lengthBelow else edgeLength,
                       edgeLengths[-beforeInsertion])
    }
  }, { # case = 3 -> y is bound on a node of x
    beforeInsertion <- seq_len(insertionEdge)
    
    if (addingNode) {
      treeEdge <- rbind(treeEdge[beforeInsertion, ],
                        c(nextNode, newTipNumber),
                        c(nextNode, treeEdge[insertionEdge, 2]),
                        treeEdge[-beforeInsertion, ])
      treeEdge[insertionEdge, 2] <- nextNode
    } else {
      treeEdge <- rbind(treeEdge[beforeInsertion, ],
                        c(treeEdge[insertionEdge, 2], newTipNumber),
                        treeEdge[-beforeInsertion, ])
    }
    
    if (hasLengths) {
      if (is.null(lengthBelow)) {
        lengthBelow <- edgeLengths[insertionEdge] / 2L
      } else if (is.na(lengthBelow)) {
        lengthBelow <- 0
      }
      edgeLengths <- if (addingNode) {
        c(edgeLengths[beforeInsertion[-insertionEdge]],
          edgeLengths[insertionEdge] - lengthBelow,
          if (is.null(edgeLength)) lengthBelow else edgeLength,
          lengthBelow, edgeLengths[-beforeInsertion])
      } else {
        c(edgeLengths[beforeInsertion], edgeLength,
          edgeLengths[-beforeInsertion])
      }
    }
    
  }
  )
  tree[["tip.label"]] <- c(tree[["tip.label"]], label)
  
  if (addingNode) {
    nNode <- nNode + 1L
    tree[["Nnode"]] <- nNode
  }
  
  ## renumber nodes:
  if (addingNode) {
    newNumbering <- integer(nNode)
    newNumbering[-rootNode] <- newTipNumber + 1L
    childNodes <- treeEdge[, 2] < 0L
    ## executed from right to left, so newNb is modified before x$edge:
    treeEdge[childNodes, 2] <-
      newNumbering[-treeEdge[childNodes, 2]] <-
      newTipNumber + 2:nNode
    treeEdge[, 1] <- newNumbering[-treeEdge[, 1]]
  } else {
    newNumbering <- newTipNumber + 1:nNode
    treeEdge[treeEdge < 0] <- newNumbering[-treeEdge[treeEdge < 0]]
  }
  
  
  
  tree[["edge"]] <- treeEdge
  if (hasLengths) {
    tree[["edge.length"]] <- edgeLengths
  }
  
  if (addingNode) {
    nodeLabels <- tree[["node.label"]]
    if (!is.null(nodeLabels)) {
      newLabels <- character(nNode)
      newLabels[newNumbering - newTipNumber] <- c(nodeLabels, "")
      tree[["node.label"]] <- newLabels
    }
  }
  
  # Return:
  tree
}

#' @rdname AddTip
#'
#' @details `AddTipEverywhere()` adds a tip to each edge in turn.
#'
#' @param includeRoot Logical; if `TRUE`, each position adjacent
#' to the root edge is considered to represent distinct edges; if `FALSE`,
#' they are treated as a single edge.
#' @return `AddTipEverywhere()` returns a list of class `multiPhylo` containing
#' the trees produced by adding `label` to each edge of `tree` in turn.
#'
#' @examples
#' # Set up multi-panel plot
#' oldPar <- par(mfrow = c(2, 4), mar = rep(0.3, 4), cex = 0.9)
#'
#' # Add leaf to each edge on a tree in turn
#' backbone <- BalancedTree(4)
#' # Treating the position of the root as instructive:
#' additions <- AddTipEverywhere(backbone, includeRoot = TRUE)
#' xx <- lapply(additions, plot)
#'
#' par(mfrow = c(2, 3))
#' # Don't treat root edges as distinct:
#' additions <- AddTipEverywhere(backbone, includeRoot = FALSE)
#' xx <- lapply(additions, plot)
#'
#' # Restore original plotting parameters
#' par(oldPar)
#'
#' @importFrom ape is.rooted
#' @export
AddTipEverywhere <- function(tree, label = "New tip", includeRoot = FALSE) {
  nTip <- NTip(tree)
  if (nTip == 0L) return(list(SingleTaxonTree(label)))
  if (nTip == 1L) return(list(StarTree(c(tree[["tip.label"]], label))))
  whichNodes <- seq_len(nTip + tree[["Nnode"]])
  edge <- tree[["edge"]]
  root <- RootNode(edge)
  if (!includeRoot) {
    parent <- edge[, 1]
    child <- edge[, 2]
    rootChildren <- child[parent == root]
    
    whichNodes <- if (length(rootChildren) == 2L && nTip > 2L) {
      rootChildrenNodes <- rootChildren[rootChildren > nTip]
      whichNodes[-c(root, rootChildrenNodes[1])]
    } else {
      whichNodes[-root]
    }
  }
  lapply(whichNodes, AddTip, tree = tree, label = label, nTip = nTip,
         rootNode = root)
}

#' Relative length of internal branches
#' 
#' Treeness (also termed stemminess) is the proportion of a tree's length found
#' on internal branches \insertCite{Lanyon1988}{TreeTools}.
#' Insofar as external branches do not contain phylogenetic (grouping) signal,
#' trees with a high treeness can be interpreted as containing a higher
#' signal:noise ratio \insertCite{Phillips2003}{TreeTools}.
#' 
#' @template tree(s)Param
#' @returns `Treeness()` returns a numeric vector reporting the treeness of each
#' `tree`.
#' 
#' @examples
#' lowTree <- BalancedTree(6, lengths = c(1, 1, 4, 4, 4, 1, 1, 4, 4, 4))
#' plot(lowTree)
#' Treeness(lowTree)
#' highTree <- BalancedTree(6, lengths = c(6, 6, 1, 1, 1, 6, 6, 1, 1, 1))
#' plot(highTree)
#' Treeness(c(lowTree, highTree))
#' @template MRS
#' @family tree properties
#' @references \insertAllCited{}
#' @export
Treeness <- function(tree) {
  UseMethod("Treeness")
}

#' @export
Treeness.phylo <- function(tree) {
  if (is.null(tree[["edge.length"]])) {
    warning("No edge lengths specified")
    return(NULL)
  }
  edge <- tree[["edge"]]
  external <- edge[, 2] <= NTip(tree)
  weights <- tree[["edge.length"]]
  sum(weights[!external]) / sum(weights)
}

#' @export
Treeness.multiPhylo <- function(tree) {
  vapply(tree, Treeness.phylo, double(1))
}

#' @export
Treeness.list <- function(tree) {
  if (all(vapply(tree, inherits, TRUE, "phylo"))) {
    Treeness.multiPhylo(tree)
  } else {
    stop("`tree` must be a list of 'phylo' objects")
  }
}

#' @export
Treeness.NULL <- function(tree) NULL

#' @rdname Treeness
#' @export
Stemminess <- Treeness

test_that("Treeness() fails safely", {
  expect_error(Treeness(list(1, 2)), "list of .phylo. objects")
  expect_warning(expect_null(Treeness(BalancedTree(8))), "edge lengths")
  expect_null(Treeness(NULL))
})

test_that("Treeness() succeeds", {
  internal8 <- c(1, 2, 5, 8, 9, 12)
  in8 <- 1:14 %in% internal8
  tr0 <- BalancedTree(8, !in8)
  tr1 <- BalancedTree(8, in8)
  expect_equal(Treeness(tr1), 1)
  expect_equal(Stemminess(tr0), 0)
  expect_equal(Treeness(BalancedTree(8, ifelse(in8, 1, 11))),
               1 * sum(in8) / sum(ifelse(in8, 1, 11)))
  expect_equal(Stemminess(c(tr0, tr1)), 0:1)
  expect_equal(Stemminess(list(tr0, tr1)), 0:1)
  expect_error(Stemminess(list(tr0, tr1, "tr1")), "phylo")
})

#' Frequency of splits
#'
#' `SplitFrequency()` provides a simple way to count the number of times that
#' bipartition splits, as defined by a reference tree, occur in a forest of
#' trees. May be used to calculate edge ("node") support for majority consensus
#' or bootstrap trees.
#'
#' If multiple calculations are required, some time can be saved by using the
#' constituent functions (see examples).
#'
#' @param reference A tree of class `phylo`, a `Splits` object. If `NULL`,
#' the frequencies of all splits in `forest` will be returned.
#' @param forest A list of trees of class `phylo`, or a `multiPhylo` object;
#' or a `Splits` object. See
#' [vignette](https://ms609.github.io/TreeTools/articles/load-trees.html) for
#' possible methods of loading trees into R.
#' @param hash Logical; if `TRUE` (default), counts the frequencies of _all_
#' splits (i.e. when `reference = NULL`) using a faster 128-bit hashing
#' approach whose results are exact with overwhelming probability (a hash
#' collision conflating two distinct splits is vanishingly unlikely).  Set
#' `hash = FALSE` for a slower but guaranteed-exact count.  Ignored when
#' `reference` is a tree or `Splits` object.
#'
#' @return `SplitFrequency()` returns the number of trees in `forest` that
#' contain each split in `reference`.
#' If `reference` is a tree of class `phylo`, then the sequence will correspond
#' to the order of nodes (use `ape::nodelabels()` to view).
#' Note that the three nodes at the root of the tree correspond to a single
#' split; see the example for how these might be plotted on a tree.
#'
#' If `reference` is `NULL`, then `SplitFrequency()` returns a list of splits 
#' (in the order encountered in forest) with attribute `"count"` stating the
#' number of times each split occurs in `forest`.
#'
#' @template exampleNodeSupport
#'
#' @template MRS
#' @family Splits operations
#' @export
SplitFrequency <- function(reference, forest = NULL, hash = TRUE) {
  if (is.null(reference) || is.null(forest)) {
    if (is.null(forest)) forest <- reference
    if (inherits(forest, "phylo")) forest <- list(forest)
    if (length(unique(lapply(lapply(forest, TipLabels), sort))) > 1) {
      stop("All trees must bear identical labels")
    }
    if (length(forest) == 0) {
      return(structure(forest, count = integer()))
    }
    tipLabels <- TipLabels(forest[[1]])
    if (length(tipLabels) < 4) {
      return(structure(matrix(raw()), nTip = length(tipLabels),
                       tip.label = tipLabels, count = integer(),
                       class = "Splits"))
    }
    forest <- RenumberTips(forest, tipLabels)
    forest <- Preorder(forest)
    result <- split_frequencies(forest, exact = !isTRUE(hash))
    splits <- result[["splits"]]
    counts <- result[["counts"]]
    nTip <- length(tipLabels)
    nbin <- ncol(splits)
    if (nrow(splits) == 0) { # Not been able to hit these lines - included just in case
      return(structure(splits, nTip = nTip, tip.label = tipLabels,              # nocov
                       count = integer(), class = "Splits"))                    # nocov
    }
    # Normalize splits: ensure bit 0 is set, filter trivial splits
    normalized <- normalize_splits(splits, nTip)
    keep <- normalized[["keep"]]
    splits <- normalized[["splits"]][keep, , drop = FALSE]
    counts <- counts[keep]
    ret <- structure(splits,
                     nTip = nTip,
                     tip.label = tipLabels,
                     class = "Splits")
    attr(ret, "count") <- counts
    ret
  } else {
    referenceSplits <- as.Splits(reference)
    refLabels <- attr(referenceSplits, "tip.label")
    forest <- lapply(forest, KeepTip, refLabels)
    forestSplits <- as.Splits(forest, tipLabels = refLabels)
  
    logicals <- vapply(forestSplits,
                       function(cf) referenceSplits %in% cf,
                       logical(length(referenceSplits)))
    ret <- if (is.null(dim(logicals))) {
      sum(logicals)
    } else {
      rowSums(logicals)
    }
    names(ret) <- rownames(referenceSplits)
  
    # Return:
    ret
  }
}

#' Label splits
#'
#' Labels the edges associated with each split on a plotted tree.
#'
#' As the two root edges of a rooted tree denote the same split, only the
#' rightmost (plotted at the bottom, by default) edge will be labelled.
#' If the position of the root is significant, add a tip at the root using
#' [`AddTip()`].
#'
#' @template treeParam
#' @param labels Named vector listing annotations for each split. Names
#' should correspond to the node associated with each split; see
#' [`as.Splits()`] for details.
#' If `NULL`, each splits will be labelled with its associated node.
#' @param unit Character specifying units of `labels`, if desired. Include a
#' leading space if necessary.
#' @param \dots Additional parameters to [`ape::edgelabels()`][ape::nodelabels].
#' @return `LabelSplits()` returns `invisible()`, after plotting `labels` on
#' each relevant edge of a plot (which should already have been produced using
#' `plot(tree)`).
#'
#' @examples
#' tree <- BalancedTree(LETTERS[1:5])
#' splits <- as.Splits(tree)
#' plot(tree)
#' LabelSplits(tree, as.character(splits), frame = "none", pos = 3L)
#' LabelSplits(tree, TipsInSplits(splits), unit = " tips", frame = "none",
#'             pos = 1L)
#'
#' @template exampleNodeSupport
#'
#' @seealso Calculate split support: [`SplitFrequency()`]
#'
#' Colour labels according to value: [`SupportColour()`]
#' @importFrom ape edgelabels
#' @importFrom stats setNames
#' @family Splits operations
#' @export
LabelSplits <- function(tree, labels = NULL, unit = "", ...) {
  splits <- as.Splits(tree)
  if (is.null(labels)) {
    splitNames <- names(splits)
    labels <- setNames(splitNames, splitNames)
  } else if (length(names(labels)) == 0) {
    if (length(splits) == length(labels)) {
      labels <- setNames(labels, names(splits))
    } else {
      stop("`labels` must bear the same names as `as.Splits(tree)`")
    }
  }
  
  if (length(setdiff(names(labels), names(splits)))) {
    warning("Label names do not correspond to splits in tree",
            immediate. = TRUE)
  }
  whichEdge <- match(as.integer(names(labels)), tree[["edge"]][, 2])
  edgelabels(paste0(labels, unit), edge = whichEdge, ...)
  # Return:
  invisible()
}

#' Colour for node support value
#'
#' Colour value with which to display node support.
#'
#' @param support A numeric vector of values in the range 0--1.
#' @param show1 Logical specifying whether to display values of 1.
#' A transparent white will be returned if `FALSE`.
#' @param scale 101-element vector listing colours in sequence. Defaults to
#' a diverging \acronym{HCL} scale.
#' @param outOfRange Colour to use if results are outside the range 0--1.
#' @return `SupportColour()` returns the appropriate value from `scale`,
#' or `outOfRange` if a value is outwith the valid range.
#' @examples
#' SupportColour((-1):4 / 4, show1 = FALSE)
#'
#' @template exampleNodeSupport
#'
#' @seealso Use in conjunction with [`LabelSplits()`] to colour split labels,
#' possibly calculated using [`SplitFrequency()`].
#'
#' @importFrom grDevices colorRampPalette
#' @export
SupportColour <- function(support,
                          show1 = TRUE,
                          # Equivalent to hcl.colors(101, "Blue-Red 2", rev = TRUE),
                          # but works safely on macOS
                          scale = colorRampPalette(c("#D33F6A", "#e2e2e2", "#4A6FE3"))(101),
                          outOfRange = "red") {
  sanitized <- support
  sanitized[!is.numeric(support) | support < 0 | support > 1] <- NA
  inRange <- !is.na(support) & support >= 0 & support <= 1 & support != ""
  ret <- character(length(support))
  ret[!inRange] <- outOfRange
  ret[inRange] <- scale[(sanitized[inRange] * 100) + 1L]
  if (!isTRUE(show1)) {
    ret[support == 1] <- "#ffffff00"
  }
  ret
}

#' @rdname SupportColour
#' @export
SupportColor <- SupportColour

#' @rdname PhyToString
#'
#' @param string String of tokens, optionally containing whitespace, with no
#'   terminating semi-colon.
#' @param tips (Optional) Character vector corresponding to the names (in order)
#' of each taxon in the matrix, or an object such as a tree from which
#' tip labels can be extracted.
#' @param byTaxon Logical; if `TRUE`, string is one **taxon's** coding at a
#' time; if `FALSE`, string is interpreted as one **character's** coding at a
#' time.
#'
#' @return `StringToPhyDat()` returns an object of class `phyDat`.
#'
#' @examples
#' StringToPhyDat("-?01231230?-", c("Lion", "Gazelle"), byTaxon = TRUE)
#' # encodes the following matrix:
#' # Lion     -?0123
#' # Gazelle  1230?-
#'
#' @export
StringToPhyDat <- function(string, tips, byTaxon = TRUE) {
  tokens <- NexusTokens(string)
  if (missing(tips)) {
    tips <- length(tokens)
  }
  tips <- TipLabels(tips)
  tokens <- matrix(tokens, nrow = length(tips), byrow = byTaxon,
                   dimnames = list(tips, NULL))
  
  # Return:
  MatrixToPhyDat(tokens)
}
#' @rdname PhyToString
StringToPhydat <- StringToPhyDat

#' Convert between strings and `phyDat` objects
#'
#' `PhyDatToString()` converts a `phyDat` object as a string;
#' `StringToPhyDat()` converts a string of character data to a `phyDat` object.
#'
#' @param phy An object of class `phyDat`.
#' @param parentheses Character specifying format of parentheses with which to
#' surround ambiguous tokens.  Choose from: \code{\{} (default), `[`, `(`, `<`.
#' @param collapse Character specifying text, perhaps `,`, with which to
#' separate multiple tokens within parentheses.
#' @param ps Character specifying text, perhaps `;`, to append to the end of
#' each string.
#' @param useIndex Logical (default: `TRUE`) specifying whether to print
#' duplicate characters multiple times, as they appeared in the original matrix.
#' @param byTaxon Logical. If `TRUE`, write one taxon followed by the next.
#' If `FALSE`, write one character followed by the next.
#' @param concatenate Logical specifying whether to concatenate all
#' characters/taxa into a single string, or to return a separate string for
#' each entry.
#'
#' @examples
#' fileName <- paste0(system.file(package = "TreeTools"),
#'                    "/extdata/input/dataset.nex")
#' phyDat <- ReadAsPhyDat(fileName)
#' PhyToString(phyDat, concatenate = FALSE)
#'
#' @return `PhyToString()` returns a character vector listing a text
#' representation of the phylogenetic character state for each taxon in turn.
#'
#' @family phylogenetic matrix conversion functions
#' @template MRS
#' @export
PhyToString <- function(phy, parentheses = "{", collapse = "", ps = "",
                        useIndex = TRUE, byTaxon = TRUE, concatenate = TRUE) {
  at <- attributes(phy)
  phyLevels <- at[["allLevels"]]
  if (sum(phyLevels == "-") > 1) {
    stop("More than one inapplicable level identified.  Is phy$levels malformed?")
  }
  phyChars <- at[["nr"]]
  phyContrast <- at[["contrast"]] == 1
  phyIndex <- if (useIndex) {
    at[["index"]]
  } else {
    seq_len(phyChars)
  }
  outLevels <- at[["levels"]]
  
  levelLengths <- vapply(outLevels, nchar, integer(1))
  longLevels <- levelLengths > 1
  if (any(longLevels)) {
    if ("10" %fin% outLevels && !(0 %fin% outLevels)) {
      outLevels[outLevels == "10"] <- "0"
      longLevels["10"] <- FALSE
    }
    outLevels[longLevels] <- LETTERS[seq_len(sum(longLevels))]
  }
  
  switch(parentheses,
         "(" = {openBracket <- "("; closeBracket = ")"},
         ")" = {openBracket <- "("; closeBracket = ")"},
         "<" = {openBracket <- "<"; closeBracket = ">"},
         ">" = {openBracket <- "<"; closeBracket = ">"},
         "[" = {openBracket <- "["; closeBracket = "]"},
         "]" = {openBracket <- "["; closeBracket = "]"},
         {openBracket <- "{"; closeBracket = "}"})
  
  levelTranslation <- apply(phyContrast, 1, function(x)
    ifelse(sum(x) == 1, as.character(outLevels[x]),
           paste0(c(openBracket, paste0(outLevels[x], collapse = collapse),
                    closeBracket), collapse = ""))
  )
  if (any(ambigToken <- apply(phyContrast, 1, all))) {
    levelTranslation[ambigToken] <- "?"
  }
  ret <- vapply(phy, USE.NAMES = FALSE,
                function(x) levelTranslation[x[phyIndex]],
                character(length(phyIndex)))
  
  ret <- if (isTRUE(concatenate) || is.null(dim(ret))) 
    { # If only one row, don't need to apply
    if (!isTRUE(byTaxon)) {
      ret <- t(ret)
    }
    if (isTRUE(concatenate)) {
      paste0(c(ret, ps), collapse = "")
    } else {
      if (isTRUE(byTaxon)) {
        paste0(ret, ps)
      } else {
        paste0(ret, ps, collapse = "")
      }
    }
  } else {
    if (isTRUE(byTaxon)) ret <- t(ret)
    paste0(apply(ret, 1, paste0, collapse = ""), ps)
  }
  # Return:
  ret
}

#' @rdname PhyToString
#' @export
PhyDatToString <- PhyToString
#' @rdname PhyToString
#' @export
PhydatToString <- PhyToString

test_that("Decompose()", {
  data("Lobo")
  
  expect_error(Decompose("Lobo.phy"), "phyDat")

  expect_error(Decompose(Lobo.phy, logical(11)),
               "length.indices.* number of characters in .dataset.")
  
  expect_warning(decomposed <- Decompose(Lobo.phy, c(1234:1235, 11)), 
                 "115 characters; .*1234, 1235 not found")

  NumberOfChars <- function(x) sum(attr(x, "weight"))
  expect_equal(NumberOfChars(decomposed),
               NumberOfChars(Lobo.phy) + 1)
  
  expect_equal(attr(decomposed, "originalIndex"),
               c(1:11, 11:115))
               
  expect_equal(attr(Decompose(Lobo.phy, 10:12), "originalIndex"),
               c(1:11, 11:115))
  
  decompMat <- as.matrix(decomposed)
  taxa <- c(8, 11, 12, 37, 36)
  if (interactive()) {
    dput(as.character(as.matrix(Lobo.phy)[taxa, 11]))
    # = c("?", "{01}", "0", "1", "2")
  }
  expect_equal(decompMat[taxa, 11:12],
               matrix(c("?", "{01}", "0", "1", "1",
                        "?", "0", "0", "0", "1"), ncol = 2,
                      dimnames = list(names(Lobo.phy)[taxa])))
  
  unchanged <- Decompose(Lobo.phy, logical(115))
  
  expect_equal(gsub("\\{\\-(\\d)\\}", "{\\1-}", perl = TRUE,
                    as.matrix(unchanged)), # Order unimportant
               as.matrix(Lobo.phy))
  
  expect_equal(Decompose(Lobo.phy[, 1], TRUE),
               structure(Lobo.phy[, 1], originalIndex = 1))
  
})

test_that("Decompose() with ambiguities", {
  gme <- structure(list(
    tax1 = c(1L, 4L, 7L), tax2 = c(2L, 5L, 4L),
    tax3 = c(3L, 6L, 8L), tax4 = c(2L, 6L, 9L)),
    weight = c(1L, 1L, 1L), nr = 3L, nc = 10L, index = 1:3,
    levels = c("-", "0", "1", "2", "3", "4", "5", "6", "7", "8"),
    allLevels = c("{0123}", "{24}", "2", "0", "{5678}", "5", "-", "{01}", "7"),
    type = "USER",
    contrast = structure(
      c(0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0,
        0, 1, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0,
        0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0), dim = 9:10, 
      dimnames = list(NULL, c("-", 0:8))),
    class = "phyDat")
  
  # "{0123}{24}2{24}"
  exp1 <- c("???0", "11??", "1100", "11??")
  expect_equal(PhyToString(Decompose(gme[, 1], 1), concatenate = FALSE), exp1)
  
  # "0{5678}55"
  exp2 <- c("00000000", "11111{01}{01}{01}", "11111000", "11111000")
  # Need to include char 1 so levels aren't lost from contrast when subsetting
  expect_equal(PhyToString(Decompose(gme[, 1:2], 2), concatenate = FALSE),
               paste0(PhyToString(gme[, 1], concat = FALSE), exp2))
  
  # "-0{01}7"
  exp3 <- c("-------", "0000000", "{01}000000", "1111111")
  expect_equal(PhyToString(Decompose(gme, 3), concatenate = FALSE),
               paste0(PhyToString(gme[, 1:2], concatenate = FALSE), exp3))
  expect_equal(PhyToString(Decompose(gme, 1:3), concatenate = FALSE),
               gsub("?", "{01}", fixed = TRUE, paste0(exp1, exp2, exp3)))
})

test_that("PhyToString() works", {
  phy <- StringToPhyDat("012[01]", letters[1:4])
  expect_equal(PhyToString(phy), "012{01}")
  expect_equal(PhyToString(phy, parentheses = "<"), "012<01>")
  expect_equal(PhyToString(phy, parentheses = ">"), "012<01>")
  expect_equal(PhyToString(phy, parentheses = "("), "012(01)")
  expect_equal(PhyToString(phy, parentheses = ")"), "012(01)")
  expect_equal(PhyToString(phy, parentheses = "]"), "012[01]")
  expect_equal(PhyToString(phy, parentheses = "["), "012[01]")
  expect_equal(PhyToString(phy, parentheses = "}"), "012{01}")
  expect_equal(PhyToString(phy, parentheses = "{"), "012{01}")
  expect_equal(PhyToString(phy, parentheses = "!"), "012{01}")
  
  expect_equal(PhyToString(phy, concatenate = FALSE), c(0, 1, 2, "{01}"))
  expect_equal(PhyToString(phy, concatenate = FALSE, ps = ";"),
               paste0(c(0, 1, 2, "{01}"), ";"))
  
  str <- "012{01}0123"
  phy <- StringToPhyDat(str, letters[1:4])
  expect_equal(str, PhyToString(StringToPhyDat(str, letters[1:4])))
  expect_equal(str,
               PhyToString(StringToPhyDat(str, letters[1:4], byTaxon = TRUE),
                           byTaxon = TRUE))
  
  singleChar <- structure(list(a = 1L, b = 1L, c = 2L, d = 2L),
                          weight = 2L, nr = 1L, nc = 2L,
                          index = c(1L, 1L), levels = c("0", "1"),
                          allLevels = c("0", "1", "?"), type = "USER",
                          contrast = structure(c(1, 0, 1, 0, 1, 1), dim = 3:2,
                                               dimnames = list(NULL, c("0", "1"))),
                          class = "phyDat")
  expect_equal(PhyToString(singleChar, concatenate = FALSE, byTaxon = FALSE,
                           useIndex = FALSE), "0011")
})

test_that("PhyToString() supports long levels", {
  ABC <- LETTERS[1:3]
  longLevels <- structure(
    list(
      x = c(8L, 15L, 1L, 2L, 3L, 4L, 5L, 6L, 7L, 9L, 10L, 11L, 12L, 13L, 14L),
      y = c(14L, 13L, 12L, 11L, 10L, 9L, 7L, 6L, 5L, 4L, 3L, 2L, 1L, 8L, 15L)),
    weight = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L),
    nr = 15L,
    nc = 14L,
    index = 1:15,
    levels = c(0:6, "-", 7:9, ABC),
    allLevels = c(0:6, "-", 7:9, ABC, "?"),
    type = "USER",
    contrast = structure(c(
      1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
      0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0,
      0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0,
      0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1,
      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0,
      0, 0, 0, 0, 0, 0, 0, 0, 1, 1), dim = 15:14),
    class = "phyDat")
  
  expect_equal(PhyToString(longLevels), "-?0123456789ABCCBA9876543210-?")
  
  expect_equal(PhyToString(longLevels, ps = ";"),
               "-?0123456789ABCCBA9876543210-?;")
  expect_equal(PhyToString(longLevels[, 1], ps = ";", byTaxon = TRUE), "-C;")
  expect_equal(PhyToString(longLevels[1, ], ps = ";", byTaxon = TRUE),
               "-?0123456789ABC;")
  expect_equal(PhyToString(longLevels, ps = ";", useIndex = FALSE,
                           byTaxon = TRUE, concatenate = TRUE),
               "-?0123456789ABCCBA9876543210-?;")
  
  # Two -s → error
  attr(longLevels, "allLevels")[1] <- "-"
  expect_error(PhyToString(longLevels))
})

#' Artificial Extinction
#'
#' Remove tokens that do not occur in a fossil "template" taxon from a living
#' taxon, to simulate the process of fossilization in removing data from
#' a phylogenetic dataset.
#'
#' Further details are provided in \insertCite{Asher2020;textual}{TreeTools}.
#'
#' Note: this simple implementation does not account for character contingency,
#' e.g. characters whose absence imposes inapplicable or absent tokens on
#' dependent characters.
#'
#' @param dataset Phylogenetic dataset of class `phyDat` or `matrix`.
#' @param subject Vector identifying subject taxa, by name or index.
#' @param template Character or integer identifying taxon to use as a template.
#' @param replaceAmbiguous,replaceCoded Character specifying whether tokens
#' that are ambiguous (`?`) or coded (not `?`) in the fossil template should
#' be replaced with:
#'  - `original`: Their original value; i.e. no change;
#'  - `ambiguous`: The ambiguous token, `?`;
#'  - `binary`: The tokens `0` or `1`, with equal probability;
#'  - `uniform`: One of the tokens present in `sampleFrom`, with equal
#'  probability;
#'  - `sample`: One of the tokens present in `sampleFrom`, sampled according
#'  to their frequency.
#' @param replaceAll Logical: if `TRUE`, replace all tokens in a subject; if
#' `FALSE`, leave any ambiguous tokens (`?`) ambiguous.
#' @param sampleFrom Vector identifying a subset of characters from which to
#' sample replacement tokens.
#' If `NULL`, replacement tokens will be sampled from the initial states of
#' all taxa not used as a template (including the subjects).
#' @return A dataset with the same class as `dataset` in which entries that
#' are ambiguous in `template` are made ambiguous in `subject`.
#'
#' @references
#' \insertAllCited{}
#' @examples
#' set.seed(1)
#' dataset <- matrix(c(sample(0:2, 4 * 8, TRUE),
#'                     "0", "0", rep("?", 6)), nrow = 5,
#'                     dimnames = list(c(LETTERS[1:4], "FOSSIL"),
#'                                     paste("char", 1:8)), byrow = TRUE)
#' artex <- ArtificialExtinction(dataset, c("A", "C"), "FOSSIL")
#' @template MRS
#' @export
ArtificialExtinction <- function(dataset, subject, template,
                                  replaceAmbiguous = "ambig",
                                  replaceCoded = "original",
                                  replaceAll = TRUE,
                                  sampleFrom = NULL) {
  UseMethod("ArtificialExtinction")
}

#' @rdname ArtificialExtinction
#' @importFrom fastmatch %fin%
#' @export
ArtificialExtinction.matrix <- function(dataset, subject, template,
                                         replaceAmbiguous = "ambig",
                                         replaceCoded = "original",
                                         replaceAll = TRUE,
                                         sampleFrom = NULL) {
  replacers <- c("original", "ambiguous", "binary", "uniform", "frequency")
  replaceA <- pmatch(replaceAmbiguous, replacers)
  if (is.na(replaceA)) stop("`replaceAmbiguous` ambiguously matched.")
  replaceC <- pmatch(replaceCoded, replacers)
  if (is.na(replaceC)) stop("`replaceCoded` ambiguously matched.")

  removes <- dataset[template, ] == "?"
  if (is.null(sampleFrom)) {
    sampleFrom <-
    if (is.numeric(template)) {
      -template
    } else {
      rownames(dataset)[!rownames(dataset) %fin% template]
    }
  }

  .DoReplace <- function(dataset, subject, columns, replace) {
    nCols <- sum(columns)
    replaceWith <- switch(replace,
      dataset[subject, columns], # Original
      "?", # Ambiguous
      sample(c("0", "1"), nCols * length(subject), replace = TRUE), # Binary
      apply(unique(dataset[sampleFrom, columns, drop = FALSE]), 2, sample,
            length(subject), replace = TRUE), # Uniform
      apply(dataset[sampleFrom, columns, drop = FALSE], 2, sample,
            length(subject), replace = TRUE) # Frequency
    )

    # Until require R >= 3.5.0
    isFALSE <- function(x) is.logical(x) && length(x) == 1L && !is.na(x) && !x
    if (isFALSE(replaceAll)) {
      replaceWith[dataset[subject, columns] == "?"] <- "?"
    }
    dataset[subject, columns] <- replaceWith
    dataset
  }

  dataset <- .DoReplace(dataset, subject, removes, replaceA)
  dataset <- .DoReplace(dataset, subject, !removes, replaceC)
  dataset
}

#' @rdname ArtificialExtinction
#' @export
ArtificialExtinction.phyDat <- function(dataset, subject, template,
                                         replaceAmbiguous = "ambig",
                                         replaceCoded = "original",
                                         replaceAll = TRUE,
                                         sampleFrom = NULL) {
  MatrixToPhyDat(ArtificialExtinction(PhyDatToMatrix(dataset), subject,
                                      template,
                                      replaceAmbiguous,
                                      replaceCoded,
                                      replaceAll,
                                      sampleFrom))
}

#' @rdname ArtificialExtinction
#' @export
ArtEx <- ArtificialExtinction

#' Parse TNT Tree
#'
#' Read a tree from [TNT](https://www.lillo.org.ar/phylogeny/tnt/)'s
#' parenthetical output.
#'
#' `ReadTntTree()` imports trees generated by the parsimony analysis program 
#' [TNT](https://www.lillo.org.ar/phylogeny/tnt/) into R, including node labels
#' written with the `ttags` command.
#' Tree files must have been saved by TNT in parenthetical notation, using the
#' TNT command `tsave *`.
#' Trees are easiest to load into R if taxa have been saved using their names
#' (TNT command `taxname =`).  In this case, the TNT `.tre` file
#' contains tip labels and can be parsed directly.  The downside is that the
#' uncompressed `.tre` files will have a larger file size.
#'
#' `ReadTntTree()` can also read `.tre` files in which taxa have been saved
#' using their numbers (`taxname -`).  Such files contain a hard-coded link to
#' the matrix file that was used to generate the trees, in the first line of the
#' `.tre` file.  This poses problems for portability: if the matrix file is
#' moved, or the `.tre` file is accessed on another computer, the taxon names
#' may be lost.  As such, it is important to check that the matrix file
#' exists in the expected location -- if it does not,
#' either use the `relativePath` argument to point to its new location, or
#' specify `tipLabels` to manually specify the tip labels.
#' 
#' `TntText2Tree()` converts text representation of a tree in TNT to an
#'  object of class `phylo`.
#'
#' @param filepath character string specifying path to TNT `.tre` file,
#' relative to the R working directory (visible with `getwd()`).
#' @param relativePath (discouraged) character string specifying location of the
#' matrix file used to generate the TNT results, relative to the current working
#' directory.  Taxon names will be read from this file if they are not specified
#'  by `tipLabels`.
#' @param keepEnd (optional, default 1) integer specifying how many elements of
#'  the file path to conserve when creating relative path (see examples).
#' @param tipLabels (optional) character vector specifying the names of the
#' taxa, in the sequence that they appear in the TNT file.  If not specified,
#' taxon names will be loaded from the data file linked in the first line of the
#'  `.tre` file specified in `filepath`.
#'
#' @return `ReadTntTree()` returns a tree of class `phylo` in 
#' \link[=TNTOrder]{TNT order},
#' corresponding to the tree in `filepath`, or NULL if no trees are found.
#'
#' @examples
#' # In the examples below, TNT has read a matrix from
#' # "c:/TreeTools/input/dataset.nex"
#' # The results of an analysis were written to
#' # "c:/TreeTools/output/results1.tnt"
#' #
#' # results1.tnt will contain a hard-coded reference to
#' # "c:/TreeTools/input/dataset.nex".
#'
#' # On the original machine (but not elsewhere), it would be possible to read
#' # this hard-coded reference from results.tnt:
#' # ReadTntTree("output/results1.tnt")
#'
#' # These datasets are provided with the "TreeTools" package, which will
#' # probably not be located at c:/TreeTools on your machine:
#'
#' oldWD <- getwd() # Remember the current working directory
#' setwd(system.file(package = "TreeTools"))
#'
#' # If taxon names were saved within the file (using `taxname=` in TNT),
#' # then our job is easy:
#' ReadTntTree("extdata/output/named.tre")
#'
#' # But if taxa were compressed to numbers (using `taxname-`), we need to
#' # look up the original matrix in order to dereference the tip names.
#' #
#' # We need to extract the relevant file path from the end of the
#' # hard-coded path in the original file.
#' #
#' # We are interested in the last two elements of
#' # c:/TreeTools/input/dataset.nex
#' #                2      1
#' #
#' # "." means "relative to the current directory"
#' ReadTntTree("extdata/output/numbered.tre", "./extdata", 2)
#'
#' # If working in a lower subdirectory
#' setwd("./extdata/otherfolder")
#'
#' # then it will be necessary to navigate up the directory path with "..":
#' ReadTntTree("../output/numbered.tre", "..", 2)
#'
#'
#' setwd(oldWD) # Restore original working directory
#'
#' TNTText2Tree("(A (B (C (D E ))));")
#'
#' @template MRS
#' @family tree import functions
#' @importFrom ape read.tree
#' @export
ReadTntTree <- function(filepath, relativePath = NULL, keepEnd = 1L,
                        tipLabels = NULL) {
  commands <- .TntCommands(filepath)
  tread <- grep("^trea?d?\\b", commands, perl = TRUE)
  if (length(tread) < 1) return(NULL)
  trees <- c(TntText2Tree(
    paste(
      gsub("tread\\s+('[^']*')*\\s*", "", commands[tread]),
      collapse = "*")))
  nRead <- unname(vapply(
    commands[tread],
    function(cmd) length(
      c(TntText2Tree(gsub("tread\\s+('[^']*')*\\s*", "", cmd)))
    ), double(1)
  ))
  
  if (!any(grepl("[A-z]", trees[[1]][["tip.label"]]))) {
    if (is.null(tipLabels)) {
      tipLabels <- rownames(ReadTntCharacters(filepath))
      if (is.null(tipLabels)) {
        if (length(tread) > 1) {
          warning("Multiple tree blocks not fully supported; ",
                  "please check tip labels and contact \"TreeTools\" ",
                  "maintainer if necessary")
        }
        treadComment <- strsplit(commands[tread[1]], "'", fixed = TRUE)[[1]]
        filePrefix <- "tree(s) from TNT, for data in "
        taxonFile <- grep(filePrefix, treadComment, fixed = TRUE)
        taxonFile <- gsub(filePrefix, "", 
                          gsub("\\", "/", treadComment[taxonFile], fixed = TRUE),
                          fixed = TRUE)
        if (!is.null(relativePath)) {
          taxonFileParts <- strsplit(taxonFile, "/")[[1]]
          nParts <- length(taxonFileParts)
          if (nParts < keepEnd) {
            stop("Taxon file path (", taxonFile,                                # nocov
                 ") contains fewer than keepEnd (", keepEnd, ") components.")   # nocov
          }
          taxonFile <- paste0(c(relativePath,
                                taxonFileParts[(nParts + 1L - keepEnd):nParts]),
                              collapse = "/")
        }
        
        if (length(taxonFile)) {
          if (!file.exists(taxonFile)) {
            warning("Cannot find linked data file:\n  ", taxonFile)             # nocov
          } else {
            tipLabels <- rownames(ReadTntCharacters(taxonFile, character_num = 1L))
            if (is.null(tipLabels)) {
              # TNT character read failed.  Perhaps taxonFile is in NEXUS format?
              tipLabels <- rownames(ReadCharacters(taxonFile, character_num = 1L))
            }
            if (is.null(tipLabels)) {
              warning("Could not read taxon names from linked TNT file:\n  ",   # nocov
                      taxonFile, "\nIs the file in TNT or Nexus format?",       # nocov
                      " If failing inexplicably, please report:",               # nocov
                      "\n  https://github.com/ms609/TreeTools/issues/new")      # nocov
            }
          }
        } else {
          warning("`filepath` does not link to taxon names; try specifying `tipLabels`")
          tipLabels <- as.character(seq_len(trees[[1]][["edge"]][1] - 1) - 1)
        }
      }
    }
    
    trees <- lapply(trees, function(tree) {
      tree$tip.label <- tipLabels[as.integer(tree$tip.label) + 1L]
      tree
    })
  }
  
  #' @rdname ReadTntTree
  #' @export
  ReadTNTTree <- ReadTntTree
  
  if (length(tipLabels)) {
    trees <- RenumberTips(trees, tipLabels)
  }
  
  trees <- TntOrder(trees)
  
  # Assign Tree Tags
  ttags <- grep("^tta?g?s?\\b", commands, perl = TRUE)
  tagLines <- commands[ttags]
  nTags <- length(ttags)
  
  tagTarget <- "^tta?g?s?\\s*\\*\\s*([\\d!]+)$"
  targetTags <- grep(tagTarget, tagLines, perl = TRUE)
  tagIsTarget <- as.logical(tabulate(targetTags, nTags))
  targetTree <- NA
  
  tagWrite <- "^tta?g?s?\\s*\\+\\s*(\\d+)\\s+(.+)$"
  writeTags <- grep(tagWrite, tagLines, perl = TRUE)
  tagIsWrite <- as.logical(tabulate(writeTags, nTags))
  
  tagClear <- "^tta?g?s?\\s*\\-$"
  clearTags <- grep(tagClear, tagLines, perl = TRUE)
  tagIsClear <- as.logical(tabulate(clearTags, nTags))
  
  for (i in seq_along(ttags)) {
    # A loop is inefficient, but matches TNT's command-driven storage structure
    # and thus may be more robust to input format -- and 
    # will make code maintenance and improvement easier in future.
    tagLine <- tagLines[i]
    if (tagIsTarget[i]) {
      target <- gsub(tagTarget, "\\1", tagLine, perl = TRUE)
      targetTree <- if (target == "!") {
        cumsum(nRead)[which.max(c(tread, Inf) > ttags[i]) - 1L]
      } else {
        as.numeric(target)
      }
    } else if (tagIsWrite[i]) {
      if (is.na(targetTree)) {
        warning("Expected `ttags *N`; applying tags to first tree");
        targetTree <- 1L
      }
      if (!length(trees[[targetTree]][["node.label"]])) {
        trees[[targetTree]][["node.label"]] <-
          character(trees[[targetTree]][["Nnode"]])
      }
      targetTips <- NTip(trees[[targetTree]])
      node <- sum(
        as.numeric(gsub(tagWrite, "\\1", tagLine, perl = TRUE)),
        - targetTips, # Tip labels stored separately
        + 1 # TNT counts from zero
      )
      tagText <- gsub(tagWrite, "\\2", tagLine, perl = TRUE)
      if (node < 1) {
        message("Ignoring tag on leaf ", node + targetTips,
                ": \"", tagText, "\"")
      } else {
        trees[[targetTree]][["node.label"]][node] <- 
          if (trees[[targetTree]][["node.label"]][node] == "") {
            tagText
          } else {
            paste(trees[[targetTree]][["node.label"]][node], tagText)
          }
      }
    } else if (tagIsClear[i]) {
      trees[[targetTree]][["node.label"]] <- NULL
    } else {
      warning("ttags command not yet supported: ", tagLine,
              ". Contact TreeTools maintainer.")
    }
  }
  
  # Return:
  if (length(trees) == 1) {
    trees[[1]]
  } else if (length(trees) == 0) {
    NULL                                                                        # nocov
  } else {
    class(trees) <- "multiPhylo"
    trees
  }
  
}

#' @rdname ReadTntTree
#' @param treeText Character string describing one or more trees,
#' in the parenthetical format output by TNT.
#' @export
TntText2Tree <- function(treeText) {
  treeText <- paste(strsplit(treeText, "\\s*\\*\\s*", perl = TRUE)[[1]], ";")
  treeText <- gsub(";(\\s*;)+", ";", treeText)
  treeText <- gsub("\\s*([\\w'\\.\\-]+)\\s*", "\\1,", treeText, perl = TRUE)
  treeText <- gsub("\\)\\s*\\(", "),(", treeText, perl = TRUE)
  treeText <- gsub(",)", ")", treeText, fixed = TRUE)
  
  tr <- read.tree(text = treeText)
  if (inherits(tr, "multiPhylo")) {
    tr[] <- lapply(tr, function (tree) {
      tree$tip.label[] <- Unquote(tree[["tip.label"]])
      tree
    })
  } else if (inherits(tr, "phylo")) {
    tr[["tip.label"]][] <- Unquote(tr[["tip.label"]])
  }
  
  # Return:
  tr
}

#' @rdname ReadTntTree
#' @export
TNTText2Tree <- TntText2Tree

.TntCommands <- function(filepath) {
  fileText <- readLines(filepath)
  commands <- strsplit(
    paste(fileText, collapse = "\UA"), #\UA = \n
    "\\s*;\\s*")[[1]]
}

#' Number of trees containing a tree
#'
#' `TreesMatchingTree()` calculates the number of unrooted binary trees that
#' are consistent with a tree topology on the same leaves.
#'
#' Remember to unroot a tree first if the position of its root is arbitrary.
#'
#' @template treeParam
#'
#' @return `TreesMatchingTree()` returns a numeric specifying the number of
#' unrooted binary trees that contain all the edges present in the input tree.
#'
#' `LnTreesMatchingTree()` gives the natural logarithm of this number.
#'
#'
#' @examples
#' partiallyResolvedTree <- CollapseNode(BalancedTree(8), 12:15)
#' TreesMatchingTree(partiallyResolvedTree)
#' LnTreesMatchingTree(partiallyResolvedTree)
#'
#' # Number of rooted trees:
#' rootedTree <- AddTip(partiallyResolvedTree, where = 0)
#' TreesMatchingTree(partiallyResolvedTree)
#' @template MRS
#'
#' @family tree information functions
#' @export
TreesMatchingTree <- function(tree) {
  prod(NUnrooted(NodeOrder(tree)))
}

#' @rdname TreesMatchingTree
#' @export
LnTreesMatchingTree <- function(tree) {
  sum(LnUnrooted(NodeOrder(tree)))
}

#' @rdname TreesMatchingTree
#' @export
Log2TreesMatchingTree <- function(tree) {
  sum(Log2Unrooted(NodeOrder(tree)))
}

#' Cladistic information content of a tree
#'
#' `CladisticInfo()` calculates the cladistic (phylogenetic) information
#' content of a phylogenetic object, _sensu_ Thorley _et al._ (1998).
#'
#' The \acronym{CIC} is the logarithm of the number of binary trees that
#' include the specified topology.
#' A base two logarithm gives an information content in bits.
#'
#' The \acronym{CIC} was originally proposed by 
#' \insertCite{Rohlf1982;textual}{TreeTools}, and formalised,
#' with an information-theoretic justification, by 
#' \insertCite{Thorley1998;textual}{TreeTools}.
#' \insertCite{Steel2006;textual}{TreeTools} term the equivalent quantity
#' "phylogenetic information content" in the context of individual characters.
#'
#' The number of binary trees consistent with a cladogram provides a more
#' satisfactory measure of the resolution of a tree than simply
#' counting the number of edges resolved \insertCite{Page1992}{TreeTools}.
#'
#' @param x Tree of class `phylo`, or a list thereof.
#'
#' @return `CladisticInfo()` returns a numeric giving the cladistic information
#' content of the input tree(s), in bits.
#' If passed a `Splits` object, it returns the information content of each
#' split in turn.
#'
#' @references
#' \insertAllCited{}
#'
#' @family tree information functions
#' @family tree characterization functions
#'
#' @template MRS
#' @export
CladisticInfo <- function(x) UseMethod("CladisticInfo")

#' @rdname CladisticInfo
#' @export
CladisticInfo.phylo <- function(x) {
  Log2Unrooted(NTip(x)) - Log2TreesMatchingTree(x)
}

#' @rdname CladisticInfo
#' @export
CladisticInfo.Splits <- function(x) {
  Log2Unrooted(NTip(x)) -
    apply(cbind(TipsInSplits(x), TipsInSplits(!x)), 1, Log2TreesMatchingSplit)
}

#' @rdname CladisticInfo
#' @export
CladisticInfo.list <- function(x) vapply(x, CladisticInfo, 0)
#' @rdname CladisticInfo
#' @export
CladisticInfo.multiPhylo <- CladisticInfo.list

#' @rdname CladisticInfo
#' @export
CladisticInformation <- CladisticInfo

test_that("Cladistic info calculated correctly", {
  bal8 <- BalancedTree(8)
  expect_equal(1, TreesMatchingTree(bal8))
  expect_equal(0, LnTreesMatchingTree(bal8))
  expect_equal(Log2Unrooted(8), CladisticInfo(bal8))
  expect_equal(0L, CladisticInfo(StarTree(8)))
  expect_equal(rep(CladisticInfo(PectinateTree(6)), 4),
               CladisticInfo(as.phylo(1:4, 6)))

  tr1 <- CollapseNode(bal8, 12:15)
  expect_equal(NUnrooted(4) * NUnrooted(5),
               TreesMatchingTree(tr1))
  expect_equal(Log2Unrooted(8) - log2(45), CladisticInfo(tr1))

  sizes <- 2:7
  expect_equal(setNames(-log2(apply(cbind(2:7, 7:2), 1, TreesMatchingSplit) /
                                NUnrooted(9)), 12:17),
               CladisticInfo(as.Splits(PectinateTree(9))))
})

#' @importFrom ape read.nexus
.ReadMrBayesTrees <- function(treeFile, burninFrac) {
  trees <- read.nexus(treeFile)
  trees[-seq_len(burninFrac * length(trees))]
}

#' Read posterior tree sample produced by MrBayes
#'
#' Read posterior trees from ['MrBayes'](
#' https://nbisweden.github.io/MrBayes/) output files, discarding burn-in
#' generations.
#'
#' `ReadMrBayesTrees()` samples trees from the posterior distributions
#' computed using the Bayesian inference software ['MrBayes'](
#' https://nbisweden.github.io/MrBayes/) 
#' 
#' @param filepath character string specifying path to `.nex` input file used
#' to initialize the MrBayes analysis, 
#' relative to the R working directory (visible with `getwd()`).
#' @param n Integer specifying number of trees to sample from posterior.
#' @param burninFrac Fraction of trees to discard from each run as burn-in.
#' If `NULL` (the default), this will be read from the last `mcmc` or `mcmcp`
#' command in `filepath`.
#'
#' @return `ReadMrBayesTrees()` returns a 'multiPhylo' object containing 
#' `n` trees sampled evenly from all runs generated by analysis of `filepath`,
#' or `NULL` if no trees are found.
#'
#' @examples
#' \dontrun{ # Download will take a few seconds
#'   url <- 
#'   "https://raw.githubusercontent.com/ms609/hyoliths/master/MrBayes/hyo.nex"
#'   trees <- ReadMrBayesTrees(url, n = 40)
#'   plot(Consensus(trees, p = 0.5))
#' }
#' @family tree import functions
#' @template MRS
#' @export
ReadMrBayesTrees <- function(filepath, n = NULL, burninFrac = NULL) {
  if (is.null(burninFrac)) {
    lines <- readLines(filepath, warn = FALSE)
    burninPattern <- ".*\\bmcmcp?\\b.*burninfr?a?c?\\s*=\\s*([\\d\\.]+)\\b.*"
    burninFrac <- rev(as.numeric(
      gsub(burninPattern, "\\1", lines[grep(burninPattern, lines, TRUE, TRUE)],
           perl = TRUE, ignore.case = TRUE)
    ))[1]
  }
  runFmt <- paste0(filepath, ".run%s.t")
  run1 <- sprintf(runFmt, 1)
  url_exists <- function(url) {
    if (requireNamespace("RCurl", quietly = TRUE)) {
      RCurl::url.exists(url)
    } else {
      tryCatch({
        con <- url(url, open = "r")
        close(con)
        TRUE
      }, error = function(e) FALSE)
    }
  }
  treeFiles <- if (file.exists(run1)) {
    list.files(dirname(filepath),
               paste0(basename(filepath), "\\.run\\d+\\.t"),
               full.names = TRUE)
  } else if (url_exists(run1)) {
    i <- 2
    repeat{
      if (!url_exists(sprintf(runFmt, i))) break
      i <- i + 1
    }
    sprintf(runFmt, seq_len(i - 1))
  } else {
    stop("Cannot find ", run1)
  }
  trees <- tryCatch(do.call(c, lapply(treeFiles, .ReadMrBayesTrees,
                                      burninFrac = burninFrac)),
                    error = function(e) {
                      warning(e, "Could not read trees from ", filepath)
                      NULL
                    })
  if (is.null(trees)) {
    return(NULL)
  }
  if (!is.null(n)) {
    trees <- trees[seq(1, length(trees), length.out = n)]
  }
}

#' @rdname ReadMrBayesTrees
#' @export
ReadMrBayes <- ReadMrBayesTrees

#' @rdname ReadMrBayesTrees
#' @export
MrBayesTrees <- ReadMrBayesTrees

#' Sort tree
#'
#' `SortTree()` sorts each node into a consistent order, so that node rotation
#' does not obscure similarities between similar trees.
#'
#' At each node, clades will be listed in `tree[["edge"]]` in decreasing size
#' order.
#'
#' Clades that contain the same number of leaves are sorted in decreasing order
#' of minimum leaf number, so (2, 3) will occur before (1, 4).
#'
#' As trees are plotted from "bottom up", the largest clades will "sink" to the
#' bottom of a plotted tree.
#'
#' @param tree One or more trees of class `phylo`, optionally as a list
#' or a `multiPhylo` object.
#' @param how Character vector specifying sort method:
#' `"Cladesize"` rotates each node such that the larger clade is first,
#' thus appearing lower when plotted;
#' `"TipLabels"` rotates nodes such that labels listed sooner in `order`
#' are listed first, and thus plot lower.
#' @param order Character vector listing tip labels in sequence they should
#' appear on tree. Clades containing a taxon earlier in this list will be listed
#' sooner and thus plot lower on a tree.  Taxa not listed in `order` will be
#' treated as if they were last in the list.
#'
#' @return `SortTree()` returns tree in the format of `tree`, with each node
#' in each tree sorted 
#'
#' @seealso `Preorder()` also rearranges trees into a consistent shape,
#' based on the index of leaves.
#'
#' @examples
#' messyTree <- as.phylo(10, 6)
#' plot(messyTree)
#'
#' sorted <- SortTree(messyTree)
#' plot(sorted)
#' ape::nodelabels()
#' ape::edgelabels()
#' ape::tiplabels(adj = c(2, 1/3))
#' 
#' plot(SortTree(messyTree, how = "tip"))
#' @template MRS
#' @family tree manipulation
#' @seealso [`sort.multiPhylo()`] sorts a list of trees stored as a `multiPhylo`
#' object.
#' @export
SortTree <- function(tree, how = "cladesize", order = TipLabels(tree)) {
  UseMethod("SortTree")
}


#' @export
SortTree.NULL <- function(tree, how, order) NULL

#' @export
#' @rdname SortTree
SortTree.phylo <- function(tree, how = "cladesize", order = TipLabels(tree)) {
  edge <- tree[["edge"]]
  weight <- tree[["edge.length"]]
  parent <- edge[, 1]
  child <- edge[, 2]
  tipLabels <- tree[["tip.label"]]
  nTip <- length(tipLabels)
  
  descendants <- .ListDescendants(tree)
  method <- pmatch(tolower(how), c("cladesize", "tiplabels"))
  if (is.na(method)) {
    warning("`how` must be an unambiguous contraction of ",
              "\"cladesize\" or \"tiplabels\"")
    method <- 1L
  }
    
  
  switch(method,
    { # Clade Size
      nDescendants <- lengths(descendants)
      MinKid <- function(tips) min(tipLabels[tips])
      for (node in nTip + seq_len(tree[["Nnode"]])) {
        childEdges <- parent == node
        kids <- child[childEdges]
        newOrder <- order(nDescendants[kids],
                          vapply(descendants[kids], MinKid, tipLabels[1]),
                          method = "radix", decreasing = TRUE)
        child[childEdges] <- kids[newOrder]
        if (!is.null(weight)) {
          weight[childEdges] <- weight[childEdges][newOrder]
        }
      }
    }, { # Tip labels
      #TODO would it be quicker to do this using RenumberTips() |> Preorder()?
      # If not, perhaps we can accelerate preorder?
      mass <- match(TipLabels(tree), order, nomatch = NTip(tree) + 1L)
      MinKid <- function(tips) min(mass[tips])
      for (node in nTip + seq_len(tree[["Nnode"]])) {
        childEdges <- parent == node
        kids <- child[childEdges]
        newOrder <- order(vapply(descendants[kids], MinKid, integer(1)),
                          method = "radix")
        child[childEdges] <- kids[newOrder]
        if (!is.null(weight)) {
          weight[childEdges] <- weight[childEdges][newOrder]
        }
      }
    }
  )
  
 
  tree[["edge"]][, 2] <- child
  tree[["edge.length"]] <- weight
  attr(tree, "order") <- NULL
  Renumber(tree)
}

.ListDescendants <- function(tree) {
  edge <- tree[["edge"]]
  parent <- edge[, 1]
  child <- edge[, 2]
  # Every node occurs once in `child` except the root
  descendants <- vector("list", length(child) + 1L)
  descendants[seq_len(NTip(tree))] <- seq_len(NTip(tree))
  for (i in postorder_order(edge)) {
    descendants[[parent[i]]] <- c(descendants[[parent[i]]], 
                                  descendants[[child[i]]])
  }
  
  # Return:
  descendants
}

#' @export
#' @rdname SortTree
SortTree.list <- function(tree, how = "cladesize",
                          order = TipLabels(tree[[1]])) {
  lapply(tree, SortTree, how = how, order = order)
}

#' @export
#' @rdname SortTree
SortTree.multiPhylo <- function(tree,
                                how = "cladesize",
                                order = TipLabels(tree[[1]])) {
  tree[] <- SortTree.list(tree)
  tree
}

#' Convert phylogenetic tree to `ClusterTable`
#'
#' `as.ClusterTable()` converts a phylogenetic tree to a `ClusterTable` object,
#' which is an internal representation of its splits suitable for rapid tree
#' distance calculation (per Day, 1985).
#' 
#' Each row of a cluster table relates to a clade on a tree rooted on tip 1.
#' Tips are numbered according to the order in which they are visited in 
#' preorder: i.e., if plotted using `plot(x)`, from the top of the page
#' downwards.  A clade containing the tips 2 .. 5 would be denoted by the
#' entry `2, 5`, in either row 2 or row 5 of the cluster table.
#'
#' @param x Object to convert into `ClusterTable`: perhaps a tree of class
#'  \code{\link[ape:read.tree]{phylo}}.
#' @param tipLabels Character vector specifying sequence in which to order
#' tip labels.
#' @param \dots Unused.
#'
#' @return `as.ClusterTable()` returns an object of class `ClusterTable`,
#' or a list thereof.
#'
#' @references \insertRef{Day1985}{TreeTools}
#' @seealso [S3 methods][ClusterTable-methods] for `ClusterTable` objects.
#' @examples
#' tree1 <- ape::read.tree(text = "(A, (B, (C, (D, E))));");
#' tree2 <- ape::read.tree(text = "(A, (B, (D, (C, E))));");
#' ct1 <- as.ClusterTable(tree1)
#' summary(ct1)
#' as.matrix(ct1)
#' 
#' # Tip label order must match ct1 to allow comparison
#' ct2 <- as.ClusterTable(tree2, tipLabels = LETTERS[1:5])
#' 
#' # It can thus be safer to use
#' ctList <- as.ClusterTable(c(tree1, tree2))
#' ctList[[2]]
#' @template MRS
#' @family utility functions
#' @name ClusterTable
NULL

#' @rdname ClusterTable
#' @export
as.ClusterTable <- function(x, tipLabels = NULL, ...) {
  UseMethod("as.ClusterTable")
}

#' @rdname ClusterTable
#' @export
as.ClusterTable.phylo <- function(x, tipLabels = NULL, ...) {
  startOrder <- attr(x, "order")
  if (length(startOrder) == 0 || startOrder != "preorder") {
    x <- Preorder(x, topologyOnly = TRUE)
  }
  if (is.null(tipLabels)) {
    tipLabels <- x[["tip.label"]]
  } else if (!identical(x[["tip.label"]], tipLabels)) {
    x <- RenumberTips(x, tipLabels)
  }
  structure(ClusterTable_new(x),
            nTip = length(tipLabels),
            tip.label = tipLabels,
            class = "ClusterTable")
}

#' @rdname ClusterTable
#' @export
as.ClusterTable.list <- function(x, tipLabels = NULL, ...) {
  lapply(x, as.ClusterTable, ..., 
         tipLabels = if (is.null(tipLabels)) TipLabels(x[[1]]) else tipLabels)
}

#' @rdname ClusterTable
#' @export
as.ClusterTable.multiPhylo <- as.ClusterTable.list

#' @keywords internal
#' @export
as.ClusterTable.NULL <- function(x, tipLabels = NULL, ...) NULL

#' S3 methods for `ClusterTable` objects
#' 
#' S3 methods for [`ClusterTable`] objects.
#' 
#' @param x,object Object of class `ClusterTable`.
#' @param \dots Additional arguments for consistency with S3 methods.
#'
#' @examples
#' clustab <- as.ClusterTable(TreeTools::BalancedTree(6))
#' as.matrix(clustab)
#' @template MRS
#' @family utility functions
#' @name ClusterTable-methods
#' @export
as.matrix.ClusterTable <- function(x, ...) {
  ClusterTable_matrix(x)
}

#' @rdname ClusterTable-methods
#' @examples
#' print(clustab)
#' @export
print.ClusterTable <- function(x, ...) {
  nTip <- attr(x, "nTip")
  labels <- attr(x, "tip.label")
  cat("ClusterTable on" , nTip, "leaves:", labels[1], "..", labels[nTip])
}

#' @rdname ClusterTable-methods
#' @examples
#' summary(clustab)
#' @export
summary.ClusterTable <- function(object, ...) {
  nTip <- attr(object, "nTip")
  mat <- ClusterTable_matrix(object)
  cat("ClusterTable on" , nTip, "leaves:\n")
  cat(" ", rep(c(1:9, " "), length.out = nTip), "\n", sep = "")
  apply(mat, 1, function(x) {
    if (x[1] > 0) {
      cat(" ", rep(".", x[1] - 1), rep("*", 1 + x[2] - x[1]),
          rep(".", nTip - x[2]), "\n", sep = "")
    }
  })
  
  cat(paste0(" ", seq_len(nTip), ": ", 
             attr(object, "tip.label")[ClusterTable_decode(object)]), "\n")
}

#' Sort a list of phylogenetic trees
#'
#' Trees are sorted by their \link[=TreeNumber]{mixed base representation},
#' treating their leaves in the order of their labels (i.e. alphabetically,
#' if leaves are labelled with text).
#'
#' @param x,decreasing,na.last,\dots As in [`sort()`].
#'
#' @examples
#' sort(as.phylo(5:0, 7))
#' @template MRS
#' @family utility functions
#' @export
sort.multiPhylo <- function(x, decreasing = FALSE, na.last = NA, ...) {
  maxTip <- max(NTip(x))
  sortable <- vapply(x, .Sortable, integer(maxTip - 3), maxTip)
  # Return:
  x[do.call(order, as.data.frame(t(sortable)), ...)]
}

#' @rdname sort.multiPhylo
#' @param e1,e2 Objects to be compared.
#' @export
`==.phylo` <- function(e1, e2) {
  isTRUE(all.equal(.Comparable(e1), .Comparable(e2)))
}

.Comparable <- function(tree) {
  as.MixedBase(RenumberTips(tree, sort(tree[["tip.label"]])))
}

.Sortable <- function(tree, maxTip = NTip(tree)) {
  c(rep(-1L, maxTip - NTip(tree)),
    as.integer(.Comparable(tree)))
}

#' @rdname sort.multiPhylo
#' @export
`<.phylo` <- function(e1, e2) {
  .Comparable(e1) < .Comparable(e2)
}

#' @rdname sort.multiPhylo
#' @export
`>.phylo` <- function(e1, e2) `<.phylo`(e2, e1)


#' @rdname sort.multiPhylo
#' @export
`==.MixedBase` <- function(e1, e2) {
  isTRUE(all.equal(e1, e2))
}

#' @rdname sort.multiPhylo
#' @export
`<.MixedBase` <- function(e1, e2) {
  s1 <- as.integer(e1)
  s2 <- as.integer(e2)
  if (length(s1) == length(s2)) {
    s1Less <- s1 < s2
    if (any(s1Less)) {
      s1More <- s1 > s2
      if (any(s1More)) {
        which.max(s1Less) < which.max(s1More)
      } else {
        TRUE
      }
    } else {
      FALSE
    }
  } else {
    length(s1) < length(s2)
  }
}

#' @rdname sort.multiPhylo
#' @export
`>.MixedBase` <- function(e1, e2) e2 < e1

#' Identify descendant edges
#'
#' `DescendantEdges()` efficiently identifies edges that are "descended" from
#' edges in a tree.
#'
#' @template treeParent
#' @template treeChild
#' @param edge Integer specifying the number of the edge whose children are
#' required (see \code{\link[ape:nodelabels]{edgelabels}()}).
#' @param node Integer specifying the number(s) of nodes whose children are
#' required.  Specify `0` to return all nodes.  If `NULL` (the default), the 
#' `edge` parameter will be used instead.
#' @param nEdge number of edges (calculated from `length(parent)` if not
#' supplied).
#' @param includeSelf Logical specifying whether to mark `edge` as its own
#' descendant.
#' @return `DescendantEdges()` returns a logical vector stating whether each
#' edge in turn is the specified edge (if `includeSelf = TRUE`)
#' or one of its descendants.
#' @examples
#' tree <- as.phylo(0, 6)
#' plot(tree)
#' desc <- DescendantEdges(tree$edge[, 1], tree$edge[, 2], edge = 5)
#' which(desc)
#' ape::edgelabels(bg = 3 + desc)
#' @family tree navigation
#' @export
DescendantEdges <- function(parent, child, edge = NULL, node = NULL,
                            nEdge = length(parent),
                            includeSelf = TRUE) {
  postorder <- PostorderOrder(cbind(parent, child))
  
  if (is.null(node)) {
    if (is.null(edge)) {
      # Dense full matrix path
      nodeDescendants <- descendant_edges(parent, child, postorder)
      entries <- pmax(0, child - min(parent) + 1)
      ret <- matrix(FALSE, nEdge, nEdge)
      ret[entries > 0, ] <- nodeDescendants[entries, ]
      if (includeSelf) {
        diag(ret) <- TRUE
      }
      ret
    } else {
      if (length(edge) == 1L) {
        descendant_edges_single(parent, child, postorder, edge,
                                include_self = includeSelf)
      } else {
        nodeDescendants <- descendant_edges(parent, child, postorder)
        entry <- child[edge] - min(parent) + 1
        ret <- ifelse(entry > 0,
                      nodeDescendants[entry, ],
                      logical(nEdge))
        if (includeSelf) {
          ret[edge] <- TRUE
        }
        ret
      }
    }
  } else {
    nodeDescendants <- descendant_edges(parent, child, postorder)
    if (length(node) == 1 && node == 0) {
      nodeDescendants
    } else {
      nodeDescendants[node - min(parent) + 1, ]
    }
  }
}

#' Identify descendant tips
#'
#' `DescendantTips()` efficiently identifies leaves (external nodes) that are
#' "descended" from edges in a tree.
#'
#' @return `DescendantTips()` returns a logical vector stating whether each
#' leaf in turn is a descendant of the specified edge.
#' @examples
#' tips <- DescendantTips(tree$edge[, 1], tree$edge[, 2], edge = 5)
#' which(tips)
#' tiplabels(bg = 3 + tips)
#' @rdname DescendantEdges
#' @export
DescendantTips <- function(parent, child, edge = NULL,
                           node = NULL,
                           nEdge = length(parent)) {
  tips <- descendant_tips(parent, child, PostorderOrder(cbind(parent, child)))
  nTip <- dim(tips)[[2]]
  if (is.null(node)) {
    if (is.null(edge)) {
      tips[child, ]
    } else {
      tips[child[edge], ]
    }
  } else {
    if (length(node) == 1 && node == 0) {
      tips
    } else {
      tips[node, ]
    }
  }
}

.AllDescendantEdges <- function(parent, child, nEdge = length(parent)) {
  ret <- diag(nEdge) == 1
  blankLogical <- logical(nEdge)
  allEdges <- seq_len(nEdge)
  for (edge in rev(allEdges[child > parent[1]])) {
    ret[edge, ] <- apply(ret[parent == child[edge], ], 2, any)
    ret[edge, edge] <- TRUE
  }
  
  # Return:
  ret
}

#' Count descendants for each node in a tree
#'
#' `NDescendants()` counts the number of nodes (including leaves) directly
#' descended from each node in a tree.
#'
#' @template treeParam
#'
#' @return `NDescendants()` returns an integer listing the number of direct
#' descendants (leaves or internal nodes) for each node in a tree.
#'
#' @examples
#' tree <- CollapseNode(BalancedTree(8), 12:15)
#' NDescendants(tree)
#' plot(tree)
#' nodelabels(NDescendants(tree))
#'
#' @seealso
#' [`CladeSizes()`] for the count of descendant leaves (rather than direct
#' children) at each node.
#' @template MRS
#' @family tree navigation
#' @export
NDescendants <- function(tree) {
  NodeOrder(tree[["edge"]], includeAncestor = FALSE)
}

#' @importFrom utils globalVariables
utils::globalVariables(c("nRootedShapes",
                         "nUnrootedShapes",
                         "unrootedKeys"),
                       "TreeTools")

#' Integer representing shape of a tree
#'
#' Returns an integer that uniquely represents the shape of an _n_-tip
#' binary tree, ignoring tip labels.
#'
#' Rooted trees are numbered working up from the root.
#'
#' The root node divides _n_ tips into two subtrees.  The smaller subtree
#' may contain $a = 1, 2, ..., n/2$ tips, leaving $b = n - a$ tips in
#' These options are worked through in turn.
#'
#' For the first shape of the smaller subtree, work through each possible shape
#' for the larger subtree.  Then, move to the next shape of the smaller subtree,
#' and work through each possible shape of the larger subtree.
#'
#' Stop when the desired topology is encountered.
#'
#' Unrooted trees are numbered less elegantly.  Each cherry (i.e. node
#' subtending a pair of tips) is treated in turn.  The subtended tips are
#' removed, and the node treated as the root of a rooted tree.  The number of
#' this rooted tree is then calculated.  The tree is assigned a _key_
#' corresponding to the lowest such value.  The keys of all unrooted tree shapes
#' on _n_ tips are ranked, and the unrooted tree shape is assigned a _number_
#' based on the rank order of its key among all possible keys, counting from
#' zero.
#'
#' If `UnrootedTreeShape()` or `UnrootedTreeKey()` is passed a rooted tree,
#' the position of the root will be ignored.
#'
#' The number of unlabelled binary rooted trees corresponds to the
#' [Wedderburn-Etherington numbers](https://oeis.org/A001190).
#'
#'
#' @template treeParam
#'
#' @return `TreeShape()` returns an integer specifying the shape of a tree,
#' ignoring tip labels.
#'
#' @examples
#' RootedTreeShape(PectinateTree(8))
#' plot(RootedTreeWithShape(0, nTip = 8L))
#'
#' NRootedShapes(8L)
#' # Shapes are numbered from 0 to NRootedShapes(n) - 1.  The maximum shape is:
#' RootedTreeShape(BalancedTree(8))
#'
#' # Unique shapes of unrooted trees:
#' NUnrootedShapes(8L)
#'
#' # Keys of these trees:
#' UnrootedKeys(8L)
#'
#' # A tree may be represented by multiple keys.
#' # For a one-to-one correspondence, use a number instead:
#' unrootedShapes8 <- as.integer(NUnrootedShapes(8L))
#' allShapes <- lapply(seq_len(unrootedShapes8) - 1L,
#'                     UnrootedTreeWithShape, 8L)
#' plot(allShapes[[1]])
#' sapply(allShapes, UnrootedTreeShape)
#' sapply(allShapes, UnrootedTreeKey, asInteger = TRUE) # Key >= number
#'
#' # If numbers larger than 2>31 are required, sapply needs a little help
#' # with 64-bit integers:
#' structure(sapply(allShapes, UnrootedTreeKey), class = "integer64")
#'
#'
#' @seealso Unique number for a labelled tree: [`TreeNumber()`]
#'
#' @template MRS
#' @name TreeShape
#' @export
RootedTreeShape <- function(tree) {
  edge <- tree[["edge"]]
  nTip <- NTip(tree)
  edge <- Postorder(edge)
  .Int64(edge_to_rooted_shape(edge[, 1], edge[, 2], nTip))
}

#' @importFrom bit64 as.integer64
#' @keywords internal
.Int64 <- function(n) {
  n <- as.integer64(n)
  if (length(n) == 2L) {
    n <- n[1] * 2147483647L + n[2]
  }

  # Return:
  n
}

#' @rdname TreeShape
#' @param shape Integer specifying shape of tree, perhaps generated by
#'  `TreeShape()`.
#' @param nTip Integer specifying number of tips.
#' @return `RootedTreeWithShape()` returns a tree of class `phylo`
#' corresponding to the shape provided.  Tips are unlabelled.
#' @export
RootedTreeWithShape <- function(shape, nTip, tipLabels)
  UseMethod("RootedTreeWithShape")

#' @export
RootedTreeWithShape.numeric <- function(shape, nTip,
                                        tipLabels = character(nTip)) {
  structure(list(edge = rooted_shape_to_edge(as.integer(shape), nTip),
                 Nnode = nTip - 1L,
                 tip.label = tipLabels),
            class = "phylo")
}

#' @export
RootedTreeWithShape.integer64 <- function(shape, nTip,
                                          tipLabels = character(nTip)) {
  if (shape < 0) {
    stop("Shape may not be negative.")
  } else if (shape > 2L^31L - 1L) {
    stop("Shapes this large are not currently implemented. ",
         "Please contact maintainer for help.")
  } else {
    RootedTreeWithShape(as.integer(shape), nTip, tipLabels)
  }
}

#' @rdname TreeShape
#' @template tipLabelsParam
#' @return `UnrootedTreeWithShape()` returns a tree of class `phylo`
#' corresponding to the shape provided.  Tips are unlabelled.
#' @export
UnrootedTreeWithShape <- function(shape, nTip, tipLabels = character(nTip)) {
  if (nTip > 30) {
    stop("Only trees with < 31 leaves are presently handled")
  }
  nShapes <- nUnrootedShapes[nTip]
  if (shape >= as.integer(nShapes)) {
    stop("Shape must be between 0 and ", nShapes - 1L)
  }

  key <- UnrootedKeys(nTip)[shape + 1L]

  UnrootedTreeWithKey(key, nTip, tipLabels)
}

#' @rdname TreeShape
#' @param key Integer specifying the _key_ (not number) of an unrooted tree.
#' @return `UnrootedTreeWithKey()` returns a tree of class `phylo` corresponding
#' to the key provided.  Tips are unlabelled.
#' @export
UnrootedTreeWithKey <- function(key, nTip, tipLabels = character(nTip)) {
  AddRoot <- function(x) {
    x[["root.edge"]] <- 1L
    x
  }
  SingleTaxonTree(tipLabels[1]) + SingleTaxonTree(tipLabels[2]) +
    AddRoot(RootedTreeWithShape(key, nTip - 2L, tipLabels[-c(1, 2)]))
}

#' @rdname TreeShape
#' @export
UnrootedTreeShape <- function(tree) {
  which(UnrootedKeys(NTip(tree)) == UnrootedTreeKey(tree)) - 1L
}

#' @rdname TreeShape
#' @param asInteger Logical specifying whether to coerce the return value to
#' mode `integer`: only possible for values < 2^31.
#' If `FALSE`, values will have class `integer64`.
#' @importFrom bit64 integer64
#' @export
UnrootedTreeKey <- function(tree, asInteger = FALSE) {
  edge <- Postorder(tree[["edge"]])
  nTip <- NTip(tree)
  parent <- edge[, 1]
  child <- edge[, 2]
  nEdge <- length(child)
  unrooted <- nEdge %% 2L
  nodeFirst <- c(rep.int(c(TRUE, FALSE), nEdge / 2L),
                 logical(as.integer(unrooted)))
  nodeSecond <- !nodeFirst
  nodeNumbers <- unique(parent)
  if (unrooted) {
    nodeFirst [nEdge - 0:2] <- FALSE
    nodeSecond[nEdge - 0:2] <- FALSE
    nodeNumbers <- nodeNumbers[-(nTip - 2L)]
  }

  RootedNumber <- function(nodeChildren) {
    RootedTreeShape(Postorder(DropTip(RootTree(tree, nodeChildren[1]), nodeChildren)))
  }

  basalTipEdges <- nEdge - (seq_len(4L - unrooted) - 1L)
  rootCandidate <- if (sum(child[basalTipEdges] <= nTip) == 2) {
    RootedNumber(child[basalTipEdges][child[basalTipEdges] <= nTip])
  } else {
    integer64(0)
  }

  cherryNodes <- nodeNumbers[child[nodeFirst] <= nTip & child[nodeSecond] <= nTip]
  allKeys <- structure(c(vapply(cherryNodes, function(node) {
    RootedNumber(child[parent == node])
  }, integer64(1)), rootCandidate), class = "integer64")

  # Return:
  min(if (asInteger) as.integer(allKeys) else allKeys)
}

#' @rdname TreeShape
#' @keywords internal
#' @importFrom bit64 integer64
#' @export
.UnrootedKeys <- function(nTip) {
  if (nTip > 28L) {
    stop("Too many shapes to calculate with ", nTip, " leaves.")
  } else if (nTip > length(unrootedKeys)) {
    #TODO make efficient - this is horrible!
    # nocov start
    shapes <- as.integer(structure(
      vapply(seq_len(as.integer(NRootedShapes(nTip))) - 1L,
             function(shape) UnrootedTreeKey(RootedTreeWithShape(shape, nTip)),
             integer64(1L)),
      class = "integer64"))
    uniqueShapes <- unique(shapes)
    # nocov end
  } else {
    uniqueShapes <- unrootedKeys[[nTip]]
  }

  # Return:
  sort(uniqueShapes)
}

# Create a cache environment for UnrootedKeys
.unrooted_keys_cache <- new.env(parent = emptyenv())

#' @rdname TreeShape
#' @param \dots Deprecated (2025-09); retained for backward compatibility.
#' @return `UnrootedKeys()` returns a vector of integers corresponding to the
#' keys (not shape numbers) of unrooted tree shapes with `nTip` tips.
#' It is a wrapper to `.UnrootedKeys()`, with memoization, meaning that results
#' once calculated are cached and need not be calculated on future calls to
#' the function.
#' @export
UnrootedKeys <- function(nTip, ...) {
  # Convert nTip to character for use as environment key
  key <- as.character(nTip)
  
  # Check if result is already cached
  if (exists(key, envir = .unrooted_keys_cache, inherits = FALSE)) {
    return(get(key, envir = .unrooted_keys_cache, inherits = FALSE))
  }
  
  # Calculate result and cache it
  result <- .UnrootedKeys(nTip)
  assign(key, result, envir = .unrooted_keys_cache)
  
  return(result)
}

#' @rdname TreeShape
#' @return `NUnrootedShapes()` returns an object of class `integer64` specifying
#' the number of unique unrooted tree shapes with `nTip` (< 61) tips.
#' @export
NUnrootedShapes <- function(nTip) {
  if (nTip > 60L) {
    stop("Too many shapes to represent as a 64-bit integer. ",
         "Consult OEIS for value: https://oeis.org/A000672/b000672.txt")
  }
  nUnrootedShapes[nTip]
}

#' @rdname TreeShape
#' @return `NRootedShapes()` returns an object of class `integer64` specifying
#' the number of unique rooted tree shapes with `nTip` (< 56) leaves.
#'
#' @export
NRootedShapes <- function(nTip) {
  if (nTip > 55L) {
    stop("Too many shapes to represent as a 64-bit integer. ",
         "Consult OEIS for value: https://oeis.org/A001190/b001190.txt")
  }
  nRootedShapes[nTip]
}

test_that("Errors are handled", {
  skip_if(Sys.getenv("USING_ASAN") != "")
  expect_error(RootedTreeWithShape(as.integer64(-1)),
               "Shape may not be negative")
  expect_error(RootedTreeWithShape(as.integer64(0), -1),
               "Tree must have at least zero leaves")
  expect_error(RootedTreeWithShape(as.integer64(2)^31),
               "Shapes this large are not.* implemented")
  expect_error(UnrootedTreeWithShape(31, 31), " < 31 leaves")
  expect_error(.UnrootedKeys(29L), " 29 leaves")

    expect_error(NRootedShapes(56L), "Too many shapes ")
  expect_error(NUnrootedShapes(61L), "Too many shapes ")
})

test_that("Tree shapes counted", {
  expect_equal(as.integer64(c(1, 1, 1, 2, 3, 6, 11, 23)),
               structure(vapply(1:8, NRootedShapes, bit64::integer64(1L)),
                         class = "integer64"))

  expect_equal(as.integer64(lengths(unrootedKeys)),
               structure(vapply(seq_along(unrootedKeys),
                                NUnrootedShapes,
                                integer64(1)),
                         class = "integer64"))
})

test_that("Nasty node order not fatal", {
  nastyBinary <- structure(list(edge = structure(
    c(9, 12, 10, 13, 11, 10, 11, 14, 15, 13, 12, 9, 14, 15,
      5, 10, 15, 14,  3, 13,  9,  4, 11,  7,  8, 6,  2,  1),
    .Dim = c(14, 2)),
    Nnode = 7L,
    tip.label = letters[1:8]),
    class = "phylo") # Danger: Do not plot!
  expect_equal(RootedTreeShape(Preorder(nastyBinary)),
               RootedTreeShape(nastyBinary))

  skip_if(Sys.getenv("USING_ASAN") != "")
  expect_error(edge_to_rooted_shape(1:10, 1:11, 6), "must be the same length")
  expect_error(edge_to_rooted_shape(1:10, 1:10, 5), "is tree binary\\?")
})

test_that("Rooted tree shapes counted", {
  # https://oeis.org/A001190
  expect_equal(as.integer64(c(1, 1, 1, 2, 3, 6, 11, 23, 46, 98, 207, 451, 983,
                              2179, 4850, 10905, 24631, 56011, 127912, 293547,
                              676157, 1563372, 3626149, 8436379, 19680277,
                              46026618, 107890609, 253450711, 596572387,
                              1406818759, "3323236238",
                              "7862958391", "18632325319", "44214569100")),
               structure(vapply(1:34, NRootedShapes, integer64(1)),
                                           class = "integer64"))
})

test_that("Rooted tree shapes fail gracefully", {
  skip_if(Sys.getenv("USING_ASAN") != "")
  expect_error(RootedTreeShape(BalancedTree(56)), "> 55 leaves")
})

test_that("Rooted tree shapes calculated", {
  expect_equal(NRootedShapes(8) - 1L, RootedTreeShape(BalancedTree(0:7)))
  expect_equal(NRootedShapes(54) - 1L, RootedTreeShape(BalancedTree(54)))
  expect_equal(NRootedShapes(55), RootedTreeShape(BalancedTree(55)))

  expect_equal(as.integer64(0L), RootedTreeShape(PectinateTree(0:3)))
  expect_equal(as.integer64(0L), RootedTreeShape(SortTree(PectinateTree(0:3))))
  expect_equal(0L, UnrootedTreeShape(PectinateTree(0:3)))
  expect_equal(as.integer64(1L), RootedTreeShape(BalancedTree(0:3)))

  expect_equal(as.integer64(0L), RootedTreeShape(PectinateTree(0:4)))
  expect_equal(0L, UnrootedTreeShape(PectinateTree(0:4)))
  expect_equal(as.integer64(1L), RootedTreeShape(as.phylo(1, 5)))
  expect_equal(as.integer64(2L), RootedTreeShape(BalancedTree(0:4)))

  expect_equal(as.integer64(0L), RootedTreeShape(PectinateTree(0:5)))
  expect_equal(0L, UnrootedTreeShape(PectinateTree(0:5)))
  expect_equal(as.integer64(1L), RootedTreeShape(as.phylo(58, 6)))
  expect_equal(as.integer64(2L), RootedTreeShape(as.phylo(1, 6)))
  expect_equal(as.integer64(3L), RootedTreeShape(
    ape::read.tree(text = "((a, b), (c, (d, (e, f))));")))
  expect_equal(as.integer64(4L), RootedTreeShape(
    ape::read.tree(text = "((a, b), ((c, d), (e, f)));")))
  expect_equal(as.integer64(5L), RootedTreeShape(BalancedTree(0:5)))

  PectinateTest <- function(i) {
    expect_equal(as.integer64(0L), RootedTreeShape(PectinateTree(i)))
  }
  lapply(4:16, PectinateTest)

  BalancedTest <- function(i) {
    expect_equal(NRootedShapes(i) - 1L,
                 RootedTreeShape(BalancedTree(i)))
  }
  lapply(c(2 ^ (1:4), 10), BalancedTest)

  expect_equal(0L, .UnrootedKeys(4))
  expect_equal(0L, .UnrootedKeys(5))
  expect_equal(0:1, .UnrootedKeys(6))
  expect_equal(0:1, .UnrootedKeys(7))
  expect_equal(c(0:2, 4), .UnrootedKeys(8))
  expect_equal(c(0:2, 4:5, 7), .UnrootedKeys(9))
})

expect_treequal <- function(...) expect_true(all.equal(...))
test_that("Rooted tree shapes built", {
  expect_error(RootedTreeWithShape(-1, 5))

  expect_treequal(RootedTreeWithShape(0, 4), PectinateTree(rep("", 4)))
  expect_treequal(RootedTreeWithShape(1, 4), BalancedTree(rep("", 4)))

  expect_treequal(RootedTreeWithShape(0L, 5L), PectinateTree(rep("", 5)))
  expect_treequal(RootedTreeWithShape(1L, 5L), as.phylo(1, 5, rep("", 5)))
  expect_treequal(RootedTreeWithShape(2L, 5L), BalancedTree(rep("", 5)))

  blank6 <- rep("", 6L)
  expect_treequal(RootedTreeWithShape(0L, 6L), PectinateTree(blank6))
  expect_treequal(RootedTreeWithShape(1L, 6L), as.phylo(58, 6, blank6))
  expect_treequal(RootedTreeWithShape(2L, 6L), as.phylo(1, 6, blank6))
  expect_treequal(RootedTreeWithShape(3L, 6L),
                  ape::read.tree(text = "((,),(,(,(,))));"))
  expect_treequal(RootedTreeWithShape(4L, 6L),
                  ape::read.tree(text = "((,),((,),(,)));"))
  expect_treequal(RootedTreeWithShape(5L, 6L), BalancedTree(blank6))

  expect_treequal(RootedTreeWithShape(0, 8), PectinateTree(rep("", 8)))
  expect_treequal(RootedTreeWithShape(NRootedShapes(8), 8),
                  BalancedTree(rep("", 8)))

  BalancedTest <- function(i) {
    expect_treequal(BalancedTree(rep("", i)),
                    RootedTreeWithShape(NRootedShapes(i) - 1L, i))
  }
  lapply(2 ^ (1:4), BalancedTest)
})

test_that("Unrooted tree shapes fail gracefully", {
  skip_if(Sys.getenv("USING_ASAN") != "")
  expect_error(UnrootedTreeWithShape(4, 8), "must be between 0 and 3")
})

test_that("Unrooted tree shapes built", {
  expect_treequal(UnrootedTreeWithShape(0, 9),
                  UnrootTree(PectinateTree(rep("", 9))))
  TestSym <- function(tree, shape) {
    expect_equal(shape, UnrootedTreeShape(tree))
    expect_equal(UnrootedTreeKey(tree),
                 UnrootedTreeKey(UnrootedTreeWithShape(shape, NTip(tree))))
  }
  blank9 <- rep("", 9)
  TestSym(PectinateTree(blank9), 0)
  TestSym(as.phylo(72292, nTip = 9, blank9), 1)
  TestSym(as.phylo(67987, nTip = 9, blank9), 2)
  TestSym(BalancedTree(blank9), 3)
  TestSym(as.phylo(72237, nTip = 9, blank9), 3)
  TestSym(as.phylo(67882, nTip = 9, blank9), 4)
  TestSym(as.phylo(72298, nTip = 9, blank9), 5)

})

#' Minimum spanning tree
#'
#' Calculate or plot the minimum spanning tree \insertCite{Gower1969}{TreeTools}
#' of a distance matrix.
#'
#' @param distances Either a matrix that can be interpreted as a distance
#' matrix, or an object of class `dist`.
#' @param plot Logical specifying whether to add the minimum spanning tree
#' to an existing plot.
#' @param x,y Numeric vectors specifying the X and Y coordinates of each
#' element in `distances`.  Necessary only if `plot = TRUE`.
#' @param \dots Additional parameters to send to `[lines()]`.
#'
#' @return `MSTEdges()` returns a matrix in which each row corresponds to an
#' edge of the minimum spanning tree, listed in non-decreasing order of length.
#' The two columns contain the indices of the entries in `distances` that
#' each edge connects, with the lower value listed first.
#'
#' @seealso
#' Slow implementation returning the association matrix of the minimum spanning
#' tree: [`ape::mst()`].
#'
#' @references \insertAllCited{}
#'
#' @examples
#' # Corners of an almost-regular octahedron
#' points <- matrix(c(0, 0, 2, 2, 1.1, 1,
#'                    0, 2, 0, 2, 1, 1.1,
#'                    0, 0, 0, 0, 1, -1), 6)
#' distances <- dist(points)
#' mst <- MSTEdges(distances)
#' MSTLength(distances, mst)
#' plot(points[, 1:2], ann = FALSE, asp = 1)
#' MSTEdges(distances, TRUE, x = points[, 1], y = points[, 2], lwd = 2)
#' @template MRS
#' @family utility functions
#' @importFrom graphics lines
#' @export
MSTEdges <- function(distances, plot = FALSE, x = NULL, y = NULL, ...) {
  ends <- MinimumSpanningTree(distances)
  if (plot) {
    apply(ends, 1, function(edge)
      lines(x[edge], y[edge], ...))
    invisible(ends)
  } else {
    ends
  }
}

#' @rdname MSTEdges
#' @param mst Optional parameter specifying the minimum spanning tree in the
#' format returned by `MSTEdges()`; if `NULL`, calculated from `distances`.
#' @return `MSTLength()` returns the length of the minimum spanning tree.
#' @export
MSTLength <- function(distances, mst = NULL) {
  distMat <- as.matrix(distances)
  if (is.null(mst)) mst <- MSTEdges(distances)
  sum(apply(mst, 1L, function(x) distMat[x[1], x[2]]))
}

MinimumSpanningTree <- function(distances) UseMethod("MinimumSpanningTree")

#' @export
MinimumSpanningTree.NULL <- function(distances) NULL

#' @export
MinimumSpanningTree.dist <- function(distances) {
  minimum_spanning_tree(order(distances, decreasing = TRUE) - 1L)
}

#' @export
MinimumSpanningTree.matrix <- function(distances) {
  dims <- dim(distances)
  if (dims[1] != dims[2]) {
    stop("`distances` must be a square matrix.")
  }
  dists <- distances[lower.tri(distances)]
  minimum_spanning_tree(order(dists, decreasing = TRUE) - 1L)
}

test_that("MSTEdges() handles bad input", {
  expect_error(MSTEdges(matrix(1:12, 6, 2)), "distance")
})

test_that("minimum_spanning_tree.cpp handles bad input", {
  expect_equal(minimum_spanning_tree(numeric(0)), matrix(0, 0, 0))
  expect_error(minimum_spanning_tree(c(1:-1)),
               "`order` contains entries < 0")
  expect_error(minimum_spanning_tree(c(3, 100, 1)),
               "`order` contains entries > `length.order.`")
  expect_error(minimum_spanning_tree(c(3, 1, NA_real_)),
               "`order` contains NA")
  expect_error(minimum_spanning_tree(0:13),
               "`length.order.`.* not.* triangular")
})

test_that("MST edges calculated correctly", {
  set.seed(0)
  points <- matrix(c(0.1, 0, 1.9, 2, 1.1, 1,
                     0.1, 2,   0, 2, 1, 1.1,
                       0, 0,   0, 0, 1, -1), 6)
  distances <- dist(points)
  apeMst <- matrix(c(5, 6, 6, 5, 5, 1, 1:4), 5)
  distMat <- as.matrix(distances)
  expect_equal(MSTLength(distances, apeMst),
               MSTLength(distances))
  expect_equal(MSTLength(distances),
               MSTLength(distMat))
  MSTPlot <- function() {
    plot(points, asp = 1, ann = FALSE)
    expect_equal(MSTEdges(distances, FALSE),
                 MSTEdges(distances, TRUE, points[, 1], points[, 2]))
  }
  skip_if_not_installed("vdiffr", minimum_version = "1.0.0")
  skip_if(packageVersion("graphics") < "4.1.0")
  vdiffr::expect_doppelganger("MST plotting", MSTPlot)
})

test_that("MST handles large distance matrices", {
  x <- dist(0:300)
  expect_equal(c(300, 2), dim(MSTEdges(x)))
})

#' "Stemwardness" of a leaf
#'
#' Functions to describe the position of a leaf relative to the root.
#' "Stemmier" leaves ought to exhibit a smaller root-node distance and a
#' larger sister size.
#'
#' `RootNodeDistance()` calculates the number of nodes between the chosen leaf
#' and the root of `tree`.
#' This is an unsatisfactory measure, as the range of possible
#' distances is a function of the shape of the tree
#' \insertCite{Asher2020}{TreeTools}.
#' As an example, leaf _X1_ in the tree `(.,(.,(.,(.,(X1,(a,b))))))`
#' falls outside the clade _(a, b)_ and has a root-node distance of 4,
#' whereas leaf _X2_ in the tree `(.,((.,(.,.)),(b,(X2,a))))`
#' falls within the clade _(a, b)_, so should be considered more "crownwards",
#' yet has a smaller root-node distance (3).
#'
#' \insertFig{Stemwardness.png}{TreeTools}{
#'   par(mfrow = c(1, 2), mar = rep(0.3, 4))
#'   plot(ape::read.tree(text="(.,(.,(.,(.,(X1,(a,b))))));"))
#'   ape::nodelabels(1:4, 9:12)
#'   ape::edgelabels(1:2, 11:12)
#'
#'   plot(ape::read.tree(text="(.,((.,(.,.)),(b,(X2,a))));"))
#'   ape::nodelabels(1:3, c(9, 12, 13))
#'   ape::edgelabels(1, 12)
#' }
#'
#' `SisterSize()` measures the number of leaves in the clade that is sister to
#' the chosen leaf, as proposed by \insertCite{Asher2020;textual}{TreeTools}.
#' In the examples above, _X1_ has a sister size of 2 leaves, whereas _X2_,
#' which is "more crownwards", has a smaller sister size (1 leaf), as desired.
#'
#' @template treeParam
#' @param tip Either a numeric specifying the index of a single tip, or a
#' character specifying its label.
#'
#' @return `SisterSize()` returns an integer specifying the number of leaves
#' in the clade that is sister to `tip`.
#' `RootNodeDist()` returns an integer specifying the number of nodes between
#' `tip` and the root node of `tree`.
#'
#' @references \insertAllCited{}
#' @examples
#' bal8 <- BalancedTree(8)
#' pec8 <- PectinateTree(8)
#'
#' SisterSize(bal8, 3)
#' SisterSize(pec8, "t3")
#' SisterSize(RootTree(pec8, "t3"), "t3")
#'
#' RootNodeDist(bal8, 3)
#' RootNodeDist(pec8, "t3")
#' RootNodeDist(RootTree(pec8, "t3"), "t3")
#' @template MRS
#' @name Stemwardness
#' @family tree characterization functions
#' @export
SisterSize <- function(tree, tip) UseMethod("SisterSize", tip)

#' @export
SisterSize.NULL <- function(tree, tip) NULL

#' @rdname Stemwardness
#' @export
SisterSize.numeric <- function(tree, tip) {
  edge <- tree[["edge"]]
  parent <- edge[edge[, 2] == tip, 1]

  # Return:
  CladeSizes(tree, nodes = parent) - 1L # Subtract the tip itself
}

#' @rdname Stemwardness
#' @export
SisterSize.character <- function(tree, tip) {
  SisterSize(tree, which(tree[["tip.label"]] == tip))
}


#' @rdname Stemwardness
#' @export
RootNodeDistance <- function(tree, tip) UseMethod("RootNodeDistance", tip)

#' @export
RootNodeDistance.NULL <- function(tree, tip) NULL

#' @rdname Stemwardness
#' @export
RootNodeDistance.numeric <- function(tree, tip) {
  edge <- tree[["edge"]]
  parent <- edge[edge[, 2] == tip, 1]
  depths <- NodeDepth(tree)

  # Return:
  depths[NTip(tree) + 1L] - depths[parent]

}

#' @rdname Stemwardness
#' @export
RootNodeDistance.character <- function(tree, tip) {
  RootNodeDistance(tree, which(tree[["tip.label"]] == tip))
}

#' @rdname Stemwardness
#' @aliases RootNodeDist
#' @export
RootNodeDist <- RootNodeDistance

#' Identify vertices retained when leaves are dropped
#' 
#' @param tree Original tree of class `phylo`, in [`Preorder`].
#' @param keptTips Either: 
#' - a logical vector stating whether each leaf should be retained, in a
#' sequence corresponding to `tree[["tip.label"]]`; or
#' - a character vector listing the leaf labels to retain; or
#' - a numeric vector listing the indices of leaves to retain.
#' @param tipLabels Optional character vector naming the leaves of `tree`,
#' if `keptTips` is not logical.  Inferred from `tree` if unspecified.
#' @examples 
#' master <- BalancedTree(12)
#' master <- Preorder(master) # Nodes must be listed in Preorder sequence
#' plot(master)
#' nodelabels()
#' 
#' allTips <- master[["tip.label"]]
#' keptTips <- sample(allTips, 8)
#' plot(KeepTip(master, keptTips))
#' kept <- KeptVerts(master, allTips %in% keptTips)
#' 
#' map <- which(kept)
#' # Node `i` in the reduced tree corresponds to node `map[i]` in the original.
#' @template MRS
#' @export
#' @family tree manipulation
KeptVerts <- function(tree, keptTips, tipLabels = TipLabels(tree)) {
  UseMethod("KeptVerts")
}

#' @rdname KeptVerts
#' @export
KeptVerts.phylo <- function(tree, keptTips, tipLabels = TipLabels(tree)) {
  order <- attr(tree, "order")
  if (is.null(order) || order != "preorder") {
    stop("`tree` must be in preorder; try `Preorder(tree)`")
  }
  KeptVerts(tree[["edge"]], keptTips, tipLabels = tipLabels)
}

#' @rdname KeptVerts
#' @export
KeptVerts.numeric <- function(tree, keptTips, tipLabels = TipLabels(tree)) {
  switch(mode(keptTips),
         "numeric" = {
           x <- logical(length(tipLabels))
           x[keptTips] <- TRUE
           keptTips <- x
         },
         "character" = {
           keptTips <- tipLabels %in% keptTips
         },
         "logical" = {},
         stop("Unrecognized format for `keptTips`")
  )
  dims <- dim(tree)
  if (is.null(dims) || dims[2] != 2L) {
    stop("`tree` must be the numeric edge matrix of a `phylo` object")
  }
  ret <- kept_vertices(tree, keptTips)
  # Return:
  ret[-1] > 1L
}

#' Unique integer indices for bifurcating tree topologies
#'
#' Functions converting between phylogenetic trees and their unique decimal
#' representation, based on a concept by John Tromp, employed in 
#' \insertCite{Li1996;textual}{TreeTools}.
#'
#' There are `NUnrooted(n)` unrooted trees with _n_ leaves.
#' As such, each _n_-leaf tree can be uniquely identified by a non-negative
#' integer _x_ < `NUnrooted(n)`.
#'
#' This integer can be converted by a tree by treating it as a mixed-base
#' number, with bases 1, 3, 5, 7, &hellip; (2&nbsp;_n_ - 5).
#'
#' Each digit of this mixed base number corresponds to a leaf, and determines
#' the location on a growing tree to which that leaf should be added.
#'
#' We start with a two-leaf tree, and treat 0 as the origin of the tree.
#'
#' ```
#'
#' 0 ---- 1
#'
#' ```
#'
#' We add leaf 2 by breaking an edge and inserting a node (numbered
#' `2 + nTip - 1`).
#' In this example, we'll work up to a six-leaf tree; this node will be numbered
#' 2 + 6 - 1 = 7.
#' There is only one edge on which leaf 2 can be added.  Let's add node 7 and
#' leaf 2:
#'
#' ```
#'
#' 0 ---- 7 ---- 1
#'        |
#'        |
#'        2
#'
#' ```
#'
#' There are now three edges on which leaf 3 can be added.  Our options are:
#'
#' Option 0: the edge leading to 1;
#'
#' Option 1: the edge leading to 2;
#'
#' Option 2: the edge leading to 7.
#'
#' If we select option 1, we produce:
#'
#' ```
#'
#' 0 ---- 7 ---- 1
#'        |
#'        |
#'        8 ---- 2
#'        |
#'        |
#'        3
#'
#' ```
#' `1` is now the final digit of our mixed-base number.
#'
#' There are five places to add leaf 4:
#'
#' Option 0: the edge leading to 1;
#'
#' Option 1: the edge leading to 2;
#'
#' Option 2: the edge leading to 3;
#'
#' Option 3: the edge leading to 7;
#'
#' Option 4: the edge leading to 8.
#'
#' If we chose option 3, then `3` would be the penultimate digit of our
#' mixed-base number.
#'
#' If we chose option 0 for the next two additions, we could specify this tree
#' with the mixed-base number 0021.  We can convert this into decimal:
#'
#' 0 &times; (1 &times; 3 &times; 5 &times; 9) +
#'
#' 0 &times; (1 &times; 3 &times; 5) +
#'
#' 3 &times; (1 &times; 3) +
#'
#' 1 &times; (1)
#'
#' = 10
#'
#' `as.TreeNumber()` supports up to 51 leaves.
#' For trees with at most 19 leaves, the number fits in a 64-bit integer and
#' the `TreeNumber` is stored as an `integer64` (via the `bit64` package),
#' enabling arithmetic and exact round-tripping via `as.MixedBase()`.
#' For trees with 20--51 leaves, there are more than 2^64 distinct topologies,
#' so the tree number is stored as a decimal character string instead.
#'
#' Package developers can use the C++ header `TreeTools/tree_number.h`
#' (via `LinkingTo: TreeTools`) for the underlying 256-bit encoding
#' (`tree_num_t`) directly.
#'
#' @param x Integer identifying the tree (see details).
#' @param nTip Integer specifying number of leaves in the tree.
#' @template tipLabelsParam
#' @param \dots Additional parameters for consistency with S3 methods (unused).
#'
#' @examples
#' tree <- as.phylo(10, nTip = 6)
#' plot(tree)
#' as.TreeNumber(tree)
#'
#' # Trees with 20--51 leaves are stored as decimal strings:
#' as.TreeNumber(BalancedTree(19))  # integer64-backed
#' as.TreeNumber(BalancedTree(51))  # character-backed
#'
#' # If > 9 digits, represent the tree number as a string.
#' treeNumber <- as.TreeNumber("1234567890123", nTip = 14)
#' tree <- as.phylo(treeNumber)
#' @references
#' \insertAllCited{}
#' @encoding UTF-8
#' @seealso Describe the shape of a tree topology, independent of leaf labels:
#' [`TreeShape()`]
#' @family tree generation functions
#' @family 'TreeNumber' utilities
#' @name TreeNumber
#

#' @rdname TreeNumber
#'
#' @return `as.TreeNumber()` returns an object of class `TreeNumber` with
#' attributes `nTip` and `tip.label`.
#' For trees with at most 19 leaves the underlying storage is a single
#' `integer64` value (class `c("TreeNumber", "integer64")`), enabling
#' `integer64` arithmetic and exact round-tripping through `as.MixedBase()`.
#' For trees with 20--51 leaves the number exceeds 2^64, so it is stored as a
#' decimal character string (class `c("TreeNumber", "character")`).
#' If `x` is a list of trees or a `multiPhylo` object,
#' `as.TreeNumber()` returns a corresponding list of `TreeNumber` objects.
#' @export
as.TreeNumber <- function(x, ...) UseMethod("as.TreeNumber")

# Maximum tips for integer64-backed TreeNumber (64-bit integers exhausted at 20 leaves).
.TT_MAX_TIP <- 19L
# Maximum tips supported by tree_number.h (256-bit integer, matches TREE_NUM_MAX_TIP).
.TT_MAX_TIP_FULL <- 51L

# Multiply a decimal string by a small positive integer, add a small
# non-negative integer, and return the result as a decimal string.
# Safe for multiplier and addend up to INT_MAX (~2.1e9): all intermediate
# values fit exactly in a 64-bit double (< 2^53).
.str_mul_add <- function(str, multiplier, addend) {
  m <- as.double(multiplier)
  digits <- as.integer(strsplit(str, "")[[1L]])
  n <- length(digits)
  # result_digits[k] holds the k-th least-significant output digit.
  result_digits <- integer(n + 12L)
  carry <- as.double(addend)
  for (k in seq_len(n)) {
    val <- digits[[n + 1L - k]] * m + carry
    result_digits[[k]] <- as.integer(val %% 10)
    carry <- floor(val / 10)
  }
  k <- n + 1L
  while (carry > 0) {
    result_digits[[k]] <- as.integer(carry %% 10)
    carry <- floor(carry / 10)
    k <- k + 1L
  }
  paste0(rev(result_digits[seq_len(k - 1L)]), collapse = "")
}

# Convert an INT_MAX-packed integer vector to a decimal string via Horner's
# method: result = chunks[1]; for each subsequent chunk c: result = result*INT_MAX + c.
.chunks_to_decimal <- function(chunks) {
  if (length(chunks) == 1L) return(as.character(chunks[[1L]]))
  result <- as.character(chunks[[1L]])
  for (i in seq_along(chunks)[-1L]) {
    result <- .str_mul_add(result, 2147483647L, chunks[[i]])
  }
  result
}

# Divide a decimal string by a small positive integer using long division.
# Returns list(quotient = character string, remainder = integer).
.str_divmod <- function(str, divisor) {
  divisor <- as.double(divisor)   # double avoids int32 overflow (max rem*10 ~ 2e10)
  digits <- as.integer(strsplit(str, "")[[1L]])
  quotient <- integer(length(digits))
  rem <- 0
  for (i in seq_along(digits)) {
    cur <- rem * 10 + digits[[i]]
    quotient[[i]] <- as.integer(cur %/% divisor)
    rem <- cur %% divisor
  }
  q_str <- sub("^0+", "", paste0(quotient, collapse = ""))
  if (nchar(q_str) == 0L) q_str <- "0"
  list(quotient = q_str, remainder = as.integer(rem))
}

# Convert a decimal string to an INT_MAX-packed integer vector.
# Inverse of .chunks_to_decimal(); used when converting character-backed
# TreeNumbers back to the form expected by num_to_parent().
.decimal_to_chunks <- function(str) {
  if (length(str) == 0 || !nzchar(str) || str == "0") {
    0L
  } else {
    chunks <- integer(0)
    while (str != "0") {
      dm <- .str_divmod(str, 2147483647L)
      chunks <- c(dm$remainder, chunks)
      str <- dm$quotient
    }
    chunks
  }
}

# Calculate with:
# base <- cumprod(as.integer64(seq(1, by = 2, length.out = 16)))
.TT_BASE <- as.integer64(c(
  "191898783962510625", "6190283353629375", "213458046676875", "7905853580625",
  "316234143225", "13749310575", "654729075", "34459425", "2027025", "135135",
  "10395", "945", "105", "15", "3", "1"))

.TTBases <- function (n) .TT_BASE[length(.TT_BASE) - n + seq_len(n)]

#' @export
as.TreeNumber.NULL <- function(x, ...) NULL

#' @rdname TreeNumber
#' @export
as.TreeNumber.phylo <- function(x, ...) {
  x <- RootTree(x, 1)
  edge <- x[["edge"]]
  nTip <- NTip(x)
  if (nTip > .TT_MAX_TIP_FULL) {
    stop("`as.TreeNumber()` supports at most ", .TT_MAX_TIP_FULL, " leaves.")
  }

  edge <- Postorder(edge)
  chunks <- edge_to_num(edge[, 1], edge[, 2], nTip)
  if (nTip <= .TT_MAX_TIP) {
    structure(.Int64(chunks),
              nTip = nTip,
              tip.label = TipLabels(x),
              class = c("TreeNumber", "integer64"))
  } else {
    structure(.chunks_to_decimal(chunks),
              nTip = nTip,
              tip.label = TipLabels(x),
              class = c("TreeNumber", "character"))
  }
}

#' @rdname TreeNumber
#' @export
as.TreeNumber.multiPhylo <- function(x, ...) {
  lapply(x, as.TreeNumber.phylo)
}

#' @rdname TreeNumber
#' @export
as.TreeNumber.character <- function(x, nTip, tipLabels = TipLabels(nTip), ...) {
  if (nTip <= .TT_MAX_TIP) {
    structure(as.integer64(x),
              nTip = nTip,
              tip.label = tipLabels,
              class = c("TreeNumber", "integer64"))
  } else {
    structure(x,
              nTip = nTip,
              tip.label = tipLabels,
              class = c("TreeNumber", "character"))
  }
}

#' @rdname TreeNumber
#' @export
as.TreeNumber.TreeNumber <- function(x, ...) x

#' @export
#' @rdname TreeNumber
as.TreeNumber.MixedBase <- function(x, ...) {
  if (NTip(x) > .TT_MAX_TIP) {
    stop("Converting a `MixedBase` to a `TreeNumber` requires `integer64` ",
         "arithmetic and is only supported for trees with <= ", .TT_MAX_TIP,
         " leaves. Use `as.phylo()` to convert larger `MixedBase` objects.")
  }

  # Return:
  structure(sum(as.integer(x) * .TTBases(length(x))),
            nTip = NTip(x),
            tip.label = TipLabels(x),
            class = c("TreeNumber", "integer64"))
}

#' @rdname TreeNumber
#' @export
as.MixedBase.TreeNumber <- function(x, ...) {
  nTip <- NTip(x)
  if (nTip > .TT_MAX_TIP) {
    stop("Converting a `TreeNumber` to `MixedBase` requires `integer64` ",
         "arithmetic and is only supported for trees with <= ", .TT_MAX_TIP,
         " leaves. Use `as.MixedBase(as.phylo(x))` for larger trees.")
  }
  outLength <- nTip - 3L
  baseLength <- min(outLength, length(.TT_BASE))

  base <- .TTBases(baseLength)
  ret <- integer(baseLength)
  n <- as.integer64(x)
  for (i in seq_len(baseLength)) {
    baseI <- base[i]
    ret[i] <- as.integer(n %/% baseI)
    n <- n %% baseI
  }

  # Return:
  structure(c(integer(outLength - baseLength), ret),
            nTip = NTip(x),
            tip.label = TipLabels(x),
            class = "MixedBase")
}

#' @rdname TreeNumber
#' @export
as.MixedBase.integer64 <- function(x, tipLabels = NULL, ...) {
  baseLength <- if (x > max(.TT_BASE)) {
    length(.TT_BASE)
  } else {
    which.max(rev(.TT_BASE) > x)
  }
  tipLabels <- TipLabels(tipLabels)
  if (is.null(tipLabels)) {
    tipLabels <- TipLabels(baseLength + 3L)
  }
  nTip <- length(tipLabels)
  outLength <- nTip - 3L
  if (outLength < baseLength) {
    stop("Number of tips too low; is tipLabels truncated?")
  }
  
  base <- .TTBases(baseLength)
  ret <- integer(baseLength)
  
  for (i in seq_len(baseLength)) {
    baseI <- base[i]
    ret[i] <- as.integer(x %/% baseI)
    x <- x %% baseI
  }
  
  # Return:
  structure(c(integer(outLength - baseLength), ret),
            nTip = outLength + 3L,
            tip.label = tipLabels,
            binary = TRUE,
            class = "MixedBase")
}

#' @rdname TreeNumber
#' @export
as.MixedBase.numeric <- function(x, tipLabels = NULL, ...) {
  as.MixedBase.integer64(as.integer64(x), tipLabels = tipLabels, ...)
}
                                  
#' @rdname TreeNumber
#' @template MRS
#' @return `as.phylo.numeric()` returns a tree of class `phylo`.
#'
#' @examples
#' as.phylo(0:2, nTip = 6, tipLabels = letters[1:6])
#'
#' @importFrom ape as.phylo
#' @export
as.phylo.numeric <- function(x, nTip = attr(x, "nTip"),
                             tipLabels = attr(x, "tip.label"), ...) {
  if (is.null(nTip)) {
    if (is.null(tipLabels)) {
      stop("Either nTip or tipLabels must be specified.")
    } else {
      nTip <- length(tipLabels)
    }
  }
  if (nTip < 2) {
    if (nTip == 0) {
      ZeroTaxonTree()
    } else if (nTip == 1) {
      SingleTaxonTree(if (is.null(tipLabels)) TipLabels(nTip) else tipLabels)
    } else {
      stop("`nTip` may not be negative")
    }
  } else {
    if (is.null(tipLabels)) {
      tipLabels <- TipLabels(nTip)
    }
    if (length(x) > 1) {
      structure(lapply(x, as.phylo.numeric, nTip = nTip, tipLabels = tipLabels),
                tip.label = tipLabels, class = "multiPhylo")
    } else {
      edge <- RenumberEdges(num_to_parent(.Int64.to.C(x), nTip),
                            seq_len(nTip + nTip - 2L))
      .PreorderTree(
        edge = do.call(cbind, edge),
        tip.label = tipLabels,
        Nnode = nTip - 1L
      )
    }
  }
}

# Copied from as.phylo.numeric except if length > 1
#' @export
as.phylo.integer64 <- function(x, nTip = attr(x, "nTip"),
                                tipLabels = attr(x, "tip.label"), ...) {
  if (is.null(nTip)) {
    if (is.null(tipLabels)) {
      stop("Either nTip or tipLabels must be specified.")
    } else {
      nTip <- length(tipLabels)
    }
  }
  if (is.null(tipLabels)) {
    tipLabels <- paste0("t", seq_len(nTip))
  }
  if (nTip == 1) {
    SingleTaxonTree(tipLabels)
  } else {
    if (length(x) > 1) {
      ret <- vector("list", length(x))
      for (i in seq_along(x)) {
        ret[[i]] <- as.phylo(x[i], nTip = nTip, tipLabels = tipLabels)
      }
      ret <- structure(ret, tip.label = tipLabels, class = "multiPhylo")
    } else {
      edge <- RenumberEdges(num_to_parent(.Int64.to.C(x), nTip),
                            seq_len(nTip + nTip - 2L))
      .PreorderTree(edge = do.call(cbind, edge),
                    tip.label = tipLabels,
                    Nnode = nTip - 1L)
    }
  }
}

.Int64.to.C <- function(i64) {
  INT_MAX <- as.integer64(2147483647L)
  i64 <- as.integer64(i64)
  if (i64 > INT_MAX) {
    if (i64 > INT_MAX * INT_MAX) {
      stop("Number too large for 64-bit representation")
    }
    as.integer(c(i64 %/% INT_MAX, i64 %% INT_MAX))
  } else {
    as.integer(i64[1])
  }
}

#' @rdname TreeNumber
#' @export
as.phylo.TreeNumber <- function(x, nTip = attr(x, "nTip"),
                                 tipLabels = attr(x, "tip.label"), ...) {
  if (is.null(tipLabels)) tipLabels <- paste0("t", seq_len(nTip))
  chunks <- if (inherits(x, "integer64")) {
    .Int64.to.C(x)
  } else {
    .decimal_to_chunks(as.character(x))
  }
  edge <- RenumberEdges(num_to_parent(chunks, nTip),
                        seq_len(nTip + nTip - 2L))
  .PreorderTree(
    edge = do.call(cbind, edge),
    tip.label = tipLabels,
    Nnode = nTip - 1L
  )
}

#' @export
as.integer64.TreeNumber <- function(x, ...) {
  if (inherits(x, "integer64")) {
    structure(x[1], class = "integer64")
  } else {
    nTip <- attr(x, "nTip")
    if (nTip > .TT_MAX_TIP) {
      stop("Cannot convert a ", nTip, "-leaf TreeNumber to integer64: ",
           "trees with > ", .TT_MAX_TIP, " leaves require more than 64 bits.")
    }
    as.integer64(as.character(x))
  }
}

#' Is an object a `TreeNumber` object?
#' 
#' @param x R object.
#' @return `is.TreeNumber()` returns a logical vector of length one specifying
#' whether `x` inherits the class `"TreeNumber"`.
#' @template MRS
#' @examples
#' is.TreeNumber(FALSE) # FALSE 
#' is.TreeNumber(as.TreeNumber(BalancedTree(5))) # TRUE
#' @family 'TreeNumber' utilities
#' @export
is.TreeNumber <- function(x) inherits(x, "TreeNumber")

#' Print `TreeNumber` object
#'
#' S3 method for objects of class `TreeNumber`.
#'
#' @param x Object of class `TreeNumber`.
#' @param \dots Additional arguments for consistency with S3 method (unused).
#'
#' @family 'TreeNumber' utilities
#' @export
print.TreeNumber <- function(x, ...) {
  nTip <- attr(x, "nTip")
  cat("Phylogenetic tree number", .PrintedTreeNumber(x), "of",
      .PrintNUnrooted(nTip), "\n",
      nTip, "tips:", paste0(attr(x, "tip.label")))
}

#' @keywords internal
.PrintNUnrooted <- function(n) {
  if (n < 15L || n > 19L) NUnrooted(n) else paste0(NUnrooted64(n))
}

# x: Object of class `TreeNumber`
#' @keywords internal
.PrintedTreeNumber <- function(x) {
  if (inherits(x, "integer64")) {
    paste0(structure(x, class = "integer64"))
  } else {
    as.character(x)   # already a decimal string for character-backed TreeNumbers
  }
}

#' @rdname TreeNumber
#' @export
as.MixedBase <- function(x, ...) UseMethod("as.MixedBase")

#' @rdname TreeNumber
#' @export
as.MixedBase.MixedBase <- function(x, ...) x

#' @export
as.MixedBase.NULL <- function(x, ...) NULL

#' @rdname TreeNumber
#' @export
as.MixedBase.phylo <- function(x, ...) {
  x <- RootTree(x, 1)
  nTip <- NTip(x)
  edge <- x[["edge"]]
  
  edge <- Postorder(edge)
  structure(edge_to_mixed_base(edge[, 1], edge[, 2], nTip),
            nTip = nTip,
            tip.label = TipLabels(x),
            binary = TRUE,
            class = "MixedBase")
}

#' @rdname TreeNumber
#' @export
as.MixedBase.multiPhylo <- function(x, ...) {
  lapply(x, as.MixedBase, ...)
}

#' @rdname TreeNumber
#' @export
as.phylo.MixedBase <- function(x, nTip = attr(x, "nTip"),
                               tipLabels = attr(x, "tip.label"), ...) {
  if (is.null(tipLabels)) {
    tipLabels <- paste0("t", seq_len(nTip))
  }
  edge <- RenumberEdges(mixed_base_to_parent(x, nTip),
                        seq_len(nTip + nTip - 2L))
  
  # Return:
  .PreorderTree(
    edge = do.call(cbind, edge),
    tip.label = tipLabels,
    Nnode = nTip - 1L
  )
}

test_that("Integers fit into integer sizes", {
  INT_MAX <- 2147483647L
  expect_equal(as.integer64(INT_MAX), as.integer64("2147483647"))
  expect_gt(.TT_BASE[1], .TT_BASE[2])
})

test_that("as.phylo.numeric()", {
  expect_true(all.equal(as.phylo(0:2, 6, letters[1:6])[[1]],
                        PectinateTree(letters[c(1, 3:6, 2)])))
  expect_true(all.equal(as.phylo(0, tipLabels = letters[1:6]),
                        PectinateTree(letters[c(1, 3:6, 2)])))
  expect_error(as.phylo(0))
  expect_equal(as.phylo(123, nTip = 8),
               as.phylo(123L, nTip = 8))
  expect_equal(as.phylo(123, nTip = 8),
               as.phylo(as.integer64(123), nTip = 8))
  expect_equal(as.phylo(123:124, nTip = 10),
               as.phylo(as.integer64(123:124), nTip = 10))
  expect_equal(as.phylo(as.integer64(10), tipLabels = paste0("t", 1:8)),
               as.phylo(10, 8))
})

test_that("as.TreeNumber() error handling", {
  expect_error(as.phylo(integer64(1), NULL, NULL))
  expect_error(as.TreeNumber(StarTree(5)))
  expect_error(as.TreeNumber(BalancedTree(52)))  # > 51 leaves not supported
})

test_that("as.TreeNumber()", {
  expect_equal(SingleTaxonTree("t1"), as.phylo(integer64(1), 1, NULL))
  expect_equal(c("Phylogenetic tree number 0 of 105 ",
                 " 6 tips: t1 t2 t3 t4 t5 t6"),
               capture.output(print(as.TreeNumber(as.phylo(105, 6)))))
  expect_equal("Phylogenetic tree number 105 of 6332659870762850625 ",
               capture.output(print(as.TreeNumber(as.phylo(105, 19))))[1])
  # 20+ leaves: no warning, decimal string storage
  expect_no_warning(as.TreeNumber(BalancedTree(20)))
  expect_equal("Phylogenetic tree number 105 of 2.216431e+20 ",
               capture.output(print(as.TreeNumber(as.phylo(105, 20))))[1])
  expect_equal(unlist(lapply(1:3, as.integer64)),
               unlist(as.TreeNumber(as.phylo(1:3, 6))))

  tn6 <- as.TreeNumber(as.phylo(1, 6, letters[1:6]))
  expect_equal(as.phylo(1, 6), as.phylo(tn6, tipLabels = NULL))

  tn16 <- as.TreeNumber(as.phylo(1, 16, letters[1:16]))
  expect_equal(as.phylo(1, 16), as.phylo(tn16, tipLabels = NULL))
  expect_equal(as.TreeNumber(as.MixedBase(tn16)), as.TreeNumber(tn16))

  expect_error(as.phylo(as.integer64(2) ^ 62, 16), "too large ")
  bigNumber <- as.integer64(2) ^ 61 + 1
  # Not necessarily equal to as.integer64(2 ^ 61 + 1), due to numeric conversion
  tn16 <- as.TreeNumber(as.phylo(bigNumber, 16, letters[1:16]))
  expect_equal(as.phylo(bigNumber, 16), as.phylo(tn16, tipLabels = NULL))
  # Rounding:
  expect_identical(as.integer64(as.TreeNumber(as.phylo(bigNumber, 16))) - 1,
                   as.integer64(as.TreeNumber(as.phylo(bigNumber - 1, 16))))
               
  expect_equal(as.TreeNumber(as.MixedBase(tn16)), tn16)

  # Large trees: character-backed, round-trips correctly.
  # as.phylo.TreeNumber returns Preorder(RootTree(x, 1)), so compare against that.
  for (nTip in c(20L, 35L, 51L)) {
    tree <- BalancedTree(nTip)
    tn <- as.TreeNumber(tree)
    expect_s3_class(tn, "TreeNumber")
    expect_true(inherits(tn, "character"))
    expect_false(inherits(tn, "integer64"))
    expect_equal(Preorder(RootTree(tree, 1)), as.phylo(tn, tipLabels = NULL))
  }
})

test_that("as.MixedBase()", {
  expect_error(as.MixedBase(RandomTree(100000)),
               "Too many \\w+ for mixed base")
  expect_equal(as.integer(as.TreeNumber(as.phylo(16, 16))), 16)
  nTip <- 9
  expect_equal(as.integer(as.MixedBase(as.TreeNumber(as.phylo(16, nTip)))),
               c(rep(0, nTip - 3 - 3), 1, 0, 1))
  
  expect_equal(as.MixedBase(as.phylo(0, 6)),
               structure(integer(3),
                         nTip = 6,
                         tip.label = paste0("t", 1:6),
                         binary = TRUE,
                         class = "MixedBase"))
  
  expect_equal(as.MixedBase(as.phylo(105, 20)),
               structure(tabulate(17 - 3, 17),
                         nTip = 20,
                         tip.label = paste0("t", 1:20),
                         binary = TRUE,
                         class = "MixedBase"))
  
  expect_equal(unlist(lapply(1:3, function(n) as.MixedBase(as.phylo(n, 6)))),
               unlist(as.MixedBase(as.phylo(1:3, 6))))
  
  mb3 <- as.MixedBase(as.phylo(1, 3, letters[1:3]))
  expect_equal(as.phylo(1, 3), as.phylo(mb3, tipLabels = NULL))
  
  mb6 <- as.MixedBase(as.phylo(1, 6, letters[1:6]))
  expect_equal(as.phylo(1, 6), as.phylo(mb6, tipLabels = NULL))
  
  mb16 <- as.TreeNumber(as.phylo(1337, 16, letters[1:16]))
  expect_equal(as.phylo(1337, 16), as.phylo(mb16, tipLabels = NULL))
  
  expect_error(as.MixedBase(42, TipLabels(2)))
  expect_true(as.MixedBase(as.phylo(1337, 16)) == as.MixedBase(1337, 16))
  expect_true(as.MixedBase(44) > as.MixedBase(as.MixedBase(42)))
  expect_true(as.MixedBase(44, 10) < as.MixedBase(42, 11))
  bigNo <- as.integer64(2) ^ 62 + 1337
  expect_equal(sum(as.integer(as.MixedBase(bigNo)) * .TT_BASE), bigNo)
})

test_that("as.MixedBase() supports larger trees", {
  expect_equal(tail(as.MixedBase(BalancedTree(100))),
               c(6, 10, 8, 3, 4, 1) # From observation, not calculation
  )
})

test_that("is.TreeNumber()", {
  expect_equal(is.TreeNumber(as.integer64(5)), FALSE)
  expect_equal(is.TreeNumber(as.TreeNumber(BalancedTree(5))), TRUE)
})

test_that("tree_number.h add_small carry propagation", {
  # The tree number 2^64 + 1 = "18446744073709551617" has packed chunks [4, 8, 5].
  # Reconstructing it via packed_to_tree_num() calls add_small(5) after mul_small()
  # leaves w[0] = 2^64 - 4, so w[0] + 5 overflows, triggering the carry-propagation
  # loop (lines 86-88 of inst/include/TreeTools/tree_number.h).
  # NUnrooted(20) > 2^64, so this is a valid 20-leaf tree number.
  tn <- as.TreeNumber("18446744073709551617", nTip = 20L)
  expect_true(inherits(tn, "character"))
  expect_equal(tn, as.TreeNumber(as.phylo(tn, tipLabels = NULL)))
})

test_that(".decimal_to_chunks() handles zero and empty input", {
  # Calling as.phylo() on a character-backed TreeNumber of "0" routes through
  # .decimal_to_chunks("0"), hitting the early-return branch.
  tn_zero <- as.TreeNumber("0", nTip = 20L)
  expect_equal(as.phylo(tn_zero, tipLabels = NULL),
               as.phylo(0L, nTip = 20L))

  # Direct internal check: both "0" and "" return 0L immediately.
  expect_equal(.decimal_to_chunks("0"), 0L)
  expect_equal(.decimal_to_chunks(""), 0L)
  expect_equal(.decimal_to_chunks(character()), 0L)

  # The else-branch is always taken when input is non-zero;
  # the `0L` branch is unreachable (while loop always appends >= 1 element).
  expect_equal(.decimal_to_chunks("1"), 1L)
})

test_that("as.TreeNumber.character() creates character-backed result for nTip > .TT_MAX_TIP", {
  tn <- as.TreeNumber("12345678901234", nTip = 25L)
  expect_s3_class(tn, "TreeNumber")
  expect_true(inherits(tn, "character"))
  expect_false(inherits(tn, "integer64"))
  expect_equal(attr(tn, "nTip"), 25L)
})

test_that("as.TreeNumber.MixedBase() errors for trees with > .TT_MAX_TIP leaves", {
  mb <- as.MixedBase(BalancedTree(20))
  expect_error(as.TreeNumber(mb), "MixedBase.*TreeNumber")
  tn <- as.TreeNumber(BalancedTree(20))
  expect_error(as.MixedBase(tn), "TreeNumber.*MixedBase")
})

test_that("as.integer64.TreeNumber() for character-backed TreeNumbers", {
  # converting a large (nTip > 19) character-backed TreeNumber
  # to integer64 is impossible; should error.
  tn_large <- as.TreeNumber(BalancedTree(20))
  expect_error(as.integer64(tn_large), "integer64")

  # a character-backed TreeNumber with nTip <= 19 (atypical but valid
  # as a manually-constructed object) can be converted exactly to integer64.
  tn_small_char <- structure("10", nTip = 6L, tip.label = paste0("t", 1:6),
                             class = c("TreeNumber", "character"))
  expect_equal(as.integer64(tn_small_char), as.integer64(10))
})

#' Force a tree to match a constraint
#'
#' Modify a tree such that it matches a specified constraint.
#' This is at present a somewhat crude implementation that attempts to retain
#' much of the structure of `tree` whilst guaranteeing compatibility with
#' each entry in `constraint`.
#'
#' @template treeParam
#' @template constraintParam
#'
#' @return `ImposeConstraint()` returns a tree of class `phylo`, consistent
#' with `constraint`.
#'
#' @examples
#' tips <- letters[1:9]
#' tree <- as.phylo(1, 9, tips)
#' plot(tree)
#'
#' constraint <- StringToPhyDat("0000?1111 000111111 0000??110", tips, FALSE)
#' plot(ImposeConstraint(tree, constraint))
#' @template MRS
#' @family tree manipulation
#' @export
ImposeConstraint <- function(tree, constraint) {
  # This function is as efficient as it is elegant: i.e. not.
  # But it just about does the job.
  tree <- Preorder(tree)
  const <- AddUnconstrained(
    constraint,
    toAdd = setdiff(tree[["tip.label"]], TipLabels(constraint)),
    asPhyDat = FALSE
  )

  info <- apply(const, 2,
                function(x) SplitInformation(sum(x == "0"), sum(x == "1")))
  smallest <- ifelse(apply(const, 2, function(x) sum(x == "0") < sum(x == "1")),
                     "0", "1")

  tips <- tree[["tip.label"]]
  nTip <- length(tips)
  for (i in order(info)) {
    constI <- const[, i]
    collapsers <- constI == smallest[i]
    collapseNames <- names(collapsers[collapsers])
    if (length(collapseNames) < 2L) {
      warning("Could not apply constraint ", i,
              ". Is it incompatible or redundant?")
      next
    }
    collapsing <- apply(const[collapsers, , drop = FALSE], 2,
                        function(x) setdiff(x, "?")[1])

    const <- const[setdiff(rownames(const), collapseNames[-1]), , drop = FALSE]
    const[collapseNames[[1]], ] <- collapsing
    rownames(const)[match(collapseNames[[1]], rownames(const))] <- paste0(
      "(", paste0(collapseNames, collapse = ","), ")")

  }

  backbone <- Preorder(RenumberTips(
    .TextToTree("(", paste0(rownames(const), collapse = ","), ");"),
    tips))

  .ChildAtEnd <- function(x) {
    if (x <= nTip) x else .ChildAtEnd(edge[match(x, edge[, 1]), 2])
  }

  edge <- backbone[["edge"]]
  polytomies <- which(tabulate(edge[, 1]) > 2)
  
  for (node in polytomies) {
    nodeKids <- edge[edge[, 1] == node, 2]
    standIns <- vapply(nodeKids, .ChildAtEnd, 1)
    kept <- KeepTip(tree, standIns)[["edge"]]
    newNodes <- kept > length(standIns)
    kept[newNodes] <- kept[newNodes] - kept[1] + max(edge[, 1])
    kept[kept == kept[[1]]] <- node

    kept2 <- kept[, 2] # don't replace twice if standins[i - 1] < i
    for (i in seq_along(standIns)) {
      kept[, 2][kept2 == i] <- nodeKids[i]
    }
    edge <- rbind(edge[edge[, 1] != node, ], kept)
  }
  edge <- edge[order(edge[, 1]), ]
  backbone[["edge"]] <- RenumberTree(edge[, 1], edge[, 2])
  backbone[["Nnode"]] <- max(backbone[["edge"]][, 1]) - nTip
  nodeLabels <- tree[["node.label"]]
  if (!is.null(nodeLabels)) {
    backbone[["node.label"]] <- .UpdateNodeLabel(backbone[["edge"]], tree)
  }

  # Return:
  backbone
}

#' @importFrom ape read.tree
.TextToTree <- function(...) {
  space <- "XXTREETOOLSSPACEXX"
  text <- gsub(" ", space, paste0(...), fixed = TRUE)
  tree <- read.tree(text = text)
  tree[["tip.label"]] <- gsub(space, " ", tree[["tip.label"]], fixed = TRUE)

  # Return:
  tree
}

#' @describeIn ImposeConstraint Expand a constraint to include unconstrained
#' taxa.
#' @param toAdd Character vector specifying taxa to add to constraint.
#' @param asPhyDat Logical: if `TRUE`, return a `phyDat` object; if `FALSE`,
#' return a matrix.
#' @export
AddUnconstrained <- function(constraint, toAdd, asPhyDat = TRUE) {
  ret <- if (inherits(constraint, "phyDat")) {
    PhyDatToMatrix(constraint)
  } else if (inherits(constraint, "phylo")) {
    t(as.matrix(constraint))
  } else if (is.null(dim(constraint))) {
    cbind(constraint)
  } else {
    constraint
  }
  
  toAdd <- setdiff(toAdd, rownames(ret))
  ret <- if (is.null(ret)) {
    matrix("?", length(toAdd), 0, dimnames = list(toAdd, NULL))
  } else {
    rbind(ret, matrix("?", length(toAdd), dim(ret)[2],
                      dimnames = list(toAdd, NULL)))
  }

  # Return:
  if (asPhyDat) {
    MatrixToPhyDat(ret)
  } else {
    ret
  }
}

test_that("AddUnconstrained() handles zero-character phyDat", {
  star <- ape::read.tree(text = "(a,b,c,d);")
  empty_pd <- MatrixToPhyDat(t(as.matrix(star)))

  # PhyDatToMatrix returns a correctly-dimensioned 0-column matrix
  mat <- PhyDatToMatrix(empty_pd)
  expect_equal(dim(mat), c(4L, 0L))
  expect_equal(rownames(mat), c("a", "b", "c", "d"))

  # AddUnconstrained returns 0-character phyDat with extra taxa
  result <- AddUnconstrained(empty_pd, c("e", "f", "g"))
  expect_s3_class(result, "phyDat")
  expect_equal(names(result), c("a", "b", "c", "d", "e", "f", "g"))
  expect_equal(attr(result, "nr"), 0L)

  # asPhyDat = FALSE returns a 0-column matrix
  result_mat <- AddUnconstrained(empty_pd, c("e", "f", "g"), asPhyDat = FALSE)
  expect_equal(dim(result_mat), c(7L, 0L))
  expect_equal(rownames(result_mat), c("a", "b", "c", "d", "e", "f", "g"))
})

test_that("AddUnconstrained() works", {
  tips <- letters[1:9]
  constraint <- StringToPhyDat("0000?1111 000111111 0000??110", tips, FALSE)
  expect_equal(
    AddUnconstrained(constraint, letters[10:12], FALSE),
    PhyDatToMatrix(AddUnconstrained(constraint, letters[10:12], TRUE))
  )
  expect_equal(
    PhyDatToMatrix(AddUnconstrained(constraint, letters[8:12], TRUE)),
    AddUnconstrained(constraint, letters[10:12], FALSE)
  )
  
  noConstraint <- matrix(NA_character_, 6, 0,
                         dimnames = list(letters[1:6], NULL))
  expect_equal(
    AddUnconstrained(SingleTaxonTree("a"), letters[2:6], FALSE),
    noConstraint
  )
  
  expect_equal(AddUnconstrained(c(), letters[1:6], FALSE), noConstraint)
  expect_equal(AddUnconstrained(NULL, letters[1:6], FALSE), noConstraint)
  
  noNode <- structure(
    list(edge = structure(integer(0), dim = c(0L, 2L)),
         Nnode = 0L, tip.label = "a"), order = "preorder", class = "phylo")
  expect_equal(AddUnconstrained(noNode, letters[1:6], FALSE), noConstraint)
  
})

test_that("ImposeConstraint() works", {
  tips <- letters[1:9]
  tree <- as.phylo(1, 9, tips)
  tree[["node.label"]] <- paste("Node", 10:17)
  
  expect_equal(ImposeConstraint(tree, c()), tree)
  expect_equal(ImposeConstraint(tree, KeepTip(tree, character(0))), tree)
  
  # Weak constraint, test node labels
  constraint <- StringToPhyDat("000000011", tips, FALSE)
  constrained <- ImposeConstraint(tree, constraint)
  expect_true(all.equal(
    constrained,
    read.tree(text = "(a, ((c, d), (e, (f, (g, (b, (h, i)))))));")))
  expect_equal(constrained[["node.label"]],
               c(paste("Node", 10:15), NA, "Node 17"))

  # Strong constraint for very different tree
  constraint <- StringToPhyDat("0000?1111 000111111 0000??110", tips, FALSE)
  constrained <- ImposeConstraint(tree, constraint)
  expect_true(all.equal(
    constrained,
    read.tree(text = "((a, (b, c)), (d, (e, (f, (i, (g, h))))));")))
  expect_equal(constrained[["node.label"]], c("Node 10", rep(NA, 7)))

  expect_equal(ImposeConstraint(tree, constraint),
               ImposeConstraint(tree, PhyDatToMatrix(constraint)))
  
  expect_equal(
    ImposeConstraint(tree, setNames(c(rep(0, 3), rep(1, 6)), tips)),
    ImposeConstraint(tree, StringToPhyDat("000111111", tips, FALSE))
  )

  constraint <- StringToPhyDat("00001111 00011111 0000?110", tips[-5], FALSE)
  expect_true(all.equal(
    ImposeConstraint(tree, constraint),
    read.tree(text = "((a, (b, c)), (d, (e, (f, (i, (g, h))))));")))

  constraint <- ape::read.tree(text = "(a, b, c, (d, (f, (g, h), i)));")
  expect_true(all.equal(
    ImposeConstraint(tree, constraint),
    read.tree(text = "((a, (b, c)), (d, (e, (f, (i, (g, h))))));")))
  
  constraint <- MatrixToPhyDat(matrix(
    c(0, 0, "?", "?", 1, 1,
      1, 1,   1, "?", 0, 0), ncol = 2,
    dimnames = list(letters[1:6], NULL)))

  expect_warning({out <- ImposeConstraint(NJTree(constraint), constraint)})
  expect_true(inherits(out, "phylo"))

  # Need to collapse splits intelligently to avoid error.  From Joe Moysiuk.
  constraint <- MatrixToPhyDat(
    structure(c(0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 1L, 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 1L, 1L, 1L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 1L, 1L, 1L, 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 1L, 1L, 1L, 1L, 1L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 1L, 1L, 1L, 1L, 1L,
                1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 1L, 0L, 0L, 1L, 1L, 1L, 1L,
                1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 1L, 1L, 1L,
                1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                1L, 1L, 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 1L, 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 1L, 0L, 0L, 1L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 1L,
                1L, 0L, 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 1L, 1L, 1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                1L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L,
                0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 0L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L),
              .Dim = c(38L, 15L),
              .Dimnames = list(c("t1", "t2", "t3", "t4", "t5", "t6", "t7", "t8",
                                 "t9", "t10", "t11", "t12", "t13", "t14", "t15",
                                 "t16", "t17", "t18", "t19", "t20", "t21",
                                 "t22", "t23", "t24", "t25", "t26", "t27",
                                 "t28", "t29", "t30", "t31", "t32", "t33",
                                 "t34", "t35", "t36", "t37", "t38"), NULL)))
  ImposeConstraint(NJTree(constraint), constraint)
})

#' Identify taxa with long branches
#' 
#' The long branch (\acronym{LB}) score \insertCite{Struck2014}{TreeTools}
#' measures the deviation of the average pairwise patristic distance of a leaf
#' from all other leaves in a tree, relative to the average leaf-to-leaf
#' distance.
#' 
#' \insertCite{Struck2014;textual}{TreeTools} proposes the standard deviation
#' of LB scores as a measure of heterogeneity that can be compared between
#' trees; and the upper quartile of LB scores as "a representative value for
#' the taxa with the longest branches".
#' 
#' @template tree(s)Param
#' @returns `LongBranch()` returns a vector giving the long branch score for 
#' each leaf in `tree`, or a list of such vectors if `tree` is a list.
#' Results are given as raw deviations, without multiplying by 100 as proposed
#' by \insertCite{Struck2014;textual}{TreeTools}.
#' @family tree properties
#' @examples
#' tree <- BalancedTree(8, lengths = c(rep(2, 4), 5:7, rep(2, 4), rep(1, 3)))
#' lb <- LongBranch(tree)
#' tree$tip.label <- paste(tree$tip.label, signif(lb, 3), sep = ": ")
#' plot(tree, tip.col = SupportColour((1 - lb) / 2), font = 2)
#' 
#' # Standard deviation of LB scores allows comparison with other trees
#' sd(lb)
#' evenLengths <- BalancedTree(8, lengths = jitter(rep(1, 14)))
#' sd(LongBranch(evenLengths))
#' 
#' # Upper quartile identifies taxa with longest branches
#' threshold <- quantile(lb, 0.75)
#' tree$tip.label[lb > threshold]
#' @template MRS
#' @export
LongBranch <- function(tree) {
  UseMethod("LongBranch")
}

#' @importFrom ape cophenetic.phylo
#' @importFrom stats cophenetic
#' @export
LongBranch.phylo <- function(tree) {
  if (is.null(tree[["edge.length"]])) {
    warning("No edge lengths specified")
    return(NULL)
  }
  
  patristic <- cophenetic(tree)
  pairs <- colSums(patristic)
  
  # Mean pairwise distance of taxon i to all other taxa in tree
  pdi <- pairs # / (length(pairs) - 1) cancels
  # Average pairwise distance across all taxa
  pda <- sum(pairs) / length(pairs) # / (length(pairs) - 1) cancels
  
  # Return:
  (pdi / pda) - 1
}

#' @export
LongBranch.multiPhylo <- function(tree) {
  lapply(tree, LongBranch.phylo)
}

#' @export
LongBranch.list <- function(tree) {
  if (all(vapply(tree, inherits, TRUE, "phylo"))) {
    LongBranch.multiPhylo(tree)
  } else {
    stop("`tree` must be a list of 'phylo' objects")
  }
}

#' @export
LongBranch.NULL <- function(tree) NULL

test_that("LongBranch() fails safely", {
  expect_error(LongBranch(list(1, 2)), "list of .phylo. objects")
  expect_warning(expect_null(LongBranch(BalancedTree(8))), "edge lengths")
  expect_null(LongBranch(NULL))
})

test_that("LongBranch() succeeds", {
  evenLength <- BalancedTree(8, rep(1, 14))
  expect_equal(unname(LongBranch(evenLength)), rep(0, 8))
  tree <- BalancedTree(8, lengths = c(rep(2, 4), 5:7, rep(2, 4), rep(1, 3)))
  lb <- LongBranch(tree)
  pat <- ape::cophenetic.phylo(tree)
  expect_equal(lb, (colSums(pat) / (NTip(tree) - 1) /
                      mean(pat[upper.tri(pat)])) - 1)
  expect_equal(LongBranch(c(tree, tree)), list(lb, lb))
  expect_equal(LongBranch(c(evenLength, tree)),
               LongBranch(list(evenLength, tree)))
})

#' Numeric index of each node in a tree
#' `NodeNumbers()` returns a sequence corresponding to the nodes in a tree
#' 
#' @template treeParam
#' @param tips Logical specifying whether to also include the indices of leaves.
#' @return `NodeNumbers()` returns an integer vector corresponding to the
#' indices of nodes within a tree.
#' @template MRS
#' @family tree properties
#' @family tree navigation
#' @export
NodeNumbers <- function(tree, tips = FALSE) {
  if (tips) {
    seq_len(NTip(tree) + tree[["Nnode"]])
  } else {
    NTip(tree) + seq_len(tree[["Nnode"]])
  }
}


#' Distance of each node from tree exterior
#'
#' `NodeDepth()` evaluates how "deep" each node is within a tree.
#'
#' For a rooted tree, the depth of a node is the minimum (if `shortest = TRUE`)
#' or maximum  (`shortest = FALSE`) number of edges that must be traversed,
#' moving away from the root, to reach a leaf.
#'
#' Unrooted trees are treated as if a root node occurs in the "middle" of the
#' tree, meaning the position that will minimise the maximum node depth.
#'
#'
#' @template xPhylo
#' @param shortest Logical specifying whether to calculate the length of the
#' shortest away-from-root path to a leaf.  If `FALSE`, the length of the
#' longest such route will be returned.
#' @param includeTips Logical specifying whether to include leaves
#' (each of depth zero) in return value.
#'
#' @return `NodeDepth()` returns an integer vector specifying the depth of
#' each external and internal node in `x`.
#'
#' @template MRS
#' @family tree navigation
#' @seealso [`ape::node.depth`] returns the number of tips descended from a
#' node.
#'
#' @examples
#' tree <- CollapseNode(BalancedTree(10), c(12:13, 19))
#' plot(tree)
#' nodelabels(NodeDepth(tree, includeTips = FALSE))
#'
#'
#' @export
NodeDepth <- function(x, shortest = FALSE, includeTips = TRUE) {
  UseMethod("NodeDepth")
}

#' @export
NodeDepth.NULL <- function(x, shortest = FALSE, includeTips = TRUE) NULL

#' @export
NodeDepth.list <- function(x, shortest = FALSE, includeTips = TRUE) {
  lapply(x, NodeDepth, shortest = shortest, includeTips = includeTips)
}

#' @export
NodeDepth.multiPhylo <- NodeDepth.list

#' @importFrom ape is.rooted
#' @export
NodeDepth.phylo <- function(x, shortest = FALSE, includeTips = TRUE) {
  if (is.rooted(x)) {
    .NodeDepth.rooted(x[["edge"]], shortest, includeTips)
  } else {
    NodeDepth(x[["edge"]], shortest, includeTips)
  }
}

#' @export
NodeDepth.matrix <- function(x, shortest = FALSE, includeTips = TRUE) {
  parent <- x[, 1]
  child <- x[, 2]
  postorder <- PostorderOrder(x)
  depths <- node_depth_unrooted(parent, child, postorder, shortest)

  # Return:
  if (includeTips) depths else depths[(min(parent)):max(parent)]
}

.NodeDepth.rooted <- function(x, shortest = FALSE, includeTips = TRUE) {

  parent <- x[, 1]
  child <- x[, 2]
  minVertex <- min(parent)
  nVertex <- max(parent)

  nLeaf <- minVertex - 1L
  nNode <- nVertex - nLeaf
  leaf0s <- integer(nLeaf)
  depths <- c(leaf0s, rep.int(NA_integer_, nNode))
  uncalculated <- is.na(depths)
  Func <- if (shortest) min else max

  while(any(uncalculated)) {
    depths[uncalculated] <- vapply(which(uncalculated), function(node) {
      Func(depths[child[parent == node]])
    }, 0L) + 1L
    uncalculated <- is.na(depths)
  }


  # Return:
  if (includeTips) depths else depths[minVertex:nVertex]

}

#' Number of edges incident to each node in a tree
#'
#' `NodeOrder()` calculates the order of each node: the number of edges
#' incident to it in a tree.
#' This value includes the root edge in rooted trees.
#'
#' @template xPhylo
#' @param includeAncestor Logical specifying whether to count edge leading to
#' ancestral node in calculation of order.
#' @param internalOnly Logical specifying whether to restrict to results
#' to internal nodes, i.e. to omit leaves. Irrelevant if
#' `includeAncestor = FALSE`.
#'
#' @return `NodeOrder()` returns an integer listing the order of each node;
#' entries are named with the number of each node.
#'
#' @examples
#' tree <- CollapseNode(BalancedTree(8), 12:15)
#' NodeOrder(tree)
#' plot(tree)
#' nodelabels(NodeOrder(tree, internalOnly = TRUE))
#'
#' @template MRS
#' @family tree navigation
#' @export
NodeOrder <- function(x, includeAncestor = TRUE, internalOnly = FALSE) {
  UseMethod("NodeOrder")
}

#' @export
NodeOrder.NULL <- function(x, includeAncestor = TRUE, internalOnly = FALSE) NULL

#' @export
NodeOrder.list <- function(x, includeAncestor = TRUE, internalOnly = FALSE) {
  lapply(x, NodeOrder, includeAncestor, internalOnly)
}

#' @export
NodeOrder.multiPhylo <- NodeOrder.list

#' @export
NodeOrder.phylo <- function(x, includeAncestor = TRUE, internalOnly = FALSE) {
  NodeOrder(x[["edge"]], includeAncestor, internalOnly)
}

#' @export
NodeOrder.matrix <- function(x, includeAncestor = TRUE, internalOnly = FALSE) {
  if (includeAncestor) {
    if (internalOnly) {
      tabulate(x)[-seq_len(min(x[, 1]) - 1L)]
    } else {
      tabulate(x)
    }
  } else {
    tabulate(x[, 1])[-seq_len(min(x[, 1]) - 1L)]
  }
}

#' Distance between edges
#'
#' Number of nodes that must be traversed to navigate from each edge to
#' each other edge within a tree
#'
#' @template treeParam
#'
#' @return `EdgeDistances()` returns a symmetrical matrix listing the number
#' of edges that must be traversed to travel from each numbered edge to each
#' other.
#' The two edges straddling the root of a rooted tree
#' are treated as a single edge.  Add a "root" tip using [`AddTip()`] if the
#' position of the root is significant.
#'
#' @examples
#'
#' tree <- BalancedTree(5)
#' plot(tree)
#' ape::edgelabels()
#'
#' EdgeDistances(tree)
#'
#' @family tree navigation
#' @importFrom fastmatch %fin%
#' @template MRS
#' @export
EdgeDistances <- function(tree) {
  edge <- tree[["edge"]]
  nEdge <- dim(edge)[1]
  edge <- RenumberTree(edge[, 1], edge[, 2])
  parent <- edge[, 1]
  child <- edge[, 2]
  ancs <- AllAncestors(parent, child)
  ret <- matrix(0L, ncol = nEdge, nrow = nEdge)
  for (i in seq_len(nEdge - 1L)) {
    ancI <- ancs[[child[i]]]
    nAncI <- length(ancI)
    for (j in seq.int(from = i + 1L, to = nEdge)) {
      ancJ <- ancs[[child[j]]]
      intersection <- intersect(ancI, ancJ)
      if (length(intersection) > 1L) {
        # On same side of root
        mrca <- max(intersection)
        if (child[i] %fin% ancJ || child[j] %fin% ancI) {
          addOne <- 0L
        } else {
          addOne <- 1L
        }
      } else {
        # On opposite sides of root
        mrca <- intersection
        addOne <- 0L
      }
      u <- union(ancI, ancJ)
      ret[i, j] <- sum(u > mrca, addOne)
    }
  }
  ret[lower.tri(ret)] <- t(ret)[lower.tri(ret)]

  origOrder <- match(tree[["edge"]][, 2], edge[, 2])

  # Return:
  ret[origOrder, origOrder]
}

#' Number of leaves in a phylogenetic tree
#'
#' `NTip()` extends [`ape::Ntip()`][ape::summary.phylo] to handle
#' objects of class `Splits` and `list`, and edge matrices
#' (equivalent to `tree$edge`).
#'
#' @param phy Object representing one or more phylogenetic trees.
#'
#' @return `NTip()` returns an integer specifying the number of tips in each
#' object in `phy`.
#'
#' @family tree properties
#' @export
NTip <- function(phy) UseMethod("NTip")

#' @rdname NTip
#' @export
NTip.default <- function(phy) attr(phy, "nTip")

#' @rdname NTip
#' @family Splits operations
#' @export
NTip.Splits <- NTip.default

#' @rdname NTip
#' @export
NTip.list <- function(phy) {
  vapply(phy, NTip, integer(1))
}

#' @rdname NTip
#' @export
NTip.phylo <- function(phy) {
  length(phy[["tip.label"]])
}

#' @rdname NTip
#' @export
NTip.multiPhylo <- function(phy) {
  vapply(phy, NTip.phylo, integer(1))
}

#' @rdname NTip
#' @export
NTip.phyDat <- function(phy) {
  length(phy)
}

#' @rdname NTip
#' @export
NTip.matrix <- function(phy) {
  if (is.numeric(phy)) {
    parent <- phy[, 1]
    child <- phy[, 2]

    # Return:
    max(child[!child %fin% parent])
  } else {
    NextMethod()
  }
}

#' Number of distinct splits
#'
#' `NSplits()` counts the unique bipartition splits in a tree or object.
#'
#' @param x A phylogenetic tree of class `phylo`; a list of such trees
#' (of class `list` or `multiPhylo`); a `Splits` object;
#' a vector of integers; or a character vector listing tips of a tree,
#' or a character of length one specifying a tree in Newick format.
#'
#' @return `NSplits()` returns an integer specifying the number of bipartitions in
#' the specified objects, or in a binary tree with `x` tips.
#'
#' @examples
#' NSplits(8L)
#' NSplits(PectinateTree(8))
#' NSplits(as.Splits(BalancedTree(8)))
#' @template MRS
#'
#' @family tree properties
#' @family Splits operations
#' @export
NSplits <- function(x) UseMethod("NSplits")

#' @rdname NSplits
#' @export
NPartitions <- NSplits

#' @rdname NSplits
#' @export
NSplits.phylo <- function(x) {
  # Use optimized C++ function that replaces collapse.singles bottleneck
  cpp_count_splits(x[["edge"]], length(x[["tip.label"]]))
}

#' @rdname NSplits
#' @export
NSplits.list <- function(x) vapply(x, NSplits, numeric(1L))

#' @rdname NSplits
#' @export
NSplits.multiPhylo <- NSplits.list

#' @rdname NSplits
#' @export
NSplits.Splits <- function(x) nrow(x)

#' @rdname NSplits
#' @export
NSplits.numeric <- function(x) pmax(0L, x - 3L)

#' @rdname NSplits
#' @export
NSplits.NULL <- function(x) NULL

#' @rdname NSplits
#' @export
NSplits.ClusterTable <- function(x) {
  # Root + Ingroup + All-leaves
  nrow(as.matrix(x)) - 3L
}

#' @rdname NSplits
#' @importFrom ape read.tree
#' @export
NSplits.character <- function(x) {
  lengthX <- length(x)
  if (lengthX == 1) {
    tree <- read.tree(text = x)
    if (!is.null(tree)) return(NSplits(tree))
  }
  NSplits(lengthX)
}

#' Maximum splits in an _n_-leaf tree
#'
#' `SplitsInBinaryTree()` is a convenience function to calculate the number of
#' splits in a fully-resolved (binary) tree with _n_ leaves.
#'
#' @param tree An object of a supported format that represents a tree or
#' set of trees, from which the number of leaves will be calculated.
#'
#' @return `SplitsInBinaryTree()` returns an integer vector detailing the number
#' of unique non-trivial splits in a binary tree with _n_ leaves.
#'
#' @examples
#' tree <- BalancedTree(8)
#' SplitsInBinaryTree(tree)
#' SplitsInBinaryTree(as.Splits(tree))
#' SplitsInBinaryTree(8)
#' SplitsInBinaryTree(list(tree, tree))
#' @template MRS
#'
#' @family tree properties
#' @family Splits operations
#' @export
SplitsInBinaryTree <- function(tree) UseMethod("SplitsInBinaryTree")

#' @rdname SplitsInBinaryTree
#' @export
SplitsInBinaryTree.list <- function(tree) vapply(tree, SplitsInBinaryTree, 0L)

#' @rdname SplitsInBinaryTree
#' @export
SplitsInBinaryTree.multiPhylo <- SplitsInBinaryTree.list

#' @rdname SplitsInBinaryTree
#' @export
SplitsInBinaryTree.numeric <- function(tree) as.integer(tree) - 3L

#' @rdname SplitsInBinaryTree
#' @export
SplitsInBinaryTree.NULL <- function(tree) NULL

#' @rdname SplitsInBinaryTree
#' @export
SplitsInBinaryTree.default <- function(tree) NTip(tree) - 3L

#' @rdname SplitsInBinaryTree
#' @export
SplitsInBinaryTree.Splits <- SplitsInBinaryTree.default

#' @rdname SplitsInBinaryTree
#' @export
SplitsInBinaryTree.phylo <- SplitsInBinaryTree.default

#' Is tree rooted?
#'
#' `TreeIsRooted()` is a fast alternative to `ape::is.rooted()`.
#'
#' @param tree A phylogenetic tree of class `phylo`.
#' @return `TreeIsRooted()` returns a logical specifying whether a root node is
#' resolved.
#'
#' @examples
#' TreeIsRooted(BalancedTree(6))
#' TreeIsRooted(UnrootTree(BalancedTree(6)))
#'
#' @family tree properties
#'
#' @template MRS
#' @export
TreeIsRooted <- function(tree) {
  edge <- tree[["edge"]]
  parent <- edge[, 1]
  sum(parent == min(parent)) < 3L
}

#' Which node is a tree's root?
#'
#' `RootNode()` identifies the root node of a (rooted or unrooted) phylogenetic
#' tree.
#' Unrooted trees are represented internally by a rooted tree with a polytomy
#' at the root.
#'
#' @param x A tree of class `phylo`, or its edge matrix; or a list or
#' `multiPhylo` object containing multiple trees.
#' @return `RootNode()` returns an integer denoting the root node for each tree.
#' Badly conformed trees trigger an error.
#' @template MRS
#'
#' @examples
#' RootNode(BalancedTree(8))
#' RootNode(UnrootTree(BalancedTree(8)))
#'
#'
#' @family tree navigation
#' @seealso
#'
#' Test whether a tree is rooted: [`TreeIsRooted()`]
#'
#' `phangorn::getRoot()`
#'
#' @export
RootNode <- function(x) UseMethod("RootNode")

#' @export
RootNode.NULL <- function(x) NULL

#' @export
RootNode.phylo <- function(x) {
  edge <- x[["edge"]]
  edgeOrder <- attr(x, "order")
  if (!is.null(edgeOrder)) {
    if (edgeOrder == "postorder") {
      return(as.integer(edge[nrow(edge), 1L]))
    } else if (edgeOrder == "preorder") {
      return(as.integer(edge[1L]))
    }
  }
  RootNode(edge)
}

#' @export
RootNode.list <- function(x) {
  vapply(x, RootNode, 0L)
}

#' @export
RootNode.multiPhylo <- RootNode.list

#' @export
RootNode.numeric <- function(x) {
  parent <- x[, 1]
  child <- x[, 2]
  ret <- unique(parent[!parent %fin% child])
  if (length(ret) != 1) {
    warning("Root not unique: found ", paste(ret, collapse = ", "))
  }

  # Return:
  as.integer(ret)
}

nasty <- structure(list(edge = structure(
  c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
    5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
  .Dim = c(12, 2)),
  Nnode = 5L,
  tip.label = letters[1:8]),
  class = "phylo") # Danger: Do not plot!
test_that("NodeNumbers() works", {
  expect_equal(NodeNumbers(BalancedTree(5)), 6:9)
  expect_equal(NodeNumbers(StarTree(5)), 6L)
  expect_equal(NodeNumbers(BalancedTree(5), TRUE), 1:9)
  expect_equal(NodeNumbers(StarTree(5), TRUE), 1:6)
})

test_that("Root node can be found", {
  rooted <- BalancedTree(8)
  postorder <- Postorder(rooted)
  unrooted <- UnrootTree(rooted)
  expect_true(TreeIsRooted(rooted))
  expect_false(TreeIsRooted(unrooted))
  expect_equal(RootNode(rooted$edge), 9L)
  expect_equal(RootNode(rooted), 9L)
  expect_equal(RootNode(as.phylo(1337L, 8L)), 9L)
  expect_equal(RootNode(c(unrooted, Preorder(postorder), postorder)),
               rep(9L, 3L))
  expect_equal(RootNode(list(Cladewise(postorder),
                             ApePostorder(rooted),
                             Pruningwise(postorder))),
               rep(9L, 3L))
  expect_warning(RootNode(matrix(1:4, 2)),
                 "Root not unique")
})

test_that("NodeOrder() works", {
  expect_equal(ignore_attr = TRUE, c(2L, rep(3L, 6)),
               as.integer(NodeOrder(BalancedTree(8), internalOnly = TRUE)))
  expect_equal(ignore_attr = TRUE, c(rep(1L, 8), 2L, rep(3L, 6)),
               as.integer(NodeOrder(BalancedTree(8), internalOnly = FALSE)))
  expect_equal(ignore_attr = TRUE, rep(2L, 7),
               as.integer(NodeOrder(BalancedTree(8), FALSE)))
  tree <- CollapseNode(BalancedTree(8), 12:15)
  expect_equal(5:3, as.integer(NodeOrder(tree$edge, internalOnly = TRUE)))
  expect_equal(list(NodeOrder(tree), NodeOrder(tree)), NodeOrder(c(tree, tree)))
  expect_equal(NodeOrder(c(tree, tree)), NodeOrder(list(tree, tree)))

  expect_equal(c(3, 2, 2), NDescendants(UnrootTree(PectinateTree(5))))
})

test_that("Rooting and partition counting", {
  set.seed(0)

  expect_false(TreeIsRooted(ape::read.tree(text="(a, b, c);")))
  expect_true(TreeIsRooted(ape::read.tree(text="(a, (b, c));")))

  tree8 <- RandomTree(8L, TRUE)
  expect_true(TreeIsRooted(tree8))
  expect_false(TreeIsRooted(UnrootTree(tree8)))

  expect_equal(NPartitions(8), 5L)
  expect_equal(NPartitions(tree8), 5L)
  expect_equal(NPartitions(UnrootTree(tree8)), 5L)

  expect_equal(NPartitions(ape::read.tree(text="(A, ((B, C)));")), 0L)

  expect_equal(NPartitions(list(tree8, UnrootTree(tree8))), c(5L, 5L))
  expect_equal(NPartitions(c(8, 8)), c(5L, 5L))
  expect_equal(NSplits("((a, b), (c, (d, e)));"), 2L)
  
  emptyTree <- structure(
    list(edge = structure(numeric(0), dim = c(0L, 2L)),
         tip.label = character(0), Nnode = 0),
    class = "phylo")
  expect_equal(NSplits(emptyTree), 0L)
  expect_null(NSplits(NULL))
  expect_equal(NSplits(letters[1:2]), 0L)
  expect_equal(NSplits(letters[1:6]), 3L)
  expect_error(NPartitions(raw(1)),
               "no applicable method")
  skip_if_not_installed("ape", "5.6-3")
  expect_equal(NSplits("a;"), 0L)
})

test_that("NTip() works", {
  Test <- function(n) {
    tr <- BalancedTree(n)
    pec <- PectinateTree(n)
    expect_identical(NTip(tr), n)
    expect_identical(NTip(tr$edge), n)
    expect_identical(NTip(Postorder(tr$edge)), n)
    expect_identical(NTip(list(tr)), n)
    expect_identical(rep(n, 2L), NTip(list(tr, tr)))
    expect_identical(rep(n, 3L),
                     NTip(structure(list(tr, tr, pec), class = "multiPhylo")))
    expect_null(NTip(matrix("", n, 2)))
  }
  Test(1L)
  Test(8L)
  Test(64L)
  Test(2000L)
  expect_identical(48L, NTip(Lobo.phy))
})

test_that("NSplits() works", {
  expect_equal(NSplits(5L), NSplits(LETTERS[1:5]))
  expect_equal(NSplits(as.ClusterTable(BalancedTree(6))),
               NSplits(BalancedTree(6)))
})

test_that("Edge distances are calculated correctly", {
  tree <- ape::read.tree(text = "(((a, b), (c, d)), (e, (f, (g, (h, i)))));")
  ed <- EdgeDistances(tree)

  expect_equal(ed, t(ed)) # Symmetry
  expect_equal(c(4, 5, 6, 6,
                 5, 6, 6, 4,
                 4, 3, 3, 2,
                 2, 1, 1, 0),
               ed[, 16])
  expect_equal(c(1, 2, 3, 3, 2, 3, 3, 1, 1, 0, 1, 1, 2, 2, 3, 3), ed[, 10])
  expect_equal(c(1, 0, 1, 1, 1, 2, 2, 1, 2, 2, 3, 3, 4, 4, 5, 5), ed[, 2])

  tree <- ape::read.tree(text="(a, (b, (c, (d, (((e, f), g), (h, (i, j)))))));")
  ed <- EdgeDistances(tree)
  expect_equal(ed, t(ed)) # Symmetry
  expect_equal(c(6, 6, 6, 5, 5, 4, 4, 3, 2, 1, 0, 1, 2, 3, 4, 4, 5, 5),
               ed[11, ])
})

test_that("Node depths calculated correctly", {
  #par(mar=rep(0.4, 4))
  expect_equal(c(rep(0, 20), 19:1), NodeDepth(PectinateTree(20)))
  expect_equal(c(rep(0, 20), rep(1, 19)),
               NodeDepth(PectinateTree(20), shortest = TRUE))
  expect_equal(c(rep(0, 20), 1:9, 9:1), NodeDepth(UnrootTree(PectinateTree(20))))
  expect_equal(c(rep(0, 20), rep(1, 18)),
               NodeDepth(UnrootTree(PectinateTree(20)), shortest = TRUE))

  tree <- BalancedTree(20)
  expect_equal(c(5,4,3,2,1,1,3,2,1,1,4,3,2,1,1,3,2,1,1),
               NodeDepth(tree, includeTips = FALSE))
  expect_equal(c(4,3,2,1,1,1,2,1,1,1,3,2,1,1,1,2,1,1,1),
               NodeDepth(tree, shortest = TRUE, includeTips = FALSE))
  expect_equal(list(c(4,3,2,1,1,1,2,1,1,1,3,2,1,1,1,2,1,1,1)),
               NodeDepth(list(tree), shortest = TRUE, includeTips = FALSE))

  tree <- UnrootTree(tree)
  expect_equal(c(4,3,2,1,1,3,2,1,1,4,3,2,1,1,3,2,1,1),
               NodeDepth(tree, includeTips = FALSE))
  expect_equal(c(3,2,1,1,1,2,1,1,1,3,2,1,1,1,2,1,1,1),
               NodeDepth(tree, shortest = TRUE, includeTips = FALSE))

  tree <- UnrootTree(RootTree(BalancedTree(20), "t10"))
  #plot(tree); nodelabels(adj = 2, bg = "yellow")
  #nodelabels(NodeDepth(tree, F, FALSE))
  #nodelabels(NodeDepth(tree, T, FALSE))
  expect_equal(c(1,3,4,3,2,1,1,4,3,2,1,1,3,2,1,1,2,1),
               NodeDepth(tree, includeTips = FALSE))
  expect_equal(c(1,2,3,2,1,1,1,3,2,1,1,1,2,1,1,1,1,1),
               NodeDepth(tree, shortest = TRUE, includeTips = FALSE))



  tree <- CollapseNode(BalancedTree(20), c(22:26, 33:35))
  expect_equal(c(4,3,2,1,1,4,1,3,2,1,1), NodeDepth(tree, FALSE, FALSE))
  expect_equal(c(1,2,1,1,1,2,1,2,1,1,1), NodeDepth(tree, TRUE, FALSE))

  tree <- CollapseNode(BalancedTree(20), c(22, 33:35))
  expect_equal(c(4,3,2,1,1,3,2,1,1,4,1,3,2,1,1),
               NodeDepth(tree, FALSE, FALSE))
  expect_equal(c(3,2,1,1,1,2,1,1,1,2,1,2,1,1,1),
               NodeDepth(tree, TRUE, FALSE))

  expect_equal(1L, NodeDepth(StarTree(8), FALSE, FALSE))
  expect_equal(1L, NodeDepth(StarTree(8), TRUE, FALSE))
})

test_that("SplitsInBinaryTree() works", {
  expect_identical(5L, SplitsInBinaryTree(8))
  expect_identical(5L, SplitsInBinaryTree(as.phylo(0, 8)))
  expect_identical(rep(5L, 4), SplitsInBinaryTree(as.phylo(0:3, 8)))
  expect_null(SplitsInBinaryTree(NULL))
})

# Manual exports are required for functions that are defined in the 
# TreeTools namespace in header files.
keep_tip <- function(edge, keep) {
  .Call(`_TreeTools_keep_tip`, edge, keep)
}

postorder_order <- function(edge) {
  .Call(`_TreeTools_postorder_order`, edge)
}

root_binary <- function(edge, outgroup) {
  .Call(`_TreeTools_root_binary`, edge, outgroup)
}

#' Wrapper for internal C function `root_on_node()`
#' 
#' Direct entry point to `root_on_node()`; recommended for expert use only.
#' `RootTree()` checks that input is properly formatted and is recommended
#' for general use.
#' @param phy Minimally, a named list with entries `edge` and `Nnode`, in the
#' format of equivalent entries in a tree of class `phylo`. `edge.length` will
#' also be considered if supplied.
#' @param outgroup Integer specifying index of leaf or node to set as the
#' outgroup.
#' @returns `root_on_node()` returns `phy` rooted on the specified node.
#' 
#' @template MRS
#' @export
#' @keywords internal
root_on_node <- function(phy, outgroup) {
  .Call(`_TreeTools_root_on_node`, phy, outgroup)
}

#' Check for mismatch between character vectors
#' 
#' Checks that entries in one character vector occur in another, suggesting
#' corrections for mismatched elements.
#' 
#' @param x,table Character vectors, in which all elements of `x` are expected
#' to occur in `table`.
#' @param Fail Function to call if a mismatch is found.
#' @param max.distance,\dots Arguments to \code{\link[base:agrep]{agrep()}},
#' used to propose possible matches to the user.
#' @return `MatchStrings()` returns the elements of `x` that occur in `table`.
#' @examples
#' tree <- BalancedTree(8)
#' MatchStrings(c("t1", "tip2", "t3"), TipLabels(tree), Fail = message)
#' @template MRS
#' @family string parsing functions
#' @export
MatchStrings <- function(x, table, Fail = stop, max.distance = 0.5, ...) {
  matches <- match(x, table)
  missing <- is.na(matches)
  if (any(missing)) {
    nearMiss <- unlist(lapply(x[missing], agrep, table,
                              max.distance = max.distance, ...),
                       use.names = FALSE, recursive = FALSE)
    message <- paste0("Could not find '", paste(x[missing], collapse = "', '"), 
                      "' in ", deparse(substitute(table)), ".  ",
                      if (length(nearMiss)) {
                        paste0("Did you mean '", 
                               paste(unique(table[nearMiss]), collapse = "', '"), "'?")
                      })
    Fail(message)
  }
  table[matches[!missing]]
}

test_that("MatchStrings()", {
  list1 <- "Mouse"
  list2 <-c("cat", "mouse", "dog")
  expect_error(MatchStrings("maus", list1, stop),
               "Could not find 'maus' in list1.  Did you mean 'Mouse'?")
  expect_warning(
    expect_equal(
      MatchStrings(c("cat", "mouse", "dawg", "maus", "non-existent taxon"),
                   list2, warning),
      c("cat", "mouse")),
    "Could not find 'dawg', 'maus', 'no.+on' in list2.+ mean 'dog', 'mouse'\\?"
  )
})

#' Subset of a split on fewer leaves
#'
#' `Subsplit()` removes leaves from a `Splits` object.
#'
#' @template splitsObjectParam
#' @param tips A vector specifying a subset of the leaf labels applied to `split`.
#' @param keepAll logical specifying whether to keep entries that define trivial
#' splits (i.e. splits of zero or one leaf) on the subset of leaves.
#' @param unique logical specifying whether to remove duplicate splits.
#'
#' @return `Subsplit()` returns an object of class `Splits`, defined on the
#' leaves `tips`.
#'
#' @examples
#' splits <- as.Splits(PectinateTree(letters[1:9]))
#' splits
#' efgh <- Subsplit(splits, tips = letters[5:8], keepAll = TRUE)
#' summary(efgh)
#'
#' TrivialSplits(efgh)
#'
#' summary(Subsplit(splits, tips = letters[5:8], keepAll = FALSE))
#' @template MRS
#' 
#' @seealso [`KeepTip()`] is a less flexible but faster equivalent.
#'
#' @family split manipulation functions
#' @export
Subsplit <- function(splits, tips, keepAll = FALSE, unique = TRUE) {
  if (is.list(splits)) {
    lapply(splits, Subsplit, tips = tips, keepAll = keepAll, unique = unique)
  } else if (length(splits) == 0) {
    ret <- splits
    attr(ret, "nTip") <- length(tips)
    attr(ret, "tip.label") <- if (is.character(tips)) {
      tips
    } else {
      attr(splits, "tip.label")[tips]
    }
    ret
  } else {
    allSplits <- as.Splits(as.logical(splits)[, tips])
    ret <- if (keepAll) allSplits else WithoutTrivialSplits(allSplits)

    # Return:
    if (unique) unique(ret) else ret
  }
}

#' Identify and remove trivial splits
#'
#' `TrivialSplits()` identifies trivial splits (which separate one or zero
#' leaves from all others); `WithoutTrivialSplits()` removes them from a
#' `Splits` object.
#'
#' @template splitsObjectParam
#' @template nTipParam
#'
#' @return `TrivialSplits()` returns a logical vector specifying whether each
#' split in `splits` is trivial, i.e. includes or excludes only a single tip or
#' no tips at all.
#'
#' @template MRS
#' @family split manipulation functions
#' @examples
#' splits <- as.Splits(PectinateTree(letters[1:9]))
#' efgh <- Subsplit(splits, tips = letters[5:8], keepAll = TRUE)
#' summary(efgh)
#'
#' TrivialSplits(efgh)
#' @export
TrivialSplits <- function(splits, nTip = attr(splits, "nTip")) {
  inSplit <- TipsInSplits(splits)
  inSplit < 2L | inSplit > nTip - 2L
}

#' @rdname TrivialSplits
#' @return `WithoutTrivialSplits()` returns a `Splits` object with trivial
#' splits removed.
#' @examples
#' summary(WithoutTrivialSplits(efgh))
#' @export
WithoutTrivialSplits <- function(splits, nTip = attr(splits, "nTip")) {
  splits[[!TrivialSplits(splits, nTip)]]
}

#' Which splits are compatible?
#'
#' @template splitsObjectParam
#' @param splits2 A second `Splits` object.
#'
#' @return `CompatibleSplits` returns a logical matrix specifying whether each
#' split in `splits` is compatible with each split in `splits2`.
#'
#' @examples
#' splits <- as.Splits(BalancedTree(8))
#' splits2 <- as.Splits(PectinateTree(8))
#'
#' summary(splits)
#' summary(splits2)
#'
#' CompatibleSplits(splits, splits2)
#'
#' @template MRS
#' @export
CompatibleSplits <- function(splits, splits2) {
  splits <- as.Splits(splits)
  nTip <- attr(splits, "nTip")
  splits2 <- as.Splits(splits2, splits)
  apply(splits2, 1, function(split)
    apply(splits, 1, .CompatibleSplit, split, nTip))
}

#' @param a,b [Raw][raw] representations of splits, from a row of a `Splits`
#' object.
#' @return `.CompatibleSplit` returns a logical vector stating whether splits
#' are compatible.
#' @rdname CompatibleSplits
#' @keywords internal
#' @export
.CompatibleSplit <- function(a, b, nTip) {
  rawMask <- if (nTip %% 8L) {
    as.raw(c(2 ^ (nTip %% 8L) - 1L, rep.int(255L, nTip %/% 8)))
  } else {
    rep.int(as.raw(255), nTip %/% 8L)
  }
  .CompatibleRaws(a, b, rawMask)
}

#' @rdname CompatibleSplits
#' @param rawA,rawB Raw representations of splits.
#' @param bitmask Raw masking bits that do not correspond to tips.
#'
#' @return `.CompatibleRaws` returns a logical vector specifying whether input
#' raws are compatible.
#' @keywords internal
#' @export
.CompatibleRaws <- function(rawA, rawB, bitmask) {
  !any(as.logical(rawA & rawB)) ||
  !any(as.logical(rawA & !rawB)) ||
  !any(as.logical(!rawA & rawB)) ||
  !any(as.logical(!rawA & !rawB & bitmask))
}

#' Probability of matching this well
#'
#' (`Ln`)`SplitMatchProbability()`calculates the probability that two random
#' splits of the sizes provided will be at least as similar as the two
#' specified.
#'
#' @template split12Params
#'
#' @return `SplitMatchProbability()` returns a numeric giving the proportion
#' of permissible non-trivial splits that divide the terminals into bipartitions
#' of the sizes given, that match as well as `split1` and `split2` do.
#'
#' @examples
#' split1 <- as.Splits(c(rep(TRUE, 4), rep(FALSE, 4)))
#' split2 <- as.Splits(c(rep(TRUE, 3), rep(FALSE, 5)))
#' SplitMatchProbability(split1, split2)
#' @family split information functions
#' @template MRS
#' @export
SplitMatchProbability <- function(split1, split2) {

  if (NTip(split1) != NTip(split2)) stop("Splits pertain to different tips")

  split1 <- as.logical(split1)
  split2 <- as.logical(as.Splits(split2, split1))
  partitions <- c(sum(split1 & split2), sum(split1 & !split2),
                  sum(!split1 & split2), sum(!split1 & !split2))
  #, dimnames=list(c("A1", "B1"), c("A2", "B2")))

  split1Size <- .rowSums(partitions, 2, 2)
  split2Size <- .colSums(partitions, 2, 2)
  A1 <- split1Size[1]
  A2 <- split2Size[1]
  B2 <- split2Size[2]
  n <- sum(partitions)

  minA1B2 <- max(0, A1 - A2)
  iMax <- min(A1, B2)
  minB1A2 <- B2 - iMax

  nArrangements <- iMax - minA1B2 + 1L
  # It turns out that this is called a confusion matrix, association matrix
  # or contingency table; see Meila 2007
  arrangements <- vapply(minA1B2:iMax,
                         function(i) c(A1 - i, i, i + A2 - A1, B2 - i),
                         double(4))

  #H <- function(p) -sum(p[p > 0] * log(p[p > 0]))
  #jointEntropies <- apply(arrangements / n, 2, H)
  #mutualInformation <- H(c(A1,n - A1) / n) + H(c(A2, B2) / n) - jointEntropies

  extraTipsInPerfectMatch.A1A2.B1B2 <- sum(arrangements[c(1, 4), nArrangements])
  extraTipsInPerfectMatch.A1B2.B1A2 <- sum(arrangements[c(2, 3), 1L])

  ranking <- if (extraTipsInPerfectMatch.A1A2.B1B2 >
                 extraTipsInPerfectMatch.A1B2.B1A2) {
    c(seq.int(1L, nArrangements, 2L),
      rev(seq.int(2L, nArrangements, 2L)))
  } else if (extraTipsInPerfectMatch.A1A2.B1B2 <
             extraTipsInPerfectMatch.A1B2.B1A2) {
    c(seq.int(2L, nArrangements, 2L),
      rev(seq.int(1L, nArrangements, 2L)))
  } else {
    c(seq.int(1L, nArrangements, 2L),
      rev(seq.int(2L, nArrangements, 2L)) - 1)
  }

  choices <- apply(arrangements, 2, NPartitionPairs)

  # Return:
  sum(choices[ranking <= ranking[partitions[3] + 1L - minA1B2]]) / choose(n, A1)
}

#' @rdname SplitMatchProbability
#' @return `LnSplitMatchProbability()` returns the natural logarithm of the
#' probability.
#' @examples
#' LnSplitMatchProbability(split1, split2)
#' @export
LnSplitMatchProbability <- function(split1, split2) {
  log(SplitMatchProbability(split1, split2))
}


#' Extract tip labels
#'
#' `TipLabels()` extracts labels from an object: for example, names of taxa in
#' a phylogenetic tree or data matrix.  `AllTipLabels()` extracts all labels,
#' where entries of a list of trees may pertain to different taxa.
#'
#' @param x An object of a supported class (see Usage section above).
#' @param single Logical specifying whether to report the labels for the first
#' object only (`TRUE`), or for each object in a list (`FALSE`).
#'
#' @return `TipLabels()` returns a character vector listing the tip labels
#' appropriate to `x`. If `x` is a single integer, this will be a vector
#' `t1`, `t2` ... `tx`, to match the default of \code{\link[ape]{rtree}()}.
#'
#' @examples
#' TipLabels(BalancedTree(letters[5:1]))
#' TipLabels(5)
#'
#' data("Lobo")
#' head(TipLabels(Lobo.phy))
#'
#' AllTipLabels(c(BalancedTree(4), PectinateTree(8)))
#'
#' @family tree properties
#' @template MRS
#' @export
TipLabels <- function(x, single = TRUE) UseMethod("TipLabels")

#' @rdname TipLabels
#' @export
TipLabels.default <- function(x, single = TRUE) {
  tla <- attr(x, "tip.label")
  # TODO when require R4.1, tla %||% names(x) %||% x
  if (length(tla) > 0) {
    tla
  } else {
    nom <- names(x)
    if (length(nom) == 0) {
      x
    } else {
      nom
    }
  }
}

#' @rdname TipLabels
#' @export
TipLabels.matrix <- function(x, single = TRUE) colnames(x)

#' @rdname TipLabels
#' @export
TipLabels.logical <- function(x, single = TRUE) TipLabels.numeric(length(x))

#' @rdname TipLabels
#' @export
TipLabels.phylo <- function(x, single = TRUE) x[["tip.label"]]

#' @rdname TipLabels
#' @export
TipLabels.phyDat <- function(x, single = TRUE) names(x)

#' @rdname TipLabels
#' @export
TipLabels.MixedBase <- TipLabels.default

#' @rdname TipLabels
#' @export
TipLabels.TreeNumber <- TipLabels.default

#' @rdname TipLabels
#' @family Splits operations
#' @export
TipLabels.Splits <- TipLabels.default

#' @rdname TipLabels
#' @export
TipLabels.list <- function(x, single = FALSE) {
  if (!is.null(attr(x, "tip.label"))) return(attr(x, "tip.label"))
  xTipLabel <- x[["tip.label"]]
  if (!is.null(xTipLabel)) {
    if (is.list(xTipLabel) && !is.null(xTipLabel[["tip.label"]])) {
      return(xTipLabel[["tip.label"]])
    } else {
      return(xTipLabel)
    }
  }
  .ListLabels(x, single, TipLabels)
}

#' @rdname TipLabels
#' @export
TipLabels.multiPhylo <- function(x, single = FALSE) {
  xTipLabel <- x[["tip.label"]]
  if (!is.null(xTipLabel)) {
    if (is.list(xTipLabel) && !is.null(xTipLabel[["tip.label"]])) {
      return(xTipLabel[["tip.label"]])
    } else {
      return(xTipLabel)
    }
  }
  if (single) {
    firstEntry <- x[[1]]
    if (!is.null(firstEntry[["tip.label"]])) {
      return(firstEntry[["tip.label"]])
    }
  } else {
    .ListLabels(x, single, TipLabels.phylo)
  }
}

#' @rdname TipLabels
#' @export
TipLabels.character <- function(x, single = TRUE) {
  if (is.null(attr(x, "tip.label"))) {
    NextMethod("TipLabels", as.character(x))
  } else {
    attr(x, "tip.label")
  }
}

#' @rdname TipLabels
#' @export
TipLabels.numeric <- function(x, single = TRUE) {
  if (length(x) == 1L) {
    if (x < 0) {
      stop("`x` may not be negative")
    }
    paste0(rep_len("t", x), seq_len(x))
  } else {
    NextMethod("TipLabels", as.character(x))
  }
}

#' @rdname TipLabels
#' @export
TipLabels.phyDat <- function(x, single = TRUE) names(x)

#' @rdname TipLabels
#' @export
AllTipLabels <- function(x) UseMethod("AllTipLabels")

#' @export
AllTipLabels.NULL <- function(x) NULL

#' @rdname TipLabels
#' @export
AllTipLabels.list <- function(x) {
  unique(unlist(lapply(x, TipLabels)))
}

#' @rdname TipLabels
#' @export
AllTipLabels.multiPhylo <- AllTipLabels.list

#' @rdname TipLabels
#' @export
AllTipLabels.phylo <- function(x) TipLabels.phylo(x)

#' @rdname TipLabels
#' @export
AllTipLabels.Splits <- function(x) TipLabels.Splits(x)

#' @rdname TipLabels
#' @export
AllTipLabels.TreeNumber <- function(x) TipLabels.TreeNumber(x)

#' @rdname TipLabels
#' @export
AllTipLabels.matrix <- function(x) TipLabels.matrix(x)

#' @keywords internal
.ListLabels <- function(x, single, Func) {
  if (length(x)) {
    if (single) {
      Func(x[[1]])
    } else {
      ret <- lapply(x, Func)
      uniqueRet <- unique(ret)
      if (length(uniqueRet) == 1) uniqueRet[[1]] else ret
    }
  } else {
    # else Return:
    NULL
  }
}

#' Distributions of tips consistent with a partition pair
#'
#' `NPartitionPairs()` calculates the number of terminal arrangements matching
#' a specified configuration of two splits.
#'
#' Consider splits that divide eight terminals, labelled A to H.
#'
#' \tabular{rcll}{
#'   Bipartition 1:\tab ABCD:EFGH\tab A1 = ABCD\tab B1 = EFGH \cr
#'   Bipartition 2:\tab ABE:CDFGH\tab A2 = ABE\tab B2 = CDFGH
#' }
#'
#' This can be represented by an association matrix:
#'
#' \tabular{rll}{
#'      \tab *A2* \tab *B2* \cr
#' *A1* \tab AB  \tab C   \cr
#' *B1* \tab E   \tab FGH
#' }
#'
#' The cells in this matrix contain 2, 1, 1 and 3 terminals respectively; this
#' four-element vector (`c(2, 1, 1, 3)`) is the `configuration` implied by
#' this pair of bipartition splits.
#'
#' @param configuration Integer vector of length four specifying the number of
#' terminals that occur in both
#'   (1) splits A1 and A2;
#'   (2) splits A1 and B2;
#'   (3) splits B1 and A2;
#'   (4) splits B1 and B2.
#'
#' @return The number of ways to distribute `sum(configuration)` taxa according
#'  to the specified pattern.
#'
#' @examples
#' NPartitionPairs(c(2, 1, 1, 3))
#' @template MRS
#' @export
NPartitionPairs <- function(configuration) {
  choose(sum(configuration[c(1, 3)]), configuration[1]) *
    choose(sum(configuration[c(2, 4)]), configuration[2])
}

test_that("Subsplits", {
  splits <- as.Splits(PectinateTree(letters[1:9]))
  efgh <- Subsplit(splits, tips = letters[5:8], keepAll = TRUE, unique = FALSE)
  expect_equal(TipsInSplits(efgh),
               setNames(c(4, 4, 4, 3, 2, 1), 12:17))
  expect_equal(DropTip(splits, letters[c(1:4, 9)]),
               KeepTip(splits, letters[5:8]))
  expect_equal(KeepTip(splits, letters[5:8]), Subsplit(splits, letters[5:8]))
  expect_equal(c("12" = TRUE, "13" = TRUE, "14" = TRUE, "15" = TRUE,
                 "16" = FALSE, "17" = TRUE), TrivialSplits(efgh))

  efghF <- Subsplit(splits, tips = letters[5:8], keepAll = FALSE)
  expect_equal(c("16" = 2), TipsInSplits(efghF))
  expect_equal(efghF, Subsplit(list(splits, splits), tips = letters[5:8],
                               keepAll = FALSE)[[1]])

  noSplit <- Subsplit(splits - splits, letters[5:8])
  expect_equal(attributes(noSplit)[3:5], attributes(efghF)[3:5])
  expect_equal(ignore_attr = TRUE, raw(1), noSplit[1])
  expect_equal(noSplit[1], Subsplit(splits - splits, 5:8)[1])


  splits <- as.Splits(PectinateTree(32 + 32 + 10))
  fourTips <- c("t32", "t33", "t64", "t65")
  sub <- Subsplit(splits, tips = fourTips)
  expect_equal(unname(sub),
               as.Splits(c("t32" = FALSE, "t33" = FALSE, "t64" = TRUE,
                           "t65" = TRUE)), ignore_attr = TRUE)

})

test_that("Bitwise logic works", {
  splits <- as.Splits(BalancedTree(8))
  splits2 <- as.Splits(PectinateTree(8))
  A <- TRUE
  B <- FALSE
  .CSB <- function(a, b) .CompatibleSplit(packBits(as.logical(a)),
                                           packBits(as.logical(b)),
                                           length(a))
  expect_true(.CSB(rep(0, 8), rep(0, 8)))
  expect_true(.CSB(rep(0, 8), rep(1, 8)))
  expect_true(.CSB(rep(1, 8), rep(1, 8)))
  expect_true(.CSB(c(0, 0, 0, 0, 0, 0, 1, 1),
                   c(0, 0, 0, 0, 0, 0, 1, 1)))
  expect_true(.CSB(c(0, 0, 0, 0, 0, 0, 1, 1),
                   c(0, 0, 0, 0, 1, 1, 1, 1)))
  expect_true(.CSB(c(0, 0, 0, 0, 1, 1, 1, 1),
                   c(0, 0, 0, 0, 0, 0, 1, 1)))
  expect_true(.CSB(c(1, 1, 1, 1, 1, 0, 0, 0),
                   c(0, 0, 0, 1, 1, 1, 1, 1)))
  expect_false(.CSB(c(0, 0, 0, 0, 1, 1, 1, 1),
                    c(1, 0, 0, 0, 0, 0, 1, 1)))
  expect_false(.CSB(c(1, 0, 0, 0, 0, 0, 1, 1),
                    c(0, 0, 0, 0, 1, 1, 1, 1)))

  expectation <- matrix(TRUE, 5, 5,
                        dimnames = list(names(splits), names(splits2)))
  expectation["12", "12"] <- FALSE
  expectation["14", "14"] <- FALSE
  expect_equal(CompatibleSplits(splits, splits2), expectation)

  expect_true(.CompatibleSplit(as.raw(3), as.raw(7), nTip = 5))
  expect_false(.CompatibleSplit(as.raw(3), as.raw(6), nTip = 5))
})

test_that("SplitMatchProbability() returns expected probabilities", {
  splitAB   <- as.Splits(c(rep(TRUE, 2), rep(FALSE, 7)))
  splitABC  <- as.Splits(c(rep(TRUE, 3), rep(FALSE, 6)))
  splitABI  <- as.Splits(c(rep(TRUE, 2), rep(FALSE, 6), TRUE))
  splitBCD  <- as.Splits(c(FALSE, rep(TRUE, 3), rep(FALSE, 5)))
  splitAEF  <- as.Splits(c(TRUE, rep(FALSE, 3), rep(TRUE, 2), rep(FALSE, 3)))
  splitABCD <- as.Splits(c(rep(TRUE, 4), rep(FALSE, 5)))
  splitABCE <- as.Splits(c(rep(TRUE, 3), FALSE, TRUE, rep(FALSE, 4)))
  splitCDEF <- as.Splits(c(rep(FALSE, 2), rep(TRUE, 4), rep(FALSE, 3)))
  splitABEF <- as.Splits(c(rep(TRUE, 2), rep(FALSE, 2), rep(TRUE, 2), rep(FALSE, 3)))
  splitABCDE<- as.Splits(c(rep(TRUE, 5), rep(FALSE, 4)))

  splitAI <- as.Splits(c(TRUE, rep(FALSE, 7), TRUE))
  splitHI   <- as.Splits(c(rep(FALSE, 7), rep(TRUE, 2)))
  splitBC <- as.Splits(c(FALSE, TRUE, TRUE, rep(FALSE, 6)))
  splitCD <- as.Splits(c(FALSE, FALSE, TRUE, TRUE, rep(FALSE, 5)))

  expect_error(SplitMatchProbability(as.Splits(TRUE), splitAB))

  expect_equal(LnSplitMatchProbability(splitABC, splitABCD),
               LnSplitMatchProbability(splitABCD,
                                       as.Splits(splitABC, rev(TipLabels(splitABC)))))

  # Possible matches to ABCD:....
  expect_true(SplitMatchProbability(splitABC, splitABCD) <
                SplitMatchProbability(splitAB, splitABCD))

  expect_true(SplitMatchProbability(splitABCD, splitABCD) <
                SplitMatchProbability(splitABC, splitABCD))

  expect_true(SplitMatchProbability(splitBCD, splitABCD) ==
                SplitMatchProbability(splitABC, splitABCD))

  expect_true(SplitMatchProbability(splitABC, splitABCD) <
                SplitMatchProbability(splitAEF, splitABCD))

  expect_true(SplitMatchProbability(splitABC, splitABCD) <
                SplitMatchProbability(splitABI, splitABCD))

  expect_true(SplitMatchProbability(splitABI, splitABCD) <
                SplitMatchProbability(splitAEF, splitABCD))

  expect_true(SplitMatchProbability(splitABC, splitABCD) <
                SplitMatchProbability(splitABCE, splitABCD))

  expect_true(SplitMatchProbability(splitCDEF, splitABCD) ==
                SplitMatchProbability(splitABEF, splitABCD))

  expect_true(SplitMatchProbability(splitABCE, splitABCD) <
                SplitMatchProbability(splitABEF, splitABCD))


  # Two splits of AB:...
  expect_true(SplitMatchProbability(splitAB, splitAB) <
                SplitMatchProbability(splitCD, splitAB))
  expect_true(SplitMatchProbability(splitCD, splitAB) <
                SplitMatchProbability(splitAI, splitAB))


  Test <- function(score, split1, split2) {
    expect_equal(ignore_attr = TRUE, score, SplitMatchProbability(split1, split2))
    expect_equal(ignore_attr = TRUE, score, SplitMatchProbability(split2, split1))

    expect_equal(ignore_attr = TRUE, score, SplitMatchProbability(split1, !split2))
    expect_equal(ignore_attr = TRUE, score, SplitMatchProbability(split2, !split1))

    expect_equal(ignore_attr = TRUE, score, SplitMatchProbability(!split1, !split2))
    expect_equal(ignore_attr = TRUE, score, SplitMatchProbability(!split2, !split1))

    expect_equal(ignore_attr = TRUE, score, SplitMatchProbability(!split1, split2))
    expect_equal(ignore_attr = TRUE, score, SplitMatchProbability(!split2, split1))

    score
  }

  Test(1L, splitAB, splitAI)
  Test(1L, splitAB, splitBC)
  Test(1L, splitBC, splitCD)
  Test(1L, splitHI, splitAI)
  Test(1L, as.Splits(c(rep(TRUE, 4), rep(FALSE, 4))), # Test even splits
       as.Splits(c(rep(TRUE, 2), rep(FALSE, 2), rep(TRUE, 2), rep(FALSE, 2))))

  Test(1/36, splitAB, splitAB)
  Test(1/36, splitBC, splitBC)

  Test(1/126, splitABCD, splitABCD)
  Test(1/126, splitABEF, splitABEF)
  Test(1/126, splitCDEF, splitCDEF)
  Test(1, splitABCD, splitABEF)

  Test(1/12, splitAB, splitABC)
  Test(1/12, splitBC, splitABC)
  Test(1/12, splitBC, splitBCD)

  Test(1, splitAEF, splitABCD)
  Test(66 / 126, splitABC, splitABEF)
  Test(66 / 126, splitBCD, splitCDEF)

  Test(4/84, splitABC, splitABCD)
  Test(4/84, splitBCD, splitABCD)
})


test_that("Tip labels are found", {
  pt4 <- PectinateTree(4L)
  bt4 <- BalancedTree(4L)
  t1..4 <- c("t1", "t2", "t3", "t4")
  expect_equal(TipLabels(0), character(0))
  expect_equal(t1..4, TipLabels(t1..4))
  expect_equal(t1..4, TipLabels(pt4))
  expect_equal(t1..4, TipLabels(as.Splits(pt4)))
  expect_equal(t1..4, TipLabels(list(pt4, bt4)))
  expect_equal(t1..4, TipLabels(structure(list(pt4, bt4),
                                          class="multiPhylo")))
  expect_equal(t1..4, TipLabels(structure(list(pt4, BalancedTree(letters[1:4])),
                                          class="multiPhylo"), single = TRUE))
  expect_equal(t1..4, TipLabels(structure(list(pt4, bt4,
                                               tip.label = t1..4),
                                          class="multiPhylo")))
  expect_equal(t1..4, TipLabels(as.TreeNumber(pt4)))

  expect_equal(t1..4, TipLabels(structure(list(), tip.label = t1..4)))
  expect_equal(t1..4, TipLabels(list(tip.label = t1..4)))
  expect_null(TipLabels(list()))
  expect_equal(TipLabels(pt4), TipLabels(list(as.Splits(pt4))))
  expect_equal(TipLabels(pt4, single = TRUE),
               TipLabels(list(as.Splits(pt4)), single = TRUE))
  expect_equal(TipLabels(pt4), TipLabels(list(as.Splits(pt4), as.Splits(bt4))))
  expect_equal(list(t1..4, c(t1..4, "t5")),
                    TipLabels(list(as.Splits(pt4),
                                   as.Splits(BalancedTree(5)))))

  expect_equal(t1..4, TipLabels(c(t1 = 1, t2 = 3, t3 = 3, t4 = 4)))
  
  # Duplicates permitted
  expect_equal(TipLabels(c(1:4, 4:1)), c(1:4, 4:1))
  expect_equal(TipLabels(c(t1..4, t1..4)), c(t1..4, t1..4))
  
  # Actual cases tested by test cases eludes me
  expect_equal(TipLabels(list(tip.label = list(tip.label = t1..4))),
               t1..4)
  expect_equal(TipLabels(structure(list(tip.label = list(tip.label = t1..4)),
                                   class = "multiPhylo")),
               t1..4)
})

test_that("AllTipLabels()", {
  expect_equal(sort(as.integer(
    AllTipLabels(c(BalancedTree(1:5), PectinateTree(6:10))))),
               1:10)
  expect_equal(sort(as.integer(
    AllTipLabels(list(BalancedTree(1:6), PectinateTree(4:10))))),
               1:10)
  bal8 <- BalancedTree(8)
  expect_equal(AllTipLabels(bal8), TipLabels(bal8))
  expect_equal(AllTipLabels(matrix(4, 4, 4, 
                                   dimnames = list(NULL, paste0("t", 1:4)))),
               paste0("t", 1:4))
  expect_equal(AllTipLabels(as.Splits(bal8)), TipLabels(bal8))
  expect_equal(AllTipLabels(as.TreeNumber(bal8)), TipLabels(bal8))
})

#' Brewer palettes
#'
#' A list of eleven Brewer palettes containing one to eleven colours that
#' are readily distinguished by colourblind viewers, followed by a twelfth
#' 12-colour palette adapted for colour blindness.
#'
#' @source{
#'
#' * [ColourBrewer2.org](https://colorbrewer2.org/#type=diverging&scheme=RdYlBu&n=3)
#'
#' * [Martin Krzywinski](https://mk.bcgsc.ca/colorblind/)
#' }
#'
#' @examples
#' data("brewer", package = "TreeTools")
#' plot(0, type = "n", xlim = c(1, 12), ylim = c(12, 1),
#'      xlab = "Colour", ylab="Palette")
#' for (i in seq_along(brewer)) text(seq_len(i), i, col = brewer[[i]])
#'
#' @keywords datasets
#' @name brewer
"brewer"

#' Double factorials
#'
#' A vector with pre-calculated values of double factorials up to 300!!,
#' and the logarithms of double factorials up to 50 000!!.
#'
#' 301!! is too large to store as an integer; use `logDoubleFactorials` instead.
#'
#' @family double factorials
#' @keywords datasets
#' @name doubleFactorials
#' @export
"doubleFactorials"

#' Natural logarithms of double factorials
#'
#' `logDoubleFactorials` is a numeric vector with pre-calculated values of
#' double factorials up to 50 000!!.
#'
#' @family double factorials
#' @keywords datasets
#' @name logDoubleFactorials
#' @export
"logDoubleFactorials"

#' @rdname TreeShape
#' @format `unrootedKeys` is a list of length `r length(unrootedKeys)`; each
#' entry is a vector of integers corresponding to the keys (not shape numbers)
#' of the different unrooted tree shapes with `nTip` leaves.
"unrootedKeys"

#' Number of rooted / unrooted tree shapes
#'
#' `nRootedShapes` and `nUnrootedShapes` give the number of (un)rooted binary
#' trees on _n_ unlabelled leaves.
#'
#' @source
#' `nRootedShapes` corresponds to the Wedderburn-Etherington numbers,
#' [\acronym{OEIS} A001190](https://oeis.org/A001190)
#'
#' `nUnrootedShapes` is [\acronym{OEIS} A000672](https://oeis.org/A000672)
#'
#' @keywords datasets
"nRootedShapes"

#' @rdname nRootedShapes
"nUnrootedShapes"

#' Data from Zhang et al. 2016
#'
#' Phylogenetic data from \insertCite{Zhang2016;textual}{TreeTools} in raw
#' (`Lobo.data`) and `phyDat` (`Lobo.phy`) formats.
#'
#' @template LoboMods
#'
#' @examples
#' data("Lobo", package = "TreeTools")
#' Lobo.data
#' Lobo.phy
#' @source \insertCite{Zhang2016;textual}{TreeTools}
#' 
#' @references \insertAllCited{}
#'
#' @keywords datasets
"Lobo.data"

#' @rdname Lobo.data
"Lobo.phy"

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

ClusterTable_new <- function(phylo) {
    .Call(`_TreeTools_ClusterTable_new`, phylo)
}

ClusterTable_matrix <- function(xp) {
    .Call(`_TreeTools_ClusterTable_matrix`, xp)
}

ClusterTable_decode <- function(xp) {
    .Call(`_TreeTools_ClusterTable_decode`, xp)
}

ape_neworder_phylo <- function(n_tips, parent, child, n_edges, order) {
    .Call(`_TreeTools_ape_neworder_phylo`, n_tips, parent, child, n_edges, order)
}

ape_neworder_pruningwise <- function(n_tips, n_node, parent, child, n_edges) {
    .Call(`_TreeTools_ape_neworder_pruningwise`, n_tips, n_node, parent, child, n_edges)
}

as_newick <- function(edge) {
    .Call(`_TreeTools_as_newick`, edge)
}

split_frequencies <- function(trees) {
    .Call(`_TreeTools_split_frequencies`, trees)
}

consensus_tree <- function(trees, p) {
    .Call(`_TreeTools_consensus_tree`, trees, p)
}

descendant_edges <- function(parent, child, postorder) {
    .Call(`_TreeTools_descendant_edges`, parent, child, postorder)
}

descendant_edges_single <- function(parent, child, postorder, edge_index, include_self = TRUE) {
    .Call(`_TreeTools_descendant_edges_single`, parent, child, postorder, edge_index, include_self)
}

descendant_tips <- function(parent, child, postorder) {
    .Call(`_TreeTools_descendant_tips`, parent, child, postorder)
}

splits_to_char <- function(membership, labels) {
    .Call(`_TreeTools_splits_to_char`, membership, labels)
}

first_matching_split_pair <- function(x, table) {
    .Call(`_TreeTools_first_matching_split_pair`, x, table)
}

first_matching_split_index <- function(x, table) {
    .Call(`_TreeTools_first_matching_split_index`, x, table)
}

num_to_parent <- function(n, nTip) {
    .Call(`_TreeTools_num_to_parent`, n, nTip)
}

random_parent <- function(nTip, seed) {
    .Call(`_TreeTools_random_parent`, nTip, seed)
}

edge_to_num <- function(parent, child, nTip) {
    .Call(`_TreeTools_edge_to_num`, parent, child, nTip)
}

edge_to_mixed_base <- function(parent, child, nTip) {
    .Call(`_TreeTools_edge_to_mixed_base`, parent, child, nTip)
}

mixed_base_to_parent <- function(n, nTip) {
    .Call(`_TreeTools_mixed_base_to_parent`, n, nTip)
}

kept_vertices <- function(edge, kept) {
    .Call(`_TreeTools_kept_vertices`, edge, kept)
}

minimum_spanning_tree <- function(order) {
    .Call(`_TreeTools_minimum_spanning_tree`, order)
}

n_cherries_wrapper <- function(parent, child, nTip) {
    .Call(`_TreeTools_n_cherries_wrapper`, parent, child, nTip)
}

node_depth_unrooted <- function(parent, child, postorder, shortest) {
    .Call(`_TreeTools_node_depth_unrooted`, parent, child, postorder, shortest)
}

path_lengths <- function(edge, weight, init_nas) {
    .Call(`_TreeTools_path_lengths`, edge, weight, init_nas)
}

renumber_tips_batch <- function(trees, perm, n_tip, new_labels) {
    .Call(`_TreeTools_renumber_tips_batch`, trees, perm, n_tip, new_labels)
}

cpp_edge_to_splits <- function(edge, order, nTip) {
    .Call(`_TreeTools_cpp_edge_to_splits`, edge, order, nTip)
}

duplicated_splits <- function(splits, fromLast) {
    .Call(`_TreeTools_duplicated_splits`, splits, fromLast)
}

mask_splits <- function(x) {
    .Call(`_TreeTools_mask_splits`, x)
}

not_splits <- function(x) {
    .Call(`_TreeTools_not_splits`, x)
}

xor_splits <- function(x, y) {
    .Call(`_TreeTools_xor_splits`, x, y)
}

and_splits <- function(x, y) {
    .Call(`_TreeTools_and_splits`, x, y)
}

or_splits <- function(x, y) {
    .Call(`_TreeTools_or_splits`, x, y)
}

split_consistent <- function(needle, haystacks, invert) {
    .Call(`_TreeTools_split_consistent`, needle, haystacks, invert)
}

thin_splits <- function(splits, drop) {
    .Call(`_TreeTools_thin_splits`, splits, drop)
}

pack_splits_logical <- function(x) {
    .Call(`_TreeTools_pack_splits_logical`, x)
}

pack_splits_logical_vec <- function(x) {
    .Call(`_TreeTools_pack_splits_logical_vec`, x)
}

cpp_count_splits <- function(edge, nTip) {
    .Call(`_TreeTools_cpp_count_splits`, edge, nTip)
}

normalize_splits <- function(splits, n_tip) {
    .Call(`_TreeTools_normalize_splits`, splits, n_tip)
}

splits_to_edge <- function(splits, nTip) {
    .Call(`_TreeTools_splits_to_edge`, splits, nTip)
}

tips_in_splits <- function(splits) {
    .Call(`_TreeTools_tips_in_splits`, splits)
}

edge_to_rooted_shape <- function(parent, child, nTip) {
    .Call(`_TreeTools_edge_to_rooted_shape`, parent, child, nTip)
}

rooted_shape_to_edge <- function(shape, nTip) {
    .Call(`_TreeTools_rooted_shape_to_edge`, shape, nTip)
}

test_that("Failures are graceful", {
  expect_error(num_to_parent(10, 1), "`nTip` must be > 1")
  expect_error(num_to_parent(10, -1), "`nTip` must be > 1")
  expect_warning(num_to_parent(10, c(2, 3)),
                 "`nTip` should be a single integer")
  expect_error(num_to_parent(NA, 10), "`n` may not contain NA")
  expect_error(num_to_parent(-1, 10), "`n` may not be negative")
  
  expect_error(mixed_base_to_parent(10, 1), "`nTip` must be > 1")
  expect_error(mixed_base_to_parent(10, -1), "`nTip` must be > 1")
  expect_warning(mixed_base_to_parent(10, c(2, 3)),
                 "`nTip` should be a single integer")
  expect_error(mixed_base_to_parent(NA, 10), "`n` may not contain NA")
  expect_error(mixed_base_to_parent(-1, 10), "`n` may not be negative")
  
  expect_error(edge_to_num(1:10, 1:11, 6), "Parent and child .* same length")
  expect_error(edge_to_num(1:10, 1:10, 5), "nEdge must == nTip . nTip - 2")
  expect_error(edge_to_num(1:110, 1:110, 56), "Too many leaves")
  expect_error(edge_to_mixed_base(1:10, 1:11, 6),
               "Parent and child .* same length")
  expect_warning(expect_error(
    edge_to_mixed_base(1:10, 1:10, c(5, 6)),
    "nEdge must == nTip . nTip - 2"),
    "`nTip` should be a single integer"
    )
  expect_equal(as.phylo(0, 0), ZeroTaxonTree())
  
  expect_error(as.phylo(0, -1), "`nTip` may not be negative")
})

test_that("Edge cases handled", {
  expect_equal(SingleTaxonTree("singleton"), as.phylo(0, 1, "singleton"))
  expect_true(all.equal(BalancedTree(2), as.phylo(0, 2)))
  expect_equal(structure(as.integer64(0L), nTip = 2L, tip.label = c("t1", "t2"),
                         class = c("TreeNumber", "integer64")),
               as.TreeNumber(as.phylo(0, 2)))
  expect_equal(structure(integer64(1), nTip = 1L, tip.label = "t1",
                         class = c("TreeNumber", "integer64")),
               as.TreeNumber(SingleTaxonTree()))
})

test_that("Trees generated okay", {
  expect_true(all.equal(
    Preorder(ape::read.tree(text=("(0, (4, ((1, 5), (2, 3))));"))),
    as.phylo(10, 6, 0:5)))
  expect_equal(as.TreeNumber("10", 6, 0:5),
               as.TreeNumber(as.phylo.numeric(10, 6, 0:5)))
  Test <- function(i, nTip) {
    expect_equal(as.integer64(i),
                 as.integer64(as.TreeNumber(as.phylo(i, nTip, seq_len(nTip) - 1L))))
  }
  xx <- lapply(0:104, Test, 6)
  xx <- lapply(seq_len(NUnrooted(7)) - 1L, Test, 7)
  xx <- lapply(floor(runif(48) * NUnrooted(8)), Test, 8)
  xx <- lapply(floor(runif(48) * NUnrooted(10)), Test, 10)
  xx <- lapply(floor(runif(48) * NUnrooted(12)), Test, 12)

  nTip <- 14L
  treeNumber <- as.TreeNumber("123456789876", nTip, seq_len(nTip) - 1L)
  expect_equal(treeNumber, as.TreeNumber(as.phylo(treeNumber)))
  treeNumber <- as.TreeNumber("123456789876", nTip, seq_len(nTip) - 1L)
  expect_equal(treeNumber, as.TreeNumber(as.phylo(treeNumber)))
  
  expect_equal
})

test_that("Bad data results in error", {
  skip_if(Sys.getenv("USING_ASAN") != "")
  expect_error(cpp_edge_to_splits(matrix(1, 3, 3), 1:3, 3),
               "must contain two col")
  expect_equal(cpp_edge_to_splits(matrix(1, 10, 2), 1:10, 0),
               matrix(raw(0), 0, 0))
  expect_error(cpp_edge_to_splits(matrix(1, 10, 2), 1:10, -10),
               "Tree must contain non-negative number of tips")
  expect_error(cpp_edge_to_splits(matrix(1, 10, 2), 1:9, 5),
               "ength of `order` must equal number of edges")
  expect_error(cpp_edge_to_splits(matrix(1, 2, 2), 1:2, 6),
               "Not enough edges") # Impossible anyway...
  
  s9 <- as.Splits(BalancedTree(9))
  expect_error(duplicated_splits(structure(s9, nTip = NULL), TRUE),
               "`nTip` attribute")
  expect_error(duplicated_splits(structure(s9, nTip = 999L), TRUE),
               "tip number")
  
  expect_error(s9 & s9[[1]], "same number of splits")
  expect_error(s9 | s9[[1]], "same number of splits")
  expect_error(xor(s9, s9[[1]]), "same number of splits")
  
  noNTip <- structure(s9, nTip = NULL)
  expect_error(noNTip & s9, "`x` lacks nTip attrib")
  expect_error(noNTip | s9, "`x` lacks nTip attrib")
  expect_error(xor(noNTip, s9), "`x` lacks nTip attrib")
  
  expect_error(s9 & noNTip, "`y` lacks nTip attrib")
  expect_error(s9 | noNTip, "`y` lacks nTip attrib")
  expect_error(xor(s9, noNTip), "`y` lacks nTip attrib")

  wrongTip <- structure(s9, nTip = 15L)
  expect_error(xor(s9, wrongTip), "`y` differ in `nTip`")
  expect_error(s9 & wrongTip, "`y` differ in `nTip`")
  expect_error(s9 | wrongTip, "`y` differ in `nTip`")
})

#' Reorder edges of a phylogenetic tree
#'
#' Wrappers for the C functions called by
#' \code{ape::\link[ape:reorder.phylo]{reorder.phylo}}.
#' These call the C functions directly, so are faster -- but don't perform
#' as many checks on user input.  Bad input could crash R.
#'
#'
#' @param nTip,nNode,nEdge Integer specifying the number of tips, nodes
#' and edges in the input tree.
#' @template treeParent
#' @template treeChild
#'
#' @return `NeworderPruningwise` returns an integer vector specifying the
#' pruningwise order of edges within a tree.
#'
#' @examples
#' nTip <- 8L
#' tree <- BalancedTree(nTip)
#' edge <- tree[["edge"]]
#' pruningwise <- NeworderPruningwise(nTip, tree$Nnode, edge[, 1], edge[, 2],
#'                                    dim(edge)[1])
#' cladewise <- NeworderPhylo(nTip, edge[, 1], edge[, 2], dim(edge)[1], 1L)
#' postorder <- NeworderPhylo(nTip, edge[, 1], edge[, 2], dim(edge)[1], 2L)
#'
#' tree[["edge"]] <- tree[["edge"]][pruningwise, ]
#'
#' @author
#'  - C algorithm: Emmanuel Paradis
#'  - R wrapper: Martin R. Smith
#' @family C wrappers
#' @keywords internal
#' @useDynLib TreeTools, .registration = TRUE
#' @name Neworder
#' @export
NeworderPruningwise <- function(nTip, nNode, parent, child, nEdge) {
  .Call(`_TreeTools_ape_neworder_pruningwise`, as.integer(nTip),
        as.integer(nNode), as.integer(parent), as.integer(child),
        as.integer(nEdge))
}

#' @rdname Neworder
#' @param whichwise Integer specifying whether to order edges (1)
#' cladewise; or (2) in postorder.
#' @return `NeworderPhylo` returns an integer vector specifying the order
#' of edges under the ordering sequence specified by `whichwise`.
#' @keywords internal
#' @export
NeworderPhylo <- function(nTip, parent, child, nEdge, whichwise) {
  .Call(`_TreeTools_ape_neworder_phylo`, as.integer(nTip), as.integer(parent),
        as.integer(child), as.integer(nEdge), as.integer(whichwise))
}

#' @rdname Reorder
#'
#' @template treeParent
#' @template treeChild
#' @param weight Optional vector specifying the weight of each edge,
#' corresponding to the `edge.length` property of a `phylo` object.
#'
#' @return `RenumberTree()` returns an edge matrix for a tree of class `phylo`
#' following the preorder convention for edge and node numbering.
#'
#' @family tree manipulation
#' @family C wrappers
#' @export
RenumberTree <- function(parent, child, weight) {
  if (missing(weight)) {
    .Call(`_TreeTools_preorder_edges_and_nodes`, parent, child)
  } else {
    .Call(`_TreeTools_preorder_weighted`, parent, child, weight)
  }
}

#' @rdname Reorder
#'
#' @return `RenumberEdges()` formats the output of `RenumberTree()` into a list
#' whose two entries correspond to the new parent and child vectors,
#' in preorder.
#' @export
RenumberEdges <- function(parent, child) {
  oenn <- .Call(`_TreeTools_preorder_edges_and_nodes`, parent, child)

  # Return:
  list(oenn[, 1], oenn[, 2])
}

#' Reorder tree edges and nodes
#'
#' Functions for systematically ordering the internal edges of trees.
#'
#' `Reorder()` is a wrapper for \code{ape:::.reorder_ape}.
#' Calling this C function directly is approximately twice as fast as using
#' \code{ape::\link[ape:reorder.phylo]{cladewise}} or
#' \code{ape::\link[ape:reorder.phylo]{postorder}}
#'
#' `Cladewise()`, `ApePostorder()` and `Pruningwise()` are convenience
#' functions to the corresponding functions in "ape".
#' Single nodes may need to be collapsed using [ape::collapse.singles] first.
#' "ape" functions can cause crashes if nodes are numbered unconventionally --
#' sometimes arising after using tree rearrangement functions,
#' e.g. `phangorn::SPR()`.
#'
#' `Preorder()` is more robust: it supports polytomies, nodes may be numbered
#' in any sequence, and edges may be listed in any order in the input tree.
#' Its output is guaranteed to be identical for any tree of an equivalent
#' leaf labelling (see `RenumberTips()`) and topology,
#' allowing unique trees to be detected by comparing sorted edge matrices alone.
#'
#' Nodes and edges in a preorder tree are numbered starting from the deepest
#' node.
#' Each node is numbered in the sequence in which it is encountered, and
#' each edge is listed in the sequence in which it is visited.
#'
#' At each node, child edges are sorted from left to right in order of the
#' lowest-numbered leaf in the subtree subtended by each edge; i.e. an edge
#' that leads eventually to tip 1 will be to the left of an edge leading to a
#' subtree containing tip 2.
#'
#' Numbering begins by following the leftmost edge of the root node,
#' and sorting its descendant subtree into preorder.
#' Then, the next edge at the root node is followed, and its descendants
#' sorted into preorder, until each edge has been visited.
#'
#' `RenumberTree()` and `RenumberEdges()` are wrappers for the C function
#' `preorder_edges_and_nodes()`; they do not perform the same checks on input
#' as `Preorder()` and are intended for use where performance is at a premium.
#'
#'
#' `Postorder()` numbers nodes as in `Preorder()`, and lists edges in
#' descending order of parent node number, breaking ties by listing child
#' nodes in increasing order.  If a tree is already in postorder, it will not
#' be rearranged unless `force = TRUE`.
#'
#' Methods applied to numeric inputs do not check input for sanity, so should
#' be used with caution: malformed input may cause undefined results, including
#' crashing R.
#'
#' Trees with >8191 leaves require additional memory and are not handled
#' by `Postorder()` at present.
#' If you need to process such large trees, please contact the maintainer for
#' advice.
#'
#' @template treeParam
#' @template nTipParam
#' @param edge Two-column matrix listing the parent and child of each edge in a
#' tree, corresponding to `tree[["edge"]]`. Optional in `Cladewise()`.
#'
#' @return `ApePostorder()`, `Cladewise()`, `Postorder()`, `Preorder()` and
#' `Pruningwise()` each return a tree of class `phylo` with nodes following the
#' specified numbering scheme.
#' @author
#' `Preorder()` and `Postorder()`: Martin R. Smith.
#'
#' `Cladewise()`, `ApePostorder()` and `Pruningwise()`: modified by Martin R.
#' Smith from \code{.reorder_ape()} in \pkg{ape} (Emmanuel Paradis).
#'
#'
#' @seealso Rotate each node into a consistent orientation with [`SortTree()`].
#'
#' @family C wrappers
#' @keywords internal
#' @name Reorder
NULL

#' @describeIn Reorder Reorder tree cladewise.
#' @export
Cladewise <- function(tree, nTip, edge) UseMethod("Cladewise")

#' @rdname Reorder
#' @export
Cladewise.phylo <- function(tree, nTip = NTip(tree), edge = tree[["edge"]]) {
  if (!is.null(attr(tree, "order")) && attr(tree, "order") == "cladewise") {
    return(tree)
  }
  nEdge <- dim(edge)[1]
  nNode <- tree[["Nnode"]]
  if (nNode == 1) {
    return(tree)
  }
  if (nNode >= nTip) {
    stop("`tree` apparently badly conformed")
  }

  newOrder <- NeworderPhylo(nTip, edge[, 1], edge[, 2], nEdge, 1)

  tree[["edge"]] <- edge[newOrder, ]
  if (!is.null(tree[["edge.length"]])) {
    tree[["edge.length"]] <- tree[["edge.length"]][newOrder]
  }
  attr(tree, "order") <- "cladewise"
  attr(tree, "suborder") <- NULL

  # Return:
  tree
}

#' @rdname Reorder
#' @export
Cladewise.list <- function(tree, nTip, edge) {
  lapply(tree, Cladewise)
}

#' @rdname Reorder
#' @export
Cladewise.multiPhylo <- function(tree, nTip, edge) {
  tree[] <- lapply(tree, Cladewise)
  attr(tree, "order") <- "cladewise"
  attr(tree, "suborder") <- NULL
  tree
}

#' @rdname Reorder
#' @export
Cladewise.matrix <- function(tree, nTip = min(tree[, 1]) - 1L, edge) {
  if (is.numeric(tree)) {
    newOrder <- NeworderPhylo(nTip, tree[, 1], tree[, 2], dim(tree)[1], 1L)

    # Return:
    tree[newOrder, ]
  } else {
    NextMethod()
  }
}

#' @rdname Reorder
#' @export
Cladewise.NULL <- function(tree, nTip = min(tree[, 1]) - 1L, edge) NULL

#' @describeIn Reorder Reorder tree in Postorder using ape's `postorder`
#' function, which is robust to unconventional node numbering.
#' @export
ApePostorder <- function(tree, nTip, edge) UseMethod("ApePostorder")

#' @rdname Reorder
#' @export
ApePostorder.phylo <- function(tree, nTip = NTip(tree), edge = tree[["edge"]]) {
  if (!is.null(attr(tree, "order")) && attr(tree, "order") == "postorder") {
    return(tree)
  }
  nEdge <- dim(edge)[1]
  nNode <- tree[["Nnode"]]
  if (nNode == 1) {
    return(tree)
  }
  if (nNode >= nTip) {
    stop("`tree` apparently badly conformed")
  }
  neworder <- NeworderPhylo(nTip, edge[, 1], edge[, 2], nEdge, 2L)
  tree[["edge"]] <- edge[neworder, ]
  if (!is.null(tree[["edge.length"]])) {
    tree[["edge.length"]] <- tree[["edge.length"]][neworder]
  }
  attr(tree, "order") <- "postorder"
  attr(tree, "suborder") <- "ape"
  tree
}

#' @rdname Reorder
#' @export
ApePostorder.list <- function(tree, nTip, edge) {
  lapply(tree, ApePostorder)
}

#' @rdname Reorder
#' @export
ApePostorder.NULL <- function(tree, nTip, edge) NULL

#' @rdname Reorder
#' @export
ApePostorder.multiPhylo <- function(tree, nTip, edge) {
  tree[] <- lapply(tree, ApePostorder)
  attr(tree, "order") <- "postorder"
  attr(tree, "suborder") <- "ape"
  tree
}

#' @rdname Reorder
#' @param force Logical specifying whether to rearrange trees already in
#' postorder, in order to ensure edges are ordered in the "TreeTools" fashion.
#' @export
Postorder <- function(tree, force = FALSE) {
  UseMethod("Postorder")
}

#' @rdname Reorder
#' @export
Postorder.phylo <- function(tree, force = FALSE) {
  if (is.null(attr(tree, "order"))
      || attr(tree, "order") != "postorder"
      || (force &&
          (is.null(attr(tree, "suborder")) ||
           attr(tree, "suborder") != "TreeTools"))) {
    weight <- tree[["edge.length"]]
    edge <- tree[["edge"]]
    parent <- edge[, 1]
    child <- edge[, 2]
    if (!is.null(weight)) {
      rnt <- RenumberTree(parent, child, weight)
      edge <- rnt[[1]]
      newOrder <- order(edge[, 1], edge[, 2], decreasing = TRUE,
                        method = "radix")
      tree[["edge.length"]] <- rnt[[2]][newOrder]
    } else {
      edge <- RenumberTree(parent, child)
      newOrder <- order(edge[, 1], edge[, 2], decreasing = TRUE,
                        method = "radix")
    }
    tree[["edge"]] <- edge[newOrder, , drop = FALSE]
    attr(tree, "order") <- "postorder"
    attr(tree, "suborder") <- "TreeTools"
  }
  tree
}

#' @rdname Reorder
#' @export
Postorder.NULL <- function(tree, force = FALSE) NULL

#' @rdname Reorder
#' @export
Postorder.list <- function(tree, force = FALSE) {
  lapply(tree, Postorder, force = force)
}


#' @rdname Reorder
#' @export
Postorder.multiPhylo <- function(tree, force = FALSE) {
  tree[] <- lapply(tree, Postorder, force = force)
  attr(tree, "order") <- "postorder"
  tree
}

#' @rdname Reorder
#' @return `Postorder.numeric` accepts a numeric matrix corresponding to the
#' `edge` entry of a tree of class `phylo`, and returns a two-column array
#' corresponding to `tree`, with edges listed in postorder
#' @export
Postorder.numeric <- function(tree, force = FALSE) {
  edge <- RenumberTree(tree[, 1], tree[, 2])
  ordr <- order(edge[, 1], edge[, 2],
                decreasing = c(TRUE, FALSE), method = "radix")
  edge[ordr, , drop = FALSE]
}

#' @rdname Reorder
#' @return `PostorderOrder()` returns an integer vector. Visiting edges in this
#' order will traverse the tree in postorder.
#' @export
PostorderOrder <- function(tree) UseMethod("PostorderOrder")

#' @rdname Reorder
#' @export
PostorderOrder.phylo <- function(tree) {
  order <- attr(tree, "order")
  if (is.null(order)) {
    order <- character(1)
  }
  edge <- tree[["edge"]]
  
  # Return:
  switch(order,
         "preorder" = dim(edge)[1]:1,
         "postorder" = seq_len(dim(edge)[1]),
         postorder_order(edge))
}

#' @export
PostorderOrder.NULL <- function(tree) NULL
  
  
#' @rdname Reorder
#' @export
PostorderOrder.numeric <- function(tree) {
  dims <- dim(tree)
  if (is.null(dims) || dims[2] != 2) {
    stop("`tree` must be the edge matrix of a `phylo` object.")
  }
  postorder_order(tree)
}

#' @describeIn Reorder Reorder tree Pruningwise.
#' @export
Pruningwise <- function(tree, nTip, edge) UseMethod("Pruningwise")

#' @rdname Reorder
#' @export
Pruningwise.phylo <- function(tree, nTip = NTip(tree),
                              edge = tree[["edge"]]) {
  if (!is.null(attr(tree, "order")) && attr(tree, "order") == "pruningwise") {
    return(tree)
  }
  nEdge <- dim(edge)[1]
  nNode <- tree[["Nnode"]]
  if (nNode == 1) {
    return(tree)
  }
  if (nNode >= nTip) {
    stop("`tree` apparently badly conformed")
  }
  tree <- Cladewise(tree, nTip, edge)
  edge <- tree[["edge"]]
  neworder <- NeworderPruningwise(nTip, nNode, edge[, 1], edge[, 2], nEdge)
  tree[["edge"]] <- tree[["edge"]][neworder, ]
  if (!is.null(tree[["edge.length"]])) {
    tree[["edge.length"]] <- tree[["edge.length"]][neworder]
  }
  attr(tree, "order") <- "pruningwise"
  attr(tree, "suborder") <- NULL
  tree
}

#' @rdname Reorder
#' @export
Pruningwise.list <- function(tree, nTip, edge) {
  lapply(tree, Pruningwise)
}

#' @rdname Reorder
#' @export
Pruningwise.multiPhylo <- function(tree, nTip, edge) {
  tree[] <- lapply(tree, Pruningwise)
  attr(tree, "order") <- "pruningwise"
  attr(tree, "suborder") <- NULL
  tree
}

#' @rdname Reorder
#' @export
Pruningwise.NULL <- function(tree, nTip, edge) NULL

#' @describeIn Reorder Reorder tree in Preorder (special case of cladewise).
#' @param topologyOnly Logical; if `TRUE`, edge weights may not be retained.
#' @export
Preorder <- function(tree, topologyOnly = FALSE) UseMethod("Preorder")

#' @rdname Reorder
#' @export
Preorder.phylo <- function(tree, topologyOnly = FALSE) {
  startOrder <- attr(tree, "order")
  if (length(startOrder) && startOrder == "preorder") {
    # length(x) is twice as fast as !is.null(x)
    tree
  } else {
    edge <- tree[["edge"]]
    parent <- edge[, 1]
    child <- edge[, 2]
    if (topologyOnly) {
      tree[["edge"]] <- RenumberTree(parent, child)
      tree[["edge.length"]] <- NULL
      tree[["node.label"]] <- NULL
    } else {
      lengths <- tree[["edge.length"]]
      if (topologyOnly || length(lengths) == 0) {
        tree[["edge"]] <- RenumberTree(parent, child)
      } else {
        newEdge <- RenumberTree(parent, child, lengths)
        tree[["edge"]] <- newEdge[[1]]
        tree[["edge.length"]] <- newEdge[[2]]
      }
      nodeLabels <- tree[["node.label"]]
      if (length(nodeLabels)) {
        tree[["node.label"]] <- .UpdateNodeLabel.numeric(edge, tree, nodeLabels)
      }
    }
    attr(tree, "order") <- "preorder"
    attr(tree, "suborder") <- NULL
    
    # Return:
    tree
  }
}

#' @rdname Reorder
#' @export
Preorder.numeric <- function(tree, topologyOnly = FALSE) {
  RenumberTree(tree[, 1], tree[, 2])
}

#' @rdname Reorder
#' @export
Preorder.multiPhylo <- function(tree, topologyOnly = FALSE) {
  tree[] <- lapply(tree, Preorder, topologyOnly)
  attr(tree, "order") <- "preorder"
  attr(tree, "suborder") <- NULL
  tree
}

#' @rdname Reorder
#' @export
Preorder.list <- function(tree, topologyOnly = FALSE) {
  lapply(tree, Preorder, topologyOnly)
}

#' @rdname Reorder
#' @export
Preorder.NULL <- function(tree, topologyOnly = FALSE) NULL

#' @describeIn Reorder Reorder tree in postorder, numbering internal nodes
#' according to [TNT's rules](https://stackoverflow.com/a/54296100/3438001),
#' which number the root node as `nTip + 1`, then the remaining nodes in
#' the sequence encountered when traversing the tree in postorder, starting from
#' each tip in sequence.
#' @export
TntOrder <- function(tree) UseMethod("TntOrder")

#' @rdname Reorder
#' @export
TNTOrder <- TntOrder

#' @rdname Reorder
#' @export
TntOrder.phylo <- function(tree) {
  startOrder <- attr(tree, "order")
  if (length(startOrder) && startOrder == "tnt") {
    # length(x) is twice as fast as !is.null(x)
    tree
  } else {
    # TODO this'd be faster in C++
    tree <- Postorder(tree)
    
    nTip <- NTip(tree)
    nNode <- tree[["Nnode"]]
    root <- nTip + 1
    edge <- tree[["edge"]]
    parent <- edge[, 1]
    child <- edge[, 2]
    parentOf <- numeric(nTip + nNode)
    parentOf[child] <- parent
    newNo <- c(seq_len(root), numeric(nNode - 1L))
    renum <- child > root
    n <- root + 1
    for (i in seq_len(nTip)) {
      ptr <- i
      while (parentOf[ptr] > root && !newNo[parentOf[ptr]]) {
        newNo[parentOf[ptr]] <- n
        n <- n + 1
        ptr <- parentOf[ptr]
      }
    }
    stopifnot(all(newNo > 0)) # TODO remove once fully tested
    tree[["edge"]][] <- newNo[edge]
    attr(tree, "order") <- "tnt"
    attr(tree, "suborder") <- NULL
    
    # Return:
    tree
  }
}

#' @rdname Reorder
#' @export
TntOrder.numeric <- function(tree) {
  stop(RenumberTree(tree[, 1], tree[, 2]))
}

#' @rdname Reorder
#' @export
TntOrder.multiPhylo <- function(tree) {
  tree[] <- lapply(Postorder(tree), TntOrder)
  attr(tree, "order") <- "tnt"
  tree
}

#' @rdname Reorder
#' @export
TntOrder.list <- function(tree) {
  lapply(tree, TntOrder)
}

#' @rdname Reorder
#' @export
TntOrder.NULL <- function(tree) NULL



#' Renumber a tree's tips
#'
#' `RenumberTips(tree, tipOrder)` sorts the tips of a phylogenetic tree `tree`
#' such that the indices in `tree[["edge"]][, 2]` correspond to the order of
#' leaves given in `tipOrder`.
#'
#' @template treeParam
#' @param tipOrder A character vector containing the values of
#' `tree[["tip.label"]]` in the desired sort order,
#' an object (perhaps of class `phylo` or `Splits`) with tip labels,
#' or a numeric specifying the desired order.
#'
#' @return `RenumberTips()` returns `tree`, with the tips' internal
#' representation numbered to match `tipOrder`.
#'
#' @examples
#' data("Lobo") # Loads the phyDat object Lobo.phy
#' tree <- RandomTree(Lobo.phy)
#' tree <- RenumberTips(tree, names(Lobo.phy))
#'
#' @family tree manipulation
#'
#' @template MRS
#' @export
RenumberTips <- function(tree, tipOrder) UseMethod("RenumberTips")

#' @rdname RenumberTips
#' @export
RenumberTips.phylo <- function(tree, tipOrder) {
  startOrder <- tree[["tip.label"]]
  newOrder <- if (is.numeric(tipOrder)) {
    startOrder[tipOrder]
  } else {
    TipLabels(tipOrder, single = TRUE)
  }
  if (!identical(startOrder, newOrder)) {
    
    if (any(duplicated(newOrder))) {
      stop("Tree labels ", paste0(newOrder[duplicated(newOrder)], collapse = ", "),
           " repeated in `tipOrder`")
    }
    
    if (length(startOrder) != length(newOrder)) {
      startOnly <- setdiff(startOrder, newOrder)
      newOnly <- setdiff(newOrder, startOrder)
      if (length(startOnly)) {
        stop("Tree labels and tipOrder must match.",
             if (length(newOnly)) "\n  Missing in `tree`: ",
             paste0(newOnly, collapse = ", "),
             if (length(startOnly)) "\n  Missing in `tipOrder`: ",
             paste0(startOnly, collapse = ", ")
             )
      }
      newOrder <- intersect(newOrder, startOrder)
    }
  
    nTip <- length(startOrder)
    child <- tree[["edge"]][, 2]
    tips <- child <= nTip
  
    matchOrder <- match(startOrder, newOrder)
    if (any(is.na(matchOrder))) {
      stop("Tree labels ", paste0(startOrder[is.na(matchOrder)], collapse = ", "),
           " missing from `tipOrder`")
    }
    tree[["edge"]][tips, 2] <- matchOrder[tree[["edge"]][tips, 2]]
    tree[["tip.label"]] <- newOrder
    
    if (attr(tree, "order") == "preorder") {
      attr(tree, "order") <- "cladewise"
    }
  }
  tree
}

#' @rdname RenumberTips
#' @export
RenumberTips.Splits <- function(tree, tipOrder) {
  if (is.character(tipOrder)) {
    as.Splits(tree, tipOrder)
  } else if (is.numeric(tipOrder)) {
    as.Splits(tree, TipLabels(tree)[tipOrder])
  }
}

#' @rdname RenumberTips
#' @export
RenumberTips.multiPhylo <- function(tree, tipOrder) {
  at <- attributes(tree)
  labelled <- !is.null(at[["TipLabel"]])

  if (!labelled) {
    # Unlabelled: each tree may have different tip orderings,
    # so must process individually
    tree <- lapply(tree, RenumberTips.phylo, tipOrder)
    attributes(tree) <- at
    return(tree)
  }

  startOrder <- at[["TipLabel"]]
  newOrder <- if (is.numeric(tipOrder)) {
    startOrder[tipOrder]
  } else {
    TipLabels(tipOrder, single = TRUE)
  }

  nTip <- length(startOrder)
  if (identical(startOrder, newOrder)) {
    # Tip numbering already correct; use identity permutation to ensure
    # each tree materializes its own tip.label element
    matchOrder <- seq_len(nTip)
  } else {
    if (any(duplicated(newOrder))) {
      stop("Tree labels ",
           paste0(newOrder[duplicated(newOrder)], collapse = ", "),
           " repeated in `tipOrder`")
    }

    if (length(startOrder) != length(newOrder)) {
      startOnly <- setdiff(startOrder, newOrder)
      newOnly <- setdiff(newOrder, startOrder)
      if (length(startOnly)) {
        stop("Tree labels and tipOrder must match.",
             if (length(newOnly)) "\n  Missing in `tree`: ",
             paste0(newOnly, collapse = ", "),
             if (length(startOnly)) "\n  Missing in `tipOrder`: ",
             paste0(startOnly, collapse = ", ")
        )
      }
      newOrder <- intersect(newOrder, startOrder)
    }

    matchOrder <- match(startOrder, newOrder)
    if (any(is.na(matchOrder))) {
      stop("Tree labels ",
           paste0(startOrder[is.na(matchOrder)], collapse = ", "),
           " missing from `tipOrder`")
    }
  }

  tree <- .Call(`_TreeTools_renumber_tips_batch`, tree, matchOrder, nTip,
                newOrder)

  at[["TipLabel"]] <- newOrder
  attributes(tree) <- at
  tree
}

#' @rdname RenumberTips
#' @export
RenumberTips.list <- function(tree, tipOrder) {
  lapply(tree, RenumberTips, tipOrder)
}

#' @rdname RenumberTips
#' @export
RenumberTips.NULL <- function(tree, tipOrder) NULL

test_that("TopologyOnly() works", {
  topol <- as.phylo(10, 5)
  tree <- topol
  tree[["edge.length"]] <- topol[["edge"]][, 1]
  tree[["node.label"]] <- seq_len(topol$Nnode)
  
  # .phylo
  expect_equal(TopologyOnly(tree), topol)
  expect_equal(TopologyOnly(Postorder(tree)), topol)
  expect_equal(Preorder(Postorder(tree), topologyOnly = TRUE), topol)
  
  # .list
  expect_equal(TopologyOnly(list(tree, tree)), list(topol, topol))
  expect_equal(Preorder(Postorder(list(tree, tree)), topologyOnly = TRUE),
               list(topol, topol))
  
  # .multiPhylo
  expect_equal(TopologyOnly(c(tree, tree)), c(topol, topol))
  expect_equal(Preorder(Postorder(c(tree, tree)), topologyOnly = TRUE),
               structure(c(topol, topol), order = "preorder"))
  
  # .Splits
  expect_equal(TopologyOnly(as.Splits(tree)), as.Splits(topol))
  
  # .NULL
  expect_null(TopologyOnly(NULL))
})

TestFile <- function(filename = "") {
  system.file("extdata", "tests", filename, package = "TreeTools")
}

test_that("ReadTntCharacter()", {
  testFile <- TestFile("tnt-trees-and-matrix.tnt")
  expect_equal(
    structure(c("0", "0", "1", "1", "1", "1", "-", "-", "0", "0",
                "0", "0", "-", "-", "0", "0", "1", "1", "-", "-", "-", "-", "0",
                "1", "-", "-", "1", "1", "1", "0", "-", "-", "0", "0", "1", "-",
                "-", "-", "0", "1", "0", "-"), .Dim = 6:7,
              .Dimnames = list(c("a", "b", "c", "d", "e", "f"), NULL)),
    ReadTntCharacters(testFile)
  )
  
  dnaTest <- TestFile("tnt-dna.tnt")
  expect_equal(ReadTntCharacters(dnaTest),
               cbind(ReadTntCharacters(dnaTest, type = "num"),
                     ReadTntCharacters(dnaTest, type = "dna")))
  expect_equal(ReadTntCharacters(dnaTest),
               ReadTntCharacters(dnaTest, type = c("NUM", "Dna")))
  expect_message(expect_null(ReadTntCharacters(dnaTest, type = "NONE")))
  
  expectedPhyDat <- structure( # Generated by phangorn::as.phyDat
    list(a = c(1L, 3L, 3L, 3L, 3L, 3L, 3L),
         b = c(1L, 3L, 3L, 3L, 3L, 3L, 3L),
         c = c(2L, 1L, 1L, 3L, 2L, 1L, 1L),
         d = c(2L, 1L, 1L, 3L, 2L, 1L, 2L),
         e = c(2L, 1L, 2L, 1L, 2L, 2L, 1L),
         f = c(2L, 1L, 2L, 2L, 1L, 3L, 3L)),
    weight = c(1L, 1L, 1L, 1L, 1L, 1L, 1L),
    nr = 7L, nc = 3L, index = 1:7,
    levels = c("-", "0", "1"),
    allLevels = c("0", "1", "-"),
    type = "USER",
    contrast = structure(c(0, 0, 1, 1, 0, 0, 0, 1, 0), dim = c(3L, 3L),
                         dimnames = list(NULL, c("-", "0", "1"))),
    class = "phyDat")
  expect_equal(ReadTntAsPhyDat(testFile), expectedPhyDat)
})

test_that("TntTextToTree()", {
  expect_equal(TNTText2Tree("(A (B (C (D E ))));"),
               ape::read.tree(text = "(A, (B, (C, (D, E))));"))
})

test_that("ReadTntTree() NULL return", {
  expect_null(ReadTntTree(TestFile("ape-tree.nex")))
})

test_that("TNT trees parsed correctly", {
  expect_message(expect_warning(expect_warning(
    trees <- ReadTntTree(
      filepath = TestFile("tnt-tree.tre"),
      relativePath = TestFile()
    ),
    "Multiple tree blocks"),
    "Expected `ttags \\*N`; applying tags to first tree"),
    "Ignoring tag on leaf 1: .Origin."
  )
  expect_equal(length(trees), 3)
  expect_equal(trees[[3]]$tip.label[2], "Novocrania")
  expect_equal(trees[[3]][["node.label"]][37:41 - NTip(trees[[3]])],
               c("Nov+1", "Nov+2", "Nov+3",
                 "Nov +4", # Concatenated from Nov & +4
                 "Nov+5"))
  expect_equal(trees[[3]][["node.label"]][69 - NTip(trees[[3]])], "Nis+1")
  
  tipLabels <- c("Dailyatia", "Novocrania", "Craniops", "Ussunia", "Gasconsia",
                 "Heliomedusa_orienta", "Micrina", "Mickwitzia_muralensis",
                 "Micromitra", "Askepasma_toddense", "Pelagodiscus_atlanticus",
                 "Lingula", "Eoobolus", "Clupeafumosus_socialis", "Phoronis",
                 "Eccentrotheca", "Yuganotheca_elegans",
                 "Longtancunella_chengjiangensis", "Paterimitra",
                 "Lingulellotreta_malongensis", "Acanthotretella",
                 "Lingulosacculus", "Pedunculotheca_diania",
                 "Haplophrentis_carinatus", "Tomteluva_perturbata",
                 "Salanygolina", "Mummpikia_nuda", "Alisina", "Coolinia_pecten",
                 "Antigonambonites_planus", "Kutorgina_chengjiangensis",
                 "Nisusia_sulcata", "Glyptoria", "Orthis", "Terebratulina")
  expect_warning(
    fromLabels <- suppressMessages(
      ReadTntTree(TestFile("tnt-tree.tre"), tipLabels = tipLabels)),
    "Expected `ttags .N`;"
  )
  expect_identical(trees, fromLabels)
  
  trees[[2]][["node.label"]] <- NULL
  trees[[3]][["node.label"]] <- NULL
  expect_equal(trees[[2]], trees[[3]])
  expect_equal(ConsensusWithout(trees, "Paterimitra")$Nnode, 32)
  
  
  namedLabels <- ReadTntTree(TestFile("tnt-namedtree.tre"))[[1]]$tip.label
  expect_equal("Flustra_sp.", namedLabels[1])
  expect_equal(74L, length(namedLabels))
  
  tam <- ReadTntTree(TestFile("tnt-trees-and-matrix.tnt"))
  expect_equal(length(tam), 4L)
  expect_equal(tam[[1]],
               TNTOrder(
                 RenumberTips(
                   ape::read.tree(text = "(a, (b, (c, (f, (d, e )))));"),
                   letters[1:6])
                 )
               )
  expect_equal(tam[[4]],
               TntOrder(ape::read.tree(text = "(a, (b, (c, (d, e, f))));")))
  
  
  oldWD <- getwd()
  setwd(system.file(package = "TreeTools"))
  expect_equal(paste0("taxon_", letters[1:5]),
               ReadTntTree("extdata/output/numbered.tre", "./extdata", 2)[[1]]$tip.label)
  
  expect_equal(paste0("taxon_", letters[1:5]),
               ReadTntTree("extdata/output/named.tre")$tip.label)
  setwd(oldWD)
})

test_that("ReadTntTree() reads bare tree", {
  bareTree <- TestFile("tnt-bare-tree.tnt")
  expect_warning(
    expect_equal(ReadTntTree(bareTree),
                 ReadTntTree(bareTree, tipLabels = as.character(0:5))),
    "does not link to taxon names")
})

test_that("ReadTntTree() follows TNT node numbering conventions", {
  a..e <- paste("taxon", letters[1:5], sep = "_")
  namedTree <- system.file("extdata/output/named.tre", package = "TreeTools")
  expect_equal(
    ReadTntTree(namedTree),
    ReadTntTree(namedTree, tipLabels = a..e)
  )
  expect_equal(ReadTntTree(namedTree)$edge,
               Postorder(ReadTntTree(namedTree))$edge)
  
  expect_equal(
    # NB this is not a perfect test, as the order in which edges are listed
    # is not stipulated by the ReadTntTree documentation.
    ReadTntTree(namedTree, tipLabels = rev(a..e))$edge,
    cbind(rep(c(7, 8, 9, 6), each = 2), c(2, 1, 7, 3, 8, 4, 9, 5))
  )
})

nastyEdge <- structure(c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
                         5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
                       .Dim = c(12, 2))
nasty <- structure(list(edge = nastyEdge, Nnode = 5L, tip.label = letters[1:8]),
                   class = "phylo")

expect_postorder <- function(edge) {
  parent <- edge[, 1]
  child <- edge[, 2]
  visited <- logical(max(parent))
  root <- min(parent)
  for (i in seq_along(parent)) {
    visited[parent[i]] <- TRUE
    expect_true(child[i] < root || visited[child[i]])
  }
  expect_equal(visited, c(logical(root - 1), !logical(max(parent) + 1 - root)))
}

test_that("RenumberTree() fails safely", {
  expect_error(RenumberTree(1:3, 1:4), "Length.*must match")
  expect_error(RenumberTree(1:3, 1:4, 5:6), "Length.*match")
  expect_error(RenumberTree(1:4, 1:4, 5:6), "Length mismatch")
})


test_that("RenumberTree() handles polytomies", {
  tr <- ape::read.tree(text = "(a, (b, d, c));")
  edge <- tr$edge
  parent <- edge[, 1]
  child <- edge[, 2]

  ret <- RenumberTree(parent, child)
  expect_equal(c(5, 5, 6, 6, 6), ret[, 1])
  expect_equal(c(1, 6, 2, 3, 4), ret[, 2])


  edge <- structure(c(6L, 7L, 5L, 7L, 6L, 5L,
                      2L, 5L, 3L, 6L, 1L, 4L),
                    .Dim = c(6L, 2L))

  # Must be in preorder; i.e. each node in left subtree before each node in
  # right subtree for each subtree
  #
  # Also, nodes should be rotated such that the lowest tip in a subtree
  # is always encountered first.
  #
  # These rules ensure a unique representation for any tree.
  expectation <- structure(c(5L, 6L, 6L, 5L, 7L, 7L,
                             6L, 1L, 2L, 7L, 3L, 4L),
                           .Dim = c(6L, 2L))
  expect_equal(RenumberTree(edge[, 1], edge[, 2]),
               expectation)

  expect_equal(c(9, 10, 10, 11, 11, 11, 10, 12, 12, 13, 13, 9,
                 10, 1, 11,  2,  4,  7, 12,  3, 13,  5,  6, 8),
               as.integer(RenumberTree(nasty$edge[, 1], nasty$edge[, 2])))
})

test_that("RenumberTree() handles singles", {
  withSingles <- ape::read.tree(text = "(a, (b, (c), (((d), (e)))));")
  expect_equal(as.integer(Preorder(withSingles)$edge),
               c(6, 6, 7, 7, 8, 7, 9, 10, 11, 10, 12,
                 1, 7, 2, 8, 3, 9, 10, 11, 4, 12, 5))
})

test_that("Replacement reorder functions work correctly", {
  ## Tree
  tree <- ape::read.tree(text = "((((((1,2),3),4),5),6),(7,(8,(9,(10,(11,12))))));")
  expect_equal(ape::reorder.phylo(tree, "cladewise"), Cladewise(tree))
  expect_equal(ape::reorder.phylo(tree, "pruningwise"), Pruningwise(tree))

  post6 <- Postorder(BalancedTree(6))$edge
  expect_postorder(post6)
  parent6 <- post6[, 1]
  child6 <- post6[, 2]

  # Order of tip pairs is arbitrary\
  expect_equal(1:2, sort(child6[parent6 == 9]))
  expect_equal(4:5, sort(child6[parent6 == 11]))
  expect_equal(c(6, 11), sort(child6[parent6 == 10]))
  expect_equal(c(3, 9), sort(child6[parent6 == 8]))
  expect_equal(c(8, 10), sort(child6[parent6 == 7]))

  star <- ape::read.tree(text = "(a, b, d, c);")
  edge <- RenumberTips(star, letters[1:4])$edge
  expect_equal(edge,
               RenumberTips(star, ape::read.tree(text = "(a, b, c, d);"))$edge)
  expect_equal(star$edge, RenumberTree(edge[, 1], edge[, 2]))
  expect_equal(list(star$edge[, 1], star$edge[, 2]),
               RenumberEdges(edge[, 1], edge[, 2]))
})

test_that("RenumberTips() with numeric tipOrder", {
  tr <- BalancedTree(6)
  order <- c(6, 5, 4, 2, 3, 1)
  expect_equal(RenumberTips(tr, order), RenumberTips(tr, TipLabels(order)))
})

test_that("RenumberTips() handles misspecification", {
  expect_error(RenumberTips(BalancedTree(8), paste0("t", 0:5)),
               "Missing in `tree`: t0.*Missing in `tipOrder`: t6, t7, t8")
  expect_error(RenumberTips(BalancedTree(8), c(1, 1:8)),
               "Tree labels t1 repeated in `tipOrder`")
})

test_that("RenumberTips() works correctly", {
  abcd <- letters[1:4]
  dcba <- letters[4:1]
  bal7b <- BalancedTree(dcba)
  bsp7b <- as.Splits(bal7b)
  bal7f <- BalancedTree(abcd)
  bsp7f <- as.Splits(bal7f)
  pec7f <- PectinateTree(abcd)
  psp7f <- as.Splits(bal7f)
  pec7b <- PectinateTree(dcba)
  psp7b <- as.Splits(bal7b)

  l7 <- list("bal7b" = bal7b, "bal7f" = bal7f, "pec7f" = pec7f)
  f7 <- list(bal7f, bal7f, pec7f)
  b7 <- list(bal7b, bal7b, pec7b)
  mp7 <- structure(l7, class = "multiPhylo")

  expect_true(all.equal(unname(RenumberTips(l7, dcba)), b7))
  expect_true(all.equal(unname(RenumberTips(l7, 4:1)), b7))
  expect_true(all.equal(unname(RenumberTips(l7, abcd)), f7))
  expect_true(all.equal(unname(RenumberTips(l7, 4:1)), f7))
  
  sl7 <- list("bsp7b" = bsp7b, "bsp7f" = bsp7f, "psp7f" = psp7f)
  sf7 <- list(bsp7f, bsp7f, psp7f)
  sb7 <- list(bsp7b, bsp7b, psp7b)
  
  expect_splits_same <- function(x, y) {
    if (is.list(x)) {
      for (i in seq_along(x)) {
        expect_splits_same(x[[i]], y[[i]])
      }
    } else {
      expect_equal(TipLabels(x), TipLabels(y))
      expect_equal(PolarizeSplits(x, 1), PolarizeSplits(y, 1))
    }
  }
  
  expect_splits_same(RenumberTips(bsp7b, abcd), bsp7f)
  expect_splits_same(RenumberTips(sl7, dcba), sb7)
  expect_splits_same(RenumberTips(sl7, abcd), sf7)
  

  expect_true(all.equal(structure(f7, class = "multiPhylo"),
                        unname(RenumberTips(mp7, abcd))))
  expect_true(all.equal(structure(b7, class = "multiPhylo"),
                        unname(RenumberTips(mp7, dcba))))
  
  expect_null(RenumberTips(NULL))

  expect_equal(Preorder(RenumberTips(l7, c("extra tip", letters[1:5]))),
               Preorder(RenumberTips(l7, letters[1:4])))
  expect_error(RenumberTips(l7, letters[2:5]))
  
  
  expect_equal(
    attr(
      RenumberTips(structure(b7, TipLabel = dcba, class = "multiPhylo"), abcd),
      "TipLabel"
    ),
    abcd
  )
    
})

test_that("RenumberTips.multiPhylo() covers edge cases", {
  mp8 <- structure(
    list(BalancedTree(8), PectinateTree(8)),
    class = "multiPhylo"
  )

  # Numeric tipOrder
  result <- RenumberTips(mp8, 8:1)
  expect_equal(TipLabels(result[[1]]), paste0("t", 8:1))

  # No-op when order already matches (unlabelled: per-tree fallback)
  result_noop <- RenumberTips(mp8, TipLabels(mp8[[1]]))
  expect_equal(result_noop[[1]][["edge"]], mp8[[1]][["edge"]])
  expect_equal(result_noop[[2]][["edge"]], mp8[[2]][["edge"]])

  # Duplicate error
  expect_error(RenumberTips(mp8, rep("t1", 8)), "repeated")

  # Length mismatch error
  expect_error(RenumberTips(mp8, paste0("t", 0:5)),
               "Missing in `tree`.*Missing in `tipOrder`")

  # Missing label error
  mp_shared <- structure(
    list(BalancedTree(8), PectinateTree(8)),
    TipLabel = paste0("t", 1:8),
    class = "multiPhylo"
  )
  expect_error(
    RenumberTips(mp_shared, c(paste0("t", 1:7), "t_unknown")),
    "missing from `tipOrder`"
  )
})

test_that("RenumberTips.multiPhylo() materializes tip.label", {
  # TipLabel-labelled multiPhylo: individual trees should gain tip.label
  # even when the order already matches (needed by RoguePlot and others
  # that later remove TipLabel)
  labs <- paste0("t", 1:6)
  mp_labelled <- structure(
    list(BalancedTree(6), PectinateTree(6)),
    TipLabel = labs,
    class = "multiPhylo"
  )
  result <- RenumberTips(mp_labelled, labs)
  attr(result, "TipLabel") <- NULL
  # After removing TipLabel, individual trees should still have tip.label
  expect_equal(result[[1]][["tip.label"]], labs)
  expect_equal(result[[2]][["tip.label"]], labs)
})

test_that("RenumberTips.multiPhylo() preserves phylo class on elements", {
  labs <- paste0("t", 1:6)
  mp <- structure(
    list(BalancedTree(6), PectinateTree(6)),
    TipLabel = labs,
    class = "multiPhylo"
  )
  result <- RenumberTips(mp, rev(labs))
  # Each element should retain class "phylo", not be downgraded to "list"
  expect_s3_class(.subset2(result, 1), "phylo")
  expect_s3_class(.subset2(result, 2), "phylo")
})

test_that("RenumberTips.multiPhylo() handles unlabelled different orderings", {
  # Trees with different tip orderings in an unlabelled multiPhylo
  tree_abc <- read.tree(text = "(a, (b, c));")
  tree_cab <- read.tree(text = "(c, (a, b));")
  mp_mixed <- structure(list(tree_abc, tree_cab), class = "multiPhylo")

  result <- RenumberTips(mp_mixed, c("a", "b", "c"))
  per_tree <- lapply(mp_mixed, RenumberTips, c("a", "b", "c"))

  expect_equal(result[[1]][["edge"]], per_tree[[1]][["edge"]])
  expect_equal(result[[2]][["edge"]], per_tree[[2]][["edge"]])
  expect_equal(result[[1]][["tip.label"]], c("a", "b", "c"))
  expect_equal(result[[2]][["tip.label"]], c("a", "b", "c"))
})

test_that("RenumberTips.multiPhylo() batch matches per-tree", {
  set.seed(7429)
  trees <- c(
    replicate(50, RandomTree(12, root = TRUE), simplify = FALSE),
    replicate(50, PectinateTree(12), simplify = FALSE)
  )
  class(trees) <- "multiPhylo"
  target <- sort(TipLabels(trees[[1]]))

  batch <- RenumberTips(trees, target)
  per_tree <- structure(
    lapply(trees, RenumberTips.phylo, target),
    class = "multiPhylo"
  )

  for (i in seq_along(batch)) {
    expect_equal(batch[[i]][["edge"]], per_tree[[i]][["edge"]],
                 info = paste("tree", i))
    expect_equal(batch[[i]][["tip.label"]], per_tree[[i]][["tip.label"]],
                 info = paste("tree", i))
  }
})

test_that("postorder_order() works", {
  edg7 <- BalancedTree(7)$edge
  expect_postorder(edg7[postorder_order(edg7), ])
  
  test2 <- edg7[c(1:4, 9, 8, 12, 11, 10, 7:5), ]
  expect_postorder(test2[postorder_order(test2), ])
  
  nastyEdge <- structure(c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
                           5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
                         .Dim = c(12, 2))
  expect_postorder(nastyEdge[postorder_order(nastyEdge), ])
  
  poly <- ape::read.tree(text = "((a, b, c), (d, (e, f)));")$edge
  expect_postorder(poly[postorder_order(poly), ])
  
  star <- ape::read.tree(text = "(a, b, c, d, e, f);")$edge
  expect_postorder(star[postorder_order(star), ])
  
  expect_error(PostorderOrder(1:5), "edge matrix of a `phylo` obj")
  expect_error(PostorderOrder(matrix(1, 3, 3)), "edge matrix of a `phylo` obj")
  expect_equal(PostorderOrder(BalancedTree(4)),
               rev(seq_len(nrow(BalancedTree(4)$edge))))
  expect_equal(PostorderOrder(Postorder(BalancedTree(4))),
               seq_len(nrow(BalancedTree(4)$edge)))
  expect_postorder(nastyEdge[PostorderOrder(nastyEdge), ])
  expect_postorder(nastyEdge[PostorderOrder(nasty), ])
  
})

test_that("Reorder methods work correctly", {
  bal7 <- BalancedTree(7)
  bal7$edge.length <- 1:12 * 10
  attr(bal7, "order") <- NULL
  pec7 <- PectinateTree(7)
  list7 <- list(bal7, pec7)
  stt <- SingleTaxonTree(1)
  bad <- bal7
  bad$Nnode <- 100
  attr(bad, "order") <- NULL
  mp7 <- structure(list7, class = "multiPhylo")
  
  Test <- function(Method, ..., testEdges = TRUE) {
    expect_identical(Method(bal7, ...), Method(list7, ...)[[1]])
    expect_identical(Method(pec7, ...), Method(mp7, ...)[[2]])
    expect_true(all.equal(Method(stt), stt))
    expect_identical(Method(bal7), Method(Method(bal7)))
    expect_equal(Method(bal7),
                 Method(Preorder(Postorder(Cladewise(Pruningwise(bal7))))))
    if (testEdges) expect_equal(Method(bal7)$edge, Method(bal7$edge))
    expect_error(Method(10))
    expect_error(Method(1:2))
    expect_error(Method(matrix("one")))
    expect_null(Method(NULL))
  }
  
  Test(ApePostorder, testEdges = FALSE)
  expect_error(ApePostorder(bad))

  Test(Postorder, testEdges = FALSE) # Different order if edge lengths present
  expect_postorder(Postorder(nastyEdge))

  Test(Cladewise)
  expect_error(Cladewise(bad))

  Test(Preorder)
  
  Test(TntOrder, testEdges = FALSE)

  Test(Pruningwise, testEdges = FALSE)
  expect_error(Pruningwise(bad))

})

test_that("Preorder() gives identical output", {
  tree1 <- BalancedTree(1:9)
  numbers <- c("One", "Two", "Three", "Four", "Five", "Six",
               "Seven", "Eight", "Nine")
  tree2 <- BalancedTree(numbers[1:9])
  # Force pre-order with NEW tip labels
  tree2 <- Preorder(Postorder(RenumberTips(tree2, sort(numbers))))
  tree2$tip.label <- as.character(match(sort(numbers), numbers))
  expect_true(all.equal(tree1, tree2))
  expect_false(identical(tree1, tree2))
  
  # Now preorder after renumbering.
  # Renumbering should drop the previous preorder attribute.
  pre2 <- Preorder(RenumberTips(tree2, as.character(1:9)))
  expect_equal(tree1, pre2)
  expect_identical(tree1, pre2)
})


test_that("Reorder methods retain edge weights", {
  bal7 <- BalancedTree(7)
  bal7$edge.length <- 1:12 * 10
  attr(bal7, "order") <- NULL
  expect_equal(Preorder(bal7)[["edge.length"]],
               bal7$edge.)
  expect_equal(Preorder(Cladewise(bal7))[["edge.length"]],
               bal7$edge.)
  expect_equal(Preorder(Postorder(bal7))[["edge.length"]],
               bal7$edge.)
})

test_that("Reorder methods retain node labels", {
  bal7 <- BalancedTree(7)
  startLabels <- paste("Node", 8:13)
  bal7[["node.label"]] <- startLabels
  attr(bal7, "order") <- NULL
  if (interactive()) {
    plot(bal7, show.node.label = TRUE, xpd = NA)
  }
  expect_equal(Preorder(bal7)[["node.label"]], startLabels)
  expect_equal(Cladewise(bal7)[["node.label"]], startLabels)
  expect_equal(Preorder(Cladewise(bal7))[["node.label"]], startLabels)
  expect_equal(Postorder(bal7)[["node.label"]], startLabels)
  expect_equal(Preorder(Postorder(bal7))[["node.label"]], startLabels)
})

test_that("Malformed trees don't cause crashes", {
  treeDoubleNode <- read.tree(text = "((((((1,2)),3),4),5),6);")
  treePolytomy   <- read.tree(text = "((((1,2,3),4),5),6);")
  treeDoublyPoly <- read.tree(text = "(((((1,2,3)),4),5),6);")
  nasty <- structure(list(edge = structure(
    c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
      5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
    .Dim = c(12, 2)),
    Nnode = 5L,
    tip.label = letters[1:8]),
    class = "phylo") # Danger: Do not plot!

  reordered <- Preorder(treeDoubleNode)$edge
  expect_equal(11L, dim(reordered)[1])
  expect_equal(5L, sum(tabulate(reordered[, 1]) == 2L))

  postordered <- Postorder(treeDoubleNode)$edge
  expect_equal(11L, dim(postordered)[1])
  expect_equal(5L, sum(tabulate(postordered[, 1]) == 2L))


  reordered <- Preorder(treePolytomy)$edge
  expect_equal(9L, dim(reordered)[1])
  expect_equal(c(2L, 2L, 2L, 3L), as.integer(table(reordered[, 1])))

  reordered <- Postorder(treePolytomy)$edge
  expect_equal(9L, dim(reordered)[1])
  expect_equal(c(2L, 2L, 2L, 3L), as.integer(table(reordered[, 1])))


  reordered <- Preorder(treeDoublyPoly)$edge
  expect_equal(10L, dim(reordered)[1])
  expect_equal(c(2L, 2L, 2L, 1L, 3L), as.integer(table(reordered[, 1])))

  reordered <- Postorder(treeDoublyPoly)$edge
  expect_equal(10L, dim(reordered)[1])
  expect_equal(c(2L, 2L, 2L, 1L, 3L), as.integer(table(reordered[, 1])))

  #C <- 0
  #plot(Preorder(nasty)); nodelabels(c(12, 10, 13, 11, 9) - C); tiplabels(1:8 - C)
  #edgelabels(c(2, 3, 6, 4, 8, 10, 9, 5, 7, 1, 12, 11) - C)
  reordered <- Preorder(nasty)$edge
  expect_equal(12L, dim(reordered)[1])
  # Nodes renumbered
  expect_equal(c(2L, 3L, 3L, 2L, 2L), tabulate(reordered[, 1])[9:13])

  reordered <- Postorder(nasty)$edge
  expect_equal(12L, dim(reordered)[1])
  expect_postorder(reordered)
})

#' Generate pectinate, balanced or random trees
#'
#' `RandomTree()`, `PectinateTree()`, `BalancedTree()` and `StarTree()`
#' generate trees with the specified shapes and leaf labels.
#'
#' @param tips An integer specifying the number of tips, or a character vector
#' naming the tips, or any other object from which [`TipLabels()`] can
#' extract leaf labels.
#'
#' @return
#' Each function returns an unweighted binary tree of class `phylo` with
#' the specified leaf labels. Trees are rooted unless `root = FALSE`.
#'
#' @family tree generation functions
#'
#' @template MRS
#' @references \insertAllCited()
#' @name GenerateTree
NULL

#' @rdname GenerateTree
#'
#' @param root Character or integer specifying tip to use as root;
#' or `TRUE` to root the tree on a random edge;
#' or `FALSE` to return an unrooted tree.
#' @param nodes Number of nodes to generate.  The default and maximum,
#' `tips - 1`, generates a binary tree; setting a lower value will induce
#' polytomies.
#'
#' @return `RandomTree()` returns a topology drawn at random from the uniform
#' distribution (i.e. each binary tree is drawn with equal probability).
#' Trees are generated by inserting
#' each tip in term at a randomly selected edge in the tree.
#' Random numbers are generated using a Mersenne Twister.
#' If `root = FALSE`, the tree will be unrooted, with the first tip in a
#' basal position.  Otherwise, the tree will be rooted on `root`.
#'
#' @examples
#' RandomTree(LETTERS[1:10])
#'
#' data("Lobo")
#' RandomTree(Lobo.phy)
#'
#' @export
RandomTree <- function(tips, root = FALSE, nodes, lengths = NULL) {
  tips <- TipLabels(tips)
  nTips <- length(tips)
  if (length(root) > 1L) {
    root <- root[[1]]
    warning("More than one entry in `root`; using ", root)
  }
  if (is.na(root)) {
    warning("Treating `root = NA` as `FALSE`")
    root <- FALSE
  }
  unrooted <- !is.logical(root) || root == FALSE
  nodesInBinary <- nTips - ifelse(unrooted, 2L, 1L)
  if (missing(nodes)) {
    nodes <- nodesInBinary
  }
  if (nodes > nodesInBinary) {
    warning("`nodes` higher than number in binary tree. Ignoring.")
  }
  if (nTips < 3) {
    return(BalancedTree(tips = tips, lengths = lengths))
  }
  if (nodes < 1L) {
    stop("A tree must contain one or more `nodes`")
  }

  edge <- do.call(cbind,
                  RenumberEdges(.RandomParent(nTips),
                                seq_len(nTips + nTips - 2L)))
  # If root identifies a specific tip, root edge on that tip
  if (!is.logical(root) && root != 1L) {
    if (is.character(root)) {
      root <- which(tips == root)
    }
    if (length(root) == 0L) {
      stop("No match found for `root`")
    }
    if (!is.integer(root)) {
      root <- as.integer(root)
    }
    edge <- root_binary(edge, root)
  }
  tree <- structure(list(edge = edge,
                         Nnode = nTips - 1L,
                         tip.label = tips),
                    order = "preorder",
                    class = "phylo")

  # If root is simply a logical T/F, then root our tree randomly - or don't!
  if (is.logical(root)) {
    if (root) {
      tree <- root_on_node(tree, sample.int(nTips + nodes, 1))
    } else {
      tree <- UnrootTree(tree)
    }
  }

  # Collapse nodes at random to obtain desired resolution
  if (nodes < nodesInBinary) {
    tree <- CollapseNode(tree,
                         nTips + 1L + sample.int(nodesInBinary - 1L,
                                                 tree[["Nnode"]] - nodes))
  }
  
  if (!is.null(lengths)) {
    tree[["edge.length"]] <- rep(lengths, length.out = dim(tree[["edge"]])[[1]])
  }
  
  # Return:
  tree
}

#' @rdname GenerateTree
#'
#' @param addInTurn Logical specifying whether to add leaves in the order of
#' `tips`.  If `FALSE`, leaves will be added in a random order.
#' @return `YuleTree()` returns a topology generated by the Yule process
#' \insertCite{Steel2001}{TreeTools},
#' i.e. adding leaves in turn adjacent to a randomly-chosen existing leaf.
#' @examples
#' YuleTree(LETTERS[1:10])
#'
#' @export
YuleTree <- function(tips, addInTurn = FALSE, root = TRUE, lengths = NULL) {
  tips <- TipLabels(tips)
  if (!addInTurn[[1]]) {
    tips <- sample(tips)
  }
  nTips <- length(tips)
  if (any(is.na(root))) {
    warning("treating `root = NA` as `FALSE`")
    root[is.na(root)] <- FALSE
  }
  
  if (nTips < 1) {
    return(ZeroTaxonTree())
  } else if (nTips == 1) {
    return(SingleTaxonTree(tips, lengths))
  }
  
  tree <- BalancedTree(tips[1:2])
  if (nTips > 2) for (i in 3:nTips) {
    tree <- AddTip(tree, sample.int(i - 1, 1), tips[[i]])
  }
  
  unrooted <- !is.logical(root) || root == FALSE
  if (unrooted) {
    tree <- UnrootTree(tree)
  }
  
  if (!is.null(lengths)) {
    tree[["edge.length"]] <- rep(lengths, length.out = dim(tree[["edge"]])[[1]])
  }
  
  # Return:
  tree
}

#' Random parent vector
#'
#' @param n Integer specifying number of leaves.
#' @param seed (Optional) Integer with which to seed Mersenne Twister random
#' number generator in C++.
#'
#' @return Integer vector corresponding to the "parent" entry of
#' `tree[["edge"]]`, where the "child" entry, i.e. column 2, is numbered
#' sequentially from `1:n`.
#' @template MRS
#' @keywords internal
#' @importFrom stats runif
#' @export
.RandomParent <- function(n, seed = sample.int(2147483647L, 1L)) {
  if (!is.numeric(n) || is.na(n) || n[1] < 3) {
    stop("nTip must be > 2");
  }
  random_parent(as.integer(n), as.integer(seed))
}

#' @rdname GenerateTree
#' @return `PectinateTree()` returns a pectinate (caterpillar) tree.
#' @examples
#' plot(PectinateTree(LETTERS[1:10]))
#'
#' @export
PectinateTree <- function(tips, lengths = NULL) {
  tips <- TipLabels(tips)
  nTip <- length(tips)
  if (nTip == 0) {
    ZeroTaxonTree()
  } else if (nTip == 1) {
    SingleTaxonTree(tips, lengths)
  } else {
  
    nEdge <- nTip + nTip - 2L
    tipSeq <- seq_len(nTip - 1L)
  
    parent <- rep(seq_len(nTip - 1L) + nTip, each = 2L)
  
    child <- integer(nEdge)
    child[tipSeq + tipSeq - 1L] <- tipSeq
    child[tipSeq + tipSeq] <- tipSeq + nTip + 1L
    child[nEdge] <- nTip
  
    tr <- list(
        edge = matrix(c(parent, child), ncol = 2L),
        Nnode = nTip - 1L,
        tip.label = tips
      )
    if (!is.null(lengths) && nTip > 1) {
      tr[["edge.length"]] <- rep(lengths, length.out = 2 * (nTip - 1))
    }
    structure(tr, order = "cladewise", class = "phylo")
  }
}


#' @rdname GenerateTree
#'
#' @inheritParams SingleTaxonTree
#' @return `BalancedTree()` returns a balanced (symmetrical) tree, in preorder.
#'
#' @examples
#' plot(BalancedTree(LETTERS[1:10]))
#' @export
BalancedTree <- function(tips, lengths = NULL) {
  tips <- TipLabels(tips)
  nTip <- length(tips)
  if (nTip < 2L) {
    if (nTip == 1L) {
      SingleTaxonTree(tips, lengths)
    } else if (nTip == 0L) {
      ZeroTaxonTree()
    } else {
      NULL
    }
  } else {
    tr <- list(edge = .BalancedBit(seq_len(nTip), nTips = nTip),
               Nnode = nTip - 1L,
               tip.label = as.character(tips))
    if (!is.null(lengths)) {
      tr[["edge.length"]] <- rep(lengths, length.out = 2 * (nTip - 1))
    }
    # Return:
    structure(tr, order = "preorder", class = "phylo")
  }
}

#' @keywords internal
.BalancedBit <- function(tips, nTips = length(tips), rootNode = nTips + 1L) {
  if (nTips < 4L) {
    if (nTips == 2L) {
      matrix(c(rootNode, rootNode, tips), 2L, 2L)
    } else if (nTips == 3L) {
      matrix(c(rootNode + c(0L, 1L, 1L, 0L, 1L), tips), 4L, 2L)
    } else {
      tips
    }
  } else {
    # Recurse:
    firstN <- as.integer(ceiling(nTips / 2L))
    firstHalf <- seq_len(firstN)
    root2 <- rootNode + firstN
    rbind(rootNode + 0:1,
          .BalancedBit(tips[firstHalf], rootNode = rootNode + 1L),
          c(rootNode, root2),
          .BalancedBit(tips[-firstHalf], rootNode = root2))
  }
}

#' @rdname GenerateTree
#' @return `StarTree()` returns a completely unresolved (star) tree.
#' @examples
#' plot(StarTree(LETTERS[1:10]))
#'
#' @export
StarTree <- function(tips, lengths = NULL) {
  tips <- TipLabels(tips)
  nTip <- length(tips)

  parent <- rep.int(nTip + 1L, nTip)
  child <- seq_len(nTip)

  tr <- if (is.null(lengths)) {
    list(
      edge = matrix(c(parent, child), ncol = 2L),
      Nnode = 1L,
      tip.label = tips
    )
  } else {
    list(
      edge = matrix(c(parent, child), ncol = 2L),
      Nnode = 1L,
      tip.label = tips,
      edge.length =  rep(lengths, length.out = nTip)
    )
  }
  structure(tr, order = "cladewise", class = "phylo")
}

#' Generate a neighbour joining tree
#'
#' `NJTree()` generates a rooted neighbour joining tree from a phylogenetic
#' dataset.
#'
#' @template datasetParam
#' @param edgeLengths Logical specifying whether to include edge lengths.
#' @param ambig,ratio Settings of `ambig` and `ratio` to be used when
#' computing [`Hamming()`] distances between sequences.
#'
#' @return `NJTree` returns an object of class \code{phylo}.
#'
#' @examples
#' data("Lobo")
#' NJTree(Lobo.phy)
#'
#' @template MRS
#' @importFrom ape nj
#' @family tree generation functions
#' @export
NJTree <- function(dataset, edgeLengths = FALSE,
                    ratio = TRUE, ambig = "mean") {
  tree <- nj(Hamming(dataset, ratio = ratio, ambig = ambig))
  tree <- RootTree(tree, names(dataset)[[1]])
  if (!edgeLengths) {
    tree[["edge.length"]] <- NULL
  }
  tree
}

#' Hamming distance between taxa in a phylogenetic dataset
#'
#' The Hamming distance between a pair of taxa is the number of characters
#' with a different coding, i.e. the smallest number of evolutionary steps
#' that must have occurred since their common ancestor.
#'
#' Tokens that contain the inapplicable state are treated as requiring no steps
#' to transform into any applicable token.
#'
#' @param dataset Object of class `phyDat`.
#' @param ratio Logical specifying whether to weight distance against
#' maximum possible, given that a token that is ambiguous in either of two taxa
#' cannot contribute to the total distance between the pair.
#' @param ambig Character specifying value to return when a pair of taxa
#' have a zero maximum distance (perhaps due to a preponderance of ambiguous
#' tokens).
#' "median", the default, take the median of all other distance values;
#' "mean", the mean;
#' "zero" sets to zero; "one" to one;
#' "NA" to `NA_integer_`; and "NaN" to `NaN`.
#'
#' @return `Hamming()` returns an object of class `dist` listing the Hamming
#' distance between each pair of taxa.
#'
#' @seealso
#' Used to construct neighbour joining trees in [`NJTree()`].
#'
#' `dist.hamming()` in the \pkg{phangorn} package provides an alternative
#' implementation.
#' 
#' @examples 
#' tokens <- matrix(c(0, 0, "0", 0, "?",
#'                    0, 0, "1", 0, 1,
#'                    0, 0, "1", 0, 1,
#'                    0, 0, "2", 0, 1,
#'                    1, 1, "-", "?", 0,
#'                    1, 1, "2", 1, "{01}"),
#'                    nrow = 6, ncol = 5, byrow = TRUE,
#'                    dimnames = list(
#'                      paste0("Taxon_", LETTERS[1:6]),
#'                      paste0("Char_", 1:5)))
#'
#' dataset <- MatrixToPhyDat(tokens)
#' Hamming(dataset)
#' @template MRS
#' @family utility functions
#' @importFrom stats median
#' @importFrom utils combn
#' @export
Hamming <- function(dataset, ratio = TRUE,
                     ambig = c("median", "mean", "zero", "one", "na", "nan")) {
  at <- attributes(dataset)
  if (!inherits(dataset, "phyDat") || is.null(at[["contrast"]])) {
    stop("`dataset` must be a valid `phyDat` object")
  }
  contrast <- at[["contrast"]] > 0L
  if ("-" %in% colnames(contrast)) {
    # TODO This is debatable but perhaps acceptable behaviour.
    # Is the distance between {0, -} and {1} necessarily zero?
    contrast[contrast[, "-"], ] <- TRUE
  }
  weight <- at[["weight"]]
  tokens <- vapply(dataset, function(codings) contrast[codings, ],
                   matrix(NA, at[["nr"]], dim(contrast)[2]))
  hamming <- apply(combn(length(dataset), 2L), 2L, function(ij) {
    sum(weight[!apply(tokens[, , ij[1], drop = FALSE] &
                      tokens[, , ij[2], drop = FALSE], 1, any)])
  })

  if (ratio) {
    informative <- apply(!contrast, 1, any)
    nonAmbig <- .vapply(dataset, function(codings) informative[codings],
                       logical(at[["nr"]]))
    bothInformative <- apply(combn(length(dataset), 2L), 2L, function(ij) {
      sum(weight[nonAmbig[, ij[1]] & nonAmbig[, ij[2]]])
    })
    hamming <- hamming / bothInformative
    if (ambig[1] == 0) {
      ambig <- "zero"
    } else if (ambig[1] == 1) {
      ambig <- "one"
    }
    nanMethod <- pmatch(tolower(ambig[1]),
                        c("median", "mean", "zero", "one", "na", "nan"))
    if (is.na(nanMethod)) {
      stop("Invalid `ambig` value specified")
    }
    hamming[is.nan(hamming)] <- switch(nanMethod,
      median(hamming[!is.na(hamming)]),
      mean(hamming[!is.na(hamming)]),
      0L,
      1L,
      NA_integer_,
      NaN)

  }
  attributes(hamming) <- list(
    Size = length(dataset),
    Labels = names(dataset),
    Diag = FALSE,
    Upper = FALSE,
    method = "hamming",
    class = "dist")
  
  # Return:
  hamming
}

#' Constrained neighbour-joining tree
#'
#' Constructs an approximation to a neighbour-joining tree, modified in order
#' to be consistent with a constraint.  Zero-length branches are collapsed
#' at random.
#'
#' @template datasetParam
#' @template constraintParam
#' @param weight Numeric specifying degree to up-weight characters in
#' `constraint`.
#'
#' @inheritParams NJTree
#' @return `ConstrainedNJ()` returns a tree of class `phylo`.
#' @examples
#' dataset <- MatrixToPhyDat(matrix(
#'   c(0, 1, 1, 1, 0, 1,
#'     0, 1, 1, 0, 0, 1), ncol = 2,
#'   dimnames = list(letters[1:6], NULL)))
#' constraint <- MatrixToPhyDat(
#'   c(a = 0, b = 0, c = 0, d = 0, e = 1, f = 1))
#' plot(ConstrainedNJ(dataset, constraint))
#' @template MRS
#' @importFrom ape nj multi2di
#' @family tree generation functions
#' @export
ConstrainedNJ <- function(dataset, constraint, weight = 1L,
                           ratio = TRUE, ambig = "mean") {
  missing <- setdiff(names(dataset), names(constraint))
  if (length(missing)) {
    constraint <- AddUnconstrained(constraint, missing)
  }
  constraint <- .SubsetPhyDat(constraint, names(dataset))
  tree <- multi2di(nj((Hamming(constraint, ratio = ratio, ambig = ambig)
                       * weight) +
                        Hamming(dataset, ratio = ratio, ambig = ambig)))
  tree[["edge.length"]] <- NULL
  tree <- ImposeConstraint(tree, constraint)
  tree <- RootTree(tree, names(dataset)[1])

  # Return:
  tree
}

# Wrap an edge matrix as a tree in preorder
.PreorderTree <- function(edge,
                          tip.label,
                          Nnode =  dim(edge)[[1]] + 1 - length(tip.label)) {
  structure(
    list(
      # Order is consistent with ape::read.tree (but not ape::rtree...)
      edge = edge,
      Nnode = Nnode,
      tip.label = tip.label
    ),
    order = "preorder",
    class = "phylo"
  )
}

test_that("Write is successful", {
  Test <- function (tree) expect_equal(ape::write.tree(tree), as.Newick(tree))
  Test(BalancedTree(0:7))
  Test(UnrootTree(BalancedTree(0:7)))
  Test(PectinateTree(0:7))
  Test(UnrootTree(PectinateTree(0:7)))
  Test(ape::read.tree(text="(0,1,2,3,4);"))
  Test(ape::read.tree(text="((0,1,2,3,4), ((5, 6), (7, 8, 9)), 10, 11);"))
  Test(as.phylo(1:4, 5, 0:4))

  nasty <- structure(list(edge = structure(
    c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
      5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
    .Dim = c(12, 2)),
    Nnode = 5L,
    tip.label = 0:7),
    class = "phylo") # Danger: Do not plot!
  Test(nasty)
})

test_that("WriteTntCharacters()", {
  nTax <- 4L
  dataset <- matrix(seq_len(nTax), nrow = nTax, ncol = 6,
                    dimnames = list(LETTERS[1:nTax], 1:6))
  dataset[4, 4:6] <- "-"
  dataset[3, 3] <- "(23)"
  dataset[3:2, 4:6] <- "?"

  expect_equal(
    WriteTntCharacters(dataset, comment = c("COM", "MENT"),
                       pre = c("PRE", "FIX"),
                       post  = c("POST", "SCRIPT")),
    "PRE\nFIX\nxread 'COM MENT'\n6 4\nA 1 1 1 1 1 1\nB 2 2 2 ? ? ?\nC 3 3 [23] ? ? ?\nD 4 4 4 - - -\n;\nPOST\nSCRIPT"
    )
  expect_equal(WriteTntCharacters(dataset),
               WriteTntCharacters(MatrixToPhyDat(dataset)))

  expect_equal(
    WriteTntCharacters(dataset, types = c(num = 1, dna = 3)),
    "\nxread 'Dataset written by `TreeTools::WriteTntCharacters()`'\n6 4\n&[num]\n A 1 1\nB 2 2\nC 3 3\nD 4 4\n&[dna]\n A 1 1 1 1\nB 2 ? ? ?\nC [23] ? ? ?\nD 4 - - -\n;\n"
  )

  written <- tempfile()
  WriteTntCharacters(dataset, written)
  on.exit(file.remove(written))
  colnames(dataset) <- NULL
  expect_equal(ReadTntCharacters(written),
               gsub("(23)", "[23]", fixed = TRUE, dataset))
})

test_that("Pectinate trees are generated", {
  expect_equal(ape::read.tree(text = "(t1, (t2, (t3, t4)));"),
               PectinateTree(4L))
  expect_equal(ape::read.tree(text = "(a, (b, (c, (d, e))));"),
               PectinateTree(letters[1:5]))
  expect_equal(ape::read.tree(text = "(a, (b, (c, (d, e))));"),
               PectinateTree(ape::read.tree(text = "(a, ((b, c), (d, e)));")))
  data("Lobo", package = "TreeTools")
  expect_equal(ape::read.tree(text = "(Cricocosmia, (Aysheaia, Siberion));"),
               PectinateTree(.SubsetPhyDat(Lobo.phy, 2:4)))
  expect_true(is.integer(PectinateTree(8)$edge))
})

test_that("Balanced trees are generated correctly", {
  expect_equal(BalancedTree(0), ZeroTaxonTree())
  # nTip even
  expect_equal(BalancedTree(8L),
               Preorder(ape::read.tree(
                 text = "(((t1, t2), (t3, t4)), ((t5, t6), (t7, t8)));")
               ))
  # nTip odd
  expect_equal(BalancedTree(9L),
               Preorder(ape::read.tree(
                 text = "((((t1, t2), t3), (t4, t5)), ((t6, t7), (t8, t9)));")
               ))
  expect_equal(BalancedTree(9L)$edge, Preorder(BalancedTree(9L)$edge))
  expect_equal(BalancedTree(as.character(1:9)), BalancedTree(1:9))
  escapees <- c("Apostrophe's", 'and quote"s')
  expect_equal(ignore_attr = TRUE,
               PectinateTree(escapees), BalancedTree(escapees))
  expect_equal(integer(0), .BalancedBit(seq_len(0)))
  expect_equal("Test", .BalancedBit("Test"))
  expect_true(is.integer(BalancedTree(8)[["edge"]]))
})

test_that("BalancedTree(lengths)", {
  expect_equal(BalancedTree(0, lengths = 1), ZeroTaxonTree())
  expect_equal(BalancedTree(1, lengths = 2), SingleTaxonTree("t1", lengths = 2))
  expect_equal(BalancedTree(2, lengths = 2)$edge.length, c(2, 2))
  expect_equal(BalancedTree(3, lengths = 1:3)$edge.length, c(1:3, 1))
  expect_equal(BalancedTree(3, lengths = seq_len)$edge.length, 1:4)
})
test_that("PectinateTree(lengths)", {
  expect_equal(PectinateTree(0, lengths = 1), ZeroTaxonTree())
  expect_equal(PectinateTree(1, lengths = 2), SingleTaxonTree("t1", lengths = 2))
  expect_equal(PectinateTree(2, lengths = 2)$edge.length, c(2, 2))
  expect_equal(PectinateTree(3, lengths = 1:3)$edge.length, c(1:3, 1))
  expect_equal(PectinateTree(3, lengths = seq_len)$edge.length, 1:4)
})

test_that("StarTree() works", {
  expect_equal(ape::read.tree(text = "(t1, t2, t3, t4, t5, t6, t7, t8);"),
               StarTree(8L))
  expect_true(is.integer(StarTree(8)$edge))
  expect_null(StarTree(8L)[["edge.length"]])
  expect_equal(StarTree(8L, 8:1)[["edge.length"]], 8:1)
  expect_equal(StarTree(4L, seq_len)[["edge.length"]], 1:4)
})

test_that("Random trees are generated correctly", {
  # These two seeds give two different node labellings
  random3 <- c(4, 4, 5, 5, 1, 5, 2, 3)
  set.seed(1)
  expect_equal(random3, RandomTree(3, root = 1)$edge[1:8])
  set.seed(4)
  expect_equal(random3, RandomTree(3, root = "t1")$edge[1:8])
  
  expect_true(all.equal(RandomTree(3, root = "t2"),
    PectinateTree(c("t2", "t3", "t1"))))
  expect_equal(c(4, 4, 4), RandomTree(3, root = FALSE)$edge[1:3])
  expect_warning(expect_equal(RandomTree(3, root = "t2"),
                              RandomTree(3, root = 2:3)))
  expect_error(RandomTree(4, root = "not_there"), "No match found for `root`")
  expect_error(RandomTree(4, root = 999), "exceeds number of nodes")
  expect_error(RandomTree(4, root = -1), "must be a positive integer")
    
  expect_warning(expect_equal({
    set.seed(1)
    RandomTree(8, root = NA_integer_)
  }, {
    set.seed(1)
    RandomTree(8, root = FALSE)
  }), "Treating `root = NA` as `FALSE`")
  expect_error(RandomTree(4, nodes = 0), "A tree must contain one .*node")

  expect_warning(RandomTree(4, nodes = 4),
                 "`nodes` higher than number in binary tree")
  expect_warning(RandomTree(4, root = FALSE, nodes = 3),
                 "`nodes` higher than number in binary tree")

  for (nNode in 1:8) {
    expect_equal(RandomTree(10, nodes = nNode)$Nnode, nNode)
  }
  for (nNode in 1:9) {
    expect_equal(RandomTree(10, root = TRUE, nodes = nNode)$Nnode, nNode)
  }
})

test_that("RandomTree(lengths)", {
  expect_equal(RandomTree(0, lengths = 1), ZeroTaxonTree())
  expect_equal(RandomTree(1, lengths = 2), SingleTaxonTree("t1", lengths = 2))
  expect_equal(RandomTree(3, lengths = 2)$edge.length, c(2, 2, 2))
  expect_equal(RandomTree(3, lengths = 1:3, root = FALSE)$edge.length, 1:3)
  expect_equal(RandomTree(3, lengths = 1:3, root = TRUE)$edge.length, c(1:3, 1))
  expect_equal(RandomTree(5, nodes = 2, lengths = 1:3)$edge.length, c(1:3, 1:3))
  expect_equal(RandomTree(3, lengths = seq_len, root = FALSE)$edge.length, 1:3)
  expect_equal(RandomTree(3, lengths = seq_len, root = TRUE)$edge.length, 1:4)
})

test_that("Small random trees are generated", {
  expect_equal(RandomTree(0, root = FALSE), ZeroTaxonTree())
  expect_equal(RandomTree(0, root = TRUE), ZeroTaxonTree())
  expect_equal(RandomTree(0), ZeroTaxonTree())
  expect_equal(RandomTree("one", root = FALSE), SingleTaxonTree("one"))
  expect_equal(RandomTree("one", root = TRUE), SingleTaxonTree("one"))
  expect_equal(RandomTree(letters[1:2], root = FALSE),
               UnrootTree(PectinateTree(letters[1:2])))
  expect_equal(RandomTree(letters[1:2], root = TRUE),
               UnrootTree(PectinateTree(letters[1:2])))
})

test_that("YuleTree() works", {
  expect_equal(YuleTree(0), ZeroTaxonTree())
  expect_equal(YuleTree("a"), SingleTaxonTree("a"))
  expect_equal(YuleTree(2, addInTurn = TRUE), BalancedTree(2))
  expect_equal(NTip(YuleTree(10)), 10)
  expect_equal(YuleTree(10)$Nnode, 9)
  expect_true(all(TipLabels(YuleTree(10)) %in% TipLabels(10)))
  expect_equal(TipLabels(YuleTree(10, addInTurn = TRUE)), TipLabels(10))
  expect_equal(
    mean(replicate(100, TotalCopheneticIndex(YuleTree(10)))),
    TCIContext(10)$yule.expected,
    tolerance = 0.1
  )
})

test_that("YuleTree() root parameter", {
  expect_equal(
    {set.seed(0); YuleTree(10, root = FALSE)},
    {set.seed(0); UnrootTree(YuleTree(10, root = TRUE))}
  )
  
  expect_warning(expect_equal(
    {set.seed(0); YuleTree(10, root = NA)},
    {set.seed(0); YuleTree(10, root = FALSE)}
  ), "root = NA")
})

test_that("YuleTree(lengths)", {
  expect_equal(YuleTree(0, lengths = 1), ZeroTaxonTree())
  expect_equal(YuleTree(1, lengths = 2), SingleTaxonTree("t1", lengths = 2))
  expect_equal(YuleTree(3, lengths = 2)$edge.length, rep(2, 4))
  expect_equal(YuleTree(3, lengths = 1:3, root = FALSE)$edge.length, 1:3)
  expect_equal(YuleTree(3, lengths = 1:3, root = TRUE)$edge.length, c(1:3, 1))
  expect_equal(YuleTree(3, lengths = seq_len, addInTurn = TRUE)$edge.length, 1:4)
})

test_that("SingleTaxonTree(lengths)", {
  expect_equal(SingleTaxonTree("t1", lengths = 2)$edge.length, 2)
  expect_equal(SingleTaxonTree("t1", lengths = seq_len)$edge.length, 1)
})

test_that("Hamming() works", {
  dataset <- StringToPhyDat("111100 ???000 ???000 111??? 10??10",
                            letters[1:5], byTaxon = TRUE)
  expected <- c(1/3, 1/3, 0, 1/2,
                0, NaN, 1/2,
                NaN, 1/2,
                1/2)
  expect_equal(as.double(Hamming(dataset, ambig = "NAN")), expected)
  ex <- expected
  ex[is.nan(expected)] <- NA
  expect_equal(as.double(Hamming(dataset, ambig = c("NA", "mean"))), ex)
  ex[is.nan(expected)] <- 0
  expect_equal(as.double(Hamming(dataset, ambig = 0)), ex)
  ex[is.nan(expected)] <- 1
  expect_equal(as.double(Hamming(dataset, ambig = "1")), ex)
  ex[is.nan(expected)] <- mean(expected[!is.nan(expected)])
  expect_equal(as.double(Hamming(dataset, ambig = "mean")), ex)
  ex[is.nan(expected)] <- median(expected[!is.nan(expected)])
  expect_equal(as.double(Hamming(dataset, ambig = "med")), ex)
  expect_error(Hamming(dataset, ambig = "ERROR"))
})

test_that("Hamming() handles inapplicables", {
  dataset <- StringToPhyDat("221100 ---000 ---000 211{-0}?? 10-?10",
                            letters[1:5], byTaxon = TRUE)
  expected <- c(1/3, 1/3, 1/3, 3/4,
                0, NaN, 1/2,
                NaN, 1/2,
                1)
  expect_equal(as.double(Hamming(dataset, ambig = "NaN")), expected)

})

test_that("NJTree() works", {
  a..f <- letters[1:6]
  bal6 <- StringToPhyDat("111100 111000 111000 110000", letters[1:6],
                         byTaxon = FALSE)
  expect_true(all.equal(
    RootTree(NJTree(bal6), a..f[1:3]),
    BalancedTree(letters[c(1:3, 6:4)])
  ))
  expect_equal(NJTree(bal6, edgeLengths = TRUE),
               Preorder(NJTree(bal6, edgeLengths = TRUE)))
  expect_equal(c(0, 1, 2, 1, rep(0, 6)),
               RootTree(NJTree(bal6, edgeLengths = TRUE), 6)$edge.length * 4L)
})

test_that("Constrained NJ trees work", {
  dataset <- MatrixToPhyDat(matrix(
    c(0, 1, 1, 1, 0, 1,
      0, 1, 1, 0, 0, 1), ncol = 2,
    dimnames = list(letters[1:6], NULL)))
  constraint <- MatrixToPhyDat(c(a = 0, b = 0, c = 0, d = 0, e = 1, f = 1))
  expect_true(all.equal(read.tree(text = "(a, (d, ((c, b), (e, f))));"),
                        ConstrainedNJ(dataset, constraint)))
  # b == c == f, so these three could be resolved in one of three ways. Drop B.
  expect_true(all.equal(DropTip(NJTree(dataset), "b"),
                        DropTip(ConstrainedNJ(dataset, dataset), "b")))

  expect_true(all.equal(
    KeepTip(ConstrainedNJ(dataset, constraint[3:6]), letters[3:6]),
    BalancedTree(letters[3:6])
  ))
})

test_that("Hamming() fails nicely", {
  expect_error(Hamming(matrix(1:4, 2, 2)))
})

test_that("Node supports calculated correctly", {
  treeSample <- list(
    correct = ape::read.tree(text = "((((((A,B),C),D),E),F),out);"),
    swapFE  = ape::read.tree(text = "((((((A,B),C),D),F),E),out);"),
    DEClade = ape::read.tree(text = "(((((A,B),C),(D,E)),F),out);"),
    swapBC  = ape::read.tree(text = "((((((A,C),B),D),E),F),out);"),
    DbyA    = ape::read.tree(text = "((((((A,D),C),B),E),F,G),out);")
  )
  expect_equal(
    SplitFrequency(treeSample$correct, treeSample),
    c("10" = 4, "11" = 4, "12" = 4, "13" = 3)
   )
  
  expect_error(SplitFrequency(NULL, treeSample), "must bear identical")
  sameTips <- KeepTip(treeSample, TipLabels(treeSample$correct))
  sameSplits <- do.call(c, as.Splits(sameTips))
  
  expect_equal(
    unname(sort(PolarizeSplits(pole = 1,
    SplitFrequency(sameTips) 
    ))),
    unname(sort(PolarizeSplits(pole = 1,
    structure(sameSplits[[!duplicated(sameSplits)]],
              count = c(4, 4, 4, 3, 1, 1, 1, 1, 1))
    )))
  )
  
  
  monoSplit <- ape::read.tree(text = "((a, b, c, d), (e, f, g));")
  expect_equal(SplitFrequency(monoSplit),
               structure(unname(as.Splits(monoSplit)), count = 1))
  
  expect_equal(SplitFrequency(SingleTaxonTree()),
               structure(as.Splits(SingleTaxonTree()), count = integer()))
  
  expect_equal(SplitFrequency(as.Splits(monoSplit)[[FALSE]]),
               structure(as.Splits(monoSplit)[[FALSE]], count = integer()))
  
  expect_equal(SplitFrequency(StarTree(67)),
               structure(as.Splits(StarTree(67)), count = integer()))
  
  # Internal nodes on each side of root
  balanced <- ape::read.tree(text="((D, (E, (F, out))), (C, (A, B)));")
  freq <- SplitFrequency(balanced, treeSample)
  expect_equal(freq,
               c("9" = 4, "10" = 4, "11" = 4, "12" = 4, "13" = 3)[names(freq)])

  skip_if(!isTRUE(options("runSlowTests")))
  expect_equal(SplitFrequency(c(PectinateTree(8200), PectinateTree(8200))),
               structure(as.Splits(PectinateTree(8200)), count = rep(2, 8197)))
})

test_that("SplitFrequency() exact and hashed counts agree", {
  set.seed(2)
  freqKey <- function(sf) sort(paste(as.character(sf), attr(sf, "count")))
  forests <- list(
    c(PectinateTree(16), PectinateTree(16)),
    lapply(1:10, function(i) RandomTree(13, root = TRUE)),
    lapply(1:30, function(i) RandomTree(8, root = TRUE))
  )
  for (f in forests) {
    expect_equal(freqKey(SplitFrequency(f)),
                 freqKey(SplitFrequency(f, hash = FALSE)))
  }
})

test_that("Node support colours consistent", {
  expect_equal(SupportColour(NA), "red")
  expect_equal(SupportColour(1:2, show1 = FALSE), c("#ffffff00", "red"))
  expect_equal(SupportColor(-2), "red") # Check alternative spelling
  expect_equal(SupportColour(1, show1 = FALSE), "#ffffff00")
  expect_equal(SupportColour((-1):3 / 3, scale = 1:101, outOfRange = "oor"),
               c("oor", "1", "34", "67", "101"))
})

test_that("normalize_splits() covers all branches", {
  # n_spare == 0 (nTip multiple of 8): complement without trailing-bit mask
  trees16 <- c(PectinateTree(16), PectinateTree(16))
  freq16 <- SplitFrequency(trees16)
  expect_equal(length(freq16), NSplits(trees16[[1]]))
  expect_true(all(attr(freq16, "count") == 2L))

  # Trivial splits are filtered: a star tree produces no non-trivial splits
  star <- c(StarTree(10), StarTree(10))
  freq_star <- SplitFrequency(star)
  expect_equal(length(freq_star), 0)
})

test_that("SplitFrequency() handles four-split trees", {
  trees <- AddTipEverywhere(BalancedTree(3))
  trees <- c(trees[1], trees)
  expect_equal(c("7" = 2L), SplitFrequency(trees[[1]], trees))
})

test_that("LabelSplits()", {
  expect_error(LabelSplits(BalancedTree(8), 1:8))
  skip_if_not_installed("vdiffr", minimum_version = "1.0.0")
  skip_if(packageVersion("graphics") < "4.1.0")
  vdiffr::expect_doppelganger("LabelSplits()",  function() {
    tree <- BalancedTree(9)
    plot(tree)
    labs <- letters[6:1]
    names(labs) <- rev(names(as.Splits(tree)))
    LabelSplits(tree, labs, frame = "circ", cex = 2, bg = "orange")
  })
  vdiffr::expect_doppelganger("LabelSplits()-nameless", function() {
    tree <- BalancedTree(9)
    plot(tree)
    LabelSplits(tree, bg = "orange")
    expect_warning(LabelSplits(BalancedTree(9), setNames(letters[11:16], 1:6)))
  })
  vdiffr::expect_doppelganger("LabelSplits()-names", function() {
    tree <- BalancedTree(9)
    plot(tree)
    labs <- letters[1:6]
    LabelSplits(tree, labs, bg = "orange")
  })
})

test_that("Split contradiction is calculated", {
  pec12 <- as.Splits(PectinateTree(12))
  bal12 <- as.Splits(BalancedTree(12))
  bal8 <- as.Splits(BalancedTree(8))
  sp_cons <- function(n, h) split_consistent(n, h, FALSE)
  expect_error(sp_cons(pec12, list(bal12)),
               "must contain exactly one split")
  expect_error(sp_cons(pec12[[1]], list(bal8)),
               "Haystack 1 has different nTip than needle")
  expect_error(sp_cons(`attr<-`(pec12[[1]], "nTip", NULL), list(1)),
               "lacks nTip")
  
  expect_equal(SplitConsistent(pec12[[2]], bal12),
               SplitConsistent(pec12[[2]], list(bal12)))
  expect_error(SplitConsistent(pec12[[2]], "bal12"), "must be a Splits object")
  expect_error(SplitConflicts(pec12[[2]], "bal12"), "must be a Splits object")
  expect_error(sp_cons(pec12[[2]], bal12), "Not a matrix")
  expect_error(sp_cons(pec12[[2]], list(BalancedTree(12))),
               "not a RawMatrix")
  
  expect_error(sp_cons(bal8[[4]], list(bal12)),
               "different nTip")
  
  expect_equal(sp_cons(pec12[[5]], list(bal12))[[1]],
               rep(TRUE, length(bal12)))
  expect_equal(sp_cons(pec12[[3]], list(bal12))[[1]],
               c(rep(TRUE, 3), FALSE, FALSE, rep(TRUE, 4)))
  expect_equal(split_consistent(pec12[[3]], list(bal12), TRUE)[[1]],
               !c(rep(TRUE, 3), FALSE, FALSE, rep(TRUE, 4)))
  expect_equal(SplitConsistent(pec12[[5]], list(bal12, bal12))[[2]],
               rep(TRUE, length(bal12)))
  expect_equal(SplitConflicts(pec12[[5]], as.Splits(StarTree(12)))[[1]],
               logical(0))
  expect_equal(SplitConflicts(pec12[[2]], list(bal12, pec12)),
               lapply(SplitConsistent(pec12[[2]], list(bal12, pec12)), `!`))
})

test_that("Consensus() errors", {
  expect_error(Consensus(trees), " a list")
  
  bal8 <- BalancedTree(8)
  oneLeaf <- Consensus(list(DropTip(bal8, 1:7))[c(1, 1, 1)])
  expect_equal(class(oneLeaf), "phylo")
  expect_equal(oneLeaf$Nnode, 0)
  expect_identical(Consensus(list(DropTip(bal8, 1:8))[c(1, 1, 1)]),
                   DropTip(bal8, 1:8))
  expect_warning(expect_true(all.equal(
    Consensus(list(PectinateTree(6), PectinateTree(8))),
    PectinateTree(6)
  )), regexp = "Tree sizes")

  halfTree <- CollapseNode(bal8, 10:12)
  expect_equal(Consensus(halfTree), halfTree)
  expect_equal(Consensus(c(halfTree)), halfTree)
  expect_equal(Consensus(list(halfTree)), halfTree)
  
  # Test that trees larger than heap limit fail gracefully
  expect_error(
    Consensus(c(BalancedTree(100001), PectinateTree(100001))),
    "too many leaves.*100000"
  )
  
  largeTree <- BalancedTree(33333)
  consensus_large <- Consensus(c(largeTree, largeTree))
  expect_equal(NTip(consensus_large), 33333)
})

test_that("Consensus()", {

  ApeTest <- function(tr, p = 1) {
    if (!expect_true(all.equal(RootTree(Consensus(tr, p = p), 1),
                               RootTree(ape::consensus(tr, p = p), 1)))) {
      dput(RootTree(Consensus(tr, p = p), 1)$edge)
      dput(RootTree(ape::consensus(tr, p = p), 1)$edge)
    }
  }

  trees <- list(BalancedTree(8), PectinateTree(8))[c(1, 1, 1, 1, 2, 2, 2)]
  trees <- RenumberTips(trees, trees[[1]])
  expect_error(Consensus(trees, 2), regexp = "`p`") # p too high
  expect_error(Consensus(trees, 0.2), regexp = "`p`") # p too low

  ApeTest(trees)
  ApeTest(trees, 0.5)
  expect_equal(Consensus(trees, 0.999), Consensus(trees, 1))

  ApeTest(as.phylo(0:2, 8))
  ApeTest(as.phylo(0:250, 8))
  ApeTest(as.phylo(0:250, 80))
  
  trees <- list(ape::read.tree(text = "((a, b), (c, d));"),
                ape::read.tree(text = "((a, c), (b, d));"))
  expect_equal(Consensus(trees), Preorder(StarTree(letters[1:4])))
})

test_that("Consensus() handles large sets of trees", {
  oneTree <- as.phylo(0, 13)
  expect_equal(
    Consensus(lapply(1:33000, function(x) oneTree)),
    oneTree
  )
  
  manyTrees <- lapply(1:33000, as.phylo, 13) # More than int16_max
  expect_equal(
    Consensus(c(manyTrees[1:16500], lapply(1:16500, function(x) oneTree)),
              p = 0.5),
    oneTree
  )
  
  skip_on_cran() # Slow!
  skip_if(!isTRUE(options("runSlowTests")))
  
  expect_true(all.equal(
    Consensus(manyTrees),
    read.tree(text = "((((((t2,t13),t12),t11),t10),t3,t4,t5,t6,t7,t8,t9),t1);")
    # i.e. write.tree(ape::consensus(manyTrees))
  ))
})

test_that("Consensus() exact and hashed counts agree", {
  # The hashed (default) and exact (opt-in) split counts must yield identical
  # consensus trees; this also guards the shared counting core.
  skip_if_not_installed("ape")
  set.seed(1)
  forests <- list(
    balPec = list(BalancedTree(8), PectinateTree(8))[c(1, 1, 1, 1, 2, 2, 2)],
    starlike = list(ape::read.tree(text = "((a, b), (c, d));"),
                    ape::read.tree(text = "((a, c), (b, d));")),
    tie = c(rep(list(BalancedTree(8)), 2L), rep(list(PectinateTree(8)), 2L)),
    rand12 = lapply(1:7, function(i) ape::rtree(12, br = NULL)),
    rand9  = lapply(1:20, function(i) ape::rtree(9, br = NULL))
  )
  for (f in forests) {
    for (p in c(0.5, 2 / 3, 1)) {
      hashed <- Consensus(f, p = p)
      exact  <- Consensus(f, p = p, hash = FALSE)
      expect_true(isTRUE(all.equal(RootTree(hashed, 1), RootTree(exact, 1))))
    }
  }
})

test_that("Consensus() handles non-preorder trees", {
  trees <- ape::read.nexus(test_path("testdata", "nonPreCons.nex"))
  expect_equal(Consensus(trees)$Nnode, 3)
  expect_equal(Consensus(trees), Preorder(trees[[1]]))
})

test_that("ConsensusWithout() is robust", {
  tr <- as.phylo(0:6, 16)
  expect_identical(ConsensusWithout(tr, paste0("t", 1:16)),
                   DropTip(tr[[1]], 1:16))
  expect_identical(ConsensusWithout(tr, paste0("t", 1:4)),
                   Consensus(DropTip(tr, 1:4)))

  expect_true(all.equal(BalancedTree(8), ConsensusWithout(BalancedTree(8))))
  expect_true(all.equal(BalancedTree(4),
                        ConsensusWithout(BalancedTree(8), paste0("t", 5:8))))
  balAndPec <- list(BalancedTree(8), PectinateTree(8))
  t25 <- paste0("t", c(2:5))
  expect_true(all.equal(PectinateTree(paste0("t", c("1", 6:8))),
                        ConsensusWithout(balAndPec, t25)))
  expect_true(all.equal(
    ConsensusWithout(structure(balAndPec, class = "multiPhylo"), t25),
    ConsensusWithout(balAndPec, t25))
  )

  nasty <- structure(list(edge = structure(
    c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
      5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
    .Dim = c(12, 2)),
    Nnode = 5L,
    tip.label = letters[1:8]),
    class = "phylo") # Danger: Do not plot!
  expect_true(all.equal(Preorder(nasty), ConsensusWithout(nasty)))
  expect_true(all.equal(DropTip(nasty, 2), ConsensusWithout(nasty, "b")))

})

test_that("as.multiPhylo()", {
  expect_equal(structure(list(BalancedTree(8), PectinateTree(8)),
                         class = "multiPhylo"),
               as.multiPhylo(list(BalancedTree(8), PectinateTree(8))))
  expect_equal(structure(list(BalancedTree(8)), class = "multiPhylo"),
               as.multiPhylo(BalancedTree(8)))

  char <- MatrixToPhyDat(matrix(c(1,1,1,0,0,0), ncol = 1,
                                dimnames = list(letters[1:6], NULL)))
  expect_equal(ape::read.tree(text = "((a, b, c), (d, e, f));"),
               as.multiPhylo(char)[[1]])

  char2 <- MatrixToPhyDat(matrix(c(1,1,1,0,0,0,
                                   0,1,1,1,1,"?",
                                   0,0,1,1,"-",2,
                                   0,1,1,1,1,"?"), ncol = 4,
                                dimnames = list(letters[1:6], NULL)))
  mpChar2 <- as.multiPhylo(char2)

  expect_equal(ape::read.tree(text = "((a, b, c), (d, e, f));"),
               mpChar2[[1]])
  expect_equal(ape::read.tree(text = "(b, c, d, e);"),
               mpChar2[[2]])
  expect_equal(ape::read.tree(text = "((a, b), (c, d));"),
               mpChar2[[3]])
  expect_equal(mpChar2[[2]], mpChar2[[4]])

  mpSplits <- as.Splits(PectinateTree(letters[1:6]))
  expect_true(all.equal(as.multiPhylo(mpSplits), structure(list(
    "9" = ape::read.tree(text = "((a, b), (c, d, e, f));"),
    "10" = ape::read.tree(text = "((a, b, c), (d, e, f));"),
    "11" = ape::read.tree(text = "((a, b, c, d), (e, f));")),
    class = "multiPhylo")))

})

test_that("Stemwardness functions", {
  bal8 <- BalancedTree(8)
  pec8 <- PectinateTree(8)
  pr3 <- RootTree(pec8, "t3")

  expect_equal(1L, SisterSize(bal8, 3))
  expect_equal(2L, RootNodeDist(bal8, 3))

  expect_equal(5L, SisterSize(pec8, "t3"))
  expect_equal(2L, RootNodeDist(pec8, "t3"))

  expect_equal(7L, SisterSize(pr3, "t3"))
  expect_equal(0L, RootNodeDist(pr3, "t3"))
})

ApeRoot <- function(tree, root, rr = TRUE) ape::root(tree, root, resolve.root = rr)

test_that("Memory leak not encountered", {
  # Example from TreeDist::ClusterTable
  tree1 <- ape::read.tree(text = "(A, (B, (C, (D, E))));");
  tree2 <- ape::read.tree(text = "(A, (B, (D, (C, E))));");
  # as.ClusterTable(tree1) calls:
  expect_equal(tree1, root_on_node(tree1, 1))

  # Check for memory leaks...
  root_on_node(RenumberTips(Preorder(tree2), LETTERS[1:5]), 1)[]
  root_on_node(RenumberTips(StarTree(LETTERS[5:1]), LETTERS[1:5]), 1)[]

  expect_error(root_on_node(tree1, 0), "`outgroup` must be a positive integer")
  expect_error(root_on_node(tree1, 999), "`outgroup` exceeds number of nodes")
})

test_that("Big trees don't fail", {
  # 2^14 + 1 is too big for int16
  expect_equal(root_on_node(PectinateTree(2^14 + 1), 1),
               PectinateTree(2^14 + 1))
})

test_that("Small trees are rootable", {
  ztt <- ZeroTaxonTree()
  expect_equal(root_on_node(ztt, 1), ztt)
  expect_equal(root_binary(ztt[["edge"]], 1), ztt$edge)
  expect_equal(root_on_node(ztt, 999), ztt)
  
  stt <- SingleTaxonTree()
  expect_equal(root_on_node(stt, 1), stt)
  expect_equal(root_on_node(stt, 2), root_on_node(stt, 1))
})

test_that("Binary trees are rootable", {
  Test <- function(tree, root) {
    expect_equal(Preorder(ApeRoot(tree, tree$tip.label[root]))$edge,
                 root_binary(tree$edge, root))
  }
  Test(BalancedTree(9), 3)
  Test(BalancedTree(9), 1)
  Test(PectinateTree(9), 1)
  Test(PectinateTree(9), 7)
  ed9 <- PectinateTree(9)$edge
  expect_equal(root_binary(ed9, 10), ed9)
  expect_equal(root_binary(ed9, 1), ed9)
  expect_error(root_binary(ed9, 9999), "exceeds number of nodes")
})

test_that("Polytomous trees are rootable", {
  Test <- function(tree, root) {
    expect_equal(Preorder(ApeRoot(tree, tree$tip.label[root])),
                 root_on_node(tree, root))
  }
  bt <- BalancedTree(9)
  pt <- PectinateTree(9)
  Test(CollapseNode(bt, 12), 1)
  Test(CollapseNode(bt, 12), 3)
  Test(CollapseNode(bt, 11), 1)
  Test(CollapseNode(pt, 11), 1)
  Test(CollapseNode(pt, c(11, 12)), 1)
  Test(CollapseNode(pt, c(11, 12)), 3)
  Test(CollapseNode(pt, c(11, 12)), 5)
  Test(CollapseNode(pt, c(11, 13, 15)), 5)
  Test(CollapseNode(pt, c(11:13, 15)), 9)
  Test(StarTree(8), 1)

  # Day 1985 examples
  t1 <- Preorder(ape::read.tree(
    text = "((10, 7), (6, (8, 11)), (12, (4, (2, 1))), 14, (5, 9, 13), 3);"))
  Test(t1, 1)
  t2 <- Preorder(ape::read.tree(
    text = "(((2, 4, 5, 7, 9, 10, 12, 13), (1, 14)), (6, (8, 11)), 3);"))
  Test(t2, 1)
})

test_that("Rooted trees report preorder accurately", {
  set.seed(1)
  nTips <- 8
  edge <- do.call(cbind,
                  RenumberEdges(.RandomParent(nTips),
                                seq_len(nTips + nTips - 2L)))
  
  expect_preorder <- function(x) {
    expect_equal(x, Preorder(`attr<-`(x, "order", "unknown")))
  }
  
  # Check that we are in preorder
  expect_preorder(edge)
  
  expect_preorder(root_binary(edge, 2))
  expect_preorder(root_binary(edge, 6))
  expect_error(root_binary(edge, 999), "exceeds number of nodes")
  
  tree <- structure(list(edge = edge,
                         Nnode = nTips - 1L,
                         tip.label = TipLabels(nTips)),
                    order = "preorder",
                    class = "phylo")
  
  rootNode <- nTips + 1L
  expect_preorder(root_on_node(tree, rootNode))
  deepNode <- 2 * nTips - 2
  expect_preorder(root_on_node(tree, deepNode))
  
  weighted <- tree
  weights <- seq_len(dim(edge)[[1]])
  weighted[["edge.length"]] <- weights
  expect_equal(sort(root_on_node(weighted, rootNode)[["edge.length"]]),
                    weighted[["edge.length"]])
  expect_preorder(root_on_node(weighted, rootNode))
  expect_preorder(root_on_node(weighted, deepNode))
})

test_that("ClusterTable fails gracefully", {
  bigTree <- PectinateTree(100001)
  expect_error(
    as.ClusterTable(bigTree),
    "Tree has too many leaves.*100000"
  )
  
  mediumTree <- PectinateTree(20000)
  ct <- as.ClusterTable(mediumTree)
  expect_equal(attr(ct, "nTip"), 20000)
})

test_that("ClusterTable class behaves", {
  tree <- RootTree(BalancedTree(6), 1)
  ct <- as.ClusterTable(tree)
  expect_equal(matrix(c(0, 2, rep(1, 4), 0, 3, 3:6), 6),
               as.matrix.ClusterTable(ct))
  expect_equal(ClusterTable_decode(ct), 6:1)
  
  expect_equal(capture.output(ct),
               "ClusterTable on 6 leaves: t1 .. t6")
  
  expect_equal(
    capture.output(summary(ct)),
    c("ClusterTable on 6 leaves:",
      " 123456",
      " .**...",
      " ***...",
      " ****..",
      " *****.",
      " ******",
      " 1: t6  2: t5  3: t4  4: t3  5: t2  6: t1 ")
  )
  
  ClusterSummary <- function(...) capture.output(summary(as.ClusterTable(...)))
  t1..8 <- paste0("t", 1:8)
  t8..1 <- rev(t1..8)
  expect_equal(
    ClusterSummary(BalancedTree(t8..1),
                   tipLabels = t1..8),
    ClusterSummary(BalancedTree(t1..8))
  )
  
  byList <- as.ClusterTable(list(BalancedTree(t8..1), PectinateTree(t8..1)),
                            tipLabels = t1..8)
  listBy <- list(as.ClusterTable(BalancedTree(t1..8)),
                 as.ClusterTable(PectinateTree(t1..8)))
  expect_equal(capture.output(summary(byList[[2]])),
               capture.output(summary(listBy[[2]])))
  
})

test_that("ClusterTable with multiple trees", {
  tree1 <- ape::read.tree(text = "(A, (B, (C, (D, E))));");
  tree2 <- ape::read.tree(text = "(E, (B, (D, (C, A))));");
  ct <- as.ClusterTable(c(tree1, tree2))
  expect_equal(
    capture.output(print(ct)),
    capture.output(print(list(as.ClusterTable(tree1),
                              as.ClusterTable(tree2, TipLabels(tree1)))))
  )
  expect_equal(
    capture.output(summary(ct)),
    capture.output(summary(list(as.ClusterTable(tree1),
                                as.ClusterTable(tree2, TipLabels(tree1)))))
  )
  
  trees <- ape::read.tree(text = 
                            "(A, (B, (C, (D, E))));(E, (B, (D, (C, A))));");
  expect_equal(
    capture.output(print(as.ClusterTable(trees))),
    capture.output(print(list(as.ClusterTable(trees[[1]]),
                              as.ClusterTable(trees[[2]], TipLabels(tree1)))))
  )
})

test_that("Attributes are correct", {
  t6 <- as.ClusterTable(BalancedTree(6))
  t7 <- as.ClusterTable(PectinateTree(7))
  t8 <- as.ClusterTable(BalancedTree(8))
  s8 <- StarTree(8)
  expect_equal(3, NSplits(t6))
  expect_equal(4:5, NSplits(list(t7, t8)))
  
  expect_equal(6, NTip(t6))
  expect_equal(7:8, NTip(list(t7, t8)))
  
  #TODO test TipLabels, SplitsInBalancedTree
})

test_that("ClusterTable with multiple trees", {
  tree1 <- ape::read.tree(text = "(A, (B, (C, (D, E))));");
  tree2 <- ape::read.tree(text = "(E, (B, (D, (C, A))));");
  ct <- as.ClusterTable(c(tree1, tree2))
  expect_equal(
    capture.output(print(ct)),
    capture.output(print(list(as.ClusterTable(tree1),
                              as.ClusterTable(tree2, TipLabels(tree1)))))
  )
  expect_equal(
    capture.output(summary(ct)),
    capture.output(summary(list(as.ClusterTable(tree1),
                                as.ClusterTable(tree2, TipLabels(tree1)))))
  )
  
  trees <- ape::read.tree(text = 
                            "(A, (B, (C, (D, E))));(E, (B, (D, (C, A))));");
  expect_equal(
    capture.output(print(as.ClusterTable(trees))),
    capture.output(print(list(as.ClusterTable(trees[[1]]),
                              as.ClusterTable(trees[[2]], TipLabels(tree1)))))
  )
})

test_that("ClusterTable with complex trees", {
  skip_if_not_installed("TreeDist", "2.9.2.9000")
  skip_on_cran()
  library("TreeDist")
  
  # Test exposes failures in C++ - constexpr not playing nicely with Rcpp
  # Specifically if replacing
  # const int16 L_COL = int16(0);
  # const int16 R_COL = int16(1);
  # const int16 X_COLS = int16(2);
  # hard-coding, using enum all fails.
  tr1 <- structure(list(
    edge = structure(c(8L, 8L, 9L, 10L, 10L, 9L, 11L, 11L, 8L, 12L, 12L,
                       1L, 9L, 10L, 2L, 3L, 11L, 4L, 5L, 12L, 6L, 7L),
                     dim = c(11L, 2L)),
    Nnode = 5L, tip.label = c("t1", "t2", "t3", "t4", "t5", "t6", "t7")),
    class = "phylo", order = "preorder")
  tr2 <- structure(list(
    edge = structure(c(8L, 9L, 10L, 10L, 9L, 11L, 11L, 8L, 12L, 12L, 8L,
                       9L, 10L, 1L, 2L, 11L, 3L, 4L, 12L, 5L, 6L, 7L),
                     dim = c(11L, 2L)),
    Nnode = 5L, tip.label = c("t1", "t2", "t3", "t4", "t5", "t6", "t7")),
    class = "phylo", order = "preorder")
  t4 <- list(a = tr1, b = tr2, c = tr1, d = tr2)
  r4 <- RootTree(t4, 1)
  
  expect_equal(as.numeric(RobinsonFoulds(r4)), c(8, 0, 8, 8, 0, 8))
})

test_that("Pairwise distances calculated correctly", {
  nTrees <- 6L
  nTip <- 16L

  set.seed(0)
  trees <- lapply(rep(nTip, nTrees), RandomTree, root = TRUE)
  trees[[1]] <- BalancedTree(nTip)
  trees[[nTrees - 1L]] <- PectinateTree(nTip)
  class(trees) <- "multiPhylo"

  # From example
  TCIRange <- function(tree1, tree2) {
   range(TotalCopheneticIndex(tree1), TotalCopheneticIndex(tree2))
  }
  tciPairs <- PairwiseDistances(trees, TCIRange, 2)
  expect_equal(length(tciPairs), 2)
  expect_equal(as.matrix(tciPairs[[1]])[3, 6],
               TCIRange(trees[[3]], trees[[6]])[1])
  
  skip_if_not_installed("phangorn")
  trees <- reorder(trees, "cladewise")
  dists <- PairwiseDistances(trees, phangorn::RF.dist)
  expect_equal(as.integer(phangorn::RF.dist(trees)), as.integer(dists))
})

test_that("Cherries works", {
  expect_equal(Cherries(BalancedTree(8)), 4L)
  expect_equal(Cherries(PectinateTree(8)), 1L)
  expect_equal(Cherries(UnrootTree(PectinateTree(8))$edge, 8L), 2L)
  expect_error(Cherries(matrix(4, 4, 4)), "edge matrix")
  expect_error(Cherries(1:3), "edge matrix")
  
  expect_error(n_cherries_wrapper(1:2, 1:3, 4), "same length")
  
  skip_if_not_installed("testthat", "3.1.5") # for expect_no_error
  expect_no_error(Cherries(BalancedTree(144)))
})

nasty <- structure(list(edge = structure(
  c(9, 12, 10, 13, 11, 10, 11, 13, 10, 13, 12, 9,
    5, 10,  1,  2,  3, 13,  9,  4, 11,  7,  8, 6),
  .Dim = c(12, 2)),
  Nnode = 5L,
  tip.label = letters[1:8]),
  class = "phylo") # Danger: Do not plot!

test_that("AddTipEverywhere() correct", {
  backbone <- PectinateTree(5)
  
  expect_equal(7L, length(AddTipEverywhere(backbone, includeRoot = FALSE)))
  expect_equal(7L, length(AddTipEverywhere(UnrootTree(backbone),
                                           includeRoot = FALSE)))
  expect_equal(9L, length(AddTipEverywhere(backbone, includeRoot = TRUE)))
  expect_equal(5L, length(AddTipEverywhere(CollapseNode(backbone, 7:9))))
  
})

test_that("AddTip() at root", {
  expect_true(all.equal(
    AddTip(DropTip(PectinateTree(8), 1), where = 0, label = "t1"),
    PectinateTree(8)))
  AddTip(nasty, 1L)
  AddTip(nasty, 12L)
  expect_true(TRUE) # no crashes!
})

test_that("AddTip() with tip name", {
  bal8 <- BalancedTree(8)
  expect_error(AddTip(bal8, "invalid tip"))
  expect_equal(AddTip(bal8, 1L), AddTip(bal8, "t1"))
  expect_equal(AddTip(nasty, 1L), AddTip(nasty, "a"))
})

test_that("AddTip() with edge lengths", {
  pec8 <- PectinateTree(8)
  pec8$edge.length <- rep(1L, 14L)
  
  add1 <- AddTip(pec8, 0, "Root")
  expect_equal(dim(add1$edge)[1], length(add1[["edge.length"]]))
  
  # case = 1 -> y is bound on the root of x
  expect_equal(
    AddTip(pec8, 0, edgeLength = NULL, lengthBelow = 0.6)$edge.length,
    c(0.6, rep(1, 14), 0.6)
  )
  expect_equal(
    AddTip(pec8, 0, edgeLength = 2, lengthBelow = 0.6)$edge.length,
    c(0.6, rep(1, 14), 2)
  )
  # case = 2 -> y is bound on a tip of x
  expect_equal(
    AddTip(pec8, 2, edgeLength = NULL, lengthBelow = 0.7)$edge.length,
    c(rep(1, 2), 0.3, 0.7, 0.7, rep(1, 11))
  )
  expect_equal(
    AddTip(pec8, 2, edgeLength = 2, lengthBelow = 0.7)$edge.length,
    c(rep(1, 2), 0.3, 0.7, 2, rep(1, 11))
  )
  expect_equal(
    AddTip(pec8, 2, edgeLength = NULL, lengthBelow = NULL)$edge.length,
    c(rep(1, 2), 0.5, 0.5, 0.5, rep(1, 11))
  )
  expect_equal(
    AddTip(pec8, 2, edgeLength = 2, lengthBelow = NULL)$edge.length,
    c(rep(1, 2), 0.5, 0.5, 2, rep(1, 11))
  )
  # case = 3 -> y is bound on a node of x
  expect_equal(
    AddTip(pec8, 15, edgeLength = NULL, lengthBelow = 0.7)$edge.length,
    c(rep(1, 11), 0.3, 0.7, 0.7, 1, 1)
  )
  expect_equal(
    AddTip(pec8, 15, edgeLength = 2, lengthBelow = 0.7)$edge.length,
    c(rep(1, 11), 0.3, 2, 0.7, 1, 1)
  )
  expect_equal(
    AddTip(pec8, 15)$edge.length,
    c(rep(1, 11), 0.5, 0, 0.5, 1, 1)
  )
  expect_equal(
    AddTip(pec8, 15, edgeLength = NULL, lengthBelow = 1)$edge.length,
    c(rep(1, 11), 0, 1, 1, 1, 1)
  )
  expect_equal(
    AddTip(pec8, 15, edgeLength = 2, lengthBelow = 1)$edge.length,
    c(rep(1, 11), 0, 2, 1, 1, 1)
  )
  expect_equal(
    AddTip(pec8, 15, edgeLength = NULL, lengthBelow = 1.6)$edge.length,
    c(rep(1, 11), -0.6, 1.6, 1.6, 1, 1)
  )
  expect_equal(
    AddTip(pec8, 15, edgeLength = 2, lengthBelow = 1.6)$edge.length,
    c(rep(1, 11), -0.6, 2, 1.6, 1, 1)
  )
})

test_that("AddTip() handles node labels", {
  bal6 <- BalancedTree(6)
  startLabels <- paste("Node", 7:11)
  bal6[["node.label"]] <- startLabels
  if (interactive()) {
    plot(bal6, show.node.label = TRUE)
  }
  
  # case = 1 -> y is bound on the root of x
  expect_equal(
    AddTip(bal6, where = 0)[["node.label"]],
    c("", startLabels)
  )
  
  # case = 2 -> y is bound on a tip of x
  expect_equal(
    AddTip(bal6, where = 1)[["node.label"]],
    c(startLabels[1:3], "", startLabels[4:5])
  )
  
  # case = 3 -> y is bound on a node of x
  expect_equal(
    AddTip(bal6, where = 11)[["node.label"]],
    c(startLabels[1:4], "", startLabels[[5]])
  )
  
  expect_equal(AddTipEverywhere(bal6)[[1]][["node.label"]],
               AddTip(bal6, where = 1)[["node.label"]])
})

test_that("AddTip(lengthBelow = NA)", {
  tree <- BalancedTree(10)
  tree$edge.length <- 1 + (1:18 / 100)
  tree$node.label <- paste("n", 11:19)
  
  # Case 1: At root
  at11 <- AddTip(tree, 11, "NEW_TIP", lengthBelow = NA)
  expect_equal(at11$edge, rbind(tree$edge + ifelse(tree$edge > 10, 1, 0),
                                c(12, 11)))
  expect_equal(at11$edge.length, c(tree$edge.length, 0))
  expect_equal(at11$node.label, tree$node.label)
  
  # Case 2: At leaf
  at5 <- AddTip(tree, 5, "NEW_TIP", lengthBelow = NA)
  new5 <- tree$edge + ifelse(tree$edge > 10, 1, 0)
  expect_equal(at5$edge, AddTip(tree, 5)$edge)
  expect_equal(at5$edge.length[-10:-11], tree$edge.length)
  expect_equal(at5$edge.length[10:11], c(0, 0))
  expect_equal(at5$node.label[-6], tree$node.label)
  expect_equal(at5$node.label[6], "")
  
  # Case 3: Internal node
  at15 <- AddTip(tree, 15, "NEW_TIP", lengthBelow = NA)
  expect_equal(at15$edge[-8, ], tree$edge + ifelse(tree$edge > 10, 1, 0))
  expect_equal(at15$edge[8, ], c(16, 11))
  expect_equal(at15$edge.length[-8], tree$edge.length)
  expect_equal(at15$edge.length[8], 0)
  expect_equal(at15$node.label, tree$node.label)
})

test_that("AddTipEverywhere() handles nasty tree", {
  added <- AddTipEverywhere(nasty)
  lapply(added, function(tr) expect_true(all(tr$edge > 0)))
  expect_true(all.equal(lapply(added, Preorder),
                        AddTipEverywhere(Preorder(nasty))))
})

test_that("AddTipEverywhere() with tiny trees", {
  added <- AddTipEverywhere(StarTree(2))
  lapply(added, function(tr) expect_true(all(tr$edge > 0)))
  expect_equal(2, length(added))
  expect_equal(3, length(AddTipEverywhere(StarTree(2), include = TRUE)))
  
  expect_equal(list(PectinateTree(c("t1", "New tip"))),
               AddTipEverywhere(StarTree(1)))
  expect_equal(list(SingleTaxonTree("New tip")),
               AddTipEverywhere(structure(list(tip.label = character(0)),
                                          class = "phylo")))
})

test_that("match.multiPhylo()", {
  t1 <- BalancedTree(6)
  t2 <- PectinateTree(7)
  t9 <- StarTree(99)
  trees <- c(t2, t1, StarTree(5), t2, t1)
  
  expect_true(t1 %in% t1)
  expect_false(t9 %in% t1)
  expect_false(t1 %in% t9)
  expect_true(t2 %in% trees)
  expect_false(t9 %in% trees)
  expect_equal(trees %in% t1, c(FALSE, TRUE, FALSE, FALSE, TRUE))
  
  expect_equal(trees[1:2] %in% trees, c(TRUE, TRUE))
  expect_equal(trees %in% trees[1:2], c(TRUE, TRUE, FALSE, TRUE, TRUE))
  
  
  expect_equal(match(t1, t9, nomatch = 123), 123)
  expect_equal(match(t1, t1, nomatch = 123), 1)
  expect_equal(match(t2, c(t1, t2)), 2)
  expect_equal(match(t9, c(t1, t2)), NA_integer_)

  expect_equal(match(c(t1, t2), t2),c(NA, 1))
  expect_equal(match(c(t1, t2), t9), rep(NA_integer_, 2))
  
  expect_equal(match(c(t1, t2), c(t2, t1, t9)), match(1:2, 2:0))
  expect_equal(match(c(t1, t2, t9), c(t2, t1)), match(1:3, 2:1))
  expect_equal(match(c(t1, t2, t1), c(t9, t9)), match(1:3, 5:6))
})

test_that("PathLengths() works", {
  expect_error(PathLengths(matrix(0, 2, 2)),
               "object of class")
  bal9 <- BalancedTree(9)
  bal9$edge.length <- 1:16
  pl <- PathLengths(bal9, fullMatrix = TRUE)
  expect_equal(dim(pl), rep(17L, 2))
  
  expect_equal(PathLengths(Postorder(bal9)), PathLengths(bal9))
  
  Test <- function(tr) {
    calculated <- PathLengths(tr, full = TRUE)
    edge <- tr$edge
    parent <- edge[, 1]
    child <- edge[, 2]
    weight <- tr[["edge.length"]]
    nVert <- tr$Nnode + NTip(tr)
    anc <- AllAncestors(parent, child)
    parentEdge <- match(seq_len(nVert), child)
    lapply(seq_len(nVert)[-(NTip(tr) + 1L)], function(end) {
      ancs <- anc[[end]]
      expect_equal(calculated[ancs, end],
                   cumsum(weight[parentEdge[c(end, ancs[-length(ancs)])]]))
    })
  }
  Test(bal9)
  
  pec8 <- CollapseNode(PectinateTree(8), 12:13)
  pec8$edge.length <- 1:12
  Test(pec8)
  
  bal9$edge.length <- rep_len(1, 16)
  expect_equal(PathLengths(BalancedTree(9)), PathLengths(bal9))
  
})

test_that("FirstMatchingSplit() fails gracefully", {
  bal13 <- BalancedTree(13)
  pec13 <- PectinateTree(13)
  expect_error(FirstMatchingSplit(bal13, raw(13)),
               "Splits")
})

test_that("FirstMatchingSplit() works", {
  bal13 <- as.Splits(BalancedTree(13))
  pec13 <- as.Splits(PectinateTree(13))
  
  expect_equal(
    FirstMatchingSplit(BalancedTree(13), PectinateTree(13), return = "both"),
    FirstMatchingSplit(bal13, pec13, return = "both")
  )
  
  
  firstMatch <- which(bal13 %in% pec13)[[1]]
  expect_equal(FirstMatchingSplit(bal13, pec13), firstMatch)
  
  expect_equal(
    FirstMatchingSplit(bal13, pec13, return = "both"),
    c(firstMatch, match(bal13[[firstMatch]], pec13))
  )
  expect_equal(FirstMatchingSplit(bal13, as.Splits(StarTree(13))), 0)
  expect_equal(FirstMatchingSplit(as.Splits(StarTree(13)), bal13, nomatch = NA),
               NA_integer_)
  
  # Check robustness to label order
  ren13 <- RenumberTips(pec13, TipLabels(13:1))[[-1]]
  # Split 1 is t1, t2, t3; not t11, t12, t13
  # So first match is split 2 (t1:4).
  expect_equal(FirstMatchingSplit(ren13, bal13, return = "both"), c(2, 2))
})

#' Write a phylogenetic tree in Newick format
#'
#' `as.Newick()` creates a character string representation of a phylogenetic
#' tree, in the Newick format, using R's internal tip numbering.
#' Use [`RenumberTips()`] to ensure that the internal numbering follows the
#' order you expect.
#'
#'
#' @param x Object to convert to Newick format.
#' See Usage section for supported classes.
#'
#' @return `as.Newick()` returns a character string representing `tree` in Newick
#' format.
#'
#' @examples
#' trees <- list(BalancedTree(1:8), PectinateTree(8:1))
#' trees <- lapply(trees, RenumberTips, 1:8)
#' as.Newick(trees)
#'
#' @seealso
#' - Retain leaf labels: [`NewickTree()`]
#'
#' - Change R's internal numbering of leaves: [`RenumberTips()`]
#'
#' - Write tree to text or file: [`ape::write.tree()`]
#'
#' @template MRS
#' @export
as.Newick <- function(x) UseMethod("as.Newick")

#' @rdname as.Newick
#' @export
as.Newick.phylo <- function(x) {
  as_newick(x[["edge"]] - 1L)
}

#' @rdname as.Newick
#' @export
as.Newick.list <- function(x) {
  vapply(x, as.Newick, character(1L))
}

#' @rdname as.Newick
#' @export
as.Newick.multiPhylo <- as.Newick.list

#' Write morphological character matrix to TNT file
#' @param dataset Morphological dataset of class `phyDat` or `matrix`.
#' @param filepath Path to file; if `NULL`, returns a character vector.
#' @param comment Optional comment with which to entitle matrix.
#' @param types Optional list specifying where different data types begin.
#' `c(num = 1, dna = 10)` sets characters 1..9 as numeric, 10..end as DNA.
#' @param pre,post Character vector listing text to print before and after the
#' character matrix.  Specify `pre = 'piwe=;` if the matrix is to be analysed
#' using extended implied weighting (`xpiwe=`).
#'
#' @seealso [`ReadTntCharacters()`]
#' @examples
#' data("Lobo", package = "TreeTools")
#'
#' WriteTntCharacters(Lobo.phy)
#'
#' # Read with extended implied weighting
#' WriteTntCharacters(Lobo.phy, pre = "piwe=10;", post = "xpiwe=;")
#'
#' # Write to a file with:
#' # WriteTntCharacters(Lobo.phy, "example_file.tnt")
#' @template MRS
#' @export
WriteTntCharacters <- function(dataset, filepath = NULL,
                                comment = "Dataset written by `TreeTools::WriteTntCharacters()`",
                                types = NULL,
                                pre = "", post = "") {
  UseMethod("WriteTntCharacters")
}

#' @rdname WriteTntCharacters
#' @export
WriteTntCharacters.phyDat <- function(
    dataset,
    filepath = NULL,
    comment = "Dataset written by `TreeTools::WriteTntCharacters()`",
    types = NULL,
    pre = "",
    post = "") {
  WriteTntCharacters(
    PhyDatToMatrix(dataset, parentheses = "[]"),
    filepath,
    comment,
    types,
    pre,
    post)
}

#' @rdname WriteTntCharacters
#' @export
WriteTntCharacters.matrix <- function(dataset, filepath = NULL,
                                       comment = "Dataset written by `TreeTools::WriteTntCharacters()`",
                                       types = NULL,
                                       pre = "", post = "") {
  EOL <- "\n"
  dataset <- gsub("(", "[", fixed = TRUE, dataset)
  dataset <- gsub(")", "]", fixed = TRUE, dataset)

  ret <- paste(
    paste(pre, collapse = "\n"),
    paste0("xread '", paste(comment, collapse = " "), "'"),
    paste(rev(dim(dataset)), collapse = " "),
    if (is.null(types)) {
      paste(rownames(dataset),
            apply(dataset, 1, paste0, collapse = " "),
            collapse = EOL)
    } else {
      typeEnds <- c(unname(types[-1]) - 1L, ncol(dataset))
      paste(paste0("&[", names(types), "]\n"),
            vapply(seq_along(types), function(i)
              paste(rownames(dataset),
                    apply(dataset[, types[i]:typeEnds[i]], 1, paste0,
                          collapse = " "),
                    collapse = EOL),
              character(1)), collapse = EOL)
    },
    ";",
    paste(post, collapse = "\n"),
    sep = EOL)
  if (is.null(filepath)) {
    ret
  } else {
    writeLines(ret, filepath)
  }
}

#' Match nodes and edges between trees
#' 
#' `MatchNodes()` and `MatchEdges()` matches nodes or edges in one tree to
#' entries in the second that denote a clade with identical tip labels.
#' 
#' The current implementation is potentially inefficient.
#' Please contact the maintainer to request a more efficient implementation if
#' this function is proving a bottleneck.
#' 
#' @param x Tree whose nodes are to be matched.
#' @param table Tree containing nodes to be matched against.
#' @param tips Logical specifying whether to return matches for tips;
#' unless `TRUE`, only the matches for internal nodes will be returned.
#' @examples
#' MatchNodes(BalancedTree(8), RootTree(BalancedTree(8)))
#' @inheritParams match
#' @template MRS
#' @family tree navigation
#' @family tree properties
#' @export
MatchEdges <- function(x, table, nomatch = NA_integer_) {
  xEdge <- .Edge(x)
  tableEdge <- .Edge(table)
  xLab <- TipLabels(x)
  tabLab <- TipLabels(table)
  tipMatch <- if (is.null(xLab) && is.null(tabLab)) {
    # Match tips by number
    xTip <- min(xEdge[, 1]) - 1
    tabTip <- min(tableEdge[, 1]) - 1
    `length<-`(seq_len(tabTip), xTip)
  } else {
    match(xLab, tabLab)
  }
  seek <- DescendantTips(xEdge[, 1], xEdge[, 2])
  find <- DescendantTips(tableEdge[, 1], tableEdge[, 2])[, tipMatch]
  find[is.na(find)] <- FALSE
  match(as.data.frame(t(seek)), as.data.frame(t(find)))
}

.Edge <- function(x) {
  UseMethod(".Edge")
}

#' @export
.Edge.phylo <- function(x) {
  x[["edge"]]
}

#' @export
.Edge.numeric <- function(x) {
  x
}

#' @export
.Edge.list <- function(x) {
  if (is.null(x[["edge"]])) {
    cbind(x[[1]], x[[2]])
  } else {
    x[["edge"]]
  }
}

#' @rdname MatchEdges
#' @export
MatchNodes <- function(x, table, nomatch = NA_integer_, tips = FALSE) {
  xEdge <- .Edge(x)
  xLab <- TipLabels(x)
  tableEdge <- .Edge(table)
  tabLab <- TipLabels(table)
  allLab <- union(xLab, tabLab)
  tipMatch <- match(xLab, tabLab)
  seek <- DescendantTips(xEdge[, 1], xEdge[, 2])[, match(allLab, xLab)]
  seek[is.na(seek)] <- FALSE
  find <- DescendantTips(tableEdge[, 1], tableEdge[, 2])[, match(allLab, tabLab)]
  find[is.na(find)] <- FALSE
  findFrame <-as.data.frame(cbind(t(find), allLab %in% tabLab),
                            optional = TRUE)
  matching <- match(as.data.frame(t(seek), optional = TRUE), findFrame)
  matchRoot <- match(as.data.frame(allLab %in% xLab, optional = TRUE),
                     findFrame)
  
  nodeIndex <- c(tableEdge[, 2], length(tabLab) + 1)
  xRoot <- length(xLab) + 1L
  ret <- nodeIndex[matching][order(c(xEdge[, 2], xRoot))]
  ret[xRoot] <- nodeIndex[matchRoot]
  if (!isTRUE(tips)) {
    ret <- ret[-seq_along(xLab)]
  }
  
  # Return:
  `[<-`(ret, is.na(ret), nomatch)
}

.UpdateNodeLabel <- function(new, old, nodeLabel = old[["node.label"]], ...) {
  UseMethod(".UpdateNodeLabel")
}

#' @export
.UpdateNodeLabel.numeric <- function(new, old, nodeLabel = old[["node.label"]],
                                     newTips = TipLabels(old), ...) {
  nodeLabel[MatchNodes(list(edge = new, tip.label = newTips),
             old, tips = FALSE) - NTip(old)]
}

#' @export
.UpdateNodeLabel.phylo <- function(new, old, nodeLabel = old[["node.label"]], ...) {
  nodeLabel[MatchNodes(new, old, tips = FALSE) - NTip(old)]
}

test_that("MatchNodes() works", {
  bal8 <- BalancedTree(8)
  bal16 <- BalancedTree(16)
  if (interactive()) {
    oPar <- par(mfrow = c(2, 1), mar = rep(1, 4), xpd = NA)
    on.exit(par(oPar))
    plot(bal8)
    edgelabels()
    nodelabels()
    par(cex = 0.75)
    plot(bal16)
    edgelabels()
    nodelabels()
  }
  expect_equal(MatchEdges(bal8, bal16), 2:15)
  e16 <- bal16[["edge"]]
  expect_equal(.Edge(bal8$edge), bal8[["edge"]])
  l16 <- structure(list(e16[, 1], e16[, 2]), tip.label = TipLabels(bal16))
  expect_equal(.Edge(l16), e16)
  
  expect_equal(MatchEdges(l16, bal8), c(NA, 1:14, rep(NA, 15)))
  expect_equal(MatchEdges(bal16$edge, bal8$edge), c(NA, 1:14, rep(NA, 15)))
  expect_equal(MatchNodes(bal8, bal16, tips = FALSE), 18:24)
  expect_equal(MatchNodes(bal8, bal16, tips = TRUE), c(1:8, 18:24))
  expect_equal(MatchNodes(BalancedTree(16), bal8, tips = TRUE, nomatch = -1),
               c(1:8, rep(-1, 8), -1, 9:15, rep(-1, 7)))
  
  bal8rn <- bal8
  bal8rn[["tip.label"]][7:8] <- letters[7:8]
  expect_equal(MatchNodes(bal8rn, bal8, tips = TRUE),
               c(1:6, rep(NA, 2), NA, 10:12, NA, 14, NA))
  
  table <- RootTree(bal8, 1)
  if (interactive()) {
    par(cex = 0.9)
    plot(bal8)
    edgelabels()
    nodelabels()
    plot(table)
    edgelabels()
    nodelabels()
  }
  expect_equal(MatchNodes(bal8, table),
               c(9, rep(NA_integer_, 2), 12:15))
  expect_equal(MatchNodes(bal8, table, nomatch = -1),
               c(9, rep(-1, 2), 12:15))
  
  
  if (interactive()) {
    plot(bal8)
    edgelabels()
    nodelabels()
    plot(Postorder(bal8))
    edgelabels()
    nodelabels()
  }
  expect_equal(MatchNodes(bal8, Postorder(bal8)),
               c(9:15))
})

test_that(".UpdateNodeLabel() works", {
  .Node <- function(n) paste("Node", n)
  bal4 <- BalancedTree(4)
  bal4[["node.label"]] <- .Node(5:7)
  bal8 <- BalancedTree(8)
  bal8[["node.label"]] <- .Node(9:15)
  expect_equal(.UpdateNodeLabel(bal4, bal8), .Node(10:12))
  expect_equal(.UpdateNodeLabel(bal8, bal4), c(NA, .Node(5:7), rep(NA, 3)))
  
  expect_equal(.UpdateNodeLabel(bal4$edge, bal8, newTips = TipLabels(4)),
               .Node(10:12))
  expect_equal(.UpdateNodeLabel(bal8$edge, bal4, newTips = TipLabels(8)),
               c(NA, .Node(5:7), rep(NA, 3)))
  
})

#' Identify consistent / conflicting splits
#' 
#' `SplitConsistent()` and `SplitConflict()` determine whether a series of
#' splits `haystack` are consistent with or contradict the focal split `needle`.
#' 
#' @param needle Splits object containing the single split to evaluate
#' @param haystack Splits object, or list thereof, containing the splits to
#' compare against `needle`.
#' @returns `SplitConsistent()` returns a list of logical vectors.
#' Each list item corresponds to an entry in `haystack`, reporting whether each
#' split is consistent with (`TRUE`) or in conflict with (`FALSE`) `needle`.
#' `SplitConflicts()` returns the inverse.
#' @examples
#' splits1 <- as.Splits(BalancedTree(8))
#' splits2 <- as.Splits(PectinateTree(8))
#' summary(splits1[[4]])
#' SplitConsistent(splits1[[4]], splits2)
#' SplitConflicts(splits1[[4]], list(splits1, splits2))
#' @template MRS
#' @family split manipulation functions
#' @export
SplitConsistent <- function(needle, haystack) {
  if (inherits(haystack, "Splits")) {
    haystack <- list(haystack)
  }
  if (!is.list(haystack)) {
    stop("`haystack` must be a Splits object, or a list thereof.")
  }
  split_consistent(needle, haystack, FALSE)
}

#' @rdname SplitConsistent
#' @export
SplitConflicts <- function(needle, haystack) {
  if (inherits(haystack, "Splits")) {
    haystack <- list(haystack)
  }
  if (!is.list(haystack)) {
    stop("`haystack` must be a Splits object, or a list thereof.")
  }
  split_consistent(needle, haystack, TRUE)
}

#' Distances between each pair of trees
#'
#' @param trees List of trees of class `phylo`.
#' @param Func Function returning a distance between two trees.
#' @param valueLength Integer specifying expected length of the value returned
#' by `Func`.
#' @param \dots Additional arguments to `Func`.
#'
#' @return Matrix detailing distance between each pair of trees.
#' Identical trees are assumed to have zero distance.
#' @examples
#' trees <- list(BalancedTree(8), PectinateTree(8), StarTree(8))
#' TCIDiff <- function(tree1, tree2) {
#'   TotalCopheneticIndex(tree1) - TotalCopheneticIndex(tree2)
#' }
#' PairwiseDistances(trees, TCIDiff, 1)
#' TCIRange <- function(tree1, tree2) {
#'   range(TotalCopheneticIndex(tree1), TotalCopheneticIndex(tree2))
#' }
#' PairwiseDistances(trees, TCIRange, 2)
#' @template MRS
#' @family pairwise tree distances
#' @importFrom stats as.dist
#' @export
PairwiseDistances <- function(trees, Func, valueLength = 1L, ...) {
  ret <- array(0, c(length(trees), length(trees), valueLength))
  for (i in seq_along(trees)) {
    trI <- trees[[i]]
    for (j in i + seq_len(length(trees) - i)) {
      val <- Func(trI, trees[[j]], ...)
      ret[j, i, ] <- unlist(val)
    }
  }

  # Return:
  if (valueLength > 1L) {
    structure(lapply(seq_len(valueLength), function(i) {
      as.dist(ret[, , i], upper = TRUE)
    }), names = names(val))
  } else {
    as.dist(ret[, , 1], upper = TRUE)
  }
}

#' Paths present in reduced tree
#' 
#' Lists which paths present in a master tree are present when leaves are
#' dropped.
#' 
#' @param paths `data.frame` of paths in master tree, perhaps generated using
#' [`PathLengths()`].
#' @param keptVerts Logical specifying whether each entry is retained in the
#' reduced tree, perhaps generated using [`KeptVerts()`].
#' @param all Logical: if `TRUE`, return all paths that occur in the reduced
#' tree; if `FALSE`, return only those paths that correspond to a single edge.
#' that correspond to edges in the reduced tree.
#' Ignored if `paths` is a matrix.
#' @return `KeptPaths()` returns a logical vector specifying whether each path
#' in `paths` occurs when `keptVerts` vertices are retained.
#' @family tree manipulation
#' @examples 
#' master <- BalancedTree(9)
#' paths <- PathLengths(master)
#' keptTips <- c(1, 5, 7, 9)
#' keptVerts <- KeptVerts(master, keptTips)
#' KeptPaths(paths, keptVerts)
#' paths[KeptPaths(paths, keptVerts, all = FALSE), ]
#' @export
#' @template MRS
KeptPaths <- function(paths, keptVerts, all = TRUE) UseMethod("KeptPaths")

#' @rdname KeptPaths
#' @export
KeptPaths.data.frame <- function(paths, keptVerts, all = TRUE) {
  kept <- apply(matrix(as.integer(unlist(paths[, 1:2])) %in% which(keptVerts),
                       ncol = 2L), 1, all)
  
  if (!all) {
    kept[kept][duplicated(paths[kept, "end"], fromLast = TRUE)] <- FALSE
  }
  # Return:
  kept
}

#' @rdname KeptPaths
#' @export
KeptPaths.matrix <- function(paths, keptVerts, all = TRUE) {
  kept <- array(FALSE, dim = dim(paths))
  kept[keptVerts, keptVerts] <- TRUE
  kept
}

#' Display time-calibrated tree using tip information only
#' 
#' `TipTimedTree()` plots a phylogenetic tree against time using an
#' _ad hoc_ approach based on dates associated with the leaves.
#' Nodes are dated to the youngest possible value, plus an additional "buffer"
#' (specified with `minEdge`) to ensure that branching order is readable.
#' 
#' This experimental function is liable to change its behaviour, or to
#' be deprecated, in coming releases.
#' Please contact the maintainer if you find it useful, so that a
#' production-ready version can be prioritized.
#'  
#' @template treeParam
#' @param tipAge Numeric vector specifying the age (in units-of-time ago)
#' associated with each tip in `tree$tip.label` in turn.
#' Older ages signify earlier tips.
#' @param minEdge Minimum length of edge to allow (in units-of-time)
#' 
#' @return `TipTimedTree()` returns a tree with edge lengths set based on the
#' ages of each tip.
#' 
#' @examples
#' tree <- BalancedTree(6)
#' plot(TipTimedTree(tree, tipAge = 1:6, minEdge = 2))
#' @family utility functions
#' @family tree manipulation
#' @export
TipTimedTree <- function(tree, tipAge, minEdge = 1) {
  edge <- tree[["edge"]]
  nEdge <- dim(edge)[1]
  if (nEdge < 2) {
    warning("`tree` does not contain multiple edges")
    tree[["edge.length"]] <- rep_len(minEdge, nEdge)
    return(tree)
  }
  nTip <- NTip(tree)
  if (nTip != length(tipAge)) {
    stop("`tipAge` must list one age per leaf in `tree`")
  }
  nNode <- nEdge - nTip
  
  age <- c(tipAge, rep_len(-Inf, max(edge) - nTip))
  
  for (i in PostorderOrder(tree)) {
    age[edge[i, 1]] <- max(age[edge[i, 2]] + minEdge, age[edge[i, 1]])
  }
  tree[["edge.length"]] <- apply(edge, 1, function(i) age[i[1]] - age[i[2]])
  # Return:
  tree
}

#' Read modification time from "ape" Nexus file
#'
#' `ApeTime()` reads the time that a tree written with "ape" was modified,
#' based on the comment in the Nexus file.
#'
#' @param filepath Character string specifying path to the file.
#' @param format Format in which to return the time: "double" as a sortable numeric;
#'               any other value to return a string in the format
#'               `YYYY-MM-DD hh:mm:ss`.
#'
#' @return `ApeTime()` returns the time that the specified file was created by
#' ape, in the format specified by `format`.
#' @template MRS
#' @export
ApeTime <- function(filepath, format = "double") {
  if (length(filepath) > 1L) {
    stop("`filepath` must be a character string of length 1")
  }
  comment <- readLines(filepath, n = 2)[2]
  Month <- function(month) {
    months <- c("Jan", "Feb", "Mar", "Apr", "May", "Jun",
                "Jul", "Aug", "Sep", "Oct", "Nov", "Dec")
    whichMonth <- months == month
    if (any(whichMonth)) {
      formatC(which(whichMonth), width = 2, flag = "0")
    } else {
      month                                                                     # nocov
    }
  }
  DATEEXP <- ".*? (\\w+)\\s(\\d+)\\s(\\d+\\:\\d\\d\\:\\d\\d)\\s(\\d\\d\\d\\d).*"
  time <- paste0(gsub(DATEEXP, "\\4-", comment),
                 Month(gsub(DATEEXP, "\\1", comment)),
                 gsub(DATEEXP, "-\\2 \\3", comment))

  # Return:
  ifelse(format == "double", as.numeric(as.POSIXct(time, tz = "GMT")), time)
}

#' Extract taxa from a matrix block
#'
#' Extract leaf labels and character states from a Nexus-formatted matrix.
#'
#' @param matrixLines Character vector containing lines of a file that include
#' a phylogenetic matrix. See [`ReadCharacters()`] for expected format.
#' @template characterNumParam
#' @param continuous Logical specifying whether characters are continuous.
#' Treated as discrete if `FALSE`.
#'
#' @return `ExtractTaxa()` returns a matrix with _n_ rows, each named for the
#' relevant taxon, and _c_ columns,
#' each corresponding to the respective character specified in `character_num`.
#'
#' @examples
#' fileName <- paste0(system.file(package = "TreeTools"),
#'                    "/extdata/input/dataset.nex")
#' matrixLines <- readLines(fileName)[6:11]
#' ExtractTaxa(matrixLines)
#'
#' @keywords internal
#' @export
ExtractTaxa <- function(matrixLines, character_num = NULL,
                         continuous = FALSE) {
  taxonLine.pattern <- "('([^']+)'|\"([^\"+])\"|(\\S+))\\s+(.+)$"
  # Also recognise taxon-name-only lines (name without data on the same line,
  # used e.g. in TNT files where data runs across multiple lines per taxon)
  nameOnly.pattern <- "^[\\p{L}\\p{Pi}\\p{Pf}][^\\s]*$"

  taxonLines <- regexpr(taxonLine.pattern, matrixLines, perl = TRUE) > -1
  nameOnlyLines <- grepl(nameOnly.pattern, matrixLines, perl = TRUE)
  taxonLines <- taxonLines | nameOnlyLines
  # If a line does not start with a taxon name, join it to the preceding line
  taxonLineNumber <- which(taxonLines)
  previousTaxon <- vapply(which(!taxonLines), function(x) {
    max(taxonLineNumber[taxonLineNumber < x])
  }, integer(1))


  taxa <- sub(taxonLine.pattern, "\\2\\3\\4", matrixLines, perl = TRUE)
  taxa <- gsub(" ", "_", taxa, fixed=TRUE)
  # Strip TNT @taxonomy classification suffixes (e.g. Name_@Family_Genus)
  taxa <- sub("@\\S*$", "", taxa, perl = TRUE)
  taxa <- sub("_+$", "", taxa, perl = TRUE)  # remove trailing underscores
  taxa[!taxonLines] <- taxa[previousTaxon]
  uniqueTaxa <- unique(taxa)

  tokens <- sub(taxonLine.pattern, "\\5", matrixLines, perl = TRUE)
  tokens[nameOnlyLines] <- ""  # name-only lines carry no character data
  if (continuous) {
    tokens <- strsplit(tokens, "\\s+")
    lengths <- lengths(tokens)
    if (length(unique(lengths)) != 1) {
      stop("Different numbers of tokens in different taxa: ",                   # nocov
           paste(lengths, collapse = ", "))                                     # nocov
    }
    tokens <- t(vapply(tokens, I, tokens[[1]]))
  } else {
    tokens <- gsub("\t", "", gsub(" ", "", tokens, fixed = TRUE), fixed = TRUE)
    tokens <- vapply(uniqueTaxa,
                     function(taxon) paste0(tokens[taxa == taxon],
                                             collapse = ""),
                     character(1))
    tokens <- NexusTokens(tokens, character_num = character_num)
  }

  rownames(tokens) <- uniqueTaxa

  # Return:
  tokens
}

#' @rdname ExtractTaxa
#' @param tokens Vector of character strings corresponding to phylogenetic
#'  tokens.
#' @return `NexusTokens()` returns a character vector in which each entry
#' corresponds to the states of a phylogenetic character, or a list containing
#' an error message if input is invalid.
#' @examples
#' NexusTokens("01[01]-?")
#' @export
NexusTokens <- function(tokens, character_num = NULL) {
  tokens.pattern <- "\\([^\\)]+\\)|\\[[^\\]]+\\]|\\{[^\\}]+\\}|\\S"
  matches <- gregexpr(tokens.pattern, tokens, perl = TRUE)

  nChar <- length(matches[[1]])

  if (!exists("character_num") || is.null(character_num)) {
    character_num <- seq_len(nChar)
  } else if (any(character_num > nChar) || any(character_num < 1)) {
    character_num[character_num < 1] <- 1L
    character_num[character_num > nChar] <- nChar
    return(list(paste0("Character number must be between 1 and ", nChar, ".")))
  }

  tokens <- t(vapply(regmatches(tokens, matches),
                     function(x) x[character_num, drop=FALSE],
                     character(length(character_num))))
  if (length(character_num) == 1) {
    tokens <- t(tokens)
  } else if (length(character_num) == 0) {
    stop("No characters selected")
  }

  # Return:
  tokens
}

#' Read phylogenetic characters from file
#'
#' Parse a Nexus \insertCite{Maddison1997}{TreeTools} or 
#' TNT \insertCite{Goloboff2008}{TreeTools} file, reading character states and
#' names.
#'
#' Tested with matrices downloaded from [MorphoBank](https://morphobank.org/)
#' \insertCite{OLeary2011}{TreeTools}, but should also work more widely; please
#' [report](https://github.com/ms609/TreeTools/issues/new?title=Error+parsing+Nexus+file&body=<!--Tell+me+more+and+attach+your+file...-->)
#' incompletely or incorrectly parsed files.
#'
#' Matrices must contain only continuous or only discrete characters;
#' maximum one matrix per file.  Continuous characters will be read as strings
#' (i.e. base type "character").
#'
#' The encoding of an input file will be automatically determined by R.
#' Errors pertaining to an `invalid multibyte string` or
#' `string invalid at that locale` indicate that R has failed to detect
#' the appropriate encoding.  Either
#' [re-save the file](
#' https://support.posit.co/hc/en-us/articles/200532197-Character-Encoding-in-the-RStudio-IDE)
#' in a supported encoding (`UTF-8` is a good choice) or
#' specify the file encoding (which you can find by, for example, opening in
#' [Notepad++](https://notepad-plus-plus.org/downloads/) and identifying
#' the highlighted option in the "Encoding" menu) following the example below.
#'
#' @param filepath character string specifying location of file, or a
#' [connection][base::connections] to the file.
#' @param type Character vector specifying categories of data to extract from
#' file. Setting `type = c("num", "dna")` will return only characters
#' following a `&[num]` or `&[dna]` tag in a TNT input file, listing `num`
#' character blocks before `dna` characters.
#' Leave as `NULL` (the default) to return all characters in their original
#' sequence.
#' @template characterNumParam
#' @param encoding Character encoding of input file.
#'
#' @return `ReadCharacters()` and `ReadTNTCharacters()` return a matrix whose
#' row names correspond to tip labels, and
#' column names correspond to character labels, with the
#' attribute `state.labels` listing the state labels for each character; or
#' a list of length one containing a character string explaining why the
#' function call was unsuccessful.
#'
#' `ReadAsPhyDat()` and `ReadTntAsPhyDat()` return a `phyDat` object.
#'
#' @references \insertAllCited{}
#'
#' @examples
#' fileName <- paste0(system.file(package = "TreeTools"),
#'                    "/extdata/input/dataset.nex")
#' ReadCharacters(fileName)
#'
#' fileName <- paste0(system.file(package = "TreeTools"),
#'                    "/extdata/tests/continuous.nex")
#'
#' continuous <- ReadCharacters(fileName, encoding = "UTF8")
#'
#' # To convert from strings to numbers:
#' at <- attributes(continuous)
#' continuous <- suppressWarnings(as.numeric(continuous))
#' attributes(continuous) <- at
#' continuous
#' @template MRS
#'
#' @seealso
#' - Convert between matrices and `phyDat` objects: [`MatrixToPhyDat()`]
#'
#' - Write characters to TNT-format file: [`WriteTntCharacters()`]
#' @export
ReadCharacters <- function(filepath, character_num = NULL, encoding = "UTF8") {

  lines <- .UTFLines(filepath, encoding)
  nexusComment.pattern <- "\\[[^\\]*?\\]"
  lines <- gsub(nexusComment.pattern, "", lines)
  lines <- trimws(lines)
  lines <- lines[lines != ""]

  semicolons <- which(RightmostCharacter(lines) == ";")
  upperLines <- toupper(lines)

  continuous <- length(grep("DATATYPE[\\S\\=]+CONTINUOUS", upperLines)) > 0
  matrixStart <- which(upperLines == "MATRIX")
  if (length(matrixStart) == 0) {
    return(list("MATRIX block not found in Nexus file."))
  } else if (length (matrixStart) > 1) {
    return(list("Multiple MATRIX blocks found in Nexus file."))                 # nocov
  } else {

    matrixEnd <- semicolons[semicolons > matrixStart][1]
    if (lines[matrixEnd] == ";") matrixEnd <- matrixEnd - 1

    matrixLines <- lines[(matrixStart + 1):matrixEnd]
    tokens <- ExtractTaxa(matrixLines, character_num = character_num,
                          continuous = continuous)

    nChar.pattern <- "DIMENSIONS\\s+.*NCHAR\\s*=\\s*(\\d+)"
    nCharLines <- grepl(nChar.pattern, upperLines, perl = TRUE)
    nChar <- .RegExpMatches("DIMENSIONS\\s+.*NCHAR\\s*=\\s*(\\d+)", upperLines[nCharLines])
    if (length(unique(nChar)) > 1) {
      return(list("Inconsistent DIMENSIONS NCHAR= counts in Nexus file."))
    }
    nChar <- as.integer(nChar[1])
    if (is.null(character_num)) {
      character_num <- seq_len(ncol(tokens))
    }

    ## Written with MorphoBank format in mind: each label on separate line,
    ## each character introduced by integer and terminated with comma.
    labelStart <- grep("^\\s*CHARLABELS\\s*$", upperLines,
                       ignore.case = TRUE, perl = TRUE)
    if (length(labelStart) == 1) {
      labelEnd <- semicolons[semicolons > labelStart][1]
      if (lines[labelEnd] == ";") labelEnd <- labelEnd - 1
      #attr(dat, "char.labels")
      colnames(tokens) <- Unquote(
        .UnescapeQuotes(lines[labelStart + character_num])
      )
    } else {
      if (length(labelStart) > 1)
        return(list("Multiple CharLabels blocks in Nexus file."))
    }

    stateStart <- grep("^\\s*STATELABELS\\s*$", upperLines,
                                 ignore.case = TRUE, perl = TRUE)
    if (length(stateStart) == 1) {
      stateEnd <- semicolons[semicolons > stateStart][1]
      if (is.na(stateEnd)) {
        stop("STATELABELS block missing closing semicolon;")
      }
      stateLines <- lines[stateStart:stateEnd]
      # MorphoBank sometimes uses a standalone ';' to close the block rather
      # than terminating the last character entry with ';' directly. Exclude
      # it so it isn't counted as a character terminator.
      if (stateLines[length(stateLines)] == ";") {
        stateLines <- stateLines[-length(stateLines)]
      }
      stateStarts <- grep("^\\d+", stateLines)
      stateEnds <- grep("[,;]$", stateLines)
      if (length(stateStarts) != length(stateEnds)) {
        warning("Could not parse character states; does each end with a ' or ;?.")
      } else {
        if (length(character_num) != length(stateStarts)) {
          stateNos <- as.integer(stateLines[stateStarts])
          if (all(!is.na(stateNos))) {
            warning("Missing character state definition for: ",
                    paste0(setdiff(character_num, stateNos), collapse = ", "))
          } else {
            warning("More characters than character state definitions.")
          }
        }

        attr(tokens, "state.labels") <-
          lapply(character_num[seq_along(stateStarts)], function(i)
            Unquote(
              .UnescapeQuotes(
                stateLines[(stateStarts[i] + 1):(stateEnds[i] - 1)]
              )
            )
          )
      }
    } else {
      if (length(labelStart) > 1) {
        return(list("Multiple StateLabels blocks in Nexus file."))
      }
    }

    charStateLabelsStart <- grep("^\\s*CHARSTATELABELS\\s*$", upperLines,
                                 ignore.case = TRUE, perl = TRUE)
    endCommand <- grep(";$", upperLines, perl = TRUE)

    if (length(charStateLabelsStart)) {
      if (length(charStateLabelsStart) > 1) {
        return(list("Multiple CHARSTATELABELS blocks found in Nexus file."))
      }
      charStateLabelsEnd <- endCommand[endCommand > charStateLabelsStart][1]
      if (length(charStateLabelsEnd) == 0) {
        return(list("Unterminated CHARSTATELABELS block in Nexus file."))
      }

      # csl: charStateLabels
      csl <- lines[(charStateLabelsStart + 1):charStateLabelsEnd]
      quote.pattern <- "'[^']*'"
      cslEscapes <- unlist(regmatches(csl, gregexpr(quote.pattern, csl, perl = TRUE)))
      cslEscapes <- substr(cslEscapes, 2L, vapply(cslEscapes, nchar, 1) - 1L)

      cslEscaped <- strsplit(paste(gsub(quote.pattern,
                                        "__TREETOOLS_ESCAPE_SEQUENCE__", csl,
                                        perl = TRUE), collapse = "  "),
                             ",\\s+\\d+", perl = TRUE)[[1]]
      nCsl <- length(cslEscaped)
      if (nCsl != nChar) {
        return(list(paste0("CHARSTATELABELS length (", nCsl,
                           ") does not match NCHAR (", nChar, ")")))
      }
      cslEscaped[1] <- sub("^\\s*1\\b", "", cslEscaped[1], perl = TRUE)
      cslEscaped[nCsl] <- sub("\\s*;\\s*$", "", cslEscaped[nCsl], perl = TRUE)
      cslEscaped <- trimws(sub("\\s*/\\s*", "__TREETOOLS_ESCAPE_SPLITTER__", cslEscaped, perl = TRUE))
      cslEscaped <- gsub("\\s+", "__TREETOOLS_ESCAPE_LABEL_SPLITTER__", cslEscaped, perl = TRUE)
      cslEscaped <- paste(cslEscaped, collapse = "__TREETOOLS_ESCAPE_CHAR_SPLITTER__")
      for (escape in cslEscapes) {
        cslEscaped <- sub("__TREETOOLS_ESCAPE_SEQUENCE__", escape, cslEscaped, fixed = TRUE)
      }
      cslEscaped <- strsplit(cslEscaped, "__TREETOOLS_ESCAPE_CHAR_SPLITTER__", fixed = TRUE)[[1]]
      cslEscaped <- do.call(rbind, strsplit(cslEscaped, "__TREETOOLS_ESCAPE_SPLITTER__", fixed = TRUE))
      colnames(tokens) <- gsub("_", " ", cslEscaped[, 1])[character_num]
      attr(tokens, "state.labels") <- lapply(
        strsplit(cslEscaped[character_num, 2], "__TREETOOLS_ESCAPE_LABEL_SPLITTER__", fixed = TRUE),
        gsub, pattern = "_", replacement = " ", fixed = TRUE)
    }
  }

  # Return:
  tokens
}


#' @rdname ReadCharacters
#' @return `ReadTntCharacters()` and `ReadTNTCharacters()` return a matrix
#' as described above for `ReadCharacters()`.  When the TNT file contains an
#' `xgroup` partition block \insertCite{Goloboff2008}{TreeTools}, the returned
#' matrix carries an `"xgroup"` attribute: a factor of length `ncol(matrix)`
#' whose levels are the partition labels (the parenthetical label, e.g.\
#' `"ANTERIOR"`, or the numeric id as a string when no label is given).
#' Characters not assigned to any partition are `NA`.  The attribute is absent
#' (not an all-`NA` vector) when no `xgroup` block is found in the file.
#'
#' @examples
#' tntFile <- paste0(system.file(package = "TreeTools"),
#'                   "/extdata/tests/tnt-xgroup.tnt")
#' mat <- ReadTntCharacters(tntFile)
#' attr(mat, "xgroup")
#' @export
ReadTntCharacters <- function(filepath, character_num = NULL,
                               type = NULL, encoding = "UTF8") {

  lines <- .UTFLines(filepath, encoding)
  tntComment.pattern <- "'[^']*'"
  lines <- gsub(tntComment.pattern, "", lines, perl = TRUE)
  multilineComments <- grep("'", lines, fixed = TRUE)
  nmlc <- length(multilineComments) / 2
  openComment <- multilineComments[seq_len(nmlc) * 2L - 1L]
  closeComment <- multilineComments[seq_len(nmlc) * 2L]
  lines[openComment] <- gsub("'.*", "", lines[openComment])
  lines[closeComment] <- gsub(".*'", "", lines[closeComment])
  if (nmlc > 0) {
    for (i in seq_len(nmlc)) {
      innerStart <- openComment[i] + 1L
      innerEnd <- closeComment[i] - 1L
      if (innerStart <= innerEnd) lines[innerStart:innerEnd] <- ""
    }
  }

  lines <- trimws(lines)
  lines <- lines[lines != ""]

  semicolons <- grep(";", lines, fixed = TRUE)
  upperLines <- toupper(lines)

  xread <- grep("\\bXREAD\\b", lines, ignore.case = TRUE, perl = TRUE)
  if (length(xread) < 1) return(NULL)
  if (length(xread) > 1) {
    message("Multiple character blocks not yet supported;",
            "contact 'TreeTools' maintainer to request.",
            "Returning first block only.")
    xread <- xread[1]
  }
  # If xread appears mid-line (e.g. after other TNT directives separated by ;),
  # strip everything before the xread keyword so dimension parsing works.
  lines[xread] <- sub("^.*\\bxread\\b", "xread", lines[xread],
                       ignore.case = TRUE, perl = TRUE)

  xreadEnd <- semicolons[semicolons > xread][1]
  if (lines[xreadEnd] == ";") {
    xreadEnd <- xreadEnd - 1L
  }
  xreadLines <- lines[xread:xreadEnd]
  xDimPos <- regexec("'?\\s*(\\d+)\\s*(\\d+)\\s*$", xreadLines)
  xDimLine <- which(vapply(xDimPos, `[`, 1, 1) > -1)[1]
  dimText <- xreadLines[xDimLine]
  dimHit <- xDimPos[[xDimLine]]
  nChar <- as.integer(substr(dimText, dimHit[2], dimHit[2] +
                               attr(dimHit, "match.length")[2] - 1L))
  nTip <- as.integer(substr(dimText, dimHit[3], dimHit[3] +
                              attr(dimHit, "match.length")[3] - 1L))
  matrixLines <- xreadLines[-seq_len(xDimLine)]

  bareAmpLines <- grep("^&\\s*$", matrixLines, perl = TRUE)
  if (length(bareAmpLines)) matrixLines <- matrixLines[-bareAmpLines]
  ctypeLines <- grep("^&\\[[\\w\\s]+\\]$", matrixLines, perl = TRUE)
  if (is.null(type)) {
    if (length(ctypeLines)) matrixLines <- matrixLines[-ctypeLines]
  } else {
    types <- matrixLines[ctypeLines]
    blocks <- lapply(paste0("\\b", type, "\\b"), grep, types, ignore.case = TRUE)
    nBlocks <- lengths(blocks)
    if (any(nBlocks == 0)) {
      message("Tags ", paste0(type[nBlocks == 0], collapse = ", "),
      " not found. Ignored: ", types[!seq_along(types) %fin% unlist(blocks)])
    }
    if (all(nBlocks == 0L)) return(NULL)
    blockSpan <- cbind(ctypeLines + 1L, c(ctypeLines[-1] - 1, length(matrixLines)))
    matrixLines <- matrixLines[unlist(
      apply(blockSpan[unlist(blocks), , drop = FALSE],  1,
            function(x) seq.int(from = x[1], to = x[2])))]
  }

  # Some TNT files pack multiple taxa on one physical line (e.g. dromaeodat.tnt).
  # Split such lines at boundaries between character data and the next taxon name.
  # Boundary: a data character (digit, ?, -, ], }) immediately followed by a taxon
  # name (letter, then eventually @). No whitespace separator between them.
  multiTaxon <- grep(
    "(?<=[?0-9\\-\\]\\}])[A-Za-z][^\t ]*@",
    matrixLines, perl = TRUE
  )
  if (length(multiTaxon)) {
    splitLines <- strsplit(
      matrixLines[multiTaxon],
      "(?<=[?0-9\\-\\]\\}])(?=[A-Za-z][^\t ]*@)",
      perl = TRUE
    )
    matrixLines <- c(
      matrixLines[-multiTaxon],
      unlist(splitLines)
    )
  }

  tokens <- ExtractTaxa(matrixLines, character_num)
  if (nrow(tokens) != nTip) {
    warning("Extracted ", nrow(tokens), " taxa, but TNT file specifies ", nTip, # nocov
            ": please check output and report bugs.")                           # nocov
  }
  labelStart <- which(upperLines == "CHARLABELS")
  if (length(labelStart) == 1) {
    labelEnd <- semicolons[semicolons > labelStart][1]
    if (lines[labelEnd] == ";") labelEnd <- labelEnd - 1
    #attr(dat, "char.labels")
    colnames(tokens) <- lines[labelStart + character_num]
  } else {
    if (length(labelStart) > 1)
      return(list("Multiple CharLabels blocks in Nexus file."))
  }


  labelStart <- which(upperLines == "CNAMES")
  if (length(labelStart) == 1) {
    labelEnd <- semicolons[semicolons > labelStart][1]
    if (lines[labelEnd] == ";") labelEnd <- labelEnd - 1
    charLines <- lines[labelStart + character_num]
    charLine.pattern <- "^\\S+\\s\\d+\\s(\\w+)(.*)\\s*;\\s*$"

    # Character labels
    charNames <- gsub(charLine.pattern, "\\1", charLines, perl = TRUE)
    colnames(tokens) <- gsub("_", " ", charNames, fixed = TRUE)

    # State labels
    stateNames <- gsub(charLine.pattern, "\\2", charLines, perl = TRUE)
    attr(tokens, "state.labels") <- lapply(stateNames, function(line) {
      states <- strsplit(trimws(line), "\\s+", perl = TRUE)[[1]]
      trimws(gsub("_", " ", states, fixed = TRUE))
    })
  } else {
    if (length(labelStart) > 1)
      return(list("Multiple cnames entries in TNT file."))
  }

  # Attach xgroup attribute if the file contains an xgroup partition block.
  # Search only within lines following the xread block (up to the next proc/ or
  # end of file) so multi-block files scope correctly.
  xgroupAttr <- .ParseXgroupLines(lines[xread:length(lines)], nChar)
  if (!is.null(xgroupAttr)) {
    attr(tokens, "xgroup") <- xgroupAttr
  }

  # Return:
  tokens
}

#' @rdname ReadCharacters
#' @export
ReadTNTCharacters <- ReadTntCharacters


#' @keywords internal
.ParseXgroupLines <- function(lines, nChar) {
  xgLines <- grep("^xgroup\\s*=", lines, ignore.case = TRUE, perl = TRUE)
  if (length(xgLines) == 0L) {
    return(NULL)
  }

  labels <- character(nChar)
  labels[] <- NA_character_

  # Collect labels in encounter order to build factor levels correctly
  levelsSeen <- character(0)

  for (idx in xgLines) {
    parsed <- .ParseOneXgroup(lines[idx], nChar)
    lbl <- parsed[["label"]]
    labels[parsed[["chars"]]] <- lbl
    if (!lbl %in% levelsSeen) {
      levelsSeen <- c(levelsSeen, lbl)
    }
  }

  # Return:
  factor(labels, levels = levelsSeen)
}

#' @keywords internal
.ParseOneXgroup <- function(line, nChar) {
  idM <- regmatches(line, regexec("xgroup\\s*=\\s*(\\d+)", line,
                                  ignore.case = TRUE, perl = TRUE))[[1]]
  id <- as.integer(idM[2])

  labelM <- regmatches(line, regexec("\\(([^)]+)\\)", line, perl = TRUE))[[1]]
  label <- if (length(labelM) > 1L) trimws(labelM[2]) else as.character(id)

  # Range tokens follow the optional label; if no label, follow the partition id
  afterLabel <- sub(".*\\)\\s*", "", line)
  if (!nzchar(afterLabel)) {
    afterLabel <- sub("xgroup\\s*=\\s*\\d+\\s*", "", line,
                      ignore.case = TRUE, perl = TRUE)
  }
  tokens <- regmatches(afterLabel,
                       gregexpr("\\d+\\.\\d*", afterLabel, perl = TRUE))[[1]]

  list(id    = id,
       label = label,
       chars = unlist(lapply(tokens, .ExpandTntRange, nChar = nChar)))
}

#' @keywords internal
.ExpandTntRange <- function(token, nChar) {
  parts <- strsplit(token, ".", fixed = TRUE)[[1]]
  # TNT is 0-indexed; convert to 1-indexed
  from <- as.integer(parts[1]) + 1L
  to <- if (length(parts) > 1L && nzchar(parts[2])) {
    as.integer(parts[2]) + 1L
  } else {
    nChar
  }
  seq.int(from, to)
}

.UTFLines <- function(filepath, encoding) {
  if (!file.exists(filepath)) {
    stop("File '", filepath, "' not found.")
  }
  con <- file(filepath, encoding = encoding)
  on.exit(close(con))
  
  tryCatch(
    # Missing EOL might occur in user-generated file, so warning not helpful
    try1 <- enc2utf8(readLines(con, warn = FALSE)),
    warning = function(e) {
      if (substr(e[["message"]], 0, 39) == 
          "invalid input found on input connection") {
        newEnc <- if (toupper(encoding) %in% c("UTF-8", "UTF8")) {
          "cp1252"
        } else {
          "UTF8"
        }
        message("Problem reading characters; trying ", newEnc, " file encoding")
        close(con)
        con <- file(filepath, encoding = newEnc)
        enc2utf8(readLines(con, warn = FALSE))
      } else {
        try1
      }
    }
  )
}

.RegExpMatches <- function(pattern, string, i = TRUE, nMatch = 1) {
  res <- regexpr(pattern, string, perl = TRUE, ignore.case = i)
  matches <- regmatches(string, res)
  res[res < 0] <- NA
  res[!is.na(res)] <-
    gsub(pattern, paste0(paste0("\\", seq_len(nMatch)), collapse = ""),
       matches, ignore.case = i, perl = TRUE)
  res
}

#' @rdname ReadCharacters
#' @return `ReadNotes()` returns a list in which each entry corresponds to a
#' single character, and itself contains a list of with two elements:
#'
#' 1. A single character object listing any notes associated with the character
#' 2. A named character vector listing the notes associated with each taxon
#' for that character, named with the names of each note-bearing taxon.
#'
#' @importFrom stats setNames
#' @export
ReadNotes <- function(filepath, encoding = "UTF8") {
  taxon.pattern <- "^\\s+[\"']?([^;]*?)[\"']?\\s*$"
  nTax.pattern <- "DIMENSIONS\\s+.*NTAX\\s*=\\s*(\\d+)"

  lines <- .UTFLines(filepath, encoding)
  upperLines <- toupper(lines)
  trimUpperLines <- trimws(upperLines)

  notesStart <- which(trimUpperLines == "BEGIN NOTES;")
  endBlocks <- which(trimUpperLines %fin% c("END;", "ENDBLOCK;"))
  taxLabels <- which(trimUpperLines == "TAXLABELS")
  semicolons <- which(trimUpperLines == ";")
  nTaxLines <- grepl(nTax.pattern, trimUpperLines, perl = TRUE)


  if (length(notesStart) == 0) {
    return(list("NOTES block not found in Nexus file."))
  } else if (length(notesStart) > 1) {
    return(list("Multiple NOTES blocks found in Nexus file."))
  } else if (length(taxLabels) > 1) {
    return(list("Multiple TAXLABELS found in Nexus file."))
  } else if (!any(nTaxLines)) {
    return(list("No DIMENSIONS NTAX= statment found in Nexus file."))
  } else {
    if (length(taxLabels) == 0) {
      taxa <- names(ReadAsPhyDat(filepath))
    } else {

      nTax <- .RegExpMatches(nTax.pattern, trimUpperLines[nTaxLines])
      if (length(unique(nTax)) > 1) {
        return(list("Inconsistent DIMENSIONS NTAX= counts in Nexus file."))
      }
      nTax <- as.integer(nTax[1])

      taxaEnd <- semicolons[semicolons > taxLabels][1] - 1L
      taxaLines <- lines[(taxLabels + 1):taxaEnd]
      taxon.matches <- grepl(taxon.pattern, taxaLines, perl = TRUE)
      if (sum(taxon.matches) == nTax) {
        taxa <- gsub(taxon.pattern, "\\1", taxaLines[taxon.matches], perl = TRUE)
        taxa <- gsub(" ", "_", taxa, fixed = TRUE)
      } else {
        taxa <- gsub(taxon.pattern, "\\1", taxaLines[taxon.matches], perl = TRUE)
        taxa <- unlist(strsplit(taxa, " "))
        if (length(taxa) != nTax) {
          return(list(paste0("Mismatch: NTAX=", nTax, ", but ", length(taxa),
                             " TAXLABELS found in Nexus file.")))
        }
      }
    }

    notesEnd <- endBlocks[endBlocks > notesStart][1] - 1L
    notesLines <- lines[(notesStart + 1):notesEnd]
    collapsedLines <- paste0(notesLines, collapse = "\r\n")
    # Remove [comments]
    collapsedLines <- gsub("(;\\s*)\\[[^\\]]*\\]", "\\1", collapsedLines,
                           perl = TRUE)
    notes <- strsplit(collapsedLines,
                      # (?i) makes perl regexp case insensitive
                      "(?i)\\r\\n\\s*TEXT\\s+", perl = TRUE)[[1]]

    noteTaxon <- as.integer(.RegExpMatches("\\bTAXON\\s*=\\s*(\\d+)", notes))
    noteChar <- as.integer(.RegExpMatches("\\bCHARACTER\\s*=\\s*(\\d+)", notes))
    noteText <- EndSentence(MorphoBankDecode(
      .RegExpMatches("\\bTEXT\\s*=\\s*['\"]([\\s\\S]+)['\"];\\s*$", notes)))

    seqAlongNotes <- if (any(!is.na(noteChar))) {
      seq_len(max(noteChar, na.rm = TRUE))
    } else {
      numeric(0)
    }

    # Return:
    setNames(lapply(seqAlongNotes, function(i) {
      byTaxon <- !is.na(noteChar) & noteChar == i & !is.na(noteTaxon)
      ret <- list(
        noteText[!is.na(noteChar) & noteChar == i & is.na(noteTaxon)],
        setNames(noteText[byTaxon], taxa[noteTaxon[byTaxon]]))

      # Return:
      ret
    }), seqAlongNotes)
  }
}

#' Add full stop to end of a sentence
#'
#' @param string Input string
#'
#' @return `EndSentence()` returns `string`, punctuated with a final full stop
#' (period).`
#'
#' @examples
#' EndSentence("Hello World") # "Hello World."
#' @author Martin R. Smith
#' @family string parsing functions
#' @export
EndSentence <- function(string) {
  if (length(string)) {
    ret <- gsub("\\s*\\.?\\s*\\.$", ".", paste0(string, "."), perl = TRUE)
    ret <- gsub("(\\.[\"'])\\.$", "\\1", ret, perl = TRUE)
    ret <- gsub("([!\\?])\\.$", "\\1", ret, perl = TRUE)
    ret[ret == "."] <- ""
    ret
  } else {
    string
  }
}

#' Remove quotation marks from a string
#'
#' @param string Input string
#'
#' @return `Unquote()` returns `string`, with any matched punctuation marks
#' and trailing whitespace removed.
#'
#' @examples
#' Unquote("'Hello World'")
#' @author Martin R. Smith
#' @family string parsing functions
#' @export
Unquote <- function(string) {
  noSingle <- vapply(string, gsub, character(1),
                     pattern = "^\\s*'\\s*(.*?)\\s*'\\s*$",
                     replacement = "\\1", USE.NAMES = FALSE)
  vapply(noSingle, gsub, character(1),
         pattern = "^\\s*\"\\s*(.*?)\\s*\"\\s*$", replacement = "\\1",
         USE.NAMES = FALSE)
}

# Unescape quotes in Nexus format
.UnescapeQuotes <- function(string) {
  gsub("(?=.)''(?=.)", "'", string, perl = TRUE)
}

#' Decode MorphoBank text
#'
#' Converts strings from MorphoBank notes into a Latex-compatible format.
#'
#' @param string String to process
#'
#' @return `MorphoBankDecode()` returns a string with new lines and punctuation
#' reformatted.
#' @family string parsing functions
#' @author Martin R. Smith
#' @export
MorphoBankDecode <- function(string) {
  string <- gsub("^n", "  \n", string, fixed = TRUE)
  string <- gsub("''", "'", string, fixed = TRUE)
  string <- gsub(" - ", " -- ", string, fixed = TRUE)
  string <- gsub("(\\d)\\-(\\d)", "\\1--\\2", string, perl = TRUE)
  string <- gsub("(\\d) ?um\\b", "\\1 \u{03BC}m", string, perl = TRUE)
  string <- gsub(" [recoded as neomorphic]",
                 " Inapplicable tokens in this neomorphic character have been replaced with the absent token, following @Brazeau2018", string, fixed = TRUE)

  # Return:
  string
}

#' Convert between matrices and `phyDat` objects
#'
#' `MatrixToPhyDat()` converts a matrix of tokens to a `phyDat` object;
#' `PhyDatToMatrix()` converts a `phyDat` object to a matrix of tokens.
#'
#' @param tokens Matrix of tokens, possibly created with [`ReadCharacters()`]
#' or [`ReadTntCharacters()`].
#' Row names should correspond to leaf labels; column names may optionally
#' correspond to character labels.
#' @param tipLabels Optionally, an object providing labels for leaves, via
#' [TipLabels()]; will override `rownames(tokens)`.
#'
#' @return `MatrixToPhyDat()` returns an object of class `phyDat`.
#'
#' @family phylogenetic matrix conversion functions
#' @examples 
#' tokens <- matrix(c(0, 0, "0", 0, 0,
#'                    0, 0, "1", 0, 1,
#'                    0, 0, "1", 0, 1,
#'                    0, 0, "2", 0, 1,
#'                    1, 1, "-", 1, 0,
#'                    1, 1, "2", 1, "{01}"),
#'                    nrow = 6, ncol = 5, byrow = TRUE,
#'                    dimnames = list(
#'                      paste0("Taxon_", LETTERS[1:6]),
#'                      paste0("Char_", 1:5)))
#'                    
#' MatrixToPhyDat(tokens)
#' @template MRS
#' @export
MatrixToPhyDat <- function(tokens, tipLabels = rownames(tokens)) {
  if (inherits(tokens, "phyDat")) {
    warning("MatrixToPhyDat() expects a matrix, not a phyDat object")
    return(tokens)
  }
  allTokens <- unique(as.character(tokens))
  if (any(nchar(allTokens) == 0)) {
    problems <- apply(tokens, 1, function(x) which(nchar(x) == 0))
    problemTaxa <- lengths(problems) > 0
    problemTaxa <- names(problemTaxa[problemTaxa])
    warning("Blank tokens ('') found in taxa: ",
            paste0(problemTaxa, collapse = ", "))
  }
  tokenNumbers <- seq_along(allTokens)
  names(tokenNumbers) <- allTokens
  matches <- gregexpr("[\\d\\-\\w]", allTokens, perl = TRUE)
  whichTokens <- regmatches(allTokens, matches)
  levels <- sort(unique(unlist(whichTokens)))
  whichTokens[allTokens == "?"] <- list(levels)
  contrast <- vapply(whichTokens, function(x) levels %fin% x,
                     logical(length(levels)))
  contrast <- 1 * if (is.null(dim(contrast))) {
    as.matrix(contrast)
  } else {
    t(contrast)
  }
  rownames(tokens) <- TipLabels(tipLabels)
  dimnames(contrast) <- list(allTokens, levels)
  dat <- .PhyDatWithContrast(tokens, contrast = contrast)

  # Return:
  dat
}


#' @export
`[.phyDat` <- .SubsetPhyDat <- function(x, i, j, ..., drop = FALSE) {
  mat <- PhyDatToMatrix(x)
  MatrixToPhyDat(mat[i, j, ..., drop = FALSE])
}


.PhyDatWithContrast <- function(dat, contrast, tipLabels = rownames(dat)) {
  if (is.null(dim(dat))) {
    dat <- t(t(dat))
  }
  if (is.null(tipLabels)) {
    stop("Data rows must be named with tip labels")
  }
  
  levelSet <- dimnames(contrast)
  allLevels <- levelSet[[1]]
  levels <- levelSet[[2]]
  rownames(contrast) <- NULL
  
  # See https://stackoverflow.com/questions/70557817
  groups <- do.call(grouping, as.data.frame(t(dat)))
  ends <- attr(groups, "ends")
  i <- rep(seq_along(ends), c(ends[1], diff(ends)))[order(groups)]
  firstOccurrence <- match(i, i)
  tab <- table(firstOccurrence)
  weight <- as.integer(tab)
  tab[] <- seq_along(tab)
  index <- as.integer(tab[as.character(firstOccurrence)])
  
  duplicate <- duplicated(firstOccurrence)
  phyMat <- matrix(match(dat[, !duplicate], allLevels),
                   dim(dat)[1], sum(!duplicate))
  
  # Return:
  structure(
    lapply(seq_along(tipLabels), function(i) phyMat[i, ]),
    #as.list(asplit(phyMat, 1)),
    names  = tipLabels,
    weight = as.integer(weight),
    nr = length(weight),
    nc = length(levels),
    index = unname(index),
    levels = levels,
    allLevels = allLevels,
    type = "USER",
    contrast = contrast,
    class = "phyDat")
}


#' @rdname MatrixToPhyDat
#' @param dataset A dataset of class `phyDat`.
#' @param ambigNA,inappNA Logical specifying whether to denote ambiguous /
#' inapplicable characters as `NA` values.
#' @param parentheses Character vector specifying style of parentheses
#' with which to enclose ambiguous characters. `c("[", "]")` or `"[]"` will
#' render `[01]`.
#' `NULL` will use the token specified in the `phyDat` object; but beware that
#' this will be treated as a distinct (non-ambiguous) token if re-encoding with
#' `PhyDatToMatrix()`.
#' @param sep Character with which to separate ambiguous tokens, e.g. `','`
#' will render `[0,1]`.
#' @return `PhyDatToMatrix()` returns a matrix corresponding to the
#' uncompressed character states within a `phyDat` object.
#' @examples 
#' data("Lobo", package = "TreeTools")
#' head(PhyDatToMatrix(Lobo.phy)[, 91:93])
#' @export
PhyDatToMatrix <- function(dataset, ambigNA = FALSE, inappNA = ambigNA,
                           parentheses = c("{", "}"), sep = "") {
  if (!is.null(parentheses) && length(parentheses) == 0) {
    parentheses <- c("", "")
  } else if (length(parentheses) == 1) {
    parentheses <- c(
      substr(parentheses, 1, 1),
      substr(parentheses, nchar(parentheses), nchar(parentheses))
    )
  }
  
  at <- attributes(dataset)
  if (at[["nr"]] == 0L) {
    return(matrix(character(0), nrow = length(at[["names"]]), ncol = 0,
                  dimnames = list(at[["names"]], NULL)))
  }
  allLevels <- as.character(at[["allLevels"]])
  if (inappNA) {
    allLevels[allLevels == "-"] <- NA_character_
  }
  if (ambigNA) {
    allLevels[rowSums(at[["contrast"]]) != 1L] <- NA_character_
  } else if (!is.null(parentheses)) {
    cont <- at[["contrast"]]
    nTokens <- rowSums(cont)
    levels <- colnames(cont)
    partAmbig <- nTokens != 1L & nTokens < dim(cont)[2]
    allLevels[partAmbig] <- paste0(
      parentheses[1],
      apply(cont[partAmbig, , drop = FALSE] > 0, 1, function(x) {
        paste0(levels[x], collapse = sep)
      }),
      parentheses[2])
  }
  matrix(allLevels[unlist(dataset, recursive = FALSE, use.names = FALSE)],
         ncol = at[["nr"]], byrow = TRUE, dimnames = list(at[["names"]], NULL)
         )[, at[["index"]], drop = FALSE]
}

#' @rdname ReadCharacters
#' @export
ReadAsPhyDat <- function(...) {
  MatrixToPhyDat(ReadCharacters(...))
}


#' @rdname ReadCharacters
#' @param \dots Parameters to pass to `Read[Tnt]Characters()`.
#' @export
ReadTntAsPhyDat <- function(...) {
  MatrixToPhyDat(ReadTntCharacters(...))
}

#' @rdname ReadCharacters
#' @export
ReadTNTAsPhyDat <- ReadTntAsPhyDat


#' @describeIn ReadCharacters A convenient wrapper for \pkg{phangorn}'s
#' `phyDat()`, which converts a **list** of morphological characters into a
#' `phyDat` object.
#' If your morphological characters are in the form of a **matrix**, perhaps
#' because they have been read using [`read.table()`], try [`MatrixToPhyDat()`]
#' instead.
#'
#' @param dataset list of taxa and characters, in the format produced by 
#' \code{\link[ape]{read.nexus.data}()}:
#'   a list of sequences each made of a single character vector,
#'   and named with the taxon name.
#'
#' @export
PhyDat <- function(dataset) {
  nChar <- length(dataset[[1]])
  if (nChar == 1) {
    mat <- matrix(unlist(dataset), dimnames = list(names(dataset), NULL))
  } else {
    mat <- t(vapply(dataset, I, dataset[[1]]))
  }
  MatrixToPhyDat(mat)
}

#' Convert Nexus token matrix to integer
#'
#' `NexusTokensToInteger()` converts the character matrix returned by
#' [`ReadCharacters()`] to an integer matrix, mapping polymorphic,
#' ambiguous (`?`), and inapplicable (`-`) tokens to `NA_integer_` or to the
#' first/last state listed in the polymorphism, depending on `polymorphism`.
#'
#' Only digit states `0`..`9` are recognised; non-digit symbols (and any
#' token whose interior contains no digits) become `NA_integer_`.
#' Polymorphism extraction (`polymorphism = "first"`/`"last"`) likewise
#' considers digits only.
#'
#' If `tokens` is a `phyDat` object it is first converted via
#' [`PhyDatToMatrix()`] with `ambigNA = TRUE, inappNA = TRUE`, so that
#' fully-ambiguous and inapplicable rows become `NA_integer_` and only
#' true partial polymorphisms are subject to the `polymorphism` rule.
#'
#' @param tokens Character matrix as returned by [`ReadCharacters()`], a
#' character vector as returned by [`NexusTokens()`], or a `phyDat` object.
#' @param polymorphism Character string specifying how to handle polymorphic
#' tokens such as `"(01)"` or `"{12}"`:
#' \describe{
#'   \item{`"?"` (default)}{Treat as the NEXUS missing-data token: map to
#'     `NA_integer_`.}
#'   \item{`"first"`}{Use the first state digit inside the brackets.}
#'   \item{`"last"`}{Use the last state digit inside the brackets.}
#' }
#' Tokens `"?"` and `"-"` always map to `NA_integer_` regardless of
#' `polymorphism`.
#'
#' @return An integer matrix (or vector) with the same dimensions and
#' `dimnames` as `tokens`.
#'
#' @examples
#' tokens <- matrix(c("0", "(12)", "1", "?", "-"),
#'                  nrow = 1,
#'                  dimnames = list("Taxon_A", paste0("C", 1:5)))
#' NexusTokensToInteger(tokens)
#' NexusTokensToInteger(tokens, polymorphism = "first")
#'
#' @family phylogenetic matrix conversion functions
#' @template MRS
#' @export
NexusTokensToInteger <- function(tokens,
                                 polymorphism = c("?", "first", "last")) {
  polymorphism <- match.arg(polymorphism)
  if (inherits(tokens, "phyDat")) {
    tokens <- PhyDatToMatrix(tokens, ambigNA = TRUE, inappNA = TRUE)
  }
  at <- attributes(tokens)

  x <- as.character(tokens)
  result <- suppressWarnings(as.integer(x))

  ambig <- is.na(result) & !is.na(x) & x != "?" & x != "-"
  if (polymorphism != "?" && any(ambig)) {
    pattern <- if (polymorphism == "first") "\\d" else "\\d(?=[^\\d]*$)"
    m <- regexpr(pattern, x[ambig], perl = TRUE)
    matched <- regmatches(x[ambig], m)
    digits <- rep(NA_character_, sum(ambig))
    digits[m != -1L] <- matched
    result[ambig] <- suppressWarnings(as.integer(digits))
  }

  attributes(result) <- at
  result
}


#' Rightmost character of string
#'
#' `RightmostCharacter()` is a convenience function that returns the final
#' character of a string.
#'
#' @param string Character string.
#' @param len (Optional) Integer specifying number of characters in `string`.
#'
#' @return `RightmostCharacter()` returns the rightmost character of a string.
#' @examples
#' RightmostCharacter("Hello, World!")
#'
#' @template MRS
#' @export
#' @family string parsing functions
RightmostCharacter <- function(string, len = nchar(string)) {
  substr(string, len, len)
}

#' Write Newick Tree
#'
#' `NewickTree()` encodes a tree as a Newick-format string.
#' This differs from [`write.tree()`][ape::write.tree] in the encoding of
#' spaces as spaces, rather than underscores.
#'
#' @template treeParam
#'
#' @return `NewickTree()` returns a character string denoting `tree` in Newick
#' format.
#'
#' @examples
#' NewickTree(BalancedTree(LETTERS[4:9]))
#'
#' @seealso Use tip numbers, rather than leaf labels: [`as.Newick`]
#' @importFrom ape write.tree
#' @export
NewickTree <- function(tree) gsub("_", " ", write.tree(tree), fixed = TRUE)

test_that("Reweight()", {
  mat <- rbind(
    a = c(c1 = 0, c2 = 0, c3 = 2, c4 = 0),
    b = c(0, 0, 2, 0),
    c = c(1, 0, 3, 0),
    d = c(1, 0, 3, 0))
  dat <- MatrixToPhyDat(mat)
  
  expect_error(Reweight(mat, letters[1:4]),
               "must be a numeric vector")
  expect_error(Reweight(mat, 1:2),
               "Length of `weights` \\(2\\) must match .* characters \\(4\\)")
  expect_error(Reweight(mat, 1:8),
               "Length of `weights` \\(8\\) must match .* characters \\(4\\)")
  expect_error(Reweight(mat, c("-1" = 2, "999" = 911)),
               "Indices in `names\\(weights\\)` .*ween 1 and 4. Found -1, 999")
  
  newWeights <- c(0, 1, 3, 1)
  expect_equal(
    Reweight(mat, newWeights),
    `colnames<-`(mat[, c(2, 3, 3, 3, 4)],
                 c("c2_1", "c3_1", "c3_2", "c3_3", "c4_1"))
  )
  
  rwDat <- Reweight(dat, c("3" = 3, "2" = 1, "1" = 0))
  expect_equal(PhyDatToMatrix(rwDat), PhyDatToMatrix(dat))
  rwAtt <- attributes(rwDat)
  sameAtt <- c("names", "nc", "levels", "allLevels", "contrast", "type",
               "nr", "index", "class")
  expect_equal(rwAtt[sameAtt], attributes(dat)[sameAtt])
  expect_equal(sum(rwAtt[["weight"]]), sum(newWeights))
  expect_equal(rwAtt[["weight"]], c(0, 2, 3))
})

test_that("Reweight() okay with tree search", {
  skip_if_not_installed("TreeSearch")
  library("TreeSearch")
  tree <- NJTree(Lobo.phy)
  expect_equal(TreeLength(tree, Reweight(Lobo.phy,
                                       rep_len(2, length(Lobo.data[[1]])))),
               TreeLength(tree, Lobo.phy) * 2)
  
  expect_equal(TreeLength(tree, Reweight(Lobo.phy[, 9:12], c(0, 100, 1, 0))),
               TreeLength(tree, Lobo.phy[, 10]) * 100 +
                 TreeLength(tree, Lobo.phy[, 11])) # = 105
})

TestFile <- function(filename = "") {
  system.file("extdata", "tests", filename, package = "TreeTools")
}

test_that("File time is read correctly", {
  fileName <- TestFile("ape-tree.nex")
  expect_equal("2018-07-18 13:47:46", ApeTime(fileName, "string"))
  expect_error(ApeTime(rep(fileName, 2)))
})

test_that("Missing files fail gracefully", {
  expect_error(ReadAsPhyDat("non-existent.file"),
               "'non-existent.file' not found")
})

test_that("Nexus file can be parsed", {
  # Errors as lists:
  expect_equal("MATRIX block not found in Nexus file.",
               ReadCharacters(TestFile("ape-tree.nex"))[[1]])

  filename <- TestFile("parse-nexus.nexus")
  read <- ReadCharacters(filename)
  expect_equal(192, ncol(read))
  expect_equal(80, nrow(read))
  expect_equal("Wiwaxia", rownames(read)[4])
  expect_equal("(01)", as.character(read[1, 27]))

  filename <- TestFile("continuous.nex")
  read <- ReadCharacters(filename)
  expect_equal(1L, unique(as.integer(read[1, ])))
  expect_equal(setNames("?", "B_alienus"), read["B_alienus", 4])
  expect_equal(3L, unique(as.integer(read[3, ])))
})

test_that("ReadCharacters handles Cingulata-style polymorphism with internal whitespace", {
  nexusContent <- "#NEXUS
BEGIN CHARACTERS;
  DIMENSIONS NCHAR=5;
  FORMAT DATATYPE=STANDARD MISSING=? GAP=-;
  MATRIX
    Taxon_A  0(1 2)1?-
    Taxon_B  1{0 1}0?-
    Taxon_C  01010
  ;"

  tf <- tempfile(fileext = ".nex")
  on.exit(unlink(tf))
  writeLines(nexusContent, tf)
  read <- ReadCharacters(tf)

  expect_equal(dim(read), c(3L, 5L))
  expect_equal(unname(read["Taxon_A", 2]), "(12)")
  expect_equal(unname(read["Taxon_B", 2]), "{01}")
  expect_equal(unname(read["Taxon_A", 4]), "?")
  expect_equal(unname(read["Taxon_A", 5]), "-")
  expect_equal(unname(read["Taxon_C", 1]), "0")

  # Continuation lines: data for a taxon split across multiple lines
  nexusMulti <- "#NEXUS
BEGIN CHARACTERS;
  DIMENSIONS NCHAR=6;
  FORMAT DATATYPE=STANDARD MISSING=? GAP=-;
  MATRIX
    Taxon_A  0(1 2)1
              ?-1
    Taxon_B  1{0 1}0
              ?-0
  ;"

  tf2 <- tempfile(fileext = ".nex")
  on.exit(unlink(tf2), add = TRUE)
  writeLines(nexusMulti, tf2)
  readMulti <- ReadCharacters(tf2)

  expect_equal(dim(readMulti), c(2L, 6L))
  expect_equal(unname(readMulti["Taxon_A", 2]), "(12)")
  expect_equal(unname(readMulti["Taxon_B", 2]), "{01}")
  expect_equal(unname(readMulti["Taxon_A", 6]), "1")
  expect_equal(unname(readMulti["Taxon_B", 6]), "0")

  # Round-trip the ReadCharacters() matrix through NexusTokensToInteger().
  intMat <- NexusTokensToInteger(read)
  expect_equal(dim(intMat), c(3L, 5L))
  expect_equal(unname(intMat["Taxon_A", 1]), 0L)
  expect_equal(unname(intMat["Taxon_A", 2]), NA_integer_) # polymorphism -> ?
  expect_equal(unname(intMat["Taxon_B", 2]), NA_integer_) # uncertainty -> ?
  expect_equal(unname(intMat["Taxon_A", 4]), NA_integer_) # ?
  expect_equal(unname(intMat["Taxon_A", 5]), NA_integer_) # -
  expect_equal(unname(intMat["Taxon_C", 5]), 0L)
  expect_equal(unname(NexusTokensToInteger(read, "first")["Taxon_A", 2]), 1L)
  expect_equal(unname(NexusTokensToInteger(read, "last")["Taxon_A", 2]), 2L)
})

test_that("NexusTokensToInteger() converts token matrix to integer", {
  tokens <- matrix(c("0", "(12)", "1", "?", "-"),
                   nrow = 1L,
                   dimnames = list("Tax", paste0("C", 1:5)))

  # Default: polymorphisms and ambiguities become NA
  result <- NexusTokensToInteger(tokens)
  expect_equal(dim(result), c(1L, 5L))
  expect_equal(dimnames(result), list("Tax", paste0("C", 1:5)))
  expect_equal(unname(result["Tax", "C1"]), 0L)
  expect_equal(unname(result["Tax", "C2"]), NA_integer_)
  expect_equal(unname(result["Tax", "C3"]), 1L)
  expect_equal(unname(result["Tax", "C4"]), NA_integer_)
  expect_equal(unname(result["Tax", "C5"]), NA_integer_)

  # polymorphism = "first": take first digit inside brackets
  result_f <- NexusTokensToInteger(tokens, polymorphism = "first")
  expect_equal(unname(result_f["Tax", "C2"]), 1L)
  expect_equal(unname(result_f["Tax", "C4"]), NA_integer_)

  # polymorphism = "last": take last digit inside brackets
  result_l <- NexusTokensToInteger(tokens, polymorphism = "last")
  expect_equal(unname(result_l["Tax", "C2"]), 2L)

  # Braces form {01}
  tokens2 <- matrix(c("{01}", "0"), nrow = 1L, dimnames = list("T1", c("C1", "C2")))
  expect_equal(unname(NexusTokensToInteger(tokens2)["T1", "C1"]), NA_integer_)
  expect_equal(unname(NexusTokensToInteger(tokens2, "first")["T1", "C1"]), 0L)
  expect_equal(unname(NexusTokensToInteger(tokens2, "last")["T1", "C1"]), 1L)

  # Named vector input (no dim attribute).
  vec <- c(a = "0", b = "(12)", c = "?", d = "1")
  vecOut <- NexusTokensToInteger(vec)
  expect_null(dim(vecOut))
  expect_equal(names(vecOut), c("a", "b", "c", "d"))
  expect_equal(unname(vecOut), c(0L, NA_integer_, NA_integer_, 1L))
  expect_equal(unname(NexusTokensToInteger(vec, "first")), c(0L, 1L, NA_integer_, 1L))
  expect_equal(unname(NexusTokensToInteger(vec, "last")), c(0L, 2L, NA_integer_, 1L))

  # No-digit polymorphism token: must not crash, must yield NA in every mode.
  noDigit <- matrix(c("(AB)", "0"), nrow = 1L,
                    dimnames = list("T", c("C1", "C2")))
  expect_silent(NexusTokensToInteger(noDigit))
  expect_equal(unname(NexusTokensToInteger(noDigit)["T", "C1"]), NA_integer_)
  expect_equal(unname(NexusTokensToInteger(noDigit, "first")["T", "C1"]),
               NA_integer_)
  expect_equal(unname(NexusTokensToInteger(noDigit, "last")["T", "C1"]),
               NA_integer_)
  expect_equal(unname(NexusTokensToInteger(noDigit, "first")["T", "C2"]), 0L)

  # state.labels and other matrix attributes round-trip through the result.
  tokens3 <- matrix(c("0", "(12)", "1"), nrow = 1L,
                    dimnames = list("Tax", c("C1", "C2", "C3")))
  attr(tokens3, "state.labels") <- list(c("absent", "present"),
                                        c("a", "b", "c"),
                                        c("absent", "present"))
  out3 <- NexusTokensToInteger(tokens3)
  expect_equal(attr(out3, "state.labels"), attr(tokens3, "state.labels"))

  # phyDat input routes through PhyDatToMatrix(ambigNA = TRUE, inappNA = TRUE).
  phy <- MatrixToPhyDat(matrix(c("0", "(12)", "1", "?", "-"),
                               nrow = 1L,
                               dimnames = list("Tax", paste0("C", 1:5))))
  viaPhy <- NexusTokensToInteger(phy)
  viaMat <- NexusTokensToInteger(PhyDatToMatrix(phy, ambigNA = TRUE,
                                                inappNA = TRUE))
  expect_equal(unname(viaPhy), unname(viaMat))
})

test_that("NexusTokens() fails gracefully", {
  expect_error(NexusTokens("0123012301230123", integer(0)))
  expect_equal("Character number must be between 1 and 16.",
               NexusTokens("0123012301230123", 0)[[1]])
})

test_that("Matrix converts to phyDat", {
  mat <- matrix(c(1,0,1,0,1,0,1,0,0,1,0,1,0,1,0,1,2,2,2,2,2,2,2,"?"),
                nrow = 3, byrow = TRUE)
  rownames(mat) <- LETTERS[1:3]
  expect_equal(mat, as.matrix(MatrixToPhyDat(mat)))
})

test_that(".PhyDatWithContrast() fails gracefully", {
  expect_error(.PhyDatWithContrast(matrix(0, 2, 2),
                                   matrix(c(1, 0, 0, 1), 2, 2, FALSE,
                                          list(0:1, 0:1))))
})

test_that("PhyDatToMatrix() with ambigs", {
  mat <- matrix(c(1,0,1,0,1,0,1,0,0,1,0,1,0,1,0,1,2,2,2,2,"{12}","(01)","-","?"),
                nrow = 3, byrow = TRUE)
  rownames(mat) <- LETTERS[1:3]
  expectation <- mat
  expectation[3, 5:8] <- NA_character_
  expect_equal(PhyDatToMatrix(MatrixToPhyDat(mat), TRUE, TRUE), expectation)
  expect_equal(PhyDatToMatrix(MatrixToPhyDat(mat), TRUE, FALSE)[3, 5:8],
               c(NA, NA, "-", NA))
  expect_equal(PhyDatToMatrix(MatrixToPhyDat(mat), FALSE, TRUE)[3, 5:8],
               c("{12}", "{01}", NA_character_, "?"))
  expect_equal(
    PhyDatToMatrix(MatrixToPhyDat(mat),
                   ambigNA = FALSE,
                   inappNA = TRUE,
                   parentheses = NULL
                   )[3, 5:8],
    c("{12}", "(01)", NA_character_, "?"))
  expect_equal(
    PhyDatToMatrix(MatrixToPhyDat(mat),
                   ambigNA = FALSE,
                   inappNA = TRUE,
                   parentheses = c("<>"),
                   sep = "/"
    )[3, 5:8],
    c("<1/2>", "<0/1>", NA_character_, "?"))
})

test_that("Modified phyDat objects can be converted", {
  # Obtained by subsetting, i.e. dataset <- biggerDataset[1:4]
  dataset <- structure(list(a = c(1L, 1L, 1L, 1L), c = c(2L, 1L, 1L, 2L),
                            d = c(2L, 1L, 2L, 1L), e = c(2L, 2L, 2L, 2L)),
                       weight = c(3L, 3L, 0L, 0L), nr = 4L, nc = 2L,
                       index = c(2L, 1L, 1L, 1L, 2L, 2L),
                       .Label = c("0", "1"), allLevels = c("0", "1"),
                       type = "USER",
                       contrast = structure(c(1, 0, 0, 1), .Dim = c(2L, 2L),
                                            .Dimnames = list(NULL, c("0", "1"))),
                       class = "phyDat")
  expect_equal(c(4, 6), dim(PhyDatToMatrix(dataset)))
})

test_that("MatrixToPhyDat() warns when characters blank", {
  # May occur when loading an excel file with empty cells
  mat <- matrix(c(1,0,1,0,1,0,1,0,0,"","","",0,1,0,1,2,2,2,2,2,2,2,"?"),
                nrow = 3, byrow = TRUE)
  rownames(mat) <- LETTERS[1:3]
  expect_warning(MatrixToPhyDat(mat))
})

test_that("MatrixToPhyDat() returns phyDat if passed", {
  expect_warning(expect_equal(Lobo.phy, MatrixToPhyDat(Lobo.phy)),
                 "phyDat")
})

test_that("MatrixToPhyDat() with tipLabels", {
  mat <- rbind(1, 1, 2, 2, 3, 3)
  expect_equal(MatrixToPhyDat(mat, tipLabels = letters[1:6]),
               MatrixToPhyDat(`rownames<-`(mat, letters[1:6])))
  expect_equal(MatrixToPhyDat(mat, tipLabels = StarTree(6)),
               MatrixToPhyDat(`rownames<-`(mat, TipLabels(6))))
})

test_that("StringToPhyDat()", {
  expect_equal(as.integer(StringToPhyDat("1111????", letters[1:8])),
               rep(1:2, each = 4))
  expect_equal(as.integer(StringToPhyDat("----????", letters[1:8])),
               rep(1:2, each = 4))
  expect_equal(as.integer(StringToPhyDat("----????")), rep(1:2, each = 4))
  expect_equal(names(StringToPhyDat("----????")), paste0("t", 1:8))
})

test_that("EndSentence() works correctly", {
  expect_equal(EndSentence("Hi"), "Hi.")
  expect_equal(EndSentence("Hi."), "Hi.")
  expect_equal(EndSentence("Hi?"), "Hi?")
  expect_equal(EndSentence("Hi!"), "Hi!")
  expect_equal(EndSentence(character(0)), character(0))
})

test_that("Unquote() unquotes", {
  expect_equal(Unquote("'Unquoted'"), "Unquoted")
  expect_equal(Unquote("\"Unquoted\""), "Unquoted")
  expect_equal(Unquote("'Unquoted '"), "Unquoted")
  expect_equal(Unquote("\" Unquoted \""), "Unquoted")
  expect_equal( Unquote("'Unquoted's '"), "Unquoted's")
  expect_equal(Unquote(.UnescapeQuotes("'Unquoted''s '")), "Unquoted's")
  expect_equal(Unquote("\"\""), "")
  expect_equal(Unquote("''"), "")
})

test_that("ReadNotes() reads notes", {
  notes <- ReadNotes(system.file("extdata/input/notes.nex",
                                 package = "TreeTools"))
  expect_equal(length(unlist(notes$`1`)), 0)
  expect_equal(notes[[2]][[2]], setNames("Taxon 2, char 2.", "taxon_b"))
  expect_equal(notes[[3]][[1]], "Three's a crowd.")
  expect_equal(notes[[3]][[2]], setNames("Tax1-Char3.", "taxon_a"))
})

test_that("ReadNotes() handles absence of character-taxon notes", {
  expect_equal(ReadNotes(system.file("extdata/tests/taxon-notes.nex",
                                     package = "TreeTools")),
               structure(list(), names = character(0)))
})

test_that("ReadNotes() handles misspecified encoding", {
  expect_message(
    expect_equal(ReadNotes(system.file("extdata/tests/encoding.nex",
                                     package = "TreeTools"))[[1]][[2]],
                 setNames("\u20acncoding.", "Two")),
    "trying cp1252 .*encoding")
})

test_that("ReadCharacters() reads CHARSTATELABELS", {
  labels <- ReadCharacters(system.file("extdata/input/dataset.nex",
                                       package = "TreeTools"))


  expect_equal(colnames(labels), c("Character one",
                                   "Character two",
                                   "lots-of-punctuation, and \"so on\"!",
                                   "Character n", "Character 5",  "Character 6",
                                   "final character"))

  ap <- c("absent", "present")
  expect_equal(attr(labels, "state.labels"),
               list(ap, ap,
                    c("here", "there", "everywhere"),
                    c("a long description", "present"),
                    c("simple", "more complex", "with (parentheses)",
                      "more complex, 6 still"),
                    c("this one has", "multiple lines"), ap));

  labels3 <- ReadCharacters(system.file("extdata/input/dataset.nex",
                                        package = "TreeTools"), 3)
  expect_equal(labels3, labels[, 3, drop = FALSE], ignore_attr = TRUE)
  expect_equal(attr(labels3, "state.labels"),
               attr(labels, "state.labels")[3])

})

test_that("MorphoBankDecode() decodes", {
  expect_equal("' -- x  \n 1--2", MorphoBankDecode("'' - x^n 1-2"))
})

test_that("NewickTree() works", {
  expect_equal("((Test taxon,Another test),(What's this?,Number 12.3));",
               NewickTree(BalancedTree(c("Test taxon", "Another_test",
                                             "What's this?", "Number 12.3"))))
})

test_that("as_newick() fails gracefully", {
  expect_equal(as_newick(matrix(0L, 0L, 2L)), ";")
  expect_equal(as_newick(matrix(1:0, 1L, 2L)), "(0);")
  expect_equal(as_newick(Postorder(BalancedTree(4)$edge) - 1L),
               as_newick(BalancedTree(4)$edge - 1L))
  expect_error(as_newick(matrix(0L, 8192 * 2L, 2L)),
               "Too many nodes")
  expect_error(as_newick(matrix(0L, 3, 3)),
               "`edge` must have two columns")
  expect_error(as_newick(matrix(c(4, 4, 4, 1:3), 3, 2)),
               "`min.edge.` must be zero")
  expect_error(as_newick(matrix(c(3, NA, 3, 0:2), 3, 2)),
               "`edge`.* NA")
  expect_error(as_newick(matrix(c(4, 4, 3, 0:2), 3, 2)),
               "`edge` is malformed")
})

nTax <- 6L
dataset <- matrix(seq_len(nTax), nrow = nTax, ncol = 6,
                  dimnames = list(LETTERS[1:nTax], 1:6))
dataset[4, 3:5] <- "-"
dataset[5:6, 4:6] <- "?"
dataset[6, 3] <- "?"

test_that("Errors are handled", {
  expect_error(ArtificialExtinction(dataset, "A", "E", replaceAmbiguous = "INVALID"))
  expect_error(ArtificialExtinction(dataset, "A", "E", replaceCoded = "INVALID"))
})

test_that("Replacements ok", {
  expectation <- matrix(rep("?", 6), 2, 3, dimnames = list(c("A", "B"), 4:6))
  expect_equal(expectation,
               ArtificialExtinction(dataset, subject = c("A", "B"),
                                    template = "E",
                                    replaceAmbiguous = "ambig")[1:2, 4:6])

  expectation[] <- 1
  colnames(expectation) <- NULL
  expect_equal(PhyDatToMatrix(ArtificialExtinction(
                 MatrixToPhyDat(dataset),
                 c("A", "B"), "F", "unif", sampleFrom = "A"
                ))[1:2, 4:6],
               expectation
               )

  expect_true(all(!"?" == ArtificialExtinction(
    dataset[-6, ], 1:2, 5, "freq"
    )[1:2, 4:6]))


  expect_true(all(ArtificialExtinction(dataset, subject = "E", template = "F",
                                       replaceAmbiguous = "binary")["E", 3:6] %in% 0:1))

  expect_equal(setNames(rep("?", 3), 4:6),
               ArtificialExtinction(dataset, subject = "E", template = "F",
                                    replaceAmbiguous = "binary",
                                    replaceAll = FALSE)["E", 4:6])

  nChar <- 100
  dataset <- rbind(subj = c(rep(2, nChar)),
                   templ = c(rep(3, nChar / 2), rep("?", nChar / 2)))
  suppressWarnings(expect_gt(chisq.test(as.integer(
    ArtEx(dataset, 1, 2, "binary", "binary")[1, ]))$p.value, 0.05))
})

#' Decompose additive (ordered) phylogenetic characters
#' 
#' `Decompose()` decomposes additive characters into a series of binary
#' characters, which is mathematically equivalent when analysed under
#' equal weights parsimony.  (This equivalence is not exact
#' under implied weights or under probabilistic tree inference methods.)
#' 
#' An ordered (additive) character can be rewritten as a mathematically
#' equivalent hierarchy of binary neomorphic characters
#' \insertCite{Farris1970}{TreeTools}.  
#' Two reasons to prefer the latter approach are:
#' 
#' - It makes explicit the evolutionary assumptions underlying an ordered 
#'   character, whether the underlying ordering is linear, reticulate or
#'   branched \insertCite{Mabee1989}{TreeTools}.
#' - It avoids having to identify characters requiring special treatment to
#'   phylogenetic software, which requires the maintenance of an up-to-date 
#'   log of which characters are treated as additive and which sequence their
#'   states occur in, a step that may be overlooked by re-users of the data.
#'   
#' Careful consideration is warranted when evaluating whether a group of
#' related characteristics ought to be treated as ordered
#' \insertCite{Wilkinson1992}{TreeTools}.
#' On the one hand, the 'principle of indifference' states that we should treat
#' all transformations as equally probable (/ surprising / informative);
#' ordered characters fail this test, as larger changes are treated as less
#' probable than smaller ones.
#' On the other hand, ordered characters allow more opportunities for homology
#' of different character states, and might thus be defended under the auspices
#' of Hennig's Auxiliary Principle \insertCite{Wilkinson1992}{TreeTools}.
#' 
#' For a case study of how ordering phylogenetic characters can affect 
#' phylogenetic outcomes in practice, see
#' \insertCite{Brady2024;textual}{TreeTools}.
#'
#' @references \insertAllCited{}
#' @template datasetParam
#' @param indices Integer or logical vector specifying indices of characters
#' that should be decomposed
#' 
#' @return `Decompose()` returns a `phyDat` object in which the specified
#' ordered characters have been decomposed into binary characters.  
#' The attribute `originalIndex` lists the index of the character in
#' `dataset` to which each element corresponds.
#' 
#' @examples
#' data("Lobo")
#' 
#' # Identify character 11 as additive
#' # Character 11 will be replaced with two characters
#' # The present codings 0, 1 and 2 will be replaced with 00, 10, and 11.
#' decomposed <- Decompose(Lobo.phy, 11)
#' 
#' NumberOfChars <- function(x) sum(attr(x, "weight"))
#' NumberOfChars(Lobo.phy)   # 115 characters in original
#' NumberOfChars(decomposed) # 116 characters in decomposed
#' @template MRS
#' @family phylogenetic matrix conversion functions
#' @export
Decompose <- function(dataset, indices) {
  if (is.matrix(dataset)) {
    dataset <- MatrixToPhyDat(dataset)
  }
  if (!inherits(dataset, "phyDat")) {
    stop("`dataset` must be an object of class `phyDat`")
  }
  levels <- attr(dataset, "levels")
  inappLevel <- levels == "-"
  appLevels <- levels[!inappLevel]
  mat <- as.matrix(dataset)
  nTaxa <- dim(mat)[[1]]
  nChar <- dim(mat)[[2]]
  if (is.numeric(indices)) {
    if (any(indices > nChar)) {
      warning("Only ", nChar, " characters; indices ",
              paste(indices[indices > nChar], collapse = ", "),
              " not found.")
    }
    indices <- seq_len(nChar) %in% indices
  } else {
    if (length(indices) != nChar) {
      stop("`length(indices)` must correspond to the number of characters in ",
           "`dataset`")
    }
  }
  
  replacements <- if (getRversion() >= "4.1") {
    # apply(simplify = FALSE) becomes available in R4.1.0
    apply(mat[, indices, drop = FALSE], 2, function(char) {
      whichLevels <- which(vapply(appLevels,
                                 function(x) any(grepl(x, char, fixed = TRUE)),
                                 logical(1)))
      if (!length(whichLevels)) {
        return(matrix(character(0), nrow = nTaxa, ncol = 0))
      }
      maxLevel <- max(whichLevels)
      vapply(seq_len(maxLevel)[-1], function(i) {
        zero <- .RegExpEscape(appLevels[1:(i - 1)])
        one <- .RegExpEscape(appLevels[i:maxLevel])
        gsub(
          "(0)0+|(1)1+", "\\1\\2",
          gsub(
            paste0(c("[", one, "]"), collapse = ""), "1",
            gsub(paste0(c("[", zero, "]"), collapse = ""), "0", char)
            )
          )
      }, char)
    }, simplify = FALSE)
  } else {                                                         # nocov start
    lapply(which(indices), function(i) {
      char <- mat[, i]
      whichLevels <- which(vapply(appLevels,
                                 function(x) any(grepl(x, char, fixed = TRUE)),
                                 logical(1)))
      if (!length(whichLevels)) {
        return(matrix(character(0), nrow = nTaxa, ncol = 0))
      }
      maxLevel <- max(whichLevels)
      vapply(seq_len(maxLevel)[-1], function(i) {
        zero <- .RegExpEscape(appLevels[1:(i - 1)])
        one <- .RegExpEscape(appLevels[i:maxLevel])
        gsub("(0)0+|(1)1+", "\\1\\2",
             gsub(paste0(c("[", one, "]"), collapse = ""), "1",
                  gsub(paste0(c("[", zero, "]"), collapse = ""), "0", char)
             )
        )
      }, char)
    })
  }                                                                # nocov end
  nNew <- `[<-`(rep(1, nChar), indices, vapply(replacements, ncol, 1))
  ret <- matrix(nrow = nTaxa, ncol = sum(nNew),
                dimnames = list(dimnames(mat)[[1]], NULL))
  newInd <- cumsum(nNew)
  ret[, newInd[!indices]] <- mat[, !indices]
  
  offset <- c(0, cumsum(nNew[indices] - 1))
  
  whichInds <- which(indices)
  replInd <- unlist(lapply(seq_along(whichInds), function(i) {
    index <- whichInds[[i]]
    index + offset[[i]] + seq_len(nNew[[index]]) - 1
    }
  ))
  if (!is.null(replInd)) {
    ret[, replInd] <- do.call(cbind, replacements)
  }
  
  # Consistent ambiguity symbols
  ret <- gsub("\\[(\\d)\\]", "\\1", 
              gsub("{", "[", fixed = TRUE,
                   gsub("}", "]", fixed = TRUE, ret)
              )
            )
  
  # Return:
  structure(
    MatrixToPhyDat(ret),
    originalIndex = rep(seq_len(nChar), nNew)
  )
}

.RegExpEscape <- function(x) {
  gsub("([\\[\\]\\{\\}\\?\\.\\+\\*\\|\\^\\$\\-\\(\\)\\\\/])", "\\\\\\1", x,
       perl = TRUE)
}

#' Remove metadata from trees
#' 
#' `TopologyOnly()` removes all information from trees except for their
#' topologies and leaf labels.
#' 
#' @inheritParams Preorder
#' @return Returns `tree`, with each tree in [`Preorder`], with edge lengths,
#' node labels and other attributes removed.
#' @template MRS
#' @export
TopologyOnly <- function(tree) UseMethod("TopologyOnly")

#' @export
TopologyOnly.phylo <- function(tree) {
  startOrder <- attr(tree, "order")
  ret <- structure(list(edge = tree[["edge"]],
                        Nnode = tree[["Nnode"]],
                        tip.label = tree[["tip.label"]]),
                   order = startOrder,
                   class = "phylo")
  if (length(startOrder) && startOrder == "preorder") {
    ret
  } else {
    Preorder(ret)
  }
}

#' @export
TopologyOnly.multiPhylo <- function(tree) {
  tree[] <- lapply(tree, TopologyOnly)
  tree
}

#' @export
TopologyOnly.list <- function(tree) {
  lapply(tree, TopologyOnly)
}

#' @export
TopologyOnly.default <- function(tree) {
  tree
}

#' @export
as.matrix.phylo <- function(x, ...) {
  sp <- as.Splits(x)
  as.logical(sp) * 1
}

#' @export
as.logical.phylo <- function(x, ...) {
  as.logical(as.Splits(x))
}

#' @export
as.matrix.phyDat <- function(x, ambigNA = FALSE, inappNA = ambigNA, ...) {
  PhyDatToMatrix(x, ambigNA, inappNA)
}

# This file is adapted from code by Rob Noble, doi:10.5281/zenodo.5873857
# https://github.com/robjohnnoble/RUtreebalance

# MIT License
# 
# Copyright (c) 2021 Rob Noble
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#   
#   The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# Find the parent of a node.
# 
# Example:
# edges1 <- data.frame(Parent = c(1, 1, 1, 3, 3), Identity = 2:6)
# .MoveUp(edges1, 3)
.MoveUp <- function(edges, identity) {
  if (!(identity %in% edges[["Identity"]]) &&
      !(identity %in% edges[["Parent"]])) {
    stop("Invalid identity.")
  }
  parent <- edges[edges[["Identity"]] == identity, "Parent"]
  if (length(parent) == 0) {
    # if identity is the root then don't move
    identity
  } else if (is.factor(parent)) {
    levels(parent)[parent]
  } else {
    parent 
  }
}

# Find the root node of a tree.
# 
# Example:
# edges1 <- data.frame(Parent = c(1,1,1,3,3), Identity = 2:6)
# .FindRoot(edges1)
.FindRoot <- function(edges) {
  start <- edges[["Parent"]][[1]] # reasonable guess
  if (is.factor(start)) {
    start <- levels(start)[start]
  }
  repeat {
    if (.MoveUp(edges, start) == start) break
    start <- .MoveUp(edges, start)
  }
  return(start)
}

# Get a list of all subtree sizes via depth-first search.
# Optionally provide root node i and adjacency list Adj if known.
# 
# Example:
# tree1 <- data.frame(Parent = c(1, 1, 1, 1, 2, 3, 4), 
#                     Identity = 1:7, 
#                     Population = c(1, rep(5, 6)))
# .GetSubtreeSizes(tree1)
.GetSubtreeSizes <- function(
    tree,
    i = which(tree[["Identity"]] == .FindRoot(tree[, 1:2])),
    Adj = .GetAdjacency(tree),
    Col = NULL, Cumul = NULL, is_leaf =  NULL) {
  n <- length(tree[["Identity"]])
  has_pops <- "Population" %in% colnames(tree)
  
  uniqueID <- unique(tree[["Identity"]])
  if (is.null(Col)) {
    Col <- setNames(rep("w", n), uniqueID)
  }
  if (is.null(Cumul)) {
    Cumul <- setNames(rep(NA, n), uniqueID)
  }
  if (is.null(is_leaf)) {
    is_leaf <- setNames(logical(n), uniqueID)
  }
  if (is.null(Adj[[i]])) {
    is_leaf[[i]] <- TRUE
  }
  for (j in Adj[[i]]) {
    if (Col[[j]] == "w"){
      L <- .GetSubtreeSizes(tree, j, Adj, Col, Cumul, is_leaf)
      Col<- L[["colour"]]
      Cumul <- L[["cumulative"]]
      is_leaf <- L[["is_leaf"]]
    }
  }
  Col[[i]] <- "b"
  Cumul[[i]] <- if (has_pops) {
    tree[["Population"]][[i]] + sum(Cumul[Adj[[i]]])
  } else {
    ifelse(is_leaf[[i]], 1, 0) + sum(Cumul[Adj[[i]]])
  }
  # Return:
  list("colour" = Col,"cumulative" = Cumul, "is_leaf" = is_leaf)
}

# Get adjacency list of a tree.
# 
# Example:
# tree1 <- data.frame(Parent = c(1, 1, 1, 1, 2, 3, 4), 
#                     Identity = 1:7, 
#                     Population = c(1, rep(5, 6)))
# .GetAdjacency(tree1)
.GetAdjacency <- function(tree) {
  n <- length(tree[["Identity"]])
  Adj <- vector(mode = "list", length = n)
  for (i in seq_len(n)) {
    if(tree$Parent[[i]] != tree$Identity[[i]]) {
      p <- match(tree$Parent[[i]], tree$Identity)
      Adj[[p]] <- c(Adj[[p]], i)
    }
  }
  
  # Return:
  Adj
}

# Add a row to the edges list to represent the root node (if not already present)
.AddRootRow <- function(tree) {
  start <- setdiff(tree[["Parent"]], tree[["Identity"]])
  if (length(start) > 1) {
    stop("Input dataframe is missing one or more rows")
  }
  if (length(start) > 0) { # add row for root node
    root_row <- if ("Population" %in% colnames(tree)) {
      data.frame(Parent = start, Identity = start, Population = 0)
    } else {
      data.frame(Parent = start, Identity = start)
    }
    tree <- rbind(root_row, tree)
  }
  
  # Return:
  tree
}

#' Robust universal tree balance index
#' 
#' Calculate tree balance index \ifelse{html}{\out{J<sup>1</sup>}}{\eqn{J^1}}
#' (when `nonRootDominance = FALSE`) or
#' \ifelse{html}{\out{J<sup>1c</sup>}}{\eqn{J^{1c}}}
#' (when `nonRootDominance = TRUE`) from \insertCite{Lemant2022}{TreeTools}.
#' 
#' If population sizes are not provided, then the function assigns size 0 to
#' internal nodes, and size 1 to leaves.
#' 
#' @param tree Either an object of class 'phylo', or a dataframe with column 
#' names Parent, Identity and (optionally) Population.
#' The latter is similar to `tree$edge`, where tree is an object of class
#' 'phylo'; the differences are in class (data.frame versus matrix) and column
#' names.
#' The dataframe may (but isn't required to) include a row for the root node.
#' If population sizes are omitted then internal nodes will be assigned
#' population size zero and leaves will be assigned population size one.
#' @param q Numeric between zero and one specifying sensitivity to type
#' frequencies.  If `q < 1`, the \ifelse{html}{\out{J<sup>q</sup>}}{\eqn{J^q}}
#' index - based on generalized entropy - will be returned; see 
#' \insertCite{Lemant2022;textual}{TreeTools}, page 1223.
#' @param nonRootDominance Logical specifying whether to use non-root dominance
#' factor.
#' 
#' @returns `J1Index()` returns a numeric specifying the
#' \ifelse{html}{\out{J<sup>1</sup>}}{\eqn{J^1}} index of `tree`.
#' \eqn{J^1(T) = 1} for a perfectly balanced tree; 
#' \eqn{J^1(T) = 0} for a pectinate (linear / caterpillar) tree.
#' 
#' @examples
#' # Using phylo object as input:
#' phylo_tree <- read.tree(text="((a:0.1)A:0.5,(b1:0.2,b2:0.1)B:0.2);")
#' J1Index(phylo_tree)
#' phylo_tree2 <- read.tree(text='((A, B), ((C, D), (E, F)));')
#' J1Index(phylo_tree2)
#' 
#' # Using edges lists as input:
#' tree1 <- data.frame(Parent = c(1, 1, 1, 1, 2, 3, 4),
#'                     Identity = 1:7,
#'                     Population = c(1, rep(5, 6)))
#' J1Index(tree1)
#' tree2 <- data.frame(Parent = c(1, 1, 1, 1, 2, 3, 4),
#'                     Identity = 1:7,
#'                     Population = c(rep(0, 4), rep(1, 3)))
#' J1Index(tree2)
#' tree3 <- data.frame(Parent = c(1, 1, 1, 1, 2, 3, 4),
#'                     Identity = 1:7,
#'                     Population = c(0, rep(1, 3), rep(0, 3)))
#' J1Index(tree3)
#' cat_tree <- data.frame(Parent = c(1, 1:14, 1:15, 15),
#'                        Identity = 1:31,
#'                        Population = c(rep(0, 15), rep(1, 16)))
#' J1Index(cat_tree)
#'
#' # If population sizes are omitted then internal nodes are assigned population
#' # size zero and leaves are assigned population size one:
#' sym_tree1 <- data.frame(Parent = c(1, rep(1:15, each = 2)),
#'                        Identity = 1:31,
#'                        Population = c(rep(0, 15), rep(1, 16)))
#' # Equivalently:                        
#' sym_tree2 <- data.frame(Parent = c(1, rep(1:15, each = 2)),
#'                        Identity = 1:31)
#' J1Index(sym_tree1)
#' J1Index(sym_tree2)
#' @author Rob Noble, adapted by Martin R. Smith
#' @references \insertAllCited{}
#' @family tree characterization functions
#' @importFrom stats na.omit
#' @export
J1Index <- function(tree, q = 1, nonRootDominance = FALSE) {
  
  if (!is.na(tree)[[1]]) {
    if(inherits(tree, "phylo")) {
      # convert from phylo object to data frame
      tree <- as.data.frame(tree[["edge"]])
      colnames(tree) <- c("Parent", "Identity")
    }
    tree <- na.omit(tree) # remove any rows containing NA
    if(is.factor(tree[["Parent"]])) {
      tree[["Parent"]] <- levels(tree[["Parent"]])[tree[["Parent"]]]
    }
    if (is.factor(tree[["Identity"]])) {
      tree[["Identity"]] <- levels(tree[["Identity"]])[tree[["Identity"]]]
    }
    tree <- .AddRootRow(tree)
  }
  
  n <- length(tree[["Identity"]])
  if (n <= 1) {
    return(0)
  }
  
  # Adjacency list
  Adj <- .GetAdjacency(tree)
  
  # Get the list of all subtree sizes
  subtree_sizes <- .GetSubtreeSizes(tree, Adj = Adj)
  
  # Subtree sizes, including the root
  Cumul <- subtree_sizes[["cumulative"]]
  
  # Vector of internal nodes
  eff_int_nodes <- which(!subtree_sizes[["is_leaf"]])
  
  # Vector of leaves
  leaves <- which(subtree_sizes[["is_leaf"]])
  
  # If population sizes are missing then assign size 0 to internal nodes,
  # and size 1 to leaves:
  if(!("Population" %in% colnames(tree))) {
    tree[["Population"]] <- double(n)
    tree[["Population"]][leaves] <- 1
  }
  if (sum(tree$Population) <= 0) {
    stop("At least one node must have Population > 0")
  }
  J <- 0
  # Subtree sizes, excluding the root
  Star <- Cumul - tree[["Population"]]
  for (i in seq_len(n)){ # Loop over all nodes
    if (Star[[i]] > 0){ # Node has at least one child with non-zero size
      K <- 0
      if(length(Adj[[i]]) > 1) { # Otherwise i has only one child and its 
                                 # balance score is 0
        eff_children <- 0 # Number of children with non-zero size
        for (j in Adj[[i]]) {
          if (Cumul[[j]] > 0){ # Otherwise child j has a 0-sized subtree and 
                               # does not count
            eff_children <- eff_children + 1
            
            # p is the ratio of the child subtree size including the root
            # (root = the child) 
            # to the parent subtree size excluding the root
            p <- Cumul[[j]] / Star[[i]]
            
            # K is the sum of the node balance scores
            K <- K + if (q == 1) {-p * log(p)} else {p ^ q}
          }
        }
        # Non-root dominance factor:
        if (isTRUE(nonRootDominance)) {
          h_factor <- Star[[i]] / Cumul[[i]]
        } else {
          h_factor <- 1
        }
        
        # Normalize the sum of balance scores, adjust for non-root dominance, 
        # and then add the result to the index
        if (eff_children > 1) { # Exclude nodes that have only one child with 
                                # size greater than zero
          J_increment <- h_factor * Star[[i]] * if (q == 1) {
             K / log(eff_children)
          } else {
            (1 - K) * eff_children ^ (q - 1) / (eff_children ^ (q - 1) - 1)
          }
          J <- J + J_increment
        }
      }
    }
  }
  # normalize the index by dividing by the sum of all subtree sizes:
  if (length(eff_int_nodes) > 0) {
    J <- J / sum(Star[eff_int_nodes])
  }
  
  # Return:
  as.numeric(J)
}

#' @rdname J1Index
#' @export
JQIndex <- J1Index

test_that("Balance examples work as expected", {
  tree1 <- data.frame(Parent = c(1,1,1,1,2,3,4),
                      Identity = 1:7,
                      Population = c(1, rep(5, 6)))
  expect_equal(J1Index(tree1), 2/3)
  tree2 <- data.frame(Parent = c(1,1,1,1,2,3,4),
                      Identity = 1:7,
                      Population = c(rep(0, 4), rep(1, 3)))
  expect_equal(J1Index(tree2), 1/2)
  tree3 <- data.frame(Parent = c(1,1,1,1,2,3,4),
                      Identity = 1:7,
                      Population = c(0, rep(1, 3), rep(0, 3)))
  expect_equal(J1Index(tree3), 1)
  cat_tree <- data.frame(Parent = c(1, 1:14, 1:15, 15),
                         Identity = 1:31,
                         Population = c(rep(0, 15), rep(1, 16)))
  expect_equal(J1Index(cat_tree), 4736/9990)
  sym_tree <- data.frame(Parent = c(1, rep(1:15, each = 2)),
                         Identity = 1:31,
                         Population = c(rep(0, 15), rep(1, 16)))
  expect_equal(J1Index(sym_tree), 1)
  
  sym_tree2 <- data.frame(Parent = c(1, rep(1:15, each = 2)),
                         Identity = 1:31)
  expect_equal(J1Index(sym_tree), J1Index(sym_tree2))
  
})

test_that("Different tree inputs work", {
  phylo_tree2 <- read.tree(text='((A, B), ((C, D), (E, F)));')
  expect_equal(J1Index(phylo_tree2), 0.96936094)
  expected <- 0.79248125
  phylo_tree <- read.tree(text="((a:0.1)A:0.5,(b1:0.2,b2:0.1)B:0.2);")
  expect_equal(J1Index(phylo_tree), expected)
  phylo_tree <- read.tree(text="((1)5,(2,3)6);")
  expect_equal(J1Index(phylo_tree), expected)
  
  # data.frame omitting population sizes:
  edges_tree <- data.frame(Parent = c(4, 5, 4, 6, 6),
                           Identity = c(5, 1, 6, 2, 3))
  expect_equal(J1Index(edges_tree), expected)
  
  # data.frame omitting population sizes and including a row for the root:
  edges_tree_with_root <- data.frame(Parent = c(4, 4, 5, 4, 6, 6),
                                     Identity = c(4, 5, 1, 6, 2, 3))
  expect_equal(J1Index(edges_tree_with_root), expected)
  
  # data.frame including population sizes:
  edges_tree_with_pops <- data.frame(
    Parent = c(4, 5, 4, 6, 6),
    Identity = c(5, 1,6 ,2, 3),
    Population = c(0, 1, 0, 1, 1)
  )
  expect_equal(J1Index(edges_tree_with_pops), expected)
})

test_that(".GetAdjacency() works", {
  tree1 <- data.frame(Parent = c(1,1,1,1,2,3,4), Identity = 1:7,
                      Population = c(1, rep(5, 6)))
  expect_equal(
    .GetAdjacency(tree1),
    list(2:4, 5, 6, 7, NULL, NULL, NULL)
  )
})

#' @importFrom utils globalVariables
utils::globalVariables(c("doubleFactorials",
                         "logDoubleFactorials",
                         "log2DoubleFactorials"),
                       "TreeTools")

#' Double factorial
#'
#' Calculate the double factorial of a number, or its logarithm.
#'
#' @param n Vector of integers.
#'
#' @return Returns the double factorial, _n_ * (_n_ - 2) * (_n_ - 4) *
#' (_n_ - 6) * ...
#'
#' @examples
#' DoubleFactorial (-4:0) # Return 1 if n < 2
#' DoubleFactorial (2) # 2
#' DoubleFactorial (5) # 1 * 3 * 5
#' exp(LnDoubleFactorial.int (8)) # log(2 * 4 * 6 * 8)
#' @template MRS
#' @family double factorials
#' @export
DoubleFactorial <- function(n) {
  if (any(n > 300)) stop("301!! is too large to represent. Use LnDoubleFactorial() instead.")

  n[n < 2] <- 1
  doubleFactorials[n]

  #
  #odds <- as.logical(x %% 2)
  #
  #oddX <- x[odds]
  #xPlusOneOverTwo <- (oddX + 1) / 2
  #evenX <- x[!odds]
  #xOverTwo <- evenX / 2
  #
  #ret <- integer(length(x))
  #ret[odds] <- gamma(oddX + 1L) / (gamma(xPlusOneOverTwo) * 2^(xPlusOneOverTwo - 1L))
  #ret[!odds] <- evenX * gamma(xOverTwo) * 2^(xOverTwo - 1L)
  #
  ## Return:
  #ret
}

#' @describeIn DoubleFactorial Returns the exact double factorial as a 64-bit
#' `integer64`, for `n` < 34.
#' @examples
#' DoubleFactorial64(31)
#' @export
DoubleFactorial64 <- function(n) {
  if (n < 2L) 1 else as.integer64(n * DoubleFactorial64(n - 2L))
}

# Memoizing this function makes it MUCH slower...
#' @describeIn DoubleFactorial Returns the logarithm of the double factorial.
#' @export
LnDoubleFactorial <- (function(n) {
  n[n < 2L] <- 1 # Much faster than pmax
  if (any(n > 49999L)) {

    odds <- as.logical(n %% 2)

    oddN <- n[odds]
    nPlusOneOverTwo <- (oddN + 1) / 2
    evenN <- n[!odds]
    nOverTwo <- evenN / 2

    ret <- integer(length(n))
    ret[odds] <- lgamma(oddN + 1L) - (lgamma(nPlusOneOverTwo) + (nPlusOneOverTwo - 1) * log(2))
    ret[!odds] <- log(evenN) + lgamma(nOverTwo) + (nOverTwo - 1) * log(2)

    # Return:
    ret

  } else {
    # Return from cache
    logDoubleFactorials[n]
  }
})

#' @describeIn DoubleFactorial Returns the logarithm of the double factorial.
#' @export
Log2DoubleFactorial <- (function(n) {
  n[n < 2L] <- 1 # Much faster than pmax
  if (any(n > 49999L)) {

    odds <- as.logical(n %% 2)

    oddN <- n[odds]
    nPlusOneOverTwo <- (oddN + 1) / 2
    evenN <- n[!odds]
    nOverTwo <- evenN / 2

    ret <- integer(length(n))
    ret[odds] <-  ((lgamma(oddN + 1L) - lgamma(nPlusOneOverTwo)) / log(2)) - nPlusOneOverTwo + 1
    ret[!odds] <- log2(evenN) + (lgamma(nOverTwo) / log(2)) + (nOverTwo - 1)

    # Return:
    ret

  } else {
    # Return from cache
    log2DoubleFactorials[n]
  }
})

#' @rdname DoubleFactorial
#' @export
LogDoubleFactorial <- LnDoubleFactorial

#' @describeIn DoubleFactorial Slightly faster, when x is known to be length one
#' and below 50001
#' @export
LnDoubleFactorial.int <- function(n) {
  if (n < 2L) {
    0
  } else {
    logDoubleFactorials[n]
  }
}

#' @rdname DoubleFactorial
#' @export
LogDoubleFactorial.int <- LnDoubleFactorial.int

#' Number of trees
#'
#' These functions return the number of rooted or unrooted binary trees
#' consistent with a given pattern of splits.
#'
#' Functions starting `N` return the number of rooted or unrooted trees.
#' Replace this initial `N` with `Ln` for the natural logarithm of this number;
#' or `Log2` for its base 2 logarithm.
#'
#' Calculations follow \insertCite{CavalliSforza1967;textual}{TreeTools} and
#' \insertCite{Carter1990;textual}{TreeTools}, Theorem 2.
#'
#' @param tips Integer specifying the number of leaves.
#' @param \dots Integer vector, or series of integers, listing the number of
#' leaves in each split.
#'
#' @template MRS
#'
#' @references \insertAllCited{}
#'
#' @examples
#' NRooted(10)
#' NUnrooted(10)
#' LnRooted(10)
#' LnUnrooted(10)
#' Log2Unrooted(10)
#' # Number of trees consistent with a character whose states are
#' # 00000 11111 222
#' NUnrootedMult(c(5,5,3))
#'
#' @family tree information functions
#' @export
NRooted     <- function(tips) DoubleFactorial(tips + tips - 3L) # addition faster than 2*

#' @describeIn NRooted Number of unrooted trees
#' @export
NUnrooted   <- function(tips) DoubleFactorial(tips + tips - 5L)

#' @describeIn NRooted Exact number of rooted trees as 64-bit integer
#' (13 < `nTip` < 19)
#' @export
NRooted64 <- function(tips) DoubleFactorial64(tips + tips - 3L)

#' @describeIn NRooted Exact number of unrooted trees as 64-bit integer
#' (14 < `nTip` < 20)
#' @examples
#' NUnrooted64(18)
#' @export
NUnrooted64 <- function(tips) DoubleFactorial64(tips + tips - 5L)

#' @describeIn NRooted  Log Number of unrooted trees
#' @export
LnUnrooted  <- function(tips) LnDoubleFactorial(tips + tips - 5L)

#' @describeIn NRooted  Log Number of unrooted trees (as integer)
#' @export
LnUnrooted.int <- function(tips) {
  ifelse(tips < 3L, 0, logDoubleFactorials[tips + tips - 5L])
}

#' @rdname NRooted
#' @export
Log2Unrooted  <- function(tips) Log2DoubleFactorial(tips + tips - 5L)

#' @rdname NRooted
#' @export
Log2Unrooted.int <- function(tips) {
  ifelse(tips < 3L, 0, log2DoubleFactorials[tips + tips - 5L])
}

#' @describeIn NRooted  Log Number of rooted trees
#' @export
LnRooted    <- function(tips) LnDoubleFactorial(tips + tips - 3L)
#' @describeIn NRooted  Log Number of rooted trees (as integer)
#' @export
LnRooted.int <- function(tips) {
  ifelse(tips < 2L, 0, logDoubleFactorials[tips + tips - 3L])
}
#' @rdname NRooted
#' @export
Log2Rooted    <- function(tips) Log2DoubleFactorial(tips + tips - 3L)
#' @rdname NRooted
#' @export
Log2Rooted.int <- function(tips) {
  ifelse(tips < 2L, 0, log2DoubleFactorials[tips + tips - 3L])
}

#' Number of trees one SPR step away
#'
#' `N1Spr()` calculates the number of trees one subtree prune-and-regraft
#' operation away from a binary input tree using the formula given by 
#' \insertCite{Allen2001;textual}{TreeTools};
#' `IC1Spr()` calculates the information content of trees at this
#' distance: i.e. the entropy corresponding to the proportion of all possible
#' _n_-tip trees whose SPR distance is at most one from a specified tree.
#'
#' @param n Integer vector specifying the number of tips in a tree.
#'
#' @return `N1Spr()` returns an integer vector denoting the number of trees one
#' SPR rearrangement away from the input tree..
#'
#' @examples
#' N1Spr(4:6)
#' IC1Spr(5)
#'
#' @references \insertAllCited{}
#'
#' @export
N1Spr <- function(n) ifelse(n > 3L, (n + n - 6L) * (n + n - 7L), 0L)

#' @rdname N1Spr
#' @return `IC1Spr()` returns an numeric vector giving the phylogenetic
#' information content of trees 0 or 1 SPR rearrangement from an _n_-leaf tree,
#' in bits.
#' @export
IC1Spr <- function(n) Log2Unrooted(n) - log2(1L + N1Spr(n))

#' @rdname NRooted
#' @examples
#' LnUnrootedSplits(c(2,4))
#' LnUnrootedSplits(3, 3)
#' @export
LnUnrootedSplits <- function(...) {
  splits <- c(...)

  if ((nSplits <- length(splits)) < 2L) {
    LnUnrooted(splits)
  } else if (nSplits == 2L) {
    LnRooted(splits[1]) + LnRooted(splits[2])
  } else {
    LnUnrootedMult(splits)
  }
}

#' @rdname NRooted
#' @examples
#' Log2UnrootedSplits(c(2,4))
#' Log2UnrootedSplits(3, 3)
#' @export
Log2UnrootedSplits <- function(...) {
  splits <- c(...)

  if ((nSplits <- length(splits)) < 2L) {
    Log2Unrooted(splits)
  } else if (nSplits == 2L) {
    Log2Rooted(splits[1]) + Log2Rooted(splits[2])
  } else {
    Log2UnrootedMult(splits)
  }
}

#' @describeIn NRooted Number of unrooted trees consistent with a bipartition
#' split.
#' @examples
#' NUnrootedSplits(c(2,4))
#' NUnrootedSplits(3, 3)
#' @family split information function
#' @export
NUnrootedSplits  <- function(...) {
  splits <- c(...)
  if ((nSplits <- length(splits)) < 2L) {
    NUnrooted(splits)
  } else if (nSplits == 2L) {
    NRooted(splits[1]) * NRooted(splits[2])
  } else {
    NUnrootedMult(splits)
  }
}

#' @rdname NRooted
#' @export
LnUnrootedMult <- function(...) {  # Carter et al. 1990, Theorem 2
  splits <- c(...)
  splits <- splits[splits > 0]
  totalTips <- sum(splits)

  # Return:
  LnDoubleFactorial(totalTips +  totalTips - 5L) -
    LnDoubleFactorial(2L * (totalTips - length(splits)) - 1L) +
    sum(LnDoubleFactorial(splits + splits - 3L))

}

#' @rdname NRooted
#' @export
Log2UnrootedMult <- function(...) {  # Carter et al. 1990, Theorem 2
  splits <- c(...)
  splits <- splits[splits > 0]
  totalTips <- sum(splits)

  # Return:
  sum(Log2DoubleFactorial(totalTips +  totalTips - 5L),
      -Log2DoubleFactorial(2L * (totalTips - length(splits)) - 1L),
      Log2DoubleFactorial(splits + splits - 3L))
}

#' @describeIn NRooted Number of unrooted trees consistent with a multi-partition
#' split.
#' @export
NUnrootedMult  <- function(...) { # Carter et al. 1990, Theorem 2
  splits <- c(...)
  splits <- splits[splits > 0]
  nSplits <- length(splits)
  if (nSplits < 1L) {
    # Return:
    0L
  } else if (nSplits == 1L) {
    # Return:
    NUnrooted(splits)
  } else if (nSplits == 2L) {
    prod(DoubleFactorial(splits + splits - 3L))
  } else {
    totalTips <- sum(splits)

    numerator <- totalTips + totalTips - 5L
    denominator <- 2L * (totalTips - length(splits)) + 1L

    # Return:
    prod(seq.int(numerator, denominator, -2L),
         DoubleFactorial(splits + splits - 3L))
  }
}

test_that("Factorials are calculated correctly", {
  expect_equal(c(1L, 1L, 1L, 2L, 3L, 2L * 4L, 3L * 5L,
                 2L * 4L * 6L, 3L * 5L * 7L,
                 2L * 4L * 6L * 8L,
                 3L * 5L * 7L * 9L,
                 2L * 4L * 6L * 8L * 10L,
                 3L * 5L * 7L * 9L * 11L),
               as.integer(DoubleFactorial((-1):11)))
  expect_equal(floor(doubleFactorials[1:19]),
               ceiling(doubleFactorials[1:19]))
  expect_equal(doubleFactorials[1:20], DoubleFactorial(1:20))
  expect_equal(LnDoubleFactorial(-1:10), log(DoubleFactorial(-1:10)))
  expect_equal(logDoubleFactorials[1:20], log(DoubleFactorial(1:20)))
  expect_equal(log2DoubleFactorials[1:20], log2(DoubleFactorial(1:20)))
  expect_equal(LnDoubleFactorial(50001) - log(50001), LnDoubleFactorial.int(49999L))
  expect_equal(Log2DoubleFactorial(50000 - 2 + -1:1) + log2(50000 + -1:1),
               Log2DoubleFactorial(50000 + -1:1))
  expect_equal(LnDoubleFactorial.int(-1L), 0L)
  expect_equal(LnRooted.int(-1L), 0L)

  expect_error(DoubleFactorial(301))
})

test_that("N consistent with splits calculated correctly", {
  Test <- function(nExpectation, ...) {
    expect_equal(nExpectation, NUnrootedSplits(...))
    expect_equal(log(nExpectation), LnUnrootedSplits(...))
    expect_equal(log2(nExpectation), Log2UnrootedSplits(...))
  }
  Test(NUnrooted(8), 8)
  Test(NRooted(3) * NRooted(5), 3, 5)
  Test(doubleFactorials[12 + 12 - 5] / doubleFactorials[17] *
         doubleFactorials[4 + 4 - 3] ^ 3, c(4, 4, 4))


  Test <- function(...) {
    splits <- c(...)
    if (length(splits) < 3L) {
      stop("Designed to test through to NUnrootedMult")
    }
    splits <- splits[splits > 0]
    totalTips <- sum(splits)
    expect_equal(
      round(DoubleFactorial(totalTips + totalTips - 5L) /
              DoubleFactorial(2L * (totalTips - length(splits)) - 1L)
              * prod(DoubleFactorial(splits + splits - 3L))),
      NUnrootedSplits(...))
    expect_equal(log(NUnrootedSplits(...)), LnUnrootedSplits(...))
  }
  Test(49:51)
  Test(29:32)
  Test(1:10)
  expect_error(NUnrootedSplits(c(10, 152)))

})

test_that("N[Un]RootedMult() works", {
  Test <- function(...) {
    splits <- c(...)
    splits <- splits[splits > 0]
    totalTips <- sum(splits)
    # Carter et al. 1990, Theorem 2 in full
    expect_equal(1, NUnrootedMult(...) /
                   round(DoubleFactorial(totalTips + totalTips - 5L) /
                         DoubleFactorial(2L * (totalTips - length(splits)) - 1L)
                       * prod(DoubleFactorial(splits + splits - 3L))),
                 tolerance = 1e-7)
  }
  expect_equal(NUnrootedMult(c(5, 10)), NUnrootedMult(5, 10))
  Test(5, 10)
  Test(5, 0, 10)
  Test(5, 5, 10)
  expect_equal(0L, NUnrootedMult(0, 0))
  Test(0:5)
  Test(c(0, 10))
  Test(10)
})

test_that("Log rooted calculated correctly", {
  expect_equal(Log2Rooted(10:30), LnRooted(10:30) / log(2))
  expect_equal(Log2Rooted(10:30), Log2Rooted.int(10:30))
})

test_that("SPR distances calculated correctly", {
  expect_equal(0L, N1Spr(0))
  expect_equal(c(rep(0L, 3), 2L * (3:10 - 3L) * (2L * 3:10 - 7L)),
               vapply(0:10, N1Spr, integer(1)))
  expect_equal(-log2((1L + N1Spr(0:10)) / NUnrooted(0:10)), IC1Spr(0:10))
})

#' TreeTools
#'
#' "TreeTools" is an R package that provides functions for creating, modifying and
#' analysing phylogenetic trees.
#' It complements packages such as
#' [\pkg{ape}](https://cran.r-project.org/package=ape),
#' [\pkg{phangorn}](https://cran.r-project.org/package=phangorn) and
#' [\pkg{phytools}](https://cran.r-project.org/package=phytools),
#' aiming for efficient and robust implementations of functions, typically
#' applied to unweighted trees (i.e. those without edge lengths).
#'
#' Full documentation is available [online](https://ms609.github.io/TreeTools/).
#'
#' @keywords internal
"_PACKAGE"

# Suppress "NOTE: Nothing imported from Rdpack":
#' @importFrom Rdpack reprompt
NULL

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
## usethis namespace: end
NULL

test_that("Trees are sorted", {
  trees <- setNames(as.phylo(0:8, 8), paste0("as.phy", 0:8))
  expect_true(trees[[1]] == trees[[1]])
  expect_false(trees[[1]] == trees[[2]])
  
  for (i in 1:8) {
    expect_true(trees[[i]] == trees[[i]])
    expect_false(trees[[i]] > trees[[i]])
    expect_false(trees[[i]] < trees[[i]])
    for (j in (i + 1):9) {
      expect_true(trees[[i]] < trees[[j]])
      expect_true(trees[[j]] > trees[[i]])
      expect_false(trees[[j]] == trees[[i]])
      expect_false(trees[[i]] == trees[[j]])
      expect_false(trees[[i]] > trees[[j]])
      expect_false(trees[[j]] < trees[[i]])
    }
  }
  
  expect_equal(sort(trees), sort(rev(trees)))
  expect_equal(names(sort(rev(trees))), names(trees))
})

---
title: "Loading phylogenetic data into R"
author: "Martin R. Smith <martin.smith@durham.ac.uk>"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{Loading phylogenetic data into R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

It can be a bit fiddly to get a phylogenetic dataset into R, particularly if you
are not used to working with files in the Nexus format.

First off, make sure that you are comfortable 
[telling R where to find a file](filesystem-navigation.html).

Then you are ready to load raw data:

# Load raw data

## From an Excel spreadsheet

If your data is in an Excel spreadsheet, one way to load it into R is using the
'[readxl](https://readxl.tidyverse.org/)' package.
First you'll have to install it:
```r
install.packages("readxl") # You only need to do this once
```
Then you should prepare your Excel spreadsheet such that each row corresponds to
a taxon, and each column to a character.

Then you can read the data from the Excel file by telling R which sheet, rows
and columns contain your data:

```r
library("readxl")
raw_data <- as.matrix(read_excel(
  filename,
  sheet = 1,           # Loads sheet number 1 from the excel file
  range = "B1:AA21",   # Extracts columns B to AA, rows 1 to 21
  # Note that the first row is interpreted as column (character) names
  col_types = "text"   # Read all columns as character strings
))

# Read row (taxon) names from column A
# Again, the first cell will be interpreted as a column name
taxon_names <- unlist(read_excel(filename, sheet = 1, range = "A1:A21"))

rownames(raw_data) <- taxon_names
```

## From a text or CSV (comma separated values) file

Characters can be read from a text file in a similar manner to Excel.
You may need to adjust the R commands to match the particular format of your
input file.

```r
raw_data <- read.table(
  filename,            # Path to your input file
  sep = ",",           # What character separates columns?
  header = TRUE,       # Does the data contain a header row?
  row.names = 1,       # Which column contains the row names?
  na.strings = "", 
  stringsAsFactors = FALSE
)
```

## From a Nexus file

TreeTools contains an inbuilt Nexus parser:

```r
raw_data <- ReadCharacters(filename)
# Or, to go straight to PhyDat format:
as_phydat <- ReadAsPhyDat(filename)
```

This will extract character names and codings from a dataset.  It's been written
to work with datasets downloaded from [MorphoBank](https://morphobank.org/), but
my aim is for this function to handle most valid (and many invalid) NEXUS files.
If you find a file that this function can't handle, please
[let me know](https://github.com/ms609/TreeTools/issues/new?title=Unsupported+NEXUS+file&body=Please%20attach%20the%20problematic%20NEXUS%20file%20and%20describe%20the%20issue&labels=enhancement)
and I'll try to fix it.

In the meantime, alternative Nexus parsers are available: try 

```R
raw_data <- ape::read.nexus.data(filename)
```
Non-standard elements of a Nexus file might be beyond the capabilities of ape's
parser.  In particular, you will need to replace spaces in taxon names with an
underscore, and to arrange all data into a single block starting `BEGIN DATA`.
You'll need to strip out comments, character definitions and separate taxon
blocks.  

The function `readNexus` in package `phylobase` uses the NCL library and 
promises to be more powerful, but I've not been able to get it to work.

## From a TNT file

A TNT format dataset downloaded from [MorphoBank](https://morphobank.org/) can be
parsed with `ReadTntCharacters`, which might also handle other TNT-compatible
files.  If there's a file that's not being read correctly, please
[let me know](https://github.com/ms609/TreeTools/issues/new?title=Unsupported+TNT+file&body=Please%20attach%20the%20problematic%20TNT%20file%20and%20describe%20the%20issue&labels=enhancement)
and I'll try to fix it.

```r
raw_data <- ReadTntCharacters(filename)
# Or, to go straight to PhyDat format:
my_data <- ReadTntAsPhyDat(filename)
```

# Processing raw data

Next, we need the raw data in the R-friendly `phyDat` format.
If you've used the `ReadAsPhyDat` or `ReadTntAsPhyDat` functions,
then you can skip this step -- you're already there.

Otherwise, you can try 

```r 
my_data <- PhyDat(raw_data)
```

or if that doesn't work, 

```r
my_data <- MatrixToPhyDat(raw_data)
```

These functions are pretty robust, but might return an error when they encounter
an unexpected dataset format -- if they don't work on your dataset, please  
[let me know](https://github.com/ms609/TreeTools/issues/new?title=data+to+PhyDat+conversion+issue&body=Please%20describe+the+problem+here:&labels=bug).  


Failing that, you can enlist the help of the 'phangorn' package:

```r
install.packages("phangorn")
library("phangorn")
my_data <- phyDat(raw_data, type = "USER", levels = c(0:9, "-"))
```

`type="USER"` tells the parser to expect morphological data.

The `levels` parameter simply lists all the states that any character might 
take.  `0:9` includes all the integer digits from 0 to 9.  If you have
inapplicable data in your matrix, you should list `-` as a separate level as it
represents
an additional state (as handled by the Morphy implementation of [@Brazeau2018]).
If you have more complicated ambiguities, you may need to use a contrast matrix
to decode your matrix.

A contrast matrix translates the tokens used in your dataset to the
character states to which they correspond: for example decoding 'A' to {01}.
For more details, see the 'phangorn-specials' vignette in the phangorn package,
   accessible by typing '?phangorn' in the R prompt and navigating to 
   index > package vignettes.

```{r Contrast matrix}
contrast.matrix <- matrix(data = c(
# 0 1 -  # Each column corresponds to a character-state
  1, 0, 0, # Each row corresponds to a token, here 0, denoting the 
           # character-state set {0} 
  0, 1, 0, # 1 | {1}
  0, 0, 1, # - | {-}
  1, 1, 0, # A | {01}
  1, 1, 0, # + | {01}
  1, 1, 1  # ? | {01-}
), ncol = 3, # ncol corresponds to the number of columns in the matrix
byrow = TRUE)
dimnames(contrast.matrix) <- list(
  c(0, 1, "-", "A", "+", "?"), # A list of the tokens corresponding to each row
                               # in the contrast matrix
  c(0, 1, "-") # A list of the character-states corresponding to the columns 
               # in the contrast matrix
)

contrast.matrix
```

If you need to use a contrast matrix, convert the data using 
```r
my.phyDat <- phyDat(my.data, type = "USER", contrast = contrast.matrix)
```

# Store processed data

To simplify subsequent analysis, or to allow the data to be read into other
contexts (e.g. via the [`TreeSearch::EasyTrees()`](
   https://ms609.github.io/TreeSearch/articles/tree-search.html) interface),
you could save your processed data in Nexus format:

```r
ape::write.nexus.data(my.phyDat, "my_data.nex", format = "standard")
```

# What next?

You might want to:

- [Load a phylogenetic tree](load-trees.html) into R.

- Conduct parsimony search using Brazeau, Guillerme & Smith's [-@Brazeau2018]
  [approach to inapplicable data](
   https://ms609.github.io/TreeSearch/articles/tree-search.html),
  or using [Profile parsimony](
   https://ms609.github.io/TreeSearch/articles/profile.html) [@Faith2001].

# References

---
title: "File system navigation in R"
author: "Martin R. Smith <martin.smith@durham.ac.uk>"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
bibliography: ../inst/REFERENCES.bib
csl: ../inst/apa-old-doi-prefix.csl
vignette: >
  %\VignetteIndexEntry{File system navigation in R}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

Before you can open a file, you need to tell R where to find it. 
You can do this by providing the full path to the file on your system.
Be careful to use forward slashes (`/`, not `\`, which you'll get if you
copy file paths in Windows).

```r
filename <- "C:/nexus/matrix.nex"
```

You can save typing by giving R a working directory.  You can 
think of R as having a file explorer window open invisibly in the
background.

You can see the folder that's open at the moment by typing `getwd()` at the console.

`setwd()` tells R to open a different folder instead.

`setwd('../')` tells R to go up to a parent directory.
(You can [do this using the Graphical User Interface](
https://www.princeton.edu/~otorres/RStudio101.pdf) in RStudio).

By setting the directory that your files are in as the working directory, 
you only need to specify the filename:

```r 
setwd("C:/nexus/") # You only need to do this once
filename <- "matrix.nex"
# Do something with this file
#

filename <- "tree.nex"
# Do something with this file
#
```

## What next?

Now you know how to locate files, you might want to load a
[dataset](load-data.html) or [phylogenetic tree](load-trees.html) into R.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{Lobo.data}
\alias{Lobo.data}
\alias{Lobo.phy}
\title{Data from Zhang et al. 2016}
\format{
An object of class \code{list} of length 48.

An object of class \code{phyDat} of length 48.
}
\source{
\insertCite{Zhang2016;textual}{TreeTools}
}
\usage{
Lobo.data

Lobo.phy
}
\description{
Phylogenetic data from \insertCite{Zhang2016;textual}{TreeTools} in raw
(\code{Lobo.data}) and \code{phyDat} (\code{Lobo.phy}) formats.
}
\examples{
data("Lobo", package = "TreeTools")
Lobo.data
Lobo.phy
}
\references{
\insertAllCited{}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TreeNumber.R
\encoding{UTF-8}
\name{TreeNumber}
\alias{TreeNumber}
\alias{as.TreeNumber}
\alias{as.TreeNumber.phylo}
\alias{as.TreeNumber.multiPhylo}
\alias{as.TreeNumber.character}
\alias{as.TreeNumber.TreeNumber}
\alias{as.TreeNumber.MixedBase}
\alias{as.MixedBase.TreeNumber}
\alias{as.MixedBase.integer64}
\alias{as.MixedBase.numeric}
\alias{as.phylo.numeric}
\alias{as.phylo.TreeNumber}
\alias{as.MixedBase}
\alias{as.MixedBase.MixedBase}
\alias{as.MixedBase.phylo}
\alias{as.MixedBase.multiPhylo}
\alias{as.phylo.MixedBase}
\title{Unique integer indices for bifurcating tree topologies}
\usage{
as.TreeNumber(x, ...)

\method{as.TreeNumber}{phylo}(x, ...)

\method{as.TreeNumber}{multiPhylo}(x, ...)

\method{as.TreeNumber}{character}(x, nTip, tipLabels = TipLabels(nTip), ...)

\method{as.TreeNumber}{TreeNumber}(x, ...)

\method{as.TreeNumber}{MixedBase}(x, ...)

\method{as.MixedBase}{TreeNumber}(x, ...)

\method{as.MixedBase}{integer64}(x, tipLabels = NULL, ...)

\method{as.MixedBase}{numeric}(x, tipLabels = NULL, ...)

\method{as.phylo}{numeric}(x, nTip = attr(x, "nTip"), tipLabels = attr(x, "tip.label"), ...)

\method{as.phylo}{TreeNumber}(x, nTip = attr(x, "nTip"), tipLabels = attr(x, "tip.label"), ...)

as.MixedBase(x, ...)

\method{as.MixedBase}{MixedBase}(x, ...)

\method{as.MixedBase}{phylo}(x, ...)

\method{as.MixedBase}{multiPhylo}(x, ...)

\method{as.phylo}{MixedBase}(x, nTip = attr(x, "nTip"), tipLabels = attr(x, "tip.label"), ...)
}
\arguments{
\item{x}{Integer identifying the tree (see details).}

\item{\dots}{Additional parameters for consistency with S3 methods (unused).}

\item{nTip}{Integer specifying number of leaves in the tree.}

\item{tipLabels}{Character vector listing the labels assigned to each tip
in a tree, perhaps obtained using \code{\link[=TipLabels]{TipLabels()}}.}
}
\value{
\code{as.TreeNumber()} returns an object of class \code{TreeNumber} with
attributes \code{nTip} and \code{tip.label}.
For trees with at most 19 leaves the underlying storage is a single
\code{integer64} value (class \code{c("TreeNumber", "integer64")}), enabling
\code{integer64} arithmetic and exact round-tripping through \code{as.MixedBase()}.
For trees with 20--51 leaves the number exceeds 2^64, so it is stored as a
decimal character string (class \code{c("TreeNumber", "character")}).
If \code{x} is a list of trees or a \code{multiPhylo} object,
\code{as.TreeNumber()} returns a corresponding list of \code{TreeNumber} objects.

\code{as.phylo.numeric()} returns a tree of class \code{phylo}.
}
\description{
Functions converting between phylogenetic trees and their unique decimal
representation, based on a concept by John Tromp, employed in
\insertCite{Li1996;textual}{TreeTools}.
}
\details{
There are \code{NUnrooted(n)} unrooted trees with \emph{n} leaves.
As such, each \emph{n}-leaf tree can be uniquely identified by a non-negative
integer \emph{x} < \code{NUnrooted(n)}.

This integer can be converted by a tree by treating it as a mixed-base
number, with bases 1, 3, 5, 7, … (2 \emph{n} - 5).

Each digit of this mixed base number corresponds to a leaf, and determines
the location on a growing tree to which that leaf should be added.

We start with a two-leaf tree, and treat 0 as the origin of the tree.

\if{html}{\out{<div class="sourceCode">}}\preformatted{
0 ---- 1

}\if{html}{\out{</div>}}

We add leaf 2 by breaking an edge and inserting a node (numbered
\code{2 + nTip - 1}).
In this example, we'll work up to a six-leaf tree; this node will be numbered
2 + 6 - 1 = 7.
There is only one edge on which leaf 2 can be added.  Let's add node 7 and
leaf 2:

\if{html}{\out{<div class="sourceCode">}}\preformatted{
0 ---- 7 ---- 1
       |
       |
       2

}\if{html}{\out{</div>}}

There are now three edges on which leaf 3 can be added.  Our options are:

Option 0: the edge leading to 1;

Option 1: the edge leading to 2;

Option 2: the edge leading to 7.

If we select option 1, we produce:

\if{html}{\out{<div class="sourceCode">}}\preformatted{
0 ---- 7 ---- 1
       |
       |
       8 ---- 2
       |
       |
       3

}\if{html}{\out{</div>}}

\code{1} is now the final digit of our mixed-base number.

There are five places to add leaf 4:

Option 0: the edge leading to 1;

Option 1: the edge leading to 2;

Option 2: the edge leading to 3;

Option 3: the edge leading to 7;

Option 4: the edge leading to 8.

If we chose option 3, then \code{3} would be the penultimate digit of our
mixed-base number.

If we chose option 0 for the next two additions, we could specify this tree
with the mixed-base number 0021.  We can convert this into decimal:

0 × (1 × 3 × 5 × 9) +

0 × (1 × 3 × 5) +

3 × (1 × 3) +

1 × (1)

= 10

\code{as.TreeNumber()} supports up to 51 leaves.
For trees with at most 19 leaves, the number fits in a 64-bit integer and
the \code{TreeNumber} is stored as an \code{integer64} (via the \code{bit64} package),
enabling arithmetic and exact round-tripping via \code{as.MixedBase()}.
For trees with 20--51 leaves, there are more than 2^64 distinct topologies,
so the tree number is stored as a decimal character string instead.

Package developers can use the C++ header \code{TreeTools/tree_number.h}
(via \code{LinkingTo: TreeTools}) for the underlying 256-bit encoding
(\code{tree_num_t}) directly.
}
\examples{
tree <- as.phylo(10, nTip = 6)
plot(tree)
as.TreeNumber(tree)

# Trees with 20--51 leaves are stored as decimal strings:
as.TreeNumber(BalancedTree(19))  # integer64-backed
as.TreeNumber(BalancedTree(51))  # character-backed

# If > 9 digits, represent the tree number as a string.
treeNumber <- as.TreeNumber("1234567890123", nTip = 14)
tree <- as.phylo(treeNumber)
as.phylo(0:2, nTip = 6, tipLabels = letters[1:6])

}
\references{
\insertAllCited{}
}
\seealso{
Describe the shape of a tree topology, independent of leaf labels:
\code{\link[=TreeShape]{TreeShape()}}

Other tree generation functions:
\code{\link[=ConstrainedNJ]{ConstrainedNJ()}},
\code{\link{GenerateTree}},
\code{\link[=NJTree]{NJTree()}},
\code{\link{TrivialTree}}

Other 'TreeNumber' utilities:
\code{\link[=is.TreeNumber]{is.TreeNumber()}},
\code{\link[=print.TreeNumber]{print.TreeNumber()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{'TreeNumber' utilities}
\concept{tree generation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports-manual.R
\name{root_on_node}
\alias{root_on_node}
\title{Wrapper for internal C function \code{root_on_node()}}
\usage{
root_on_node(phy, outgroup)
}
\arguments{
\item{phy}{Minimally, a named list with entries \code{edge} and \code{Nnode}, in the
format of equivalent entries in a tree of class \code{phylo}. \code{edge.length} will
also be considered if supplied.}

\item{outgroup}{Integer specifying index of leaf or node to set as the
outgroup.}
}
\value{
\code{root_on_node()} returns \code{phy} rooted on the specified node.
}
\description{
Direct entry point to \code{root_on_node()}; recommended for expert use only.
\code{RootTree()} checks that input is properly formatted and is recommended
for general use.
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Splits.R
\name{edge_to_splits}
\alias{edge_to_splits}
\title{Efficiently convert edge matrix to splits}
\usage{
edge_to_splits(
  edge,
  edgeOrder,
  tipLabels = NULL,
  asSplits = TRUE,
  nTip = NTip(edge),
  ...
)
}
\arguments{
\item{edge}{A matrix with two columns, with each row listing the parent and
child node of an edge in a phylogenetic tree.  Property \code{edge} of objects
of class \code{phylo}.}

\item{edgeOrder}{Integer vector such that \code{edge[edgeOrder, ]} returns a
postorder ordering of edges.}

\item{tipLabels}{Character vector specifying sequence in which to order
tip labels.  Label order must (currently) match to combine or compare separate
\code{Splits} objects.}

\item{asSplits}{Logical specifying whether to return a \code{Splits} object,
or an unannotated two-dimensional array (useful where performance is
paramount).}

\item{nTip}{Integer specifying number of leaves in tree.}

\item{...}{Presently unused.}
}
\value{
\code{edge_to_splits()} uses the same return format as \code{as.Splits()}.
}
\description{
Wrapper for internal C++ function for maximum efficiency.
Improper input may crash R.  Behaviour not guaranteed.
It is advisable to contact the package maintainers before
relying on this function.
}
\seealso{
\code{\link[=Splits]{as.Splits()}} offers a safe access point to this
function that should be suitable for most users.
}
\concept{C++ wrappers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{ExtractTaxa}
\alias{ExtractTaxa}
\alias{NexusTokens}
\title{Extract taxa from a matrix block}
\usage{
ExtractTaxa(matrixLines, character_num = NULL, continuous = FALSE)

NexusTokens(tokens, character_num = NULL)
}
\arguments{
\item{matrixLines}{Character vector containing lines of a file that include
a phylogenetic matrix. See \code{\link[=ReadCharacters]{ReadCharacters()}} for expected format.}

\item{character_num}{Index of character(s) to return.
\code{NULL}, the default, returns all characters.}

\item{continuous}{Logical specifying whether characters are continuous.
Treated as discrete if \code{FALSE}.}

\item{tokens}{Vector of character strings corresponding to phylogenetic
tokens.}
}
\value{
\code{ExtractTaxa()} returns a matrix with \emph{n} rows, each named for the
relevant taxon, and \emph{c} columns,
each corresponding to the respective character specified in \code{character_num}.

\code{NexusTokens()} returns a character vector in which each entry
corresponds to the states of a phylogenetic character, or a list containing
an error message if input is invalid.
}
\description{
Extract leaf labels and character states from a Nexus-formatted matrix.
}
\examples{
fileName <- paste0(system.file(package = "TreeTools"),
                   "/extdata/input/dataset.nex")
matrixLines <- readLines(fileName)[6:11]
ExtractTaxa(matrixLines)

NexusTokens("01[01]-?")
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SplitFunctions.R
\name{TipLabels}
\alias{TipLabels}
\alias{TipLabels.default}
\alias{TipLabels.matrix}
\alias{TipLabels.logical}
\alias{TipLabels.phylo}
\alias{TipLabels.phyDat}
\alias{TipLabels.MixedBase}
\alias{TipLabels.TreeNumber}
\alias{TipLabels.Splits}
\alias{TipLabels.list}
\alias{TipLabels.multiPhylo}
\alias{TipLabels.character}
\alias{TipLabels.numeric}
\alias{AllTipLabels}
\alias{AllTipLabels.list}
\alias{AllTipLabels.multiPhylo}
\alias{AllTipLabels.phylo}
\alias{AllTipLabels.Splits}
\alias{AllTipLabels.TreeNumber}
\alias{AllTipLabels.matrix}
\title{Extract tip labels}
\usage{
TipLabels(x, single = TRUE)

\method{TipLabels}{default}(x, single = TRUE)

\method{TipLabels}{matrix}(x, single = TRUE)

\method{TipLabels}{logical}(x, single = TRUE)

\method{TipLabels}{phylo}(x, single = TRUE)

\method{TipLabels}{phyDat}(x, single = TRUE)

\method{TipLabels}{MixedBase}(x, single = TRUE)

\method{TipLabels}{TreeNumber}(x, single = TRUE)

\method{TipLabels}{Splits}(x, single = TRUE)

\method{TipLabels}{list}(x, single = FALSE)

\method{TipLabels}{multiPhylo}(x, single = FALSE)

\method{TipLabels}{character}(x, single = TRUE)

\method{TipLabels}{numeric}(x, single = TRUE)

\method{TipLabels}{phyDat}(x, single = TRUE)

AllTipLabels(x)

\method{AllTipLabels}{list}(x)

\method{AllTipLabels}{multiPhylo}(x)

\method{AllTipLabels}{phylo}(x)

\method{AllTipLabels}{Splits}(x)

\method{AllTipLabels}{TreeNumber}(x)

\method{AllTipLabels}{matrix}(x)
}
\arguments{
\item{x}{An object of a supported class (see Usage section above).}

\item{single}{Logical specifying whether to report the labels for the first
object only (\code{TRUE}), or for each object in a list (\code{FALSE}).}
}
\value{
\code{TipLabels()} returns a character vector listing the tip labels
appropriate to \code{x}. If \code{x} is a single integer, this will be a vector
\code{t1}, \code{t2} ... \code{tx}, to match the default of \code{\link[ape]{rtree}()}.
}
\description{
\code{TipLabels()} extracts labels from an object: for example, names of taxa in
a phylogenetic tree or data matrix.  \code{AllTipLabels()} extracts all labels,
where entries of a list of trees may pertain to different taxa.
}
\examples{
TipLabels(BalancedTree(letters[5:1]))
TipLabels(5)

data("Lobo")
head(TipLabels(Lobo.phy))

AllTipLabels(c(BalancedTree(4), PectinateTree(8)))

}
\seealso{
Other tree properties:
\code{\link[=Cherries]{Cherries()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=EdgeRatio]{EdgeRatio()}},
\code{\link[=LongBranch]{LongBranch()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=PathLengths]{PathLengths()}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TreeIsRooted]{TreeIsRooted()}},
\code{\link[=Treeness]{Treeness()}}

Other Splits operations:
\code{\link[=LabelSplits]{LabelSplits()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=PolarizeSplits]{PolarizeSplits()}},
\code{\link[=SplitFrequency]{SplitFrequency()}},
\code{\link{Splits}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipsInSplits]{TipsInSplits()}},
\code{\link{match,Splits,Splits-method}},
\code{\link[=xor]{xor()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{Splits operations}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SplitFunctions.R
\name{CompatibleSplits}
\alias{CompatibleSplits}
\alias{.CompatibleSplit}
\alias{.CompatibleRaws}
\title{Which splits are compatible?}
\usage{
CompatibleSplits(splits, splits2)

.CompatibleSplit(a, b, nTip)

.CompatibleRaws(rawA, rawB, bitmask)
}
\arguments{
\item{splits}{An object of class \code{\link[=as.Splits]{Splits}}.}

\item{splits2}{A second \code{Splits} object.}

\item{a, b}{\link[=raw]{Raw} representations of splits, from a row of a \code{Splits}
object.}

\item{rawA, rawB}{Raw representations of splits.}

\item{bitmask}{Raw masking bits that do not correspond to tips.}
}
\value{
\code{CompatibleSplits} returns a logical matrix specifying whether each
split in \code{splits} is compatible with each split in \code{splits2}.

\code{.CompatibleSplit} returns a logical vector stating whether splits
are compatible.

\code{.CompatibleRaws} returns a logical vector specifying whether input
raws are compatible.
}
\description{
Which splits are compatible?
}
\examples{
splits <- as.Splits(BalancedTree(8))
splits2 <- as.Splits(PectinateTree(8))

summary(splits)
summary(splits2)

CompatibleSplits(splits, splits2)

}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SplitFunctions.R
\name{SplitMatchProbability}
\alias{SplitMatchProbability}
\alias{LnSplitMatchProbability}
\title{Probability of matching this well}
\usage{
SplitMatchProbability(split1, split2)

LnSplitMatchProbability(split1, split2)
}
\arguments{
\item{split1, split2}{Logical vectors listing terminals in same order, such that
each terminal is identified as a member of the ingroup (\code{TRUE}) or outgroup
(\code{FALSE}) of the respective bipartition split.}
}
\value{
\code{SplitMatchProbability()} returns a numeric giving the proportion
of permissible non-trivial splits that divide the terminals into bipartitions
of the sizes given, that match as well as \code{split1} and \code{split2} do.

\code{LnSplitMatchProbability()} returns the natural logarithm of the
probability.
}
\description{
(\code{Ln})\code{SplitMatchProbability()}calculates the probability that two random
splits of the sizes provided will be at least as similar as the two
specified.
}
\examples{
split1 <- as.Splits(c(rep(TRUE, 4), rep(FALSE, 4)))
split2 <- as.Splits(c(rep(TRUE, 3), rep(FALSE, 5)))
SplitMatchProbability(split1, split2)
LnSplitMatchProbability(split1, split2)
}
\seealso{
Other split information functions:
\code{\link[=CharacterInformation]{CharacterInformation()}},
\code{\link[=SplitInformation]{SplitInformation()}},
\code{\link[=TreesMatchingSplit]{TreesMatchingSplit()}},
\code{\link[=UnrootedTreesMatchingSplit]{UnrootedTreesMatchingSplit()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_ancestors.R
\name{EdgeAncestry}
\alias{EdgeAncestry}
\title{Ancestors of an edge}
\usage{
EdgeAncestry(edge, parent, child, stopAt = (parent == min(parent)))
}
\arguments{
\item{edge}{Integer specifying the number of the edge whose child edges
should be returned.}

\item{parent}{Integer vector corresponding to the first column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 1]}}

\item{child}{Integer vector corresponding to the second column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 2]}.}

\item{stopAt}{Integer or logical vector specifying the edge(s) at which to
terminate the search; defaults to the edges with the smallest parent,
which will be the root edges if nodes are numbered \link{Cladewise} or in
\link{Preorder}.}
}
\value{
\code{EdgeAncestry()} returns a logical vector stating whether each edge
in turn is a descendant of the specified edge.
}
\description{
Quickly identify edges that are "ancestral" to a particular edge in a tree.
}
\examples{
tree <- PectinateTree(6)
plot(tree)
ape::edgelabels()
parent <- tree$edge[, 1]
child <- tree$edge[, 2]
EdgeAncestry(7, parent, child)
which(EdgeAncestry(7, parent, child, stopAt = 4))

}
\seealso{
Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=PaintTree]{PaintTree()}},
\code{\link[=RootNode]{RootNode()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ImposeConstraint.R
\name{ImposeConstraint}
\alias{ImposeConstraint}
\alias{AddUnconstrained}
\title{Force a tree to match a constraint}
\usage{
ImposeConstraint(tree, constraint)

AddUnconstrained(constraint, toAdd, asPhyDat = TRUE)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{constraint}{Either an object of class \code{phyDat}, in which case returned
trees will be perfectly compatible with each character in \code{constraint};
or a tree of class \code{phylo}, in which each node in \code{constraint} will occur in
the returned tree.
See \href{https://ms609.github.io/TreeSearch/articles/tree-search.html}{vignette}
for further examples.}

\item{toAdd}{Character vector specifying taxa to add to constraint.}

\item{asPhyDat}{Logical: if \code{TRUE}, return a \code{phyDat} object; if \code{FALSE},
return a matrix.}
}
\value{
\code{ImposeConstraint()} returns a tree of class \code{phylo}, consistent
with \code{constraint}.
}
\description{
Modify a tree such that it matches a specified constraint.
This is at present a somewhat crude implementation that attempts to retain
much of the structure of \code{tree} whilst guaranteeing compatibility with
each entry in \code{constraint}.
}
\section{Functions}{
\itemize{
\item \code{AddUnconstrained()}: Expand a constraint to include unconstrained
taxa.

}}
\examples{
tips <- letters[1:9]
tree <- as.phylo(1, 9, tips)
plot(tree)

constraint <- StringToPhyDat("0000?1111 000111111 0000??110", tips, FALSE)
plot(ImposeConstraint(tree, constraint))
}
\seealso{
Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{CollapseNode}()},
\code{\link{ConsensusWithout}()},
\code{\link{DropTip}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{MakeTreeBinary}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SplitFunctions.R
\name{Subsplit}
\alias{Subsplit}
\title{Subset of a split on fewer leaves}
\usage{
Subsplit(splits, tips, keepAll = FALSE, unique = TRUE)
}
\arguments{
\item{splits}{An object of class \code{\link[=as.Splits]{Splits}}.}

\item{tips}{A vector specifying a subset of the leaf labels applied to \code{split}.}

\item{keepAll}{logical specifying whether to keep entries that define trivial
splits (i.e. splits of zero or one leaf) on the subset of leaves.}

\item{unique}{logical specifying whether to remove duplicate splits.}
}
\value{
\code{Subsplit()} returns an object of class \code{Splits}, defined on the
leaves \code{tips}.
}
\description{
\code{Subsplit()} removes leaves from a \code{Splits} object.
}
\examples{
splits <- as.Splits(PectinateTree(letters[1:9]))
splits
efgh <- Subsplit(splits, tips = letters[5:8], keepAll = TRUE)
summary(efgh)

TrivialSplits(efgh)

summary(Subsplit(splits, tips = letters[5:8], keepAll = FALSE))
}
\seealso{
\code{\link[=KeepTip]{KeepTip()}} is a less flexible but faster equivalent.

Other split manipulation functions:
\code{\link[=DropTip]{DropTip()}},
\code{\link[=SplitConsistent]{SplitConsistent()}},
\code{\link[=TrivialSplits]{TrivialSplits()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split manipulation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{SampleOne}
\alias{SampleOne}
\title{Select element at random}
\usage{
SampleOne(x, len = length(x))
}
\arguments{
\item{x}{A vector to sample.}

\item{len}{(Optional) Integer specifying length of \code{x}.}
}
\value{
\code{SampleOne()} returns a length one vector, randomly sampled from \code{x}.
}
\description{
\code{SampleOne()} is a fast alternative to  \code{\link[=sample]{sample()}} that avoids some checks.
}
\examples{
SampleOne(9:10)
SampleOne(letters[1:4])

}
\seealso{
Other utility functions:
\code{\link{ClusterTable}},
\code{\link{ClusterTable-methods}},
\code{\link[=Hamming]{Hamming()}},
\code{\link[=MSTEdges]{MSTEdges()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link[=UnshiftTree]{UnshiftTree()}},
\code{\link[=as.multiPhylo]{as.multiPhylo()}},
\code{\link{match,phylo,phylo-method}},
\code{\link[=sapply64]{sapply64()}},
\code{\link[=sort.multiPhylo]{sort.multiPhylo()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{ReadCharacters}
\alias{ReadCharacters}
\alias{ReadTntCharacters}
\alias{ReadTNTCharacters}
\alias{ReadNotes}
\alias{ReadAsPhyDat}
\alias{ReadTntAsPhyDat}
\alias{ReadTNTAsPhyDat}
\alias{PhyDat}
\title{Read phylogenetic characters from file}
\usage{
ReadCharacters(filepath, character_num = NULL, encoding = "UTF8")

ReadTntCharacters(
  filepath,
  character_num = NULL,
  type = NULL,
  encoding = "UTF8"
)

ReadTNTCharacters(
  filepath,
  character_num = NULL,
  type = NULL,
  encoding = "UTF8"
)

ReadNotes(filepath, encoding = "UTF8")

ReadAsPhyDat(...)

ReadTntAsPhyDat(...)

ReadTNTAsPhyDat(...)

PhyDat(dataset)
}
\arguments{
\item{filepath}{character string specifying location of file, or a
\link[base:connections]{connection} to the file.}

\item{character_num}{Index of character(s) to return.
\code{NULL}, the default, returns all characters.}

\item{encoding}{Character encoding of input file.}

\item{type}{Character vector specifying categories of data to extract from
file. Setting \code{type = c("num", "dna")} will return only characters
following a \verb{&[num]} or \verb{&[dna]} tag in a TNT input file, listing \code{num}
character blocks before \code{dna} characters.
Leave as \code{NULL} (the default) to return all characters in their original
sequence.}

\item{\dots}{Parameters to pass to \verb{Read[Tnt]Characters()}.}

\item{dataset}{list of taxa and characters, in the format produced by
\code{\link[ape]{read.nexus.data}()}:
a list of sequences each made of a single character vector,
and named with the taxon name.}
}
\value{
\code{ReadCharacters()} and \code{ReadTNTCharacters()} return a matrix whose
row names correspond to tip labels, and
column names correspond to character labels, with the
attribute \code{state.labels} listing the state labels for each character; or
a list of length one containing a character string explaining why the
function call was unsuccessful.

\code{ReadAsPhyDat()} and \code{ReadTntAsPhyDat()} return a \code{phyDat} object.

\code{ReadNotes()} returns a list in which each entry corresponds to a
single character, and itself contains a list of with two elements:
\enumerate{
\item A single character object listing any notes associated with the character
\item A named character vector listing the notes associated with each taxon
for that character, named with the names of each note-bearing taxon.
}
}
\description{
Parse a Nexus \insertCite{Maddison1997}{TreeTools} or
TNT \insertCite{Goloboff2008}{TreeTools} file, reading character states and
names.
}
\details{
Tested with matrices downloaded from \href{https://morphobank.org/}{MorphoBank}
\insertCite{OLeary2011}{TreeTools}, but should also work more widely; please
\href{https://github.com/ms609/TreeTools/issues/new?title=Error+parsing+Nexus+file&body=<!--Tell+me+more+and+attach+your+file...-->}{report}
incompletely or incorrectly parsed files.

Matrices must contain only continuous or only discrete characters;
maximum one matrix per file.  Continuous characters will be read as strings
(i.e. base type "character").

The encoding of an input file will be automatically determined by R.
Errors pertaining to an \verb{invalid multibyte string} or
\verb{string invalid at that locale} indicate that R has failed to detect
the appropriate encoding.  Either
\href{https://support.posit.co/hc/en-us/articles/200532197-Character-Encoding-in-the-RStudio-IDE}{re-save the file}
in a supported encoding (\code{UTF-8} is a good choice) or
specify the file encoding (which you can find by, for example, opening in
\href{https://notepad-plus-plus.org/downloads/}{Notepad++} and identifying
the highlighted option in the "Encoding" menu) following the example below.
}
\section{Functions}{
\itemize{
\item \code{PhyDat()}: A convenient wrapper for \pkg{phangorn}'s
\code{phyDat()}, which converts a \strong{list} of morphological characters into a
\code{phyDat} object.
If your morphological characters are in the form of a \strong{matrix}, perhaps
because they have been read using \code{\link[=read.table]{read.table()}}, try \code{\link[=MatrixToPhyDat]{MatrixToPhyDat()}}
instead.

}}
\examples{
fileName <- paste0(system.file(package = "TreeTools"),
                   "/extdata/input/dataset.nex")
ReadCharacters(fileName)

fileName <- paste0(system.file(package = "TreeTools"),
                   "/extdata/tests/continuous.nex")

continuous <- ReadCharacters(fileName, encoding = "UTF8")

# To convert from strings to numbers:
at <- attributes(continuous)
continuous <- suppressWarnings(as.numeric(continuous))
attributes(continuous) <- at
continuous
}
\references{
\insertAllCited{}
}
\seealso{
\itemize{
\item Convert between matrices and \code{phyDat} objects: \code{\link[=MatrixToPhyDat]{MatrixToPhyDat()}}
\item Write characters to TNT-format file: \code{\link[=WriteTntCharacters]{WriteTntCharacters()}}
}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Splits.R
\name{xor}
\alias{xor}
\alias{xor,Splits,Splits-method}
\title{Exclusive OR operation}
\usage{
xor(x, y)

\S4method{xor}{Splits,Splits}(x, y)
}
\arguments{
\item{x, y}{Objects to be compared.}
}
\description{
Exclusive OR operation
}
\seealso{
Other Splits operations:
\code{\link[=LabelSplits]{LabelSplits()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=PolarizeSplits]{PolarizeSplits()}},
\code{\link[=SplitFrequency]{SplitFrequency()}},
\code{\link{Splits}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TipsInSplits]{TipsInSplits()}},
\code{\link{match,Splits,Splits-method}}
}
\concept{Splits operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sort.R
\name{sort.multiPhylo}
\alias{sort.multiPhylo}
\alias{==.phylo}
\alias{<.phylo}
\alias{>.phylo}
\alias{==.MixedBase}
\alias{<.MixedBase}
\alias{>.MixedBase}
\title{Sort a list of phylogenetic trees}
\usage{
\method{sort}{multiPhylo}(x, decreasing = FALSE, na.last = NA, ...)

\method{==}{phylo}(e1, e2)

\method{<}{phylo}(e1, e2)

\method{>}{phylo}(e1, e2)

\method{==}{MixedBase}(e1, e2)

\method{<}{MixedBase}(e1, e2)

\method{>}{MixedBase}(e1, e2)
}
\arguments{
\item{x, decreasing, na.last, \dots}{As in \code{\link[=sort]{sort()}}.}

\item{e1, e2}{Objects to be compared.}
}
\description{
Trees are sorted by their \link[=TreeNumber]{mixed base representation},
treating their leaves in the order of their labels (i.e. alphabetically,
if leaves are labelled with text).
}
\examples{
sort(as.phylo(5:0, 7))
}
\seealso{
Other utility functions:
\code{\link{ClusterTable}},
\code{\link{ClusterTable-methods}},
\code{\link[=Hamming]{Hamming()}},
\code{\link[=MSTEdges]{MSTEdges()}},
\code{\link[=SampleOne]{SampleOne()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link[=UnshiftTree]{UnshiftTree()}},
\code{\link[=as.multiPhylo]{as.multiPhylo()}},
\code{\link{match,phylo,phylo-method}},
\code{\link[=sapply64]{sapply64()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phylo.R
\name{Subtree}
\alias{Subtree}
\title{Extract a subtree}
\usage{
Subtree(tree, node)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}, with internal
numbering in cladewise order (use \code{\link{Preorder}(tree)} or (slower)
\code{\link{Cladewise}(tree)}) .}

\item{node}{The number of the node at the base of the clade to be extracted.}
}
\value{
\code{Subtree()} returns a tree of class \code{phylo} that represents a
clade extracted from the original tree.
}
\description{
\code{Subtree()} safely extracts a clade from a phylogenetic tree.
}
\details{
Modified from the \pkg{ape} function \code{\link[ape]{extract.clade}}, which
sometimes behaves unpredictably.
Unlike extract.clade, this function supports the extraction of "clades"
that constitute a single tip.
}
\examples{
tree <- Preorder(BalancedTree(8))
plot(tree)
ape::nodelabels()
ape::nodelabels(13, 13, bg="yellow")

plot(Subtree(tree, 13))

}
\seealso{
Other tree manipulation:
\code{\link[=AddTip]{AddTip()}},
\code{\link[=CollapseNode]{CollapseNode()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=DropTip]{DropTip()}},
\code{\link[=ImposeConstraint]{ImposeConstraint()}},
\code{\link[=KeptPaths]{KeptPaths()}},
\code{\link[=KeptVerts]{KeptVerts()}},
\code{\link[=LeafLabelInterchange]{LeafLabelInterchange()}},
\code{\link[=MakeTreeBinary]{MakeTreeBinary()}},
\code{\link[=Renumber]{Renumber()}},
\code{\link[=RenumberTips]{RenumberTips()}},
\code{\link[=RenumberTree]{RenumberTree()}},
\code{\link[=RootTree]{RootTree()}},
\code{\link[=SortTree]{SortTree()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_rearrangement.R
\name{CollapseNode}
\alias{CollapseNode}
\alias{CollapseNode.phylo}
\alias{CollapseEdge}
\title{Collapse nodes on a phylogenetic tree}
\usage{
CollapseNode(tree, nodes)

\method{CollapseNode}{phylo}(tree, nodes)

CollapseEdge(tree, edges)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{nodes, edges}{Integer vector specifying the nodes or edges in the tree
to be dropped.
(Use \code{\link[ape]{nodelabels}()} or
\code{\link[ape:nodelabels]{edgelabels}()}
to view numbers on a plotted tree.)}
}
\value{
\code{CollapseNode()} and \code{CollapseEdge()} return a tree of class \code{phylo},
corresponding to \code{tree} with the specified nodes or edges collapsed.
The length of each dropped edge will (naively) be added to each descendant
edge.
}
\description{
Collapses specified nodes or edges on a phylogenetic tree, resulting in
polytomies.
}
\examples{
oldPar <- par(mfrow = c(3, 1), mar = rep(0.5, 4))

tree <- as.phylo(898, 7)
tree$edge.length <- 11:22
plot(tree)
nodelabels()
edgelabels()
edgelabels(round(tree$edge.length, 2),
           cex = 0.6, frame = "n", adj = c(1, -1))

# Collapse by node number
newTree <- CollapseNode(tree, c(12, 13))
plot(newTree)
nodelabels()
edgelabels(round(newTree$edge.length, 2),
           cex = 0.6, frame = "n", adj = c(1, -1))

# Collapse by edge number
newTree <- CollapseEdge(tree, c(2, 4))
plot(newTree)

par(oldPar)

}
\seealso{
Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{ConsensusWithout}()},
\code{\link{DropTip}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{MakeTreeBinary}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}
}
\author{
Martin R. Smith
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_properties.R
\name{NodeDepth}
\alias{NodeDepth}
\title{Distance of each node from tree exterior}
\usage{
NodeDepth(x, shortest = FALSE, includeTips = TRUE)
}
\arguments{
\item{x}{A tree of class \code{phylo}, its \verb{$edge} property, or a list thereof.}

\item{shortest}{Logical specifying whether to calculate the length of the
shortest away-from-root path to a leaf.  If \code{FALSE}, the length of the
longest such route will be returned.}

\item{includeTips}{Logical specifying whether to include leaves
(each of depth zero) in return value.}
}
\value{
\code{NodeDepth()} returns an integer vector specifying the depth of
each external and internal node in \code{x}.
}
\description{
\code{NodeDepth()} evaluates how "deep" each node is within a tree.
}
\details{
For a rooted tree, the depth of a node is the minimum (if \code{shortest = TRUE})
or maximum  (\code{shortest = FALSE}) number of edges that must be traversed,
moving away from the root, to reach a leaf.

Unrooted trees are treated as if a root node occurs in the "middle" of the
tree, meaning the position that will minimise the maximum node depth.
}
\examples{
tree <- CollapseNode(BalancedTree(10), c(12:13, 19))
plot(tree)
nodelabels(NodeDepth(tree, includeTips = FALSE))


}
\seealso{
\code{\link[ape:node.depth]{ape::node.depth}} returns the number of tips descended from a
node.

Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=PaintTree]{PaintTree()}},
\code{\link[=RootNode]{RootNode()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Information.R
\name{TreesMatchingSplit}
\alias{TreesMatchingSplit}
\alias{LnTreesMatchingSplit}
\alias{Log2TreesMatchingSplit}
\title{Number of trees matching a bipartition split}
\usage{
TreesMatchingSplit(A, B = A[2])

LnTreesMatchingSplit(A, B = A[2])

Log2TreesMatchingSplit(A, B = A[2])
}
\arguments{
\item{A, B}{Integer specifying the number of taxa in each partition.}
}
\value{
\code{TreesMatchingSplit()} returns a numeric specifying the number of trees
that are compatible with the given split.

\code{LnTreesMatchingSplit()} and \code{Log2TreesMatchingSplit()} give the natural
and base-2 logarithms of this number.
}
\description{
Calculates the number of unrooted bifurcated trees that are consistent with
a bipartition split that divides taxa into groups of size \code{A} and \code{B}.
}
\examples{
TreesMatchingSplit(5, 6)
LnTreesMatchingSplit(5, 6)
Log2TreesMatchingSplit(5, 6)

}
\seealso{
Other split information functions:
\code{\link[=CharacterInformation]{CharacterInformation()}},
\code{\link[=SplitInformation]{SplitInformation()}},
\code{\link[=SplitMatchProbability]{SplitMatchProbability()}},
\code{\link[=UnrootedTreesMatchingSplit]{UnrootedTreesMatchingSplit()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match.R
\name{match,phylo,phylo-method}
\alias{match,phylo,phylo-method}
\alias{match,multiPhylo,phylo-method}
\alias{match,phylo,multiPhylo-method}
\alias{match,multiPhylo,multiPhylo-method}
\alias{\%in\%,multiPhylo,multiPhylo-method}
\alias{\%in\%,multiPhylo,phylo-method}
\alias{\%in\%,phylo,multiPhylo-method}
\alias{\%in\%,phylo,phylo-method}
\title{Tree matching}
\usage{
\S4method{match}{phylo,phylo}(x, table, nomatch = NA_integer_, incomparables = NULL)

\S4method{match}{multiPhylo,phylo}(x, table, nomatch = NA_integer_, incomparables = NULL)

\S4method{match}{phylo,multiPhylo}(x, table, nomatch = NA_integer_, incomparables = NULL)

\S4method{match}{multiPhylo,multiPhylo}(x, table, nomatch = NA_integer_, incomparables = NULL)

\S4method{\%in\%}{multiPhylo,multiPhylo}(x, table)

\S4method{\%in\%}{multiPhylo,phylo}(x, table)

\S4method{\%in\%}{phylo,multiPhylo}(x, table)

\S4method{\%in\%}{phylo,phylo}(x, table)
}
\arguments{
\item{x, table}{Object of class \code{phylo} or \code{multiPhylo}.}

\item{nomatch}{Integer value that will be used in place of \code{NA} in the case
where no match is found.}

\item{incomparables}{Ignored. (Included for consistency with generic.)}
}
\value{
\code{match()} returns an integer vector specifying the position in
\code{table} that matches each element in \code{x}, or \code{nomatch} if no match is found.
}
\description{
\code{match()} returns a vector of the positions of (first) matches of trees in
its first argument in its second.
\code{\%in\%} is a more intuitive interface as a binary operator, which returns
a logical vector indicating whether there is a match or not for each
tree in its left operand.
}
\examples{
tree1 <- BalancedTree(7)
trees <- c(PectinateTree(7), BalancedTree(7))

match(tree1, trees)
}
\seealso{
Corresponding base functions are documented in
\code{\link[base:match]{match()}}.

Other utility functions: 
\code{\link{ClusterTable}},
\code{\link{ClusterTable-methods}},
\code{\link{Hamming}()},
\code{\link{MSTEdges}()},
\code{\link{SampleOne}()},
\code{\link{TipTimedTree}()},
\code{\link{UnshiftTree}()},
\code{\link{as.multiPhylo}()},
\code{\link{sapply64}()},
\code{\link{sort.multiPhylo}()}
}
\concept{utility functions}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Support.R
\name{SupportColour}
\alias{SupportColour}
\alias{SupportColor}
\title{Colour for node support value}
\usage{
SupportColour(
  support,
  show1 = TRUE,
  scale = colorRampPalette(c("#D33F6A", "#e2e2e2", "#4A6FE3"))(101),
  outOfRange = "red"
)

SupportColor(
  support,
  show1 = TRUE,
  scale = colorRampPalette(c("#D33F6A", "#e2e2e2", "#4A6FE3"))(101),
  outOfRange = "red"
)
}
\arguments{
\item{support}{A numeric vector of values in the range 0--1.}

\item{show1}{Logical specifying whether to display values of 1.
A transparent white will be returned if \code{FALSE}.}

\item{scale}{101-element vector listing colours in sequence. Defaults to
a diverging \acronym{HCL} scale.}

\item{outOfRange}{Colour to use if results are outside the range 0--1.}
}
\value{
\code{SupportColour()} returns the appropriate value from \code{scale},
or \code{outOfRange} if a value is outwith the valid range.
}
\description{
Colour value with which to display node support.
}
\examples{
SupportColour((-1):4 / 4, show1 = FALSE)

# An example forest of 100 trees, some identical
forest <- as.phylo(c(1, rep(10, 79), rep(100, 15), rep(1000, 5)), nTip = 9)

# Generate an 80\% consensus tree
cons <- ape::consensus(forest, p = 0.8)
plot(cons)

# Calculate split frequencies
splitFreqs <- SplitFrequency(cons, forest)

# Optionally, colour edges by corresponding frequency.
# Note that not all edges are associated with a unique split
# (and two root edges may be associated with one split - not handled here)
edgeSupport <- rep(1, nrow(cons$edge)) # Initialize trivial splits to 1
childNode <- cons$edge[, 2]
edgeSupport[match(names(splitFreqs), childNode)] <- splitFreqs / 100

plot(cons, edge.col = SupportColour(edgeSupport), edge.width = 3)

# Annotate nodes by frequency 
LabelSplits(cons, splitFreqs, unit = "\%",
            col = SupportColor(splitFreqs / 100),
            frame = "none", pos = 3L)

}
\seealso{
Use in conjunction with \code{\link[=LabelSplits]{LabelSplits()}} to colour split labels,
possibly calculated using \code{\link[=SplitFrequency]{SplitFrequency()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_properties.R
\name{NodeOrder}
\alias{NodeOrder}
\title{Number of edges incident to each node in a tree}
\usage{
NodeOrder(x, includeAncestor = TRUE, internalOnly = FALSE)
}
\arguments{
\item{x}{A tree of class \code{phylo}, its \verb{$edge} property, or a list thereof.}

\item{includeAncestor}{Logical specifying whether to count edge leading to
ancestral node in calculation of order.}

\item{internalOnly}{Logical specifying whether to restrict to results
to internal nodes, i.e. to omit leaves. Irrelevant if
\code{includeAncestor = FALSE}.}
}
\value{
\code{NodeOrder()} returns an integer listing the order of each node;
entries are named with the number of each node.
}
\description{
\code{NodeOrder()} calculates the order of each node: the number of edges
incident to it in a tree.
This value includes the root edge in rooted trees.
}
\examples{
tree <- CollapseNode(BalancedTree(8), 12:15)
NodeOrder(tree)
plot(tree)
nodelabels(NodeOrder(tree, internalOnly = TRUE))

}
\seealso{
Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=PaintTree]{PaintTree()}},
\code{\link[=RootNode]{RootNode()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/LongBranchScore.R
\name{LongBranch}
\alias{LongBranch}
\title{Identify taxa with long branches}
\usage{
LongBranch(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}, or a list of trees
of class \code{list} or \code{multiPhylo}.}
}
\value{
\code{LongBranch()} returns a vector giving the long branch score for
each leaf in \code{tree}, or a list of such vectors if \code{tree} is a list.
Results are given as raw deviations, without multiplying by 100 as proposed
by \insertCite{Struck2014;textual}{TreeTools}.
}
\description{
The long branch (\acronym{LB}) score \insertCite{Struck2014}{TreeTools}
measures the deviation of the average pairwise patristic distance of a leaf
from all other leaves in a tree, relative to the average leaf-to-leaf
distance.
}
\details{
\insertCite{Struck2014;textual}{TreeTools} proposes the standard deviation
of LB scores as a measure of heterogeneity that can be compared between
trees; and the upper quartile of LB scores as "a representative value for
the taxa with the longest branches".
}
\examples{
tree <- BalancedTree(8, lengths = c(rep(2, 4), 5:7, rep(2, 4), rep(1, 3)))
lb <- LongBranch(tree)
tree$tip.label <- paste(tree$tip.label, signif(lb, 3), sep = ": ")
plot(tree, tip.col = SupportColour((1 - lb) / 2), font = 2)

# Standard deviation of LB scores allows comparison with other trees
sd(lb)
evenLengths <- BalancedTree(8, lengths = jitter(rep(1, 14)))
sd(LongBranch(evenLengths))

# Upper quartile identifies taxa with longest branches
threshold <- quantile(lb, 0.75)
tree$tip.label[lb > threshold]
}
\seealso{
Other tree properties:
\code{\link[=Cherries]{Cherries()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=EdgeRatio]{EdgeRatio()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=PathLengths]{PathLengths()}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TreeIsRooted]{TreeIsRooted()}},
\code{\link[=Treeness]{Treeness()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/EdgeRatio.R
\name{EdgeRatio}
\alias{EdgeRatio}
\alias{EdgeRatio.phylo}
\title{Ratio of external:internal edge length}
\usage{
EdgeRatio(x)

\method{EdgeRatio}{phylo}(x)
}
\arguments{
\item{x}{A tree of class \code{\link[ape:read.tree]{phylo}}.}
}
\value{
\code{EdgeRatio()} returns a numeric specifying the ratio of external
to internal edge length (> 1 means the length of a tree is predominantly
in external edges), with attributes \code{external}, \code{internal}, and \code{total}
specifying the total length associated with edges of that nature.
}
\description{
Reports the ratio of tree length associated with external edges (i.e.
edges whose child is a leaf) and internal edges.
Where tree length is dominated by internal edges, variation between tips
is dominantly controlled by phylogenetic history.
}
\seealso{
Other tree properties:
\code{\link[=Cherries]{Cherries()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=LongBranch]{LongBranch()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=PathLengths]{PathLengths()}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TreeIsRooted]{TreeIsRooted()}},
\code{\link[=Treeness]{Treeness()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_descendants.R
\name{NDescendants}
\alias{NDescendants}
\title{Count descendants for each node in a tree}
\usage{
NDescendants(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}
}
\value{
\code{NDescendants()} returns an integer listing the number of direct
descendants (leaves or internal nodes) for each node in a tree.
}
\description{
\code{NDescendants()} counts the number of nodes (including leaves) directly
descended from each node in a tree.
}
\examples{
tree <- CollapseNode(BalancedTree(8), 12:15)
NDescendants(tree)
plot(tree)
nodelabels(NDescendants(tree))

}
\seealso{
\code{\link[=CladeSizes]{CladeSizes()}} for the count of descendant leaves (rather than direct
children) at each node.

Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=PaintTree]{PaintTree()}},
\code{\link[=RootNode]{RootNode()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Support.R
\name{SplitFrequency}
\alias{SplitFrequency}
\title{Frequency of splits}
\usage{
SplitFrequency(reference, forest = NULL)
}
\arguments{
\item{reference}{A tree of class \code{phylo}, a \code{Splits} object. If \code{NULL},
the frequencies of all splits in \code{forest} will be returned.}

\item{forest}{A list of trees of class \code{phylo}, or a \code{multiPhylo} object;
or a \code{Splits} object. See
\href{https://ms609.github.io/TreeTools/articles/load-trees.html}{vignette} for
possible methods of loading trees into R.}
}
\value{
\code{SplitFrequency()} returns the number of trees in \code{forest} that
contain each split in \code{reference}.
If \code{reference} is a tree of class \code{phylo}, then the sequence will correspond
to the order of nodes (use \code{ape::nodelabels()} to view).
Note that the three nodes at the root of the tree correspond to a single
split; see the example for how these might be plotted on a tree.

If \code{reference} is \code{NULL}, then \code{SplitFrequency()} returns a list of splits
(in the order encountered in forest) with attribute \code{"count"} stating the
number of times each split occurs in \code{forest}.
}
\description{
\code{SplitFrequency()} provides a simple way to count the number of times that
bipartition splits, as defined by a reference tree, occur in a forest of
trees. May be used to calculate edge ("node") support for majority consensus
or bootstrap trees.
}
\details{
If multiple calculations are required, some time can be saved by using the
constituent functions (see examples).
}
\examples{
# An example forest of 100 trees, some identical
forest <- as.phylo(c(1, rep(10, 79), rep(100, 15), rep(1000, 5)), nTip = 9)

# Generate an 80\% consensus tree
cons <- ape::consensus(forest, p = 0.8)
plot(cons)

# Calculate split frequencies
splitFreqs <- SplitFrequency(cons, forest)

# Optionally, colour edges by corresponding frequency.
# Note that not all edges are associated with a unique split
# (and two root edges may be associated with one split - not handled here)
edgeSupport <- rep(1, nrow(cons$edge)) # Initialize trivial splits to 1
childNode <- cons$edge[, 2]
edgeSupport[match(names(splitFreqs), childNode)] <- splitFreqs / 100

plot(cons, edge.col = SupportColour(edgeSupport), edge.width = 3)

# Annotate nodes by frequency 
LabelSplits(cons, splitFreqs, unit = "\%",
            col = SupportColor(splitFreqs / 100),
            frame = "none", pos = 3L)

}
\seealso{
Other Splits operations: 
\code{\link{LabelSplits}()},
\code{\link{NSplits}()},
\code{\link{NTip}()},
\code{\link{PolarizeSplits}()},
\code{\link{Splits}},
\code{\link{SplitsInBinaryTree}()},
\code{\link{TipLabels}()},
\code{\link{TipsInSplits}()},
\code{\link{match,Splits,Splits-method}},
\code{\link{xor}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{Splits operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Splits.R
\name{PolarizeSplits}
\alias{PolarizeSplits}
\title{Polarize splits on a single taxon}
\usage{
PolarizeSplits(x, pole = 1L)
}
\arguments{
\item{x}{Object that can be coerced into class \code{\link{Splits}}.}

\item{pole}{Numeric, character or logical vector identifying tip that will
polarize each split.}
}
\value{
\code{PolarizeSplits()} returns a \code{Splits} object in which \code{pole} is
represented by a zero bit
}
\description{
Polarize splits on a single taxon
}
\seealso{
Other Splits operations: 
\code{\link{LabelSplits}()},
\code{\link{NSplits}()},
\code{\link{NTip}()},
\code{\link{SplitFrequency}()},
\code{\link{Splits}},
\code{\link{SplitsInBinaryTree}()},
\code{\link{TipLabels}()},
\code{\link{TipsInSplits}()},
\code{\link{match,Splits,Splits-method}},
\code{\link{xor}()}
}
\concept{Splits operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_write.R
\name{as.Newick}
\alias{as.Newick}
\alias{as.Newick.phylo}
\alias{as.Newick.list}
\alias{as.Newick.multiPhylo}
\title{Write a phylogenetic tree in Newick format}
\usage{
as.Newick(x)

\method{as.Newick}{phylo}(x)

\method{as.Newick}{list}(x)

\method{as.Newick}{multiPhylo}(x)
}
\arguments{
\item{x}{Object to convert to Newick format.
See Usage section for supported classes.}
}
\value{
\code{as.Newick()} returns a character string representing \code{tree} in Newick
format.
}
\description{
\code{as.Newick()} creates a character string representation of a phylogenetic
tree, in the Newick format, using R's internal tip numbering.
Use \code{\link[=RenumberTips]{RenumberTips()}} to ensure that the internal numbering follows the
order you expect.
}
\examples{
trees <- list(BalancedTree(1:8), PectinateTree(8:1))
trees <- lapply(trees, RenumberTips, 1:8)
as.Newick(trees)

}
\seealso{
\itemize{
\item Retain leaf labels: \code{\link[=NewickTree]{NewickTree()}}
\item Change R's internal numbering of leaves: \code{\link[=RenumberTips]{RenumberTips()}}
\item Write tree to text or file: \code{\link[ape:write.tree]{ape::write.tree()}}
}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TopologyOnly.R
\name{TopologyOnly}
\alias{TopologyOnly}
\title{Remove metadata from trees}
\usage{
TopologyOnly(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}
}
\value{
Returns \code{tree}, with each tree in \code{\link{Preorder}}, with edge lengths,
node labels and other attributes removed.
}
\description{
\code{TopologyOnly()} removes all information from trees except for their
topologies and leaf labels.
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AddTip.R
\name{AddTip}
\alias{AddTip}
\alias{AddTipEverywhere}
\title{Add a tip to a phylogenetic tree}
\usage{
AddTip(
  tree,
  where = sample.int(tree[["Nnode"]] * 2 + 2L, size = 1) - 1L,
  label = "New tip",
  nodeLabel = "",
  edgeLength = 0,
  lengthBelow = NULL,
  nTip = NTip(tree),
  nNode = tree[["Nnode"]],
  rootNode = RootNode(tree)
)

AddTipEverywhere(tree, label = "New tip", includeRoot = FALSE)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{where}{The node or tip that should form the sister taxon to the new
node.  To add a new tip at the root, use \code{where = 0}.  By default, the
new tip is added to a random edge.}

\item{label}{Character string providing the label to apply to the new tip.}

\item{nodeLabel}{Character string providing a label to apply to the newly
created node, if \code{tree$node.label} is specified.}

\item{edgeLength}{Numeric specifying length of new edge. If \code{NULL},
defaults to \code{lengthBelow}.
This will become the default behaviour in a
future release; please manually specify the desired behaviour in your code.}

\item{lengthBelow}{Numeric specifying length below neighbour at which to
graft new edge. Values greater than the length of the edge will result
in negative edge lengths. If \code{NULL}, the default, the new tip will be added
at the midpoint of the broken edge. If inserting at the root (\code{where = 0}),
a new edge of length \code{lengthBelow} will be inserted.
If \code{NA}, the new leaf will be attached adjacent to \code{where}; at internal
nodes, this will result in polytomy.}

\item{nTip, nNode, rootNode}{Optional integer vectors specifying number of tips
and nodes in \code{tree}, and index of root node.
Not checked for correctness: specifying values here yields a marginal speed
increase at the cost of code safety.}

\item{includeRoot}{Logical; if \code{TRUE}, each position adjacent
to the root edge is considered to represent distinct edges; if \code{FALSE},
they are treated as a single edge.}
}
\value{
\code{AddTip()} returns a tree of class \code{phylo} with an additional tip
at the desired location.

\code{AddTipEverywhere()} returns a list of class \code{multiPhylo} containing
the trees produced by adding \code{label} to each edge of \code{tree} in turn.
}
\description{
\code{AddTip()} adds a tip to a phylogenetic tree at a specified location.
}
\details{
\code{AddTip()} extends \code{\link[ape]{bind.tree}}, which cannot handle
single-taxon trees.

\code{AddTipEverywhere()} adds a tip to each edge in turn.
}
\examples{
tree <- BalancedTree(10)

# Add a leaf below an internal node
plot(tree)
ape::nodelabels() # Identify node numbers 
node <- 15        # Select location to add leaf
ape::nodelabels(bg = ifelse(NodeNumbers(tree) == node, "green", "grey"))

plot(AddTip(tree, 15, "NEW_TIP"))

# Add edge lengths for an ultrametric tree
tree$edge.length <- rep(c(rep(1, 5), 2, 1, 2, 2), 2)

# Add a leaf to an external edge
leaf <- 5
plot(tree)
ape::tiplabels(bg = ifelse(seq_len(NTip(tree)) == leaf, "green", "grey"))

plot(AddTip(tree, 5, "NEW_TIP", edgeLength = NULL))

# Create a polytomy, rather than a new node
plot(AddTip(tree, 5, "NEW_TIP", edgeLength = NA))

# Set up multi-panel plot
oldPar <- par(mfrow = c(2, 4), mar = rep(0.3, 4), cex = 0.9)

# Add leaf to each edge on a tree in turn
backbone <- BalancedTree(4)
# Treating the position of the root as instructive:
additions <- AddTipEverywhere(backbone, includeRoot = TRUE)
xx <- lapply(additions, plot)

par(mfrow = c(2, 3))
# Don't treat root edges as distinct:
additions <- AddTipEverywhere(backbone, includeRoot = FALSE)
xx <- lapply(additions, plot)

# Restore original plotting parameters
par(oldPar)

}
\seealso{
Add one tree to another: \code{\link[ape]{bind.tree}()}

Other tree manipulation: 
\code{\link{CollapseNode}()},
\code{\link{ConsensusWithout}()},
\code{\link{DropTip}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{MakeTreeBinary}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}
\keyword{tree}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phylo.R
\name{ListAncestors}
\alias{ListAncestors}
\alias{AllAncestors}
\title{List ancestors}
\usage{
ListAncestors(parent, child, node = NULL)

AllAncestors(parent, child)
}
\arguments{
\item{parent}{Integer vector corresponding to the first column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 1]}}

\item{child}{Integer vector corresponding to the second column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 2]}.}

\item{node}{Integer giving the index of the node or tip whose ancestors are
required, or \code{NULL} to return ancestors of all nodes.}
}
\value{
If \code{node = NULL}, \code{ListAncestors()} returns a list. Each entry \emph{i} contains
a vector containing, in order, the nodes encountered when traversing the tree
from node \emph{i} to the root node.
The last entry of each member of the list is therefore the root node,
with the exception of the entry for the root node itself, which is a
zero-length integer.

If \code{node} is an integer, \code{ListAncestors()} returns a vector of the numbers of
the nodes ancestral to the given \code{node}, including the root node.
}
\description{
\code{ListAncestors()} reports all ancestors of a given node.
}
\details{
Note that if \code{node = NULL}, the tree's edges must be listed such that each
internal node (except the root) is listed as a child before it is listed
as a parent, i.e. its index in \code{child} is less than its index in \code{parent}.
This will be true of trees listed in \link{Preorder}.
}
\section{Functions}{
\itemize{
\item \code{AllAncestors()}: Alias for \code{ListAncestors(node = NULL)}.

}}
\examples{
tree <- PectinateTree(5)
edge <- tree[["edge"]]

# Identify desired node with:
plot(tree)
nodelabels()
tiplabels()

# Ancestors of specific nodes:
ListAncestors(edge[, 1], edge[, 2], 4L)
ListAncestors(edge[, 1], edge[, 2], 8L)

# Ancestors of each node, if tree numbering system is uncertain:
lapply(seq_len(max(edge)), ListAncestors,
       parent = edge[, 1], child = edge[, 2])

# Ancestors of each node, if tree is in preorder:
ListAncestors(edge[, 1], edge[, 2])

# Alias:
AllAncestors(edge[, 1], edge[, 2])

}
\seealso{
Implemented less efficiently in \code{phangorn:::Ancestors}, on which this
code is based.

Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=PaintTree]{PaintTree()}},
\code{\link[=RootNode]{RootNode()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KeptVerts.R
\name{KeptVerts}
\alias{KeptVerts}
\alias{KeptVerts.phylo}
\alias{KeptVerts.numeric}
\title{Identify vertices retained when leaves are dropped}
\usage{
KeptVerts(tree, keptTips, tipLabels = TipLabels(tree))

\method{KeptVerts}{phylo}(tree, keptTips, tipLabels = TipLabels(tree))

\method{KeptVerts}{numeric}(tree, keptTips, tipLabels = TipLabels(tree))
}
\arguments{
\item{tree}{Original tree of class \code{phylo}, in \code{\link{Preorder}}.}

\item{keptTips}{Either:
\itemize{
\item a logical vector stating whether each leaf should be retained, in a
sequence corresponding to \code{tree[["tip.label"]]}; or
\item a character vector listing the leaf labels to retain; or
\item a numeric vector listing the indices of leaves to retain.
}}

\item{tipLabels}{Optional character vector naming the leaves of \code{tree},
if \code{keptTips} is not logical.  Inferred from \code{tree} if unspecified.}
}
\description{
Identify vertices retained when leaves are dropped
}
\examples{
master <- BalancedTree(12)
master <- Preorder(master) # Nodes must be listed in Preorder sequence
plot(master)
nodelabels()

allTips <- master[["tip.label"]]
keptTips <- sample(allTips, 8)
plot(KeepTip(master, keptTips))
kept <- KeptVerts(master, allTips \%in\% keptTips)

map <- which(kept)
# Node `i` in the reduced tree corresponds to node `map[i]` in the original.
}
\seealso{
Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{CollapseNode}()},
\code{\link{ConsensusWithout}()},
\code{\link{DropTip}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{MakeTreeBinary}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_descendants.R
\name{DescendantEdges}
\alias{DescendantEdges}
\alias{DescendantTips}
\title{Identify descendant edges}
\usage{
DescendantEdges(
  parent,
  child,
  edge = NULL,
  node = NULL,
  nEdge = length(parent),
  includeSelf = TRUE
)

DescendantTips(parent, child, edge = NULL, node = NULL, nEdge = length(parent))
}
\arguments{
\item{parent}{Integer vector corresponding to the first column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 1]}}

\item{child}{Integer vector corresponding to the second column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 2]}.}

\item{edge}{Integer specifying the number of the edge whose children are
required (see \code{\link[ape:nodelabels]{edgelabels}()}).}

\item{node}{Integer specifying the number(s) of nodes whose children are
required.  Specify \code{0} to return all nodes.  If \code{NULL} (the default), the
\code{edge} parameter will be used instead.}

\item{nEdge}{number of edges (calculated from \code{length(parent)} if not
supplied).}

\item{includeSelf}{Logical specifying whether to mark \code{edge} as its own
descendant.}
}
\value{
\code{DescendantEdges()} returns a logical vector stating whether each
edge in turn is the specified edge (if \code{includeSelf = TRUE})
or one of its descendants.

\code{DescendantTips()} returns a logical vector stating whether each
leaf in turn is a descendant of the specified edge.
}
\description{
\code{DescendantEdges()} efficiently identifies edges that are "descended" from
edges in a tree.

\code{DescendantTips()} efficiently identifies leaves (external nodes) that are
"descended" from edges in a tree.
}
\examples{
tree <- as.phylo(0, 6)
plot(tree)
desc <- DescendantEdges(tree$edge[, 1], tree$edge[, 2], edge = 5)
which(desc)
ape::edgelabels(bg = 3 + desc)
tips <- DescendantTips(tree$edge[, 1], tree$edge[, 2], edge = 5)
which(tips)
tiplabels(bg = 3 + tips)
}
\seealso{
Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=PaintTree]{PaintTree()}},
\code{\link[=RootNode]{RootNode()}}
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Decompose.R
\name{Decompose}
\alias{Decompose}
\title{Decompose additive (ordered) phylogenetic characters}
\usage{
Decompose(dataset, indices)
}
\arguments{
\item{dataset}{A phylogenetic data matrix of \pkg{phangorn} class \code{phyDat},
whose names correspond to the labels of any accompanying tree.}

\item{indices}{Integer or logical vector specifying indices of characters
that should be decomposed}
}
\value{
\code{Decompose()} returns a \code{phyDat} object in which the specified
ordered characters have been decomposed into binary characters.
The attribute \code{originalIndex} lists the index of the character in
\code{dataset} to which each element corresponds.
}
\description{
\code{Decompose()} decomposes additive characters into a series of binary
characters, which is mathematically equivalent when analysed under
equal weights parsimony.  (This equivalence is not exact
under implied weights or under probabilistic tree inference methods.)
}
\details{
An ordered (additive) character can be rewritten as a mathematically
equivalent hierarchy of binary neomorphic characters
\insertCite{Farris1970}{TreeTools}.
Two reasons to prefer the latter approach are:
\itemize{
\item It makes explicit the evolutionary assumptions underlying an ordered
character, whether the underlying ordering is linear, reticulate or
branched \insertCite{Mabee1989}{TreeTools}.
\item It avoids having to identify characters requiring special treatment to
phylogenetic software, which requires the maintenance of an up-to-date
log of which characters are treated as additive and which sequence their
states occur in, a step that may be overlooked by re-users of the data.
}

Careful consideration is warranted when evaluating whether a group of
related characteristics ought to be treated as ordered
\insertCite{Wilkinson1992}{TreeTools}.
On the one hand, the 'principle of indifference' states that we should treat
all transformations as equally probable (/ surprising / informative);
ordered characters fail this test, as larger changes are treated as less
probable than smaller ones.
On the other hand, ordered characters allow more opportunities for homology
of different character states, and might thus be defended under the auspices
of Hennig's Auxiliary Principle \insertCite{Wilkinson1992}{TreeTools}.

For a case study of how ordering phylogenetic characters can affect
phylogenetic outcomes in practice, see
\insertCite{Brady2024;textual}{TreeTools}.
}
\examples{
data("Lobo")

# Identify character 11 as additive
# Character 11 will be replaced with two characters
# The present codings 0, 1 and 2 will be replaced with 00, 10, and 11.
decomposed <- Decompose(Lobo.phy, 11)

NumberOfChars <- function(x) sum(attr(x, "weight"))
NumberOfChars(Lobo.phy)   # 115 characters in original
NumberOfChars(decomposed) # 116 characters in decomposed
}
\references{
\insertAllCited{}
}
\seealso{
Other phylogenetic matrix conversion functions:
\code{\link[=MatrixToPhyDat]{MatrixToPhyDat()}},
\code{\link[=NexusTokensToInteger]{NexusTokensToInteger()}},
\code{\link[=Reweight]{Reweight()}},
\code{\link[=StringToPhyDat]{StringToPhyDat()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{phylogenetic matrix conversion functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ReadMrBayes.R
\name{ReadMrBayesTrees}
\alias{ReadMrBayesTrees}
\alias{ReadMrBayes}
\alias{MrBayesTrees}
\title{Read posterior tree sample produced by MrBayes}
\usage{
ReadMrBayesTrees(filepath, n = NULL, burninFrac = NULL)

ReadMrBayes(filepath, n = NULL, burninFrac = NULL)

MrBayesTrees(filepath, n = NULL, burninFrac = NULL)
}
\arguments{
\item{filepath}{character string specifying path to \code{.nex} input file used
to initialize the MrBayes analysis,
relative to the R working directory (visible with \code{getwd()}).}

\item{n}{Integer specifying number of trees to sample from posterior.}

\item{burninFrac}{Fraction of trees to discard from each run as burn-in.
If \code{NULL} (the default), this will be read from the last \code{mcmc} or \code{mcmcp}
command in \code{filepath}.}
}
\value{
\code{ReadMrBayesTrees()} returns a 'multiPhylo' object containing
\code{n} trees sampled evenly from all runs generated by analysis of \code{filepath},
or \code{NULL} if no trees are found.
}
\description{
Read posterior trees from \href{https://nbisweden.github.io/MrBayes/}{'MrBayes'} output files, discarding burn-in
generations.
}
\details{
\code{ReadMrBayesTrees()} samples trees from the posterior distributions
computed using the Bayesian inference software \href{https://nbisweden.github.io/MrBayes/}{'MrBayes'}
}
\examples{
\dontrun{ # Download will take a few seconds
  url <- 
  "https://raw.githubusercontent.com/ms609/hyoliths/master/MrBayes/hyo.nex"
  trees <- ReadMrBayesTrees(url, n = 40)
  plot(Consensus(trees, p = 0.5))
}
}
\seealso{
Other tree import functions: 
\code{\link{ReadTntTree}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree import functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ReadTntTree.R
\name{ReadTntTree}
\alias{ReadTntTree}
\alias{TntText2Tree}
\alias{TNTText2Tree}
\title{Parse TNT Tree}
\usage{
ReadTntTree(filepath, relativePath = NULL, keepEnd = 1L, tipLabels = NULL)

TntText2Tree(treeText)

TNTText2Tree(treeText)
}
\arguments{
\item{filepath}{character string specifying path to TNT \code{.tre} file,
relative to the R working directory (visible with \code{getwd()}).}

\item{relativePath}{(discouraged) character string specifying location of the
matrix file used to generate the TNT results, relative to the current working
directory.  Taxon names will be read from this file if they are not specified
by \code{tipLabels}.}

\item{keepEnd}{(optional, default 1) integer specifying how many elements of
the file path to conserve when creating relative path (see examples).}

\item{tipLabels}{(optional) character vector specifying the names of the
taxa, in the sequence that they appear in the TNT file.  If not specified,
taxon names will be loaded from the data file linked in the first line of the
\code{.tre} file specified in \code{filepath}.}

\item{treeText}{Character string describing one or more trees,
in the parenthetical format output by TNT.}
}
\value{
\code{ReadTntTree()} returns a tree of class \code{phylo} in
\link[=TNTOrder]{TNT order},
corresponding to the tree in \code{filepath}, or NULL if no trees are found.
}
\description{
Read a tree from \href{https://www.lillo.org.ar/phylogeny/tnt/}{TNT}'s
parenthetical output.
}
\details{
\code{ReadTntTree()} imports trees generated by the parsimony analysis program
\href{https://www.lillo.org.ar/phylogeny/tnt/}{TNT} into R, including node labels
written with the \code{ttags} command.
Tree files must have been saved by TNT in parenthetical notation, using the
TNT command \verb{tsave *}.
Trees are easiest to load into R if taxa have been saved using their names
(TNT command \verb{taxname =}).  In this case, the TNT \code{.tre} file
contains tip labels and can be parsed directly.  The downside is that the
uncompressed \code{.tre} files will have a larger file size.

\code{ReadTntTree()} can also read \code{.tre} files in which taxa have been saved
using their numbers (\verb{taxname -}).  Such files contain a hard-coded link to
the matrix file that was used to generate the trees, in the first line of the
\code{.tre} file.  This poses problems for portability: if the matrix file is
moved, or the \code{.tre} file is accessed on another computer, the taxon names
may be lost.  As such, it is important to check that the matrix file
exists in the expected location -- if it does not,
either use the \code{relativePath} argument to point to its new location, or
specify \code{tipLabels} to manually specify the tip labels.

\code{TntText2Tree()} converts text representation of a tree in TNT to an
object of class \code{phylo}.
}
\examples{
# In the examples below, TNT has read a matrix from
# "c:/TreeTools/input/dataset.nex"
# The results of an analysis were written to
# "c:/TreeTools/output/results1.tnt"
#
# results1.tnt will contain a hard-coded reference to
# "c:/TreeTools/input/dataset.nex".

# On the original machine (but not elsewhere), it would be possible to read
# this hard-coded reference from results.tnt:
# ReadTntTree("output/results1.tnt")

# These datasets are provided with the "TreeTools" package, which will
# probably not be located at c:/TreeTools on your machine:

oldWD <- getwd() # Remember the current working directory
setwd(system.file(package = "TreeTools"))

# If taxon names were saved within the file (using `taxname=` in TNT),
# then our job is easy:
ReadTntTree("extdata/output/named.tre")

# But if taxa were compressed to numbers (using `taxname-`), we need to
# look up the original matrix in order to dereference the tip names.
#
# We need to extract the relevant file path from the end of the
# hard-coded path in the original file.
#
# We are interested in the last two elements of
# c:/TreeTools/input/dataset.nex
#                2      1
#
# "." means "relative to the current directory"
ReadTntTree("extdata/output/numbered.tre", "./extdata", 2)

# If working in a lower subdirectory
setwd("./extdata/otherfolder")

# then it will be necessary to navigate up the directory path with "..":
ReadTntTree("../output/numbered.tre", "..", 2)


setwd(oldWD) # Restore original working directory

TNTText2Tree("(A (B (C (D E ))));")

}
\seealso{
Other tree import functions: 
\code{\link{ReadMrBayesTrees}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree import functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RUtreebalance.R
\name{J1Index}
\alias{J1Index}
\alias{JQIndex}
\title{Robust universal tree balance index}
\usage{
J1Index(tree, q = 1, nonRootDominance = FALSE)

JQIndex(tree, q = 1, nonRootDominance = FALSE)
}
\arguments{
\item{tree}{Either an object of class 'phylo', or a dataframe with column
names Parent, Identity and (optionally) Population.
The latter is similar to \code{tree$edge}, where tree is an object of class
'phylo'; the differences are in class (data.frame versus matrix) and column
names.
The dataframe may (but isn't required to) include a row for the root node.
If population sizes are omitted then internal nodes will be assigned
population size zero and leaves will be assigned population size one.}

\item{q}{Numeric between zero and one specifying sensitivity to type
frequencies.  If \code{q < 1}, the \ifelse{html}{\out{J<sup>q</sup>}}{\eqn{J^q}}
index - based on generalized entropy - will be returned; see
\insertCite{Lemant2022;textual}{TreeTools}, page 1223.}

\item{nonRootDominance}{Logical specifying whether to use non-root dominance
factor.}
}
\value{
\code{J1Index()} returns a numeric specifying the
\ifelse{html}{\out{J<sup>1</sup>}}{\eqn{J^1}} index of \code{tree}.
\eqn{J^1(T) = 1} for a perfectly balanced tree;
\eqn{J^1(T) = 0} for a pectinate (linear / caterpillar) tree.
}
\description{
Calculate tree balance index \ifelse{html}{\out{J<sup>1</sup>}}{\eqn{J^1}}
(when \code{nonRootDominance = FALSE}) or
\ifelse{html}{\out{J<sup>1c</sup>}}{\eqn{J^{1c}}}
(when \code{nonRootDominance = TRUE}) from \insertCite{Lemant2022}{TreeTools}.
}
\details{
If population sizes are not provided, then the function assigns size 0 to
internal nodes, and size 1 to leaves.
}
\examples{
# Using phylo object as input:
phylo_tree <- read.tree(text="((a:0.1)A:0.5,(b1:0.2,b2:0.1)B:0.2);")
J1Index(phylo_tree)
phylo_tree2 <- read.tree(text='((A, B), ((C, D), (E, F)));')
J1Index(phylo_tree2)

# Using edges lists as input:
tree1 <- data.frame(Parent = c(1, 1, 1, 1, 2, 3, 4),
                    Identity = 1:7,
                    Population = c(1, rep(5, 6)))
J1Index(tree1)
tree2 <- data.frame(Parent = c(1, 1, 1, 1, 2, 3, 4),
                    Identity = 1:7,
                    Population = c(rep(0, 4), rep(1, 3)))
J1Index(tree2)
tree3 <- data.frame(Parent = c(1, 1, 1, 1, 2, 3, 4),
                    Identity = 1:7,
                    Population = c(0, rep(1, 3), rep(0, 3)))
J1Index(tree3)
cat_tree <- data.frame(Parent = c(1, 1:14, 1:15, 15),
                       Identity = 1:31,
                       Population = c(rep(0, 15), rep(1, 16)))
J1Index(cat_tree)

# If population sizes are omitted then internal nodes are assigned population
# size zero and leaves are assigned population size one:
sym_tree1 <- data.frame(Parent = c(1, rep(1:15, each = 2)),
                       Identity = 1:31,
                       Population = c(rep(0, 15), rep(1, 16)))
# Equivalently:                        
sym_tree2 <- data.frame(Parent = c(1, rep(1:15, each = 2)),
                       Identity = 1:31)
J1Index(sym_tree1)
J1Index(sym_tree2)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree characterization functions:
\code{\link[=CladisticInfo]{CladisticInfo()}},
\code{\link[=Consensus]{Consensus()}},
\code{\link{Stemwardness}},
\code{\link[=TotalCopheneticIndex]{TotalCopheneticIndex()}}
}
\author{
Rob Noble, adapted by Martin R. Smith
}
\concept{tree characterization functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{EndSentence}
\alias{EndSentence}
\title{Add full stop to end of a sentence}
\usage{
EndSentence(string)
}
\arguments{
\item{string}{Input string}
}
\value{
\code{EndSentence()} returns \code{string}, punctuated with a final full stop
(period).`
}
\description{
Add full stop to end of a sentence
}
\examples{
EndSentence("Hello World") # "Hello World."
}
\seealso{
Other string parsing functions:
\code{\link[=MatchStrings]{MatchStrings()}},
\code{\link[=MorphoBankDecode]{MorphoBankDecode()}},
\code{\link[=RightmostCharacter]{RightmostCharacter()}},
\code{\link[=Unquote]{Unquote()}}
}
\author{
Martin R. Smith
}
\concept{string parsing functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Combinatorics.R
\name{NRooted}
\alias{NRooted}
\alias{NUnrooted}
\alias{NRooted64}
\alias{NUnrooted64}
\alias{LnUnrooted}
\alias{LnUnrooted.int}
\alias{Log2Unrooted}
\alias{Log2Unrooted.int}
\alias{LnRooted}
\alias{LnRooted.int}
\alias{Log2Rooted}
\alias{Log2Rooted.int}
\alias{LnUnrootedSplits}
\alias{Log2UnrootedSplits}
\alias{NUnrootedSplits}
\alias{LnUnrootedMult}
\alias{Log2UnrootedMult}
\alias{NUnrootedMult}
\title{Number of trees}
\usage{
NRooted(tips)

NUnrooted(tips)

NRooted64(tips)

NUnrooted64(tips)

LnUnrooted(tips)

LnUnrooted.int(tips)

Log2Unrooted(tips)

Log2Unrooted.int(tips)

LnRooted(tips)

LnRooted.int(tips)

Log2Rooted(tips)

Log2Rooted.int(tips)

LnUnrootedSplits(...)

Log2UnrootedSplits(...)

NUnrootedSplits(...)

LnUnrootedMult(...)

Log2UnrootedMult(...)

NUnrootedMult(...)
}
\arguments{
\item{tips}{Integer specifying the number of leaves.}

\item{\dots}{Integer vector, or series of integers, listing the number of
leaves in each split.}
}
\description{
These functions return the number of rooted or unrooted binary trees
consistent with a given pattern of splits.
}
\details{
Functions starting \code{N} return the number of rooted or unrooted trees.
Replace this initial \code{N} with \code{Ln} for the natural logarithm of this number;
or \code{Log2} for its base 2 logarithm.

Calculations follow \insertCite{CavalliSforza1967;textual}{TreeTools} and
\insertCite{Carter1990;textual}{TreeTools}, Theorem 2.
}
\section{Functions}{
\itemize{
\item \code{NUnrooted()}: Number of unrooted trees

\item \code{NRooted64()}: Exact number of rooted trees as 64-bit integer
(13 < \code{nTip} < 19)

\item \code{NUnrooted64()}: Exact number of unrooted trees as 64-bit integer
(14 < \code{nTip} < 20)

\item \code{LnUnrooted()}: Log Number of unrooted trees

\item \code{LnUnrooted.int()}: Log Number of unrooted trees (as integer)

\item \code{LnRooted()}: Log Number of rooted trees

\item \code{LnRooted.int()}: Log Number of rooted trees (as integer)

\item \code{NUnrootedSplits()}: Number of unrooted trees consistent with a bipartition
split.

\item \code{NUnrootedMult()}: Number of unrooted trees consistent with a multi-partition
split.

}}
\examples{
NRooted(10)
NUnrooted(10)
LnRooted(10)
LnUnrooted(10)
Log2Unrooted(10)
# Number of trees consistent with a character whose states are
# 00000 11111 222
NUnrootedMult(c(5,5,3))

NUnrooted64(18)
LnUnrootedSplits(c(2,4))
LnUnrootedSplits(3, 3)
Log2UnrootedSplits(c(2,4))
Log2UnrootedSplits(3, 3)
NUnrootedSplits(c(2,4))
NUnrootedSplits(3, 3)
}
\references{
\insertAllCited{}
}
\seealso{
Other tree information functions:
\code{\link[=CladisticInfo]{CladisticInfo()}},
\code{\link[=TreesMatchingTree]{TreesMatchingTree()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split information function}
\concept{tree information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TreeNumber.R
\name{print.TreeNumber}
\alias{print.TreeNumber}
\title{Print \code{TreeNumber} object}
\usage{
\method{print}{TreeNumber}(x, ...)
}
\arguments{
\item{x}{Object of class \code{TreeNumber}.}

\item{\dots}{Additional arguments for consistency with S3 method (unused).}
}
\description{
S3 method for objects of class \code{TreeNumber}.
}
\seealso{
Other 'TreeNumber' utilities: 
\code{\link{TreeNumber}},
\code{\link{is.TreeNumber}()}
}
\concept{'TreeNumber' utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_generation.R
\name{NJTree}
\alias{NJTree}
\title{Generate a neighbour joining tree}
\usage{
NJTree(dataset, edgeLengths = FALSE, ratio = TRUE, ambig = "mean")
}
\arguments{
\item{dataset}{A phylogenetic data matrix of \pkg{phangorn} class \code{phyDat},
whose names correspond to the labels of any accompanying tree.}

\item{edgeLengths}{Logical specifying whether to include edge lengths.}

\item{ambig, ratio}{Settings of \code{ambig} and \code{ratio} to be used when
computing \code{\link[=Hamming]{Hamming()}} distances between sequences.}
}
\value{
\code{NJTree} returns an object of class \code{phylo}.
}
\description{
\code{NJTree()} generates a rooted neighbour joining tree from a phylogenetic
dataset.
}
\examples{
data("Lobo")
NJTree(Lobo.phy)

}
\seealso{
Other tree generation functions:
\code{\link[=ConstrainedNJ]{ConstrainedNJ()}},
\code{\link{GenerateTree}},
\code{\link{TreeNumber}},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree generation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phylo.R
\name{CladeSizes}
\alias{CladeSizes}
\title{Clade sizes}
\usage{
CladeSizes(tree, internal = FALSE, nodes = NULL)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{internal}{Logical specifying whether internal nodes should be counted
towards the size of each clade.}

\item{nodes}{Integer specifying indices of nodes at the base of clades
whose sizes should be returned.
If unspecified, counts will be provided for all nodes (including leaves).}
}
\value{
\code{CladeSizes()} returns the number of nodes (including leaves) that
are descended from each node, not including the node itself.
}
\description{
\code{CladeSizes()} reports the number of nodes in each clade in a tree.
}
\examples{
tree <- BalancedTree(6)
plot(tree)
ape::nodelabels()
CladeSizes(tree, nodes = c(1, 8, 9))

}
\seealso{
Other tree navigation: 
\code{\link{AncestorEdge}()},
\code{\link{DescendantEdges}()},
\code{\link{EdgeAncestry}()},
\code{\link{EdgeDistances}()},
\code{\link{ListAncestors}()},
\code{\link{MRCA}()},
\code{\link{MatchEdges}()},
\code{\link{NDescendants}()},
\code{\link{NodeDepth}()},
\code{\link{NodeNumbers}()},
\code{\link{NodeOrder}()},
\code{\link{RootNode}()}
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SplitFunctions.R
\name{TrivialSplits}
\alias{TrivialSplits}
\alias{WithoutTrivialSplits}
\title{Identify and remove trivial splits}
\usage{
TrivialSplits(splits, nTip = attr(splits, "nTip"))

WithoutTrivialSplits(splits, nTip = attr(splits, "nTip"))
}
\arguments{
\item{splits}{An object of class \code{\link[=as.Splits]{Splits}}.}

\item{nTip}{Integer specifying number of tips (leaves).}
}
\value{
\code{TrivialSplits()} returns a logical vector specifying whether each
split in \code{splits} is trivial, i.e. includes or excludes only a single tip or
no tips at all.

\code{WithoutTrivialSplits()} returns a \code{Splits} object with trivial
splits removed.
}
\description{
\code{TrivialSplits()} identifies trivial splits (which separate one or zero
leaves from all others); \code{WithoutTrivialSplits()} removes them from a
\code{Splits} object.
}
\examples{
splits <- as.Splits(PectinateTree(letters[1:9]))
efgh <- Subsplit(splits, tips = letters[5:8], keepAll = TRUE)
summary(efgh)

TrivialSplits(efgh)
summary(WithoutTrivialSplits(efgh))
}
\seealso{
Other split manipulation functions: 
\code{\link{DropTip}()},
\code{\link{SplitConsistent}()},
\code{\link{Subsplit}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split manipulation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mst.R
\name{MSTEdges}
\alias{MSTEdges}
\alias{MSTLength}
\title{Minimum spanning tree}
\usage{
MSTEdges(distances, plot = FALSE, x = NULL, y = NULL, ...)

MSTLength(distances, mst = NULL)
}
\arguments{
\item{distances}{Either a matrix that can be interpreted as a distance
matrix, or an object of class \code{dist}.}

\item{plot}{Logical specifying whether to add the minimum spanning tree
to an existing plot.}

\item{x, y}{Numeric vectors specifying the X and Y coordinates of each
element in \code{distances}.  Necessary only if \code{plot = TRUE}.}

\item{\dots}{Additional parameters to send to \verb{[lines()]}.}

\item{mst}{Optional parameter specifying the minimum spanning tree in the
format returned by \code{MSTEdges()}; if \code{NULL}, calculated from \code{distances}.}
}
\value{
\code{MSTEdges()} returns a matrix in which each row corresponds to an
edge of the minimum spanning tree, listed in non-decreasing order of length.
The two columns contain the indices of the entries in \code{distances} that
each edge connects, with the lower value listed first.

\code{MSTLength()} returns the length of the minimum spanning tree.
}
\description{
Calculate or plot the minimum spanning tree \insertCite{Gower1969}{TreeTools}
of a distance matrix.
}
\examples{
# Corners of an almost-regular octahedron
points <- matrix(c(0, 0, 2, 2, 1.1, 1,
                   0, 2, 0, 2, 1, 1.1,
                   0, 0, 0, 0, 1, -1), 6)
distances <- dist(points)
mst <- MSTEdges(distances)
MSTLength(distances, mst)
plot(points[, 1:2], ann = FALSE, asp = 1)
MSTEdges(distances, TRUE, x = points[, 1], y = points[, 2], lwd = 2)
}
\references{
\insertAllCited{}
}
\seealso{
Slow implementation returning the association matrix of the minimum spanning
tree: \code{\link[ape:mst]{ape::mst()}}.

Other utility functions: 
\code{\link{ClusterTable}},
\code{\link{ClusterTable-methods}},
\code{\link{Hamming}()},
\code{\link{SampleOne}()},
\code{\link{TipTimedTree}()},
\code{\link{UnshiftTree}()},
\code{\link{as.multiPhylo}()},
\code{\link{match,phylo,phylo-method}},
\code{\link{sapply64}()},
\code{\link{sort.multiPhylo}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phylo.R
\name{Renumber}
\alias{Renumber}
\title{Renumber a tree's nodes and tips}
\usage{
Renumber(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}
}
\value{
\code{Renumber()} returns a tree of class \code{phylo}, numbered in a
\link{Cladewise} fashion consistent with the expectations of \pkg{ape} functions.
}
\description{
\code{Renumber()} numbers the nodes and tips in a tree to conform with the
\code{phylo} standards.
}
\details{
The \pkg{ape} class \code{phylo} is not formally defined, but expects trees' internal
representation to conform to certain principles: for example, nodes should
be numbered sequentially, with values increasing away from the root.

\code{Renumber()} attempts to reformat any tree into a representation that will
not cause \pkg{ape} functions to produce unwanted results or to crash R.
}
\examples{
tree <- RandomTree(letters[1:10])
Renumber(tree)

}
\seealso{
\code{Preorder()} provides a faster and simpler alternative, but also
rotates nodes.

Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{CollapseNode}()},
\code{\link{ConsensusWithout}()},
\code{\link{DropTip}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{MakeTreeBinary}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_properties.R
\name{NSplits}
\alias{NSplits}
\alias{NPartitions}
\alias{NSplits.phylo}
\alias{NSplits.list}
\alias{NSplits.multiPhylo}
\alias{NSplits.Splits}
\alias{NSplits.numeric}
\alias{NSplits.NULL}
\alias{NSplits.ClusterTable}
\alias{NSplits.character}
\title{Number of distinct splits}
\usage{
NSplits(x)

NPartitions(x)

\method{NSplits}{phylo}(x)

\method{NSplits}{list}(x)

\method{NSplits}{multiPhylo}(x)

\method{NSplits}{Splits}(x)

\method{NSplits}{numeric}(x)

\method{NSplits}{`NULL`}(x)

\method{NSplits}{ClusterTable}(x)

\method{NSplits}{character}(x)
}
\arguments{
\item{x}{A phylogenetic tree of class \code{phylo}; a list of such trees
(of class \code{list} or \code{multiPhylo}); a \code{Splits} object;
a vector of integers; or a character vector listing tips of a tree,
or a character of length one specifying a tree in Newick format.}
}
\value{
\code{NSplits()} returns an integer specifying the number of bipartitions in
the specified objects, or in a binary tree with \code{x} tips.
}
\description{
\code{NSplits()} counts the unique bipartition splits in a tree or object.
}
\examples{
NSplits(8L)
NSplits(PectinateTree(8))
NSplits(as.Splits(BalancedTree(8)))
}
\seealso{
Other tree properties:
\code{\link[=Cherries]{Cherries()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=EdgeRatio]{EdgeRatio()}},
\code{\link[=LongBranch]{LongBranch()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=PathLengths]{PathLengths()}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TreeIsRooted]{TreeIsRooted()}},
\code{\link[=Treeness]{Treeness()}}

Other Splits operations:
\code{\link[=LabelSplits]{LabelSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=PolarizeSplits]{PolarizeSplits()}},
\code{\link[=SplitFrequency]{SplitFrequency()}},
\code{\link{Splits}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TipsInSplits]{TipsInSplits()}},
\code{\link{match,Splits,Splits-method}},
\code{\link[=xor]{xor()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{Splits operations}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_generation.R
\name{.RandomParent}
\alias{.RandomParent}
\title{Random parent vector}
\usage{
.RandomParent(n, seed = sample.int(2147483647L, 1L))
}
\arguments{
\item{n}{Integer specifying number of leaves.}

\item{seed}{(Optional) Integer with which to seed Mersenne Twister random
number generator in C++.}
}
\value{
Integer vector corresponding to the "parent" entry of
\code{tree[["edge"]]}, where the "child" entry, i.e. column 2, is numbered
sequentially from \code{1:n}.
}
\description{
Random parent vector
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_properties.R
\name{EdgeDistances}
\alias{EdgeDistances}
\title{Distance between edges}
\usage{
EdgeDistances(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}
}
\value{
\code{EdgeDistances()} returns a symmetrical matrix listing the number
of edges that must be traversed to travel from each numbered edge to each
other.
The two edges straddling the root of a rooted tree
are treated as a single edge.  Add a "root" tip using \code{\link[=AddTip]{AddTip()}} if the
position of the root is significant.
}
\description{
Number of nodes that must be traversed to navigate from each edge to
each other edge within a tree
}
\examples{

tree <- BalancedTree(5)
plot(tree)
ape::edgelabels()

EdgeDistances(tree)

}
\seealso{
Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=PaintTree]{PaintTree()}},
\code{\link[=RootNode]{RootNode()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MatchNodes.R
\name{MatchEdges}
\alias{MatchEdges}
\alias{MatchNodes}
\title{Match nodes and edges between trees}
\usage{
MatchEdges(x, table, nomatch = NA_integer_)

MatchNodes(x, table, nomatch = NA_integer_, tips = FALSE)
}
\arguments{
\item{x}{Tree whose nodes are to be matched.}

\item{table}{Tree containing nodes to be matched against.}

\item{nomatch}{Integer value that will be used in place of \code{NA} in the case
where no match is found.}

\item{tips}{Logical specifying whether to return matches for tips;
unless \code{TRUE}, only the matches for internal nodes will be returned.}
}
\description{
\code{MatchNodes()} and \code{MatchEdges()} matches nodes or edges in one tree to
entries in the second that denote a clade with identical tip labels.
}
\details{
The current implementation is potentially inefficient.
Please contact the maintainer to request a more efficient implementation if
this function is proving a bottleneck.
}
\examples{
MatchNodes(BalancedTree(8), RootTree(BalancedTree(8)))
}
\seealso{
Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=PaintTree]{PaintTree()}},
\code{\link[=RootNode]{RootNode()}}

Other tree properties:
\code{\link[=Cherries]{Cherries()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=EdgeRatio]{EdgeRatio()}},
\code{\link[=LongBranch]{LongBranch()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=PathLengths]{PathLengths()}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TreeIsRooted]{TreeIsRooted()}},
\code{\link[=Treeness]{Treeness()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_properties.R
\name{TreeIsRooted}
\alias{TreeIsRooted}
\title{Is tree rooted?}
\usage{
TreeIsRooted(tree)
}
\arguments{
\item{tree}{A phylogenetic tree of class \code{phylo}.}
}
\value{
\code{TreeIsRooted()} returns a logical specifying whether a root node is
resolved.
}
\description{
\code{TreeIsRooted()} is a fast alternative to \code{ape::is.rooted()}.
}
\examples{
TreeIsRooted(BalancedTree(6))
TreeIsRooted(UnrootTree(BalancedTree(6)))

}
\seealso{
Other tree properties: 
\code{\link{Cherries}()},
\code{\link{ConsensusWithout}()},
\code{\link{EdgeRatio}()},
\code{\link{LongBranch}()},
\code{\link{MatchEdges}()},
\code{\link{NSplits}()},
\code{\link{NTip}()},
\code{\link{NodeNumbers}()},
\code{\link{PathLengths}()},
\code{\link{SplitsInBinaryTree}()},
\code{\link{TipLabels}()},
\code{\link{Treeness}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{sapply64}
\alias{sapply64}
\alias{vapply64}
\alias{replicate64}
\title{Apply a function that returns 64-bit integers over a list or vector}
\usage{
sapply64(X, FUN, ..., simplify = TRUE, USE.NAMES = TRUE)

vapply64(X, FUN, FUN.LEN = 1, ...)

replicate64(n, expr, simplify = "array")
}
\arguments{
\item{X}{a vector (atomic or list) or an \code{\link[base]{expression}}
    object.  Other objects (including classed objects) will be coerced
    by \code{base::\link[base]{as.list}}.}

\item{FUN}{the function to be applied to each element of \code{X}:
    see \sQuote{Details}.  In the case of functions like
    \code{+}, \code{\%*\%}, the function name must be backquoted or quoted.}

\item{...}{optional arguments to \code{FUN}.}

\item{simplify}{logical or character string; should the result be
    simplified to a vector, matrix or higher dimensional array if
    possible?  For \code{sapply} it must be named and not abbreviated.
    The default value, \code{TRUE}, returns a vector or matrix if appropriate,
    whereas if \code{simplify = "array"} the result may be an
    \code{\link[base]{array}} of \dQuote{rank}
    (\eqn{=}\code{length(dim(.))}) one higher than the result
    of \code{FUN(X[[i]])}.}

\item{USE.NAMES}{logical; if \code{TRUE} and if \code{X} is character,
    use \code{X} as \code{\link[base]{names}} for the result unless it had names
    already.  Since this argument follows \code{\dots} its name cannot
    be abbreviated.}

\item{FUN.LEN}{Integer specifying the length of the output of \code{FUN}.}

\item{n}{integer: the number of replications.}

\item{expr}{the expression (a \link[base]{language object}, usually a call)
    to evaluate repeatedly.}
}
\description{
Wrappers for members of the \code{\link[=lapply]{lapply()}} family intended for use when a
function \code{FUN} returns a vector of \code{integer64} objects.
\code{vapply()}, \code{sapply()} or \code{replicate()} drop the \code{integer64} class,
resulting in a vector of numerics that require conversion back to
64-bit integers.  These functions restore the missing \code{class} attribute.
}
\details{
For details of the underlying functions, see \code{\link[base:lapply]{base::lapply()}}.
}
\examples{
sapply64(as.phylo(1:6, 6), as.TreeNumber)
vapply64(as.phylo(1:6, 6), as.TreeNumber, 1)
set.seed(0)
replicate64(6, as.TreeNumber(RandomTree(6)))
}
\seealso{
\code{\link[bit64]{integer64}()}

Other utility functions:
\code{\link{ClusterTable}},
\code{\link{ClusterTable-methods}},
\code{\link[=Hamming]{Hamming()}},
\code{\link[=MSTEdges]{MSTEdges()}},
\code{\link[=SampleOne]{SampleOne()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link[=UnshiftTree]{UnshiftTree()}},
\code{\link[=as.multiPhylo]{as.multiPhylo()}},
\code{\link{match,phylo,phylo-method}},
\code{\link[=sort.multiPhylo]{sort.multiPhylo()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Support.R
\name{LabelSplits}
\alias{LabelSplits}
\title{Label splits}
\usage{
LabelSplits(tree, labels = NULL, unit = "", ...)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{labels}{Named vector listing annotations for each split. Names
should correspond to the node associated with each split; see
\code{\link[=as.Splits]{as.Splits()}} for details.
If \code{NULL}, each splits will be labelled with its associated node.}

\item{unit}{Character specifying units of \code{labels}, if desired. Include a
leading space if necessary.}

\item{\dots}{Additional parameters to \code{\link[ape:nodelabels]{ape::edgelabels()}}.}
}
\value{
\code{LabelSplits()} returns \code{invisible()}, after plotting \code{labels} on
each relevant edge of a plot (which should already have been produced using
\code{plot(tree)}).
}
\description{
Labels the edges associated with each split on a plotted tree.
}
\details{
As the two root edges of a rooted tree denote the same split, only the
rightmost (plotted at the bottom, by default) edge will be labelled.
If the position of the root is significant, add a tip at the root using
\code{\link[=AddTip]{AddTip()}}.
}
\examples{
tree <- BalancedTree(LETTERS[1:5])
splits <- as.Splits(tree)
plot(tree)
LabelSplits(tree, as.character(splits), frame = "none", pos = 3L)
LabelSplits(tree, TipsInSplits(splits), unit = " tips", frame = "none",
            pos = 1L)

# An example forest of 100 trees, some identical
forest <- as.phylo(c(1, rep(10, 79), rep(100, 15), rep(1000, 5)), nTip = 9)

# Generate an 80\% consensus tree
cons <- ape::consensus(forest, p = 0.8)
plot(cons)

# Calculate split frequencies
splitFreqs <- SplitFrequency(cons, forest)

# Optionally, colour edges by corresponding frequency.
# Note that not all edges are associated with a unique split
# (and two root edges may be associated with one split - not handled here)
edgeSupport <- rep(1, nrow(cons$edge)) # Initialize trivial splits to 1
childNode <- cons$edge[, 2]
edgeSupport[match(names(splitFreqs), childNode)] <- splitFreqs / 100

plot(cons, edge.col = SupportColour(edgeSupport), edge.width = 3)

# Annotate nodes by frequency 
LabelSplits(cons, splitFreqs, unit = "\%",
            col = SupportColor(splitFreqs / 100),
            frame = "none", pos = 3L)

}
\seealso{
Calculate split support: \code{\link[=SplitFrequency]{SplitFrequency()}}

Colour labels according to value: \code{\link[=SupportColour]{SupportColour()}}

Other Splits operations:
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=PolarizeSplits]{PolarizeSplits()}},
\code{\link[=SplitFrequency]{SplitFrequency()}},
\code{\link{Splits}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TipsInSplits]{TipsInSplits()}},
\code{\link{match,Splits,Splits-method}},
\code{\link[=xor]{xor()}}
}
\concept{Splits operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{nRootedShapes}
\alias{nRootedShapes}
\alias{nUnrootedShapes}
\title{Number of rooted / unrooted tree shapes}
\format{
An object of class \code{integer64} of length 55.

An object of class \code{integer64} of length 60.
}
\source{
\code{nRootedShapes} corresponds to the Wedderburn-Etherington numbers,
\href{https://oeis.org/A001190}{\acronym{OEIS} A001190}

\code{nUnrootedShapes} is \href{https://oeis.org/A000672}{\acronym{OEIS} A000672}
}
\usage{
nRootedShapes

nUnrootedShapes
}
\description{
\code{nRootedShapes} and \code{nUnrootedShapes} give the number of (un)rooted binary
trees on \emph{n} unlabelled leaves.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TipTimedTree.R
\name{TipTimedTree}
\alias{TipTimedTree}
\title{Display time-calibrated tree using tip information only}
\usage{
TipTimedTree(tree, tipAge, minEdge = 1)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{tipAge}{Numeric vector specifying the age (in units-of-time ago)
associated with each tip in \code{tree$tip.label} in turn.
Older ages signify earlier tips.}

\item{minEdge}{Minimum length of edge to allow (in units-of-time)}
}
\value{
\code{TipTimedTree()} returns a tree with edge lengths set based on the
ages of each tip.
}
\description{
\code{TipTimedTree()} plots a phylogenetic tree against time using an
\emph{ad hoc} approach based on dates associated with the leaves.
Nodes are dated to the youngest possible value, plus an additional "buffer"
(specified with \code{minEdge}) to ensure that branching order is readable.
}
\details{
This experimental function is liable to change its behaviour, or to
be deprecated, in coming releases.
Please contact the maintainer if you find it useful, so that a
production-ready version can be prioritized.
}
\examples{
tree <- BalancedTree(6)
plot(TipTimedTree(tree, tipAge = 1:6, minEdge = 2))
}
\seealso{
Other utility functions: 
\code{\link{ClusterTable}},
\code{\link{ClusterTable-methods}},
\code{\link{Hamming}()},
\code{\link{MSTEdges}()},
\code{\link{SampleOne}()},
\code{\link{UnshiftTree}()},
\code{\link{as.multiPhylo}()},
\code{\link{match,phylo,phylo-method}},
\code{\link{sapply64}()},
\code{\link{sort.multiPhylo}()}

Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{CollapseNode}()},
\code{\link{ConsensusWithout}()},
\code{\link{DropTip}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{MakeTreeBinary}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TrivialTree}}
}
\concept{tree manipulation}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ClusterTable.R
\name{ClusterTable}
\alias{ClusterTable}
\alias{as.ClusterTable}
\alias{as.ClusterTable.phylo}
\alias{as.ClusterTable.list}
\alias{as.ClusterTable.multiPhylo}
\title{Convert phylogenetic tree to \code{ClusterTable}}
\usage{
as.ClusterTable(x, tipLabels = NULL, ...)

\method{as.ClusterTable}{phylo}(x, tipLabels = NULL, ...)

\method{as.ClusterTable}{list}(x, tipLabels = NULL, ...)

\method{as.ClusterTable}{multiPhylo}(x, tipLabels = NULL, ...)
}
\arguments{
\item{x}{Object to convert into \code{ClusterTable}: perhaps a tree of class
\code{\link[ape:read.tree]{phylo}}.}

\item{tipLabels}{Character vector specifying sequence in which to order
tip labels.}

\item{\dots}{Unused.}
}
\value{
\code{as.ClusterTable()} returns an object of class \code{ClusterTable},
or a list thereof.
}
\description{
\code{as.ClusterTable()} converts a phylogenetic tree to a \code{ClusterTable} object,
which is an internal representation of its splits suitable for rapid tree
distance calculation (per Day, 1985).
}
\details{
Each row of a cluster table relates to a clade on a tree rooted on tip 1.
Tips are numbered according to the order in which they are visited in
preorder: i.e., if plotted using \code{plot(x)}, from the top of the page
downwards.  A clade containing the tips 2 .. 5 would be denoted by the
entry \verb{2, 5}, in either row 2 or row 5 of the cluster table.
}
\examples{
tree1 <- ape::read.tree(text = "(A, (B, (C, (D, E))));");
tree2 <- ape::read.tree(text = "(A, (B, (D, (C, E))));");
ct1 <- as.ClusterTable(tree1)
summary(ct1)
as.matrix(ct1)

# Tip label order must match ct1 to allow comparison
ct2 <- as.ClusterTable(tree2, tipLabels = LETTERS[1:5])

# It can thus be safer to use
ctList <- as.ClusterTable(c(tree1, tree2))
ctList[[2]]
}
\references{
\insertRef{Day1985}{TreeTools}
}
\seealso{
\link[=ClusterTable-methods]{S3 methods} for \code{ClusterTable} objects.

Other utility functions:
\code{\link{ClusterTable-methods}},
\code{\link[=Hamming]{Hamming()}},
\code{\link[=MSTEdges]{MSTEdges()}},
\code{\link[=SampleOne]{SampleOne()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link[=UnshiftTree]{UnshiftTree()}},
\code{\link[=as.multiPhylo]{as.multiPhylo()}},
\code{\link{match,phylo,phylo-method}},
\code{\link[=sapply64]{sapply64()}},
\code{\link[=sort.multiPhylo]{sort.multiPhylo()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_properties.R
\name{SplitsInBinaryTree}
\alias{SplitsInBinaryTree}
\alias{SplitsInBinaryTree.list}
\alias{SplitsInBinaryTree.multiPhylo}
\alias{SplitsInBinaryTree.numeric}
\alias{SplitsInBinaryTree.NULL}
\alias{SplitsInBinaryTree.default}
\alias{SplitsInBinaryTree.Splits}
\alias{SplitsInBinaryTree.phylo}
\title{Maximum splits in an \emph{n}-leaf tree}
\usage{
SplitsInBinaryTree(tree)

\method{SplitsInBinaryTree}{list}(tree)

\method{SplitsInBinaryTree}{multiPhylo}(tree)

\method{SplitsInBinaryTree}{numeric}(tree)

\method{SplitsInBinaryTree}{`NULL`}(tree)

\method{SplitsInBinaryTree}{default}(tree)

\method{SplitsInBinaryTree}{Splits}(tree)

\method{SplitsInBinaryTree}{phylo}(tree)
}
\arguments{
\item{tree}{An object of a supported format that represents a tree or
set of trees, from which the number of leaves will be calculated.}
}
\value{
\code{SplitsInBinaryTree()} returns an integer vector detailing the number
of unique non-trivial splits in a binary tree with \emph{n} leaves.
}
\description{
\code{SplitsInBinaryTree()} is a convenience function to calculate the number of
splits in a fully-resolved (binary) tree with \emph{n} leaves.
}
\examples{
tree <- BalancedTree(8)
SplitsInBinaryTree(tree)
SplitsInBinaryTree(as.Splits(tree))
SplitsInBinaryTree(8)
SplitsInBinaryTree(list(tree, tree))
}
\seealso{
Other tree properties:
\code{\link[=Cherries]{Cherries()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=EdgeRatio]{EdgeRatio()}},
\code{\link[=LongBranch]{LongBranch()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=PathLengths]{PathLengths()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TreeIsRooted]{TreeIsRooted()}},
\code{\link[=Treeness]{Treeness()}}

Other Splits operations:
\code{\link[=LabelSplits]{LabelSplits()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=PolarizeSplits]{PolarizeSplits()}},
\code{\link[=SplitFrequency]{SplitFrequency()}},
\code{\link{Splits}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TipsInSplits]{TipsInSplits()}},
\code{\link{match,Splits,Splits-method}},
\code{\link[=xor]{xor()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{Splits operations}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_numbering.R
\name{Neworder}
\alias{Neworder}
\alias{NeworderPruningwise}
\alias{NeworderPhylo}
\title{Reorder edges of a phylogenetic tree}
\usage{
NeworderPruningwise(nTip, nNode, parent, child, nEdge)

NeworderPhylo(nTip, parent, child, nEdge, whichwise)
}
\arguments{
\item{nTip, nNode, nEdge}{Integer specifying the number of tips, nodes
and edges in the input tree.}

\item{parent}{Integer vector corresponding to the first column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 1]}}

\item{child}{Integer vector corresponding to the second column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 2]}.}

\item{whichwise}{Integer specifying whether to order edges (1)
cladewise; or (2) in postorder.}
}
\value{
\code{NeworderPruningwise} returns an integer vector specifying the
pruningwise order of edges within a tree.

\code{NeworderPhylo} returns an integer vector specifying the order
of edges under the ordering sequence specified by \code{whichwise}.
}
\description{
Wrappers for the C functions called by
\code{ape::\link[ape:reorder.phylo]{reorder.phylo}}.
These call the C functions directly, so are faster -- but don't perform
as many checks on user input.  Bad input could crash R.
}
\examples{
nTip <- 8L
tree <- BalancedTree(nTip)
edge <- tree[["edge"]]
pruningwise <- NeworderPruningwise(nTip, tree$Nnode, edge[, 1], edge[, 2],
                                   dim(edge)[1])
cladewise <- NeworderPhylo(nTip, edge[, 1], edge[, 2], dim(edge)[1], 1L)
postorder <- NeworderPhylo(nTip, edge[, 1], edge[, 2], dim(edge)[1], 2L)

tree[["edge"]] <- tree[["edge"]][pruningwise, ]

}
\seealso{
Other C wrappers: 
\code{\link{RenumberTree}()}
}
\author{
\itemize{
\item C algorithm: Emmanuel Paradis
\item R wrapper: Martin R. Smith
}
}
\concept{C wrappers}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RoguePlot.R
\name{RoguePlot}
\alias{RoguePlot}
\title{Visualize position of rogue taxa}
\usage{
RoguePlot(
  trees,
  tip,
  p = 1,
  plot = TRUE,
  Palette = colorRampPalette(c(par("fg"), "#009E73"), space = "Lab"),
  nullCol = rgb(colorRamp(unlist(par(c("fg", "bg"))), space = "Lab")(0.8)/255),
  edgeLength = NULL,
  thin = par("lwd"),
  fat = thin + 1L,
  outgroupTips,
  sort = FALSE,
  legend = "none",
  legend.inset = 0,
  ...
)
}
\arguments{
\item{trees}{List or \code{multiPhylo} object containing phylogenetic trees
of class \code{phylo} to be summarized.}

\item{tip}{Numeric or character identifying rogue leaf, in format accepted
by \code{\link[=DropTip]{DropTip()}}.}

\item{p}{A numeric value between 0.5 and 1 giving the proportion for a clade
to be represented in the consensus tree (see \code{\link[=Consensus]{Consensus()}}).}

\item{plot}{Logical specifying whether to plot the tree.}

\item{Palette}{Function that takes a parameter \code{n} and generates a colour
palette with \code{n} entries.}

\item{nullCol}{Colour to paint regions of the tree on which the rogue is
never found.}

\item{edgeLength}{Numeric specifying edge lengths of consensus tree;
\code{NULL} aligns tips by scaling edges proportional to clade size; \code{1} sets
all edges to unit length.}

\item{thin, fat}{Numeric specifying width to plot edges if the rogue tip
never / sometimes does attach to them.}

\item{outgroupTips}{Vector of type character, integer or logical, specifying
the names or indices of the tips to include in the outgroup.
If \code{outgroupTips} is a of type character, and a tree contains multiple tips
with a matching label, the first will be used.}

\item{sort}{Logical specifying whether to sort consensus tree using
\code{\link[=SortTree]{SortTree()}}.}

\item{legend}{Character vector specifying position of legend (e.g.
\code{"bottomleft"}), or \code{"none"} to suppress legend.
For fine-grained control of legend, use \code{\link[PlotTools:SpectrumLegend]{PlotTools::SpectrumLegend()}}.}

\item{legend.inset}{Numeric specifying fraction of plot width / height
by which the legend's position should be inset.}

\item{\dots}{Additional parameters to \code{plot.phylo()}.}
}
\value{
\code{RoguePlot()} invisibly returns a list whose elements are:
\itemize{
\item \code{cons}: The reduced consensus tree, sorted if \code{sort = TRUE}, otherwise
in preorder;
\item \code{onEdge}: a vector of integers specifying the number of
trees in \code{trees} in which the rogue leaf is attached to each edge in turn
of the consensus tree;
\item \code{atNode}: a vector of integers specifying the number of trees in \code{trees}
in which the rogue leaf is attached to an edge collapsed into each node
of the consensus tree.
\item \code{legendLabels}: A character vector suggesting labels for a plot legend;
suitable for \code{PlotTools::SpectrumLegend(legend = x$legendLabels)}.
}
}
\description{
Plots a consensus of trees with a rogue taxon omitted, with edges coloured
according to the proportion of trees in which the taxon attaches to that
edge, after \insertCite{Klopfstein2019;textual}{TreeTools}.
}
\details{
Rogue taxa can be identified using the package \pkg{Rogue}
\insertCite{SmithCons}{TreeTools}.
}
\examples{
trees <- list(read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),
              read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),
              read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),
              read.tree(text = "(a, (b, (c, (rogue, (d, (e, f))))));"),
              read.tree(text = "(rogue, (a, (b, (c, (d, (e, f))))));"),
              read.tree(text = "((rogue, a), (b, (c, (d, (e, f)))));"),
              read.tree(text = "(a, (b, ((c, d), (rogue, (e, f)))));"),
              read.tree(text = "(a, (b, ((c, (rogue, d)), (e, f))));"),
              read.tree(text = "(a, (b, (c, (d, (rogue, (e, f))))));"))
plotted <- RoguePlot(trees, "rogue", legend = "topleft", legend.inset = 0.02)
PlotTools::SpectrumLegend(
  "bottomleft",
  palette = colorRampPalette(c(par("fg"), "#009E73"), space = "Lab")(100),
  legend = plotted$legendLabels,
  cex = 0.4
)
}
\references{
\insertAllCited{}
}
\seealso{
Other consensus tree functions:
\code{\link[=Consensus]{Consensus()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{consensus tree functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DropTip.R
\name{DropTip}
\alias{DropTip}
\alias{KeepTip}
\alias{DropTip.phylo}
\alias{KeepTip.phylo}
\alias{KeepTip.Splits}
\alias{DropTip.Splits}
\alias{DropTipPhylo}
\alias{DropTip.multiPhylo}
\alias{KeepTip.multiPhylo}
\alias{DropTip.list}
\alias{KeepTip.list}
\alias{DropTip.NULL}
\alias{KeepTip.NULL}
\alias{KeepTipPreorder}
\alias{KeepTipPostorder}
\title{Drop leaves from tree}
\usage{
DropTip(tree, tip, preorder = TRUE, check = TRUE)

KeepTip(tree, tip, preorder = TRUE, check = TRUE)

\method{DropTip}{phylo}(tree, tip, preorder = TRUE, check = TRUE)

\method{KeepTip}{phylo}(tree, tip, preorder = TRUE, check = TRUE)

\method{KeepTip}{Splits}(tree, tip, preorder = TRUE, check = TRUE)

\method{DropTip}{Splits}(tree, tip, preorder, check = TRUE)

DropTipPhylo(tree, tip, preorder = TRUE, check = TRUE)

\method{DropTip}{multiPhylo}(tree, tip, preorder = TRUE, check = TRUE)

\method{KeepTip}{multiPhylo}(tree, tip, preorder = TRUE, check = TRUE)

\method{DropTip}{list}(tree, tip, preorder = TRUE, check = TRUE)

\method{KeepTip}{list}(tree, tip, preorder = TRUE, check = TRUE)

\method{DropTip}{`NULL`}(tree, tip, preorder = TRUE, check = TRUE)

\method{KeepTip}{`NULL`}(tree, tip, preorder = TRUE, check = TRUE)

KeepTipPreorder(tree, tip)

KeepTipPostorder(tree, tip)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{tip}{Character vector specifying labels of leaves in tree to be
dropped or kept, or integer vector specifying the indices of leaves to be
dropped or kept.
Specifying the index of an internal node will drop all descendants of that
node.}

\item{preorder}{Logical specifying whether to \link{Preorder} \code{tree} before
dropping tips.  Specifying \code{FALSE} saves a little time, but will result in
undefined behaviour if \code{tree} is not in preorder.}

\item{check}{Logical specifying whether to check validity of \code{tip}. If
\code{FALSE} and \code{tip} contains entries that do not correspond to leaves of the
tree, undefined behaviour may occur.}
}
\value{
\code{DropTip()} returns a tree of class \code{phylo}, with the requested
leaves removed. The edges of the tree will be numbered in preorder,
but their sequence may not conform to the conventions of \code{\link[=Preorder]{Preorder()}}.

\code{KeepTip()} returns \code{tree} with all leaves not in \code{tip} removed,
in preorder.
}
\description{
\code{DropTip()} removes specified leaves from a phylogenetic tree, collapsing
incident branches.
}
\details{
This function differs from \code{\link[ape:drop.tip]{ape::drop.tip()}}, which roots unrooted trees,
and which can crash when trees' internal numbering follows unexpected schema.
}
\section{Functions}{
\itemize{
\item \code{DropTipPhylo()}: Direct call to \code{DropTip.phylo()}, to avoid overhead of
querying object's class.

\item \code{KeepTipPreorder()}: Faster version with no checks.
Does not retain labels or edge weights.
Edges must be listed in preorder.
May crash if improper input is specified.

\item \code{KeepTipPostorder()}: Faster version with no checks.
Does not retain labels or edge weights.
Edges must be listed in postorder.
May crash if improper input is specified.

}}
\examples{
tree <- BalancedTree(9)
plot(tree)
plot(DropTip(tree, c("t5", "t6")))

unrooted <- UnrootTree(tree)
plot(unrooted)
plot(DropTip(unrooted, 4:5))

summary(DropTip(as.Splits(tree), 4:5))
}
\seealso{
Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{CollapseNode}()},
\code{\link{ConsensusWithout}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{MakeTreeBinary}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}

Other split manipulation functions: 
\code{\link{SplitConsistent}()},
\code{\link{Subsplit}()},
\code{\link{TrivialSplits}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split manipulation functions}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_numbering.R
\name{RenumberTree}
\alias{RenumberTree}
\alias{RenumberEdges}
\alias{Reorder}
\alias{Cladewise}
\alias{Cladewise.phylo}
\alias{Cladewise.list}
\alias{Cladewise.multiPhylo}
\alias{Cladewise.matrix}
\alias{Cladewise.NULL}
\alias{ApePostorder}
\alias{ApePostorder.phylo}
\alias{ApePostorder.list}
\alias{ApePostorder.NULL}
\alias{ApePostorder.multiPhylo}
\alias{Postorder}
\alias{Postorder.phylo}
\alias{Postorder.NULL}
\alias{Postorder.list}
\alias{Postorder.multiPhylo}
\alias{Postorder.numeric}
\alias{PostorderOrder}
\alias{PostorderOrder.phylo}
\alias{PostorderOrder.numeric}
\alias{Pruningwise}
\alias{Pruningwise.phylo}
\alias{Pruningwise.list}
\alias{Pruningwise.multiPhylo}
\alias{Pruningwise.NULL}
\alias{Preorder}
\alias{Preorder.phylo}
\alias{Preorder.numeric}
\alias{Preorder.multiPhylo}
\alias{Preorder.list}
\alias{Preorder.NULL}
\alias{TntOrder}
\alias{TNTOrder}
\alias{TntOrder.phylo}
\alias{TntOrder.numeric}
\alias{TntOrder.multiPhylo}
\alias{TntOrder.list}
\alias{TntOrder.NULL}
\title{Reorder tree edges and nodes}
\usage{
RenumberTree(parent, child, weight)

RenumberEdges(parent, child)

Cladewise(tree, nTip, edge)

\method{Cladewise}{phylo}(tree, nTip = NTip(tree), edge = tree[["edge"]])

\method{Cladewise}{list}(tree, nTip, edge)

\method{Cladewise}{multiPhylo}(tree, nTip, edge)

\method{Cladewise}{matrix}(tree, nTip = min(tree[, 1]) - 1L, edge)

\method{Cladewise}{`NULL`}(tree, nTip = min(tree[, 1]) - 1L, edge)

ApePostorder(tree, nTip, edge)

\method{ApePostorder}{phylo}(tree, nTip = NTip(tree), edge = tree[["edge"]])

\method{ApePostorder}{list}(tree, nTip, edge)

\method{ApePostorder}{`NULL`}(tree, nTip, edge)

\method{ApePostorder}{multiPhylo}(tree, nTip, edge)

Postorder(tree, force = FALSE)

\method{Postorder}{phylo}(tree, force = FALSE)

\method{Postorder}{`NULL`}(tree, force = FALSE)

\method{Postorder}{list}(tree, force = FALSE)

\method{Postorder}{multiPhylo}(tree, force = FALSE)

\method{Postorder}{numeric}(tree, force = FALSE)

PostorderOrder(tree)

\method{PostorderOrder}{phylo}(tree)

\method{PostorderOrder}{numeric}(tree)

Pruningwise(tree, nTip, edge)

\method{Pruningwise}{phylo}(tree, nTip = NTip(tree), edge = tree[["edge"]])

\method{Pruningwise}{list}(tree, nTip, edge)

\method{Pruningwise}{multiPhylo}(tree, nTip, edge)

\method{Pruningwise}{`NULL`}(tree, nTip, edge)

Preorder(tree, topologyOnly = FALSE)

\method{Preorder}{phylo}(tree, topologyOnly = FALSE)

\method{Preorder}{numeric}(tree, topologyOnly = FALSE)

\method{Preorder}{multiPhylo}(tree, topologyOnly = FALSE)

\method{Preorder}{list}(tree, topologyOnly = FALSE)

\method{Preorder}{`NULL`}(tree, topologyOnly = FALSE)

TntOrder(tree)

TNTOrder(tree)

\method{TntOrder}{phylo}(tree)

\method{TntOrder}{numeric}(tree)

\method{TntOrder}{multiPhylo}(tree)

\method{TntOrder}{list}(tree)

\method{TntOrder}{`NULL`}(tree)
}
\arguments{
\item{parent}{Integer vector corresponding to the first column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 1]}}

\item{child}{Integer vector corresponding to the second column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 2]}.}

\item{weight}{Optional vector specifying the weight of each edge,
corresponding to the \code{edge.length} property of a \code{phylo} object.}

\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{nTip}{Integer specifying number of tips (leaves).}

\item{edge}{Two-column matrix listing the parent and child of each edge in a
tree, corresponding to \code{tree[["edge"]]}. Optional in \code{Cladewise()}.}

\item{force}{Logical specifying whether to rearrange trees already in
postorder, in order to ensure edges are ordered in the "TreeTools" fashion.}

\item{topologyOnly}{Logical; if \code{TRUE}, edge weights may not be retained.}
}
\value{
\code{RenumberTree()} returns an edge matrix for a tree of class \code{phylo}
following the preorder convention for edge and node numbering.

\code{RenumberEdges()} formats the output of \code{RenumberTree()} into a list
whose two entries correspond to the new parent and child vectors,
in preorder.

\code{ApePostorder()}, \code{Cladewise()}, \code{Postorder()}, \code{Preorder()} and
\code{Pruningwise()} each return a tree of class \code{phylo} with nodes following the
specified numbering scheme.

\code{Postorder.numeric} accepts a numeric matrix corresponding to the
\code{edge} entry of a tree of class \code{phylo}, and returns a two-column array
corresponding to \code{tree}, with edges listed in postorder

\code{PostorderOrder()} returns an integer vector. Visiting edges in this
order will traverse the tree in postorder.
}
\description{
Functions for systematically ordering the internal edges of trees.
}
\details{
\code{Reorder()} is a wrapper for \code{ape:::.reorder_ape}.
Calling this C function directly is approximately twice as fast as using
\code{ape::\link[ape:reorder.phylo]{cladewise}} or
\code{ape::\link[ape:reorder.phylo]{postorder}}

\code{Cladewise()}, \code{ApePostorder()} and \code{Pruningwise()} are convenience
functions to the corresponding functions in "ape".
Single nodes may need to be collapsed using \link[ape:collapse.singles]{ape::collapse.singles} first.
"ape" functions can cause crashes if nodes are numbered unconventionally --
sometimes arising after using tree rearrangement functions,
e.g. \code{phangorn::SPR()}.

\code{Preorder()} is more robust: it supports polytomies, nodes may be numbered
in any sequence, and edges may be listed in any order in the input tree.
Its output is guaranteed to be identical for any tree of an equivalent
leaf labelling (see \code{RenumberTips()}) and topology,
allowing unique trees to be detected by comparing sorted edge matrices alone.

Nodes and edges in a preorder tree are numbered starting from the deepest
node.
Each node is numbered in the sequence in which it is encountered, and
each edge is listed in the sequence in which it is visited.

At each node, child edges are sorted from left to right in order of the
lowest-numbered leaf in the subtree subtended by each edge; i.e. an edge
that leads eventually to tip 1 will be to the left of an edge leading to a
subtree containing tip 2.

Numbering begins by following the leftmost edge of the root node,
and sorting its descendant subtree into preorder.
Then, the next edge at the root node is followed, and its descendants
sorted into preorder, until each edge has been visited.

\code{RenumberTree()} and \code{RenumberEdges()} are wrappers for the C function
\code{preorder_edges_and_nodes()}; they do not perform the same checks on input
as \code{Preorder()} and are intended for use where performance is at a premium.

\code{Postorder()} numbers nodes as in \code{Preorder()}, and lists edges in
descending order of parent node number, breaking ties by listing child
nodes in increasing order.  If a tree is already in postorder, it will not
be rearranged unless \code{force = TRUE}.

Methods applied to numeric inputs do not check input for sanity, so should
be used with caution: malformed input may cause undefined results, including
crashing R.

Trees with >8191 leaves require additional memory and are not handled
by \code{Postorder()} at present.
If you need to process such large trees, please contact the maintainer for
advice.
}
\section{Functions}{
\itemize{
\item \code{Cladewise()}: Reorder tree cladewise.

\item \code{ApePostorder()}: Reorder tree in Postorder using ape's \code{postorder}
function, which is robust to unconventional node numbering.

\item \code{Pruningwise()}: Reorder tree Pruningwise.

\item \code{Preorder()}: Reorder tree in Preorder (special case of cladewise).

\item \code{TntOrder()}: Reorder tree in postorder, numbering internal nodes
according to \href{https://stackoverflow.com/a/54296100/3438001}{TNT's rules},
which number the root node as \code{nTip + 1}, then the remaining nodes in
the sequence encountered when traversing the tree in postorder, starting from
each tip in sequence.

}}
\seealso{
Rotate each node into a consistent orientation with \code{\link[=SortTree]{SortTree()}}.

Other tree manipulation:
\code{\link[=AddTip]{AddTip()}},
\code{\link[=CollapseNode]{CollapseNode()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=DropTip]{DropTip()}},
\code{\link[=ImposeConstraint]{ImposeConstraint()}},
\code{\link[=KeptPaths]{KeptPaths()}},
\code{\link[=KeptVerts]{KeptVerts()}},
\code{\link[=LeafLabelInterchange]{LeafLabelInterchange()}},
\code{\link[=MakeTreeBinary]{MakeTreeBinary()}},
\code{\link[=Renumber]{Renumber()}},
\code{\link[=RenumberTips]{RenumberTips()}},
\code{\link[=RootTree]{RootTree()}},
\code{\link[=SortTree]{SortTree()}},
\code{\link[=Subtree]{Subtree()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link{TrivialTree}}

Other C wrappers:
\code{\link{Neworder}}
}
\author{
\code{Preorder()} and \code{Postorder()}: Martin R. Smith.

\code{Cladewise()}, \code{ApePostorder()} and \code{Pruningwise()}: modified by Martin R.
Smith from \code{.reorder_ape()} in \pkg{ape} (Emmanuel Paradis).
}
\concept{C wrappers}
\concept{tree manipulation}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.multiPhylo.R
\name{as.multiPhylo}
\alias{as.multiPhylo}
\alias{as.multiPhylo.phylo}
\alias{as.multiPhylo.list}
\alias{as.multiPhylo.phyDat}
\alias{as.multiPhylo.Splits}
\title{Convert object to \code{multiPhylo} class}
\usage{
as.multiPhylo(x)

\method{as.multiPhylo}{phylo}(x)

\method{as.multiPhylo}{list}(x)

\method{as.multiPhylo}{phyDat}(x)

\method{as.multiPhylo}{Splits}(x)
}
\arguments{
\item{x}{Object to be converted}
}
\value{
\code{as.multiPhylo} returns an object of class \code{multiPhylo}

\code{as.multiPhylo.phyDat()} returns a list of trees, each corresponding
to the partitions implied by each non-ambiguous character in \code{x}.
}
\description{
Converts representations of phylogenetic trees to an object of the "ape"
class \code{multiPhylo}.
}
\examples{
as.multiPhylo(BalancedTree(8))
as.multiPhylo(list(BalancedTree(8), PectinateTree(8)))
data("Lobo")
as.multiPhylo(Lobo.phy)
}
\seealso{
Other utility functions:
\code{\link{ClusterTable}},
\code{\link{ClusterTable-methods}},
\code{\link[=Hamming]{Hamming()}},
\code{\link[=MSTEdges]{MSTEdges()}},
\code{\link[=SampleOne]{SampleOne()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link[=UnshiftTree]{UnshiftTree()}},
\code{\link{match,phylo,phylo-method}},
\code{\link[=sapply64]{sapply64()}},
\code{\link[=sort.multiPhylo]{sort.multiPhylo()}}
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_rearrangement.R
\name{LeafLabelInterchange}
\alias{LeafLabelInterchange}
\title{Leaf label interchange}
\usage{
LeafLabelInterchange(tree, n = 2L)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{n}{Integer specifying number of leaves whose positions should be
exchanged.}
}
\value{
\code{LeafLabelInterchange()} returns a tree of class \code{phylo} on which
the position of \code{n} leaves have been exchanged.
The tree's internal topology will not change.
}
\description{
\code{LeafLabelInterchange()} exchanges the position of leaves within a tree.
}
\details{
Modifies a tree by switching the positions of \emph{n} leaves.  To avoid
later swaps undoing earlier exchanges, all \emph{n} leaves are guaranteed
to change position.  Note, however, that no attempt is made to avoid
swapping equivalent leaves, for example, a pair that are each others'
closest relatives.  As such, the relationships within a tree are not
guaranteed to be changed.
}
\examples{
tree <- PectinateTree(8)
plot(LeafLabelInterchange(tree, 3L))

}
\seealso{
Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{CollapseNode}()},
\code{\link{ConsensusWithout}()},
\code{\link{DropTip}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{MakeTreeBinary}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_comparison.R
\name{PairwiseDistances}
\alias{PairwiseDistances}
\title{Distances between each pair of trees}
\usage{
PairwiseDistances(trees, Func, valueLength = 1L, ...)
}
\arguments{
\item{trees}{List of trees of class \code{phylo}.}

\item{Func}{Function returning a distance between two trees.}

\item{valueLength}{Integer specifying expected length of the value returned
by \code{Func}.}

\item{\dots}{Additional arguments to \code{Func}.}
}
\value{
Matrix detailing distance between each pair of trees.
Identical trees are assumed to have zero distance.
}
\description{
Distances between each pair of trees
}
\examples{
trees <- list(BalancedTree(8), PectinateTree(8), StarTree(8))
TCIDiff <- function(tree1, tree2) {
  TotalCopheneticIndex(tree1) - TotalCopheneticIndex(tree2)
}
PairwiseDistances(trees, TCIDiff, 1)
TCIRange <- function(tree1, tree2) {
  range(TotalCopheneticIndex(tree1), TotalCopheneticIndex(tree2))
}
PairwiseDistances(trees, TCIRange, 2)
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{pairwise tree distances}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split_analysis.R
\name{TipsInSplits}
\alias{TipsInSplits}
\alias{TipsInSplits.Splits}
\alias{TipsInSplits.phylo}
\alias{SplitImbalance}
\alias{SplitImbalance.Splits}
\alias{SplitImbalance.phylo}
\title{Tips contained within splits}
\usage{
TipsInSplits(splits, keep.names = TRUE, smallest = FALSE, ...)

\method{TipsInSplits}{Splits}(splits, keep.names = TRUE, smallest = FALSE, ...)

\method{TipsInSplits}{phylo}(splits, keep.names = TRUE, smallest = FALSE, ...)

SplitImbalance(splits, keep.names = TRUE, ...)

\method{SplitImbalance}{Splits}(splits, keep.names = TRUE, ...)

\method{SplitImbalance}{phylo}(splits, keep.names = TRUE, ...)
}
\arguments{
\item{splits}{Object of class \code{Splits} or \code{phylo}.}

\item{keep.names}{Logical specifying whether to include the names of \code{splits}
in the output.}

\item{smallest}{Logical; if \code{TRUE}, return the number of leaves in the
smaller bipartition.}

\item{\dots}{Additional parameters to pass to \code{as.Splits()}.}
}
\value{
\code{TipsInSplits()} returns a named vector of integers, specifying the
number of tips contained within each split in \code{splits}.

\code{SplitImbalance()} returns a named vector of integers, specifying the
number of leaves within a split that are not "balanced" by a leaf outside it;
i.e. a split that divides leaves evenly has an imbalance of zero; one that
splits two tips from ten has an imbalance of 10 - 2 = 8.
}
\description{
\code{TipsInSplits()} specifies the number of tips that occur within each
bipartition split in a \code{Splits} object.
}
\examples{
tree <- PectinateTree(8)
splits <- as.Splits(tree)
TipsInSplits(splits)

plot(tree)
LabelSplits(tree, as.character(splits), frame = "none", pos = 3L, cex = 0.7)
LabelSplits(tree, TipsInSplits(splits), unit = " tips", frame = "none",
            pos = 1L)

}
\seealso{
Other Splits operations:
\code{\link[=LabelSplits]{LabelSplits()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=PolarizeSplits]{PolarizeSplits()}},
\code{\link[=SplitFrequency]{SplitFrequency()}},
\code{\link{Splits}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link{match,Splits,Splits-method}},
\code{\link[=xor]{xor()}}
}
\concept{Splits operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/KeptPaths.R
\name{KeptPaths}
\alias{KeptPaths}
\alias{KeptPaths.data.frame}
\alias{KeptPaths.matrix}
\title{Paths present in reduced tree}
\usage{
KeptPaths(paths, keptVerts, all = TRUE)

\method{KeptPaths}{data.frame}(paths, keptVerts, all = TRUE)

\method{KeptPaths}{matrix}(paths, keptVerts, all = TRUE)
}
\arguments{
\item{paths}{\code{data.frame} of paths in master tree, perhaps generated using
\code{\link[=PathLengths]{PathLengths()}}.}

\item{keptVerts}{Logical specifying whether each entry is retained in the
reduced tree, perhaps generated using \code{\link[=KeptVerts]{KeptVerts()}}.}

\item{all}{Logical: if \code{TRUE}, return all paths that occur in the reduced
tree; if \code{FALSE}, return only those paths that correspond to a single edge.
that correspond to edges in the reduced tree.
Ignored if \code{paths} is a matrix.}
}
\value{
\code{KeptPaths()} returns a logical vector specifying whether each path
in \code{paths} occurs when \code{keptVerts} vertices are retained.
}
\description{
Lists which paths present in a master tree are present when leaves are
dropped.
}
\examples{
master <- BalancedTree(9)
paths <- PathLengths(master)
keptTips <- c(1, 5, 7, 9)
keptVerts <- KeptVerts(master, keptTips)
KeptPaths(paths, keptVerts)
paths[KeptPaths(paths, keptVerts, all = FALSE), ]
}
\seealso{
Other tree manipulation:
\code{\link[=AddTip]{AddTip()}},
\code{\link[=CollapseNode]{CollapseNode()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=DropTip]{DropTip()}},
\code{\link[=ImposeConstraint]{ImposeConstraint()}},
\code{\link[=KeptVerts]{KeptVerts()}},
\code{\link[=LeafLabelInterchange]{LeafLabelInterchange()}},
\code{\link[=MakeTreeBinary]{MakeTreeBinary()}},
\code{\link[=Renumber]{Renumber()}},
\code{\link[=RenumberTips]{RenumberTips()}},
\code{\link[=RenumberTree]{RenumberTree()}},
\code{\link[=RootTree]{RootTree()}},
\code{\link[=SortTree]{SortTree()}},
\code{\link[=Subtree]{Subtree()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{UnshiftTree}
\alias{UnshiftTree}
\title{Add tree to start of list}
\usage{
UnshiftTree(add, treeList)
}
\arguments{
\item{add}{Tree to add to the list, of class \code{\link[ape:read.tree]{phylo}}.}

\item{treeList}{A list of trees, of class \code{list},
\code{\link[ape:multiphylo]{multiPhylo}}, or, if a single tree,
\code{\link[ape:read.tree]{phylo}}.}
}
\value{
\code{UnshiftTree()} returns a list of class \code{list} or \code{multiPhylo}
(following the original class of \code{treeList}), whose first element is the
tree specified as `add.
}
\description{
\code{UnshiftTree()} adds a phylogenetic tree to the start of a list of trees.
This is useful where the class of a list of trees is unknown, or where
names of trees should be retained.
}
\details{
Caution: adding a tree to a \code{multiPhylo} object whose own attributes apply
to all trees, for example trees read from a Nexus file, causes data to be
lost.
}
\examples{
forest <- as.phylo(0:5, 6)
tree <- BalancedTree(6)

UnshiftTree(tree, forest)
UnshiftTree(tree, tree)
}
\seealso{
\code{\link[=c]{c()}} joins a tree or series of trees to a \code{multiPhylo} object, but loses
names and does not handle lists of trees.

Other utility functions: 
\code{\link{ClusterTable}},
\code{\link{ClusterTable-methods}},
\code{\link{Hamming}()},
\code{\link{MSTEdges}()},
\code{\link{SampleOne}()},
\code{\link{TipTimedTree}()},
\code{\link{as.multiPhylo}()},
\code{\link{match,phylo,phylo-method}},
\code{\link{sapply64}()},
\code{\link{sort.multiPhylo}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PaintTree.R
\name{PaintTree}
\alias{PaintTree}
\title{Colour a tree by topology}
\usage{
PaintTree(tree, palette = "default")
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{palette}{Either a character string naming one of the built-in
palettes (\code{"default"}, \code{"protanopia"}, \code{"tritanopia"},
or a function \verb{function(h, s)} taking vectors of hues in
\verb{[0, 360]} and saturations in \verb{[0, 1]} and returning a character vector of
colours.  The \code{"protanopia"} and \code{"tritanopia"} palettes simulate
dichromatic perception of the default HCL spectrum using the
Viénot–Brettel–Mollon projection.}
}
\value{
A list with three character vectors of hex colours:
\describe{
\item{\code{edgeCol}}{One colour per edge in \code{tree$edge}, taken from the
child node's \verb{(hue, saturation)}.}
\item{\code{tipCol}}{One colour per leaf, indexed \code{1:NTip(tree)}.}
\item{\code{nodeCol}}{One colour per internal node, indexed so that entry \code{i}
corresponds to node \code{NTip(tree) + i}; the root (entry 1) is grey
because its saturation is zero.}
}
}
\description{
\code{PaintTree()} assigns a colour to every edge, leaf, and internal node of a
tree such that sister clades occupy adjacent hue bands proportional to their
tip counts, and saturation grows from zero at the root to one at every tip.
The result is a list of vectors that correspond to
\code{\link[ape:plot.phylo]{plot.phylo()}}'s
\code{edge.color}, \code{tip.color}, and \code{node.color} arguments.
}
\details{
Hue is allocated recursively from a 0–360° budget at the root.  At each
internal node, the parent's budget is partitioned across its descendant
edges in proportion to the number of leaves descended from each child; the
colour reported for an edge or node is the midpoint of its assigned range.
Saturation \code{s = (nTip - nDesc) / (nTip - 1)}, where \code{nDesc} is the number of
leaves descended from the node (including itself for tips), so the root is
achromatic (\code{s = 0}) and every leaf is fully saturated (\code{s = 1}).
}
\examples{
tree <- BalancedTree(1:8) + PectinateTree(9:14)
tree <- ape::bind.tree(BalancedTree(1:8), PectinateTree(9:12), 8, 4)
paint <- PaintTree(tree)
plot(tree,
     edge.color = paint$edgeCol,
     tip.color = paint$tipCol,
     edge.width = 3)

# Colour-blind-safe variants
paintP <- PaintTree(tree, "protanopia")
plot(tree,
     edge.color = paintP$edgeCol,
     tip.color = paintP$tipCol,
     edge.width = 3)

paintT <- PaintTree(tree, "tritanopia")
plot(tree,
     edge.color = paintT$edgeCol,
     tip.color = paintT$tipCol,
     edge.width = 3)

# User-supplied palette function (greyscale)
grey_pal <- function(h, s) grey(1 - s * 0.8)
paintG <- PaintTree(tree, grey_pal)
plot(tree, edge.color = paintG$edgeCol, edge.width = 3)

}
\seealso{
\code{\link[=CladeSizes]{CladeSizes()}}, \code{\link[=DescendantEdges]{DescendantEdges()}}

Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=RootNode]{RootNode()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Combinatorics.R
\name{DoubleFactorial}
\alias{DoubleFactorial}
\alias{DoubleFactorial64}
\alias{LnDoubleFactorial}
\alias{Log2DoubleFactorial}
\alias{LogDoubleFactorial}
\alias{LnDoubleFactorial.int}
\alias{LogDoubleFactorial.int}
\title{Double factorial}
\usage{
DoubleFactorial(n)

DoubleFactorial64(n)

LnDoubleFactorial(n)

Log2DoubleFactorial(n)

LogDoubleFactorial(n)

LnDoubleFactorial.int(n)

LogDoubleFactorial.int(n)
}
\arguments{
\item{n}{Vector of integers.}
}
\value{
Returns the double factorial, \emph{n} * (\emph{n} - 2) * (\emph{n} - 4) *
(\emph{n} - 6) * ...
}
\description{
Calculate the double factorial of a number, or its logarithm.
}
\section{Functions}{
\itemize{
\item \code{DoubleFactorial64()}: Returns the exact double factorial as a 64-bit
\code{integer64}, for \code{n} < 34.

\item \code{LnDoubleFactorial()}: Returns the logarithm of the double factorial.

\item \code{Log2DoubleFactorial()}: Returns the logarithm of the double factorial.

\item \code{LnDoubleFactorial.int()}: Slightly faster, when x is known to be length one
and below 50001

}}
\examples{
DoubleFactorial (-4:0) # Return 1 if n < 2
DoubleFactorial (2) # 2
DoubleFactorial (5) # 1 * 3 * 5
exp(LnDoubleFactorial.int (8)) # log(2 * 4 * 6 * 8)
DoubleFactorial64(31)
}
\seealso{
Other double factorials:
\code{\link{doubleFactorials}},
\code{\link{logDoubleFactorials}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{double factorials}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TreeTools-package.R
\docType{package}
\name{TreeTools-package}
\alias{TreeTools}
\alias{TreeTools-package}
\title{TreeTools}
\description{
"TreeTools" is an R package that provides functions for creating, modifying and
analysing phylogenetic trees.
It complements packages such as
\href{https://cran.r-project.org/package=ape}{\pkg{ape}},
\href{https://cran.r-project.org/package=phangorn}{\pkg{phangorn}} and
\href{https://cran.r-project.org/package=phytools}{\pkg{phytools}},
aiming for efficient and robust implementations of functions, typically
applied to unweighted trees (i.e. those without edge lengths).
}
\details{
Full documentation is available \href{https://ms609.github.io/TreeTools/}{online}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://ms609.github.io/TreeTools/}
  \item \url{https://github.com/ms609/TreeTools/}
  \item Report bugs at \url{https://github.com/ms609/TreeTools/issues/}
}

}
\author{
\strong{Maintainer}: Martin R. Smith \email{martin.smith@durham.ac.uk} (\href{https://orcid.org/0000-0001-5660-1727}{ORCID}) [copyright holder]

Other contributors:
\itemize{
  \item Emmanuel Paradis (\href{https://orcid.org/0000-0003-3092-2199}{ORCID}) (ape library) [copyright holder]
  \item Robert Noble (\href{https://orcid.org/0000-0002-8057-4252}{ORCID}) (RUtreebalance) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Splits.R
\name{Splits}
\alias{Splits}
\alias{as.Splits}
\alias{as.Splits.phylo}
\alias{as.Splits.multiPhylo}
\alias{as.Splits.Splits}
\alias{as.Splits.list}
\alias{as.Splits.matrix}
\alias{as.Splits.integer}
\alias{as.Splits.numeric}
\alias{as.Splits.logical}
\alias{as.Splits.character}
\alias{as.logical.Splits}
\title{Convert object to \code{Splits}}
\usage{
as.Splits(x, tipLabels = NULL, ...)

\method{as.Splits}{phylo}(x, tipLabels = NULL, asSplits = TRUE, ...)

\method{as.Splits}{multiPhylo}(x, tipLabels = unique(unlist(TipLabels(x))), asSplits = TRUE, ...)

\method{as.Splits}{Splits}(x, tipLabels = NULL, ...)

\method{as.Splits}{list}(x, tipLabels = NULL, asSplits = TRUE, ...)

\method{as.Splits}{matrix}(x, tipLabels = NULL, ...)

\method{as.Splits}{integer}(x, tipLabels = NULL, ...)

\method{as.Splits}{numeric}(x, tipLabels = NULL, ...)

\method{as.Splits}{logical}(x, tipLabels = NULL, ...)

\method{as.Splits}{character}(x, tipLabels = NULL, ...)

\method{as.logical}{Splits}(x, tipLabels = attr(x, "tip.label"), ...)
}
\arguments{
\item{x}{Object to convert into splits: perhaps a tree of class
\code{\link[ape:read.tree]{phylo}}.
If a logical matrix is provided, each row will be considered as a separate
split.
If an integer matrix is provided, entries assigned the same integer will
be assigned to the same split.}

\item{tipLabels}{Character vector specifying sequence in which to order
tip labels.  Label order must (currently) match to combine or compare separate
\code{Splits} objects.}

\item{\dots}{Presently unused.}

\item{asSplits}{Logical specifying whether to return a \code{Splits} object,
or an unannotated two-dimensional array (useful where performance is
paramount).}
}
\value{
\code{as.Splits()} returns an object of class \code{Splits}, or
(if \code{asSplits = FALSE}) a two-dimensional array of \code{raw} objects, with
each bit specifying whether or not the leaf corresponding to the respective
bit position is a member of the split.
Splits are named according to the node at the non-root end of the edge that
defines them. In rooted trees, the child of the rightmost root edge names
the split.
}
\description{
\code{as.Splits()} converts a phylogenetic tree to a \code{Splits} object representing
its constituent bipartition splits.
}
\examples{
splits <- as.Splits(BalancedTree(letters[1:6]))
summary(splits)
TipsInSplits(splits)
summary(!splits)
TipsInSplits(!splits)

length(splits + !splits)
length(unique(splits + !splits))

summary(c(splits[[2:3]], !splits[[1:2]]))

moreSplits <- as.Splits(PectinateTree(letters[6:1]), tipLabel = splits)
print(moreSplits, details = TRUE)
match(splits, moreSplits)
moreSplits \%in\% splits

as.Splits("....**", letters[1:6])

}
\seealso{
Other Splits operations:
\code{\link[=LabelSplits]{LabelSplits()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=PolarizeSplits]{PolarizeSplits()}},
\code{\link[=SplitFrequency]{SplitFrequency()}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TipsInSplits]{TipsInSplits()}},
\code{\link{match,Splits,Splits-method}},
\code{\link[=xor]{xor()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{Splits operations}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ArtificialExtinction.R
\name{ArtificialExtinction}
\alias{ArtificialExtinction}
\alias{ArtificialExtinction.matrix}
\alias{ArtificialExtinction.phyDat}
\alias{ArtEx}
\title{Artificial Extinction}
\usage{
ArtificialExtinction(
  dataset,
  subject,
  template,
  replaceAmbiguous = "ambig",
  replaceCoded = "original",
  replaceAll = TRUE,
  sampleFrom = NULL
)

\method{ArtificialExtinction}{matrix}(
  dataset,
  subject,
  template,
  replaceAmbiguous = "ambig",
  replaceCoded = "original",
  replaceAll = TRUE,
  sampleFrom = NULL
)

\method{ArtificialExtinction}{phyDat}(
  dataset,
  subject,
  template,
  replaceAmbiguous = "ambig",
  replaceCoded = "original",
  replaceAll = TRUE,
  sampleFrom = NULL
)

ArtEx(
  dataset,
  subject,
  template,
  replaceAmbiguous = "ambig",
  replaceCoded = "original",
  replaceAll = TRUE,
  sampleFrom = NULL
)
}
\arguments{
\item{dataset}{Phylogenetic dataset of class \code{phyDat} or \code{matrix}.}

\item{subject}{Vector identifying subject taxa, by name or index.}

\item{template}{Character or integer identifying taxon to use as a template.}

\item{replaceAmbiguous, replaceCoded}{Character specifying whether tokens
that are ambiguous (\verb{?}) or coded (not \verb{?}) in the fossil template should
be replaced with:
\itemize{
\item \code{original}: Their original value; i.e. no change;
\item \code{ambiguous}: The ambiguous token, \verb{?};
\item \code{binary}: The tokens \code{0} or \code{1}, with equal probability;
\item \code{uniform}: One of the tokens present in \code{sampleFrom}, with equal
probability;
\item \code{sample}: One of the tokens present in \code{sampleFrom}, sampled according
to their frequency.
}}

\item{replaceAll}{Logical: if \code{TRUE}, replace all tokens in a subject; if
\code{FALSE}, leave any ambiguous tokens (\verb{?}) ambiguous.}

\item{sampleFrom}{Vector identifying a subset of characters from which to
sample replacement tokens.
If \code{NULL}, replacement tokens will be sampled from the initial states of
all taxa not used as a template (including the subjects).}
}
\value{
A dataset with the same class as \code{dataset} in which entries that
are ambiguous in \code{template} are made ambiguous in \code{subject}.
}
\description{
Remove tokens that do not occur in a fossil "template" taxon from a living
taxon, to simulate the process of fossilization in removing data from
a phylogenetic dataset.
}
\details{
Further details are provided in \insertCite{Asher2020;textual}{TreeTools}.

Note: this simple implementation does not account for character contingency,
e.g. characters whose absence imposes inapplicable or absent tokens on
dependent characters.
}
\examples{
set.seed(1)
dataset <- matrix(c(sample(0:2, 4 * 8, TRUE),
                    "0", "0", rep("?", 6)), nrow = 5,
                    dimnames = list(c(LETTERS[1:4], "FOSSIL"),
                                    paste("char", 1:8)), byrow = TRUE)
artex <- ArtificialExtinction(dataset, c("A", "C"), "FOSSIL")
}
\references{
\insertAllCited{}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/SplitFunctions.R
\name{NPartitionPairs}
\alias{NPartitionPairs}
\title{Distributions of tips consistent with a partition pair}
\usage{
NPartitionPairs(configuration)
}
\arguments{
\item{configuration}{Integer vector of length four specifying the number of
terminals that occur in both
(1) splits A1 and A2;
(2) splits A1 and B2;
(3) splits B1 and A2;
(4) splits B1 and B2.}
}
\value{
The number of ways to distribute \code{sum(configuration)} taxa according
to the specified pattern.
}
\description{
\code{NPartitionPairs()} calculates the number of terminal arrangements matching
a specified configuration of two splits.
}
\details{
Consider splits that divide eight terminals, labelled A to H.

\tabular{rcll}{
Bipartition 1:\tab ABCD:EFGH\tab A1 = ABCD\tab B1 = EFGH \cr
Bipartition 2:\tab ABE:CDFGH\tab A2 = ABE\tab B2 = CDFGH
}

This can be represented by an association matrix:

\tabular{rll}{
\tab \emph{A2} \tab \emph{B2} \cr
\emph{A1} \tab AB  \tab C   \cr
\emph{B1} \tab E   \tab FGH
}

The cells in this matrix contain 2, 1, 1 and 3 terminals respectively; this
four-element vector (\code{c(2, 1, 1, 3)}) is the \code{configuration} implied by
this pair of bipartition splits.
}
\examples{
NPartitionPairs(c(2, 1, 1, 3))
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MatchStrings.R
\name{MatchStrings}
\alias{MatchStrings}
\title{Check for mismatch between character vectors}
\usage{
MatchStrings(x, table, Fail = stop, max.distance = 0.5, ...)
}
\arguments{
\item{x, table}{Character vectors, in which all elements of \code{x} are expected
to occur in \code{table}.}

\item{Fail}{Function to call if a mismatch is found.}

\item{max.distance, \dots}{Arguments to \code{\link[base:agrep]{agrep()}},
used to propose possible matches to the user.}
}
\value{
\code{MatchStrings()} returns the elements of \code{x} that occur in \code{table}.
}
\description{
Checks that entries in one character vector occur in another, suggesting
corrections for mismatched elements.
}
\examples{
tree <- BalancedTree(8)
MatchStrings(c("t1", "tip2", "t3"), TipLabels(tree), Fail = message)
}
\seealso{
Other string parsing functions:
\code{\link[=EndSentence]{EndSentence()}},
\code{\link[=MorphoBankDecode]{MorphoBankDecode()}},
\code{\link[=RightmostCharacter]{RightmostCharacter()}},
\code{\link[=Unquote]{Unquote()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{string parsing functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_properties.R
\name{NodeNumbers}
\alias{NodeNumbers}
\title{Numeric index of each node in a tree
\code{NodeNumbers()} returns a sequence corresponding to the nodes in a tree}
\usage{
NodeNumbers(tree, tips = FALSE)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{tips}{Logical specifying whether to also include the indices of leaves.}
}
\value{
\code{NodeNumbers()} returns an integer vector corresponding to the
indices of nodes within a tree.
}
\description{
Numeric index of each node in a tree
\code{NodeNumbers()} returns a sequence corresponding to the nodes in a tree
}
\seealso{
Other tree properties:
\code{\link[=Cherries]{Cherries()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=EdgeRatio]{EdgeRatio()}},
\code{\link[=LongBranch]{LongBranch()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=PathLengths]{PathLengths()}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TreeIsRooted]{TreeIsRooted()}},
\code{\link[=Treeness]{Treeness()}}

Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=PaintTree]{PaintTree()}},
\code{\link[=RootNode]{RootNode()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Information.R
\name{UnrootedTreesMatchingSplit}
\alias{UnrootedTreesMatchingSplit}
\alias{LnUnrootedTreesMatchingSplit}
\alias{Log2UnrootedTreesMatchingSplit}
\title{Number of trees consistent with split}
\usage{
UnrootedTreesMatchingSplit(...)

LnUnrootedTreesMatchingSplit(...)

Log2UnrootedTreesMatchingSplit(...)
}
\arguments{
\item{\dots}{A series or vector of integers listing the number of tips in
each of a number of tree splits (e.g. bipartitions).
For example, \verb{3, 5} states that a character divides a set of eight tips into
a group of three and a group of five.}
}
\value{
\code{UnrootedTreesMatchingSplit()} returns an integer specifying the
number of unrooted bifurcating trees consistent with the specified split.
}
\description{
Calculates the number of unrooted bifurcating trees consistent with the
specified multi-partition split, using theorem two of
\insertCite{Carter1990;textual}{TreeTools}.
}
\examples{
UnrootedTreesMatchingSplit(c(3, 5))
UnrootedTreesMatchingSplit(3, 2, 1, 2)
}
\references{
\insertAllCited{}
}
\seealso{
Other split information functions: 
\code{\link{CharacterInformation}()},
\code{\link{SplitInformation}()},
\code{\link{SplitMatchProbability}()},
\code{\link{TreesMatchingSplit}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_display.R
\name{SortTree}
\alias{SortTree}
\alias{SortTree.phylo}
\alias{SortTree.list}
\alias{SortTree.multiPhylo}
\title{Sort tree}
\usage{
SortTree(tree, how = "cladesize", order = TipLabels(tree))

\method{SortTree}{phylo}(tree, how = "cladesize", order = TipLabels(tree))

\method{SortTree}{list}(tree, how = "cladesize", order = TipLabels(tree[[1]]))

\method{SortTree}{multiPhylo}(tree, how = "cladesize", order = TipLabels(tree[[1]]))
}
\arguments{
\item{tree}{One or more trees of class \code{phylo}, optionally as a list
or a \code{multiPhylo} object.}

\item{how}{Character vector specifying sort method:
\code{"Cladesize"} rotates each node such that the larger clade is first,
thus appearing lower when plotted;
\code{"TipLabels"} rotates nodes such that labels listed sooner in \code{order}
are listed first, and thus plot lower.}

\item{order}{Character vector listing tip labels in sequence they should
appear on tree. Clades containing a taxon earlier in this list will be listed
sooner and thus plot lower on a tree.  Taxa not listed in \code{order} will be
treated as if they were last in the list.}
}
\value{
\code{SortTree()} returns tree in the format of \code{tree}, with each node
in each tree sorted
}
\description{
\code{SortTree()} sorts each node into a consistent order, so that node rotation
does not obscure similarities between similar trees.
}
\details{
At each node, clades will be listed in \code{tree[["edge"]]} in decreasing size
order.

Clades that contain the same number of leaves are sorted in decreasing order
of minimum leaf number, so (2, 3) will occur before (1, 4).

As trees are plotted from "bottom up", the largest clades will "sink" to the
bottom of a plotted tree.
}
\examples{
messyTree <- as.phylo(10, 6)
plot(messyTree)

sorted <- SortTree(messyTree)
plot(sorted)
ape::nodelabels()
ape::edgelabels()
ape::tiplabels(adj = c(2, 1/3))

plot(SortTree(messyTree, how = "tip"))
}
\seealso{
\code{Preorder()} also rearranges trees into a consistent shape,
based on the index of leaves.

\code{\link[=sort.multiPhylo]{sort.multiPhylo()}} sorts a list of trees stored as a \code{multiPhylo}
object.

Other tree manipulation:
\code{\link[=AddTip]{AddTip()}},
\code{\link[=CollapseNode]{CollapseNode()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=DropTip]{DropTip()}},
\code{\link[=ImposeConstraint]{ImposeConstraint()}},
\code{\link[=KeptPaths]{KeptPaths()}},
\code{\link[=KeptVerts]{KeptVerts()}},
\code{\link[=LeafLabelInterchange]{LeafLabelInterchange()}},
\code{\link[=MakeTreeBinary]{MakeTreeBinary()}},
\code{\link[=Renumber]{Renumber()}},
\code{\link[=RenumberTips]{RenumberTips()}},
\code{\link[=RenumberTree]{RenumberTree()}},
\code{\link[=RootTree]{RootTree()}},
\code{\link[=Subtree]{Subtree()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_rearrangement.R
\name{RootTree}
\alias{RootTree}
\alias{RootOnNode}
\alias{UnrootTree}
\title{Root or unroot a phylogenetic tree}
\usage{
RootTree(tree, outgroupTips, fallback = NULL)

RootOnNode(tree, node, resolveRoot = FALSE)

UnrootTree(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}, or a list of trees
of class \code{list} or \code{multiPhylo}.}

\item{outgroupTips}{Vector of type character, integer or logical, specifying
the names or indices of the tips to include in the outgroup.
If \code{outgroupTips} is a of type character, and a tree contains multiple tips
with a matching label, the first will be used.}

\item{fallback}{Vector corresponding to \code{outgroupTips} determining behaviour
when \code{outgroupTips} do not root a tree.

Where the smallest clade that contains \code{outgroupTips} includes all taxa,
\code{RootTree()} will not change the topology of a tree.
If \code{fallback = NULL}, \code{RootTree()} will return \code{tree}.
Otherwise, taxa will be excluded from \code{outgroupTips} in the sequence
specified (\code{fallback[1]} first) by until the clade containing all outgroup
tips is smaller than \code{tree}.}

\item{node}{Integer specifying node (internal or tip) to set as the root.}

\item{resolveRoot}{Logical specifying whether to resolve the root node.}
}
\value{
\code{RootTree()} returns a tree of class \code{phylo}, rooted on the smallest
clade that contains the specified tips, with edges and nodes numbered in
preorder. Node labels are not retained.

\code{RootOnNode()} returns a tree of class \code{phylo}, rooted on the
requested \code{node} and ordered in \code{\link{Preorder}}.

\code{UnrootTree()} returns \code{tree}, in preorder,
having collapsed the first child of the root node in each tree.
}
\description{
\code{RootTree()} roots a tree on the smallest clade containing the specified
tips;
\code{RootOnNode()} roots a tree on a specified internal node;
\code{UnrootTree()} collapses a root node, without the undefined behaviour
encountered when using \code{\link[ape:root]{ape::unroot}()} on trees in
preorder.
}
\examples{
tree <- PectinateTree(8)
plot(tree)
ape::nodelabels()

plot(RootTree(tree, c("t6", "t7")))

plot(RootOnNode(tree, 12))
plot(RootOnNode(tree, 2))

}
\seealso{
\itemize{
\item \code{\link[ape:root]{ape::root()}}
}

Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{CollapseNode}()},
\code{\link{ConsensusWithout}()},
\code{\link{DropTip}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{MakeTreeBinary}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Treeness.R
\name{Treeness}
\alias{Treeness}
\alias{Stemminess}
\title{Relative length of internal branches}
\usage{
Treeness(tree)

Stemminess(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}, or a list of trees
of class \code{list} or \code{multiPhylo}.}
}
\value{
\code{Treeness()} returns a numeric vector reporting the treeness of each
\code{tree}.
}
\description{
Treeness (also termed stemminess) is the proportion of a tree's length found
on internal branches \insertCite{Lanyon1988}{TreeTools}.
Insofar as external branches do not contain phylogenetic (grouping) signal,
trees with a high treeness can be interpreted as containing a higher
signal:noise ratio \insertCite{Phillips2003}{TreeTools}.
}
\examples{
lowTree <- BalancedTree(6, lengths = c(1, 1, 4, 4, 4, 1, 1, 4, 4, 4))
plot(lowTree)
Treeness(lowTree)
highTree <- BalancedTree(6, lengths = c(6, 6, 1, 1, 1, 6, 6, 1, 1, 1))
plot(highTree)
Treeness(c(lowTree, highTree))
}
\references{
\insertAllCited{}
}
\seealso{
Other tree properties: 
\code{\link{Cherries}()},
\code{\link{ConsensusWithout}()},
\code{\link{EdgeRatio}()},
\code{\link{LongBranch}()},
\code{\link{MatchEdges}()},
\code{\link{NSplits}()},
\code{\link{NTip}()},
\code{\link{NodeNumbers}()},
\code{\link{PathLengths}()},
\code{\link{SplitsInBinaryTree}()},
\code{\link{TipLabels}()},
\code{\link{TreeIsRooted}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_generation.R
\name{Hamming}
\alias{Hamming}
\title{Hamming distance between taxa in a phylogenetic dataset}
\usage{
Hamming(
  dataset,
  ratio = TRUE,
  ambig = c("median", "mean", "zero", "one", "na", "nan")
)
}
\arguments{
\item{dataset}{Object of class \code{phyDat}.}

\item{ratio}{Logical specifying whether to weight distance against
maximum possible, given that a token that is ambiguous in either of two taxa
cannot contribute to the total distance between the pair.}

\item{ambig}{Character specifying value to return when a pair of taxa
have a zero maximum distance (perhaps due to a preponderance of ambiguous
tokens).
"median", the default, take the median of all other distance values;
"mean", the mean;
"zero" sets to zero; "one" to one;
"NA" to \code{NA_integer_}; and "NaN" to \code{NaN}.}
}
\value{
\code{Hamming()} returns an object of class \code{dist} listing the Hamming
distance between each pair of taxa.
}
\description{
The Hamming distance between a pair of taxa is the number of characters
with a different coding, i.e. the smallest number of evolutionary steps
that must have occurred since their common ancestor.
}
\details{
Tokens that contain the inapplicable state are treated as requiring no steps
to transform into any applicable token.
}
\examples{
tokens <- matrix(c(0, 0, "0", 0, "?",
                   0, 0, "1", 0, 1,
                   0, 0, "1", 0, 1,
                   0, 0, "2", 0, 1,
                   1, 1, "-", "?", 0,
                   1, 1, "2", 1, "{01}"),
                   nrow = 6, ncol = 5, byrow = TRUE,
                   dimnames = list(
                     paste0("Taxon_", LETTERS[1:6]),
                     paste0("Char_", 1:5)))

dataset <- MatrixToPhyDat(tokens)
Hamming(dataset)
}
\seealso{
Used to construct neighbour joining trees in \code{\link[=NJTree]{NJTree()}}.

\code{dist.hamming()} in the \pkg{phangorn} package provides an alternative
implementation.

Other utility functions:
\code{\link{ClusterTable}},
\code{\link{ClusterTable-methods}},
\code{\link[=MSTEdges]{MSTEdges()}},
\code{\link[=SampleOne]{SampleOne()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link[=UnshiftTree]{UnshiftTree()}},
\code{\link[=as.multiPhylo]{as.multiPhylo()}},
\code{\link{match,phylo,phylo-method}},
\code{\link[=sapply64]{sapply64()}},
\code{\link[=sort.multiPhylo]{sort.multiPhylo()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_information.R
\name{CladisticInfo}
\alias{CladisticInfo}
\alias{CladisticInfo.phylo}
\alias{CladisticInfo.Splits}
\alias{CladisticInfo.list}
\alias{CladisticInfo.multiPhylo}
\alias{CladisticInformation}
\title{Cladistic information content of a tree}
\usage{
CladisticInfo(x)

\method{CladisticInfo}{phylo}(x)

\method{CladisticInfo}{Splits}(x)

\method{CladisticInfo}{list}(x)

\method{CladisticInfo}{multiPhylo}(x)

CladisticInformation(x)
}
\arguments{
\item{x}{Tree of class \code{phylo}, or a list thereof.}
}
\value{
\code{CladisticInfo()} returns a numeric giving the cladistic information
content of the input tree(s), in bits.
If passed a \code{Splits} object, it returns the information content of each
split in turn.
}
\description{
\code{CladisticInfo()} calculates the cladistic (phylogenetic) information
content of a phylogenetic object, \emph{sensu} Thorley \emph{et al.} (1998).
}
\details{
The \acronym{CIC} is the logarithm of the number of binary trees that
include the specified topology.
A base two logarithm gives an information content in bits.

The \acronym{CIC} was originally proposed by
\insertCite{Rohlf1982;textual}{TreeTools}, and formalised,
with an information-theoretic justification, by
\insertCite{Thorley1998;textual}{TreeTools}.
\insertCite{Steel2006;textual}{TreeTools} term the equivalent quantity
"phylogenetic information content" in the context of individual characters.

The number of binary trees consistent with a cladogram provides a more
satisfactory measure of the resolution of a tree than simply
counting the number of edges resolved \insertCite{Page1992}{TreeTools}.
}
\references{
\insertAllCited{}
}
\seealso{
Other tree information functions:
\code{\link[=NRooted]{NRooted()}},
\code{\link[=TreesMatchingTree]{TreesMatchingTree()}}

Other tree characterization functions:
\code{\link[=Consensus]{Consensus()}},
\code{\link[=J1Index]{J1Index()}},
\code{\link{Stemwardness}},
\code{\link[=TotalCopheneticIndex]{TotalCopheneticIndex()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree characterization functions}
\concept{tree information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_generation.R
\name{GenerateTree}
\alias{GenerateTree}
\alias{RandomTree}
\alias{YuleTree}
\alias{PectinateTree}
\alias{BalancedTree}
\alias{StarTree}
\title{Generate pectinate, balanced or random trees}
\usage{
RandomTree(tips, root = FALSE, nodes, lengths = NULL)

YuleTree(tips, addInTurn = FALSE, root = TRUE, lengths = NULL)

PectinateTree(tips, lengths = NULL)

BalancedTree(tips, lengths = NULL)

StarTree(tips, lengths = NULL)
}
\arguments{
\item{tips}{An integer specifying the number of tips, or a character vector
naming the tips, or any other object from which \code{\link[=TipLabels]{TipLabels()}} can
extract leaf labels.}

\item{root}{Character or integer specifying tip to use as root;
or \code{TRUE} to root the tree on a random edge;
or \code{FALSE} to return an unrooted tree.}

\item{nodes}{Number of nodes to generate.  The default and maximum,
\code{tips - 1}, generates a binary tree; setting a lower value will induce
polytomies.}

\item{lengths}{a numeric vector specifying the edge lengths of the tree.}

\item{addInTurn}{Logical specifying whether to add leaves in the order of
\code{tips}.  If \code{FALSE}, leaves will be added in a random order.}
}
\value{
Each function returns an unweighted binary tree of class \code{phylo} with
the specified leaf labels. Trees are rooted unless \code{root = FALSE}.

\code{RandomTree()} returns a topology drawn at random from the uniform
distribution (i.e. each binary tree is drawn with equal probability).
Trees are generated by inserting
each tip in term at a randomly selected edge in the tree.
Random numbers are generated using a Mersenne Twister.
If \code{root = FALSE}, the tree will be unrooted, with the first tip in a
basal position.  Otherwise, the tree will be rooted on \code{root}.

\code{YuleTree()} returns a topology generated by the Yule process
\insertCite{Steel2001}{TreeTools},
i.e. adding leaves in turn adjacent to a randomly-chosen existing leaf.

\code{PectinateTree()} returns a pectinate (caterpillar) tree.

\code{BalancedTree()} returns a balanced (symmetrical) tree, in preorder.

\code{StarTree()} returns a completely unresolved (star) tree.
}
\description{
\code{RandomTree()}, \code{PectinateTree()}, \code{BalancedTree()} and \code{StarTree()}
generate trees with the specified shapes and leaf labels.
}
\examples{
RandomTree(LETTERS[1:10])

data("Lobo")
RandomTree(Lobo.phy)

YuleTree(LETTERS[1:10])

plot(PectinateTree(LETTERS[1:10]))

plot(BalancedTree(LETTERS[1:10]))
plot(StarTree(LETTERS[1:10]))

}
\references{
\insertAllCited()
}
\seealso{
Other tree generation functions: 
\code{\link{ConstrainedNJ}()},
\code{\link{NJTree}()},
\code{\link{TreeNumber}},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree generation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_properties.R
\name{RootNode}
\alias{RootNode}
\title{Which node is a tree's root?}
\usage{
RootNode(x)
}
\arguments{
\item{x}{A tree of class \code{phylo}, or its edge matrix; or a list or
\code{multiPhylo} object containing multiple trees.}
}
\value{
\code{RootNode()} returns an integer denoting the root node for each tree.
Badly conformed trees trigger an error.
}
\description{
\code{RootNode()} identifies the root node of a (rooted or unrooted) phylogenetic
tree.
Unrooted trees are represented internally by a rooted tree with a polytomy
at the root.
}
\examples{
RootNode(BalancedTree(8))
RootNode(UnrootTree(BalancedTree(8)))


}
\seealso{
Test whether a tree is rooted: \code{\link[=TreeIsRooted]{TreeIsRooted()}}

\code{phangorn::getRoot()}

Other tree navigation:
\code{\link[=AncestorEdge]{AncestorEdge()}},
\code{\link[=CladeSizes]{CladeSizes()}},
\code{\link[=DescendantEdges]{DescendantEdges()}},
\code{\link[=EdgeAncestry]{EdgeAncestry()}},
\code{\link[=EdgeDistances]{EdgeDistances()}},
\code{\link[=ListAncestors]{ListAncestors()}},
\code{\link[=MRCA]{MRCA()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NDescendants]{NDescendants()}},
\code{\link[=NodeDepth]{NodeDepth()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=NodeOrder]{NodeOrder()}},
\code{\link[=PaintTree]{PaintTree()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_ancestors.R
\name{AncestorEdge}
\alias{AncestorEdge}
\title{Ancestral edge}
\usage{
AncestorEdge(edge, parent, child)
}
\arguments{
\item{edge}{Number of an edge}

\item{parent}{Integer vector corresponding to the first column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 1]}}

\item{child}{Integer vector corresponding to the second column of the edge
matrix of a tree of class \code{\link[ape]{phylo}}, i.e.
\code{tree[["edge"]][, 2]}.}
}
\value{
\code{AncestorEdge} returns a logical vector identifying whether each edge
is the immediate ancestor of the given edge.
}
\description{
Ancestral edge
}
\examples{
tree <- BalancedTree(6)
parent <- tree$edge[, 1]
child <- tree$edge[, 2]
plot(tree)
ape::edgelabels()
AncestorEdge(5, parent, child)
which(AncestorEdge(5, parent, child))

}
\seealso{
Other tree navigation: 
\code{\link{CladeSizes}()},
\code{\link{DescendantEdges}()},
\code{\link{EdgeAncestry}()},
\code{\link{EdgeDistances}()},
\code{\link{ListAncestors}()},
\code{\link{MRCA}()},
\code{\link{MatchEdges}()},
\code{\link{NDescendants}()},
\code{\link{NodeDepth}()},
\code{\link{NodeNumbers}()},
\code{\link{NodeOrder}()},
\code{\link{RootNode}()}
}
\concept{tree navigation}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_properties.R
\name{NTip}
\alias{NTip}
\alias{NTip.default}
\alias{NTip.Splits}
\alias{NTip.list}
\alias{NTip.phylo}
\alias{NTip.multiPhylo}
\alias{NTip.phyDat}
\alias{NTip.matrix}
\title{Number of leaves in a phylogenetic tree}
\usage{
NTip(phy)

\method{NTip}{default}(phy)

\method{NTip}{Splits}(phy)

\method{NTip}{list}(phy)

\method{NTip}{phylo}(phy)

\method{NTip}{multiPhylo}(phy)

\method{NTip}{phyDat}(phy)

\method{NTip}{matrix}(phy)
}
\arguments{
\item{phy}{Object representing one or more phylogenetic trees.}
}
\value{
\code{NTip()} returns an integer specifying the number of tips in each
object in \code{phy}.
}
\description{
\code{NTip()} extends \code{\link[ape:summary.phylo]{ape::Ntip()}} to handle
objects of class \code{Splits} and \code{list}, and edge matrices
(equivalent to \code{tree$edge}).
}
\seealso{
Other tree properties:
\code{\link[=Cherries]{Cherries()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=EdgeRatio]{EdgeRatio()}},
\code{\link[=LongBranch]{LongBranch()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=PathLengths]{PathLengths()}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TreeIsRooted]{TreeIsRooted()}},
\code{\link[=Treeness]{Treeness()}}

Other Splits operations:
\code{\link[=LabelSplits]{LabelSplits()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=PolarizeSplits]{PolarizeSplits()}},
\code{\link[=SplitFrequency]{SplitFrequency()}},
\code{\link{Splits}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TipsInSplits]{TipsInSplits()}},
\code{\link{match,Splits,Splits-method}},
\code{\link[=xor]{xor()}}
}
\concept{Splits operations}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Consensus.R
\name{Consensus}
\alias{Consensus}
\title{Construct consensus trees}
\usage{
Consensus(trees, p = 1, check.labels = TRUE, hash = TRUE)
}
\arguments{
\item{trees}{List of trees, optionally of class \code{multiPhylo}.}

\item{p}{Proportion of trees that must contain a split for it to be reported
in the consensus.  \code{p = 0.5} gives the majority-rule consensus; \code{p = 1} (the
default) gives the strict consensus.}

\item{check.labels}{Logical specifying whether to check that all trees have
identical labels.  Defaults to \code{TRUE}, which is slower.}

\item{hash}{Logical; if \code{TRUE} (default), majority/threshold consensus
counts splits using 128-bit hashing, which is exact with overwhelming
probability (a collision conflating two distinct splits is vanishingly
unlikely).  Set \code{hash = FALSE} for a slower but guaranteed-exact count.
Ignored when \code{p = 1}, which is always exact.}
}
\value{
\code{Consensus()} returns an object of class \code{phylo}, rooted as in the
first entry of \code{trees}.
}
\description{
\code{Consensus()} calculates the majority-rule or strict consensus of a set of
trees, using the cluster-table approach of \insertCite{Day1985}{TreeTools}.
}
\details{
The strict consensus (\code{p = 1}) compares the clusters of the first tree
against every other tree in linear time.  The majority-rule and threshold
consensus (\verb{0.5 <= p < 1}) instead count the frequency of every split across
all trees in a single pass and retain those occurring in a proportion \code{p} or
more of trees; this runs in time linear in the number of trees, after
\insertCite{Jansson2016}{TreeTools}.  By default the count uses
a 128-bit hash, whose results are exact with overwhelming probability; set
\code{hash = FALSE} for a slower but guaranteed-exact count.
}
\examples{
Consensus(as.phylo(0:2, 8))
}
\references{
\insertAllCited{}
}
\seealso{
\itemize{
\item \href{https://constree.github.io/}{\pkg{ConsTree}} implements
other consensus tree algorithms.
\item \href{https://ms609.github.io/Rogue/}{\pkg{Rogue}} increases the resolution of
consensus trees by dropping wildcard taxa.
\item \code{TreeDist::ConsensusInfo()} calculates the information content of a
consensus tree.
}

Other consensus tree functions:
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=RoguePlot]{RoguePlot()}}

Other tree characterization functions:
\code{\link[=CladisticInfo]{CladisticInfo()}},
\code{\link[=J1Index]{J1Index()}},
\code{\link{Stemwardness}},
\code{\link[=TotalCopheneticIndex]{TotalCopheneticIndex()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{consensus tree functions}
\concept{tree characterization functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_numbering.R
\name{RenumberTips}
\alias{RenumberTips}
\alias{RenumberTips.phylo}
\alias{RenumberTips.Splits}
\alias{RenumberTips.multiPhylo}
\alias{RenumberTips.list}
\alias{RenumberTips.NULL}
\title{Renumber a tree's tips}
\usage{
RenumberTips(tree, tipOrder)

\method{RenumberTips}{phylo}(tree, tipOrder)

\method{RenumberTips}{Splits}(tree, tipOrder)

\method{RenumberTips}{multiPhylo}(tree, tipOrder)

\method{RenumberTips}{list}(tree, tipOrder)

\method{RenumberTips}{`NULL`}(tree, tipOrder)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{tipOrder}{A character vector containing the values of
\code{tree[["tip.label"]]} in the desired sort order,
an object (perhaps of class \code{phylo} or \code{Splits}) with tip labels,
or a numeric specifying the desired order.}
}
\value{
\code{RenumberTips()} returns \code{tree}, with the tips' internal
representation numbered to match \code{tipOrder}.
}
\description{
\code{RenumberTips(tree, tipOrder)} sorts the tips of a phylogenetic tree \code{tree}
such that the indices in \code{tree[["edge"]][, 2]} correspond to the order of
leaves given in \code{tipOrder}.
}
\examples{
data("Lobo") # Loads the phyDat object Lobo.phy
tree <- RandomTree(Lobo.phy)
tree <- RenumberTips(tree, names(Lobo.phy))

}
\seealso{
Other tree manipulation:
\code{\link[=AddTip]{AddTip()}},
\code{\link[=CollapseNode]{CollapseNode()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=DropTip]{DropTip()}},
\code{\link[=ImposeConstraint]{ImposeConstraint()}},
\code{\link[=KeptPaths]{KeptPaths()}},
\code{\link[=KeptVerts]{KeptVerts()}},
\code{\link[=LeafLabelInterchange]{LeafLabelInterchange()}},
\code{\link[=MakeTreeBinary]{MakeTreeBinary()}},
\code{\link[=Renumber]{Renumber()}},
\code{\link[=RenumberTree]{RenumberTree()}},
\code{\link[=RootTree]{RootTree()}},
\code{\link[=SortTree]{SortTree()}},
\code{\link[=Subtree]{Subtree()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ConsistentSplits.R
\name{SplitConsistent}
\alias{SplitConsistent}
\alias{SplitConflicts}
\title{Identify consistent / conflicting splits}
\usage{
SplitConsistent(needle, haystack)

SplitConflicts(needle, haystack)
}
\arguments{
\item{needle}{Splits object containing the single split to evaluate}

\item{haystack}{Splits object, or list thereof, containing the splits to
compare against \code{needle}.}
}
\value{
\code{SplitConsistent()} returns a list of logical vectors.
Each list item corresponds to an entry in \code{haystack}, reporting whether each
split is consistent with (\code{TRUE}) or in conflict with (\code{FALSE}) \code{needle}.
\code{SplitConflicts()} returns the inverse.
}
\description{
\code{SplitConsistent()} and \code{SplitConflict()} determine whether a series of
splits \code{haystack} are consistent with or contradict the focal split \code{needle}.
}
\examples{
splits1 <- as.Splits(BalancedTree(8))
splits2 <- as.Splits(PectinateTree(8))
summary(splits1[[4]])
SplitConsistent(splits1[[4]], splits2)
SplitConflicts(splits1[[4]], list(splits1, splits2))
}
\seealso{
Other split manipulation functions:
\code{\link[=DropTip]{DropTip()}},
\code{\link[=Subsplit]{Subsplit()}},
\code{\link[=TrivialSplits]{TrivialSplits()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split manipulation functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Information.R
\name{SplitInformation}
\alias{SplitInformation}
\alias{SplitInformation.numeric}
\alias{SplitInformation.Splits}
\alias{SplitInformation.phylo}
\alias{MultiSplitInformation}
\title{Phylogenetic information content of splitting leaves into two partitions}
\usage{
SplitInformation(A, B = A[1])

\method{SplitInformation}{numeric}(A, B = A[1])

\method{SplitInformation}{Splits}(A, B)

\method{SplitInformation}{phylo}(A, B)

MultiSplitInformation(partitionSizes)
}
\arguments{
\item{A, B}{Integer specifying the number of taxa in each partition.}

\item{partitionSizes}{Integer vector specifying the number of taxa in each
partition of a multi-partition split.}
}
\value{
\code{SplitInformation()} and \code{MultiSplitInformation()} return the
phylogenetic information content, in bits, of a split that subdivides leaves
into partitions of the specified sizes.
}
\description{
Calculate the phylogenetic information content (\emph{sensu}
\insertCite{Steel2006;nobrackets}{TreeTools}) of a split, which
reflects the probability that a uniformly selected random tree will contain
the split: a split that is consistent with a smaller number of trees will
have a higher information content.
}
\details{
\code{SplitInformation()} addresses bipartition splits, which correspond to
edges in an unrooted phylogeny; \code{MultiSplitInformation()} supports splits
that subdivide taxa into multiple partitions, which may correspond to
multi-state characters in a phylogenetic matrix.

A simple way to characterise trees is to count the number of edges.
(Edges are almost, but not quite, equivalent to nodes.)
Counting edges (or nodes) provides a quick measure of a tree's resolution,
and underpins the Robinson-Foulds tree distance measure.
Not all edges, however, are created equal.

An edge splits the leaves of a tree into two subdivisions.  The more equal
these subdivisions are in size, the more instructive this edge is.
Intuitively, the division of mammals from reptiles is a profound revelation
that underpins much of zoology; recognizing that two species of bat are more
closely related to each other than to any other mammal or reptile is still
instructive, but somewhat less fundamental.

Formally, the phylogenetic (Shannon) information content of a split \emph{S},
\emph{h(S)}, corresponds to the probability that a uniformly selected random tree
will contain the split, \emph{P(S)}: \emph{h(S)} = -log \emph{P(S)}.
Base 2 logarithms are typically employed to yield an information content in
bits.

As an example, the split \code{AB|CDEF} occurs in 15 of the 105 six-leaf trees;
\emph{h}(\code{AB|CDEF}) = -log \emph{P}(\code{AB|CDEF}) = -log(15/105) ~ 2.81 bits.  The split
\code{ABC|DEF} subdivides the leaves more evenly, and is thus more instructive:
it occurs in just nine of the 105 six-leaf trees, and
\emph{h}(\code{ABC|DEF}) = -log(9/105) ~ 3.54 bits.

As the number of leaves increases, a single even split may contain more
information than multiple uneven splits -- see the examples section below.

Summing the information content of all splits within a tree, perhaps using
the '\href{https://ms609.github.io/TreeDist/}{TreeDist}' function
\href{https://ms609.github.io/TreeDist/reference/TreeInfo.html}{\code{SplitwiseInfo()}},
arguably gives a more instructive picture of its resolution than simply
counting the number of splits that are present -- though with the caveat
that splits within a tree are not independent of one another, so some
information may be double counted.  (This same charge applies to simply
counting nodes, too.)

Alternatives would be to count the number of quartets that are resolved,
perhaps using the '\href{https://ms609.github.io/Quartet/}{Quartet}' function
\href{https://ms609.github.io/Quartet/reference/QuartetState.html}{\code{QuartetStates()}},
or to use a different take on the information contained within a split, the
clustering information: see the 'TreeDist' function
\href{https://ms609.github.io/TreeDist/reference/TreeInfo.html}{\code{ClusteringInfo()}}
for details.
}
\examples{
# Eight leaves can be split evenly:
SplitInformation(4, 4)

# or unevenly, which is less informative:
SplitInformation(2, 6)

# A single split that evenly subdivides 50 leaves contains more information
# that seven maximally uneven splits on the same leaves:
SplitInformation(25, 25)
7 * SplitInformation(2, 48)
# Three ways to split eight leaves into multiple partitions:
MultiSplitInformation(c(2, 2, 4))
MultiSplitInformation(c(2, 3, 3))
MultiSplitInformation(rep(2, 4))


}
\references{
\insertAllCited{}
}
\seealso{
Sum the phylogenetic information content of splits within a tree:
\href{https://ms609.github.io/TreeDist/reference/TreeInfo.html}{\code{TreeDist::SplitwiseInfo()}}

Sum the clustering information content of splits within a tree:
\href{https://ms609.github.io/TreeDist/reference/TreeInfo.html}{\code{TreeDist::ClusteringInfo()}}

Other split information functions: 
\code{\link{CharacterInformation}()},
\code{\link{SplitMatchProbability}()},
\code{\link{TreesMatchingSplit}()},
\code{\link{UnrootedTreesMatchingSplit}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{doubleFactorials}
\alias{doubleFactorials}
\title{Double factorials}
\format{
An object of class \code{numeric} of length 300.
}
\usage{
doubleFactorials
}
\description{
A vector with pre-calculated values of double factorials up to 300!!,
and the logarithms of double factorials up to 50 000!!.
}
\details{
301!! is too large to store as an integer; use \code{logDoubleFactorials} instead.
}
\seealso{
Other double factorials:
\code{\link[=DoubleFactorial]{DoubleFactorial()}},
\code{\link{logDoubleFactorials}}
}
\concept{double factorials}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TotalCopheneticIndex.R
\encoding{UTF-8}
\name{TotalCopheneticIndex}
\alias{TotalCopheneticIndex}
\alias{TCIContext}
\alias{TCIContext.numeric}
\title{Total Cophenetic Index}
\usage{
TotalCopheneticIndex(x)

TCIContext(x)

\method{TCIContext}{numeric}(x)
}
\arguments{
\item{x}{A tree of class \code{phylo}, its \verb{$edge} property, or a list thereof.}
}
\value{
\code{TotalCopheneticIndex()} returns an integer denoting the total cophenetic index.

\code{TCIContext()} returns a data frame detailing the maximum and minimum value
obtainable for the Total Cophenetic Index for rooted binary trees with the
number of leaves of the given tree, and the expected value under the Yule
and Uniform models.
The variance of the expected value is given under the Yule model, but cannot
be obtained by calculation for the Uniform model.
}
\description{
\code{TotalCopheneticIndex()} calculates the total cophenetic index
\insertCite{Mir2013}{TreeTools} for any tree, a measure of its balance;
\code{TCIContext()} lists its possible values.
}
\details{
The Total Cophenetic Index is a measure of tree balance -- i.e. whether
a (phylogenetic) tree comprises symmetric pairs of nodes, or has a pectinate
"caterpillar" shape.
The index has a greater resolution power than Sackin's and Colless' indices,
and can be applied to trees that are not perfectly resolved.

For a tree with \emph{n} leaves, the Total Cophenetic Index can take values of
0 to \code{choose(n, 3)}.
The minimum value is higher for a perfectly resolved (i.e. dichotomous) tree
(see Lemma 14 of Mir \emph{et al.} 2013).
Formulae to calculate the expected values under the Yule and Uniform models
of evolution are given in Theorems 17 and 23.

Full details are provided by \insertCite{Mir2013;textual}{TreeTools}.

The \ifelse{html}{\out{J<sup>1</sup>}}{\eqn{J^1}} index
\insertCite{Lemant2022}{TreeTools} has advantages over the Total Cophenetic
Index, particularly when comparing trees with different numbers of leaves,
or where the population size of nodes is meaningful; see \code{\link[=J1Index]{J1Index()}}.
}
\examples{
# Balanced trees have the minimum index for a binary tree;
# Pectinate trees the maximum:
TCIContext(8)
TotalCopheneticIndex(PectinateTree(8))
TotalCopheneticIndex(BalancedTree(8))
TotalCopheneticIndex(StarTree(8))


# Examples from Mir et al. (2013):
tree12 <- ape::read.tree(text="(1, (2, (3, (4, 5))));")  #Fig. 4, tree 12
TotalCopheneticIndex(tree12) # 10
tree8  <- ape::read.tree(text="((1, 2, 3, 4), 5);")      #Fig. 4, tree 8
TotalCopheneticIndex(tree8)  # 6
TCIContext(tree8)
TCIContext(5L) # Context for a tree with 5 leaves.

}
\references{
\insertAllCited{}
}
\seealso{
\itemize{
\item \code{\link[=J1Index]{J1Index()}} provides a more robust, universal tree balance index.
\item \code{cophen.index()} in the package
\href{https://github.com/LuciaRotger/CollessLike}{\pkg{CollessLike}}
provides an alternative implementation of this index and its predecessors.
}

Other tree characterization functions:
\code{\link[=CladisticInfo]{CladisticInfo()}},
\code{\link[=Consensus]{Consensus()}},
\code{\link[=J1Index]{J1Index()}},
\code{\link{Stemwardness}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree characterization functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Stemwardness.R
\name{Stemwardness}
\alias{Stemwardness}
\alias{SisterSize}
\alias{SisterSize.numeric}
\alias{SisterSize.character}
\alias{RootNodeDistance}
\alias{RootNodeDistance.numeric}
\alias{RootNodeDistance.character}
\alias{RootNodeDist}
\title{"Stemwardness" of a leaf}
\usage{
SisterSize(tree, tip)

\method{SisterSize}{numeric}(tree, tip)

\method{SisterSize}{character}(tree, tip)

RootNodeDistance(tree, tip)

\method{RootNodeDistance}{numeric}(tree, tip)

\method{RootNodeDistance}{character}(tree, tip)

RootNodeDist(tree, tip)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}

\item{tip}{Either a numeric specifying the index of a single tip, or a
character specifying its label.}
}
\value{
\code{SisterSize()} returns an integer specifying the number of leaves
in the clade that is sister to \code{tip}.
\code{RootNodeDist()} returns an integer specifying the number of nodes between
\code{tip} and the root node of \code{tree}.
}
\description{
Functions to describe the position of a leaf relative to the root.
"Stemmier" leaves ought to exhibit a smaller root-node distance and a
larger sister size.
}
\details{
\code{RootNodeDistance()} calculates the number of nodes between the chosen leaf
and the root of \code{tree}.
This is an unsatisfactory measure, as the range of possible
distances is a function of the shape of the tree
\insertCite{Asher2020}{TreeTools}.
As an example, leaf \emph{X1} in the tree \verb{(.,(.,(.,(.,(X1,(a,b))))))}
falls outside the clade \emph{(a, b)} and has a root-node distance of 4,
whereas leaf \emph{X2} in the tree \verb{(.,((.,(.,.)),(b,(X2,a))))}
falls within the clade \emph{(a, b)}, so should be considered more "crownwards",
yet has a smaller root-node distance (3).

\insertFig{Stemwardness.png}{TreeTools}{
par(mfrow = c(1, 2), mar = rep(0.3, 4))
plot(ape::read.tree(text="(.,(.,(.,(.,(X1,(a,b))))));"))
ape::nodelabels(1:4, 9:12)
ape::edgelabels(1:2, 11:12)

plot(ape::read.tree(text="(.,((.,(.,.)),(b,(X2,a))));"))
ape::nodelabels(1:3, c(9, 12, 13))
ape::edgelabels(1, 12)
}

\code{SisterSize()} measures the number of leaves in the clade that is sister to
the chosen leaf, as proposed by \insertCite{Asher2020;textual}{TreeTools}.
In the examples above, \emph{X1} has a sister size of 2 leaves, whereas \emph{X2},
which is "more crownwards", has a smaller sister size (1 leaf), as desired.
}
\examples{
bal8 <- BalancedTree(8)
pec8 <- PectinateTree(8)

SisterSize(bal8, 3)
SisterSize(pec8, "t3")
SisterSize(RootTree(pec8, "t3"), "t3")

RootNodeDist(bal8, 3)
RootNodeDist(pec8, "t3")
RootNodeDist(RootTree(pec8, "t3"), "t3")
}
\references{
\insertAllCited{}
}
\seealso{
Other tree characterization functions:
\code{\link[=CladisticInfo]{CladisticInfo()}},
\code{\link[=Consensus]{Consensus()}},
\code{\link[=J1Index]{J1Index()}},
\code{\link[=TotalCopheneticIndex]{TotalCopheneticIndex()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree characterization functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Information.R
\name{CharacterInformation}
\alias{CharacterInformation}
\title{Character information content}
\usage{
CharacterInformation(tokens)
}
\arguments{
\item{tokens}{Character vector specifying the tokens assigned to each taxon for
a character.  Example: \code{c(0, 0, 0, 1, 1, 1, "?", "-")}.

Note that ambiguous tokens such as \code{(01)} are not supported, and should be
replaced with \verb{?}.}
}
\value{
\code{CharacterInformation()} returns a numeric specifying the
phylogenetic information content of the character
(\emph{sensu} \insertCite{Steel2006;nobrackets}{TreeTools}), in bits.
}
\description{
\code{CharacterInformation()} calculates the cladistic information content
\insertCite{Steel2006}{TreeTools} of a given character, in bits.
The total information in all characters gives a measure of the potential
utility of a dataset \insertCite{Cotton2008}{TreeTools}, which can be
compared with a profile parsimony score \insertCite{Faith2001}{TreeTools} to
evaluate the degree of homoplasy within a dataset.
}
\references{
\insertAllCited{}
}
\seealso{
Other split information functions: 
\code{\link{SplitInformation}()},
\code{\link{SplitMatchProbability}()},
\code{\link{TreesMatchingSplit}()},
\code{\link{UnrootedTreesMatchingSplit}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{split information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Combinatorics.R
\name{N1Spr}
\alias{N1Spr}
\alias{IC1Spr}
\title{Number of trees one SPR step away}
\usage{
N1Spr(n)

IC1Spr(n)
}
\arguments{
\item{n}{Integer vector specifying the number of tips in a tree.}
}
\value{
\code{N1Spr()} returns an integer vector denoting the number of trees one
SPR rearrangement away from the input tree..

\code{IC1Spr()} returns an numeric vector giving the phylogenetic
information content of trees 0 or 1 SPR rearrangement from an \emph{n}-leaf tree,
in bits.
}
\description{
\code{N1Spr()} calculates the number of trees one subtree prune-and-regraft
operation away from a binary input tree using the formula given by
\insertCite{Allen2001;textual}{TreeTools};
\code{IC1Spr()} calculates the information content of trees at this
distance: i.e. the entropy corresponding to the proportion of all possible
\emph{n}-tip trees whose SPR distance is at most one from a specified tree.
}
\examples{
N1Spr(4:6)
IC1Spr(5)

}
\references{
\insertAllCited{}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Cherries.R
\name{Cherries}
\alias{Cherries}
\alias{Cherries.phylo}
\alias{Cherries.numeric}
\title{Count cherries in a tree}
\usage{
Cherries(tree, nTip)

\method{Cherries}{phylo}(tree, nTip = NTip(tree))

\method{Cherries}{numeric}(tree, nTip)
}
\arguments{
\item{tree}{A binary tree, of class \code{phylo}; or a matrix corresponding to its
edge matrix.}

\item{nTip}{Number of leaves in tree.}
}
\value{
\code{Cherries()} returns an integer specifying the number of nodes whose
children are both leaves.

Equations for the number of cherries expected under uniform and Yule tree
models are derived by \insertCite{McKenzie2000;textual}{TreeTools}.
}
\description{
\code{Cherries()} counts the number of vertices in a binary tree whose children
are both leaves.
}
\references{
\insertAllCited{}
}
\seealso{
Other tree properties:
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=EdgeRatio]{EdgeRatio()}},
\code{\link[=LongBranch]{LongBranch()}},
\code{\link[=MatchEdges]{MatchEdges()}},
\code{\link[=NSplits]{NSplits()}},
\code{\link[=NTip]{NTip()}},
\code{\link[=NodeNumbers]{NodeNumbers()}},
\code{\link[=PathLengths]{PathLengths()}},
\code{\link[=SplitsInBinaryTree]{SplitsInBinaryTree()}},
\code{\link[=TipLabels]{TipLabels()}},
\code{\link[=TreeIsRooted]{TreeIsRooted()}},
\code{\link[=Treeness]{Treeness()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{Unquote}
\alias{Unquote}
\title{Remove quotation marks from a string}
\usage{
Unquote(string)
}
\arguments{
\item{string}{Input string}
}
\value{
\code{Unquote()} returns \code{string}, with any matched punctuation marks
and trailing whitespace removed.
}
\description{
Remove quotation marks from a string
}
\examples{
Unquote("'Hello World'")
}
\seealso{
Other string parsing functions:
\code{\link[=EndSentence]{EndSentence()}},
\code{\link[=MatchStrings]{MatchStrings()}},
\code{\link[=MorphoBankDecode]{MorphoBankDecode()}},
\code{\link[=RightmostCharacter]{RightmostCharacter()}}
}
\author{
Martin R. Smith
}
\concept{string parsing functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{MorphoBankDecode}
\alias{MorphoBankDecode}
\title{Decode MorphoBank text}
\usage{
MorphoBankDecode(string)
}
\arguments{
\item{string}{String to process}
}
\value{
\code{MorphoBankDecode()} returns a string with new lines and punctuation
reformatted.
}
\description{
Converts strings from MorphoBank notes into a Latex-compatible format.
}
\seealso{
Other string parsing functions: 
\code{\link{EndSentence}()},
\code{\link{MatchStrings}()},
\code{\link{RightmostCharacter}()},
\code{\link{Unquote}()}
}
\author{
Martin R. Smith
}
\concept{string parsing functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match.R
\name{match,Splits,Splits-method}
\alias{match,Splits,Splits-method}
\alias{match}
\alias{\%in\%,Splits,Splits-method}
\alias{FirstMatchingSplit}
\title{Split matching}
\usage{
\S4method{match}{Splits,Splits}(x, table, nomatch = NA_integer_, incomparables = NULL)

match(x, table, nomatch = NA_integer_, incomparables = NULL)

\S4method{\%in\%}{Splits,Splits}(x, table)

FirstMatchingSplit(x, table, nomatch, return = c("x", "table", "both"))
}
\arguments{
\item{x, table}{Splits objects}

\item{nomatch}{Integer value that will be used in place of \code{NA} in the case
where no match is found.}

\item{incomparables}{Ignored. (Included for consistency with generic.)}

\item{return}{Which index to return: in \code{x}, in \code{table}, or both}
}
\value{
\code{match()} returns an integer vector specifying the position in
\code{table} that matches each element in \code{x}, or \code{nomatch} if no match is found.

\code{FirstMatchingSplit()} returns an integer
(or length-2 integer if \code{return = "both"}) specifying the first split in \code{x}
to have a match in \code{table} (\code{return = "x"}),
or the index of that match (\code{return = "table"}).
\code{nomatch} (default \code{0}) is returned in the absence of a match.
}
\description{
\code{match()} returns a vector of the positions of (first) matches of splits in
its first argument in its second.
\code{\%in\%} is a more intuitive interface as a binary operator, which returns
a logical vector indicating whether there is a match or not for each
split in its left operand.
}
\examples{
splits1 <- as.Splits(BalancedTree(7))
splits2 <- as.Splits(PectinateTree(7))

match(splits1, splits2)
}
\seealso{
Corresponding base functions are documented in
\code{\link[base:match]{match()}}.

Other Splits operations: 
\code{\link{LabelSplits}()},
\code{\link{NSplits}()},
\code{\link{NTip}()},
\code{\link{PolarizeSplits}()},
\code{\link{SplitFrequency}()},
\code{\link{Splits}},
\code{\link{SplitsInBinaryTree}()},
\code{\link{TipLabels}()},
\code{\link{TipsInSplits}()},
\code{\link{xor}()}
}
\concept{Splits operations}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{ApeTime}
\alias{ApeTime}
\title{Read modification time from "ape" Nexus file}
\usage{
ApeTime(filepath, format = "double")
}
\arguments{
\item{filepath}{Character string specifying path to the file.}

\item{format}{Format in which to return the time: "double" as a sortable numeric;
any other value to return a string in the format
\verb{YYYY-MM-DD hh:mm:ss}.}
}
\value{
\code{ApeTime()} returns the time that the specified file was created by
ape, in the format specified by \code{format}.
}
\description{
\code{ApeTime()} reads the time that a tree written with "ape" was modified,
based on the comment in the Nexus file.
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ClusterTable.R
\name{ClusterTable-methods}
\alias{ClusterTable-methods}
\alias{as.matrix.ClusterTable}
\alias{print.ClusterTable}
\alias{summary.ClusterTable}
\title{S3 methods for \code{ClusterTable} objects}
\usage{
\method{as.matrix}{ClusterTable}(x, ...)

\method{print}{ClusterTable}(x, ...)

\method{summary}{ClusterTable}(object, ...)
}
\arguments{
\item{x, object}{Object of class \code{ClusterTable}.}

\item{\dots}{Additional arguments for consistency with S3 methods.}
}
\description{
S3 methods for \code{\link{ClusterTable}} objects.
}
\examples{
clustab <- as.ClusterTable(TreeTools::BalancedTree(6))
as.matrix(clustab)
print(clustab)
summary(clustab)
}
\seealso{
Other utility functions:
\code{\link{ClusterTable}},
\code{\link[=Hamming]{Hamming()}},
\code{\link[=MSTEdges]{MSTEdges()}},
\code{\link[=SampleOne]{SampleOne()}},
\code{\link[=TipTimedTree]{TipTimedTree()}},
\code{\link[=UnshiftTree]{UnshiftTree()}},
\code{\link[=as.multiPhylo]{as.multiPhylo()}},
\code{\link{match,phylo,phylo-method}},
\code{\link[=sapply64]{sapply64()}},
\code{\link[=sort.multiPhylo]{sort.multiPhylo()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{utility functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/phylo.R
\name{TrivialTree}
\alias{TrivialTree}
\alias{SingleTaxonTree}
\alias{ZeroTaxonTree}
\title{Generate trivial trees}
\usage{
SingleTaxonTree(label = "t1", lengths = NULL)

ZeroTaxonTree()
}
\arguments{
\item{label}{a character vector specifying the label of the tip.}

\item{lengths}{Numeric vector of edge lengths, or a function that returns
such a vector when passed the number of edges as its argument (e.g. \code{runif}).}
}
\value{
\code{SingleTaxonTree()} returns a \code{phylo} object containing a single
tip with the specified label.

\code{ZeroTaxonTree()} returns an empty \code{phylo} object.
}
\description{
\code{SingleTaxonTree()} creates a phylogenetic "tree" that contains a single
taxon.
\code{ZeroTaxonTree()} creates an empty \code{phylo} object with zero leaves or edges.
}
\examples{
SingleTaxonTree("Homo_sapiens")
plot(SingleTaxonTree("root") + BalancedTree(4))

ZeroTaxonTree()
}
\seealso{
Other tree manipulation:
\code{\link[=AddTip]{AddTip()}},
\code{\link[=CollapseNode]{CollapseNode()}},
\code{\link[=ConsensusWithout]{ConsensusWithout()}},
\code{\link[=DropTip]{DropTip()}},
\code{\link[=ImposeConstraint]{ImposeConstraint()}},
\code{\link[=KeptPaths]{KeptPaths()}},
\code{\link[=KeptVerts]{KeptVerts()}},
\code{\link[=LeafLabelInterchange]{LeafLabelInterchange()}},
\code{\link[=MakeTreeBinary]{MakeTreeBinary()}},
\code{\link[=Renumber]{Renumber()}},
\code{\link[=RenumberTips]{RenumberTips()}},
\code{\link[=RenumberTree]{RenumberTree()}},
\code{\link[=RootTree]{RootTree()}},
\code{\link[=SortTree]{SortTree()}},
\code{\link[=Subtree]{Subtree()}},
\code{\link[=TipTimedTree]{TipTimedTree()}}

Other tree generation functions:
\code{\link[=ConstrainedNJ]{ConstrainedNJ()}},
\code{\link{GenerateTree}},
\code{\link[=NJTree]{NJTree()}},
\code{\link{TreeNumber}}
}
\concept{tree generation functions}
\concept{tree manipulation}
\keyword{tree}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{NewickTree}
\alias{NewickTree}
\title{Write Newick Tree}
\usage{
NewickTree(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}
}
\value{
\code{NewickTree()} returns a character string denoting \code{tree} in Newick
format.
}
\description{
\code{NewickTree()} encodes a tree as a Newick-format string.
This differs from \code{\link[ape:write.tree]{write.tree()}} in the encoding of
spaces as spaces, rather than underscores.
}
\examples{
NewickTree(BalancedTree(LETTERS[4:9]))

}
\seealso{
Use tip numbers, rather than leaf labels: \code{\link{as.Newick}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_rearrangement.R
\name{MakeTreeBinary}
\alias{MakeTreeBinary}
\title{Generate binary tree by collapsing polytomies}
\usage{
MakeTreeBinary(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}
}
\value{
\code{MakeTreeBinary()} returns a rooted binary tree of class \code{phylo},
corresponding to tree uniformly selected from all those compatible with
the input tree topologies.
}
\description{
\code{MakeTreeBinary()} resolves, at random, all polytomies in a tree or set of
trees, such that all trees compatible with the input topology are drawn
with equal probability. Edge lengths are not yet supported, so are removed.
}
\examples{
MakeTreeBinary(CollapseNode(PectinateTree(7), c(9, 11, 13)))
UnrootTree(MakeTreeBinary(StarTree(5)))
}
\seealso{
Since ape v5.5, this functionality is available through
\code{\link[ape:multi2di]{ape::multi2di()}}; previous versions of "ape" did not return topologies
in equal frequencies.  \code{MakeTreeBinary()} is often somewhat faster;
\code{multi2di()} retains edge lengths.

Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{CollapseNode}()},
\code{\link{ConsensusWithout}()},
\code{\link{DropTip}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree manipulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_ancestors.R
\name{MRCA}
\alias{MRCA}
\title{Most recent common ancestor}
\usage{
MRCA(x1, x2, ancestors)
}
\arguments{
\item{x1, x2}{Integer specifying index of leaves or nodes whose most
recent common ancestor should be found.}

\item{ancestors}{List of ancestors for each node in a tree. Perhaps
produced by \code{\link[=ListAncestors]{ListAncestors()}}.}
}
\value{
\code{MRCA()} returns an integer specifying the node number of the last
common ancestor of \code{x1} and \code{x2}.
}
\description{
\code{MRCA()} calculates the last common ancestor of specified nodes.
}
\details{
\code{MRCA()} requires that node values within a tree increase away from the root,
which will be true of trees listed in \code{Preorder}.
No warnings will be given if trees do not fulfil this requirement.
}
\examples{
tree <- BalancedTree(7)

# Verify that node numbering increases away from root
plot(tree)
nodelabels()

# ListAncestors expects a tree in Preorder
tree <- Preorder(tree)
edge <- tree$edge
ancestors <- ListAncestors(edge[, 1], edge[, 2])
MRCA(1, 4, ancestors)

# If a tree must be in postorder, use:
tree <- Postorder(tree)
edge <- tree$edge
ancestors <- lapply(seq_len(max(edge)), ListAncestors,
                    parent = edge[, 1], child = edge[, 2])

}
\seealso{
Other tree navigation: 
\code{\link{AncestorEdge}()},
\code{\link{CladeSizes}()},
\code{\link{DescendantEdges}()},
\code{\link{EdgeAncestry}()},
\code{\link{EdgeDistances}()},
\code{\link{ListAncestors}()},
\code{\link{MatchEdges}()},
\code{\link{NDescendants}()},
\code{\link{NodeDepth}()},
\code{\link{NodeNumbers}()},
\code{\link{NodeOrder}()},
\code{\link{RootNode}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree navigation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{RightmostCharacter}
\alias{RightmostCharacter}
\title{Rightmost character of string}
\usage{
RightmostCharacter(string, len = nchar(string))
}
\arguments{
\item{string}{Character string.}

\item{len}{(Optional) Integer specifying number of characters in \code{string}.}
}
\value{
\code{RightmostCharacter()} returns the rightmost character of a string.
}
\description{
\code{RightmostCharacter()} is a convenience function that returns the final
character of a string.
}
\examples{
RightmostCharacter("Hello, World!")

}
\seealso{
Other string parsing functions:
\code{\link[=EndSentence]{EndSentence()}},
\code{\link[=MatchStrings]{MatchStrings()}},
\code{\link[=MorphoBankDecode]{MorphoBankDecode()}},
\code{\link[=Unquote]{Unquote()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{string parsing functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Consensus.R
\name{ConsensusWithout}
\alias{ConsensusWithout}
\alias{ConsensusWithout.phylo}
\alias{ConsensusWithout.multiPhylo}
\alias{ConsensusWithout.list}
\alias{MarkMissing}
\title{Reduced consensus, omitting specified taxa}
\usage{
ConsensusWithout(trees, tip = character(0), ...)

\method{ConsensusWithout}{phylo}(trees, tip = character(0), ...)

\method{ConsensusWithout}{multiPhylo}(trees, tip = character(0), ...)

\method{ConsensusWithout}{list}(trees, tip = character(0), ...)

MarkMissing(tip, position = "bottomleft", ...)
}
\arguments{
\item{trees}{A list of phylogenetic trees, of class \code{multiPhylo} or \code{list}.}

\item{tip}{A character vector specifying the names (or numbers) of tips to
drop (using \code{ape::drop.tip()}).}

\item{\dots}{Additional parameters to pass on to \code{\link[ape:consensus]{ape::consensus()}} or
\code{\link[=legend]{legend()}}.}

\item{position}{Where to plot the missing taxa.
See \code{\link[=legend]{legend()}} for options.}
}
\value{
\code{ConsensusWithout()} returns a consensus tree (of class \code{phylo})
without the excluded taxa.

\code{MarkMissing()} provides a null return, after plotting the specified
\code{tip}s as a legend.
}
\description{
\code{ConsensusWithout()} displays a consensus plot with specified taxa excluded,
which can be a useful way to increase the resolution of a consensus tree
when a few wildcard taxa obscure a consistent set of relationships.
\code{MarkMissing()} adds missing taxa as loose leaves on the plot.
}
\examples{
oldPar <- par(mfrow = c(1, 2), mar = rep(0.5, 4))

# Two trees differing only in placement of tip 2:
trees <- as.phylo(c(0, 53), 6)
plot(trees[[1]])
plot(trees[[2]])

# Strict consensus (left panel) lacks resolution:
plot(ape::consensus(trees))

# But omitting tip two (right panel) reveals shared structure in common:
plot(ConsensusWithout(trees, "t2"))
MarkMissing("t2")

par(oldPar)
}
\seealso{
Other tree manipulation: 
\code{\link{AddTip}()},
\code{\link{CollapseNode}()},
\code{\link{DropTip}()},
\code{\link{ImposeConstraint}()},
\code{\link{KeptPaths}()},
\code{\link{KeptVerts}()},
\code{\link{LeafLabelInterchange}()},
\code{\link{MakeTreeBinary}()},
\code{\link{Renumber}()},
\code{\link{RenumberTips}()},
\code{\link{RenumberTree}()},
\code{\link{RootTree}()},
\code{\link{SortTree}()},
\code{\link{Subtree}()},
\code{\link{TipTimedTree}()},
\code{\link{TrivialTree}}

Other tree properties: 
\code{\link{Cherries}()},
\code{\link{EdgeRatio}()},
\code{\link{LongBranch}()},
\code{\link{MatchEdges}()},
\code{\link{NSplits}()},
\code{\link{NTip}()},
\code{\link{NodeNumbers}()},
\code{\link{PathLengths}()},
\code{\link{SplitsInBinaryTree}()},
\code{\link{TipLabels}()},
\code{\link{TreeIsRooted}()},
\code{\link{Treeness}()}

Other consensus tree functions: 
\code{\link{Consensus}()},
\code{\link{RoguePlot}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{consensus tree functions}
\concept{tree manipulation}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PhyToString.R
\name{StringToPhyDat}
\alias{StringToPhyDat}
\alias{StringToPhydat}
\alias{PhyToString}
\alias{PhyDatToString}
\alias{PhydatToString}
\title{Convert between strings and \code{phyDat} objects}
\usage{
StringToPhyDat(string, tips, byTaxon = TRUE)

StringToPhydat(string, tips, byTaxon = TRUE)

PhyToString(
  phy,
  parentheses = "{",
  collapse = "",
  ps = "",
  useIndex = TRUE,
  byTaxon = TRUE,
  concatenate = TRUE
)

PhyDatToString(
  phy,
  parentheses = "{",
  collapse = "",
  ps = "",
  useIndex = TRUE,
  byTaxon = TRUE,
  concatenate = TRUE
)

PhydatToString(
  phy,
  parentheses = "{",
  collapse = "",
  ps = "",
  useIndex = TRUE,
  byTaxon = TRUE,
  concatenate = TRUE
)
}
\arguments{
\item{string}{String of tokens, optionally containing whitespace, with no
terminating semi-colon.}

\item{tips}{(Optional) Character vector corresponding to the names (in order)
of each taxon in the matrix, or an object such as a tree from which
tip labels can be extracted.}

\item{byTaxon}{Logical. If \code{TRUE}, write one taxon followed by the next.
If \code{FALSE}, write one character followed by the next.}

\item{phy}{An object of class \code{phyDat}.}

\item{parentheses}{Character specifying format of parentheses with which to
surround ambiguous tokens.  Choose from: \code{\{} (default), \code{[}, \code{(}, \code{<}.}

\item{collapse}{Character specifying text, perhaps \verb{,}, with which to
separate multiple tokens within parentheses.}

\item{ps}{Character specifying text, perhaps \verb{;}, to append to the end of
each string.}

\item{useIndex}{Logical (default: \code{TRUE}) specifying whether to print
duplicate characters multiple times, as they appeared in the original matrix.}

\item{concatenate}{Logical specifying whether to concatenate all
characters/taxa into a single string, or to return a separate string for
each entry.}
}
\value{
\code{StringToPhyDat()} returns an object of class \code{phyDat}.

\code{PhyToString()} returns a character vector listing a text
representation of the phylogenetic character state for each taxon in turn.
}
\description{
\code{PhyDatToString()} converts a \code{phyDat} object as a string;
\code{StringToPhyDat()} converts a string of character data to a \code{phyDat} object.
}
\examples{
StringToPhyDat("-?01231230?-", c("Lion", "Gazelle"), byTaxon = TRUE)
# encodes the following matrix:
# Lion     -?0123
# Gazelle  1230?-

fileName <- paste0(system.file(package = "TreeTools"),
                   "/extdata/input/dataset.nex")
phyDat <- ReadAsPhyDat(fileName)
PhyToString(phyDat, concatenate = FALSE)

}
\seealso{
Other phylogenetic matrix conversion functions: 
\code{\link{Decompose}()},
\code{\link{MatrixToPhyDat}()},
\code{\link{Reweight}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{phylogenetic matrix conversion functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{MatrixToPhyDat}
\alias{MatrixToPhyDat}
\alias{PhyDatToMatrix}
\title{Convert between matrices and \code{phyDat} objects}
\usage{
MatrixToPhyDat(tokens, tipLabels = rownames(tokens))

PhyDatToMatrix(
  dataset,
  ambigNA = FALSE,
  inappNA = ambigNA,
  parentheses = c("{", "}"),
  sep = ""
)
}
\arguments{
\item{tokens}{Matrix of tokens, possibly created with \code{\link[=ReadCharacters]{ReadCharacters()}}
or \code{\link[=ReadTntCharacters]{ReadTntCharacters()}}.
Row names should correspond to leaf labels; column names may optionally
correspond to character labels.}

\item{tipLabels}{Optionally, an object providing labels for leaves, via
\code{\link[=TipLabels]{TipLabels()}}; will override \code{rownames(tokens)}.}

\item{dataset}{A dataset of class \code{phyDat}.}

\item{ambigNA, inappNA}{Logical specifying whether to denote ambiguous /
inapplicable characters as \code{NA} values.}

\item{parentheses}{Character vector specifying style of parentheses
with which to enclose ambiguous characters. \code{c("[", "]")} or \code{"[]"} will
render \verb{[01]}.
\code{NULL} will use the token specified in the \code{phyDat} object; but beware that
this will be treated as a distinct (non-ambiguous) token if re-encoding with
\code{PhyDatToMatrix()}.}

\item{sep}{Character with which to separate ambiguous tokens, e.g. \code{','}
will render \verb{[0,1]}.}
}
\value{
\code{MatrixToPhyDat()} returns an object of class \code{phyDat}.

\code{PhyDatToMatrix()} returns a matrix corresponding to the
uncompressed character states within a \code{phyDat} object.
}
\description{
\code{MatrixToPhyDat()} converts a matrix of tokens to a \code{phyDat} object;
\code{PhyDatToMatrix()} converts a \code{phyDat} object to a matrix of tokens.
}
\examples{
tokens <- matrix(c(0, 0, "0", 0, 0,
                   0, 0, "1", 0, 1,
                   0, 0, "1", 0, 1,
                   0, 0, "2", 0, 1,
                   1, 1, "-", 1, 0,
                   1, 1, "2", 1, "{01}"),
                   nrow = 6, ncol = 5, byrow = TRUE,
                   dimnames = list(
                     paste0("Taxon_", LETTERS[1:6]),
                     paste0("Char_", 1:5)))
                   
MatrixToPhyDat(tokens)
data("Lobo", package = "TreeTools")
head(PhyDatToMatrix(Lobo.phy)[, 91:93])
}
\seealso{
Other phylogenetic matrix conversion functions: 
\code{\link{Decompose}()},
\code{\link{Reweight}()},
\code{\link{StringToPhyDat}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{phylogenetic matrix conversion functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse_files.R
\name{NexusTokensToInteger}
\alias{NexusTokensToInteger}
\title{Convert Nexus token matrix to integer}
\usage{
NexusTokensToInteger(tokens, polymorphism = c("?", "first", "last"))
}
\arguments{
\item{tokens}{Character matrix as returned by \code{\link[=ReadCharacters]{ReadCharacters()}}, a
character vector as returned by \code{\link[=NexusTokens]{NexusTokens()}}, or a \code{phyDat} object.}

\item{polymorphism}{Character string specifying how to handle polymorphic
tokens such as \code{"(01)"} or \code{"{12}"}:
\describe{
\item{\code{"?"} (default)}{Treat as the NEXUS missing-data token: map to
\code{NA_integer_}.}
\item{\code{"first"}}{Use the first state digit inside the brackets.}
\item{\code{"last"}}{Use the last state digit inside the brackets.}
}
Tokens \code{"?"} and \code{"-"} always map to \code{NA_integer_} regardless of
\code{polymorphism}.}
}
\value{
An integer matrix (or vector) with the same dimensions and
\code{dimnames} as \code{tokens}.
}
\description{
\code{NexusTokensToInteger()} converts the character matrix returned by
\code{\link[=ReadCharacters]{ReadCharacters()}} to an integer matrix, mapping polymorphic,
ambiguous (\verb{?}), and inapplicable (\code{-}) tokens to \code{NA_integer_} or to the
first/last state listed in the polymorphism, depending on \code{polymorphism}.
}
\details{
Only digit states \code{0}..\code{9} are recognised; non-digit symbols (and any
token whose interior contains no digits) become \code{NA_integer_}.
Polymorphism extraction (\code{polymorphism = "first"}/\code{"last"}) likewise
considers digits only.

If \code{tokens} is a \code{phyDat} object it is first converted via
\code{\link[=PhyDatToMatrix]{PhyDatToMatrix()}} with \verb{ambigNA = TRUE, inappNA = TRUE}, so that
fully-ambiguous and inapplicable rows become \code{NA_integer_} and only
true partial polymorphisms are subject to the \code{polymorphism} rule.
}
\examples{
tokens <- matrix(c("0", "(12)", "1", "?", "-"),
                 nrow = 1,
                 dimnames = list("Taxon_A", paste0("C", 1:5)))
NexusTokensToInteger(tokens)
NexusTokensToInteger(tokens, polymorphism = "first")

}
\seealso{
Other phylogenetic matrix conversion functions:
\code{\link[=Decompose]{Decompose()}},
\code{\link[=MatrixToPhyDat]{MatrixToPhyDat()}},
\code{\link[=Reweight]{Reweight()}},
\code{\link[=StringToPhyDat]{StringToPhyDat()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{phylogenetic matrix conversion functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Reweight.R
\name{Reweight}
\alias{Reweight}
\title{Reweight phylogenetic characters}
\usage{
Reweight(dataset, weights)
}
\arguments{
\item{dataset}{A phylogenetic data matrix of \pkg{phangorn} class \code{phyDat},
or as a matrix in the format produced by \code{\link[=PhyDatToMatrix]{PhyDatToMatrix()}}.}

\item{weights}{Unnamed integer vector specifying desired weight of each
character in turn; or named integer vector specifying weights of each
character; unnamed entries will be assigned weight 1.}
}
\value{
\code{Reweight()} returns \code{dataset} after adjusting the weights of
the specified characters.
For a matrix, this is attained by repeating each column the \code{weights} times.
For a \code{phyDat} object, the "weight" attribute will be modified.
}
\description{
\code{Reweight()} allows the weights of specific characters in phylogenetic
datasets to be arbitrarily adjusted.
}
\details{
This functionality should be employed with care.
The underlying principle of parsimony is that all evolutionary steps are
equivalent.
Setting different weights to different characters is at odds with that
principle, so analysis of a re-weighted matrix using a parsimony-based
framework is arguably no longer parsimony analysis; on the most permissive
view, the criteria used to determine a weighting scheme will always
be arbitrary.

It can be useful to relax the criterion that all evolutionary steps are
equivalent -- for example, implied weighting
\insertCite{Goloboff1997}{TreeTools} typically recovers better trees than
equal-weights parsimony \insertCite{Smith2019}{TreeTools}.
This said, assigning different weights to different characters tacitly
imposes a model of evolution that differs from that implicit in equal-weights
parsimony.  Whereas probabilistic models can be evaluated by various methods
(e.g. fit, marginal likelihood, posterior predictive power), there are no
principled methods of comparing different models under a parsimony framework.

As such, \code{Reweight()} is likely to be useful for a narrow set of uses.
Examples may include:
\itemize{
\item informal robustness testing, to explore whether certain characters are
more or less influential on the resulting tree;
\item Imposing constraints on a dataset, by adding each constraint as a column
in a dataset whose weight exceeds the total amount of data.
}
}
\examples{
mat <- rbind(a = c(0, 2, 0), b = c(0, 2, 0), c = c(1, 3, 0), d = c(1, 3, 0))
dat <- MatrixToPhyDat(mat)

 # Set character 1 to weight 1, character 2 to weight 2; omit character 3
Reweight(mat, c(1, 2, 0))
# Equivalently:
Reweight(dat, c("3" = 0, "2" = 2))
}
\references{
\insertAllCited{}
}
\seealso{
Other phylogenetic matrix conversion functions: 
\code{\link{Decompose}()},
\code{\link{MatrixToPhyDat}()},
\code{\link{StringToPhyDat}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{phylogenetic matrix conversion functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_write.R
\name{WriteTntCharacters}
\alias{WriteTntCharacters}
\alias{WriteTntCharacters.phyDat}
\alias{WriteTntCharacters.matrix}
\title{Write morphological character matrix to TNT file}
\usage{
WriteTntCharacters(
  dataset,
  filepath = NULL,
  comment = "Dataset written by `TreeTools::WriteTntCharacters()`",
  types = NULL,
  pre = "",
  post = ""
)

\method{WriteTntCharacters}{phyDat}(
  dataset,
  filepath = NULL,
  comment = "Dataset written by `TreeTools::WriteTntCharacters()`",
  types = NULL,
  pre = "",
  post = ""
)

\method{WriteTntCharacters}{matrix}(
  dataset,
  filepath = NULL,
  comment = "Dataset written by `TreeTools::WriteTntCharacters()`",
  types = NULL,
  pre = "",
  post = ""
)
}
\arguments{
\item{dataset}{Morphological dataset of class \code{phyDat} or \code{matrix}.}

\item{filepath}{Path to file; if \code{NULL}, returns a character vector.}

\item{comment}{Optional comment with which to entitle matrix.}

\item{types}{Optional list specifying where different data types begin.
\code{c(num = 1, dna = 10)} sets characters 1..9 as numeric, 10..end as DNA.}

\item{pre, post}{Character vector listing text to print before and after the
character matrix.  Specify \verb{pre = 'piwe=;} if the matrix is to be analysed
using extended implied weighting (\verb{xpiwe=}).}
}
\description{
Write morphological character matrix to TNT file
}
\examples{
data("Lobo", package = "TreeTools")

WriteTntCharacters(Lobo.phy)

# Read with extended implied weighting
WriteTntCharacters(Lobo.phy, pre = "piwe=10;", post = "xpiwe=;")

# Write to a file with:
# WriteTntCharacters(Lobo.phy, "example_file.tnt")
}
\seealso{
\code{\link[=ReadTntCharacters]{ReadTntCharacters()}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_information.R
\name{TreesMatchingTree}
\alias{TreesMatchingTree}
\alias{LnTreesMatchingTree}
\alias{Log2TreesMatchingTree}
\title{Number of trees containing a tree}
\usage{
TreesMatchingTree(tree)

LnTreesMatchingTree(tree)

Log2TreesMatchingTree(tree)
}
\arguments{
\item{tree}{A tree of class \code{\link[ape:read.tree]{phylo}}.}
}
\value{
\code{TreesMatchingTree()} returns a numeric specifying the number of
unrooted binary trees that contain all the edges present in the input tree.

\code{LnTreesMatchingTree()} gives the natural logarithm of this number.
}
\description{
\code{TreesMatchingTree()} calculates the number of unrooted binary trees that
are consistent with a tree topology on the same leaves.
}
\details{
Remember to unroot a tree first if the position of its root is arbitrary.
}
\examples{
partiallyResolvedTree <- CollapseNode(BalancedTree(8), 12:15)
TreesMatchingTree(partiallyResolvedTree)
LnTreesMatchingTree(partiallyResolvedTree)

# Number of rooted trees:
rootedTree <- AddTip(partiallyResolvedTree, where = 0)
TreesMatchingTree(partiallyResolvedTree)
}
\seealso{
Other tree information functions: 
\code{\link{CladisticInfo}()},
\code{\link{NRooted}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree information functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/TreeNumber.R
\name{is.TreeNumber}
\alias{is.TreeNumber}
\title{Is an object a \code{TreeNumber} object?}
\usage{
is.TreeNumber(x)
}
\arguments{
\item{x}{R object.}
}
\value{
\code{is.TreeNumber()} returns a logical vector of length one specifying
whether \code{x} inherits the class \code{"TreeNumber"}.
}
\description{
Is an object a \code{TreeNumber} object?
}
\examples{
is.TreeNumber(FALSE) # FALSE 
is.TreeNumber(as.TreeNumber(BalancedTree(5))) # TRUE
}
\seealso{
Other 'TreeNumber' utilities: 
\code{\link{TreeNumber}},
\code{\link{print.TreeNumber}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{'TreeNumber' utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PathLengths.R
\name{PathLengths}
\alias{PathLengths}
\title{Calculate length of paths between each pair of vertices within tree}
\usage{
PathLengths(tree, fullMatrix = FALSE, use.na = TRUE)
}
\arguments{
\item{tree}{Original tree of class \code{phylo}, in \code{\link{Preorder}}.}

\item{fullMatrix}{Logical specifying return format; see "value" section`.}

\item{use.na}{Logical specifying whether to set non-existent paths to \code{NA},
or to leave uninitialized.  Set to \code{FALSE} to maximize performance.}
}
\value{
If \code{fullMatrix = TRUE}, \code{PathLengths()} returns a square matrix in
which entry \verb{[i, j]} denotes the distance from internal node \code{i} to the
descendant vertex \code{j}.
Vertex pairs without a continuous directed path are denoted \code{NA} if \code{use.na}
is \code{TRUE}.
If \code{fullMatrix = FALSE}, \code{PathLengths()} returns a \code{data.frame} with three
columns: \code{start} lists the deepest node in each path (i.e. that closest
to the root); \code{end} lists the shallowest node (i.e. that closest to a leaf);
\code{length} lists the total length of that path.
}
\description{
Given a weighted rooted tree \code{tree}, \code{PathLengths()} returns the distance
from each vertex to each of its descendant vertices.
}
\examples{
tree <- rtree(6)
plot(tree)
add.scale.bar()
nodelabels()
tiplabels()
PathLengths(tree)
}
\seealso{
Other tree properties: 
\code{\link{Cherries}()},
\code{\link{ConsensusWithout}()},
\code{\link{EdgeRatio}()},
\code{\link{LongBranch}()},
\code{\link{MatchEdges}()},
\code{\link{NSplits}()},
\code{\link{NTip}()},
\code{\link{NodeNumbers}()},
\code{\link{SplitsInBinaryTree}()},
\code{\link{TipLabels}()},
\code{\link{TreeIsRooted}()},
\code{\link{Treeness}()}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree properties}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{brewer}
\alias{brewer}
\title{Brewer palettes}
\format{
An object of class \code{list} of length 12.
}
\source{
{
\itemize{
\item \href{https://colorbrewer2.org/#type=diverging&scheme=RdYlBu&n=3}{ColourBrewer2.org}
\item \href{https://mk.bcgsc.ca/colorblind/}{Martin Krzywinski}
}
}
}
\usage{
brewer
}
\description{
A list of eleven Brewer palettes containing one to eleven colours that
are readily distinguished by colourblind viewers, followed by a twelfth
12-colour palette adapted for colour blindness.
}
\examples{
data("brewer", package = "TreeTools")
plot(0, type = "n", xlim = c(1, 12), ylim = c(12, 1),
     xlab = "Colour", ylab="Palette")
for (i in seq_along(brewer)) text(seq_len(i), i, col = brewer[[i]])

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{logDoubleFactorials}
\alias{logDoubleFactorials}
\title{Natural logarithms of double factorials}
\format{
An object of class \code{numeric} of length 50000.
}
\usage{
logDoubleFactorials
}
\description{
\code{logDoubleFactorials} is a numeric vector with pre-calculated values of
double factorials up to 50 000!!.
}
\seealso{
Other double factorials:
\code{\link[=DoubleFactorial]{DoubleFactorial()}},
\code{\link{doubleFactorials}}
}
\concept{double factorials}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_generation.R
\name{ConstrainedNJ}
\alias{ConstrainedNJ}
\title{Constrained neighbour-joining tree}
\usage{
ConstrainedNJ(dataset, constraint, weight = 1L, ratio = TRUE, ambig = "mean")
}
\arguments{
\item{dataset}{A phylogenetic data matrix of \pkg{phangorn} class \code{phyDat},
whose names correspond to the labels of any accompanying tree.}

\item{constraint}{Either an object of class \code{phyDat}, in which case returned
trees will be perfectly compatible with each character in \code{constraint};
or a tree of class \code{phylo}, in which each node in \code{constraint} will occur in
the returned tree.
See \href{https://ms609.github.io/TreeSearch/articles/tree-search.html}{vignette}
for further examples.}

\item{weight}{Numeric specifying degree to up-weight characters in
\code{constraint}.}

\item{ambig, ratio}{Settings of \code{ambig} and \code{ratio} to be used when
computing \code{\link[=Hamming]{Hamming()}} distances between sequences.}
}
\value{
\code{ConstrainedNJ()} returns a tree of class \code{phylo}.
}
\description{
Constructs an approximation to a neighbour-joining tree, modified in order
to be consistent with a constraint.  Zero-length branches are collapsed
at random.
}
\examples{
dataset <- MatrixToPhyDat(matrix(
  c(0, 1, 1, 1, 0, 1,
    0, 1, 1, 0, 0, 1), ncol = 2,
  dimnames = list(letters[1:6], NULL)))
constraint <- MatrixToPhyDat(
  c(a = 0, b = 0, c = 0, d = 0, e = 1, f = 1))
plot(ConstrainedNJ(dataset, constraint))
}
\seealso{
Other tree generation functions:
\code{\link{GenerateTree}},
\code{\link[=NJTree]{NJTree()}},
\code{\link{TreeNumber}},
\code{\link{TrivialTree}}
}
\author{
\href{https://orcid.org/0000-0001-5660-1727}{Martin R. Smith}
(\href{mailto:martin.smith@durham.ac.uk}{martin.smith@durham.ac.uk})
}
\concept{tree generation functions}
